//------------------------------------------------------------------------------
//
// Client classes
//
//------------------------------------------------------------------------------

unit ACRCommunication;

interface

{$I ACRVer.inc}

uses SysUtils, Classes,

// Accuracer units

{$IFDEF DEBUG_LOG}
     ACRDebug,
{$ENDIF}
{$IFNDEF D5H}
     ACRD4Routines,
{$ENDIF}
     ACRCompression,
     ACRConnection,
     ACRBase,
     ACRMemory,
     ACRExcept,
     ACRTypes,
     ACRConst;


type

//------------------------------------------------------------------------------
//
//  Communication Command Format:
//
//  CommandHeader | [CommandData]
//
//------------------------------------------------------------------------------

  TACRCommunicationCommandType = (acctRequest,acctReply);

  TACRCommunicationRequest = (
//0
     accrqNULL // not a request

    // Session requests to the RemoteDatabase
// 1
    ,accrqConnectDatabase
    ,accrqDisconnectDatabase   // not used since v 4.02
    ,accrqIsDatabaseExists
    ,accrqGetTablesList
    ,accrqIsTableExists
    ,accrqGetFormatVersion
    ,accrqGetTotalPageCount
    ,accrqGetFreePageCount
    ,accrqIsDatabaseEncrypted
    ,accrqIsDatabaseEncryptedByPassword
    ,accrqIsCryptoParamsValid
    ,accrqIsInTransaction
    ,accrqStartTransaction
    ,accrqCommit
// 15
    ,accrqRollback


    // Cursor requests
// 16
    ,accrqCreateTable
    ,accrqDeleteTable
    ,accrqEmptyTable
    ,accrqRenameTable
    ,accrqRenameField
// 21
    ,accrqOpenTable
    ,accrqCloseTable
    ,accrqApplyDefaultValues
    ,accrqAddIndex
    ,accrqDeleteIndex
    ,accrqDeleteAllIndexes
    ,accrqGetRecordCount
// 28
    ,accrqGetRecordBuffer
    ,accrqSetRecNo
    ,accrqGetRecNo 
    ,accrqInternalEdit 
    ,accrqInternalCancel
// 33
    ,accrqInternalPost 
    ,accrqInternalDelete 
    ,accrqActivateFilters
    ,accrqDeactivateFilters
    ,accrqLocate 
    ,accrqFindKey
    ,accrqResetRange 
    ,accrqApplyRange
// 41 
    ,accrqReadBLOBValue
 
// 42
    // SQLProcessor requests 
    ,accrqExecSQL 
	
// 43 
    ,accrqIsRecordExists 
 
  ); 
  TACRCommunicationReply = ( 
     accrplNULL // not a reply
    ,accrplYes
    ,accrplNo
// 3
    ,accrplOperationSucceed
// 4 
    ,accrplOperationFailed
  ); 
	
  TACRCommunicationCommandHeader = packed record 
    Request:          TACRCommunicationRequest; 
    Reply:            TACRCommunicationReply;
    Reserved:         array [0..1] of Byte; 
    Reserved1:        array [0..1] of Integer;
  end; 
  PACRCommunicationCommandHeader = ^TACRCommunicationCommandHeader; 
	
	
//////////////////////////////////////////////////////////////////////////////// 
// 
// TACRNetworkSession 
// 
//////////////////////////////////////////////////////////////////////////////// 
 
 
 TACRNetworkSession = class(TACRBaseSession)
  protected
   FSentCommandHeader:          TACRCommunicationCommandHeader; 
   FSentCommandDataStream:      TACRMemoryStream;
   FReceivedCommandHeader:      TACRCommunicationCommandHeader;
   FReceivedCommandDataStream:  TACRMemoryStream; 
   FCommandReceived:            Boolean;
  protected
   FConnected:                  Boolean; 
  public 
   constructor Create; 
   destructor Destroy; override;
   procedure DoOnError(ErrorCode: Integer; NativeError: Integer = -1; ErrorMessage: String = ''); virtual; abstract;
   procedure DoCloseSessionOnNetworkError; virtual; abstract; 
   // receive data from network and move it to ReceivedCommandHeader and ReceivedCommandDataStream
   procedure ReceiveData(Buffer: PChar; BufferSize: Integer); override;
   // send command specified by SentCommandHeader [ optionally SentCommandDataStream ] 
   function SendCommand: Boolean; virtual; 
   // send buffer via established connection using connection manager
   procedure SendBuffer(var Buffer: PChar; BufferSize: Integer); virtual; abstract;
   // send custom message
   procedure SendMessage(Buffer: PChar; BufferSize: Integer); virtual; abstract; 
   procedure ReceiveMessage(Buffer: PChar; BufferSize: Integer); virtual; abstract; 
  public 
   property SentCommandHeader: TACRCommunicationCommandHeader 
              read FSentCommandHeader write FSentCommandHeader;
   property SentCommandDataStream: TACRMemoryStream 
              read FSentCommandDataStream; 
   property ReceivedCommandHeader: TACRCommunicationCommandHeader 
              read FReceivedCommandHeader;
   property ReceivedCommandDataStream: TACRMemoryStream
              read FReceivedCommandDataStream;
 end; // TACRNetworkSession 
	
implementation
	
 
//////////////////////////////////////////////////////////////////////////////// 
//
// TACRNetworkSession
//
////////////////////////////////////////////////////////////////////////////////
	
	
//------------------------------------------------------------------------------ 
// Create
//------------------------------------------------------------------------------
constructor TACRNetworkSession.Create; 
begin
  inherited;
  FCommandReceived := False; 
  FSentCommandDataStream := TACRMemoryStream.Create; 
  FReceivedCommandDataStream := TACRMemoryStream.Create; 
  FConnected := False; 
end; // Create
 
 
//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRNetworkSession.Destroy;
begin 
  FSentCommandDataStream.Free; 
  FReceivedCommandDataStream.Free;
  inherited;
end; // Destroy 
 
 
//------------------------------------------------------------------------------
// receive data from network and move it to ReceivedCommandHeader and ReceivedCommandDataStream
//------------------------------------------------------------------------------
procedure TACRNetworkSession.ReceiveData(Buffer: PChar; BufferSize: Integer);
var Size: Integer;
begin 
//aaWriteToLog('start TACRNetworkSession.ReceiveData, Buffer = '+IntToStr(Integer(Buffer))+
//', length = '+ 
//IntToStr(MemoryManager.GetMemoryBufferSize(Buffer)));
  // ignore invalid buffers
  if (BufferSize >= SizeOf(TACRCommunicationCommandHeader)) then
   begin
    FCommandReceived := True;
    Size := BufferSize - SizeOf(TACRCommunicationCommandHeader);
    FReceivedCommandDataStream.Size := Size;
    Move (Buffer^, FReceivedCommandHeader, SizeOf(TACRCommunicationCommandHeader));
    if (Size > 0) then
      Move(PChar(Buffer + SizeOf(TACRCommunicationCommandHeader))^,
           FReceivedCommandDataStream.Buffer^,Size);
    FReceivedCommandDataStream.Position := 0;
   end
  else
   raise EACRException.Create(11511,ErrorLInvalidCommandBufferReceived,[BufferSize,SizeOf(TACRCommunicationCommandHeader)]);
end; // ReceiveData


//------------------------------------------------------------------------------
// send command specified by SentCommandHeader [ optionally SentCommandDataStream ]
//------------------------------------------------------------------------------
function TACRNetworkSession.SendCommand: Boolean;
var Buffer:     PChar;
    BufferSize: Integer;
    Size:       Integer;

 procedure Finalize;
 begin
   if (Buffer <> nil) and (BufferSize > 0) then
    MemoryManager.FreeAndNilMem(Buffer);
   FSentCommandDataStream.Size := 0;
 end;

begin
  Result := False;
  BufferSize := FSentCommandDataStream.Size + SizeOf(TACRCommunicationCommandHeader);
  Buffer := MemoryManager.GetMem(BufferSize);
  Move(FSentCommandHeader,Buffer^,SizeOf(TACRCommunicationCommandHeader));
  Size := FSentCommandDataStream.Size;
  if (Size > 0) then
    Move(FSentCommandDataStream.Buffer^,
         PChar(Buffer + SizeOf(TACRCommunicationCommandHeader))^,
         Size);
  FSentCommandDataStream.Position := 0;
  try
    SendBuffer(Buffer,BufferSize);
    Finalize;
    Result := True;
  except
    on e: EACRException do
     begin
      DoOnError(ACR_CS_ErrorSendCommandFailed,e.NativeError,
                 Format(ErrorL_CS_ErrorSendCommandFailed,
                       [Integer(FSentCommandHeader.Request),Integer(FSentCommandHeader.Reply),FSessionID,e.Message]));
      Finalize;
      DoCloseSessionOnNetworkError;
     end;
    on e: Exception do
     begin
      DoOnError(ACR_CS_ErrorSendCommandFailed,-1,
                 Format(ErrorL_CS_ErrorSendCommandFailed,
                       [Integer(FSentCommandHeader.Request),Integer(FSentCommandHeader.Reply),FSessionID,e.Message]));
      Finalize;
      DoCloseSessionOnNetworkError;
     end
    else
     begin
      DoOnError(ACR_CS_ErrorSendCommandFailed,-1,
                 Format(ErrorL_CS_ErrorSendCommandFailed,
                       [Integer(FSentCommandHeader.Request),Integer(FSentCommandHeader.Reply),FSessionID,'']));
      Finalize;
      DoCloseSessionOnNetworkError;
     end;
  end;
end; // SendCommand


initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRCommunication> initialized');
{$ENDIF}

end.
