unit ACRExpressions;

interface

{$I ACRVer.inc}

uses Classes, SysUtils, DB,
{$IFDEF D6H}
     Variants,
{$ELSE}
     ACRD4Routines,
{$ENDIF}

     {$IFDEF DEBUG_LOG}
     ACRDebug,
     {$ENDIF}

     ACRBase,
     ACRVariant,
     ACRLexer,
     ACRExcept,
     ACRConst,
     ACRTypes,
     ACRConverts,
     ACRRelationalAlgebra,
     ACRMemory,
     ACRDateFormat;


type

 TACRExprNodeType = (entField, entConst, entOperator, entSet, entNull);
 TSign = (sgnUndefined, sgnPlus, sgnMinus);

 // unary / binary Value operators
 TACRDataOperator = (
    doUNDEFINED,                       // error
    //George doCMP,                             // compare <0,=0,>0
    {Comparison}
    doEQ,                              // equal
    doNE,                              // NOT equal
    doGT,                              // greater than
    doLT,                              // less than
    doGE,                              // greater or equal
    doLE,                              // less or equal
    {Boolean}
    doNOT,                             // NOT
    doAND,                             // AND
    doOR,                              // OR
    doIN,                              // in (a,b,c, ...)
    doNOTIN,                           // not in (a,b,c, ...)
    doBETWEEN,                         // between (a,b)
    doNOTBETWEEN,                      // not between (a,b)
    doLIKE,                            // like 'a?b%'
    doNOTLIKE,                         // not like
    doISNULL,                          // is null
    doISNOTNULL,                       // is not null
    doTRUE,                            // TRUE const
    doFALSE,                           // FALSE const
    {Arithmetic}
    doADD,                             // addition
    doSUB,                             // subtraction
    doMUL,                             // multiplication
    doDIV,                             // division
    {Functions}
    //doPOSITION,                        // Position (str1) in (str2)
    doCONCAT,                          // str1 || str2
    doUPPER,                           // Upper(str)
    doLOWER,                           // Lower(str)
    doTRIM,                            // TRIM
    doLTRIM,                           // LTRIM
    doRTRIM,                           // RTRIM
    doLENGTH,                          // LENGTH(str)
    doPOS,                             // POS(substr in|, str) { 1 - first char, 0 - not found }
    doSUBSTRING,                       // SUBSTRING(str from|, startindex [for|, length])
    {aggregated functions}
    doSUM,                             // SUM([distinct]expression)
    doAVG,                             // AVG([distinct]expression)
    doMIN,                             // min(expression)
    doMAX,                             // max(expression)
    doCOUNT,                           // COUNT([distinct]expression)
    doCOUNTALL,                        // count(*)
    {datetime functions}
    doSYSDATE,                         // SYSDATE - return current DateTime
    doCURRENT_DATE,                    // CURRENT_DATE
    doCURRENT_TIME,                    // CURRENT_TIME
    doTODATE,                          // TODATE(string, format)
    doTOSTRING,                        // TOSTRING(date, format)
    {cast types}
    doCAST,                            // Cast(expression, type)
    {system}
    doLASTAUTOINC,
    doTOBLOB                           // TOBLOB(string, format)
 );

  TACRBLOBValueFormat = (bvfMIME64,bvfHEX);
  // forward declarations
  TACRExprNode = class;
  TACRExprNodeComparison = class;

////////////////////////////////////////////////////////////////////////////////
//
// TACRExpression
//
////////////////////////////////////////////////////////////////////////////////


  TACRExpression = class (TObject)
   private
    LLex: TACRLexer;   // lexer with expression to parse
    Token: TToken;     // current token
    LCursor: TACRCursor;
    FRootExprNode: TACRExprNode;
    FCaseInsensitive: Boolean;
    FPartialKey: Boolean;
    F3ValueLogic: Boolean;
    FInMemory:        Boolean;
    FDatabaseName:    String;
    FSessionName:     String;
    FParams:          TParams;
   private
    // gets current token
    //function GetCurrentToken: Boolean;
    // gets next token
    function GetNextToken: Boolean; overload;
    // gets next token
    //function GetNextToken(NativeErrorCode: integer): Boolean; overload;
    // gets token and lokks at next token with check for token type restrictions
    function GetNextToken(PermittedTypes: TTokenTypes;
                          RaiseExceptions: Boolean = False): Boolean; overload;

   private
    // saves internal state (to restore in case of not successful forward parsing)
    procedure SaveState(var SavedTokenNo: integer);// var SavedRootNode: TACRExprNode);
    // restores internal state (in case of not successful forward parsing)
    procedure RestoreState(SavedTokenNo: integer);// SavedRootNode: TACRExprNode);

    // parses <,>,=,<>,>=,<=
    function ParseCompOp: TACRDataOperator;
    // parses <row value constructor element>
    function ParseRowValueConstructorElement: TACRExprNode;
    // parses <row value constructor element> | <row subquery>
    function ParseRowValueConstructor: TACRExprNode;
    // parses <row value constructor> <comp op> <row value constructor>
    function ParseComparisonPredicate: TACRExprNode;
    // parses <match value> [ NOT ] LIKE <pattern> [ ESCAPE <escape character> ]
    function ParseLikePredicate: TACRExprNode;
    // parses <row value constructor> IS [ NOT ] NULL
    function ParseNullPredicate: TACRExprNode;
    // parses <between predicate>
    function ParseBetweenPredicate: TACRExprNode;
    // parses <in predicate>
    function ParseInPredicate: TACRExprNode;
    // parses <Exists predicate>
    function ParseExistsPredicate: TACRExprNode;
    // parses <true/false>
    function ParseTrueFalseConst: TACRExprNode;
    // parses <NULL> const
    function ParseNullConst: TACRExprNode;
    // parses <comparison predicate> | <between predicate> | <in predicate>  |
    // <like predicate>  | <null predicate> | <quantified comparison predicate> |
    // <exists predicate> | <unique predicate> | <match predicate> | <overlaps predicate>
    function ParsePredicate: TACRExprNode;
    // parses <predicate>  | <left paren> <search condition> <right paren>
    function ParseBooleanPrimary: TACRExprNode;
    // parses <boolean primary> [ IS [ NOT ] <truth value> ]
    function ParseBooleanTest: TACRExprNode;
    // parses [ NOT ] <boolean test>
    function ParseBooleanFactor: TACRExprNode;
    // parses <boolean factor> | <boolean term> AND <boolean factor>
    function ParseBooleanTerm: TACRExprNode;
    // parses <boolean term> | <search condition> OR <boolean term>
    function ParseSearchCondition: TACRExprNode;

    // parses <value expression>
    function ParseValueExpression: TACRExprNode;
    // parses <numeric value expression>
    function ParseNumericValueExpression: TACRExprNode;
    // parses <term>
    function ParseTerm: TACRExprNode;
    // parses <factor>
    function ParseFactor: TACRExprNode;
    // parses <sign> ('+' or '-')
    function ParseSign: TSign;
    // parses <numeric primary>
    function ParseNumericPrimary: TACRExprNode;
    // parses <value expression primary>
    function ParseValueExpressionPrimary: TACRExprNode;
    // parses <unsigned value specification>
    function ParseUnsignedValueSpecification: TACRExprNode;
    // parses <unsigned literal>
    function ParseUnsignedLiteral: TACRExprNode;
    // parses <unsigned numeric literal>
    function ParseUnsignedNumericLiteral: TACRExprNode;
    // parses <column reference>
    function ParseColumnReference: TACRExprNode;
    // parses <SUB QUERY>
    function ParseSubQuery(bExists:       Boolean = False;
                           bIn:           Boolean = False;
                           bNot:          Boolean = False;
                           SourceNode:    TACRExprNode = nil
                           ): TACRExprNode;
    // parses <set function specification>
    function ParseSetFunctionSpecification: TACRExprNode;
    // parses <general set function>
    function ParseGeneralSetFunction: TACRExprNode;
    // parses <cast specification>
    function ParseCastSpecification: TACRExprNode;
    // parses <cast operand>
    function ParseCastOperand: TACRExprNode;

    // parses <numeric value function>
    function ParseNumericValueFunction: TACRExprNode;
    // parses <position expression>
    function ParsePositionExpression: TACRExprNode;
    // parses <length expression>
    function ParseLengthExpression: TACRExprNode;
    // parses <LastAutoInc expression>
    function ParseLastAutoIncExpression: TACRExprNode;

    // parses <string value expression>
    function ParseStringValueExpression: TACRExprNode;
    // parses <character value expression>
    function ParseCharacterValueExpression: TACRExprNode;
    // parses <concatenation>
    function ParseConcatination: TACRExprNode;
    // parses <character factor>
    function ParseCharacterFactor: TACRExprNode;
    // parses <character primary>
    function ParseCharacterPrimary: TACRExprNode;
    // parses <general literal>
    function ParseGeneralLiteral: TACRExprNode;
    // parses <character string literal>
    function ParseCharacterStringLiteral: TACRExprNode;
    // parses <string value function>
    function ParseStringValueFunction: TACRExprNode;
    // parses <character value function>
    function ParseCharacterValueFunction: TACRExprNode;
    // parses <BLOB value function>
    function ParseBLOBValueFunction: TACRExprNode;

    // parse <datetime value expression>
    function ParseDatetimeValueExpression: TACRExprNode;
    // parse <datetime sysdate function>
    function ParseSysdateFunction: TACRExprNode;
    // parse <datetime CURRENT_TIME function>
    function ParseCurrentTimeFunction: TACRExprNode;
    // parse <datetime CURRENT_DATE function>
    function ParseCurrentDateFunction: TACRExprNode;
    // parse <datetime TODATE function>
    function ParseToDateFunction: TACRExprNode;

   private
    procedure Clear;
   public
    // Constructors
    constructor Create; overload;
    // creates
    constructor Create(RootNode: TACRExprNode); overload;
    //constructor Create(Lexer: TACRLexer); overload;
    // Destructor
    destructor Destroy; override;
   public
    procedure PrepareForeignKeyCheck(
                                        ReferencedCursor: TACRCursor;
                                        Cursor:           TACRCursor;
                                        ConstraintDef:    TACRConstraintDefForeignKey;
                                        PrimaryIndex:     TACRIndexDef
                                     );
    procedure PrepareForeignKeyActionFilter(
                                        ReferencedCursor: TACRCursor;
                                        Cursor:           TACRCursor;
                                        ConstraintDef:    TACRConstraintDefForeignKeyAction;
                                        PrimaryIndex:     TACRIndexDef
                                     );
    // Parsing for Locate
    procedure ParseForLocate(
                              Cursor: TACRCursor;
                              FieldNames: String;
                              KeyValues: Variant;
                              CaseInsensitive: boolean = true;
                              PartialKey: boolean = false
                            );
    // Parsing for Filter
    procedure ParseForFilter(
                              Cursor: TACRCursor;
                              Filter: String;
                              CaseInsensitive: boolean = true;
                              PartialKey: boolean = false
                            );
    // Parse For Boolean Expression (Filter, Where-Clause)
    procedure ParseForBooleanExpression(
                                        //Cursor: TACRCursor;
                                        Lexer: TACRLexer
                                        //CaseInsensitive: boolean = true;
                                        //PartialKey: boolean = false
                                       );
    // Parse ValueExpression
    procedure ParseForValueExpression(
                                        //Cursor: TACRCursor;
                                        Lexer: TACRLexer
                                        //CaseInsensitive: boolean = true;
                                        //PartialKey: boolean = false
                                       );
    // Return Variant
    function GetValue(TrueFalseNullLogic: boolean = true): TACRVariant;
    // Get Boolean result
    function GetResult: Boolean;

    // is expression contains aggregated function
    function IsAggregated: Boolean;
    // Init for aggregated functions
    procedure Init;
    // Accumulate for group functions
    procedure Accumulate;
    // process assign AO
    procedure AssignAO(AO: TACRAO);
    // process assign Cursor and its RecordBuffer
    procedure AssignCursor(Cursor: TACRCursor);
    // process assign New Cursor Buffer
    procedure AssignCursorBuffer(Buffer: TACRRecordBuffer);
    // return Type of Data
    function GetDataType: TACRAdvancedFieldType;
    // return Size of Data (for strings and arrays)
    function GetDataSize: Integer;
    // return Size of Data (for strings and arrays)
    function GetPrecision: Integer;
    // is expression contains no nodes
    function IsEmpty: Boolean;
    // is expression a Field (for join)
    function IsField: Boolean;
    // is expression a constant
    function IsConst: Boolean;
    // Move AndNodes To RootNode
    procedure MoveAndNodesToRoot;
    // Field Name, Table Name
    procedure GetFieldInfo(var TableName: String; var FieldName: String);
    // makes filter string from related parts and sets it to AO
    procedure ApplyFilterParts(AO: TACRAO; bHaving: Boolean);
    // replace pseudonyms to original names (f1 -> table1.field1)
    procedure ReplacePseudonyms(SelectList: array of TACRSelectListItem);
    // makes join field lists
    function ExtractJoinConditions(
                                    AO1, AO2: TACRAO;
                                    FieldList1, FieldList2: TACRFields
                                  ): Integer;
    // extract conditions for index scan
    procedure TryExtractIndexScanConditionsFromNode(
                             Node:           TACRExprNodeComparison;
                             IndexDefs:      TACRIndexDefs;
                             ScanConditions: TACRScanSearchConditionArray;
                             ExtractedConditionsInfo: TList
                                        );
    // remove extracted nodes (they included in list, but list contains more)
    procedure RemoveExtractedNodes(ExtractedConditionsInfo: TList);
    // extract conditions for index scan
    procedure ExtractIndexScanConditions(
                             IndexDefs:               TACRIndexDefs;
                             ScanConditions:          TACRScanSearchConditionArray;
                             ExtractedConditionsInfo: TList
                                        );
    // return true if expression have incompatible conditions (like compare = for string fields with constant longer then field size)
    function IsIncompatible: Boolean;
    // add extracted node
    procedure AddNode(Node: TACRExprNode);
   public
    property InMemory: Boolean read FInMemory write FInMemory;
    property DatabaseName: String read FDatabaseName write FDatabaseName;
    property SessionName: String read FSessionName write FSessionName;
    property Params: TParams read FParams write FParams;
  end;




  TACRExprNode = class(TObject)
   private
    Children:         TList;                  // Children nodes
    Operator:         TACRDataOperator;      // '<',  '>', AND, NOT, ...
    Value:            TACRVariant;
    FCaseInsensitive: Boolean;
    FPartialKey:      Boolean;
   private
    // converts some WideStrings to Strings
    procedure PatchWideStrings; virtual;
   public
    // constructors
    constructor Create(
                       CaseInsensitive: boolean = true;
                       PartialKey: boolean = false
                      ); overload;
    constructor Create(
                       Op: TACRDataOperator;
                       CaseInsensitive: boolean = true;
                       PartialKey: boolean = false
                      ); overload;
    constructor Create(
                       Op: TACRDataOperator;
                       Node: TACRExprNode;
                       CaseInsensitive: boolean = true;
                       PartialKey: boolean = false
                      ); overload;
    constructor Create(
                       Op: TACRDataOperator;
                       Node1, Node2: TACRExprNode;
                       CaseInsensitive: boolean = true;
                       PartialKey: boolean = false
                       ); overload;
    constructor Create(
                       Op: TACRDataOperator;
                       Node1, Node2, Node3: TACRExprNode;
                       CaseInsensitive: boolean = true;
                       PartialKey: boolean = false
                       ); overload;
    // destroys
    destructor Destroy; override;
    // return Value
    function GetDataValue: TACRVariant;  virtual; abstract;
    // is expression contains aggregated function
    function IsAggregated: Boolean; virtual;
    // Init for group function
    procedure Init; virtual;
    // Accumulate for group functions
    procedure Accumulate; virtual;

    // process assign AO
    procedure AssignAO(AO: TACRAO); virtual;
    // process assign Cursor
    procedure AssignCursor(Cursor: TACRCursor); virtual;
    // process assign New Cursor Buffer
    procedure AssignCursorBuffer(Buffer: TACRRecordBuffer); virtual;

    // return Type of Data
    function GetDataType: TACRAdvancedFieldType; virtual;
    // return Size of Data (for strings and arrays)
    function GetDataSize: Integer; virtual;
    // return Data Precision
    function GetPrecision: Integer; virtual;

    // can be used by the AO?
    function CanBeAssigned(AO: TACRAO): Boolean; virtual;

    // is node a join condition?
    function IsJoinCondition(AO1, AO2: TACRAO): Boolean; virtual;

    // replace pseudonyms to original names (f1 -> table1.field1)
    procedure ReplacePseudonyms(SelectList: array of TACRSelectListItem); virtual;
  end;

  TACRExprNodeConst = class(TACRExprNode)
   public
    // return Value
    function GetDataValue: TACRVariant; override;
    // return Type of Data
    function GetDataType: TACRAdvancedFieldType; override;
    // return Size of Data (for strings and arrays)
    function GetDataSize: Integer; override;

    // can be used by the AO?
    function CanBeAssigned(AO: TACRAO): Boolean; override;
  end;


  TACRExprNodeField = class(TACRExprNode)
   private
    LCursor:              TACRCursor; // Cursor
    LAO:                  TACRAO;
    LRecordBuffer:        PChar;
    FTableName:           string;  // table name - Table1 (Table1.Field1)
    FFieldName:           string;  // field name - Field1
    FFieldNo:             Integer;
    FFieldOffsetInBuffer: Integer;
    FFieldType:           TACRAdvancedFieldType;
    FBaseFieldType:       TACRBaseFieldType;
    FFieldSize:           Integer;
    FFieldPrecision:      Integer;
   public
    constructor Create(Cursor: TACRCursor; FieldName: String; TableName: string = ''); overload;
    // Return Data Value
    function GetDataValue: TACRVariant;  override;
    // process assign AO
    procedure AssignAO(AO: TACRAO); override;
    // process assign Cursor and its buffer
    procedure AssignCursor(Cursor: TACRCursor); override;
    // process assign New Cursor Buffer
    procedure AssignCursorBuffer(Buffer: TACRRecordBuffer); override;
    // return Type of Data
    function GetDataType: TACRAdvancedFieldType; override;
    // return Data Size
    function GetDataSize: Integer;  override;

    // can be used by the AO?
    function CanBeAssigned(AO: TACRAO): Boolean; override;
    // fills Field Item
    procedure FillItem(var Item: TACRSelectListItem);
    // replace pseudonyms to original names (f1 -> table1.field1)
    procedure ReplacePseudonyms(SelectList: array of TACRSelectListItem); override;

   public
    property TableName: String read FTableName;
    property FieldName: String read FFieldName;
    property FieldNo: Integer read FFieldNo;
    property FieldSize: Integer read FFieldSize;
    property BaseFieldType: TACRBaseFieldType read FBaseFieldType;
  end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeSubQuery
//
////////////////////////////////////////////////////////////////////////////////


  // expression node SubQuery
  TACRExprNodeSubQuery = class (TACRExprNode)
  private
    FQuery:           TDataset; // query
    FNot:             Boolean;  // NOT option for IN or EXISTS
    FSourceNode:      TACRExprNode; // source node for IN
    FQueryExecuted:   Boolean;
  public
    constructor Create(QueryText:         String;
                       Params:            TParams;
                       InMemory:          Boolean;
                       DatabaseName:      String;
                       SessionName:       String;
                       bNot:              Boolean = False;
                       SourceNode:        TACRExprNode = nil
                       ); overload;
    // destroy
    destructor Destroy; override;
    // return Data Value
    function GetDataValue: TACRVariant; override;
  end;



////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeSubQueryExists
//
////////////////////////////////////////////////////////////////////////////////



  // expression node SubQueryIN
  TACRExprNodeSubQueryIN = class (TACRExprNodeSubQuery)
  public
    // return Data Value
    function GetDataValue: TACRVariant; override;
  end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeSubQueryExists
//
////////////////////////////////////////////////////////////////////////////////


  // expression node SubQueryExists
  TACRExprNodeSubQueryExists = class (TACRExprNodeSubQuery)
  public
    // return Data Value
    function GetDataValue: TACRVariant; override;
  end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeComparison
//
////////////////////////////////////////////////////////////////////////////////


  TACRExprNodeComparison = class(TACRExprNode)
   private
    F3ValueLogic: Boolean; // 3 Value Logic (TRUE, FALSE, NULL)
   private
    procedure PatchWideStrings; override;
   public
    // Constructor
    constructor Create(
                       Op: TACRDataOperator;
                       Node1, Node2: TACRExprNode;
                       TrueFalseNullLogic: boolean = true;
                       CaseInsensitive: boolean = true;
                       PartialKey: boolean = false
                       ); overload;
    // Return Data Value
    function GetDataValue: TACRVariant;  override;
    // can use index for comparison?
    function CanUseIndex(IndexDef: TACRIndexDef): Boolean;
    // add index scan condition
    procedure AddAsIndexScanCondition(
                   ScanConditions: TACRScanSearchConditionArray;
                   IndexDef:       TACRIndexDef;
                   ExtractedConditionsInfo: TList;
                   Expression:  TACRExpression
                              );
    // return true if expr node is incompatible (node comparison)
    function IsIncompatible: Boolean;
    // is node a join condition?
    function IsJoinCondition(AO1, AO2: TACRAO): Boolean; override;
  end;


  TACRExprNodeBoolean = class(TACRExprNode)
   private
    TempVal: TACRVariant;
   private
    // process a Like b
    procedure Like;
    // process IN (...)
    procedure bIn;
    // process A BETWEEN B AND C
    procedure Between;
   public
    constructor Create(
                       CaseInsensitive: boolean = true;
                       PartialKey: boolean = false
                      ); overload;
    constructor Create(
                       Op: TACRDataOperator;
                       CaseInsensitive: boolean = true;
                       PartialKey: boolean = false
                      ); overload;
    constructor Create(
                       Op: TACRDataOperator;
                       Node: TACRExprNode;
                       CaseInsensitive: boolean = true;
                       PartialKey: boolean = false
                      ); overload;
    constructor Create(
                       Op: TACRDataOperator;
                       Node1, Node2: TACRExprNode;
                       CaseInsensitive: boolean = true;
                       PartialKey: boolean = false
                       ); overload;
    constructor Create(
                       Op: TACRDataOperator;
                       Node1, Node2, Node3: TACRExprNode;
                       CaseInsensitive: boolean = true;
                       PartialKey: boolean = false
                       ); overload;
    destructor Destroy; override;
    // Return Data Value
    function GetDataValue: TACRVariant;  override;
  end;


  TACRExprNodeArithmetic = class(TACRExprNode)
   protected
    // SUB operation
    procedure SubData;
    // ADD operation
    procedure AddData;
    // MUL operation
    procedure MulData;
    // DIV operation
    procedure DivData;
   public
    // Return Data Value
    function GetDataValue: TACRVariant;  override;
  end;


  TACRExprNodeAggregated = class(TACRExprNode)
   private
    count:        integer;
    distinct:     Boolean;
    AvgSum:       TACRVariant;
    FTempTable:   TDataset;
   public
    // Constructors
    constructor Create(Op: TACRDataOperator); overload;
    constructor Create(Op: TACRDataOperator;
                       distinct: boolean;
                       Node: TACRExprNode); overload;
    // Destructor
    destructor Destroy; override;
    // Return Data Value
    function GetDataValue: TACRVariant;  override;
    // return Type of Data
    function GetDataType: TACRAdvancedFieldType; override;
    // is expression contains aggregated function
    function IsAggregated: Boolean; override;
    // Init for aggregated functions
    procedure Init; override;
    // Accumulate for group functions
    procedure Accumulate; override;
  end;


  TACRExprNodeCast = class(TACRExprNode)
  private
    FCastType: TACRAdvancedFieldType;  // Type for convertation
   public
    constructor Create(Node: TACRExprNode; CastType: TACRAdvancedFieldType); overload;
    // Return Data Value
    function GetDataValue: TACRVariant;  override;
    // return Type of Data
    function GetDataType: TACRAdvancedFieldType; override;
  end;


  TACRExprNodeStringFunction = class(TACRExprNode)
   protected
    function InitStringValue: Boolean;
   protected
    procedure Concat;
    procedure Upper;
    procedure Lower;
    procedure Trim;
    procedure LTrim;
    procedure RTrim;
    procedure Length;
    procedure Pos;
    procedure Substring;
   public
    // Return Data Value
    function GetDataValue: TACRVariant;  override;
    // return Type of Data
    function GetDataType: TACRAdvancedFieldType; override;
    // return Size of Data (for strings and arrays)
    function GetDataSize: Integer; override;
  end;


  TACRExprNodeSystem = class(TACRExprNode)
   private
    FSessionName:   String;
    FDatabaseName:   String;
    FTableName:      String;
    FColumnName:     String;
    FInMemory:       Boolean;
    FTable:          TDataSet;
    FFieldType:      TACRAdvancedFieldType;
   private
    procedure LastAutoInc;
   public
    // creates
    constructor Create(Operator:     TACRDataOperator;
                       TableName:    String;
                       ColumnName:   String;
                       InMemory:     Boolean;
                       SessionName:  String;
                       DatabaseName: String); overload;
    // Destructor
    destructor Destroy; override;
    // Return Data Value
    function GetDataValue: TACRVariant;  override;
    // return Type of Data
    function GetDataType: TACRAdvancedFieldType; override;
    // return Size of Data (for strings and arrays)
    function GetDataSize: Integer; override;
  end;


  TACRExprNodeDateFunction = class(TACRExprNode)
   private
    DateFormater: TDateFormater;
    // ToDate
    procedure ToDate;
    // ToString
    procedure ToString;
   public
    // Constructor
    constructor Create(Op: TACRDataOperator;
                       Node: TACRExprNode;
                       FormatStr: String); overload;
    // Destructor
    destructor Destroy; override;
    // Return Data Value
    function GetDataValue: TACRVariant;  override;
    // return Type of Data
    function GetDataType: TACRAdvancedFieldType; override;
    // return Size of Data (for strings and arrays)
    function GetDataSize: Integer; override;
  end;


  TACRExprNodeBLOBFunction = class(TACRExprNode)
   public
    // Constructor
    constructor Create(Op:                    TACRDataOperator;
                       const BLOBValue:       String;
                       const BLOBValueFormat: TACRBLOBValueFormat); overload;
    // Destructor
    destructor Destroy; override;
    // Return Data Value
    function GetDataValue: TACRVariant;  override;
    // return Type of Data
    function GetDataType: TACRAdvancedFieldType; override;
    // return Size of Data (for strings and arrays)
    function GetDataSize: Integer; override;
  end; // TACRExprNodeBLOBFunction

// get operator name
function GetOperatorName(op: TACRDataOperator): String;

// return FieldType
function GetFieldType(const TypeName: String): TACRAdvancedFieldType;
// return SQL field name
function GetFieldTypeSQLName(AdvFieldType: TACRAdvancedFieldType): String;

procedure FillFieldNames(List: TStringList; const Names: String);

implementation

uses ACRMain;

//------------------------------------------------------------------------------
// get operator name
//------------------------------------------------------------------------------
function GetOperatorName(op: TACRDataOperator): String;
begin
  case op of
    doNOT:          Result := 'NOT';
    doAND:          Result := 'AND';
    doOR:           Result := 'OR';
    doEQ:           Result := '=';
    doNE:           Result := '<>';
    doLT:           Result := '<';
    doGT:           Result := '>';
    doLE:           Result := '<=';
    doGE:           Result := '>=';
    doLIKE:         Result := 'LIKE';
    doNOTLIKE:      Result := 'NOT LIKE';
    doISNULL:       Result := 'IS NULL';
    doISNOTNULL:    Result := 'IS NOT NULL';
    doADD:          Result := '+';
    doSUB:          Result := '-';
    doMUL:          Result := '*';
    doDIV:          Result := '/';
    doCONCAT:       Result := '||';
    doUPPER:        Result := 'UPPER';
    doLOWER:        Result := 'LOWER';
    doTRIM:         Result := 'TRIM';
    doLTRIM:        Result := 'LTRIM';
    doRTRIM:        Result := 'RTRIM';
    doSUM:          Result := 'SUM';
    doAVG:          Result := 'AVG';
    doMIN:          Result := 'MIN';
    doMAX:          Result := 'MAX';
    doCOUNT:        Result := 'COUNT';
    doCOUNTALL:     Result := 'COUNT(*)';
    else            Result:='unknown';
  end;
end;//GetOperatorName


//------------------------------------------------------------------------------
// Return FieldType
//------------------------------------------------------------------------------
function GetFieldType(const TypeName: String): TACRAdvancedFieldType;
var
  i: Integer;
  s: String;
begin
  Result := aftUnknown;
  s := AnsiUpperCase(TypeName);
  for i := Low(SQLFieldTypes) to High(SQLFieldTypes) do
    if SQLFieldTypes[i].SqlName = s then
      begin
        Result := SQLFieldTypes[i].AdvancedFieldType;
        break;
      end;
end;//GetFieldType


//------------------------------------------------------------------------------
// return SQL field name
//------------------------------------------------------------------------------
function GetFieldTypeSQLName(AdvFieldType: TACRAdvancedFieldType): String;
var
  i: Integer;
begin
  Result := '';
  for i := Low(SQLFieldTypes) to High(SQLFieldTypes) do
    if SQLFieldTypes[i].AdvancedFieldType = AdvFieldType then
     begin
      Result := SQLFieldTypes[i].SqlName;
      break;
     end;
end; // GetFieldTypeSQLDefinition


////////////////////////////////////////////////////////////////////////////////
//
// TACRExpression
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExpression.Create;
begin
  FRootExprNode := nil;
  LLex := nil;
  FCaseInsensitive := False;
  FPartialKey := False;
  F3ValueLogic := True;
  Params := nil;
end;//Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExpression.Create(RootNode: TACRExprNode);
begin
  Params := nil;
  Create;
  FRootExprNode := RootNode;
end;//Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
//constructor TACRExpression.Create(Lexer: TACRLexer);
//begin
//  Create;
//  LLex := Lexer;
//end;//Create



//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRExpression.Destroy;
begin
  Clear;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// Clear all Variables
//------------------------------------------------------------------------------
procedure TACRExpression.Clear;
begin
  // free nodes tree
  if (FRootExprNode <> nil) then
    begin
      FRootExprNode.Free;
      FRootExprNode := nil;
    end;
end;//Clear


//------------------------------------------------------------------------------
// PrepareForeignKeyCheck
//------------------------------------------------------------------------------
procedure TACRExpression.PrepareForeignKeyCheck(
                                  ReferencedCursor: TACRCursor;
                                  Cursor:           TACRCursor;
                                  ConstraintDef:    TACRConstraintDefForeignKey;
                                  PrimaryIndex:     TACRIndexDef
                               );
var
    i,h,FieldNo:    Integer;
    value:          TACRVariant;
    EqNode:         TACRExprNode;
    NodeConst:      TACRExprNodeConst;
    NodeField:      TACRExprNodeField;
begin
  Clear;
  LCursor := ReferencedCursor;
  F3ValueLogic := False;
  FCaseInsensitive := False;
  FPartialKey := False;
  value := TACRVariant.Create;
  try
    h := High(ConstraintDef.Columns);
    for i := 0 to h do
     begin
      FieldNo := Cursor.FieldDefs.GetDefNumberByObjectId(ConstraintDef.Columns[i].ColumnObjectID);
      Cursor.GetFieldValue(value,FieldNo,True,False);
      if (not value.IsNull) then
       begin
        // const (from record in detail table - Cursor)
        NodeConst := TACRExprNodeConst.Create(False,False);
        NodeConst.Value.Assign(value,True);
        // field (from referenced table)
        NodeField := TACRExprNodeField.Create(LCursor, PrimaryIndex.Columns[i].FieldName);
        EqNode := TACRExprNodeComparison.Create(doEQ,NodeField,NodeConst,False,False,False);
        if (FRootExprNode = nil) then
         begin
          if (i = h) then
           FRootExprNode := EqNode
          else
           FRootExprNode := TACRExprNodeBoolean.Create(doAND,EqNode,False,False);
         end
        else
         FRootExprNode.Children.Add(EqNode);
       end;
     end; // for (scan constraint columns)
  finally
    value.Free;
  end;
  FRootExprNode.PatchWideStrings;
end; // PrepareForeignKeyCheck


//------------------------------------------------------------------------------
// foreign key action
//------------------------------------------------------------------------------
procedure TACRExpression.PrepareForeignKeyActionFilter(
                                    ReferencedCursor: TACRCursor;
                                    Cursor:           TACRCursor;
                                    ConstraintDef:    TACRConstraintDefForeignKeyAction;
                                    PrimaryIndex:     TACRIndexDef
                                 );
var
    i,h,FieldNo:  Integer;
    value:          TACRVariant;
    EqNode:         TACRExprNode;
    NodeConst:      TACRExprNodeConst;
    NodeField:      TACRExprNodeField;
begin
  Clear;
  LCursor := ReferencedCursor;
  F3ValueLogic := (ConstraintDef.MatchType = cfkmtPartial);
  FCaseInsensitive := False;
  FPartialKey := False;
  value := TACRVariant.Create;
  try
    h := High(ConstraintDef.Columns);
    for i := 0 to h do
     begin
      FieldNo := Cursor.FieldDefs.GetDefNumberByName(PrimaryIndex.Columns[i].FieldName);
      Cursor.GetFieldValue(value,FieldNo,True,False);
      if ((not value.IsNull) or (F3ValueLogic)) then
       begin
        // const (from record in master table - Cursor)
        NodeConst := TACRExprNodeConst.Create(False,False);
        NodeConst.Value.Assign(value,True);
        // field (from detail table)
        NodeField := TACRExprNodeField.Create(LCursor, ConstraintDef.Columns[i].ColumnName);
        EqNode := TACRExprNodeComparison.Create(doEQ,NodeField,NodeConst,F3ValueLogic,False,False);
        if (FRootExprNode = nil) then
         begin
          if (i = h) then
           FRootExprNode := EqNode
          else
           FRootExprNode := TACRExprNodeBoolean.Create(doAND,EqNode,False,False);
         end
        else
         FRootExprNode.Children.Add(EqNode);
       end;
     end; // for (scan constraint columns)
  finally
    value.Free;
  end;
  FRootExprNode.PatchWideStrings;
end; // PrepareForeignKeyActionFilter


//------------------------------------------------------------------------------
// Parsing for Locate
//------------------------------------------------------------------------------
procedure TACRExpression.ParseForLocate(
                                          Cursor: TACRCursor;
                                          FieldNames: String; // name1;name2
                                          KeyValues: Variant; //
                                          CaseInsensitive: boolean = true;
                                          PartialKey: boolean = false
                                       );


var
  i, ArrLen: Integer;
  FieldNamesList: TStringList;
  //Arr: Pointer;
  LeftNField: TACRExprNodeField;
  RightNConst: TACRExprNodeConst;
  EqNode: TACRExprNodeComparison;
  //AndNode: TACRExprNodeBoolean;
  //val: TACRVariant;
begin
  Clear;
  LCursor := Cursor;
  F3ValueLogic := False;

  FieldNamesList := TStringList.Create;
  try
    FillFieldNames(FieldNamesList, FieldNames);

    // Determinate KeyValues count (can be array)
    if  (VarType(KeyValues) and varArray ) <> 0 then
      ArrLen := VarArrayHighBound(KeyValues,1) - VarArrayLowBound(KeyValues,1) + 1
    else
      ArrLen := 1;

    if (FieldNamesList.Count <> ArrLen) then
      raise EACRException.Create(30112, ErrorGNotEqualCountsOfFieldNamesAndKeyValues,
                                      [FieldNamesList.Count, ArrLen]);

    if (ArrLen = 1) then
      begin
        // Field
        LeftNField := TACRExprNodeField.Create(LCursor, FieldNamesList[0]);
        // Const
        RightNConst := TACRExprNodeConst.Create;
        RightNConst.Value.AsVariant := KeyValues;
        // RootNode
        FRootExprNode := TACRExprNodeComparison.Create(doEQ, LeftNField,
                                      RightNConst, F3ValueLogic, CaseInsensitive, PartialKey);
      end
    else
      begin
        FRootExprNode := TACRExprNodeBoolean.Create(doAND);
        for i:=0 to ArrLen-1 do
          begin
            // Field
            LeftNField := TACRExprNodeField.Create(LCursor, FieldNamesList[i]);
            // Check FieldExists
            // ...
            // Const
            RightNConst := TACRExprNodeConst.Create;
            RightNConst.Value.AsVariant := VarArrayGet(KeyValues, i);

            // EqNode
            EqNode := TACRExprNodeComparison.Create(doEQ, LeftNField, RightNConst,
                                       F3ValueLogic, CaseInsensitive, PartialKey);
            // Add to RootNode
            FRootExprNode.Children.Add(EqNode);
          end;
      end;
  finally
    FieldNamesList.Free;
  end;

  if LCursor <> nil then
    FRootExprNode.PatchWideStrings;

end;//ParseForLocate


//------------------------------------------------------------------------------
// Parsing for Filter
//------------------------------------------------------------------------------
procedure TACRExpression.ParseForFilter(Cursor: TACRCursor; Filter: String;
  CaseInsensitive, PartialKey: boolean);
var
  Lexer: TACRLexer;
begin
  LCursor := Cursor;
  FCaseInsensitive := CaseInsensitive;
  FPartialKey := PartialKey;

  Lexer := TACRLexer.Create(Filter);
  try
    if (not Lexer.GetNextCommand) then
       raise EACRException.Create(30119, ErrorGBlankSqlCommand);

    ParseForBooleanExpression(Lexer);
  finally
    Lexer.Free;
  end;
end;//ParseForFilter


//------------------------------------------------------------------------------
// Parse For Boolean Expression (Filter, Where-Clause)
//------------------------------------------------------------------------------
procedure TACRExpression.ParseForBooleanExpression(
                                        //Cursor: TACRCursor;
                                        Lexer: TACRLexer
                                        //CaseInsensitive: boolean = true;
                                        //PartialKey: boolean = false
                                                   );
begin
  Clear;
  LLex := Lexer;

  if (not LLex.GetCurrentToken(Token)) then
    raise EACRException.Create(30118, ErrorGUnexpectedEndOfCommand,
                                              [Token.LineNum, Token.ColumnNum]);
  // Parse...
  FRootExprNode := ParseSearchCondition;

  MoveAndNodesToRoot;

  if LCursor <> nil then
    FRootExprNode.PatchWideStrings;
end;//ParseForBooleanExpression


//------------------------------------------------------------------------------
// Parse ValueExpression
//------------------------------------------------------------------------------
procedure TACRExpression.ParseForValueExpression(
                                        //Cursor: TACRCursor;
                                        Lexer: TACRLexer
                                        //CaseInsensitive: boolean = true;
                                        //PartialKey: boolean = false
                                                 );
begin
 LLex := Lexer;
 //LCursor := Cursor;
 //FCaseInsensitive := CaseInsensitive;
 //FPartialKey := PartialKey;

 // get first token (for very beginning of the query) or current token
 if (not LLex.GetCurrentToken(Token)) then
  raise EACRException.Create(30136, ErrorGUnexpectedEndOfCommand, [Token.LineNum, Token.ColumnNum]);

 // parse
 FRootExprNode := ParseValueExpression;
end;//ParseForValueExpression


//------------------------------------------------------------------------------
// Return Variant
//------------------------------------------------------------------------------
function TACRExpression.GetValue(TrueFalseNullLogic: boolean): TACRVariant;
begin
  if FRootExprNode <> nil then
    begin
      Result := FRootExprNode.GetDataValue;
      if ( Result.IsNull and (not TrueFalseNullLogic) ) then
        Result.AsBoolean := False;
    end
  else
    Result := nil;
end;//GetValue


//------------------------------------------------------------------------------
// Get Boolean result
//------------------------------------------------------------------------------
function TACRExpression.GetResult: Boolean;
var
  Value: TACRVariant;
begin
  Value := GetValue(false);
  if Value = nil then
    raise EACRException.Create(30110, ErrorGValueIsNull);
  Result := Value.AsBoolean;
end;//GetResult


//------------------------------------------------------------------------------
// is expression contains aggregated function
//------------------------------------------------------------------------------
function TACRExpression.IsAggregated: Boolean;
begin
 if FRootExprNode <> nil then
   Result := FRootExprNode.IsAggregated
 else
   Result := False;
end;//IsAggregated


//------------------------------------------------------------------------------
// Init for aggregated functions
//------------------------------------------------------------------------------
procedure TACRExpression.Init;
begin
  if FRootExprNode <> nil then
    FRootExprNode.Init;
end;//Init


//------------------------------------------------------------------------------
// Accumulate for group functions
//------------------------------------------------------------------------------
procedure TACRExpression.Accumulate;
begin
  if FRootExprNode <> nil then
    FRootExprNode.Accumulate;
end;//Accumulate



//------------------------------------------------------------------------------
// process assign AO
//------------------------------------------------------------------------------
procedure TACRExpression.AssignAO(AO: TACRAO);
begin
  if (FRootExprNode <> nil) then
    begin
      FRootExprNode.AssignAO(AO);
      FRootExprNode.PatchWideStrings;
    end;
end;//AssignAO


//------------------------------------------------------------------------------
// process assign Cursor and its RecordBuffer
//------------------------------------------------------------------------------
procedure TACRExpression.AssignCursor(Cursor: TACRCursor);
begin
  if (FRootExprNode <> nil) then
    begin
      FRootExprNode.AssignCursor(Cursor);
      FRootExprNode.PatchWideStrings;
    end;
end;//AssignCursor


//------------------------------------------------------------------------------
// process assign New Cursor Buffer
//------------------------------------------------------------------------------
procedure TACRExpression.AssignCursorBuffer(Buffer: TACRRecordBuffer);
begin
  if (FRootExprNode <> nil) then
    begin
      FRootExprNode.AssignCursorBuffer(Buffer);
      FRootExprNode.PatchWideStrings;
    end;
end;//AssignCursorBuffer


//------------------------------------------------------------------------------
// return Expression DataSize
//------------------------------------------------------------------------------
function TACRExpression.GetDataSize: Integer;
begin
 if FRootExprNode <> nil then
   Result := FRootExprNode.getDataSize
 else
   Result := 0;
end;//getDataSize


//------------------------------------------------------------------------------
// return Size of Data (for strings and arrays)
//------------------------------------------------------------------------------
function TACRExpression.GetPrecision: Integer;
begin
 if FRootExprNode <> nil then
   Result := FRootExprNode.GetPrecision
 else
   Result := 0;
end;//GetPrecision


//------------------------------------------------------------------------------
// return Expression DataType
//------------------------------------------------------------------------------
function TACRExpression.GetDataType: TACRAdvancedFieldType;
begin
 if FRootExprNode <> nil then
   Result := FRootExprNode.getDataType
 else
   Result := aftUnknown;
end;//GetDataType


//------------------------------------------------------------------------------
// is expression contains no nodes
//------------------------------------------------------------------------------
function TACRExpression.IsEmpty: Boolean;
begin
  Result := (FRootExprNode = nil);
end;//IsEmpty


//------------------------------------------------------------------------------
// is expression a Field (for join)
//------------------------------------------------------------------------------
function TACRExpression.IsField: Boolean;
begin
  Result := False;
  if FRootExprNode <> nil then
    Result := FRootExprNode is TACRExprNodeField;
end;//IsField


//------------------------------------------------------------------------------
// is expression a constant
//------------------------------------------------------------------------------
function TACRExpression.IsConst: Boolean;
begin
  Result := False;
  if FRootExprNode <> nil then
    Result := FRootExprNode is TACRExprNodeConst;
end; // IsConst


//------------------------------------------------------------------------------
// MoveAndNodesToRoot
//------------------------------------------------------------------------------
procedure TACRExpression.MoveAndNodesToRoot;
var
  i: Integer;
begin
  if (FRootExprNode.Operator = doAND) then
   begin
    i := 0;
    while i < FRootExprNode.Children.Count-1 do
     begin
      if (TACRExprNode(FRootExprNode.Children[i]).Operator = doAND) then
       begin
        while TACRExprNode(FRootExprNode.Children[i]).Children.Count > 0 do
         begin
          FRootExprNode.Children.Add(TACRExprNode(FRootExprNode.Children[i]).Children[0]);
          TACRExprNode(FRootExprNode.Children[i]).Children.Delete(0);
         end;
        TACRExprNode(FRootExprNode.Children[i]).Free;
        FRootExprNode.Children.Delete(i);
        continue;
       end;
      Inc(i);
     end;
   end;
end; // MoveAndNodesToRoot


//------------------------------------------------------------------------------
// Field Name, Table Name
//------------------------------------------------------------------------------
procedure TACRExpression.GetFieldInfo(var TableName, FieldName: String);
begin
  if (IsField) then
   begin
    TableName := TACRExprNodeField(FRootExprNode).FTableName;
    FieldName := TACRExprNodeField(FRootExprNode).FFieldName;
   end
  else
   raise EACRException.Create(30198, ErrorGNodeIsNotField);
end;// GetFieldInfo


//------------------------------------------------------------------------------
// makes filter string from related parts and sets it to AO
//------------------------------------------------------------------------------
procedure TACRExpression.ApplyFilterParts(AO: TACRAO; bHaving: Boolean);
var
  NodesToDelete:    TList;
  Node:             TACRExprNode;
  bRecurseLeft:     Boolean;
  bRecurseRight:    Boolean;
  i:                Integer;
  FilterRootNode:   TACRExprNode;
begin
  FilterRootNode := nil;
  bRecurseLeft := True;
  bRecurseRight := True;
  // don't set filter to outer join childs
  if (AO is TACRAOJoin) then
   begin
    if (TACRAOJoin(AO).JoinType = ajtFullOuter) then
     begin
      bRecurseLeft := False;
      bRecurseRight := False;
     end
    else
    if (TACRAOJoin(AO).JoinType = ajtLeftOuter) then
     begin
      bRecurseLeft := True;
      bRecurseRight := False;
     end
    else
    if (TACRAOJoin(AO).JoinType = ajtRightOuter) then
     begin
      bRecurseLeft := False;
      bRecurseRight := True;
     end
   end;

  if (bHaving) then
   begin
    bRecurseLeft := False;
    bRecurseRight := False;
   end;


  // try to apply to AO children
  if (bRecurseLeft or bRecurseRight) then
   begin
    if (bRecurseLeft) then
     if (AO.FLeftAO <> nil) then
      ApplyFilterParts(AO.FLeftAO,bHaving);
    if (bRecurseRight) then
     if (AO.FRightAO <> nil) then
      ApplyFilterParts(AO.FRightAO,bHaving);
   end;

  // traverse tree by factors: (...) AND (...) AND (...)
  if (FRootExprNode <> nil) then
   begin
     NodesToDelete := TList.Create;
     try
      // -> (...) AND (...) AND (...) ?
      if (FRootExprNode is TACRExprNodeBoolean) and
         (FRootExprNode.Operator = doAND) then
        for i:=0 to FRootExprNode.Children.Count-1 do
         begin
          Node := FRootExprNode.Children[i];
          // (...) can be used by the AO?
          if (Node.CanBeAssigned(AO)) then
           begin
            // no FilterRootNode
            if FilterRootNode = nil then
              FilterRootNode := Node
            else
            // FilterRootNode=AND ( ex: a=b AND c=d )
            if (FilterRootNode is TACRExprNodeBoolean) and
               (FilterRootNode.Operator = doAND) then
               FilterRootNode.Children.Add(Node)
            else
              // single FilterRootNode ( ex: a=b ) - converting to AND
              FilterRootNode := TACRExprNodeBoolean.Create(doAND,
                                                FilterRootNode, Node);
            NodesToDelete.Add(Node);
           end;
          // -> AND (...)
         end
      else
        // (...)
        begin
          if (FRootExprNode.CanBeAssigned(AO)) then
            begin
              FilterRootNode := FRootExprNode;
              FRootExprNode := nil;
            end;
        end;
      // delete used nodes
      for i:=0 to NodesToDelete.Count-1 do
       begin
        FRootExprNode.Children.Delete(
          FRootExprNode.Children.IndexOf(NodesToDelete[i]));
       end;
      // set extracted filter
      if FilterRootNode <> nil then
       AO.SetFilter(TACRExpression.Create(FilterRootNode));

      // no children?
      if (FRootExprNode <> nil) then
       if (FRootExprNode.Children.Count = 0) then
        begin
         FRootExprNode.Free;
         FRootExprNode := nil;
        end;
//      if (FilterString <> '') then
//       AO.SetFilter(FilterString, []);
     finally
      NodesToDelete.Free;
     end;
   end;
end;// ApplyFilterParts


//------------------------------------------------------------------------------
// gets current token
//------------------------------------------------------------------------------
procedure TACRExpression.ReplacePseudonyms(SelectList: array of TACRSelectListItem);
begin
 if (FRootExprNode <> nil) then
    FRootExprNode.ReplacePseudonyms(SelectList);
end;//ReplacePseudonyms


//------------------------------------------------------------------------------
// makes join field lists
//------------------------------------------------------------------------------
function TACRExpression.ExtractJoinConditions(
                                    AO1, AO2: TACRAO;
                                    FieldList1, FieldList2: TACRFields
                                             ): Integer;
var
  NodesToDelete: TList;
  Node: TACRExprNode;
  Item1,Item2: TACRSelectListItem;
  i: Integer;
begin
  Result := 0;
  if FRootExprNode = nil then Exit;
  // If 'AND'
  if ((FRootExprNode is TACRExprNodeBoolean) and
      (FRootExprNode.Operator = doAND)) then
   begin
     NodesToDelete := TList.Create;
     try
      for i:=0 to FRootExprNode.Children.Count-1 do
       begin
        Node := FRootExprNode.Children[i];
        // (...) can be used by the AO?
        if (Node.IsJoinCondition(AO1,AO2)) then
         begin
          // store extracted filter
          TACRExprNodeField(Node.Children[0]).FillItem(Item1);
          TACRExprNodeField(Node.Children[1]).FillItem(Item2);
          FieldList1.Append(Item1);
          FieldList2.Append(Item2);
          Inc(Result);
          // will delete this node
          NodesToDelete.Add(Node);
         end;
        // -> AND (...)
        //Node := Node.NextFactor;
       end;
      // delete used nodes
      for i:=0 to NodesToDelete.Count-1 do
       begin
        FRootExprNode.Children.Delete(FRootExprNode.Children.IndexOf(NodesToDelete[i]));
        TACRExprNode(NodesToDelete[i]).Free;
       end;
      if (FRootExprNode.Children.Count = 0) then
       begin
         FRootExprNode.Free;
         FRootExprNode := nil;
       end;
     finally
      NodesToDelete.Free;
     end;
   end
 else
 // IF '='
 if ((FRootExprNode is TACRExprNodeComparison) and
    (FRootExprNode.Operator = doEQ)) then
  begin
    if (FRootExprNode.IsJoinCondition(AO1,AO2)) then
     begin
      // store extracted filter
      TACRExprNodeField(FRootExprNode.Children[0]).FillItem(Item1);
      TACRExprNodeField(FRootExprNode.Children[1]).FillItem(Item2);
      FieldList1.Append(Item1);
      FieldList2.Append(Item2);
      Inc(Result);
      // will delete this node
      FRootExprNode.Free;
      FRootExprNode := nil;
     end;
  end;
end; // ExtractJoinConditions


//------------------------------------------------------------------------------
// extract conditions for index scan
//------------------------------------------------------------------------------
procedure TACRExpression.TryExtractIndexScanConditionsFromNode(
                         Node:           TACRExprNodeComparison;
                         IndexDefs:      TACRIndexDefs;
                         ScanConditions: TACRScanSearchConditionArray;
                         ExtractedConditionsInfo: TList
                                    );
var
  j:  Integer;
begin
  for j := 0 to IndexDefs.Count-1 do
  if (Node.CanUseIndex(IndexDefs[j])) then
    begin
      Node.AddAsIndexScanCondition(
                                  ScanConditions,
                                  IndexDefs[j],
                                  ExtractedConditionsInfo,
                                  Self
                                  );
      break;
    end;
end;// TryExtractIndexScanConditionsFromNode


//------------------------------------------------------------------------------
// remove extracted nodes (they included in list, but list contains more)
//------------------------------------------------------------------------------
procedure TACRExpression.RemoveExtractedNodes(ExtractedConditionsInfo: TList);
var
  Node: TACRExprNode;
  i, k: Integer;
begin
  for i := 0 to ExtractedConditionsInfo.Count-1 do
   if (PACRExtractedConditionInfo(ExtractedConditionsInfo[i])^.Expression = Self) then
    begin
      Node := TACRExprNode(PACRExtractedConditionInfo(ExtractedConditionsInfo[i])^.
                           ExtractedExpressionNode);
      if (Node <> FRootExprNode) then
       begin
         k := FRootExprNode.Children.IndexOf(Node);
         if (k < 0) then
           raise EACRException.Create(20059, ErrorAInvalidIndexOfNo)
         else
          FRootExprNode.Children.Delete(k);
         if (FRootExprNode.Children.Count <= 0) then
          FreeAndNil(FRootExprNode);
       end
      else
       FRootExprNode := nil;
    end;
end;// RemoveExtractedNodes


//------------------------------------------------------------------------------
// extract conditions for index scan
//------------------------------------------------------------------------------
procedure TACRExpression.ExtractIndexScanConditions(
                             IndexDefs:               TACRIndexDefs;
                             ScanConditions:          TACRScanSearchConditionArray;
                             ExtractedConditionsInfo: TList
                                        );
var
  Node: TACRExprNode;
  i:    Integer;
begin
  if ((FRootExprNode is TACRExprNodeBoolean) and
      (FRootExprNode.Operator = doAND)) then
    for i := 0 to FRootExprNode.Children.Count-1 do
     begin
      Node := FRootExprNode.Children[i];
      if (Node is TACRExprNodeComparison) then
        TryExtractIndexScanConditionsFromNode(
                                           TACRExprNodeComparison(Node),
                                           IndexDefs,
                                           ScanConditions,
                                           ExtractedConditionsInfo
                                             );
     end
  else
   if (FRootExprNode is TACRExprNodeComparison) then
        TryExtractIndexScanConditionsFromNode(
                                 TACRExprNodeComparison(FRootExprNode),
                                 IndexDefs,
                                 ScanConditions,
                                 ExtractedConditionsInfo
                                             );

  RemoveExtractedNodes(ExtractedConditionsInfo);
end;// ExtractIndexScanConditions


//------------------------------------------------------------------------------
// return true if expression have incompatible conditions (like compare = for string fields with constant longer then field size)
//------------------------------------------------------------------------------
function TACRExpression.IsIncompatible: Boolean;
var
  Node: TACRExprNode;
  i:    Integer;
begin
  Result := False;
  if ((FRootExprNode is TACRExprNodeBoolean) and
      (FRootExprNode.Operator = doAND)) then
    for i := 0 to FRootExprNode.Children.Count-1 do
     begin
      Node := FRootExprNode.Children[i];
      if (Node is TACRExprNodeComparison) then
       begin
        Result := TACRExprNodeComparison(Node).IsIncompatible;
        if (Result) then break;
       end;
     end
  else
   if (FRootExprNode is TACRExprNodeComparison) then
    begin
      Result := TACRExprNodeComparison(FRootExprNode).IsIncompatible;
    end;
end; // IsExpressionHaveIncompatibleConditions


//------------------------------------------------------------------------------
// add extracted node
//------------------------------------------------------------------------------
procedure TACRExpression.AddNode(Node: TACRExprNode);
begin
   if (FRootExprNode <> nil) then
     FRootExprNode.Children.Add(Node)
   else
     FRootExprNode := Node;
end;// AddNode


//------------------------------------------------------------------------------
// gets current token
//------------------------------------------------------------------------------
//function TACRExpression.GetCurrentToken: Boolean;
//begin
// Result := LLex.GetCurrentToken(Token);
//end;// GetCurrentToken


//------------------------------------------------------------------------------
// gets next token
//------------------------------------------------------------------------------
function TACRExpression.GetNextToken: Boolean;
begin
 Result := LLex.GetNextToken(Token);
end;// GetNextToken


//------------------------------------------------------------------------------
// gets token and looks at next token with check for token type restrictions
//------------------------------------------------------------------------------
function TACRExpression.GetNextToken(PermittedTypes: TTokenTypes;
                                      RaiseExceptions: Boolean = False): Boolean;
begin
  // get next token
  Result := LLex.GetNextToken(Token);
  if (not Result) then
   if (RaiseExceptions) then
      raise EACRException.Create(30063, ErrorGUnexpectedEndOfCommand,
                                 [Token.LineNum, Token.ColumnNum]);
  // check token type
  if (Result) then
   if ((PermittedTypes <> [])) then
    if not (Token.TokenType in PermittedTypes) then
      if (RaiseExceptions) then
        raise EACRException.Create(30064, ErrorGUnexpectedToken,
                                   [Token.Text, Token.LineNum, Token.ColumnNum])
      else
         Result := False;
end;// GetNextToken


//------------------------------------------------------------------------------
// saves internal state (to restore in case of not successful forward parsing)
//------------------------------------------------------------------------------
procedure TACRExpression.SaveState(var SavedTokenNo: integer);
//  var SavedRootNode: TACRExprNode);
begin
 // save token No
 SavedTokenNo := LLex.GetCurrentTokenNo;
 // save root node
 //SavedRootNode := FRootNode;
end;// SaveState


//------------------------------------------------------------------------------
// restores internal state (in case of not successful forward parsing)
//------------------------------------------------------------------------------
procedure TACRExpression.RestoreState(SavedTokenNo: integer);
//  SavedRootNode: TACRExprNode);
begin
 // restore token No
 LLex.SetCurrentTokenNo(SavedTokenNo, Token);
 // restore root node
 //FRootNode.DeleteTo(SavedRootNode);
 //FRootNode := SavedRootNode;
end;// RestoreState


//------------------------------------------------------------------------------
// parses <,>,=,<>,>=,<=
//------------------------------------------------------------------------------
function TACRExpression.ParseCompOp: TACRDataOperator;
begin
  if (Token.Text = '<') then
   Result := doLT
  else
  if (Token.Text = '>') then
   Result := doGT
  else
  if (Token.Text = '=') then
   Result := doEQ
  else
  if (Token.Text = '<>') then
   Result := doNE
  else
  if (Token.Text = '<=') then
   Result := doLE
  else
  if (Token.Text = '>=') then
   Result := doGE
  else
   Result := doUNDEFINED;

  if (Result <> doUNDEFINED) then
   GetNextToken;
end;// ParseCompOp


//------------------------------------------------------------------------------
// parses <row value constructor element>
//------------------------------------------------------------------------------
function TACRExpression.ParseRowValueConstructorElement: TACRExprNode;
begin
  Result := ParseNullConst;
  if (Result = nil) then
    Result := ParseValueExpression;
end;// ParseRowValueConstructorElement


//------------------------------------------------------------------------------
// parses <row value constructor element> | <row subquery>
//------------------------------------------------------------------------------
function TACRExpression.ParseRowValueConstructor: TACRExprNode;
begin
 Result := ParseRowValueConstructorElement;
end;// ParseRowValueConstructor


//------------------------------------------------------------------------------
// parses <row value constructor> <comp op> <row value constructor>
//------------------------------------------------------------------------------
function TACRExpression.ParseComparisonPredicate: TACRExprNode;
var
  LeftNode, RightNode: TACRExprNode;
  CompOp: TACRDataOperator;
  pk: Boolean;

 function IsPartialKey(Node: TACRExprNode): integer;
 var i: Integer;
 begin
  Result := 0;
  if Node.ClassType = TACRExprNodeConst then
   begin
    i := Node.Value.StrLen;
    if (i <> -1) then
     if Node.Value.AsString[i] = '*' then
      begin
       if Node.Value.IsWideStringDataType then
         Node.Value.AsWideString := Copy(Node.Value.AsWideString,1,i-1)
       else
         Node.Value.AsString := Copy(Node.Value.AsString,1,i-1);
       Result := 1;
      end;
   end;
 end;

begin
 // parse <row value constructor>
 LeftNode := ParseRowValueConstructor;
 if (LeftNode <> nil) then
  begin
   // <,>,=,<>,>=,<=
   CompOp := ParseCompOp;
   // parse 2nd <row value constructor>
   if (CompOp <> doUndefined) then
    begin
     RightNode := ParseRowValueConstructor;
     if (RightNode <> nil) then
      begin
        pk := False;
       // Detect Partial Key
       if FPartialKey then
        pk := (IsPartialKey(LeftNode) + IsPartialKey(RightNode) > 0);
       // create <CompOp> node
       LeftNode := TACRExprNodeComparison.Create(CompOp, LeftNode, RightNode,
                                             F3ValueLogic, FCaseInsensitive, pk);

      end
     else
      begin
       LeftNode.Free;
       LeftNode := nil;
      end;
    end
   else
    begin
     LeftNode.Free;
     LeftNode := nil;
    end;
  end;
 Result := LeftNode;
end;// ParseComparisonPredicate


//------------------------------------------------------------------------------
// parses <match value> [ NOT ] LIKE <pattern> [ ESCAPE <escape character> ]
//------------------------------------------------------------------------------
function TACRExpression.ParseLikePredicate: TACRExprNode;
var
  Node, LeftNode, RightNode: TACRExprNode;
  bNot: Boolean;
begin
 // parse <character value expression>
 LeftNode := ParseRowValueConstructor;
 if (LeftNode <> nil) then
  begin
   // [NOT]
   if (IsReservedWord(Token, rwNOT)) then
     begin
      bNot := True;
      GetNextToken;
     end
   else
     bNot := False;
   // LIKE
   if (IsReservedWord(Token, rwLIKE)) then
    begin
     // get <pattern>
     GetNextToken;
     RightNode := ParseRowValueConstructor;
     if (not bNot) then
       Node := TACRExprNodeBoolean.Create(doLIKE)
     else
       Node := TACRExprNodeBoolean.Create(doNOTLIKE);
     Node.Children.Add(LeftNode);
     Node.Children.Add(RightNode);
     LeftNode := Node;
    end
   else
    begin
      LeftNode.Free;
      LeftNode := nil;
    end;
  end;
 Result := LeftNode;
end;// ParseLikePredicate


//------------------------------------------------------------------------------
// parses <row value constructor> IS [ NOT ] NULL
//------------------------------------------------------------------------------
function TACRExpression.ParseNullPredicate: TACRExprNode;
var
  LeftNode: TACRExprNode;
  bNot: Boolean;
  bOK: Boolean;
begin
 // parse <row value constructor>
 LeftNode := ParseRowValueConstructor;
 bOK := False;
 if (LeftNode <> nil) then
  begin
   // IS?
   if (IsReservedWord(Token, rwIS)) then
    begin
     // get next token
     GetNextToken;
     // [NOT]
     if (IsReservedWord(Token, rwNOT)) then
      begin
       bNot := True;
       GetNextToken;
      end
     else
      bNot := False;
     // NULL
     if (IsReservedWord(Token, rwNULL)) then
      begin
       // skip token NULL
       GetNextToken;
       if (not bNot) then
        LeftNode := TACRExprNodeBoolean.Create(doISNULL, LeftNode)
       else
        LeftNode := TACRExprNodeBoolean.Create(doISNOTNULL, LeftNode);
       bOk := True;
      end;
    end;
  end;
 if (not bOK) then
    begin
      if (LeftNode <> nil) then
       LeftNode.Free;
      LeftNode := nil;
    end;
 Result := LeftNode;
end;// ParseNullPredicate


//------------------------------------------------------------------------------
// parses <between predicate>
//------------------------------------------------------------------------------
function TACRExpression.ParseBetweenPredicate: TACRExprNode;
var
  arg1, arg2, arg3: TACRExprNode;
  bNot: boolean;
begin
{
<between predicate> ::=
    <row value constructor> [ NOT ] BETWEEN
      <row value constructor> AND <row value constructor>
}
 Result := nil;
 arg2:=nil; arg3:=nil;
 arg1 := ParseRowValueConstructor;
 if arg1 <> nil then
  begin
   // [NOT]
   if (IsReservedWord(Token, rwNOT)) then
    begin
     bNot := True;
     // and Get next token
     GetNextToken;
    end
   else
    bNot := False;
   // BETWEEN
   if (IsReservedWord(Token, rwBETWEEN)) then
    begin
     // get next token
     GetNextToken;
     arg2 := ParseRowValueConstructor;
     if arg2 <> nil then
      begin
       // AND
       if (IsReservedWord(Token, rwAND)) then
        begin
         // get next token
         GetNextToken;
         arg3 := ParseRowValueConstructor;
         if arg3 <> nil then
          if bNot then
           Result := TACRExprNodeBoolean.Create(doNOTBETWEEN, arg1, arg2, arg3)
          else
           Result := TACRExprNodeBoolean.Create(doBETWEEN, arg1, arg2, arg3);
        end;
      end
    end;
  end;
 if Result = nil then
  begin
   if arg1 <> nil then arg1.Free;
   if arg2 <> nil then arg2.Free;
   if arg3 <> nil then arg3.Free;
  end;
end;//ParseBetWeenPredicate


//------------------------------------------------------------------------------
// parses <in predicate>
//------------------------------------------------------------------------------
function TACRExpression.ParseInPredicate: TACRExprNode;
var
  bNot:         Boolean;
  arg1:         TACRExprNode;
  SavedTokenNo: Integer;
begin
{
<in predicate> ::=
    <row value constructor>
      [ NOT ] IN <in predicate value>
}
  Result := nil;
  arg1 := ParseRowValueConstructor;
  if arg1 <> nil then
   begin
    // [NOT]
    if (IsReservedWord(Token, rwNOT)) then
     begin
      bNot := True;
      // and Get next token
      GetNextToken;
     end
    else
     bNot := False;

    // IN
    if (IsReservedWord(Token, rwIN)) then
     begin
      // get next token
      GetNextToken;

      SaveState(SavedTokenNo);
      // IN (SELECT ...)
      Result := ParseSubQuery(False,True,bNot,arg1);
      // '(' ?
      if (Result = nil) then
       begin
        RestoreState(SavedTokenNo);
        // '(' ?
        if (Token.TokenType = tktLeftParenthesis) then
         begin
          // get next token
          GetNextToken;
          if bNot then
            Result := TACRExprNodeBoolean.Create(doNOTIN, arg1)
          else
            Result := TACRExprNodeBoolean.Create(doIN, arg1);
          repeat
           arg1 := ParseValueExpression;
           // if error
           if arg1 = nil then
            begin
             Result.Free;
             Result := nil;
            end;
           // add argument to IN node
           Result.Children.add(arg1);
           arg1 := nil;
           // ',' ?
           if Token.TokenType = tktComma then
            // get next token
            GetNextToken
           else
           // ')'
           if Token.TokenType = tktRightParenthesis then
            break;
          until false;
          // get next token
          GetNextToken;
         end;
       end; // not a Subquery
     end;

    if (Result = nil) and (arg1 <> nil) then
     begin
      arg1.Free;
     end;
   end;
end;//ParseInPredicate


//------------------------------------------------------------------------------
// parses <Exists predicate>
//------------------------------------------------------------------------------
function TACRExpression.ParseExistsPredicate: TACRExprNode;
var
  bNot: boolean;
begin
{
<Exists predicate> ::=
      [ NOT ] EXISTS (<subquery>)
}
  Result := nil;
  // [NOT]
  if (IsReservedWord(Token, rwNOT)) then
   begin
    bNot := True;
    // and Get next token
    GetNextToken;
   end
  else
   bNot := False;
  if (IsReservedWord(Token, rwEXISTS)) then
   begin
    if (GetNextToken) then
      Result := ParseSubQuery(True,False,bNot);
   end;
end; // ParseExistsPredicate


//------------------------------------------------------------------------------
// parses <true/false>
//------------------------------------------------------------------------------
function TACRExpression.ParseTrueFalseConst: TACRExprNode;
begin
 Result := nil;
 if Token.ReservedWord in [rwTRUE, rwFALSE] then
   begin
     Result := TACRExprNodeConst.Create;
     Result.Value.AsBoolean := (Token.ReservedWord = rwTRUE);
     GetNextToken;
   end
 else if (Token.TokenType = tktParameter) then
   begin
     if (Token.ParamValue.DataType = bftLogical) then
       begin
         Result := TACRExprNodeConst.Create;
         Result.Value.Assign(Token.ParamValue);
         GetNextToken;
       end;
   end
end;//ParseTrueFalseConst


//------------------------------------------------------------------------------
// parses NULL const
//------------------------------------------------------------------------------
function TACRExpression.ParseNullConst: TACRExprNode;
begin
 if Token.ReservedWord = rwNULL then
   begin
     Result := TACRExprNodeConst.Create;
     Result.Value.DataType := bftSignedInt32;
     GetNextToken;
   end
 else
   Result := nil;
end;//ParseNullConst


//------------------------------------------------------------------------------
// parses <comparison predicate> | <between predicate> | <in predicate>  |
// <like predicate>  | <null predicate> | <quantified comparison predicate> |
// <exists predicate> | <unique predicate> | <match predicate> | <overlaps predicate>
//------------------------------------------------------------------------------
function TACRExpression.ParsePredicate: TACRExprNode;
var
  SavedTokenNo: integer;        // saved/restored information
  //SavedRootNode: TACRExprNode; // saved/restored information
begin
 // save current state
 SaveState(SavedTokenNo);
 Result := ParseTrueFalseConst;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    // try to parse next tokens as a comparison predicate
    Result := ParseComparisonPredicate;
  end;
 // if failed comparison - try like
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    // try to parse next tokens as a like predicate
    Result := ParseLikePredicate;
  end;
 // if failed like - try null
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    // try to parse next tokens as a null predicate
    Result := ParseNullPredicate;
  end;
 // if failed null - try between
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    // try to parse next tokens as a between predicate
    Result := ParseBetweenPredicate;
  end;
 // if failed between - try in
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    // try to parse next tokens as a IN predicate
    Result := ParseInPredicate;
  end;
 // if failed in - try Exists
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    // try to parse next tokens as a IN predicate
    Result := ParseExistsPredicate;
  end;
{ TODO : 
add boolean column parsing here!!!
otherwise 
SELECT * from table WHERE BooleanField
does not work!!! }  
{
 // check for boolean column reference
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseColumnReference;
    if (Result <> nil) then
     begin
      if (Result.GetDataType <> aftBoolean) then
       begin
        Result.Free;
        Result := nil;
        RestoreState(SavedTokenNo);
       end;
     end;
  end;
}
end;// ParsePredicate


//------------------------------------------------------------------------------
// parses <predicate>  | <left paren> <search condition> <right paren>
//------------------------------------------------------------------------------
function TACRExpression.ParseBooleanPrimary: TACRExprNode;
begin
 // '(' ?
 if (Token.TokenType = tktLeftParenthesis) then
  begin
   // skip '('
   GetNextToken;
   // parse <search condition>
   Result := ParseSearchCondition;
   // skip ')'
   GetNextToken;
  end
 else
  Result := ParsePredicate;
end;// ParseBooleanPrimary


//------------------------------------------------------------------------------
// parses <boolean primary> [ IS [ NOT ] <truth value> ]
//------------------------------------------------------------------------------
function TACRExpression.ParseBooleanTest: TACRExprNode;
var
  bNot: Boolean;
  bIsNull: Boolean;
  bIsNotNull: Boolean;
  LeftNode: TACRExprNode;
begin
  // parse <boolean primary>
  LeftNode := ParseBooleanPrimary;
  if (LeftNode = nil) then
    raise EACRException.Create(30066, ErrorGBooleanExpressionExpected,
                               [Token.Text, Token.LineNum, Token.ColumnNum]);

  bNot := False;
  bIsNull := False;
  bIsNotNull := False;
  // [ IS [ NOT ] <truth value> ] ?
  if (IsReservedWord(Token, rwIS)) then
    begin
      // skip 'IS' token
      GetNextToken([tktReservedWord]);
      if (IsReservedWord(Token, rwNOT)) then
       begin
        bNot := True;
        // skip 'NOT' token
        GetNextToken([tktReservedWord]);
       end;
      // TRUE | FALSE | UNKNOWN ?
      if (IsReservedWord(Token, rwTRUE)) then
        // skip 'TRUE' token
        GetNextToken
      else
      if (IsReservedWord(Token, rwFALSE)) then
       begin
        bNot := not bNot;
        // skip 'FALSE' token
        GetNextToken
       end
      else
      if (IsReservedWord(Token, rwUNKNOWN)) then
       begin
        if (bNot) then
          bIsNotNull := True
        else
          bIsNull := True;
        bNot := False;
        // skip 'UNKNOWN' token
        GetNextToken
       end
      else
       raise EACRException.Create(30067, ErrorGOtherTokenExpected,
         ['TRUE, FALSE or UNKNOWN', Token.Text, Token.LineNum, Token.ColumnNum]);
    end;

  // create node?
  if (bNot) then
   Result := TACRExprNodeBoolean.Create(doNOT, LeftNode)
  else
  if (bIsNull) then
   Result := TACRExprNodeBoolean.Create(doISNULL, LeftNode)
  else
  if (bIsNotNull) then
   Result := TACRExprNodeBoolean.Create(doISNOTNULL, LeftNode)
  else
   Result := LeftNode;
end;// ParseBooleanTest


//------------------------------------------------------------------------------
// parses [ NOT ] <boolean test>
//------------------------------------------------------------------------------
function TACRExpression.ParseBooleanFactor: TACRExprNode;
var
  bNot: Boolean;
  LeftNode: TACRExprNode;
begin
  Result := ParseExistsPredicate;
  if (Result = nil) then
   begin
    // [NOT]?
    if (IsReservedWord(Token, rwNOT)) then
      begin
        // skip 'NOT' token
        GetNextToken;
        bNot := True;
      end
    else
      bNot := False;
    // parse <boolean test>
    LeftNode := ParseBooleanTest;
    if (LeftNode = nil) then
      raise EACRException.Create(30068, ErrorGBooleanExpressionExpected,
                                 [Token.Text, Token.LineNum, Token.ColumnNum]);
    // create <NOT> node?
    if (bNot) then
     Result := TACRExprNodeBoolean.Create(doNOT, LeftNode)
    else
     Result := LeftNode;
   end; // not EXISTS
end;// ParseBooleanFactor


//------------------------------------------------------------------------------
// parses <boolean factor> | <boolean term> AND <boolean factor>
//------------------------------------------------------------------------------
function TACRExpression.ParseBooleanTerm: TACRExprNode;
var
  Node: TACRExprNode;
  AndOperatorNode: TACRExprNode;
begin
 AndOperatorNode := nil;
 // parse <boolean factor>
 Node := ParseBooleanFactor;
 if (Node <> nil) then
  begin
   // { AND <boolean term> }
   if IsReservedWord(Token, rwAND) then
    begin
     AndOperatorNode := TACRExprNodeBoolean.Create(doAND, Node);
     while (IsReservedWord(Token, rwAND)) do
      begin
        // skip 'AND' token
        GetNextToken;
        // parse <boolean factor>
        Node := ParseBooleanFactor;
        if (Node = nil) then
         raise EACRException.Create(30069, ErrorGBooleanExpressionExpected,
                                  [Token.Text, Token.LineNum, Token.ColumnNum]);
        // add into <AND> node
        AndOperatorNode.Children.Add(Node);
      end;
    end
   else
    AndOperatorNode := Node;
  end;
 Result := AndOperatorNode;
end;// ParseBooleanTerm


//------------------------------------------------------------------------------
// parses <boolean term> | <search condition> OR <boolean term>
//------------------------------------------------------------------------------
function TACRExpression.ParseSearchCondition: TACRExprNode;
var
  LeftNode, RightNode: TACRExprNode;
begin
 // <boolean term>
 LeftNode := ParseBooleanTerm;
 if (LeftNode <> nil) then
  begin
   // { OR <boolean term> }
   while (IsReservedWord(Token, rwOR)) do
    begin
      // skip 'OR' token
      GetNextToken;
      // parse <boolean term>
      RightNode := ParseBooleanTerm;
      if (RightNode = nil) then
       raise EACRException.Create(30070, ErrorGBooleanExpressionExpected,
                                  [Token.Text, Token.LineNum, Token.ColumnNum]);
      // create <OR> node
      LeftNode := TACRExprNodeBoolean.Create(doOR, LeftNode, RightNode);
    end;
  end;
 Result := LeftNode;
end;// ParseSearchCondition







//------------------------------------------------------------------------------
// parses <value expression>
//------------------------------------------------------------------------------
function TACRExpression.ParseValueExpression: TACRExprNode;
var
 SavedTokenNo: integer;        // saved/restored information
begin
{
<value expression> ::=
      <numeric value expression>
    | <string value expression>
    | <datetime value expression>
    | <interval value expression>
}
 SaveState(SavedTokenNo);
 Result := ParseNumericValueExpression;
 {if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseSearchCondition;
  end;}
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseStringValueExpression;
  end;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseBLOBValueFunction;
  end;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseDatetimeValueExpression;
  end;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseTrueFalseConst;
  end;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseNullConst;
  end;
end;// ParseValueExpression

//------------------------------------------------------------------------------
// parses <numeric value expression>
//------------------------------------------------------------------------------
function TACRExpression.ParseNumericValueExpression: TACRExprNode;
var
  Term: TACRExprNode;
begin
{
<numeric value expression> ::=
      <term>
    | <numeric value expression> <plus sign> <term>
    | <numeric value expression> <minus sign> <term>
}
  Result:=nil;
  Term := ParseTerm;
  if Term <> nil then
   begin
    if (Token.Text='+') or (Token.Text='-') then
     begin
      Result := TACRExprNodeArithmetic.Create(doADD, Term);
      while (Token.Text='+') or (Token.Text='-') do
       begin
         Term := ParseTerm;
         if Term = nil then
          begin
           Result.Free;
           Result := nil;
           break;
          end;
         Result.Children.Add(Term);
       end;
     end
    else
     Result := Term;
    if Token.Text='||' then
     begin
      Result.Free;
      Result := nil;
     end;
   end
end;//ParseNumericValueExpression

//------------------------------------------------------------------------------
// parses <term>
//------------------------------------------------------------------------------
function TACRExpression.ParseTerm: TACRExprNode;
var
  Factor: TACRExprNode;
  op: TACRDataOperator;
begin
{
<term> ::=
      <factor>
    | <term> <asterisk> <factor>
    | <term> <solidus> <factor>
}
  Result:=ParseFactor;
  if Result <> nil then
    while (Token.Text='*') or (Token.Text='/') do
     begin
      if Token.Text='*' then
        op := doMUL
      else
        op := doDIV;
      // get next token
      GetNextToken;
      Result := TACRExprNodeArithmetic.Create(op, Result);
      Factor := ParseFactor;
//!!      if Factor = nil then ;
      Result.Children.Add(Factor);
     end;
end;//ParseTerm

//------------------------------------------------------------------------------
// parses <factor>
//------------------------------------------------------------------------------
function TACRExpression.ParseFactor: TACRExprNode;
var
  minus: boolean;
begin
{
<factor> ::=
    [ <sign> ] <numeric primary>
}
  minus := false;
  if ParseSign = sgnMinus then minus := true;
  Result:=ParseNumericPrimary;
  if (result <> nil) and minus then
    Result := TACRExprNodeArithmetic.Create(doSUB, Result);

end;//ParseFactor

//------------------------------------------------------------------------------
// parses <sign>
//------------------------------------------------------------------------------
function TACRExpression.ParseSign: TSign;
begin
  if Token.Text = '+' then Result := sgnPlus
  else
  if Token.Text = '-' then Result := sgnMinus
  else Result := sgnUndefined;
  if Result <> sgnUndefined then GetNextToken;
end;//ParseSign

//------------------------------------------------------------------------------
// parses <numeric primary>
//------------------------------------------------------------------------------
function TACRExpression.ParseNumericPrimary: TACRExprNode;
var
  SavedTokenNo: integer;        // saved/restored information
begin
{
<numeric primary> ::=
      <value expression primary>
    | <numeric value function>
}
  Result := ParseValueExpressionPrimary;
  // save current state
  SaveState(SavedTokenNo);
  if Result = nil then
   begin
    RestoreState(SavedTokenNo);
    Result := ParseNumericValueFunction;
   end;
  // Check for numeric content condition
  if Result <> nil then
   begin
    if Result.ClassType = TACRExprNodeConst then
     if (not Result.Value.IsNumericDataType) then
      begin
       RestoreState(SavedTokenNo);
       Result.Free;
       Result := nil;
      end;
   end;

end;//ParseNumericPrimary

//------------------------------------------------------------------------------
// parses <value expression primary>
//------------------------------------------------------------------------------
function TACRExpression.ParseValueExpressionPrimary: TACRExprNode;
var
  SavedTokenNo: integer;        // saved/restored information
begin
{
<value expression primary> ::=
      <unsigned value specification>
    | <column reference>
    | <set function specification>
    | <scalar subquery>
    | <case expression>
    | <left paren> <value expression> <right paren>
    | <cast specification>
}
  SaveState(SavedTokenNo);
  // parses <unsigned value specification>
  Result := ParseUnsignedValueSpecification;
  // parses <column reference>
  if Result=nil then
   begin
    RestoreState(SavedTokenNo);
    Result := ParseColumnReference
   end;
  // parses <SubQuery>
  if Result = nil then
   begin
    RestoreState(SavedTokenNo);
    Result := ParseSubQuery;
   end;
  // parses <set function specification>
  if Result=nil then
   begin
    RestoreState(SavedTokenNo);
    Result := ParseSetFunctionSpecification;
   end;
  // parses ( <value expression> )
  if Result=nil then
   begin
    RestoreState(SavedTokenNo);
    // '('
    if Token.TokenType = tktLeftParenthesis then
     begin
      // get next token
      GetNextToken;
      Result := ParseValueExpression;
      if (Result<>nil) and
         (Token.TokenType <> tktRightParenthesis) then
       begin
        Result.Free;
        raise EACRException.Create(30071, ErrorGOtherTokenExpected,
                              [')', Token.Text, Token.LineNum,Token.ColumnNum]);
       end;
      // get next token
      GetNextToken;
     end;
   end;
  // parses <cast specification>
  if Result = nil then
   begin
    RestoreState(SavedTokenNo);
    Result := ParseCastSpecification;
   end;
end;//ParseValueExpressionPrimary

//------------------------------------------------------------------------------
// parses <unsigned value specification>
//------------------------------------------------------------------------------
function TACRExpression.ParseUnsignedValueSpecification: TACRExprNode;
begin
{
<unsigned value specification> ::=
      <unsigned literal>
    | <general value specification>
}
  Result := ParseUnsignedLiteral;
end;//ParseUnsignedValueSpecification

//------------------------------------------------------------------------------
// parses <unsigned literal>
//------------------------------------------------------------------------------
function TACRExpression.ParseUnsignedLiteral: TACRExprNode;
var
  SavedTokenNo: integer;        // saved/restored information
begin
{
<unsigned literal> ::=
      <unsigned numeric literal>
    | <general literal>
}
  // save current state
  SaveState(SavedTokenNo);
  Result := ParseUnsignedNumericLiteral;
  if Result = nil then
   begin
    RestoreState(SavedTokenNo);
    Result := ParseGeneralLiteral;
   end;
end;//ParseUnsignedLiteral


//------------------------------------------------------------------------------
// parses <unsigned numeric literal>
//------------------------------------------------------------------------------
function TACRExpression.ParseUnsignedNumericLiteral: TACRExprNode;
begin
{
<unsigned numeric literal> ::=
      <exact numeric literal>
    | <approximate numeric literal>
}
 Result := nil;
 // integer
 if (Token.TokenType = tktInt) then
  begin
   Result := TACRExprNodeConst.Create;
   {$IFDEF D6H}
   Result.Value.DataType := bftSignedInt64;
   {$ELSE}
   Result.Value.DataType := bftSignedInt32;
   {$ENDIF}
   if Token.Text <> '' then
     Result.Value.AsString := Token.Text;
   // get next token
   GetNextToken;
  end
 else
 // float
 if (Token.TokenType = tktFloat) then
  begin
   Result := TACRExprNodeConst.Create;
   Result.Value.DataType := bftDouble;
   Result.Value.AsString := Token.Text;
   // get next token
   GetNextToken;
  end
 else
 // Parameter
 if (Token.TokenType = tktParameter) then
  begin
   if (Token.ParamValue.IsNull) then
    begin
     Result := TACRExprNodeConst.Create;
     Result.Value.SetNull(Token.ParamValue.DataType);
     // get next token
     GetNextToken;
    end
   else
   if (Token.ParamValue.IsNumericDataType) then
    begin
     Result := TACRExprNodeConst.Create;
     Result.Value.Assign(Token.ParamValue);
     // get next token
     GetNextToken;
    end
  end // param
 else
  Result := nil;
end;//ParseUnsignedNumericLiteral


//------------------------------------------------------------------------------
// parses <column reference>
//------------------------------------------------------------------------------
function TACRExpression.ParseColumnReference: TACRExprNode;
var
  FieldName, TableName: string;
  IsResWord: boolean;
begin
 Result := nil;
 //InitDataValue(Value);
 if ((Token.TokenType = tktString) or
     (Token.TokenType = tktBracketedString) or
     (Token.TokenType = tktReservedWord)) then
  begin
   //SavedTokenNo := FLex.GetCurrentTokenNo;
   FieldName := Token.Text;
   TableName := '';
   IsResWord := Token.ReservedWord <> rwNone;
   // get next token
   GetNextToken;
   // table.field?
   if (Token.TokenType = tktDot) then
    begin
     TableName := FieldName;
     GetNextToken;
     FieldName := Token.Text;
     IsResWord := Token.ReservedWord <> rwNone;
     // get next token
     GetNextToken;
    end;
   // function ?
   if (Token.TokenType <> tktLeftParenthesis) and (not IsResWord) then
     Result := TACRExprNodeField.Create(LCursor, FieldName, TableName);
  end;
 //FinalizeDataValue(Value);
end;//ParseColumnReference


//------------------------------------------------------------------------------
// parses <SUB QUERY>
//------------------------------------------------------------------------------
function TACRExpression.ParseSubQuery(bExists:       Boolean = False;
                                       bIn:           Boolean = False;
                                       bNot:          Boolean = False;
                                       SourceNode:    TACRExprNode = nil
                                       ): TACRExprNode;
var SubQueryText:           String;
    LeftParenthesisCount:   Integer;
    bSkipSpace:             Boolean;
    
begin
  Result := nil;
  if (Token.TokenType = tktLeftParenthesis) then
   begin
    LeftParenthesisCount := 1;
    if (GetNextToken) then
      if (Token.TokenType = tktReservedWord) then
       if (Token.ReservedWord = rwSELECT) then
        begin
         SubQueryText := Token.Text;
         bSkipSpace := False;
         while (GetNextToken) do
          begin
            if (Token.TokenType = tktLeftParenthesis) then
             Inc(LeftParenthesisCount)
            else
            if (Token.TokenType = tktRightParenthesis) then
             Dec(LeftParenthesisCount);
            if (LeftParenthesisCount = 0) then
             begin
              GetNextToken;
              break;
             end;
            if (Token.TokenType = tktBracketedString) then
             begin
              SubQueryText := SubQueryText + Space + '[' + Token.Text + ']';
              bSkipSpace := False;
             end
            else
            if (Token.TokenType = tktQuotedString) then
             begin
              SubQueryText := SubQueryText + Space + SingleQuote + Token.Text + SingleQuote;
              bSkipSpace := False;
             end
            else
            if (Token.TokenType = tktDot) then
             begin
              SubQueryText := SubQueryText + Token.Text;
              bSkipSpace := True;
             end
            else
            if (Token.TokenType = tktParameter) then
             begin
              SubQueryText := SubQueryText + Space + Colon + Token.Text;
              bSkipSpace := False;
             end
            else
             begin
              if (bSkipSpace) then
               SubQueryText := SubQueryText + Token.Text
              else
               SubQueryText := SubQueryText + Space + Token.Text;
              bSkipSpace := False;
             end;
          end; // building SubQuery text
         if (LeftParenthesisCount = 0) then
          begin
           if (bExists) then
            Result := TACRExprNodeSubQueryExists.Create(SubQueryText,FParams,FInMemory,
                        FDatabaseName,FSessionName, bNot,SourceNode)
           else
           if (bIn) then
            Result := TACRExprNodeSubQueryIn.Create(SubQueryText,FParams,FInMemory,
                        FDatabaseName,FSessionName, bNot,SourceNode)
           else
            Result := TACRExprNodeSubQuery.Create(SubQueryText,FParams,FInMemory,
                        FDatabaseName,FSessionName, bNot,SourceNode);
          end; // while GetNextToken
        end; // SubQuery
   end; // Left Parenthesis
end; // ParseSubQuery


//------------------------------------------------------------------------------
// parses <set function specification>
//------------------------------------------------------------------------------
function TACRExpression.ParseSetFunctionSpecification: TACRExprNode;
var
  SavedTokenNo: integer;        // saved/restored information
begin
{
<set function specification> ::=
      COUNT <left paren> <asterisk> <right paren>
    | <general set function>
}
 Result := nil;
 SaveState(SavedTokenNo);
 if IsReservedWord(Token, rwCOUNT) then
  begin
   // get next token
   GetNextToken;
   // '('
   if Token.TokenType = tktLeftParenthesis then
    begin
     // get next token
     GetNextToken;
     // '*'
     if Token.Text='*' then
      begin
       // get next token
       GetNextToken;
       // ')'
       if Token.TokenType <> tktRightParenthesis then
       raise EACRException.Create(30073, ErrorGOtherTokenExpected,
                             [')', Token.Text, Token.LineNum, Token.ColumnNum]);
       // get next token
       GetNextToken;
       Result := TACRExprNodeAggregated.Create(doCOUNTALL);
      end;
    end;
  end;
 if Result = nil then
  begin
   RestoreState(SavedTokenNo);
   Result := ParseGeneralSetFunction;
  end;
end;//ParseSetFunctionSpecification


//------------------------------------------------------------------------------
// parses <general set function>
//------------------------------------------------------------------------------
function TACRExpression.ParseGeneralSetFunction: TACRExprNode;
var
  SavedTokenNo: integer;        // saved/restored information
  expr: TACRExprNode;
  distinct: boolean;
  oper: TACRDataOperator;
begin
{ 
<general set function> ::= 
      <set function type>
          <left paren> [ <set quantifier> ] <value expression> <right paren>
} 
 Result := nil; 
 SaveState(SavedTokenNo);
	
 oper := doUNDEFINED; 
 if IsReservedWord(Token, rwCOUNT) then
  oper:=doCOUNT 
 else if IsReservedWord(Token, rwAVG) then
  oper:=doAVG 
 else if IsReservedWord(Token, rwMAX) then
  oper:=doMAX 
 else if IsReservedWord(Token, rwMIN) then
  oper:=doMIN 
 else if IsReservedWord(Token, rwSUM) then 
  oper:=doSUM;
 
 if oper = doUNDEFINED then Exit; 
 
 // get next token 
 GetNextToken; 
	
 // DISTINCT
 distinct := IsReservedWord(Token, rwDISTINCT);
 if distinct then GetNextToken;
	
 // '(' 
 if Token.TokenType <> tktLeftParenthesis then
   raise EACRException.Create(30074, ErrorGOtherTokenExpected, 
                            ['(', Token.Text, Token.LineNum, Token.ColumnNum]);
 // get next token 
 GetNextToken; 
	
 if (IsReservedWord(Token, rwDISTINCT)) then 
  begin
   distinct := True; 
   GetNextToken; 
  end;
	
 // EXPRESSION 
 expr := ParseValueExpression; 
 if expr = nil then 
   raise EACRException.Create(30102, ErrorGArgumentExpected, 
            [GetOperatorName(oper), Token.LineNum, Token.ColumnNum]); 
 
 // ')' 
 if Token.TokenType <> tktRightParenthesis then
  begin
   expr.Free; 
   raise EACRException.Create(30075, ErrorGOtherTokenExpected,
                            [')', Token.Text, Token.LineNum, Token.ColumnNum]);
  end; 
 // get next token
 GetNextToken;
 
 // Return GroupFunctionNode 
 Result := TACRExprNodeAggregated.Create(oper, distinct, expr); 
end;//ParseGeneralSetFunction
	
 
	
//------------------------------------------------------------------------------
// parses <cast specification> 
//------------------------------------------------------------------------------ 
function TACRExpression.ParseCastSpecification: TACRExprNode;
var
  Operand: TACRExprNode;
  fType: TACRAdvancedFieldType; 
begin 
{ 
<cast specification> ::= 
    CAST <left paren> <cast operand> [as | , ]
        <cast target> <right paren> 
} 
 //raise Exception.Create('Not supported yet'); 
 Result := nil;
 if IsReservedWord(Token, rwCAST) then
  begin
   // get next token 
   GetNextToken;
   // '('
   if Token.TokenType = tktLeftParenthesis then
    begin 
     // get next token 
     GetNextToken;
     // Operand
     Operand := ParseCastOperand;
     if Operand <> nil then
      begin
       if (Token.TokenType = tktComma) or
          IsReservedWord(Token, rwAS) then 
        begin
         // get next token
         GetNextToken; 
         // Get Field Type
         fType := GetFieldType(Token.Text);
         // If correct fType 
         if IsConvertableFieldType(fType) then 
          begin 
           // get next token 
           GetNextToken;
           // ')' 
           if Token.TokenType <> tktRightParenthesis then 
            begin
              Operand.Free;
              raise EACRException.Create(30076, ErrorGOtherTokenExpected,
                              [')', Token.Text,Token.LineNum,Token.ColumnNum]);
            end; 
           Result := TACRExprNodeCast.Create(Operand, fType); 
           // get next token
           GetNextToken;
          end 
         else 
          begin 
           Operand.Free;
           raise EACRException.Create(30077, ErrorGNotApplicableCastType,
                                      [AftToStr(fType)]);
          end;
        end;
      end; 
    end;
  end; 
end;//ParseCastSpecification


//------------------------------------------------------------------------------
// parses <cast operand>
//------------------------------------------------------------------------------
function TACRExpression.ParseCastOperand: TACRExprNode;
var
  SavedTokenNo: integer;
begin
{
<cast operand> ::=
      <value expression>
    | NULL
}
 // NULL
 SaveState(SavedTokenNo);

 Result := ParseNullConst;
 GetNextToken;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseValueExpression;
  end;
end;//ParseCastOperand


//------------------------------------------------------------------------------
// parses <numeric value function>
//------------------------------------------------------------------------------
function TACRExpression.ParseNumericValueFunction: TACRExprNode;
var
  SavedTokenNo: integer;        // saved/restored information
begin
{
<numeric value function> ::=
      <position expression>
    | <extract expression>
    | <length expression>
    | <LastAutoInc expression>
}
 // save current state
 SaveState(SavedTokenNo);
 Result := ParsePositionExpression;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseLengthExpression;
  end;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseLastAutoIncExpression;
  end;
end;//ParseNumericValueFunction


//------------------------------------------------------------------------------
// parses <position expression>
//------------------------------------------------------------------------------
function TACRExpression.ParsePositionExpression: TACRExprNode;
var
  arg1, arg2: TACRExprNode;
begin
 if IsReservedWord(Token, rwPOSITION) or IsReservedWord(Token, rwPOS) then
  begin
    // get next token
    GetNextToken;

    // '('
    if Token.TokenType <> tktLeftParenthesis then
     raise EACRException.Create(30078, ErrorGOtherTokenExpected,
                             ['(', Token.Text, Token.LineNum, Token.ColumnNum]);
    // get next token
    GetNextToken;

    // arg
    arg1 := ParseCharacterValueExpression;
    if arg1 = nil then
     raise EACRException.Create(30079, ErrorGArgumentExpected,
                                       ['POS', Token.LineNum, Token.ColumnNum]);

    // ',' or 'in'
    if (Token.TokenType <> tktComma) and not IsReservedWord(Token, rwIN) then
     begin
       arg1.Free;
       raise EACRException.Create(30080, ErrorGOtherTokenExpected,
                [','' or ''IN', Token.Text, Token.LineNum, Token.ColumnNum]);
     end;
    // get next token
    GetNextToken;

    arg2 := ParseCharacterValueExpression;
    if arg2 = nil then
     begin
      arg1.Free;
      raise EACRException.Create(30081, ErrorGArgumentExpected,
                                       ['POS', Token.LineNum, Token.ColumnNum]);
     end;

    Result := TACRExprNodeStringFunction.Create(doPOS, arg1, arg2);

    // ')'
    if Token.TokenType <> tktRightParenthesis then
     begin
      Result.Free;
      raise EACRException.Create(30082, ErrorGOtherTokenExpected,
                             [')', Token.Text, Token.LineNum, Token.ColumnNum]);
     end;
    // get next token
    GetNextToken;
  end
 else
  Result := nil;
end;//ParsePositionExpression


//------------------------------------------------------------------------------
// parses <length expression>
//------------------------------------------------------------------------------
function TACRExpression.ParseLengthExpression: TACRExprNode;
begin
 if IsReservedWord(Token, rwLENGTH) then
  begin
    // get next token
    GetNextToken;

    // '('
    if Token.TokenType <> tktLeftParenthesis then
     raise EACRException.Create(30083, ErrorGOtherTokenExpected,
                             ['(', Token.Text, Token.LineNum, Token.ColumnNum]);
    // get next token
    GetNextToken;

    // argument
    Result := ParseStringValueExpression;
    if result = nil then
      raise EACRException.Create(30085, ErrorGArgumentExpected,
                                     ['LENGTH', Token.LineNum, Token.ColumnNum]);

    Result := TACRExprNodeStringFunction.Create(doLENGTH, Result);

    // ')'
    if Token.TokenType <> tktRightParenthesis then
     begin
      Result.Free;
      raise EACRException.Create(30084, ErrorGOtherTokenExpected,
                             [')', Token.Text, Token.LineNum, Token.ColumnNum]);
     end;

    // get next token
    GetNextToken;
  end
 else
  Result := nil;
end;//ParseLengthExpression


//------------------------------------------------------------------------------
// parses <LastAutoInc expression>
//------------------------------------------------------------------------------
function TACRExpression.ParseLastAutoIncExpression: TACRExprNode;
var
  TableName, ColumnName: String;
  MemoryWord: Boolean;
begin

 //  LastAutoinc(TableName, ColumnName)

 if IsReservedWord(Token, rwLAST_AUTOINC) then
  begin
    TableName := '';
    ColumnName := '';
    MemoryWord := False;

    // get next token
    GetNextToken;

    // '('
    if (Token.TokenType <> tktLeftParenthesis) then
      raise EACRException.Create(30086, ErrorGOtherTokenExpected,
                             ['(', Token.Text, Token.LineNum, Token.ColumnNum]);
      // table_name
    GetNextToken;
    // MEMORY
    if (IsReservedWord(Token, rwMEMORY)) then
     begin
       MemoryWord := True;
      GetNextToken;
     end;

    TableName := Token.Text;
    GetNextToken;


    // ','
    if (Token.TokenType <> tktComma) then
      raise EACRException.Create(30334, ErrorGOtherTokenExpected,
                           [',', Token.Text, Token.LineNum, Token.ColumnNum]);


    // column_name
        GetNextToken;
    ColumnName := Token.Text;

      // ')'
    GetNextToken;
      if Token.TokenType <> tktRightParenthesis then
        raise EACRException.Create(30089, ErrorGOtherTokenExpected,
                             [')', Token.Text, Token.LineNum, Token.ColumnNum]);


    // get next token
    GetNextToken;

    Result := TACRExprNodeSystem.Create(doLASTAUTOINC, TableName,
                                        ColumnName, (InMemory or MemoryWord),
                                        FSessionName,
                                        FDatabaseName);
  end
 else
  Result := nil;
end;//ParseLastAutoIncExpression


//------------------------------------------------------------------------------
// parses <string value expression>
//------------------------------------------------------------------------------
function TACRExpression.ParseStringValueExpression: TACRExprNode;
begin
{
<string value expression> ::=
      <character value expression>
    | <bit value expression>
}
  Result := ParseCharacterValueExpression;
end;//ParseStringValueExpression


//------------------------------------------------------------------------------
// parses <character value expression>
//------------------------------------------------------------------------------
function TACRExpression.ParseCharacterValueExpression: TACRExprNode;
var
  SavedTokenNo: integer;        // saved/restored information
begin
{
<character value expression> ::=
      <concatenation>
    | <character factor>
}
  // save current state
  SaveState(SavedTokenNo);
  Result := ParseConcatination;
  if Result = nil then
   begin
    RestoreState(SavedTokenNo);
    Result := ParseCharacterFactor;
   end;
end;//ParseCharacterValueExpression


//------------------------------------------------------------------------------
// parses <concatenation>
//------------------------------------------------------------------------------
function TACRExpression.ParseConcatination: TACRExprNode;
var
  left, right: TACRExprNode;
begin
{
<concatenation> ::=
    <character value expression> <concatenation operator>
        <character factor>
}
  Result := nil;
  left := ParseCharacterFactor;
  //left := ParseCharacterValueExpression;
  if left <> nil then
   if (Token.Text = '||') or (Token.Text = '+') then
    begin
     Result := TACRExprNodeStringFunction.Create(doConcat, left);
     while (Token.Text = '||') or (Token.Text = '+') do
      begin
       GetNextToken;
       right := ParseCharacterFactor;
       if right <> nil then
        Result.Children.add(right);
      end;
    end
   else
    left.Free;
end;//ParseConcatination


//------------------------------------------------------------------------------
// parses <character factor>
//------------------------------------------------------------------------------
function TACRExpression.ParseCharacterFactor: TACRExprNode;
begin
{
<character factor> ::=
    <character primary> [ <collate clause> ]
}
  Result := ParseCharacterPrimary;
end;//ParseCharacterFactor


//------------------------------------------------------------------------------
// parses <character primary>
//------------------------------------------------------------------------------
function TACRExpression.ParseCharacterPrimary: TACRExprNode;
begin
{
<character primary> ::=
      <value expression primary>
    | <string value function>
}
  Result := ParseValueExpressionPrimary;
  if Result = nil then
   begin
    Result := ParseStringValueFunction;
   end;
end;//ParseCharacterPrimary


//------------------------------------------------------------------------------
// parses <general literal>
//------------------------------------------------------------------------------
function TACRExpression.ParseGeneralLiteral: TACRExprNode;
begin
{
<general literal> ::=
      <character string literal>
    | <national character string literal>
    | <bit string literal>
    | <hex string literal>
    | <datetime literal>
    | <interval literal>
}
  Result := ParseCharacterStringLiteral;
end;//ParseGeneralLiteral


//------------------------------------------------------------------------------
// parses <character string literal>
//------------------------------------------------------------------------------
function TACRExpression.ParseCharacterStringLiteral: TACRExprNode;
begin
(*
<character string literal> ::=
    [ <introducer><character set specification> ]
    <quote> [ <character representation>... ] <quote>
      [ { <separator>... <quote> [ <character representation>... ] <quote> }...]
*)
 Result := nil;

 if (Token.TokenType = tktQuotedString) then
  begin
   Result := TACRExprNodeConst.Create;
    Result.Value.AsString := Token.Text;
   // get next token
   GetNextToken;
  end
 else if (Token.TokenType = tktParameter) and
          ((Token.ParamValue.IsStringDataType) or
           (Token.ParamValue.IsBlobDataType)) then
  begin
    Result := TACRExprNodeConst.Create;
    Result.Value.Assign(Token.ParamValue);
    // get next token
    GetNextToken;
  end

end;//ParseCharacterStringLiteral


//------------------------------------------------------------------------------
// parses <string value function>
//------------------------------------------------------------------------------
function TACRExpression.ParseStringValueFunction: TACRExprNode;
begin
{
<string value function> ::=
      <character value function>
    | <bit value function>
}
  Result := ParseCharacterValueFunction;
end;//ParseStringValueFunction


//------------------------------------------------------------------------------
// parses <character value function>
//------------------------------------------------------------------------------
function TACRExpression.ParseCharacterValueFunction: TACRExprNode;
var
  arg1,arg2,arg3: TACRExprNode;
  op: TACRDataOperator;
begin
{
<character value function> ::=
      <character substring function>
    | <fold>
    | <form-of-use conversion>
    | <character translation>
    | <trim function>
}
  Result := nil;
  op := doUNDEFINED;
  if IsReservedWord(Token, rwUPPER) then op:=doUPPER
  else if IsReservedWord(Token, rwLOWER) then op:=doLOWER
  else if IsReservedWord(Token, rwTRIM) then op:=doTRIM
  else if IsReservedWord(Token, rwLTRIM) then op:=doLTRIM
  else if IsReservedWord(Token, rwRTRIM) then op:=doRTRIM
  else if IsReservedWord(Token, rwSUBSTRING) then op:=doSUBSTRING
  else if IsReservedWord(Token, rwTOSTRING) then op:=doTOSTRING
  ;
  if op = doUNDEFINED then Exit;
  // get next token
  GetNextToken;

  // Functions with 1 agrument
  if op in [doUPPER, doLOWER, doTRIM, doLTRIM, doRTRIM] then
   begin
    // '('
    if Token.TokenType <> tktLeftParenthesis then
      raise EACRException.Create(30090, ErrorGOtherTokenExpected,
                             ['(', Token.Text, Token.LineNum, Token.ColumnNum]);
    // get next token
    GetNextToken;

    // arg1
    arg1 := ParseCharacterValueExpression;

    // ')'
    if Token.TokenType <> tktRightParenthesis then
      raise EACRException.Create(30091, ErrorGOtherTokenExpected,
                             [')', Token.Text, Token.LineNum, Token.ColumnNum]);
    // get next token
    GetNextToken;

    if arg1 = nil then
      raise EACRException.Create(30092, ErrorGStringArgumentExpected,
             [GetOperatorName(op), Token.Text, Token.LineNum, Token.ColumnNum]);
    Result := TACRExprNodeStringFunction.Create(op, arg1);
   end
  else
  // Functions with 2 agruments
  if op = doTOSTRING then
   begin
    // '('
    if Token.TokenType <> tktLeftParenthesis then
      raise EACRException.Create(30093, ErrorGOtherTokenExpected,
                             ['(', Token.Text, Token.LineNum, Token.ColumnNum]);
    // get next token
    GetNextToken;

    //TOSTRING(str, format)
    // string
    arg1 := ParseValueExpression;
    if arg1 = nil then
     raise EACRException.Create(30094, ErrorGArgumentExpected,
                        [GetOperatorName(op), Token.LineNum, Token.ColumnNum]);
    // ','
    if (Token.TokenType <> tktComma) and not IsReservedWord(Token, rwFrom) then
     begin
      arg1.Free;
      raise EACRException.Create(30095, ErrorGOtherTokenExpected,
                             [',', Token.Text, Token.LineNum, Token.ColumnNum]);
     end;
    // get next token
    GetNextToken;

    // format
    if Token.TokenType <> tktQuotedString then
     begin
      arg1.Free;
      raise EACRException.Create(30096, ErrorGConstDateFormatExpected,
                             [Token.Text, Token.LineNum, Token.ColumnNum]);
     end;

    Result := TACRExprNodeDateFunction.Create(doTOSTRING, arg1, Token.Text);

    // get next token
    GetNextToken;

    // ')'
    if Token.TokenType <> tktRightParenthesis then
      raise EACRException.Create(30097, ErrorGOtherTokenExpected,
                             [')', Token.Text, Token.LineNum, Token.ColumnNum]);
    // get next token
    GetNextToken;
   end
  else
  // Functions with 3 agruments
  if op = doSUBSTRING then
   begin
    // '('
    if Token.TokenType <> tktLeftParenthesis then
      raise EACRException.Create(30098, ErrorGOtherTokenExpected,
                             ['(', Token.Text, Token.LineNum, Token.ColumnNum]);

    // get next token
    GetNextToken;

    //SUBSTRING(str from|, startindex [for|, length])
    // string
    arg1 := ParseCharacterValueExpression;
    if arg1 = nil then
     raise EACRException.Create(30099, ErrorGArgumentExpected,
                       [GetOperatorName(op), Token.LineNum, Token.ColumnNum]);
    // ','
    if (Token.TokenType <> tktComma) and not IsReservedWord(Token, rwFrom) then
     begin
      arg1.Free;
      raise EACRException.Create(30100, ErrorGOtherTokenExpected,
                            [','' or ''from', Token.LineNum, Token.ColumnNum]);
     end;
    // get next token
    GetNextToken;

    // StartIndex
    arg2 := ParseNumericValueExpression;
    if arg2 = nil then
     begin
      arg1.Free;
     raise EACRException.Create(30101, ErrorGArgumentExpected,
                       [GetOperatorName(op), Token.LineNum, Token.ColumnNum]);
     end;

    // ','
    if (Token.TokenType <> tktComma) and not IsReservedWord(Token, rwFor) and
       (Token.TokenType <> tktRightParenthesis) then
     begin
      arg1.Free;
      arg2.Free;
      raise EACRException.Create(30103, ErrorGOtherTokenExpected,
                     [')'' or '','' or ''FOR', Token.LineNum, Token.ColumnNum]);
     end;
    // lenghth
    if Token.TokenType = tktRightParenthesis then
     arg3 := nil
    else
     begin
      // get next token
      GetNextToken;
      arg3 := ParseNumericValueExpression;
     end;

    // ')'
    if Token.TokenType <> tktRightParenthesis then
      raise EACRException.Create(30104, ErrorGOtherTokenExpected,
                             [')', Token.Text, Token.LineNum, Token.ColumnNum]);
    // get next token
    GetNextToken;
    if arg3 = nil then
      Result := TACRExprNodeStringFunction.Create(op, arg1,arg2)
    else
    Result := TACRExprNodeStringFunction.Create(op, arg1,arg2,arg3);
   end;
end;//ParseCharacterValueFunction


//------------------------------------------------------------------------------
// parses <BLOB value function>
//------------------------------------------------------------------------------
function TACRExpression.ParseBLOBValueFunction: TACRExprNode;
var
  op:              TACRDataOperator;
  BLOBValue:       string;
  BLOBValueFormat: TACRBLOBValueFormat;
begin
{
<BLOB value function> ::=
  TOBLOB ('BLOBValue' [, MIME64 | HEX])
}
  Result := nil;
  op := doUNDEFINED;
  if IsReservedWord(Token, rwTOBLOB) then op:=doTOBLOB;
  if op = doUNDEFINED then Exit;
  // get next token
  GetNextToken;
  case op of
   doTOBLOB:
    begin
      BLOBValueFormat := bvfMIME64;
      if Token.TokenType <> tktLeftParenthesis then
        raise EACRException.Create(11345, ErrorGOtherTokenExpected,
                               ['(', Token.Text, Token.LineNum, Token.ColumnNum]);
      // get next token
      GetNextToken;
      if Token.TokenType <> tktQuotedString then
        raise EACRException.Create(11346, ErrorLConstBLOBValueExpected,
                               [Token.Text, Token.LineNum, Token.ColumnNum]);
      BLOBValue := Token.Text;
      Token.Text := ''; // to avoid unnecessary memory usage
      // get next token
      GetNextToken;
      if Token.TokenType = tktComma then
       begin
        // get next token
        GetNextToken;
        if (IsReservedWord(Token,rwHEX)) then
         BLOBValueFormat := bvfHEX
        else
        if (IsReservedWord(Token,rwMIME64)) then
         BLOBValueFormat := bvfMIME64
        else
         raise EACRException.Create(11347, ErrorLBLOBValueFormatExpected,
                               [Token.Text, Token.LineNum, Token.ColumnNum]);
        // get next token
        GetNextToken;
       end;
      if Token.TokenType <> tktRightParenthesis then
        raise EACRException.Create(11348, ErrorGOtherTokenExpected,
                               [')', Token.Text, Token.LineNum, Token.ColumnNum]);
      Result := TACRExprNodeBLOBFunction.Create(op,BLOBValue,BLOBValueFormat);
    end;
  end;
 // get next token
 GetNextToken;
end; // ParseBLOBValueFunction


//------------------------------------------------------------------------------
// parses <datetime value expression>
//------------------------------------------------------------------------------
function TACRExpression.ParseDatetimeValueExpression: TACRExprNode;
var
 SavedTokenNo: integer;        // saved/restored information
begin
{
 <datetime value expression> ::=
   <datetime CURRENT_TIMESTAMP function>
   | <datetime CURRENT_TIME function>
   | <datetime CURRENT_DATE function>
   | <datetime to_date function>
   | :ParameterName
}

 SaveState(SavedTokenNo);
 Result := ParseSysdateFunction;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseCurrentTimeFunction;
  end;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseCurrentDateFunction;
  end;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    Result := ParseToDateFunction;
  end;
 if (Result = nil) then
  begin
    // restore original state
    RestoreState(SavedTokenNo);
    if (Token.TokenType = tktParameter) then
      begin
        if Token.ParamValue.IsDateTimeDataType then
          begin
            Result := TACRExprNodeConst.Create;
            Result.Value.Assign(Token.ParamValue);
            // get next token
            GetNextToken;
          end;
      end;
  end;
end;//ParseDatetimeValueExpression



//------------------------------------------------------------------------------
// parse <datetime sysdate function>
//------------------------------------------------------------------------------
function TACRExpression.ParseSysdateFunction: TACRExprNode;
begin
 if IsReservedWord(Token, rwSYSDATE) or
    IsReservedWord(Token, rwNOW) or
    IsReservedWord(Token, rwCURRENT_TIMESTAMP) then
  begin
   // get next token
   GetNextToken;
   Result := TACRExprNodeDateFunction.Create(doSYSDATE);
  end
 else
  Result := nil;
end;//ParseSysdateFunction


//------------------------------------------------------------------------------
// parse <datetime CURRENT_TIME function>
//------------------------------------------------------------------------------
function TACRExpression.ParseCurrentTimeFunction: TACRExprNode;
begin
 if IsReservedWord(Token, rwCURRENT_TIME) then
  begin
   // get next token
   GetNextToken;
   Result := TACRExprNodeDateFunction.Create(doCURRENT_TIME);
  end
 else
  Result := nil;
end;//ParseCurrentTimeFunction


//------------------------------------------------------------------------------
// parse <datetime CURRENT_DATE function>
//------------------------------------------------------------------------------
function TACRExpression.ParseCurrentDateFunction: TACRExprNode;
begin
 if IsReservedWord(Token, rwCURRENT_DATE) then
  begin
   // get next token
   GetNextToken;
   Result := TACRExprNodeDateFunction.Create(doCURRENT_DATE);
  end
 else
  Result := nil;
end;//ParseCurrentDateFunction


//------------------------------------------------------------------------------
// parse <datetime TODATE function>
//------------------------------------------------------------------------------
function TACRExpression.ParseToDateFunction: TACRExprNode;
begin
{
   ToDate(string, format)
}
 if IsReservedWord(Token, rwTODATE) then
  begin
   // get next token
   GetNextToken;
   // '('
   if Token.TokenType <> tktLeftParenthesis then
      raise EACRException.Create(30105, ErrorGOtherTokenExpected,
                             ['(', Token.Text, Token.LineNum, Token.ColumnNum]);
   // get next token
   GetNextToken;

   // string
   Result := ParseCharacterValueExpression;
   if Result = nil then
    raise EACRException.Create(30107, ErrorGArgumentExpected,
                                    ['ToDate', Token.LineNum, Token.ColumnNum]);
   // ','
   if Token.TokenType <> tktComma then
    begin
      Result.Free;
      raise EACRException.Create(30108, ErrorGOtherTokenExpected,
                             [',', Token.Text, Token.LineNum, Token.ColumnNum]);
    end;
   // get next token
   GetNextToken;

   // format
   if Token.TokenType <> tktQuotedString then
    begin
     Result.Free;
     raise EACRException.Create(30109, ErrorGConstDateFormatExpected,
                                  [Token.Text, Token.LineNum, Token.ColumnNum]);
    end;

   Result := TACRExprNodeDateFunction.Create(doTODATE, Result, Token.Text);

   // get next token
   GetNextToken;

   // ')'
   if Token.TokenType <> tktRightParenthesis then
    begin
     Result.Free;
      raise EACRException.Create(30106, ErrorGOtherTokenExpected,
                             [')', Token.Text, Token.LineNum, Token.ColumnNum]);
    end;
   // get next token
   GetNextToken;
  end
 else
  Result := nil;
end;//ParseToDateFunction




////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNode
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNode.Create(CaseInsensitive, PartialKey: Boolean);
begin
  Operator := doUNDEFINED;
  Value := TACRVariant.Create;
  Children := TList.Create;
  FCaseInsensitive := CaseInsensitive;
  FPartialKey := PartialKey;
end;//Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNode.Create(Op: TACRDataOperator; CaseInsensitive, PartialKey: Boolean);
begin
  Create(CaseInsensitive, PartialKey);
  Operator := Op;
end;//Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNode.Create(Op: TACRDataOperator; Node: TACRExprNode; CaseInsensitive, PartialKey: Boolean);
begin
  Create(Op, CaseInsensitive, PartialKey);
  Children.Add(Node);
end;//Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNode.Create(Op: TACRDataOperator; Node1, Node2: TACRExprNode; CaseInsensitive, PartialKey: Boolean);
begin
  Create(Op, Node1, CaseInsensitive, PartialKey);
  Children.Add(Node2);
end;//Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNode.Create(Op: TACRDataOperator; Node1, Node2, Node3: TACRExprNode; CaseInsensitive, PartialKey: Boolean);
begin
  Create(Op, Node1, Node2, CaseInsensitive, PartialKey);
  Children.Add(Node3);
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRExprNode.Destroy;
var i: Integer;
begin
  for i:=0 to Children.Count-1 do
    TACRExprNode(Children[i]).Free;
  Children.Free;
  Value.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// converts some WideStrings to Strings
//------------------------------------------------------------------------------
procedure TACRExprNode.PatchWideStrings;
var i: Integer;
begin
  for i:=0 to Children.Count-1 do
    TACRExprNode(Children[i]).PatchWideStrings;
end;//PatchWideStrings


//------------------------------------------------------------------------------
// is expression contains aggregated function
//------------------------------------------------------------------------------
function TACRExprNode.IsAggregated: Boolean;
var i: Integer;
begin
  Result := False;
  for i:=0 to Children.Count-1 do
   begin
    if Children[i] = nil then continue;
    Result := TACRExprNode(Children[i]).IsAggregated;
    if Result then Break;
   end;
end;//isAggregated


//------------------------------------------------------------------------------
// Init for group function
//------------------------------------------------------------------------------
procedure TACRExprNode.Init;
var i: integer;
begin
  for i:=0 to Children.Count-1 do
   TACRExprNode(Children[i]).Init;
end;//Init


//------------------------------------------------------------------------------
// Accumulate for group functions
//------------------------------------------------------------------------------
procedure TACRExprNode.Accumulate;
var i: integer;
begin
  for i:=0 to Children.Count-1 do
   TACRExprNode(Children[i]).Accumulate;
end;//Accumulate


//------------------------------------------------------------------------------
// assign AO
//------------------------------------------------------------------------------
procedure TACRExprNode.AssignAO(AO: TACRAO);
var
  i: integer;
begin
  for i:=0 to Children.Count-1 do
    TACRExprNode(Children[i]).AssignAO(AO);
end;//AssignAO


//------------------------------------------------------------------------------
// Assign Cursor
//------------------------------------------------------------------------------
procedure TACRExprNode.AssignCursor(Cursor: TACRCursor);
var
  i: integer;
begin
  for i:=0 to Children.Count-1 do
    TACRExprNode(Children[i]).AssignCursor(Cursor);
end;//AssignCursor


//------------------------------------------------------------------------------
// process assign New Cursor Buffer
//------------------------------------------------------------------------------
procedure TACRExprNode.AssignCursorBuffer(Buffer: TACRRecordBuffer);
var i: integer;
begin
  for i:=0 to Children.Count-1 do
    TACRExprNode(Children[i]).AssignCursorBuffer(Buffer);
end;//AssignCursorBuffer


//------------------------------------------------------------------------------
// get Data Size
//------------------------------------------------------------------------------
function TACRExprNode.GetDataSize: Integer;
var
  i: integer;
  ftype: TACRBaseFieldType;
  child: TACRExprNode;
  size:  Integer;
begin
  Result := 0;
  ftype := AdvancedFieldTypeToBaseFieldType(getDataType);
  //if ftype in [ftString, ftWideString, ftBytes] then
  if IsStringFieldType(ftype) then
   begin
    if Operator in [doADD, doCONCAT, doCAST] then
     for i:=0 to Children.Count-1 do
      begin
       child := TACRExprNode(Children[i]);
       size := child.getDataSize;
       if (size <= 0) then
        size := ACRExpressionMaxStringSize;
       Result := Result + size;
       {case child.getDataType of
         ftString,
         ftBytes,
         ftWideString:
           Result := Result + child.getDataSize;
         ftAutoInc,
         ftInteger:
           Result := Result + 11;
         ftSmallInt:
           Result := Result + 6;
         ftWord:
           Result := Result + 5;
         ftLargeInt:
           Result := Result + 20;
         ftFloat,
         ftDate,
         ftTime,
         ftDateTime:
           Result := Result + 26;
         ftBoolean:
           Result := Result + 5;
         ftCurrency:
           Result := Result + 21;
       end;
       }
      end;
   end;
//  if ftype = ftWideString then
//    Result := Result * 2;
end;//GetDataSize


//------------------------------------------------------------------------------
// return Type of Data
//------------------------------------------------------------------------------
function TACRExprNode.GetDataType: TACRAdvancedFieldType;
var
  i: integer;
begin
  // default ftUnknown - Error flag
  Result := aftUnknown;
  if Children.Count <> 0 then
   begin
    // Take first type
    Result := TACRExprNode(Children[0]).getDataType;
    // and compare it with other
    for i:=1 to Children.Count-1 do
      begin
        Result := getCommonDataType(Result, TACRExprNode(Children[i]).getDataType);
      end;
   end;
  if (Result = aftUnknown) then
    raise EACRException.Create(30283, ErrorGCannotDetermineExpressionType);
end;//GetDataType


//------------------------------------------------------------------------------
// return Data Precision
//------------------------------------------------------------------------------
function TACRExprNode.GetPrecision: Integer;
begin
  Result := 0;
end;//GetPrecision


//------------------------------------------------------------------------------
// can be used by the AO?
//------------------------------------------------------------------------------
function TACRExprNode.CanBeAssigned(AO: TACRAO): Boolean;
var
  i: Integer;
begin
  Result := True;
  // get childs cans
  for i:=0 to Children.Count-1 do
    begin
      Result := Result and TACRExprNode(Children[i]).CanBeAssigned(AO);
      if not Result then break;
    end;
end;//CanBeAssigned


//------------------------------------------------------------------------------
// is node a join condition?
//------------------------------------------------------------------------------
function TACRExprNode.IsJoinCondition(AO1, AO2: TACRAO): Boolean;
begin
  Result := False;
end;// IsJoinConditions


//------------------------------------------------------------------------------
// replace pseudonyms to original names (f1 -> table1.field1)
//------------------------------------------------------------------------------
procedure TACRExprNode.ReplacePseudonyms(SelectList: array of TACRSelectListItem);
var
  i: integer;
begin
  for i:=0 to Children.Count-1 do
    TACRExprNode(Children[i]).ReplacePseudonyms(SelectList);
end;//ReplacePseudonyms


////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeField
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeField.Create(Cursor: TACRCursor; FieldName, TableName: string);
begin
  inherited Create;

  LCursor := Cursor;
  FFieldName := FieldName;
  FTableName := TableName;

  FFieldNo := -1;
  FFieldOffsetInBuffer := 0;
  FFieldType := aftUnknown;
  FBaseFieldType := bftUnknown;
  FFieldSize := 0;
  FFieldPrecision := 0;
  LRecordBuffer := nil;

  if LCursor <> nil then
    AssignCursor(Cursor);
end;//Create


//------------------------------------------------------------------------------
// Return Data Value
//------------------------------------------------------------------------------
function TACRExprNodeField.GetDataValue: TACRVariant;
var TempBuffer: PChar;
begin
  Value.SetNull;
  if (LRecordBuffer <> nil) then
    begin
    // If field not null then set Value variable
    if (not CheckNullFlag(FFieldNo, LRecordBuffer)) then
     begin
      if (LCursor = nil) then
       raise EACRException.Create(11300,ErrorLNilPointer);
//      if (IsBLOBFieldType(FBaseFieldType)) then
      TempBuffer := LCursor.CurrentRecordBuffer;
      try
        LCursor.CurrentRecordBuffer := LRecordBuffer;
        LCursor.GetFieldValue(Value,FFieldNo,True,False);
      finally
        LCursor.CurrentRecordBuffer := TempBuffer;
      end;
//      else
//      Value.SetData(LRecordBuffer + FFieldOffsetInBuffer,
//                    FFieldSize, FBaseFieldType, False)
     end;
    end
  else
    if ( (LAO <> nil) and (FFieldNo <> -1) ) then
      LAO.GetFieldValue(Value, FFieldNo, true, true)
    else
      raise EACRException.Create(30196, ErrorGDatasetAndAONotAssigned);
  Result := Value;
end;//GetDataValue


//------------------------------------------------------------------------------
// process assign AO
//------------------------------------------------------------------------------
procedure TACRExprNodeField.AssignAO(AO: TACRAO);
var
  Fields: TACRIntegerArray;
begin
  LCursor := nil;
  LRecordBuffer := nil;
  LAO := AO;
  if LAO <> nil then
   begin
    Fields := TACRIntegerArray.Create;
    try
      LAO.FieldExists(FFieldName, FTableName, True, Fields, True);
      if Fields.ItemCount = 1 then
        begin
          // Store Field Parameters
          FFieldNo := Fields.items[0];
          FFieldOffsetInBuffer := 0;
          FFieldType := LAO.GetFieldType(FFieldNo);
          FBaseFieldType := AdvancedFieldTypeToBaseFieldType(FFieldType);
          FFieldSize := LAO.GetFieldSize(FFieldNo);
          FFieldPrecision := LAO.GetFieldPrecision(FFieldNo);
        end
      else
        raise EACRException.Create(30197, ErrorGCannotFindField,
                                   [FTableName+'.'+FFieldName]);
    finally
      Fields.Free;
    end;
  end;
end;//AssignAO


//------------------------------------------------------------------------------
// process assign Cursor and its buffer
//------------------------------------------------------------------------------
procedure TACRExprNodeField.AssignCursor(Cursor: TACRCursor);
var n: Integer;
begin
  if Cursor = nil then
    raise EACRException.Create(30202, ErrorGValueIsNull);

  LAO := nil;
  LCursor := Cursor;

  // if field not exists then raise exception
  n := LCursor.FieldDefs.GetDefNumberByName(FFieldName);
  if (n = -1) then
   begin
    n := LCursor.VisibleFieldDefs.GetDefNumberByName(FFieldName);
    if (n = -1) then
     raise EACRException.Create(30115, ErrorGFieldWithNameNotFound, [FFieldName]);
    n := LCursor.VisibleFieldDefs[n].FieldNoReference;
   end;
  // Store Field Parameters
  FFieldNo := n;
  FFieldOffsetInBuffer := LCursor.FieldDefs[n].MemoryOffset;
  FFieldType := LCursor.FieldDefs[n].AdvancedFieldType;
  FBaseFieldType := AdvancedFieldTypeToBaseFieldType(FFieldType);
  FFieldSize := LCursor.FieldDefs[n].MemoryDataSize;
  AssignCursorBuffer(LCursor.CurrentRecordBuffer);
end;//AssignCursor


//------------------------------------------------------------------------------
// process assign New Cursor Buffer
//------------------------------------------------------------------------------
procedure TACRExprNodeField.AssignCursorBuffer(Buffer: TACRRecordBuffer);
begin
  LRecordBuffer := Buffer;
end;//AssignCursorBuffer


//------------------------------------------------------------------------------
// return Type of Data
//------------------------------------------------------------------------------
function TACRExprNodeField.GetDataType: TACRAdvancedFieldType;
begin
  Result := aftUnknown;
  if LCursor <> nil then
    Result := LCursor.FieldDefs[FFieldNo].AdvancedFieldType
  else
    if LAO <> nil then
      Result := LAO.GetFieldType(FFieldNo);

  if (Result = aftUnknown) then
    raise EACRException.Create(30284, ErrorGCannotDetermineExpressionType);
end;//GetDataType


//------------------------------------------------------------------------------
// return Data Size
//------------------------------------------------------------------------------
function TACRExprNodeField.GetDataSize: Integer;
begin
  {Result := 0;
  if LCursor <> nil then
    Result := LCursor.FieldDefs[FFieldNo].MemoryDataSize
  else
    if LAO <> nil then
      Result := LAO.GetFieldSize(FFieldNo);
  }
  Result := FFieldSize;
end;//GetDataSize


//------------------------------------------------------------------------------
// can be used by the AO?
//------------------------------------------------------------------------------
function TACRExprNodeField.CanBeAssigned(AO: TACRAO): Boolean;
var
  Fields: TACRIntegerArray;
  n: integer;
begin
  n := -1;
  if AO <> nil then
    begin
      Fields := TACRIntegerArray.Create;
      try
        AO.FieldExists(FFieldName, FTableName, False, Fields);
        if Fields.ItemCount = 1 then
          n := Fields.items[0];
      finally
        Fields.Free;
      end;
    end;
  Result := (n <> -1);
end;


//------------------------------------------------------------------------------
// fills Field Item
//------------------------------------------------------------------------------
procedure TACRExprNodeField.FillItem(var Item: TACRSelectListItem);
begin
  Item.TableName := FTableName;
  Item.FieldName := FFieldName;
end;// FillItem


//------------------------------------------------------------------------------
// replace pseudonyms to original names (f1 -> table1.field1)
//------------------------------------------------------------------------------
procedure TACRExprNodeField.ReplacePseudonyms(SelectList: array of TACRSelectListItem);
var
  i: Integer;
begin
  for i := 0 to Length(SelectList)-1 do
  if (AnsiUpperCase(FFieldName) = AnsiUpperCase(SelectList[i].Pseudonym)) then
   if (not SelectList[i].IsExpression) then
    if (FTableName = '') then
     begin
      FFieldName := SelectList[i].FieldName;
      FTableName := SelectList[i].TableName;
     end;
end;//ReplacePseudonyms


////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeSubQuery
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeSubQuery.Create(
                       QueryText:         String;
                       Params:            TParams;
                       InMemory:          Boolean;
                       DatabaseName:      String;
                       SessionName:       String;
                       bNot:              Boolean = False;
                       SourceNode:        TACRExprNode = nil
                       );
var i:      Integer;
    Param:  TParam;
begin
  inherited Create;
  FQueryExecuted := False;
  FQuery := TACRQuery.Create(nil);
  TACRQuery(FQuery).SQL.Text := QueryText;
  TACRQuery(FQuery).DatabaseName := DatabaseName;
  TACRQuery(FQuery).SessionName := SessionName;
  TACRQuery(FQuery).InMemory := InMemory;
  TACRQuery(FQuery).Prepare;
  for i := 0 to TACRQuery(FQuery).Params.Count-1 do
   begin
    Param := Params.FindParam(TACRQuery(FQuery).Params[i].Name);
    if (Param <> nil) then
     TACRQuery(FQuery).Params[i].Assign(Param);
   end;
  FNot := bNot;
  FSourceNode := SourceNode;
  if (SourceNode <> nil) then
   Children.Add(SourceNode);
end; // Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRExprNodeSubQuery.Destroy;
begin
  if (FQuery <> nil) then
   FQuery.Free;
  inherited;
end; // Destroy


//------------------------------------------------------------------------------
// return Data Value
//------------------------------------------------------------------------------
function TACRExprNodeSubQuery.GetDataValue: TACRVariant;
begin
  if (not FQueryExecuted) then
   begin
    FQueryExecuted := True;
    try
      TACRQuery(FQuery).Open;
      if (TACRQuery(FQuery).RecordCount < 1) then
       raise EACRException.Create(10848,ErrorLSubqueryReturnsNoRows);
      if (TACRQuery(FQuery).RecordCount > 1) then
       raise EACRException.Create(10849,ErrorLSubqueryReturnsMultipleRows);

      TACRQuery(FQuery).GetFieldValue(Value,0,False);
      FQuery.Close;
      FreeAndNil(FQuery);
    except
      FQuery.Free;
      Fquery := nil;
      raise;
    end;
   end;
  Result := Value;
end; // GetDataValue


////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeSubQueryIn
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// return Data Value
//------------------------------------------------------------------------------
function TACRExprNodeSubQueryIn.GetDataValue: TACRVariant;
var v:         variant;
    FieldName: String;
    bvalue:    Boolean;
begin
  if (not FQueryExecuted) then
   begin
    FQueryExecuted := True;
    try
      TACRQuery(FQuery).Open;
    except
      FQuery.Free;
      Fquery := nil;
      raise;
    end;
   end;
  FieldName := FQuery.Fields[0].FieldName;
  if (FSourceNode = nil) then
   raise EACRException.Create(10850,ErrorLSubqueryINnoArgumentPassed);
  Result := TACRExprNode(FSourceNode).GetDataValue;
  if (Result.IsNull) then
   v := varNull
  else
   v := Result.AsVariant;
  bvalue := TACRQuery(FQuery).Locate(FieldName,v,[]);
  if (FNot) then
   bvalue := not bvalue;
  Value.AsBoolean := bvalue;
  Result := Value;
end; // GetDataValue


////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeSubQueryExists
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// return Data Value
//------------------------------------------------------------------------------
function TACRExprNodeSubQueryExists.GetDataValue: TACRVariant;
var bvalue: Boolean;
begin
  if (not FQueryExecuted) then
   begin
    FQueryExecuted := True;
    try
      TACRQuery(FQuery).Open;
      if (FNot) then
        bvalue := TACRQuery(FQuery).Eof
      else
        bvalue := not TACRQuery(FQuery).Eof;
      Value.AsBoolean := bvalue;
      FQuery.Close;
      FreeAndNil(FQuery);
    except
      FQuery.Free;
      Fquery := nil;
      raise;
    end;
   end;
  Result := Value;
end; // GetDataValue


////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeComparison
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeComparison.Create(Op: TACRDataOperator; Node1,
  Node2: TACRExprNode; TrueFalseNullLogic, CaseInsensitive,
  PartialKey: boolean);
begin
  inherited Create(Op, Node1, Node2,CaseInsensitive, PartialKey);
  F3ValueLogic := TrueFalseNullLogic;
  Value.SetNull(bftLogical);
end;//Create


//------------------------------------------------------------------------------
// converts some WideStrings to Strings
//------------------------------------------------------------------------------
procedure TACRExprNodeComparison.PatchWideStrings;
var
   i:      Integer;
   NField: TACRExprNodeField;
   NConst: TACRExprNodeConst;
   dt: TACRAdvancedFieldType;
begin
  NField := nil;
  NConst := nil;
  for i:=0 to Children.Count-1 do
    if TACRExprNode(Children[i]) is TACRExprNodeField then
      NField := Children[i]
    else if TACRExprNode(Children[i]) is TACRExprNodeConst then
      NConst := Children[i];

  if (NConst <> nil) and (NField <> nil) then
    begin
      dt := NField.GetDataType;
      if ( IsStringFieldType(dt) and (not IsWideStringFieldType(dt)) ) then
        if IsWideStringFieldType(NConst.GetDataType) then
          NConst.Value.Cast(bftVarchar);
    end;
end;//PatchWideStrings


//------------------------------------------------------------------------------
// return Value
//------------------------------------------------------------------------------
function TACRExprNodeComparison.GetDataValue: TACRVariant;
var
  cmpRes: TACRCompareResult;
  arg1, arg2: TACRVariant;
begin
    arg1 := TACRExprNode(Children[0]).GetDataValue;
    arg2 := TACRExprNode(Children[1]).GetDataValue;
    cmpRes := arg1.Compare(arg2, F3ValueLogic , FCaseInsensitive, FPartialKey);

    case Operator of
     doEQ:   Value.AsBoolean := (cmpRes = cmprEqual);
     doNE:   Value.AsBoolean := (cmpRes <> cmprEqual);
     doLT:   Value.AsBoolean := (cmpRes = cmprLower);
     doGT:   Value.AsBoolean := (cmpRes = cmprGreater);
     doLE:   Value.AsBoolean := (cmpRes in [cmprEqual, cmprLower]);
     doGE:   Value.AsBoolean := (cmpRes in [cmprEqual, cmprGreater]);
     else
       raise EACRException.Create(30117, ErrorGUnknownOperator, [GetOperatorName(Operator)]);
    end;
  Result := Value;
end;//GetDataValue


//------------------------------------------------------------------------------
// can use index for comparison?
//------------------------------------------------------------------------------
function TACRExprNodeComparison.CanUseIndex(IndexDef: TACRIndexDef): Boolean;
var
  FieldNode: TACRExprNodeField;
  ConstNode: TACRExprNodeConst;
begin
  FieldNode := nil;
  ConstNode := nil;
  if (TObject(Children[0]) is TACRExprNodeField) then
    FieldNode := Children[0];
  if (TObject(Children[1]) is TACRExprNodeField) then
    FieldNode := Children[1];
  if (TObject(Children[0]) is TACRExprNodeConst) then
    ConstNode := Children[0];
  if (TObject(Children[1]) is TACRExprNodeConst) then
    ConstNode := Children[1];
  if ((FieldNode <> nil) and (ConstNode <> nil)) then
   begin
     Result := (AnsiUpperCase(FieldNode.FieldName) =
                AnsiUpperCase(IndexDef.Columns[0].FieldName));
     if (Result) then
      if (IsStringFieldType(FieldNode.BaseFieldType)) then
       Result := (FCaseInsensitive = IndexDef.Columns[0].CaseInsensitive);
     if (Result) then
       Result := (FPartialKey = False);
     if (Result) then
       Result := (Operator = doEQ) or
                 (Operator = doGT) or
                 (Operator = doLT) or
                 (Operator = doGE) or
                 (Operator = doLE);
     if (Result) then
       Result := (FieldNode.FieldSize >= ConstNode.GetDataSize);
   end
  else
   Result := False;
end;// CanUseIndex


//------------------------------------------------------------------------------
// add index scan condition
//------------------------------------------------------------------------------
procedure TACRExprNodeComparison.AddAsIndexScanCondition(
               ScanConditions:          TACRScanSearchConditionArray;
               IndexDef:                TACRIndexDef;
               ExtractedConditionsInfo: TList;
               Expression:              TACRExpression
                          );
var
  FieldNode:     TACRExprNodeField;
  ConstNode:     TACRExprNodeConst;
  ScanCondition: TACRScanSearchCondition;
  ExCondInfo:    PACRExtractedConditionInfo;
begin
  FieldNode := nil;
  ConstNode := nil;
  if (TObject(Children[0]) is TACRExprNodeField) then
    FieldNode := Children[0];
  if (TObject(Children[1]) is TACRExprNodeField) then
    FieldNode := Children[1];
  if (TObject(Children[0]) is TACRExprNodeConst) then
    ConstNode := Children[0];
  if (TObject(Children[1]) is TACRExprNodeConst) then
    ConstNode := Children[1];
  if ((FieldNode <> nil) and (ConstNode <> nil)) then
   begin
    if (FieldNode.LCursor = nil) then
      raise EACRException.Create(20054, ErrorANilPointer);
    if (IndexDef.ColumnCount < 1) then
      raise EACRException.Create(20057, ErrorAInvalidIndexForSearch);

    ScanCondition := TACRScanSearchCondition.Create;
    try
      if (IndexDef.Columns[0].Descending) then
        case Operator of
         doEQ: ScanCondition.Condition := scEqual;
         doGT: ScanCondition.Condition := scLower;
         doLT: ScanCondition.Condition := scGreater;
         doGE: ScanCondition.Condition := scLowerEqual;
         doLE: ScanCondition.Condition := scGreaterEqual;
        end
      else
        case Operator of
         doEQ: ScanCondition.Condition := scEqual;
         doGT: ScanCondition.Condition := scGreater;
         doLT: ScanCondition.Condition := scLower;
         doGE: ScanCondition.Condition := scGreaterEqual;
         doLE: ScanCondition.Condition := scLowerEqual;
        end;

      ScanCondition.KeyRecordBuffer := FieldNode.LCursor.AllocateRecordBuffer;
      FieldNode.LCursor.SetFieldValue(
                                       ConstNode.Value,
                                       FieldNode.FieldNo,
                                       True,
                                       ScanCondition.KeyRecordBuffer
                                     );
      ScanCondition.KeyFieldCount := 1;
      ScanCondition.IndexID := IndexDef.ObjectID;
      ScanCondition.Expression := nil;
      ScanConditions.AddCondition(ScanCondition);

      New(ExCondInfo);
      ExCondInfo^.KeyRecordBuffer := ScanCondition.KeyRecordBuffer;
      ExCondInfo^.ExtractedExpressionNode := Self;
      ExCondInfo^.Expression := Expression;
      ExtractedConditionsInfo.Add(ExCondInfo);
    finally
      ScanCondition.Free;
    end;
   end
  else
   raise EACRException.Create(20055, ErrorANilPointer);
end;// AddScanCondition


//------------------------------------------------------------------------------
// incompatible
//------------------------------------------------------------------------------
function TACRExprNodeComparison.IsIncompatible: Boolean;
var
  FieldNode: TACRExprNodeField;
  ConstNode: TACRExprNodeConst;
begin
  Result := False;
  FieldNode := nil;
  ConstNode := nil;
  if (TObject(Children[0]) is TACRExprNodeField) then
    FieldNode := Children[0];
  if (TObject(Children[1]) is TACRExprNodeField) then
    FieldNode := Children[1];
  if (TObject(Children[0]) is TACRExprNodeConst) then
    ConstNode := Children[0];
  if (TObject(Children[1]) is TACRExprNodeConst) then
    ConstNode := Children[1];

  if ((FieldNode <> nil) and (ConstNode <> nil)) then
   begin
    if (Operator = doEQ) then
     if (IsStringFieldType(FieldNode.BaseFieldType)) then
       Result := (FieldNode.FieldSize < ConstNode.GetDataSize);
   end;
end; // IsIncompatible


//------------------------------------------------------------------------------
// is node a join condition?
//------------------------------------------------------------------------------
function TACRExprNodeComparison.IsJoinCondition(AO1,
  AO2: TACRAO): Boolean;
var
  Left, Right: TACRExprNodeField;
  FieldNumber: TACRIntegerArray;
  Field1LeftCount, Field1RightCount: integer;
  Field2LeftCount, Field2RightCount: integer;
begin
  Result := false;
  if not ((TACRExprNode(Children[0]) is TACRExprNodeField) and
          (TACRExprNode(Children[1]) is TACRExprNodeField)) then Exit;

  Left := TACRExprNodeField(Children[0]);
  Right := TACRExprNodeField(Children[1]);
  if Operator = doEQ then
   if ((Left is TACRExprNodeField) and
       (Right is TACRExprNodeField)) then
    begin
     FieldNumber := TACRIntegerArray.Create;
     try
      // map fields
      FieldNumber.SetSize(0);
      Field1LeftCount := AO1.FieldExists(Left.FieldName, Left.TableName, True, FieldNumber);
      FieldNumber.SetSize(0);
      Field2LeftCount := AO1.FieldExists(Right.FieldName, Right.TableName, True, FieldNumber);
      FieldNumber.SetSize(0);
      Field1RightCount := AO2.FieldExists(Left.FieldName, Left.TableName, True, FieldNumber);
      FieldNumber.SetSize(0);
      Field2RightCount := AO2.FieldExists(Right.FieldName, Right.TableName, True, FieldNumber);
      // ambiguous field reference?
      if (Field1LeftCount+Field1RightCount > 1) then
        raise EACRException.Create(10404, ErrorLAmbiguousFieldReference, [Left.FieldName]);
      if (Field2LeftCount+Field2RightCount > 1) then
        raise EACRException.Create(10405, ErrorLAmbiguousFieldReference, [Right.FieldName]);
      // each field macths only 1 AO?
      if ((Field1LeftCount+Field1RightCount = 1) and
          (Field2LeftCount+Field2RightCount = 1)) then
       // boths fields match different AO?
       if (Field1LeftCount <> Field2LeftCount) then
        Result := True;
     finally
      FieldNumber.Free;
     end;
    end;
end; // IsJoinCondition


////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeSystem
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// LastAutoInc
//------------------------------------------------------------------------------
procedure TACRExprNodeSystem.LastAutoInc;
begin
  Value.SetNull(AdvancedFieldTypeToBaseFieldType(GetDataType));
  Value.AsInt64 := TACRTable(FTable).LastAutoincValue(FColumnName);
end;//LastAutoInc


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeSystem.Create(Operator:     TACRDataOperator;
                                      TableName:    String;
                                      ColumnName:   String;
                                      InMemory:     Boolean;
                                      SessionName:  String;
                                      DatabaseName: String);
var
  FieldDef: TACRAdvFieldDef;
begin
  inherited Create(Operator);
  FTableName := TableName;
  FColumnName := ColumnName;
  FSessionName := SessionName;
  FFieldType := aftUnknown;
  FInMemory := InMemory;
  FDatabaseName := DatabaseName;

  FTable := TACRTable.Create(nil);
  TACRTable(FTable).DatabaseName := FDatabaseName;
  TACRTable(FTable).SessionName := FSessionName;
  TACRTable(FTable).TableName := FTableName;
  TACRTable(FTable).InMemory := FInMemory;
  TACRTable(FTable).ReadOnly := True;
  TACRTable(FTable).Open;

  // Field Type
  FieldDef := TACRTable(FTable).AdvFieldDefs.Find(FColumnName);
  if (FieldDef = nil) then
    raise EACRException.Create(30337, ErrorGFieldWithNameNotFound, [FColumnName]);
  FFieldType := FieldDef.DataType;

end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRExprNodeSystem.Destroy;
begin
  TACRTable(FTable).Close;
  FTable.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// Return Data Value
//------------------------------------------------------------------------------
function TACRExprNodeSystem.GetDataValue: TACRVariant;
begin
  case Operator of
    doLASTAUTOINC: LastAutoInc;
    else
      raise EACRException.Create(30338, ErrorGUnknownOperator, [GetOperatorName(Operator)]);
  end;

  Result := Value;
end;//GetDataValue


//------------------------------------------------------------------------------
// return Type of Data
//------------------------------------------------------------------------------
function TACRExprNodeSystem.GetDataType: TACRAdvancedFieldType;
begin
  Result := FFieldType;
end;//GetDataType


//------------------------------------------------------------------------------
// return Size of Data (for strings and arrays)
//------------------------------------------------------------------------------
function TACRExprNodeSystem.GetDataSize: Integer;
begin
  Result := 0;
end;//GetDataSize



////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeAggregated
//
////////////////////////////////////////////////////////////////////////////////



//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeAggregated.Create(Op: TACRDataOperator);
begin
  inherited Create(Op);
  Count := 0;
  AvgSum := TACRVariant.Create;
  distinct := False;
  FTempTable := nil;
end;//Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeAggregated.Create(Op: TACRDataOperator;
  distinct: boolean; Node: TACRExprNode);
begin
  inherited Create(op, Node);
  self.distinct := distinct;
  AvgSum := TACRVariant.Create;
  FTempTable := nil;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRExprNodeAggregated.Destroy;
begin
  AvgSum.Free;
  if (FTempTable <> nil) then
   FTempTable.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// Return Data Value
//------------------------------------------------------------------------------
function TACRExprNodeAggregated.GetDataValue: TACRVariant;
begin
  case Operator of
    doCOUNT,
    doCOUNTALL:
      Value.AsInteger := Count;
    doAVG:
      begin
        //   = AvgSum / Count
        Value.Assign(AvgSum);
        AvgSum.AsDouble := Count;
        Value.Division(AvgSum);
      end;
  end;
  Result := Value;
end;//GetDataValue


//------------------------------------------------------------------------------
// return Type of Data
//------------------------------------------------------------------------------
function TACRExprNodeAggregated.GetDataType: TACRAdvancedFieldType;
begin
  Result := aftUnknown;
  case Operator of
    doCOUNTALL,
    doCOUNT:
      Result := aftInteger;
    doMIN,
    doMAX:
      if Children.Count <> 0 then
             Result := TACRExprNode(Children[0]).getDataType;
    doSUM:
      if Children.Count <> 0 then
       begin
        Result := TACRExprNode(Children[0]).getDataType;
        if not (Result in [aftSingle, aftDouble, aftExtended,
                            aftCurrency] ) then
          Result := aftLargeint;
       end;
    doAVG:
      Result := aftDouble;
  end;
end;//GetDataType


//------------------------------------------------------------------------------
// is expression contains aggregated function
//------------------------------------------------------------------------------
function TACRExprNodeAggregated.isAggregated: Boolean;
begin
  Result := True;
end;//isAggregated


//------------------------------------------------------------------------------
// Init for aggregated functions
//------------------------------------------------------------------------------
procedure TACRExprNodeAggregated.Init;
var FieldDef: TACRAdvFieldDef;
begin
  count := 0;
  Value.SetNull;
  if Operator = doAVG then
    AvgSum.SetNull;
  if (distinct) then
   begin
    if (Children.Count <= 0) then
     raise EACRException.Create(11338,ErrorLNoChildren);
    FTempTable := TACRTable.Create(nil);
    TACRTable(FTempTable).Temporary := True;
    repeat
      TACRTable(FTempTable).TableName := GetTemporaryName(ACRTemporaryTableName);
    until (not TACRTable(FTempTable).Exists);
    TACRTable(FTempTable).FieldDefs.Clear;
    TACRTable(FTempTable).AdvFieldDefs.Clear;
    TACRTable(FTempTable).IndexDefs.Clear;
    TACRTable(FTempTable).AdvIndexDefs.Clear;
    FieldDef := TACRTable(FTempTable).AdvFieldDefs.AddFieldDef;
    FieldDef.Name := ACRExpressionFieldName;

    FieldDef.DataType := TACRExprNode(Children[0]).GetDataType;
    FieldDef.Size := TACRExprNode(Children[0]).GetDataSize;
    TACRTable(FTempTable).IndexDefs.Add(ACRDefaultIndexName,FieldDef.Name,[]);
    TACRTable(FTempTable).CreateTable;
    TACRTable(FTempTable).IndexName := ACRDefaultIndexName;
    TACRTable(FTempTable).Open;
   end;
end;//Init


//------------------------------------------------------------------------------
// Accumulate for group functions
//------------------------------------------------------------------------------
procedure TACRExprNodeAggregated.Accumulate;
var
  TmpVal:  TACRVariant;

  function IsDistinctHaveNewValue: Boolean;
  begin
    if (TACRTable(FTempTable).RecordCount = 0) then
     Result := True
    else
     begin
      TACRTable(FTempTable).SetKey;
      TACRTable(FTempTable).SetFieldValue(TmpVal,0,True);
      Result := (not TACRTable(FTempTable).GotoKey);
     end;
  end; // IsDistinctHaveNewValue

  procedure InsertNewDistinctValue;
  begin
    TACRTable(FTempTable).Insert;
    TACRTable(FTempTable).SetFieldValue(TmpVal,0,True);
    TACRTable(FTempTable).Post;
  end; // InsertNewDistinctValue

begin
  TmpVal := TACRVariant.Create;
  try
    if Children.Count <> 0 then
      TmpVal.Assign(TACRExprNode(Children[0]).GetDataValue, False);
    case Operator of
      doMIN:
        begin
         if (Value.IsNull) then
           Value.Assign(TmpVal)
         else
           if (not TmpVal.IsNull) then
             if (TmpVal.Compare(Value) = cmprLower) then
               Value.Assign(TmpVal);
        end;
      doMAX:
        begin
         if (Value.IsNull) then
           Value.Assign(TmpVal)
         else
           if (not TmpVal.IsNull) then
             if (TmpVal.Compare(Value) = cmprGreater) then
               Value.Assign(TmpVal);
        end;
      doCOUNTALL:
        begin
          Inc(count);
        end;
      doCOUNT:
        begin
          if (not TmpVal.IsNull) then
           begin
            if (distinct) then
             begin
              if (IsDistinctHaveNewValue) then
               begin
                InsertNewDistinctValue;
                Inc(count);
               end;
             end
            else
              Inc(count);
           end;
        end;
      doSUM:
        begin
          // If not numeric type (string for example) - raise
          if (not TmpVal.IsNumericDataType) then
            raise EACRException.Create(30132, ErrorGNotNumericDataType, ['SUM']);
          // if value not null
          if (not TmpVal.IsNull) then
            if (distinct) then
             begin
              if (IsDistinctHaveNewValue) then
               begin
                InsertNewDistinctValue;
                Value.Add(TmpVal);
               end;
             end
            else
             Value.Add(TmpVal);
        end;
      doAVG:
        begin
          // If not numeric type (string for example) - raise
          if (not TmpVal.IsNumericDataType) then
            raise EACRException.Create(30133, ErrorGNotNumericDataType, ['AVG']);
          // if value not null
          if (not TmpVal.IsNull) then
            if (distinct) then
             begin
              if (IsDistinctHaveNewValue) then
               begin
                InsertNewDistinctValue;
                AvgSum.Add(TmpVal);
                inc(count);
               end;
             end
            else
             begin
              AvgSum.Add(TmpVal);
              inc(count);
             end;
        end;
      else
        raise EACRException.Create(30134, ErrorGUnknownOperator,
                                   [GetOperatorName(Operator)]);
    end;
  finally
    TmpVal.Free;
  end;
end;//Accumulate



////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeCast
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeCast.Create(Node: TACRExprNode;
  CastType: TACRAdvancedFieldType);
begin
  inherited Create(doCAST, Node);
  FCastType := CastType;
end;//Create


//------------------------------------------------------------------------------
// Return Data Value
//------------------------------------------------------------------------------
function TACRExprNodeCast.GetDataValue: TACRVariant;
begin
  Value.Assign(TACRExprNode(Children[0]).GetDataValue, False);
  Value.Cast(AdvancedFieldTypeToBaseFieldType(FCastType));
  Result := Value;
end;//GetDataValue


//------------------------------------------------------------------------------
// return Type of Data
//------------------------------------------------------------------------------
function TACRExprNodeCast.GetDataType: TACRAdvancedFieldType;
begin
  Result := FCastType;
end;//GetDataType


////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeDateFunction
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeDateFunction.Create(Op: TACRDataOperator;
  Node: TACRExprNode; FormatStr: String);
begin
  inherited Create(op, Node);
  DateFormater := TDateFormater.Create(FormatStr);
end;


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRExprNodeDateFunction.Destroy;
begin
  DateFormater.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// To Date
//------------------------------------------------------------------------------
procedure TACRExprNodeDateFunction.ToDate;
var
  TmpVal: TACRVariant;
  dt: TDateTime;
begin
  Value.SetNull(bftDateTime);
  TmpVal := TACRVariant.Create;
  try
    TmpVal.Assign(TACRExprNode(Children[0]).GetDataValue, False);
    if (not TmpVal.IsNull) then
      begin
        dt := DateFormater.ToDate(TmpVal.AsString);
        Value.AsTDateTime := dt;
        if (dt <> 0) then
          if (trunc(dt) = 0 ) then
            Value.Cast(bftTime);
        if (frac(dt) = 0 ) then
           Value.Cast(bftDate);
      end
  finally
    TmpVal.Free;
  end;
end;//ToDate


//------------------------------------------------------------------------------
// To String
//------------------------------------------------------------------------------
procedure TACRExprNodeDateFunction.ToString;
var
  TmpVal: TACRVariant;
begin
  Value.SetNull(bftVarchar);
  TmpVal := TACRVariant.Create;
  try
    TmpVal.Assign(TACRExprNode(Children[0]).GetDataValue, False);
    if (not TmpVal.IsNull) then
      Value.AsString := DateFormater.ToString(TmpVal.AsTDateTime);
  finally
    TmpVal.Free;
  end;
end;//ToString


//------------------------------------------------------------------------------
// Return Data Value
//------------------------------------------------------------------------------
function TACRExprNodeDateFunction.GetDataValue: TACRVariant;
begin
  case Operator of
    doSYSDATE:        Value.AsTDateTime := now;
    doCURRENT_DATE:   Value.AsTDate := now;
    doCURRENT_TIME:   Value.AsTTime := now;
    doTODATE:         ToDate;
    doTOSTRING:       ToString;
   else raise EACRException.Create(30130, ErrorGUnknownOperator,
                                   [GetOperatorName(Operator)]);
  end;
  Result := Value;
end;//GetDataValue


//------------------------------------------------------------------------------
// return Type of Data
//------------------------------------------------------------------------------
function TACRExprNodeDateFunction.GetDataType: TACRAdvancedFieldType;
begin
  case Operator of
    doSYSDATE,
    doCURRENT_DATE,
    doCURRENT_TIME,
    doTODATE:
                    Result := aftDateTime;
    doTOSTRING:
                    Result := aftString;
    else
                    raise EACRException.Create(30287, ErrorGUnknownOperator,
                                               [GetOperatorName(Operator)]);
  end;
end;//GetDataType


//------------------------------------------------------------------------------
// return Size of Data (for strings and arrays)
//------------------------------------------------------------------------------
function TACRExprNodeDateFunction.GetDataSize: Integer;
begin
  if Operator = doTOSTRING then
    Result := DateFormater.GetStringMaxSize
  else
    Result:=0
end;//GetDataSize



////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeBLOBFunction
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeBLOBFunction.Create(
                       Op:                    TACRDataOperator;
                       const BLOBValue:       String;
                       const BLOBValueFormat: TACRBLOBValueFormat);
var buffer: PChar;
begin
  inherited Create(op, False, False);
  Value.SetNull;
  try
    case BLOBValueFormat of
     bvfMIME64: buffer := ACRMIME64ToBinary(BLOBValue);
     bvfHEX: buffer := ACRHEXToBinary(BLOBValue);
    end;
  except
    raise EACRException.Create(11344,ErrorLCannotConvertBLOBValue,[BLOBValue,Integer(BLOBValueFormat)]);
  end;
  Value.SetData(buffer,MemoryManager.GetMemoryBufferSize(buffer),bftBLOB,False);
end; // Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRExprNodeBLOBFunction.Destroy;
begin
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// Return Data Value
//------------------------------------------------------------------------------
function TACRExprNodeBLOBFunction.GetDataValue: TACRVariant;
begin
  case Operator of
    doTOBLOB: ;
   else raise EACRException.Create(11343, ErrorGUnknownOperator,
                                   [GetOperatorName(Operator)]);
  end;
  Result := Value;
end;//GetDataValue


//------------------------------------------------------------------------------
// return Type of Data
//------------------------------------------------------------------------------
function TACRExprNodeBLOBFunction.GetDataType: TACRAdvancedFieldType;
begin
  Result := aftBLOB;
end;//GetDataType


//------------------------------------------------------------------------------
// return Size of Data (for strings and arrays)
//------------------------------------------------------------------------------
function TACRExprNodeBLOBFunction.GetDataSize: Integer;
begin
  Result:=0
end;//GetDataSize



////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeConst
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// return Value
//------------------------------------------------------------------------------
function TACRExprNodeConst.GetDataValue: TACRVariant;
begin
  Result := Value;
end;//GetDataValue


//------------------------------------------------------------------------------
// return Type of Data
//------------------------------------------------------------------------------
function TACRExprNodeConst.GetDataType: TACRAdvancedFieldType;
begin
  Result := BaseFieldTypeToAdvancedFieldType(Value.DataType);
end;//GetDataType


//------------------------------------------------------------------------------
// return Size of Data (for strings and arrays)
//------------------------------------------------------------------------------
function TACRExprNodeConst.GetDataSize: Integer;
begin
  Result := 0;
  if (not Value.IsNull) then
    if (Value.IsStringDataType) then
      begin
        Result := Value.DataSize;
        // skip last zeros
        Dec(Result);
        if Value.IsWideStringDataType then
          Dec(Result);
      end;
end;//GetDataSize


//------------------------------------------------------------------------------
// can be used by the AO?
//------------------------------------------------------------------------------
function TACRExprNodeConst.CanBeAssigned(AO: TACRAO): Boolean;
begin
  Result := True;
end;//CanBeAssigned


////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeBoolean
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeBoolean.Create(CaseInsensitive, PartialKey: Boolean);
begin
  inherited;
  TempVal := TACRVariant.Create(bftLogical);
end;//Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeBoolean.Create(Op: TACRDataOperator;
  CaseInsensitive, PartialKey: boolean);
begin
  Create(CaseInsensitive, PartialKey);
  Operator := Op;
end;//Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeBoolean.Create(Op: TACRDataOperator;
                     Node: TACRExprNode; CaseInsensitive, PartialKey: boolean);
begin
  Create(Op, CaseInsensitive, PartialKey);
  Children.Add(Node);
end;//Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeBoolean.Create(Op: TACRDataOperator;
            Node1, Node2: TACRExprNode; CaseInsensitive, PartialKey: boolean);
begin
  Create(Op, Node1, CaseInsensitive, PartialKey);
  Children.Add(Node2);
end;//Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRExprNodeBoolean.Create(Op: TACRDataOperator;
       Node1, Node2, Node3: TACRExprNode;CaseInsensitive, PartialKey: boolean);
begin
  Create(Op, Node1, Node2, CaseInsensitive, PartialKey);
  Children.Add(Node3);
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRExprNodeBoolean.Destroy;
begin
  TempVal.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// return Value
//------------------------------------------------------------------------------
procedure TACRExprNodeBoolean.Like;
var
  Str, Pattern: TACRVariant;
  CommonType: TACRBaseFieldType;
begin
  Str := TACRVariant.Create(bftLogical);
  Pattern := TACRVariant.Create(bftLogical);
  try
    Str.Assign(TACRExprNode(Children[0]).GetDataValue, False);
    Pattern.Assign(TACRExprNode(Children[1]).GetDataValue, False);


    // If Nulls
    if (Pattern.IsNull or Str.IsNull) then
      begin
        Exit;
      end;

    // Get Common Data Type
    CommonType := GetCommonDataType(Str.DataType, Pattern.DataType);
    // Check for strings
    if (not IsStringFieldType(CommonType)) then
     raise EACRException.Create(30120, ErrorGNotStringArgument, ['LIKE']);

    // Do Like
    if IsWideStringFieldType(CommonType) then
      Value.AsBoolean := IsWideStrMatchPattern(
                                            PWideChar(Str.AsWideString),
                                            PWideChar(Pattern.AsWideString),
                                            FCaseInsensitive
                                          )
    else
      Value.AsBoolean := IsStrMatchPattern(
                                            PChar(Str.AsString),
                                            Pchar(Pattern.AsString),
                                            FCaseInsensitive
                                          );

  finally
    Str.Free;
    Pattern.Free;
  end;
end;//Like


//------------------------------------------------------------------------------
// process IN (...)
//------------------------------------------------------------------------------
procedure TACRExprNodeBoolean.bIn;
var
  Val1, Val2: TACRVariant;
  //CommonType: TACRBaseFieldType;
  i: Integer;
begin
  Val1 := TACRVariant.Create;
  Val2 := TACRVariant.Create;
  try
    Val1.Assign(TACRExprNode(Children[0]).GetDataValue, False);
    Value.AsBoolean := False;
    // Val is null ?
    if (Val1.IsNull) then
      Value.SetNull(bftLogical)
    else
      for i:=1 to Children.Count-1 do
        begin
          Val2.Assign(TACRExprNode(Children[i]).GetDataValue, False);
          if Val1.Compare(Val2) = cmprEqual then
            begin
              Value.AsBoolean := true;
              Break;
            end;
        end;
  finally
    Val1.Free;
    Val2.Free;
  end;
end;//bIn


//------------------------------------------------------------------------------
// process A BETWEEN B AND C
//------------------------------------------------------------------------------
procedure TACRExprNodeBoolean.Between;
var
  ValCur, ValMin, ValMax: TACRVariant;
begin
  ValCur := TACRVariant.Create;
  ValMin := TACRVariant.Create;
  ValMax := TACRVariant.Create;
  try
    ValCur.Assign(TACRExprNode(Children[0]).GetDataValue, False);
    ValMin.Assign(TACRExprNode(Children[1]).GetDataValue, False);
    ValMax.Assign(TACRExprNode(Children[2]).GetDataValue, False);
    // Nulls ?
    if (ValCur.IsNull or ValMin.IsNull or ValMax.IsNull) then
      Exit;
    Value.AsBoolean := False;
    if ValCur.Compare(ValMin) in [cmprEqual, cmprGreater] then
      if ValCur.Compare(ValMax) in [cmprEqual, cmprLower] then
         Value.AsBoolean := True;
  finally
    ValCur.Free;
    ValMin.Free;
    ValMax.Free;
  end;
end;//Between


//------------------------------------------------------------------------------
// return Value
//------------------------------------------------------------------------------
function TACRExprNodeBoolean.GetDataValue: TACRVariant;
var
  i: Integer;
begin
  Value.SetNull(bftLogical);

   case Operator of
    doNOT: // NOT
      begin
        TempVal.Assign(TACRExprNode(Children[0]).GetDataValue, false);
        // NOT <null> = NULL
        if not TempVal.IsNull then
          Value.AsBoolean := not TempVal.AsBoolean;
      end;
    doAND: // AND
      begin
        // possible result is TRUE
        Value.AsBoolean := True;
        // Loop for all childs
        for i:=0 to Children.Count-1 do
         begin
          TempVal.Assign(TACRExprNode(Children[i]).GetDataValue, false);

          // x AND NULL = NULL (and check later)
          if TempVal.IsNull then
           begin
            Value.SetNull(bftLogical);
            //break;
            continue;
           end;

          // x AND FALSE = FALSE
          if (not TempVal.AsBoolean) then
           begin
            Value.AsBoolean := False;
            break;
           end;
         end;//for
      end;
    doOR: // OR
      begin
        // possible result is FALSE
        Value.AsBoolean := False;
        // Loop for all childs
        for i:=0 to Children.Count-1 do
         begin
          TempVal.Assign(TACRExprNode(Children[i]).GetDataValue, false);

          // False OR NULL = NULL (and check later)
          if TempVal.IsNull then
           begin
            Value.SetNull(bftLogical);
            continue;
           end;

          // x OR TRUE=TRUE;
          if (TempVal.AsBoolean) then
           begin
            Value.AsBoolean := True;
            break;
           end;
         end;
      end;
    doISNULL:  // IS NULL
      begin
        Value.AsBoolean := TACRExprNode(Children[0]).GetDataValue.IsNull;
      end;
    doISNOTNULL: // IS NOT NULL
      begin
        Value.AsBoolean := not TACRExprNode(Children[0]).GetDataValue.IsNull;
      end;
    doLIKE:
      begin
        Like;
      end;
    doNOTLIKE:
      begin
        Like;
        if (not Value.IsNull) then
          Value.AsBoolean := not Value.AsBoolean;
      end;
    doIN:
      begin
        bIn;
      end;
    doNOTIN:
      begin
        bIn;
        if (not Value.IsNull) then
          Value.AsBoolean := not Value.AsBoolean;
      end;
    doBETWEEN:
      begin
        Between;
      end;
    doNOTBETWEEN:
      begin
        Between;
        if (not Value.IsNull) then
          Value.AsBoolean := not Value.AsBoolean;
      end;
    else
      raise EACRException.Create(30116, ErrorGUnknownOperator,
                                                   [GetOperatorName(Operator)]);
   end;
  Result := Value;
end;//GetDataValue



////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeArithmetic
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// ADD operation
//------------------------------------------------------------------------------
procedure TACRExprNodeArithmetic.AddData;
var
  i: Integer;
begin
  // Get first argument
  Value.Assign(TACRExprNode(Children[0]).GetDataValue, False);
  // Add next values
  for i:=1 to Children.Count-1 do
    Value.Add(TACRExprNode(Children[i]).GetDataValue);
end;//AddData


//------------------------------------------------------------------------------
// SUB operation
//------------------------------------------------------------------------------
procedure TACRExprNodeArithmetic.SubData;
begin
  Value.Assign(TACRExprNode(Children[0]).GetDataValue, False);
  Value.InvertValue;
end;//SubData


//------------------------------------------------------------------------------
// MUL operation
//------------------------------------------------------------------------------
procedure TACRExprNodeArithmetic.MulData;
var
  i: Integer;
begin
  // Get first argument
  Value.Assign(TACRExprNode(Children[0]).GetDataValue, False);
  // Mul next values
  for i:=1 to Children.Count-1 do
    Value.Mul(TACRExprNode(Children[i]).GetDataValue);
end;//MulData


//------------------------------------------------------------------------------
// DIV operation
//------------------------------------------------------------------------------
procedure TACRExprNodeArithmetic.DivData;
begin
  // Get first argument
  Value.Assign(TACRExprNode(Children[0]).GetDataValue, False);
  Value.Division(TACRExprNode(Children[1]).GetDataValue);
end;//DivData


//------------------------------------------------------------------------------
// Return Data Value
//------------------------------------------------------------------------------
function TACRExprNodeArithmetic.GetDataValue: TACRVariant;
begin
  case Operator of
   doSUB:       SubData;
   doADD:       AddData;
   doMUL:       MulData;
   doDIV:       DivData;
   else
     raise EACRException.Create(30121, ErrorGUnknownOperator,
                                [GetOperatorName(Operator)]);
  end;
  Result := Value;
end;//GetDataValue



////////////////////////////////////////////////////////////////////////////////
//
// TACRExprNodeStringFunction
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// InitStringValue and return IsNull
//------------------------------------------------------------------------------
function TACRExprNodeStringFunction.InitStringValue;
begin
  Value.Assign(TACRExprNode(Children[0]).GetDataValue, False);
  // Guarantee String or WideString type
  if not Value.IsStringDataType then
    Value.Cast(bftVarchar);
  Result := not Value.IsNull;
end;//InitStringValue


//------------------------------------------------------------------------------
// Concat
//------------------------------------------------------------------------------
procedure TACRExprNodeStringFunction.Concat;
var i: Integer;
begin
  InitStringValue;
  // Concatenation
  for i:=1 to Children.Count-1 do
   Value.Add(TACRExprNode(Children[i]).GetDataValue);
end;//Concat


//------------------------------------------------------------------------------
// Upper
//------------------------------------------------------------------------------
procedure TACRExprNodeStringFunction.Upper;
begin
  if (not InitStringValue) then
    Exit;
  // WideString ?
  if Value.IsWideStringDataType then
    Value.AsWideString := AnsiUpperCase(Value.AsWideString)
  else
    Value.AsString := AnsiUpperCase(Value.AsString)
end;//Upper


//------------------------------------------------------------------------------
// Lower
//------------------------------------------------------------------------------
procedure TACRExprNodeStringFunction.Lower;
begin
  if (not InitStringValue) then
    Exit;
  // WideString ?
  if Value.IsWideStringDataType then
    Value.AsWideString := AnsiLowerCase(Value.AsWideString)
  else
    Value.AsString := AnsiLowerCase(Value.AsString)
end;//Lower


//------------------------------------------------------------------------------
// Trim
//------------------------------------------------------------------------------
procedure TACRExprNodeStringFunction.Trim;
begin
  if (not InitStringValue) then
    Exit;
  // WideString ?
  if Value.IsWideStringDataType then
    Value.AsWideString := SysUtils.Trim(Value.AsWideString)
  else
    Value.AsString := SysUtils.Trim(Value.AsString)
end;//Trim


//------------------------------------------------------------------------------
// LTrim
//------------------------------------------------------------------------------
procedure TACRExprNodeStringFunction.LTrim;
begin
  if (not InitStringValue) then
    Exit;
  // WideString ?
  if Value.IsWideStringDataType then
    Value.AsWideString := SysUtils.TrimLeft(Value.AsWideString)
  else
    Value.AsString := SysUtils.TrimLeft(Value.AsString)
end;//LTrim


//------------------------------------------------------------------------------
// RTrim
//------------------------------------------------------------------------------
procedure TACRExprNodeStringFunction.RTrim;
begin
  if (not InitStringValue) then
    Exit;
  // WideString ?
  if Value.IsWideStringDataType then
    Value.AsWideString := SysUtils.TrimRight(Value.AsWideString)
  else
    Value.AsString := SysUtils.TrimRight(Value.AsString)
end;//RTrim


//------------------------------------------------------------------------------
// Length
//------------------------------------------------------------------------------
procedure TACRExprNodeStringFunction.Length;
var
  TmpVal: TACRVariant;
begin
  //Value.SetNull(bftSignedInt32);
  Value.AsInteger := 0;
  TmpVal := TACRVariant.Create;
  try
    TmpVal.Assign(TACRExprNode(Children[0]).GetDataValue, False);
    // Guarantee String or WideString type
    if (not TmpVal.IsStringDataType) then
      TmpVal.Cast(bftVarchar);
    // IsNull ?
    if (not TmpVal.IsNull) then
      case TmpVal.DataType of
        bftChar,
        bftVarchar:
          Value.AsInteger := System.Length(TmpVal.AsString);
        bftWideChar,
        bftWideVarchar:
          Value.AsInteger := System.Length(TmpVal.AsWideString);
      end;
  finally
    TmpVal.Free;
  end;
end;//Length


//------------------------------------------------------------------------------
// Pos
//------------------------------------------------------------------------------
procedure TACRExprNodeStringFunction.Pos;
var
  StrValue, SubstrValue: TACRVariant;
begin
  StrValue := TACRVariant.Create;
  SubstrValue := TACRVariant.Create;
  try
    StrValue.Assign(TACRExprNode(Children[0]).GetDataValue, False);
    SubstrValue.Assign(TACRExprNode(Children[1]).GetDataValue, False);
    if (StrValue.IsNull or SubstrValue.IsNull) then
     begin
      Value.AsInteger := 0;
      Exit;
     end;    // Guarantee String or WideString type
    if (not StrValue.IsStringDataType) then
      StrValue.Cast(bftVarchar);
    if (not SubStrValue.IsStringDataType) then
      SubStrValue.Cast(bftVarchar);
    case StrValue.DataType of
      bftChar,
      bftVarchar:
        case SubStrValue.DataType of
          bftChar,
          bftVarchar:
            Value.AsInteger := System.pos(StrValue.AsString, SubstrValue.AsString);
          bftWideChar,
          bftWideVarchar:
            Value.AsInteger := System.pos(StrValue.AsString, SubstrValue.AsWideString);
        end;
      bftWideChar,
      bftWideVarchar:
        case SubStrValue.DataType of
          bftChar,
          bftVarchar:
            Value.AsInteger := System.pos(StrValue.AsWideString, SubstrValue.AsString);
          bftWideChar,
          bftWideVarchar:
            Value.AsInteger := System.pos(StrValue.AsWideString, SubstrValue.AsWideString);
        end
    end;
  finally
    StrValue.Free;
    SubstrValue.Free;
  end;
end;//Pos


//------------------------------------------------------------------------------
// Substring
//------------------------------------------------------------------------------
procedure TACRExprNodeStringFunction.Substring;
var
  Start, Len: Integer;
begin
  // Start Position
  Start := TACRExprNode(Children[1]).GetDataValue.AsInteger;

  // length
  if Children.Count >=3 then
   Len := TACRExprNode(Children[2]).GetDataValue.AsInteger
  else
   begin
     Length;
     len := Value.AsInteger;
   end;

  // String
  Value.Assign(TACRExprNode(Children[0]).GetDataValue, False);
  // String is null ?
  if (Value.IsNull) then
    Exit;
  // Guarantee String or WideString Type
  if (not Value.IsStringDataType) then
    Value.Cast(bftVarchar);

  // Get Substring
  if Value.IsWideStringDataType then
    Value.AsWideString := copy(Value.AsWideString, Start, Len)
  else
    Value.AsString := copy(Value.AsString, Start, Len);

end;//Substring


//------------------------------------------------------------------------------
// Return Data Value
//------------------------------------------------------------------------------
function TACRExprNodeStringFunction.GetDataValue: TACRVariant;
begin
  case Operator of
   doCONCAT:    Concat;
   doUPPER:     Upper;
   doLOWER:     Lower;
   doTRIM:      Trim;
   doLTRIM:     LTrim;
   doRTRIM:     RTrim;
   doSUBSTRING: Substring;
   doPOS:       Pos;
   doLENGTH:    Length;
   else
     raise EACRException.Create(30131, ErrorGUnknownOperator,
                                 [GetOperatorName(Operator)]);
  end;
  Result := Value;
end;//GetDataValue


//------------------------------------------------------------------------------
// return Type of Data
//------------------------------------------------------------------------------
function TACRExprNodeStringFunction.GetDataType: TACRAdvancedFieldType;
begin
  if Operator in [doPOS, doLENGTH] then
    Result := aftInteger
  else
    Result := inherited GetDataType;
end;//GetDataType


//------------------------------------------------------------------------------
// get Data Size
//------------------------------------------------------------------------------
function TACRExprNodeStringFunction.GetDataSize: Integer;
var i: Integer;
begin
  Result := 0;
  case Operator of
    doUPPER,
    doLOWER,
    doSUBSTRING,
    doTRIM,
    doLTRIM,
    doRTRIM:
      Result :=  TACRExprNode(Children[0]).GetDataSize;
    doCONCAT:
      begin
        for i:=0 to Children.Count-1 do
          Result := Result + TACRExprNode(Children[i]).GetDataSize;
      end;
  end;
end;//GetDataSize




//------------------------------------------------------------------------------
//FillFieldNames
//------------------------------------------------------------------------------
procedure FillFieldNames(List: TStringList; const Names: String);

  function ExtractFieldName(const Fields: string; var Pos: Integer): string;
  var
    I: Integer;
  begin
    I := Pos;
    while (I <= Length(Fields)) and (Fields[I] <> ';') do Inc(I);
    Result := Trim(Copy(Fields, Pos, I - Pos));
    if (I <= Length(Fields)) and (Fields[I] = ';') then Inc(I);
    Pos := I;
  end;//ExtractFieldName

var
  Pos: Integer;
begin
  List.Clear;
  Pos := 1;
  while Pos <= Length(Names) do
    List.Add(AnsiUpperCase(ExtractFieldName(Names, Pos)));
end;//FillFieldNames


initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRBaseEngine> initialized');
{$ENDIF}

end.
