unit ACRLexer;

{$I ACRVer.inc}

interface

uses Classes, SysUtils, DB,
{$IFDEF MSWINDOWS}
     Windows,
{$ENDIF}
{$IFDEF DEBUG_LOG}
     ACRDebug,
{$ENDIF}
     ACRVariant,
     ACRCompression,
     ACRExcept,
     ACRConst;

const 
EACRTokenType: array [0..11] of string =
(
 'tktNone', 'tktString', 'tktQuotedString', 'tktBracketedString',
 'tktInt', 'tktFloat', 'tktReservedWord',
 'tktParameter', 'tktLeftParenthesis', 'tktRightParenthesis',
 'tktComma', 'tktDot'
);

type
 TTokenType = (tktNone, tktString, tktQuotedString, tktBracketedString,
 tktInt, tktFloat, tktReservedWord,
 tktParameter, tktLeftParenthesis, tktRightParenthesis,
 tktComma, tktDot);

 TTokenTypes = set of TTokenType;

 TToken = record
  TokenType:      TTokenType;
  ReservedWord:   TReservedWord;
  Text:           String;
  ParamValue:     TACRVariant;
  LineNum:        Integer; // number of line in script where token begins
  ColumnNum:      Integer; // number of column in script where token begins
 end;

 PToken = ^TToken;

 TSQLCommand = record
  Tokens:     array of TToken;
  NumTokens:  Integer;
  CurrentTokenNo: Integer;
 end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLParam
//
////////////////////////////////////////////////////////////////////////////////

  TACRSQLParam = class (TACRVariant)
   public
    Name:  String;
   public
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
    procedure Assign(Source: TACRVariant; CopyDataFlag: boolean = true); override;
  end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLParams
//
////////////////////////////////////////////////////////////////////////////////

  TACRSQLParams = class (TObject)
   private
    FParamList: TList;
   private
    function GetCount: Integer;
    function GetValue(Index: Integer): TACRSQLParam;
    procedure SetValue(Index: Integer; Value: TACRSQLParam);
   public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function AddCreated: TACRSQLParam;
    procedure Assign(Source: TACRSQLParams); 
    function GetParamByName(Name: String): TACRSQLParam;
    procedure LoadFromStream(Stream: TACRStream);
    procedure SaveToStream(Stream: TACRStream);
   public
    property Count: Integer read GetCount;
    property Items[Index: Integer]: TACRSQLParam read GetValue write SetValue;
  end;//TACRSQLParams




 TACRLexer = class (TObject)
  private
   FSQL:              String;
   FSQLParams:        TACRSQLParams;
   FSQLScriptLength:  Integer;
   FCurrentPos:       Integer;
   FCurrentCommandNo: Integer;
   FCurrentCommand:   TSQLCommand;
  private
   procedure FreeCurrentCommand;
   function Parse: Boolean;
  public
   constructor Create(SQLScript: string; SQLParams: TACRSQLParams = nil);

   // makes next command current
   function GetNextCommand: Boolean;
   // gets next token in current command
   function GetNextToken(var Token: TToken): Boolean;
   // gets current token in current command
   function GetCurrentToken(var Token: TToken): Boolean;
   // looks at next token
   function LookNextToken(var Token: TToken): Boolean;
   // gets current token No
   function GetCurrentTokenNo: integer;
   // sets current token No
   function SetCurrentTokenNo(TokenNo: integer; var Token: TToken): Boolean;
  public
   property CurrentCommand: TSQLCommand read FCurrentCommand;
 end;

 // checks whether token is reserved word
 function IsReservedWord(Token: TToken; ReservedWord: TReservedWord=rwNone): Boolean;

implementation


////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLParam
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// LoadFromStream
//------------------------------------------------------------------------------
procedure TACRSQLParam.LoadFromStream(Stream: TStream);
var Len: Integer;
begin
  LoadDataFromStream(Len,SizeOf(Len),Stream,11190);
  SetLength(Name,Len);
  LoadDataFromStream(PChar(@Name[1])^,Len,Stream,11191);
  inherited LoadFromStream(Stream);
end; // LoadFromStream


//------------------------------------------------------------------------------
// SaveToStream
//------------------------------------------------------------------------------
procedure TACRSQLParam.SaveToStream(Stream: TStream);
var Len: Integer;
begin
  Len := Length(Name);
  SaveDataToStream(Len,Sizeof(Len),Stream,11192);
  SaveDataToStream(PChar(@Name[1])^,Len,Stream,11193);
  inherited SaveToStream(Stream);
end; // SaveToStream


//------------------------------------------------------------------------------
// assign
//------------------------------------------------------------------------------
procedure TACRSQLParam.Assign(Source: TACRVariant; CopyDataFlag: boolean = true);
begin
  inherited Assign(Source,CopyDataFlag);
  if (Source is TACRSQLParam) then
    Name := TACRSQLParam(Source).Name;
end; // Assign


////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLParams
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSQLParams.Create;
begin
  FParamList := TList.Create;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRSQLParams.Destroy;
begin
  Clear;
  FParamList.Free;
end;//Destroy


//------------------------------------------------------------------------------
// return Count
//------------------------------------------------------------------------------
function TACRSQLParams.GetCount: Integer;
begin
  Result := FParamList.Count;
end;//GetCount


//------------------------------------------------------------------------------
// GetValue
//------------------------------------------------------------------------------
function TACRSQLParams.GetValue(Index: Integer): TACRSQLParam;
begin
  Result := TACRSQLParam(FParamList[Index]);
end;//GetValue


//------------------------------------------------------------------------------
// SetValue
//------------------------------------------------------------------------------
procedure TACRSQLParams.SetValue(Index: Integer; Value: TACRSQLParam);
begin
  TACRSQLParam(FParamList[Index]).Free;
  FParamList[Index] := Value;
end;//SetValue


//------------------------------------------------------------------------------
// AddCreated
//------------------------------------------------------------------------------
function TACRSQLParams.AddCreated: TACRSQLParam;
begin
  Result := TACRSQLParam.Create;
  FParamList.Add(Result);
end;//AddCreated


//------------------------------------------------------------------------------
// assign
//------------------------------------------------------------------------------
procedure TACRSQLParams.Assign(Source: TACRSQLParams);
var i: Integer;
begin
  Clear;
  for i := 0 to Source.Count - 1 do
    AddCreated.Assign(Source.Items[i]);
end; // Assign


//------------------------------------------------------------------------------
// Clear
//------------------------------------------------------------------------------
procedure TACRSQLParams.Clear;
var i: Integer;
begin
  for i:=0 to Count-1 do
    TACRSQLParam(FParamList[i]).Free;
  FParamList.Clear;
end;//Clear


//------------------------------------------------------------------------------
// GetParamByName
//------------------------------------------------------------------------------
function TACRSQLParams.GetParamByName(Name: String): TACRSQLParam;
var i: Integer;
begin
  Result := nil;
  for i:=0 to Count-1 do
   if (AnsiUpperCase(items[i].Name) = AnsiUpperCase(Name)) then
    begin
     Result := items[i];
     Break;
    end;
end;//GetParamByName


//------------------------------------------------------------------------------
// LoadFromStream
//------------------------------------------------------------------------------
procedure TACRSQLParams.LoadFromStream(Stream: TACRStream);
var num,i: Integer;
begin
  LoadDataFromStream(num,SizeOf(num),Stream,11188);
  Clear();
  for i := 0 to num - 1 do
    AddCreated.LoadFromStream(Stream);
end; // LoadFromStream


//------------------------------------------------------------------------------
// SaveToStream
//------------------------------------------------------------------------------
procedure TACRSQLParams.SaveToStream(Stream: TACRStream);
var num,i: Integer;
begin
  num := Count;
  SaveDataToStream(num,SizeOf(num),Stream,11189);
  for i := 0 to num - 1 do
    Items[i].SaveToStream(Stream);
end; // SaveToStream


////////////////////////////////////////////////////////////////////////////////
//
// TACRLexer
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// free current command
//------------------------------------------------------------------------------
procedure TACRLexer.FreeCurrentCommand;
var j: Integer;
begin
  for j := 0 to FCurrentCommand.NumTokens-1 do
   if (FCurrentCommand.Tokens[j].ParamValue <> nil) then
     FCurrentCommand.Tokens[j].ParamValue.Free;
end; // FreeCurrentCommand


//------------------------------------------------------------------------------
// Parse SQL script
//------------------------------------------------------------------------------
function TACRLexer.Parse: Boolean;
var i,l:                Integer;
    line,token,column:  Integer;
    c,priorSymbol:      Char;
    numParenthesis,
    LeftParenthesisLine,
    LeftParenthesisColumn,
    RightParenthesisLine,
    RightParenthesisColumn,
    quoteLine,quoteColumn: Integer;
    bIsDelimiter:       Boolean;
    bTokenStarted:      Boolean;
    bCommandStarted:    Boolean;
    bNewTokenStarted:   Boolean;
    bTokenFinished:     Boolean;
    bQuoteNotClosed:    Boolean;
    FNewCommandStarted: Boolean;

  function GetNextSymbol: char;
  begin
   result := ' ';
    if (i < l-1) then
     result := pChar(pChar(FSQL)+i+1)^;
  end;


 function IsNewTokenStarted: Boolean;
 begin
  result := false;
  if (bTokenStarted) then
   begin
      if (c = '>') or (c = '<') or (c = '=') or (c = '(') or (c = ')') or
         (c = ',') or (c = '/') or (c = '*') or (c = Colon) or (c = '|') or
         ((c=Dot) and (FCurrentCommand.Tokens[token].TokenType<>tktInt)) then
           result := true
      else
      if (FCurrentCommand.Tokens[token].TokenType = tktDot) then
       begin
         if not ((c >= '0') and (c <= '9')) then
          result := true;
       end
      else
      if (FCurrentCommand.Tokens[token].TokenType = tktString) then
       begin
        if (c = '+') or (c = '-') then
         result := true;
       end
      else
      if (FCurrentCommand.Tokens[token].TokenType = tktInt) then
       begin
        if (c = '+') and (AnsiUpperCase(priorSymbol) <> 'E') then
         result := true 
        else 
        if (c = '-') and (AnsiUpperCase(priorSymbol) <> 'E') then
         result := true;
       end; // Int or Float token 
   end // some token already started 
  else
   begin
    if (not bIsDelimiter) then 
     result := true;
   end; // no token started 
 end; // IsNewTokenStarted
 
 function IsTokenFinished: Boolean;
 begin 
  result := false;
  if (bTokenStarted) then 
   begin 
    if (bIsDelimiter) then
     result := true; 
   end; 
 end; // IsTokenFinished 
 
 function FindReservedWord: Integer; 
 var i: integer;
     s: string;
 begin
  result := -1;
  s := AnsiUpperCase(FCurrentCommand.Tokens[token].Text);
  for i := 0 to ACRMaxSQLReservedWords do 
    if (s = ACRSQLReservedWords[i]) then
     begin 
      result := i;
      break; 
     end; 
 end; // FindReservedWord
 
 procedure CreateCommand;
 begin 
  FNewCommandStarted := True; 
  FCurrentCommand.NumTokens := 0;
  FCurrentCommand.Tokens := nil;
  token := 0; 
  bCommandStarted := true; 
  numParenthesis := 0; 
 end; // CreateCommand; 
 
 procedure CloseToken; 
 var i:         Integer; 
     Param:     TACRSQLParam;
     ParamName: String;
//     oldSeparator: char; 
 begin
  if (not bTokenStarted) then
   Exit; 
  // check int or float
  if (FCurrentCommand.Tokens[token].TokenType = tktInt) then
   begin 
    // check int or float 
    if (Pos('.',FCurrentCommand.Tokens[token].Text) > 0) then 
     FCurrentCommand.Tokens[token].TokenType := tktFloat
    else
    if (Pos('E',AnsiUpperCase(FCurrentCommand.Tokens[token].Text)) > 0) then 
     FCurrentCommand.Tokens[token].TokenType := tktFloat;
   end;
  // String 
  if (FCurrentCommand.Tokens[token].TokenType = tktString) then 
   begin
    // check for SQLParameter ( :name)
    if (FCurrentCommand.Tokens[token].Text[1] = Colon) then
     begin 
      FCurrentCommand.Tokens[token].TokenType := tktParameter; 
      // param name without ':' 
      ParamName :=Copy(FCurrentCommand.Tokens[token].Text, 2, 
                       Length(FCurrentCommand.Tokens[token].Text)-1);
      FCurrentCommand.Tokens[token].Text := ParamName; 
 
      // try to set param value 
      Param := FSQLParams.GetParamByName(ParamName);
      if (Param <> nil) then
        begin
          FCurrentCommand.Tokens[token].ParamValue := TACRVariant.Create; 
          FCurrentCommand.Tokens[token].ParamValue.Assign(Param);
        end
      else
        raise EACRException.Create(30358, ErrorGParameterValueNotFound, [ParamName]); 
     end 
    else
     begin
      // check for reserved word
      i := FindReservedWord;
      if (i > 0) then
       begin
        FCurrentCommand.Tokens[token].ReservedWord := TReservedWord(i); 
        FCurrentCommand.Tokens[token].TokenType := tktReservedWord;
       end;
     end; 
   end;// if String
  bTokenStarted := false;
 end; // CloseToken; 
 
 procedure CreateToken; 
 var NextSymbol: char; 
     quoteSymbol: char;
 begin 
  if (not bCommandStarted) then 
   CreateCommand;
  bTokenStarted := true;
  inc(FCurrentCommand.NumTokens);
  SetLength(FCurrentCommand.Tokens,FCurrentCommand.NumTokens);
  FCurrentCommand.Tokens[FCurrentCommand.NumTokens-1].ParamValue := nil; 
 
  token := FCurrentCommand.NumTokens - 1;
  FCurrentCommand.Tokens[token].Text := c;
  FCurrentCommand.Tokens[token].ColumnNum := column; 
  FCurrentCommand.Tokens[token].LineNum := line; 
  FCurrentCommand.Tokens[token].ReservedWord := rwNone; 
  FCurrentCommand.Tokens[token].ParamValue := nil;
	
  // here will be scanning quoted string to its end
  if (c = SingleQuote) or (c = DoubleQuote) or (c = LeftBracket) then
   begin
    FCurrentCommand.Tokens[token].Text := ''; 
    if (c <> LeftBracket) then
     FCurrentCommand.Tokens[token].TokenType := tktQuotedString 
    else
     FCurrentCommand.Tokens[token].TokenType := tktBracketedString;
    if (c <> LeftBracket) then
     quoteSymbol := c
    else
     quoteSymbol := RightBracket;
    quoteLine := line;
    quoteColumn := column;
    bQuoteNotClosed := true;
    inc(i);
    while (i < l) do
     begin
      c := pChar(pChar(FSQL)+i)^;
      NextSymbol := GetNextSymbol;
      if (c = quoteSymbol) then
       begin
        if (NextSymbol = c) then
         begin
          FCurrentCommand.Tokens[token].Text :=
            FCurrentCommand.Tokens[token].Text + c;
          inc(i,2);
          inc(column,2);
          continue;
         end // ''
        else
         begin
          // end of QuotedStr
          bQuoteNotClosed := false;
          break;
         end; // '
       end; // quote symbol

      FCurrentCommand.Tokens[token].Text :=
         FCurrentCommand.Tokens[token].Text + c;
      if (c = lf) then
       begin
        inc(line);
        column := 1;
       end
      else
       if (c <> cr) then
        inc(column);
      inc(i);
     end; // while
    CloseToken;
   end
  else
  if (c = LeftParenthesis) then
   begin
    FCurrentCommand.Tokens[token].TokenType := tktLeftParenthesis;
    CloseToken;
    inc(numParenthesis);
    LeftParenthesisLine := line;
    LeftParenthesisColumn := column;
   end
  else
  if (c = RightParenthesis) then
   begin
    dec(numParenthesis);
    RightParenthesisLine := line;
    RightParenthesisColumn := column;
    FCurrentCommand.Tokens[token].TokenType := tktRightParenthesis;
    CloseToken;
   end
  else
  if (c = Dot) then //!!
   begin
    FCurrentCommand.Tokens[token].TokenType := tktDot;
    //CloseToken;
   end
  else
  if (c = Comma) then
   begin
    FCurrentCommand.Tokens[token].TokenType := tktComma;
    CloseToken;
   end
  else
  if ((c >= '0') and (c <= '9')) then
    FCurrentCommand.Tokens[token].TokenType := tktInt
  else
   FCurrentCommand.Tokens[token].TokenType := tktString;
  if (c = '-') or (c = '+') or (c = '*') or (c = '/') or (c = '=') then
   CloseToken;
  if (c = '|') then
   begin
    NextSymbol := GetNextSymbol;
    if (NextSymbol = '|') then
     begin
      FCurrentCommand.Tokens[token].Text :=
        FCurrentCommand.Tokens[token].Text + NextSymbol;
      inc(i);
      inc(column);
     end;
    CloseToken;
   end; // <
  if (c = '<') then
   begin
    NextSymbol := GetNextSymbol;
    if (NextSymbol = '=') or (NextSymbol = '>') then
     begin
      FCurrentCommand.Tokens[token].Text :=
        FCurrentCommand.Tokens[token].Text + NextSymbol;
      inc(i);
      inc(column);
     end;
    CloseToken;
   end; // <
  if (c = '>') then
   begin
    NextSymbol := GetNextSymbol;
    if (NextSymbol = '=') then
     begin
      FCurrentCommand.Tokens[token].Text :=
        FCurrentCommand.Tokens[token].Text + NextSymbol;
      inc(i);
      inc(column);
     end;
    CloseToken;
   end; // >
 end; // CreateToken;

 procedure CloseCommand;
 begin
  CloseToken;
  bCommandStarted := false;

  if (numParenthesis > 0) then
    raise EACRException.Create(30059, ErrorGMissingRightParenthesis,
                                  [LeftParenthesisLine,LeftParenthesisColumn]);
  if (numParenthesis < 0) then
    raise EACRException.Create(30060, ErrorGUnexpectedRightParenthesis,
                                  [LeftParenthesisLine,LeftParenthesisColumn]);
  priorSymbol := ' ';

  if (bQuoteNotClosed) then
    raise EACRException.Create(30061, ErrorGUnterminatedString,
                                                       [quoteLine,quoteColumn]);
 end; // CloseCommand;

 var NextSymbol: char;
begin
 FNewCommandStarted := False;
 l := FSQLScriptLength;
 line := 1;
 column := 1;
 bTokenStarted := false;
 bQuoteNotClosed := false;
 bCommandStarted := false;
 token := 0;
 priorSymbol := ' ';
 numParenthesis := 0;
 LeftParenthesisLine := 0;
 LeftParenthesisColumn := 0;
 RightParenthesisLine := 0;
 RightParenthesisColumn := 0;
 i := FCurrentPos;
 while (i < l) do
  begin
   // get current character
   c := pChar(pChar(FSQL)+i)^;

   // end of the current comand
   if (c = ';') then
    begin
     CloseCommand;
     inc(i);
     break;
    end;

   // check for comment
   if (c = '-') then
    begin
     NextSymbol := GetNextSymbol;
     if (NextSymbol = '-') then
      begin
       // seek for end of line
       inc(i);
       column := 1;
       inc(line);
       while (i < l) do
        begin
         c := pChar(pChar(FSQL)+i)^;
         if (c = lf) then
           break;
         inc(i);
        end;
       inc(i);
       continue;
      end;
    end; // comment --

   // check for comment
   if (c = '/') then
    begin
     NextSymbol := GetNextSymbol;
     if (NextSymbol = '*') then
      begin
       // commentary started
       inc(i,2);
       inc(column,2);
       while (i < l) do
        begin
          c := pChar(pChar(FSQL)+i)^;
          if (c = '*') then
           begin
             NextSymbol := GetNextSymbol;
             if (NextSymbol = '/') then
              begin
               inc(i,2);
               inc(column,2);
               break;
              end;
           end;
          if (c = lf) then
           begin
            inc(line);
            column := 1;
           end
          else
           if (c <> cr) then
            inc(column);
          inc(i);
        end; // while
       continue;
      end; // comment /* */ started
    end;


   // is this symbol a delimiter?
   bIsDelimiter := false;
   if (c = Tab) or (c = CR) or (c = LF) or (c = Space) then
    bIsDelimiter := true;

   bTokenFinished := IsTokenFinished;
   bNewTokenStarted := IsNewTokenStarted;
   if (bTokenFinished) or (bTokenStarted and bNewTokenStarted) then
    CloseToken;
   if (bNewTokenStarted) then
    CreateToken;

   // add current symbol to token
   if (not bNewTokenStarted) and (bTokenStarted) and (not bIsDelimiter) then
    begin
     // check for number
     if FCurrentCommand.Tokens[token].TokenType in [tktInt,tktFloat] then
      if not ((c >= '0') and (c <= '9') or (c = Dot) or
              (AnsiUpperCase(c)='E') or (c = '+') or (c = '-') ) then
        raise EACRException.Create(30062, ErrorGInvalidNumericSymbol,
                                                              [c,line,column]);
      // check float type .22
      if (FCurrentCommand.Tokens[token].TokenType = tktDot) and
         (c >= '0') and (c <= '9') then
       begin
        FCurrentCommand.Tokens[token].TokenType := tktInt;
       end;
     // add current symbol to token
    FCurrentCommand.Tokens[token].Text :=
      FCurrentCommand.Tokens[token].Text + c;
    end;

   if (c = lf) then
    begin
     inc(line);
     column := 1;
    end
   else
    if (c <> cr) then
     inc(column);
   // skip delimiters
   if (bIsDelimiter) then
    begin
     if (bTokenStarted) then
      CloseToken;
    end;
   // next symbol
   priorSymbol := c;
   inc(i);
  end; // for all symbols in FSQL
 FCurrentPos := i;
 CloseCommand;
 Result := FNewCommandStarted;
end; // Parse


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRLexer.Create(SQLScript: string; SQLParams: TACRSQLParams);
begin
  FSQL := SQLScript;
  FSQLParams := SQLParams;
  FCurrentPos := 0;
  FCurrentCommandNo := -1;
  FSQLScriptLength := Length(FSQL);
end; // Create


//------------------------------------------------------------------------------
// makes next command current
//------------------------------------------------------------------------------
function TACRLexer.GetNextCommand: Boolean;
begin
 if (FCurrentCommandNo >= 0) then
  FreeCurrentCommand;
 Inc(FCurrentCommandNo);
 Result := Parse;
 FCurrentCommand.CurrentTokenNo := 0;
end;// GetNextCommand

//------------------------------------------------------------------------------
// gets next token in current command
//------------------------------------------------------------------------------
function TACRLexer.GetNextToken(var Token: TToken): Boolean;
begin
  Result := LookNextToken(Token);
  Inc(FCurrentCommand.CurrentTokenNo);
end;// GetNextToken


//------------------------------------------------------------------------------
// gets current token in current command
//------------------------------------------------------------------------------
function TACRLexer.GetCurrentToken(var Token: TToken): Boolean;
begin
   Result := (FCurrentCommand.CurrentTokenNo < FCurrentCommand.NumTokens);
   if (Result) then
    Token := FCurrentCommand.Tokens[FCurrentCommand.CurrentTokenNo]
   else
    Token.TokenType := tktNone;
end;// GetCurrentToken


//------------------------------------------------------------------------------
// looks at next token in current command
//------------------------------------------------------------------------------
function TACRLexer.LookNextToken(var Token: TToken): Boolean;
begin
   Result := (FCurrentCommand.CurrentTokenNo+1 < FCurrentCommand.NumTokens);
   if (Result) then
    Token := FCurrentCommand.Tokens[FCurrentCommand.CurrentTokenNo+1]
   else
    Token.TokenType := tktNone;
end;// LookNextToken


//------------------------------------------------------------------------------
// gets current token No
//------------------------------------------------------------------------------
function TACRLexer.GetCurrentTokenNo: integer;
begin
  Result := FCurrentCommand.CurrentTokenNo;
end;// GetCurrentTokenNo


//------------------------------------------------------------------------------
// sets current token No
//------------------------------------------------------------------------------
function TACRLexer.SetCurrentTokenNo(TokenNo: integer; var Token: TToken): Boolean;
begin
 FCurrentCommand.CurrentTokenNo := TokenNo;
 Result := GetCurrentToken(Token);
end;// SetCurrentTokenNo


// checks whether token is reserved word
function IsReservedWord(Token: TToken; ReservedWord: TReservedWord=rwNone): Boolean;
begin
 Result := False;
 if (Token.TokenType = tktReservedWord) then
  if ((Token.ReservedWord = ReservedWord) or
      (ReservedWord = rwNone)) then
   Result := True;
end;

initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRLexer> initialized');
{$ENDIF}

end.

