unit ACRMemory;

{$I ACRVer.inc}

interface

uses SysUtils, Classes,
{$IFDEF MSWINDOWS}
     Windows,
{$ENDIF}
{$IFDEF LINUX}
     Libc,
{$ENDIF}
     {$IFDEF DEBUG_LOG}
     ACRDebug,
     {$ENDIF}
     ACRCriticalSection,
     ACRExcept, ACRConst;


type

 TACRMemorySize = Integer;

////////////////////////////////////////////////////////////////////////////////
//
// TACRMemoryManager
//
////////////////////////////////////////////////////////////////////////////////

 TACRMemoryManager = class (TObject)
  private
   FMaxMemorySize:        TACRMemorySize;  // Max Memory Limit
   FTotalMemAllocated:    TACRMemorySize;  // Total allocated memory
   FMaxMemAllocated:      TACRMemorySize;

   FFreeSystemMemorySize: TACRMemorySize;  // Free Memory size in system

   FAllocMemCallCount:    Integer;  // count of allocmem calls
   FGetMemCallCount:      Integer;  // count of getmem calls
   FFreeMemCallCount:     Integer;  // count of freemem calls
   FReallocMemCallCount:  Integer;  // count of reallocmem calls

   MMLock: TRTLCriticalSection;
  public
   // Constructor
   constructor Create; overload;
   // Constructor
   constructor Create(MaxMemorySize: TACRMemorySize); overload;
   // Destructor
   destructor Destroy; override;
   // GetMem analog
   function GetMem(BufferSize: TACRMemorySize): Pointer;
   // AllocMem analog
   function AllocMem(BufferSize: TACRMemorySize): Pointer;
   // ReAllocMem analog
   procedure ReallocMem(var Buffer; BufferSize: TACRMemorySize; ClearTail: Boolean = False);
   // ReAllocMem and clear Tail of Buffer
   procedure ReallocMemAndClearTail(var Buffer; BufferSize: TACRMemorySize);
   // FreeMem and set pointer to nil
   procedure FreeAndNilMem(var Buffer);
   // Return buffer size
   function GetMemoryBufferSize(Buffer: Pointer): TACRMemorySize;
   // Get min from free system memory size and (FMaxMemorySize - FTotalMemAllocated)
   function GetFreeMemorySize:  TACRMemorySize;
  public
   property MaxMemorySize:        TACRMemorySize read FMaxMemorySize;
   property TotalMemAllocated:    TACRMemorySize read FTotalMemAllocated;
   property MaxMemAllocated:    TACRMemorySize read FMaxMemAllocated;
   // statistics usage
   property AllocMemCallCount:    Integer read FAllocMemCallCount;
   property GetMemCallCount:      Integer read FGetMemCallCount;
   property FreeMemCallCount:     Integer read FFreeMemCallCount;
   property ReallocMemCallCount:  Integer read FReallocMemCallCount;
 end; // TACRMemoryManager

// Memory Manager variable
var MemoryManager: TACRMemoryManager = nil;

// move memory block
procedure ACRMove(const Source; var Dest; count : Integer );

implementation

uses Math;

type
  TGetMemType = (gmtGetMem, gmtVirtualAlloc, gmtGlobalAlloc);

  // Memory Block Header
  PACRMemoryBlockHeader = ^TACRMemoryBlockHeader;
  TACRMemoryBlockHeader = packed record
    Size: Integer;
    GetMemType: TGetMemType;
    Signature: Cardinal;  // Last 4 byte
  end;

  // Memory Block Footer
  PACRMemoryBlockFooter = ^TACRMemoryBlockFooter;
  TACRMemoryBlockFooter = packed record
    Signature: Cardinal;  // = ACRMemoryEndSignature
  end;


const  ACRMemorySignature:    Cardinal = $ACCACCAC;
       ACRMemoryEndSignature: Cardinal = $ACCEACCE;

type
   TGetMemFunction = packed record
     Size: Integer;
     GetMemType: TGetMemType;
   end;

const

// If Enable MemChecker then use only GetMem function
{$IFDEF DEBUG_MEMCHECK}
   GetMemTypes: array[1..1] of TGetMemFunction =
      (
        (Size: MaxInt;  GetMemType: gmtGetMem)
      );
{$ELSE}
 {$IFDEF LINUX}
   GetMemTypes: array[1..1] of TGetMemFunction =
      (
        (Size: MaxInt;  GetMemType: gmtGetMem)
      );
 {$ENDIF}
 {$IFDEF MSWINDOWS}
   GetMemTypes: array[1..4] of TGetMemFunction =
      (
        (Size: 726;   GetMemType: gmtGetMem),          // 0     - 1024  ==> GetMem
        (Size: 64500;  GetMemType: gmtGlobalAlloc),    // 1025  - 64500 ==> GlobalAlloc
        (Size: 1048576;  GetMemType: gmtGetMem),       // 64500 - 1 MB ...  ==> GetMem
        (Size: MaxInt;  GetMemType: gmtGlobalAlloc)    // 1 MB -  ...  ==> GlobalAlloc
      );
 {$ENDIF}
{$ENDIF}

function GetMemFunctionType(MemSize: Integer): TGetMemType;
var i: Integer;
begin
  // Set GetMem Function for Maximum Size
  Result := GetMemTypes[High(GetMemTypes)].GetMemType;
  // Find lower diapason
  for i:= High(GetMemTypes) downto Low(GetMemTypes)  do
    if (MemSize <= GetMemTypes[i].Size) then
     begin
      Result := GetMemTypes[i].GetMemType;
      Break;
     end;
end;

////////////////////////////////////////////////////////////////////////////////
//
// TACRMemoryManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRMemoryManager.Create;
begin
 FMaxMemorySize := 0; 
 FMaxMemAllocated := 0; 
 FTotalMemAllocated := 0;
 InitializeCriticalSection(MMLock);
end;//Create 
 
//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------ 
constructor TACRMemoryManager.Create(MaxMemorySize: TACRMemorySize);
begin 
 Create;
 FMaxMemorySize := MaxMemorySize; 
end;//Create
 
	
//------------------------------------------------------------------------------ 
// Destructor 
//------------------------------------------------------------------------------
destructor TACRMemoryManager.Destroy; 
begin 
 DeleteCriticalSection(MMLock); 
{$IFNDEF DEBUG_MEMCHECK} 
// if FTotalMemAllocated <> 0 then 
//   raise EACRException.Create(30003, ErrorGMemoryLeakFound, [FTotalMemAllocated]);
{$ENDIF}
end;//Destructor
	
	
//------------------------------------------------------------------------------ 
// GetMemoryBufferSize
//------------------------------------------------------------------------------ 
function TACRMemoryManager.GetMemoryBufferSize(Buffer: Pointer): TACRMemorySize;
var 
  Block: PACRMemoryBlockHeader; 
begin
  if Buffer = nil then 
    Result := 0
  else 
    begin 
  Block := PACRMemoryBlockHeader(PChar(Buffer) - SizeOf(TACRMemoryBlockHeader));
  if (Block.Signature = ACRMemorySignature) then
    Result := Block.Size 
  else 
    raise EACRException.Create(30005, ErrorGInvalidPointer); 
    end; 
end;//GetMemoryBufferSize 
 
 
//------------------------------------------------------------------------------
// GetMem
//------------------------------------------------------------------------------ 
function TACRMemoryManager.GetMem(BufferSize: TACRMemorySize): Pointer;
var
  NewSize: TACRMemorySize; 
  BlockHeader: PACRMemoryBlockHeader;
  BlockFooter: PACRMemoryBlockFooter;
  GetMemType: TGetMemType; 
begin 
  // Increment Counter 
  Inc(FGetMemCallCount);
	
  // Allocate 0 bytes ? 
  if (BufferSize = 0) then
    raise EACRException.Create(30286, ErrorGCannotAllocateZeroBytes);
 
  // Mem Limit ? 
  if ((FMaxMemorySize <> 0) and
      (BufferSize + FTotalMemAllocated > FMaxMemorySize)) then
    raise EACRException.Create(30004, ErrorGMemoryLimitExceeded, [FMaxMemorySize]);
 
  try 
    // Calculate New Size of Buffer 
    NewSize := BufferSize + SizeOf(TACRMemoryBlockHeader) + SizeOf(TACRMemoryBlockFooter); 
	
    // GetMem 
    GetMemType := GetMemFunctionType(NewSize); 
    case GetMemType of 
      gmtGetMem:
          System.GetMem(BlockHeader, NewSize)
 {$IFDEF MSWINDOWS}
      ; 
      gmtVirtualAlloc:
          BlockHeader := VirtualAlloc(nil, NewSize, MEM_COMMIT, PAGE_READWRITE);
      gmtGlobalAlloc:
          BlockHeader := Pointer(GlobalAlloc(GMEM_FIXED, NewSize)) 
 {$ENDIF} 
      else
          raise EACRException.Create(30340, ErrorGUnknownGetMemType, [Integer(GetMemType)]);
    end;
    // Fill Block Header
    BlockHeader.GetMemType := GetMemType;
    BlockHeader.Signature := ACRMemorySignature;
    BlockHeader.Size := BufferSize; 
    // Fill Block Footer
    BlockFooter := Pointer(PChar(BlockHeader) + SizeOf(TACRMemoryBlockHeader) + BufferSize);
    BlockFooter.Signature := ACRMemoryEndSignature; 
	
	
    Result := Pointer(Cardinal(BlockHeader) + SizeOf(TACRMemoryBlockHeader)); 
    EnterCriticalSection(MMLock); 
    Inc(FTotalMemAllocated, BufferSize); 
    if (FTotalMemAllocated > FMaxMemAllocated) then 
     FMaxMemAllocated := FTotalMemAllocated;
    LeaveCriticalSection(MMLock); 
  except 
    on e: Exception do
      raise EACRException.Create(30015, ErrorGGetMemError, [e.Message]);
  end;
end;//GetMem
 
 
//------------------------------------------------------------------------------
// AllocMem
//------------------------------------------------------------------------------ 
function TACRMemoryManager.AllocMem(BufferSize:TACRMemorySize):Pointer; 
begin 
  Inc(FAllocMemCallCount);
  try
    Result := self.GetMem(BufferSize);
    FillChar(Result^, BufferSize, 0);
  finally
    Dec(FGetMemCallCount); 
  end;
end;//AllocMem 
	

//------------------------------------------------------------------------------
// FreeMem and set Pointer to nil
//------------------------------------------------------------------------------
procedure TACRMemoryManager.FreeAndNilMem(var Buffer);
var
  BlockHeader: PACRMemoryBlockHeader;
  BlockFooter: PACRMemoryBlockFooter;
  FooterIncorrect: Boolean;
begin
  // Increment Counter
  Inc(FFreeMemCallCount);
  try
    // Check Header Signature
    BlockHeader := PACRMemoryBlockHeader(PChar(Buffer) - SizeOf(TACRMemoryBlockHeader));
    if (BlockHeader.Signature <> ACRMemorySignature) then
      raise EACRException.Create(30001, ErrorGInvalidPointer);

    // Check Footer Signature
    BlockFooter := Pointer(PChar(Buffer) + BlockHeader.Size);
    FooterIncorrect := (BlockFooter.Signature <> ACRMemoryEndSignature);

    // Calculate TotalMemAllocated
    EnterCriticalSection(MMLock);
    Dec(FTotalMemAllocated, BlockHeader.Size);
    LeaveCriticalSection(MMLock);

//if (DebugStarted) then
//aaWriteToLog('6 TACRMemoryManager.FreeAndNilMem, type = '+IntToStr(Integer(BlockHeader.GetMemType)));

    // FreeMem
    case BlockHeader.GetMemType of
      gmtGetMem:
          System.FreeMem(BlockHeader);
 {$IFDEF MSWINDOWS}
      gmtVirtualAlloc:
          VirtualFree(BlockHeader, 0, MEM_RELEASE);
      gmtGlobalAlloc:
          GlobalFree(Cardinal(Pointer(BlockHeader)));
 {$ENDIF}
    end;

    // Clear Buffer Pointer
    Pointer(Buffer) := nil;

    // if Footer Signature incorrect then raise
    if (FooterIncorrect) then
      raise EACRException.Create(30137, ErrorGMemoryOverrunDetected);

  except
    on EACRException do raise;
    on e: Exception do
      raise EACRException.Create(30138, ErrorGFreeMemError, [e.Message]);
  end;
end;//FreeAndNilMem


//------------------------------------------------------------------------------
// ReallocMem
//------------------------------------------------------------------------------
procedure TACRMemoryManager.ReallocMem(var Buffer; BufferSize: TACRMemorySize; ClearTail: Boolean);
var
  BlockHeader: PACRMemoryBlockHeader;
  NewBuffer: Pointer;
begin
  // Increment Counter
  Inc(FReallocMemCallCount);
  try
    // Check Header Signature
    BlockHeader := PACRMemoryBlockHeader(PChar(Buffer) - SizeOf(TACRMemoryBlockHeader));
    if (BlockHeader.Signature <> ACRMemorySignature) then
      raise EACRException.Create(30002, ErrorGInvalidPointer);

    // GetMem
    NewBuffer := Self.GetMem(BufferSize);
    // Copy OldBuffer to NewBuffer
    Move(PChar(Buffer)^, NewBuffer^,
       min(BufferSize, BlockHeader.Size));
    // Clear Tail
    if (ClearTail) then
      if (BufferSize > BlockHeader.Size) then
        FillChar(PChar(PChar(NewBuffer) + BlockHeader.Size)^, BufferSize-BlockHeader.Size, 0);

    // Free old buffer
    Self.FreeAndNilMem(PChar(Buffer));
    // Set Buffer to NewBuffer
    Pointer(Buffer) := NewBuffer;

    // Correct call counters
    Dec(FGetMemCallCount);
    Dec(FFreeMemCallCount);
  except
    on e: Exception do
      raise EACRException.Create(30014, ErrorGReallocMemError, [e.Message]);
  end;
end;//ReallocMem


//------------------------------------------------------------------------------
// ReAllocMem and clear Tail of Buffer
//------------------------------------------------------------------------------
procedure TACRMemoryManager.ReallocMemAndClearTail(var Buffer; BufferSize: TACRMemorySize);
begin
  ReallocMem(Buffer, BufferSize, True);
end;//ReallocMemAndClearTail

//------------------------------------------------------------------------------
// GetFreeMemorySize
//------------------------------------------------------------------------------
function TACRMemoryManager.GetFreeMemorySize: TACRMemorySize;
{$IFDEF MSWINDOWS}
var
  Status: TMemoryStatus;
{$ENDIF}
begin
{$IFDEF MSWINDOWS}
  GlobalMemoryStatus(Status);
  FFreeSystemMemorySize := Status.dwAvailPhys;
  if (FMaxMemorySize = 0) then
    Result := FFreeSystemMemorySize
  else
    Result := Min(FFreeSystemMemorySize, FMaxMemorySize);
{$ENDIF}
{$IFDEF LINUX}
    Result := FMaxMemorySize;
{$ENDIF}
end;//GetFreeMemorySize


//------------------------------------------------------------------------------
// move memory block
//------------------------------------------------------------------------------
procedure ACRMove(const Source; var Dest; Count : Integer );
var
  S, D: PChar;
  I, Offset: Integer;
begin
  S := PChar(@Source);
  D := PChar(@Dest);
  Offset := D - S;
  if ((Offset > 0) and (Offset < 4)) then
    for i := Count-1 downto 0 do
      (D+i)^ := (S+i)^
  else
  if ((Offset > -4) and (Offset < 0)) then
    for i := 0 to Count-1 do
      (D+i)^ := (S+i)^
  else
    Move(Source, Dest, Count);
end;


initialization

 MemoryManager := TACRMemoryManager.Create;

finalization

 MemoryManager.Free;
 MemoryManager := nil;

end.


