//------------------------------------------------------------------------------
//
// Server classes, Server component
// Uses Local Engine
//
//------------------------------------------------------------------------------

unit ACRServer;

interface

{$I ACRVer.inc}

uses SysUtils,Classes,IniFiles,DB,

// Accuracer units

{$IFDEF DEBUG_LOG}
     ACRDebug,
{$ENDIF}
{$IFNDEF D5H}
     ACRD4Routines,
{$ENDIF}
     ACRConnection,
     ACRLexer,
     ACRExcept,
     ACRMemory,
     ACRCommunication,
     ACRBase,
     ACRBaseEngine,
     ACRLocalEngine,
     ACRCompression,
     ACRCrypto,
     ACRVariant,
     ACRTypes,
     ACRSQLProcessor,
     ACRMain,
     ACRConst;

type

  TACRServerCursor = class;
  TACRServerSession = class;

  TACRClientInfo = packed record
   Host:             String;
   Port:             Integer;
   Application:      String;
   DatabaseName:     String;
   DatabaseFileName: String;
   SessionID:        TACRSessionID;
  end; // TACRClientInfo

  TACRClientInfoArray = array of TACRClientInfo;

  TACROnSQL = procedure (
                          Sender:     TComponent;
                          ClientInfo: TACRClientInfo;
                          Params:     TACRSQLParams;
                          var SQL:    String;
                          var Abort:  Boolean
                                   ) of object;
 TACROnServerReceiveTextMessage = procedure (const Client: TACRClientInfo; const Text: String)  of object;
 TACROnServerReceiveBinaryMessage = procedure (const Client: TACRClientInfo; Buffer: PChar; Size: Integer) of object;
 TACROnServerReceiveStreamMessage = procedure (const Client: TACRClientInfo; Stream: TStream) of object;



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerNetworkSettingsEditor
//
////////////////////////////////////////////////////////////////////////////////


  TACRServerNetworkSettingsEditor = class (TACRNetworkSettingsEditor)
   private
    FServerReceiveTimeOut:           Integer;
    FServerReceiveSleep:             Integer;
    FMinServerSendTimeOut:           Integer;
    FServerSendTimeOut:              Integer;
    FServerWaitForSendSleep:            Integer;
    FServerResendDelay:              Integer;
    FServerRequestDelay:             Integer;
    FWaitForMessagesSend:            Integer;
    FWaitForServerSessionThreadTimeOut: Integer;
    FServerThreadsTerminateDelay:    Integer;
    FServerSessionTerminatorSleep:   Integer;
    FWaitForPingAnswer:              Integer;
    FServerPingSleep:                Integer;
    FPingClients:                       Boolean;
   public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure CopySettingsToConnectParams(var ConnectParams: TACRConnectParams); override;
   published
    property ServerReceiveTimeOut:           Integer read FServerReceiveTimeOut write FServerReceiveTimeOut;
    property ServerReceiveSleep:             Integer read FServerReceiveSleep write FServerReceiveSleep;
    property MinServerSendTimeOut:           Integer read FMinServerSendTimeOut write FMinServerSendTimeOut;
    property ServerSendTimeOut:              Integer read FServerSendTimeOut write FServerSendTimeOut;
    property ServerWaitForSendSleep:         Integer read FServerWaitForSendSleep write FServerWaitForSendSleep;
    property ServerResendDelay:              Integer read FServerResendDelay write FServerResendDelay;
    property ServerRequestDelay:             Integer read FServerRequestDelay write FServerRequestDelay;
    property WaitForMessagesSend:            Integer read FWaitForMessagesSend write FWaitForMessagesSend;
    property WaitForServerSessionThreadTimeOut: Integer read FWaitForServerSessionThreadTimeOut write FWaitForServerSessionThreadTimeOut;
    property ServerThreadsTerminateDelay:    Integer read FServerThreadsTerminateDelay write FServerThreadsTerminateDelay;
    property ServerSessionTerminatorSleep:   Integer read FServerSessionTerminatorSleep write FServerSessionTerminatorSleep;
    property WaitForPingAnswer:              Integer read FWaitForPingAnswer write FWaitForPingAnswer;
    property ServerPingSleep:                Integer read FServerPingSleep write FServerPingSleep;
    property PingClients:                    Boolean read FPingClients write FPingClients;
  end; // TACRServerClientNetworkSettingsEditor


////////////////////////////////////////////////////////////////////////////////
//
// TACRServer
//
////////////////////////////////////////////////////////////////////////////////


  TACRServer = class (TComponent)
   private
    FUseConfigFile:                 Boolean;
    FNetworkSettings:               TACRServerNetworkSettingsEditor;
    FOnError:                       TACROnError;
    FBeforeServerStart:             TNotifyEvent;
    FAfterServerStart:              TNotifyEvent;
    FBeforeServerStop:              TNotifyEvent;
    FAfterServerStop:               TNotifyEvent;
    FCurrentVersion:                String;
    FActive:                        Boolean;
    FConfigFileName:                String;
    FDatabaseNames:                 TStrings;
    FDatabaseFileNames:             TStrings;
    FConnectionManager:             TACRServerConnectionManager;
    FOpenDatabasesInExclusiveMode:  Boolean;
    FLockParams:                    TACRLockParams;
    FLocalHost:                     String;
    FLocalPort:                     Integer;
    FServerID:                      Integer;
    FOnSQL:                         TACROnSQL;
    FOnReceiveTextMessage:          TACROnServerReceiveTextMessage;
    FOnReceiveBinaryMessage:        TACROnServerReceiveBinaryMessage;
    FOnReceiveStreamMessage:        TACROnServerReceiveStreamMessage;
    FCryptoParamsEditor:            TACRCryptoParamsEditor;
   protected
    function GetDatabaseCount: Integer;
    procedure StartServer;
    procedure StopServer;
    procedure DoOnSQL(
                          ClientInfo: TACRClientInfo;
                          Params:     TACRSQLParams;
                          var SQL:    String;
                          var Abort:  Boolean);
    procedure DoOnError(
                       const ErrorCode:    Integer;
                       const NativeError:  Integer;
                       const ErrorMessage: String
                       );
   public
    procedure DoOnConnectionError(
                       const ErrorCode:    Integer;
                       const NativeError:  Integer;
                       const ErrorMessage: String
                       );
    procedure LoadDefaultSettings;
    procedure LoadSettingsFromConfigFile;
    procedure SaveSettingsToConfigFile;
   protected
    procedure LoadServerSettings;
    function GetCurrentVersion: String;
    procedure SetActive(Value: Boolean);
    procedure SetDatabaseNames(Value: TStrings);
    procedure SetDatabaseFileNames(Value: TStrings);
    function GetClientSession(const Client: TACRClientInfo): TACRServerSession;
    procedure ReceiveMessage(ServerSession: TACRServerSession; Buffer: PChar; Size: Integer);
   public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function FindDatabaseByName(Name: String): Integer;
    procedure GetClients(var Clients: TACRClientInfoArray);
   public
    function IsClientConnected(const Client: TACRClientInfo): Boolean;
    procedure SendMessage(const Client: TACRClientInfo; const Text: String); overload;
    procedure SendMessage(const Client: TACRClientInfo; Buffer: PChar; Size: Integer); overload;
    procedure SendMessage(const Client: TACRClientInfo; Stream: TStream); overload;
   public
    property DatabaseCount: Integer read GetDatabaseCount;
    property ConnectionManager: TACRServerConnectionManager read FConnectionManager;
    property LockParams: TACRLockParams read FLockParams;
   published
    property UseConfigFile: Boolean read FUseConfigFile write FUseConfigFile;
    property Active: Boolean read FActive write SetActive;
    property CurrentVersion: String read GetCurrentVersion write FCurrentVersion;
    property DatabaseNames: TStrings read FDatabaseNames write SetDatabaseNames;
    property DatabaseFileNames: TStrings read FDatabaseFileNames write SetDatabaseFileNames;
    property ConfigFileName: String read FConfigFileName write FConfigFileName;
    property CryptoParams: TACRCryptoParamsEditor read FCryptoParamsEditor write FCryptoParamsEditor;
    property NetworkSettings: TACRServerNetworkSettingsEditor read FNetworkSettings write FNetworkSettings;
    property BeforeServerStart: TNotifyEvent read FBeforeServerStart write FBeforeServerStart;
    property AfterServerStart: TNotifyEvent read FAfterServerStart write FAfterServerStart;
    property BeforeServerStop: TNotifyEvent read FBeforeServerStop write FBeforeServerStop;
    property AfterServerStop: TNotifyEvent read FAfterServerStop write FAfterServerStop;
    property OpenDatabasesInExclusiveMode: Boolean read FOpenDatabasesInExclusiveMode
              write FOpenDatabasesInExclusiveMode;
    property LocalHost: String read FLocalHost write FLocalHost;
    property LocalPort: Integer read FLocalPort write FLocalPort;
    property ServerID: Integer read FServerID write FServerID;
    property OnSQL: TACROnSQL read FOnSQL write FOnSQL;
    property OnReceiveTextMessage: TACROnServerReceiveTextMessage read FOnReceiveTextMessage write FOnReceiveTextMessage;
    property OnReceiveBinaryMessage: TACROnServerReceiveBinaryMessage read FOnReceiveBinaryMessage write FOnReceiveBinaryMessage;
    property OnReceiveStreamMessage: TACROnServerReceiveStreamMessage read FOnReceiveStreamMessage write FOnReceiveStreamMessage;
    property OnError: TACROnError read FOnError write FOnError;
  end; // TACRServer


////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSession
//
////////////////////////////////////////////////////////////////////////////////


  TACRServerSession = class (TACRNetworkSession)
   private
    FServer:        TACRServer;
    FDatabase:      TACRDatabase;
    FHandle:        TACRLocalSession;
    FLastCursorID:  TACRObjectID;
    FCursorList:    TList;
   protected
    // get error message and raises an excpetion
    procedure SendErrorMessage(ErrorCode:    Integer;
                               ErrorMessage: String);
    // find server cursor by id
    function FindServerCursor(CursorID: TACRObjectID): TACRServerCursor;
    // send vsisiblefielddefs.fieldnoreference
    procedure SendFieldNoReferences(aCursor: TACRLocalCursor);
    // create server cursor, put information about it to FSentCommandDataStream and add it to the CursorList
    function CreateServerCursor(aCursor: TACRLocalCursor; aQuery: TACRQuery = nil): TACRServerCursor;
    //---------------------- SQL support ---------------------------------------
    // receive SQL Params
    procedure ReceiveParams(FQuery: TACRQuery);
    procedure SendLiveQuery(FQuery: TACRQuery; LocalCursor: TACRLocalCursor);
    procedure SendNotLiveQuery(FQuery: TACRQuery; LocalCursor: TACRLocalCursor);
    //--------------------- base methods ---------------------------------------
    procedure SetConnected(Value: Boolean); override;
    // db connected?
    function GetConnected: Boolean; override;
    // connect / disconnect
   public
    // check if database exists
    function GetDatabaseExists: Boolean; override;
    procedure GetTablesList(List: TStrings); override;
    function TableExists(TableName: String): Boolean; override;
    // return database format version
    function GetFormatVersion: Double; override;
    // return total number of pages
    function GetTotalPageCount : Integer; override;
    // return number of free pages
    function GetFreePageCount : Integer; override;
    // return true if database is encrypted
    function IsDatabaseEncrypted: Boolean; override;
    // return true if database is encrypted by password or by key
    function IsDatabaseEncryptedByPassword: Boolean; override;
    // return true if CryptoParams are valid
    function IsCryptoParamsValid: Boolean; override;
    //------------------------ Transactions ------------------------------------
   protected
    // retrun true if database has active transaction
    function GetInTransaction: Boolean; override;
   public
    // start a transaction
    procedure StartTransaction; override;
    // apply changes made by transaction
    procedure Commit(FlushFileBuffers: Boolean = True); override;
    // cancel changes made by transaction
    procedure Rollback; override;
    // flush file buffers
    procedure FlushFileBuffers; override;

   public
    constructor Create(aServer: TComponent);
    destructor Destroy; override;
    // call OnError event handler
    procedure DoOnError(ErrorCode: Integer; NativeError: Integer = -1; ErrorMessage: String = ''); override;
    procedure DoCloseSessionOnNetworkError; override;
    // execute received command
    procedure ExecuteReceivedCommand;
    // receive data from network and move it to ReceivedCommandHeader and ReceivedCommandDataStream
    procedure ReceiveData(Buffer: PChar; BufferSize: Integer); override;
    // send buffer via established connection using connection manager
    procedure SendBuffer(var Buffer: PChar; BufferSize: Integer); override;
    // send custom message to client
    procedure SendMessage(Buffer: PChar; BufferSize: Integer); override;
    // receive custom message from client
    procedure ReceiveMessage(Buffer: PChar; BufferSize: Integer); override;
   protected
    function GetClientInfo: TACRClientInfo;
    //--------------------- Executing client requests --------------------------
    // execute ConnectDatabase request
    procedure ExecuteConnectDatabase;
    // execute IsDatabaseExists
    procedure ExecuteIsDatabaseExists;
    // execute GetTablesList
    procedure ExecuteGetTablesList;
    // execute IsTableExists
    procedure ExecuteIsTableExists;
    // execute GetFormatVersion
    procedure ExecuteGetFormatVersion;
    // execute GetTotalPageCount
    procedure ExecuteGetTotalPageCount;
    // execute GetFreePageCount
    procedure ExecuteGetFreePageCount;
    // execute IsDatabaseEncrypted
    procedure ExecuteIsDatabaseEncrypted;
    // execute IsDatabaseEncryptedByPassword
    procedure ExecuteIsDatabaseEncryptedByPassword;
    // execute IsCryptoParamsValid
    procedure ExecuteIsCryptoParamsValid;
    // execute IsInTransaction
    procedure ExecuteIsInTransaction;
    // execute StartTransaction
    procedure ExecuteStartTransaction;
    // execute Commit
    procedure ExecuteCommit;
    // execute Rollback
    procedure ExecuteRollback;

    //------------------------- Cursor Commands --------------------------------

    // execute CreateTable
    procedure ExecuteCreateTable;
    // execute DeleteTable
    procedure ExecuteDeleteTable;
    // execute EmptyTable
    procedure ExecuteEmptyTable;
    // execute RenameTable
    procedure ExecuteRenameTable;
    // execute RenameField
    procedure ExecuteRenameField;
    // execute OpenTable
    procedure ExecuteOpenTable;
    // execute CloseTable
    procedure ExecuteCloseTable;
    // execute ApplyDefaultValues
    procedure ExecuteApplyDefaultValues;
    // execute AddIndex
    procedure ExecuteAddIndex;
    // execute DeleteIndex
    procedure ExecuteDeleteIndex;
    // execute DeleteAllIndexes
    procedure ExecuteDeleteAllIndexes;
    // execute GetRecordCount
    procedure ExecuteGetRecordCount;
    // execute GetRecordBuffer
    procedure ExecuteGetRecordBuffer;
    // execute SetRecNo
    procedure ExecuteSetRecNo;
    // execute GetRecNo
    procedure ExecuteGetRecNo;
    // execute InternalEdit
    procedure ExecuteInternalEdit;
    // execute InternalCancel
    procedure ExecuteInternalCancel;
    // execute InternalPost
    procedure ExecuteInternalPost;
    // execute InternalDelete
    procedure ExecuteInternalDelete;
    // execute ActivateFilters
    procedure ExecuteActivateFilters;
    // execute DeactivateFilters
    procedure ExecuteDeactivateFilters;
    // execute Locate
    procedure ExecuteLocate;
    // execute FindKey
    procedure ExecuteFindKey;
    // execute ResetRange
    procedure ExecuteResetRange;
    // execute ApplyRange
    procedure ExecuteApplyRange;
    // execute ReadBLOBValue
    procedure ExecuteReadBLOBValue;

    // execute ExecSQL
    procedure ExecuteExecSQL;

    // execute IsRecordExists
    procedure ExecuteIsRecordExists;
  end; // TACRServer


////////////////////////////////////////////////////////////////////////////////
//
// TACRServerCursor
//
////////////////////////////////////////////////////////////////////////////////


  TACRServerCursor = class (TObject)
   private
    FCursorID:            TACRObjectID;
    FLocalCursor:         TACRLocalCursor;
    FQuery:               TACRQuery;
    FCurrentRecordBuffer: TACRRecordBuffer;
    FEditRecordBuffer:    TACRRecordBuffer;
    FSession:             TACRServerSession;
    FForcedDestroy:       Boolean;
   protected
    procedure SendRecordBuffer(const ErrorCode: Integer; RecordBuffer: TACRRecordBuffer);
    procedure ReceiveRecordBuffer(const ErrorCode: Integer; RecordBuffer: TACRRecordBuffer);
    procedure ReceiveModifiedBLOBValues(ToInsert: Boolean);
   public
    constructor Create(aCursor: TACRLocalCursor; aCursorID: TACRObjectID; aSession: TACRServerSession; aQuery: TACRQuery = nil);
    destructor Destroy; override;
    procedure ApplyDefaultValues;
    procedure GetRecordCount;
    procedure GetRecordBuffer;
    procedure SetRecNo;
    procedure GetRecNo;
    procedure InternalEdit;
    procedure InternalCancel;
    procedure InternalPost;
    procedure InternalDelete;
    procedure ActivateFilters;
    procedure DeactivateFilters;
    procedure Locate;
    procedure FindKey;
    procedure ResetRange(DoNotSendReply: Boolean = False);
    procedure ApplyRange;
    procedure ReadBLOBValue;
    procedure IsRecordExists;
   public
    property CursorID: TACRObjectID read FCursorID;
    property LocalCursor: TACRLocalCursor read FLocalCursor;
    property Session: TACRServerSession read FSession;
    property Query: TACRQuery read FQuery;
    property ForcedDestroy: Boolean read FForcedDestroy write FForcedDestroy;
  end; // TACRServer


 function InstallingService: Boolean;
 function InteractiveService: Boolean;

implementation

{$IFDEF D6H}
uses
Variants;
{$ENDIF}


////////////////////////////////////////////////////////////////////////////////
//
// TACRServerNetworkSettingsEditor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerNetworkSettingsEditor.Create;
begin
  inherited Create;
  FServerReceiveTimeOut := ACRServerReceiveTimeOut;
  FServerReceiveSleep := ACRServerReceiveSleep;
  FMinServerSendTimeOut := ACRMinServerSendTimeOut;
  FServerSendTimeOut := ACRServerSendTimeOut;
  FServerWaitForSendSleep := ACRServerWaitForSendSleep;
  FServerResendDelay := ACRServerResendDelay;
  FServerRequestDelay := ACRServerRequestDelay;
  FWaitForMessagesSend := ACRWaitForMessagesSend;
  FWaitForServerSessionThreadTimeOut := ACRWaitForServerSessionThreadTimeOut; 
  FServerThreadsTerminateDelay := ACRServerThreadsTerminateDelay;
  FServerSessionTerminatorSleep := ACRServerSessionTerminatorSleep;
  FWaitForPingAnswer := ACRWaitForPingAnswer;
  FPingClients := ACRPingClients;
  FServerPingSleep := ACRServerPingSleep;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRServerNetworkSettingsEditor.Destroy;
begin
  inherited Destroy;
end; // Destroy


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRServerNetworkSettingsEditor.Assign(Source: TPersistent);
begin
  if (Source <> nil) then
   if (Source is TACRServerNetworkSettingsEditor) then
    begin
      inherited Assign(Source);
      FServerReceiveTimeOut := TACRServerNetworkSettingsEditor(Source).ServerReceiveTimeOut;
      FServerReceiveSleep := TACRServerNetworkSettingsEditor(Source).ServerReceiveSleep;
      FMinServerSendTimeOut := TACRServerNetworkSettingsEditor(Source).MinServerSendTimeOut;
      FServerSendTimeOut := TACRServerNetworkSettingsEditor(Source).ServerSendTimeOut;
      FServerWaitForSendSleep := TACRServerNetworkSettingsEditor(Source).ServerWaitForSendSleep;
      FServerResendDelay := TACRServerNetworkSettingsEditor(Source).ServerResendDelay;
      FServerRequestDelay := TACRServerNetworkSettingsEditor(Source).ServerRequestDelay;
      FWaitForMessagesSend := TACRServerNetworkSettingsEditor(Source).WaitForMessagesSend;
      FWaitForServerSessionThreadTimeOut := TACRServerNetworkSettingsEditor(Source).WaitForServerSessionThreadTimeOut;
      FServerThreadsTerminateDelay := TACRServerNetworkSettingsEditor(Source).ServerThreadsTerminateDelay;
      FServerSessionTerminatorSleep := TACRServerNetworkSettingsEditor(Source).ServerSessionTerminatorSleep;
      FPingClients := TACRServerNetworkSettingsEditor(Source).PingClients;
      FWaitForPingAnswer := TACRServerNetworkSettingsEditor(Source).WaitForPingAnswer;
      FServerPingSleep := TACRServerNetworkSettingsEditor(Source).ServerPingSleep;
    end;
end; // Assign


//------------------------------------------------------------------------------
// Copy ServerNetwork settings to ConnectParams
//------------------------------------------------------------------------------
procedure TACRServerNetworkSettingsEditor.CopySettingsToConnectParams(var ConnectParams: TACRConnectParams);
begin
  inherited CopySettingsToConnectParams(ConnectParams);
  ConnectParams.ServerReceiveTimeOut := FServerReceiveTimeOut;
  ConnectParams.ServerReceiveSleep := FServerReceiveSleep;
  ConnectParams.MinServerSendTimeOut := FMinServerSendTimeOut;
  ConnectParams.ServerSendTimeOut := FServerSendTimeOut;
  ConnectParams.ServerWaitForSendSleep := FServerWaitForSendSleep;
  ConnectParams.ServerResendDelay := FServerResendDelay;
  ConnectParams.ServerRequestDelay := FServerRequestDelay;
  ConnectParams.WaitForMessagesSend := FWaitForMessagesSend;
  ConnectParams.WaitForServerSessionThreadTimeOut := FWaitForServerSessionThreadTimeOut;
  ConnectParams.ServerThreadsTerminateDelay := FServerThreadsTerminateDelay;
  ConnectParams.ServerSessionTerminatorSleep := FServerSessionTerminatorSleep;
  ConnectParams.PingClients := FPingClients;
  ConnectParams.WaitForPingAnswer := FWaitForPingAnswer;
  ConnectParams.ServerPingSleep := FServerPingSleep;
end; // CopySettingsToConnectParams


////////////////////////////////////////////////////////////////////////////////
//
// TACRServer
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// retunr number of databases in this server
//------------------------------------------------------------------------------
function TACRServer.GetDatabaseCount: Integer;
begin
  Result := FDatabaseNames.Count;
end; // GetDatabaseCount


//------------------------------------------------------------------------------
// start server
//------------------------------------------------------------------------------
procedure TACRServer.StartServer;
begin
  if (not FActive) then
   begin
    if (Assigned(FBeforeServerStart)) then
     FBeforeServerStart(Self);
    LoadServerSettings;
    FConnectionManager := TACRServerConnectionManager.Create(Self);
    FActive := True;
    if (Assigned(FAfterServerStart)) then
     FAfterServerStart(Self);
   end;
end; // StartServer


//------------------------------------------------------------------------------
// stop server
//------------------------------------------------------------------------------
procedure TACRServer.StopServer;
begin
  if (FActive) then
   begin
    if (Assigned(FBeforeServerStart)) then
     FBeforeServerStop(Self);
    FConnectionManager.Free;
    FConnectionManager := nil;
    FActive := False;
    if (Assigned(FAfterServerStart)) then
     FAfterServerStop(Self);
   end;
end; // StopServer;


//------------------------------------------------------------------------------
// do on sql
//------------------------------------------------------------------------------
procedure TACRServer.DoOnSQL(
                      ClientInfo: TACRClientInfo;
                      Params:     TACRSQLParams;
                      var SQL:    String;
                      var Abort:  Boolean);
begin
 Abort := False;
 if (Assigned(FOnSQL)) then
   FOnSQL(Self,ClientInfo,Params,SQL,Abort);
end; // DoOnSQL


//------------------------------------------------------------------------------
// Do on error
//------------------------------------------------------------------------------
procedure TACRServer.DoOnConnectionError(
                       const ErrorCode:    Integer;
                       const NativeError:  Integer;
                       const ErrorMessage: String
                       );
begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog('Error in ACRConnection module. Session is not existing.');
aaWriteToLog('------------------------------------------------------------------');
aaWriteToLog('ErrorCode='+IntToStr(Integer(ErrorCode)));
aaWriteToLog('NativeError='+IntToStr(Integer(NativeError)));
aaWriteToLog('ErrorMessage:"'+ErrorMessage+'"');
aaWriteToLog('GetTickCount = '+IntToStr(aaGetTickCount));
aaWriteToLog('==================================================================');
{$ENDIF}
  DoOnError(ErrorCode,NativeError,ErrorMessage);
end; // DoOnError


//------------------------------------------------------------------------------
// Do on error
//------------------------------------------------------------------------------
procedure TACRServer.DoOnError(
                       const ErrorCode:    Integer;
                       const NativeError:  Integer;
                       const ErrorMessage: String
                       );
begin
  if (Assigned(FOnError)) then
   FOnError(Self,ErrorCode,NativeError,ErrorMessage);
end; // DoOnError


//------------------------------------------------------------------------------
// load default settings;
//------------------------------------------------------------------------------
procedure TACRServer.LoadDefaultSettings;
begin
  if (UseConfigFile) then
   if (not FileExists(FConfigFileName)) then
    SaveSettingsToConfigFile;
end; // LoadDefaultSettings


//------------------------------------------------------------------------------
// load server settings from config file
//------------------------------------------------------------------------------
procedure TACRServer.LoadSettingsFromConfigFile;
const NetworkCaption = 'Server Network Settings';
var IniFile:      TIniFile;
    i,n:          Integer;
    s:            String;
    FileName:     String;
    CryptoParams: TACRCryptoParams;
begin
  // ini file cannot be created without full path
  if (ExtractFilePath(FConfigFileName) <> '') then
   FileName := FConfigFileName
  else
   FileName := ExtractFilePath(ParamStr(0))+ FConfigFileName;
  FDatabaseNames.Clear;
  FDatabaseFileNames.Clear;
  IniFile := TIniFile.Create(FileName);
  try
    // connection params
    FLocalHost := IniFile.ReadString('Server Connection Parameters','LocalHost',FLocalHost);
    FLocalPort := IniFile.ReadInteger('Server Connection Parameters','LocalPort',FLocalPort);
    FServerID := IniFile.ReadInteger('Server Connection Parameters','ServerID',FServerID);
    FNetworkSettings.ServerRequestDelay := IniFile.ReadInteger('Server Connection Parameters','ServerReceiveDelay',FNetworkSettings.ServerRequestDelay);
    FNetworkSettings.ServerResendDelay := IniFile.ReadInteger('Server Connection Parameters','ServerResendRequestDelay',FNetworkSettings.ServerResendDelay);
    FNetworkSettings.DisconnectRetryCount := IniFile.ReadInteger('Server Connection Parameters','ServerDisconnectRetryCount',FNetworkSettings.DisconnectRetryCount);
    FNetworkSettings.DisconnectDelay := IniFile.ReadInteger('Server Connection Parameters','ServerDisconnectDelay',FNetworkSettings.DisconnectDelay);
    // crypto parameters
    CryptoParams := FCryptoParamsEditor.GetCryptoParams;
    IniFile.ReadInteger('Server Crypto Parameters','CryptoAlgorithm',CryptoParams.CryptoAlgorithm);
    IniFile.ReadInteger('Server Crypto Parameters','CryptoMode',CryptoParams.CryptoMode);
    IniFile.ReadString('Server Crypto Parameters','Password',CryptoParams.Password);
    IniFile.ReadBool('Server Crypto Parameters','UseInitVector',CryptoParams.UseInitVector);
    if (CryptoParams.UseInitVector) then
      for i := 0 to High(CryptoParams.InitVector) do
       IniFile.ReadInteger('Server Crypto Parameters','InitVector'+IntToStr(i),CryptoParams.InitVector[i]);
    if ((CryptoParams.Password = '') and (CryptoParams.CryptoAlgorithm <> acr_Cipher_None)) then
     begin
       IniFile.ReadInteger('Server Crypto Parameters','KeySize',CryptoParams.KeyInfo.KeySize);
       for i := 0 to CryptoParams.KeyInfo.KeySize do
        IniFile.ReadInteger('Server Crypto Parameters','Key'+IntToStr(i),CryptoParams.KeyInfo.Key[i]);
     end;
    // server databases
    n := IniFile.ReadInteger('Server Databases','DatabaseCount',0);
    for i := 0 to n - 1 do
     begin
      s := IniFile.ReadString('Database '+IntToStr(i+1),'DatabaseName','');
      FDatabaseNames.Add(s);
      s := IniFile.ReadString('Database '+IntToStr(i+1),'DatabaseFileName','');
      FDatabaseFileNames.Add(s);
     end;
    // database access params
    FOpenDatabasesInExclusiveMode := IniFile.ReadBool('Server Database Access Parameters',
      'OpenDatabasesInExclusiveMode',FOpenDatabasesInExclusiveMode);
    FLockParams.RetryCount := IniFile.ReadInteger('Server Database Access Parameters','LockParamsRetryCount',FLockParams.RetryCount);
    FLockParams.Delay := IniFile.ReadInteger('Server Database Access Parameters','LockParamsDelay',FLockParams.Delay);
    // network settings

    FNetworkSettings.PacketSize := IniFile.ReadInteger(NetworkCaption,'PacketSize',FNetworkSettings.PacketSize);
    FNetworkSettings.MaxThreadCount := IniFile.ReadInteger(NetworkCaption,'MaxThreadCount',FNetworkSettings.MaxThreadCount);
    FNetworkSettings.ConnectionParamsTunning := IniFile.ReadBool(NetworkCaption,'ConnectionParamsTunning',FNetworkSettings.ConnectionParamsTunning);
    FNetworkSettings.TestPacketCount := IniFile.ReadInteger(NetworkCaption,'TestPacketCount',FNetworkSettings.TestPacketCount);
    FNetworkSettings.MinServerSendTimeOut := IniFile.ReadInteger(NetworkCaption,'MinSendTimeOut',FNetworkSettings.MinServerSendTimeOut);
    FNetworkSettings.DisconnectRetryCount := IniFile.ReadInteger(NetworkCaption,'DisconnectRetryCount',FNetworkSettings.DisconnectRetryCount);
    FNetworkSettings.DisconnectDelay := IniFile.ReadInteger(NetworkCaption,'DisconnectDelay',FNetworkSettings.DisconnectDelay);
    FNetworkSettings.ServerReceiveTimeOut := IniFile.ReadInteger(NetworkCaption,'ServerReceiveTimeOut',FNetworkSettings.ServerReceiveTimeOut);
    FNetworkSettings.ServerReceiveSleep := IniFile.ReadInteger(NetworkCaption,'ServerReceiveSleep',FNetworkSettings.ServerReceiveSleep);
    FNetworkSettings.MinServerSendTimeOut := IniFile.ReadInteger(NetworkCaption,'MinServerSendTimeOut',FNetworkSettings.MinServerSendTimeOut);
    FNetworkSettings.ServerSendTimeOut := IniFile.ReadInteger(NetworkCaption,'ServerSendTimeOut',FNetworkSettings.ServerSendTimeOut);
    FNetworkSettings.ServerResendDelay := IniFile.ReadInteger(NetworkCaption,'ServerResendDelay',FNetworkSettings.ServerResendDelay);
    FNetworkSettings.ServerRequestDelay := IniFile.ReadInteger(NetworkCaption,'ServerRequestDelay',FNetworkSettings.ServerRequestDelay);
    FNetworkSettings.WaitForMessagesSend := IniFile.ReadInteger(NetworkCaption,'WaitForMessagesSend',FNetworkSettings.WaitForMessagesSend);
    FNetworkSettings.ServerThreadsTerminateDelay := IniFile.ReadInteger(NetworkCaption,'ServerThreadsTerminateDelay',FNetworkSettings.ServerThreadsTerminateDelay);
    FNetworkSettings.ServerSessionTerminatorSleep := IniFile.ReadInteger(NetworkCaption,'ServerSessionTerminatorSleep',FNetworkSettings.ServerSessionTerminatorSleep);
    FNetworkSettings.WaitForPingAnswer := IniFile.ReadInteger(NetworkCaption,'WaitForPingAnswer',FNetworkSettings.WaitForPingAnswer);
    FNetworkSettings.ServerPingSleep := IniFile.ReadInteger(NetworkCaption,'ServerPingSleep',FNetworkSettings.ServerPingSleep);
  finally
    IniFile.Free;
  end;
  FillChar(CryptoParams,SizeOf(CryptoParams),$00);
end; // LoadSettingsFromConfigFile


//------------------------------------------------------------------------------
// SaveSettingsToConfigFile
//------------------------------------------------------------------------------
procedure TACRServer.SaveSettingsToConfigFile;
const NetworkCaption = 'Server Network Settings';
var IniFile:      TIniFile;
    i:            Integer;
    FileName:     String;
    CryptoParams: TACRCryptoParams;
begin
  if (not UseConfigFile) then
   Exit;
  // ini file cannot be created without full path
  if (ExtractFilePath(FConfigFileName) <> '') then
   FileName := FConfigFileName
  else
   FileName := ExtractFilePath(ParamStr(0))+ FConfigFileName;
  if (SysUtils.FileExists(FileName)) then
   SysUtils.DeleteFile(FileName);
  IniFile := TIniFile.Create(FileName);
  try
    // connection params
    IniFile.WriteString('Server Connection Parameters','LocalHost',FLocalHost);
    IniFile.WriteInteger('Server Connection Parameters','LocalPort',FLocalPort);
    IniFile.WriteInteger('Server Connection Parameters','ServerID',FServerID);
    // crypto parameters
    CryptoParams := FCryptoParamsEditor.GetCryptoParams;
    IniFile.WriteInteger('Server Crypto Parameters','CryptoAlgorithm',CryptoParams.CryptoAlgorithm);
    IniFile.WriteInteger('Server Crypto Parameters','CryptoMode',CryptoParams.CryptoMode);
    IniFile.WriteString('Server Crypto Parameters','Password',CryptoParams.Password);
    IniFile.WriteBool('Server Crypto Parameters','UseInitVector',CryptoParams.UseInitVector);
    if (CryptoParams.UseInitVector) then
      for i := 0 to High(CryptoParams.InitVector) do
       IniFile.WriteInteger('Server Crypto Parameters','InitVector'+IntToStr(i),CryptoParams.InitVector[i]);
    if ((CryptoParams.Password = '') and (CryptoParams.CryptoAlgorithm <> acr_Cipher_None)) then
     begin
       IniFile.WriteInteger('Server Crypto Parameters','KeySize',CryptoParams.KeyInfo.KeySize);
       for i := 0 to CryptoParams.KeyInfo.KeySize do
        IniFile.WriteInteger('Server Crypto Parameters','Key'+IntToStr(i),CryptoParams.KeyInfo.Key[i]);
     end;
    // server databases
    IniFile.WriteInteger('Server Databases','DatabaseCount',DatabaseCount);
    for i := 0 to DatabaseCount - 1 do
     begin
      IniFile.WriteString('Database '+IntToStr(i+1),'DatabaseName',FDatabaseNames.Strings[i]);
      IniFile.WriteString('Database '+IntToStr(i+1),'DatabaseFileName',FDatabaseFileNames.Strings[i]);
     end;
    // database access params
    IniFile.WriteBool('Server Database Access Parameters',
      'OpenDatabasesInExclusiveMode',FOpenDatabasesInExclusiveMode);
    IniFile.WriteInteger('Server Database Access Parameters','LockParamsRetryCount',FLockParams.RetryCount);
    IniFile.WriteInteger('Server Database Access Parameters','LockParamsDelay',FLockParams.Delay);
    // network settings
    IniFile.WriteInteger(NetworkCaption,'PacketSize',FNetworkSettings.PacketSize);
    IniFile.WriteInteger(NetworkCaption,'MaxThreadCount',FNetworkSettings.MaxThreadCount);
    IniFile.WriteBool(NetworkCaption,'ConnectionParamsTunning',FNetworkSettings.ConnectionParamsTunning);
    IniFile.WriteInteger(NetworkCaption,'TestPacketCount',FNetworkSettings.TestPacketCount);
    IniFile.WriteInteger(NetworkCaption,'MinSendTimeOut',FNetworkSettings.MinServerSendTimeOut);
    IniFile.WriteInteger(NetworkCaption,'DisconnectRetryCount',FNetworkSettings.DisconnectRetryCount);
    IniFile.WriteInteger(NetworkCaption,'DisconnectDelay',FNetworkSettings.DisconnectDelay);
    IniFile.WriteInteger(NetworkCaption,'ServerReceiveTimeOut',FNetworkSettings.ServerReceiveTimeOut);
    IniFile.WriteInteger(NetworkCaption,'ServerReceiveSleep',FNetworkSettings.ServerReceiveSleep);
    IniFile.WriteInteger(NetworkCaption,'MinServerSendTimeOut',FNetworkSettings.MinServerSendTimeOut);
    IniFile.WriteInteger(NetworkCaption,'ServerSendTimeOut',FNetworkSettings.ServerSendTimeOut);
    IniFile.WriteInteger(NetworkCaption,'ServerResendDelay',FNetworkSettings.ServerResendDelay);
    IniFile.WriteInteger(NetworkCaption,'ServerRequestDelay',FNetworkSettings.ServerRequestDelay);
    IniFile.WriteInteger(NetworkCaption,'WaitForMessagesSend',FNetworkSettings.WaitForMessagesSend);
    IniFile.WriteInteger(NetworkCaption,'ServerThreadsTerminateDelay',FNetworkSettings.ServerThreadsTerminateDelay);
    IniFile.WriteInteger(NetworkCaption,'ServerSessionTerminatorSleep',FNetworkSettings.ServerSessionTerminatorSleep);
    IniFile.WriteInteger(NetworkCaption,'WaitForPingAnswer',FNetworkSettings.WaitForPingAnswer);
    IniFile.WriteInteger(NetworkCaption,'ServerPingSleep',FNetworkSettings.ServerPingSleep);
    IniFile.UpdateFile;
  finally
    IniFile.Free;
  end;
end; // SaveSettingsToConfigFile


//------------------------------------------------------------------------------
// load server settings from ini file or set them to default
//------------------------------------------------------------------------------
procedure TACRServer.LoadServerSettings;
begin
  if (not FileExists(FConfigFileName)) then
    LoadDefaultSettings
  else
   if (UseConfigFile) then
    LoadSettingsFromConfigFile;
end; // LoadServerSettings


//------------------------------------------------------------------------------
// return current version
//------------------------------------------------------------------------------
function TACRServer.GetCurrentVersion: String;
var c : char;
begin
 c := DecimalSeparator;
 DecimalSeparator := '.';
 Result := FloatToStrF(ACRVersion,ffFixed,3,2) + ' ' + ACRVersionText;
 DecimalSeparator := c;
end; // GetCurrentVersion


//------------------------------------------------------------------------------
// active
//------------------------------------------------------------------------------
procedure TACRServer.SetActive(Value: Boolean);
begin
 if (Value <> FActive) then
  begin
   if (Value) then
    StartServer
   else
    StopServer;
  end;
end; // SetActive


//------------------------------------------------------------------------------
// set database names
//------------------------------------------------------------------------------
procedure TACRServer.SetDatabaseNames(Value: TStrings);
begin
  FDatabaseNames.Assign(Value);
end; // SetDatabaseFileNames


//------------------------------------------------------------------------------
// set database filenames
//------------------------------------------------------------------------------
procedure TACRServer.SetDatabaseFileNames(Value: TStrings);
begin
  FDatabaseFileNames.Assign(Value);
end; // SetDatabaseFileNames


//------------------------------------------------------------------------------
// get client session
//------------------------------------------------------------------------------
function TACRServer.GetClientSession(const Client: TACRClientInfo): TACRServerSession;
var ClientSessions: TACRSessionsArray;
    i:              Integer;
begin
  Result := nil;
  FConnectionManager.GetClientsList(ClientSessions);
  if (Length(ClientSessions) > 0) then
   for i := Low(ClientSessions) to High(ClientSessions) do
    if (ClientSessions[i].SessionID = Client.SessionID) then
     begin
      Result := TACRServerSession(ClientSessions[i]);
      break;
     end;
  ClientSessions := nil;
end; // GetClientSession


//------------------------------------------------------------------------------
// receive custom message from client
//------------------------------------------------------------------------------
procedure TACRServer.ReceiveMessage(ServerSession: TACRServerSession; Buffer: PChar; Size: Integer);
var MessageType: TACRMessageType;
    ms,ms1:      TACRMemoryStream;
    Text:        String;
    len:         Integer;
    ClientInfo:  TACRClientInfo;
    Buf:         PChar;
    StreamSize:  Int64;
begin
  ClientInfo := ServerSession.GetClientInfo;
  ms := TACRMemoryStream.Create(Buffer);
  try
    LoadDataFromStream(MessageType,SizeOf(MessageType),ms,11376);
    case MessageType of
     aamtText:
      if (Assigned(FOnReceiveTextMessage)) then
       begin
         LoadDataFromStream(len,SizeOf(len),ms,11377);
         SetLength(Text,len);
         if (len > 0) then
          LoadDataFromStream(PChar(@Text[1])^,len,ms,11378);
         FOnReceiveTextMessage(ClientInfo,Text);
       end;
     aamtBinary:
      if (Assigned(FOnReceiveBinaryMessage)) then
       begin
         LoadDataFromStream(len,SizeOf(len),ms,11379);
         if (len > 0) then
          begin
           Buf := MemoryManager.GetMem(len);
           try
             LoadDataFromStream(Buf^,len,ms,11380);
             FOnReceiveBinaryMessage(ClientInfo,Buf,len);
           finally
             MemoryManager.FreeAndNilMem(Buf);
           end;
          end
         else
          FOnReceiveBinaryMessage(ClientInfo,nil,len);
       end;
     aamtStream:
      if (Assigned(FOnReceiveStreamMessage)) then
       begin
        ms1 := TACRMemoryStream.Create;
        try
          LoadDataFromStream(StreamSize,SizeOf(StreamSize),ms,11381);
          if (StreamSize > 0) then
           ms1.LoadFromStreamWithPosition(ms,ms.Position,StreamSize);
          FOnReceiveStreamMessage(ClientInfo,ms1);
        finally
          ms1.Free;
        end;
       end;
    end;
  finally
    ms.Free;
  end;
end; // ReceiveMessage


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRServer.Create(AOwner: TComponent);
begin
  FUseConfigFile := True;
  FConfigFileName := ACRDefaultServerConfigFileName;
  FActive := False;
  FDatabaseNames := TStringList.Create;
  FDatabaseFileNames := TStringList.Create;
  FConnectionManager := nil;
  FCryptoParamsEditor := TACRCryptoParamsEditor.Create;
  FNetworkSettings := TACRServerNetworkSettingsEditor.Create;
  FLockParams.RetryCount := ACRDefaultServerLockRetryCount;
  FLockParams.Delay := ACRDefaultServerLockDelay;
  FLocalHost := ACRDefaultServerHost;
  FLocalPort := ACRDefaultServerPort;
  FOpenDatabasesInExclusiveMode := True;
  FServerID := ACRDefaultServerID;
  inherited;
end; // Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRServer.Destroy;
begin
 Active := False;
 FDatabaseNames.Free;
 FDatabaseFileNames.Free;
 FCryptoParamsEditor.Free;
 FNetworkSettings.Free;
 if (FConnectionManager <> nil) then
  begin
   FConnectionManager.Free;
   FConnectionManager := nil;
  end;
inherited;
end; // Destroy


//------------------------------------------------------------------------------
// returns -1 if database was not found, otherwise index in FDatabaseNames (FDatabaseFileNames)
//------------------------------------------------------------------------------
function TACRServer.FindDatabaseByName(Name: String): Integer;
var DBName: String;
begin
  DBName := AnsiUpperCase(Name);
  Result := FDatabaseNames.IndexOf(DBName);
end; // FindDatabaseByName


//------------------------------------------------------------------------------
// fills array with client info
//------------------------------------------------------------------------------
procedure TACRServer.GetClients(var Clients: TACRClientInfoArray);
var ClientSessions: TACRSessionsArray;
    i,l:            Integer;
begin
  FConnectionManager.GetClientsList(ClientSessions);
  l := Length(ClientSessions);
  SetLength(Clients,l);
  if (l > 0) then
   for i := Low(ClientSessions) to High(ClientSessions) do
    Clients[i] := TACRServerSession(ClientSessions[i]).GetClientInfo;
  ClientSessions := nil;
end; // GetClients


//------------------------------------------------------------------------------
// return ture if client is still connected to server
//------------------------------------------------------------------------------
function TACRServer.IsClientConnected(const Client: TACRClientInfo): Boolean;
begin
  Result := (GetClientSession(Client) <> nil);
end; // IsClientConnected


//------------------------------------------------------------------------------
// Send text message
//------------------------------------------------------------------------------
procedure TACRServer.SendMessage(const Client: TACRClientInfo; const Text: String);
var
  ClientSession:  TACRServerSession;
  ms:             TACRMemoryStream;
  MessageType:    TACRMessageType;
  len:            Integer;
begin
  ClientSession := GetClientSession(Client);
  if (ClientSession = nil) then
   raise EACRException.Create(11250,ErrorLClientIsNotConnected);
  MessageType := aamtText;
  len := Length(Text);
  ms := TACRMemoryStream.Create();
  try
   SaveDataToStream(MessageType,SizeOf(MessageType),ms,11253);
   SaveDataToStream(len,SizeOf(len),ms,11254);
   if (len > 0) then
     SaveDataToStream(PChar(@Text[1])^,len,ms,11255);
   ClientSession.SendMessage(ms.Buffer,ms.Size);
  finally
    ms.Free;
  end;
end; // SendMessage


//------------------------------------------------------------------------------
// Send binary message
//------------------------------------------------------------------------------
procedure TACRServer.SendMessage(const Client: TACRClientInfo; Buffer: PChar; Size: Integer);
var
  ClientSession:  TACRServerSession;
  ms:             TACRMemoryStream;
  MessageType:    TACRMessageType;
begin
  ClientSession := GetClientSession(Client);
  if (ClientSession = nil) then
   raise EACRException.Create(11251,ErrorLClientIsNotConnected);
  MessageType := aamtBinary;
  ms := TACRMemoryStream.Create();
  try
   SaveDataToStream(MessageType,SizeOf(MessageType),ms,11256);
   SaveDataToStream(Size,SizeOf(Size),ms,11257);
   if (Size > 0) then
     SaveDataToStream(Buffer^,Size,ms,11258);
   ClientSession.SendMessage(ms.Buffer,ms.Size);
  finally
    ms.Free;
  end;
end; // SendMessage


//------------------------------------------------------------------------------
// Send stream
//------------------------------------------------------------------------------
procedure TACRServer.SendMessage(const Client: TACRClientInfo; Stream: TStream);
var
  ClientSession:  TACRServerSession;
  ms:             TACRMemoryStream;
  MessageType:    TACRMessageType;
  Size:           Int64;
begin
  ClientSession := GetClientSession(Client);
  if (ClientSession = nil) then
   raise EACRException.Create(11252,ErrorLClientIsNotConnected);
  MessageType := aamtStream;
  Size := Stream.Size;
  ms := TACRMemoryStream.Create();
  try
   SaveDataToStream(MessageType,SizeOf(MessageType),ms,11259);
   SaveDataToStream(Size,SizeOf(Size),ms,11260);
   if (Size > 0) then
    begin
     if (Stream is TACRStream) then
      TACRStream(Stream).SaveToStream(ms)
     else
      ms.CopyFrom(Stream,Size);
    end;
   ClientSession.SendMessage(ms.Buffer,ms.Size);
  finally
    ms.Free;
  end;
end; // SendMessage


////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSession
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// get error message and raises an excpetion
//------------------------------------------------------------------------------
procedure TACRServerSession.SendErrorMessage(ErrorCode:          Integer;
                                             ErrorMessage:       String);
var Len: Integer;
begin
  FSentCommandHeader.Reply := accrplOperationFailed;
  FSentCommandDataStream.Size := 0;
  SaveDataToStream(ErrorCode,Sizeof(ErrorCode),FSentCommandDataStream,10863);
  Len := Length(ErrorMessage);
  SaveDataToStream(Len,Sizeof(Len),FSentCommandDataStream,10864);
  SaveDataToStream(PChar(@ErrorMessage[1])^,Len,FSentCommandDataStream,10865);
  SendCommand;
end; // SendErrorMessage


//------------------------------------------------------------------------------
// find server cursor by id
//------------------------------------------------------------------------------
function TACRServerSession.FindServerCursor(CursorID: TACRObjectID): TACRServerCursor;
var i: Integer;
begin
  Result := nil;
  for i := 0 to FCursorList.Count - 1 do
   if (TACRServerCursor(FCursorList.Items[i]).CursorID = CursorID) then
    begin
     Result := FCursorList.Items[i];
     break;
    end;
end; // FindServerCursor


//------------------------------------------------------------------------------
// send vsisiblefielddefs.fieldnoreference
//------------------------------------------------------------------------------
procedure TACRServerSession.SendFieldNoReferences(aCursor: TACRLocalCursor);
var i,num: Integer;
begin
  for i := 0 to aCursor.VisibleFieldDefs.Count - 1 do
   begin
    num := aCursor.VisibleFieldDefs.Items[i].FieldNoReference;
    SaveDataToStream(num,SizeOf(num),FSentCommandDataStream,11185);
   end;
end; // SendFieldNoReferences


//------------------------------------------------------------------------------
// create server cursor, put information about it to FSentCommandDataStream and add it to the CursorList
//------------------------------------------------------------------------------
function TACRServerSession.CreateServerCursor(aCursor: TACRLocalCursor; aQuery: TACRQuery = nil): TACRServerCursor;
var LocalCursor: TACRLocalCursor;
begin
  LocalCursor := aCursor;
  Inc(FLastCursorID);
  // create server cursor and add it to cursor list of this server session
  Result := TACRServerCursor.Create(LocalCursor,FLastCursorID,Self,aQuery);
  FCursorList.Add(Result);
  SaveDataToStream(FLastCursorID,SizeOf(FLastCursorID),FSentCommandDataStream,10968);
  LocalCursor.FieldDefs.SaveToStream(FSentCommandDataStream);
  LocalCursor.VisibleFieldDefs.SaveToStream(FSentCommandDataStream);
  LocalCursor.IndexDefs.SaveToStream(FSentCommandDataStream);
  LocalCursor.ConstraintDefs.SaveToStream(FSentCommandDataStream);
  if (aQuery <> nil) then
   SendFieldNoReferences(LocalCursor);
end; // CreateServerCursor


//------------------------------------------------------------------------------
// receive SQL Params
//------------------------------------------------------------------------------
procedure TACRServerSession.ReceiveParams(FQuery: TACRQuery);
begin
  TACRQuery(FQuery).StmtHandle.SQLParams.LoadFromStream(FReceivedCommandDataStream);
end; // SendNotLiveQuery


//------------------------------------------------------------------------------
// send live query
//------------------------------------------------------------------------------
procedure TACRServerSession.SendLiveQuery(FQuery: TACRQuery; LocalCursor: TACRLocalCursor);
begin
  CreateServerCursor(LocalCursor,FQuery);
end; // SendNotLiveQuery


//------------------------------------------------------------------------------
// send not live query
//------------------------------------------------------------------------------
procedure TACRServerSession.SendNotLiveQuery(FQuery: TACRQuery; LocalCursor: TACRLocalCursor);
begin
  LocalCursor.SaveTableToStream(FSentCommandDataStream);
end; // SendNotLiveQuery


//------------------------------------------------------------------------------
// db connected?
//------------------------------------------------------------------------------
function TACRServerSession.GetConnected: Boolean;
begin
  Result := (FDatabase <> nil);
end; // GetConnected


//------------------------------------------------------------------------------
// connect / disconnect
//------------------------------------------------------------------------------
procedure TACRServerSession.SetConnected(Value: Boolean);
var i: Integer;
begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog(#13#10+'Server start SetConnected SessionID = '+IntToStr(FSessionID)+
              ', DatabaseName = '+FDatabaseName+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True));
 {$ENDIF}
  if (Value <> GetConnected) then
   begin
    if (Value) then
     begin
      FCursorList.Clear;
      if (FDatabase <> nil) then
       begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected closing database ... SessionID = '+IntToStr(FSessionID)+
              ', DatabaseName = '+FDatabaseName+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True)+#13#10);
 {$ENDIF}
        FreeAndNil(FDatabase);
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected database closed ... SessionID = '+IntToStr(FSessionID)+
              ', DatabaseName = '+FDatabaseName+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True)+#13#10);
 {$ENDIF}
       end;
      FDatabase := TACRDatabase.Create(nil);
      try
        FDatabase.DatabaseFileName := Self.DatabaseFileName;
        FDatabase.DatabaseName := GetTemporaryName('ACRServerDBName'+'_'+IntToStr(SessionID));
        FDatabase.CryptoParams.SetCryptoParams(Self.CryptoParams);
        FDatabase.LockParams.SetLockParams(FServer.LockParams);
        FDatabase.ReadOnly := False;
        FDatabase.Exclusive := FServer.OpenDatabasesInExclusiveMode;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected opening database ... SessionID = '+IntToStr(FSessionID)+
              ', DatabaseName = '+FDatabaseName+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True)+#13#10);
 {$ENDIF}
        FDatabase.Open;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected database opened... SessionID = '+IntToStr(FSessionID)+
              ', DatabaseName = '+FDatabaseName+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True)+#13#10);
 {$ENDIF}
        FHandle := TACRLocalSession(FDatabase.Handle);
        Self.FOptions := FHandle.Options;
        Self.CryptoParams := FHandle.CryptoParams;
      except
        FreeAndNil(FDatabase);
        raise;
      end;
     end // Connect
    else
     begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected starting diconnect... SessionID = '+IntToStr(FSessionID)+
              ', FDatabase = '+IntToStr(Integer(FDatabase))+
              ', DatabaseName = '+FDatabaseName+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True)+#13#10);

 {$ENDIF}
      if (FDatabase <> nil) then
       begin
        try
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected starting rollback... SessionID = '+IntToStr(FSessionID)+
              ', FDatabase = '+IntToStr(Integer(FDatabase))+
              ', FDatabase.Connected = '+BoolToStr(FDatabase.Connected,True)+
              ', DatabaseName = '+FDatabaseName+
              ', InTransaction = '+BoolToStr(FDatabase.InTransaction,True)+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True)+#13#10);
 {$ENDIF}
          if (FDatabase.InTransaction) then
           FDatabase.Rollback;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected rollback finished. SessionID = '+IntToStr(FSessionID)+
              ', FDatabase = '+IntToStr(Integer(FDatabase))+
              ', DatabaseName = '+FDatabaseName+
              ', InTransaction = '+BoolToStr(FDatabase.InTransaction,True)+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True)+#13#10);
 {$ENDIF}
        except
         on e: EACRException do
          DoOnError(ACR_CS_ErrorClosingServerSessionRollbackFailed,e.NativeError,
                   Format(ErrorL_CS_ErrorClosingServerSessionRollbackFailed,
                         [FSessionID,FDatabaseName,e.Message]));
         on e: Exception do
          DoOnError(ACR_CS_ErrorClosingServerSessionRollbackFailed,-1,
                   Format(ErrorL_CS_ErrorClosingServerSessionRollbackFailed,
                         [FSessionID,FDatabaseName,e.Message]))
         else
          DoOnError(ACR_CS_ErrorClosingServerSessionRollbackFailed,-1,
                   Format(ErrorL_CS_ErrorClosingServerSessionRollbackFailed,
                         [FSessionID,FDatabaseName,'']));
        end;
        try
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected starting remove all locks ... SessionID = '+IntToStr(FSessionID)+
              ', FDatabase = '+IntToStr(Integer(FDatabase))+
              ', DatabaseName = '+FDatabaseName+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True)+#13#10);
 {$ENDIF}
          FDatabase.RemoveAllLocks;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected remove all locks OK. SessionID = '+IntToStr(FSessionID)+
              ', FDatabase = '+IntToStr(Integer(FDatabase))+
              ', DatabaseName = '+FDatabaseName+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True)+#13#10);
 {$ENDIF}
        except
         on e: EACRException do
          DoOnError(ACR_CS_ErrorClosingServerSessionRemoveAllLocks,e.NativeError,
                   Format(ErrorL_CS_ErrorClosingServerSessionRemoveAllLocks,
                         [FSessionID,FDatabaseName,e.Message]));
         on e: Exception do
          DoOnError(ACR_CS_ErrorClosingServerSessionRollbackFailed,-1,
                   Format(ErrorL_CS_ErrorClosingServerSessionRemoveAllLocks,
                         [FSessionID,FDatabaseName,e.Message]))
         else
          DoOnError(ACR_CS_ErrorClosingServerSessionRollbackFailed,-1,
                   Format(ErrorL_CS_ErrorClosingServerSessionRemoveAllLocks,
                         [FSessionID,FDatabaseName,'']));
        end;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected starting closing cursors ... SessionID = '+IntToStr(FSessionID)+
              ', FCursorList.Count = '+IntToStr(FCursorList.Count)+
              ', DatabaseName = '+FDatabaseName+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True));
 {$ENDIF}
        for i := 0 to FCursorList.Count - 1 do
         try
          if (FCursorList.Items[i] <> nil) then
           begin
            TACRServerCursor(FCursorList.Items[i]).ForcedDestroy := True;
            TACRServerCursor(FCursorList.Items[i]).Free;
           end;
         except
           on e: EACRException do
            DoOnError(ACR_CS_ErrorClosingServerSessionCursorsClosingFailed,e.NativeError,
                     Format(ErrorL_CS_ErrorClosingServerSessionCursorsClosingFailed,
                            [FSessionID,FDatabaseName,e.Message]));
           on e: Exception do
            DoOnError(ACR_CS_ErrorClosingServerSessionCursorsClosingFailed,-1,
                     Format(ErrorL_CS_ErrorClosingServerSessionCursorsClosingFailed,
                            [FSessionID,FDatabaseName,e.Message]))
           else
            DoOnError(ACR_CS_ErrorClosingServerSessionCursorsClosingFailed,-1,
                     Format(ErrorL_CS_ErrorClosingServerSessionCursorsClosingFailed,
                            [FSessionID,FDatabaseName,'']));
         end;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected cursors closed ... SessionID = '+IntToStr(FSessionID)+
              ', FCursorList.Count = '+IntToStr(FCursorList.Count)+
              ', DatabaseName = '+FDatabaseName+#13#10);
 {$ENDIF}
        try
          FCursorList.Clear;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected starting closing database ... SessionID = '+IntToStr(FSessionID)+
              ', FDatabase = '+IntToStr(Integer(FDatabase))+
              ', FDatabase.Connected = '+BoolToStr(FDatabase.Connected,True)+
              ', DatabaseName = '+FDatabaseName+#13#10);
 {$ENDIF}
          FreeAndNil(FDatabase);
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server SetConnected database closed ... SessionID = '+IntToStr(FSessionID)+
              ', FDatabase = '+IntToStr(Integer(FDatabase))+
              ', DatabaseName = '+FDatabaseName+ #13#10);
 {$ENDIF}
          FHandle := nil;
        except
         on e: EACRException do
           DoOnError(ACR_CS_ErrorClosingServerSessionDestroyingDatabaseFailed,e.NativeError,
                   Format(ErrorL_CS_ErrorClosingServerSessionCursorsClosingFailed,
                          [FSessionID,FDatabaseName,e.Message]));
         on e: Exception do
           DoOnError(ACR_CS_ErrorClosingServerSessionDestroyingDatabaseFailed,-1,
                   Format(ErrorL_CS_ErrorClosingServerSessionCursorsClosingFailed,
                          [FSessionID,FDatabaseName,e.Message]))
         else
           DoOnError(ACR_CS_ErrorClosingServerSessionDestroyingDatabaseFailed,-1,
                   Format(ErrorL_CS_ErrorClosingServerSessionCursorsClosingFailed,
                          [FSessionID,FDatabaseName,'']));
        end;
       end;
     end; // Disconnect
   end;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server finish SetConnected SessionID = '+IntToStr(FSessionID)+
              ', DatabaseName = '+FDatabaseName+
              ', Value = '+BoolToStr(Value,True)+
              ', Connected = '+BoolToStr(GetConnected,True)+#13#10+#13#10);
 {$ENDIF}
end; // SetConnected


//------------------------------------------------------------------------------
// check if database exists
//------------------------------------------------------------------------------
function TACRServerSession.GetDatabaseExists: Boolean;
begin
  Result := False; // never used
end; // GetDatabaseExists


//------------------------------------------------------------------------------
// get tables list
//------------------------------------------------------------------------------
procedure TACRServerSession.GetTablesList(List: TStrings);
begin
  FHandle.GetTablesList(List);
end; // GetTablesList


//------------------------------------------------------------------------------
// return true if table exists
//------------------------------------------------------------------------------
function TACRServerSession.TableExists(TableName: String): Boolean;
begin
  Result := FHandle.TableExists(TableName);
end; // TableExists


//------------------------------------------------------------------------------
// return database format version
//------------------------------------------------------------------------------
function TACRServerSession.GetFormatVersion: Double;
begin
  Result := FHandle.GetFormatVersion;
end; // GetFormatVersion


//------------------------------------------------------------------------------
// return total number of pages
//------------------------------------------------------------------------------
function TACRServerSession.GetTotalPageCount: Integer;
begin
  Result := FHandle.GetTotalPageCount;
end; // GetTotalPageCount


//------------------------------------------------------------------------------
// return number of free pages
//------------------------------------------------------------------------------
function TACRServerSession.GetFreePageCount: Integer;
begin
  Result := FHandle.GetFreePageCount;
end; // GetFreePageCount


//------------------------------------------------------------------------------
// return true if database is encrypted
//------------------------------------------------------------------------------
function TACRServerSession.IsDatabaseEncrypted: Boolean;
begin
  Result := FHandle.IsDatabaseEncrypted;
end; // IsDatabaseEncrypted


//------------------------------------------------------------------------------
// return true if database is encrypted by password or by key
//------------------------------------------------------------------------------
function TACRServerSession.IsDatabaseEncryptedByPassword: Boolean;
begin
  Result := FHandle.IsDatabaseEncryptedByPassword;
end; // IsDatabaseEncryptedByPassword


//------------------------------------------------------------------------------
// return true if CryptoParams are valid
//------------------------------------------------------------------------------
function TACRServerSession.IsCryptoParamsValid: Boolean;
begin
  FHandle.CryptoParams := FCryptoParams;
  Result := FHandle.IsCryptoParamsValid;
end; // IsCryptoParamsValid


//------------------------------------------------------------------------------
// retrun true if database has active transaction
//------------------------------------------------------------------------------
function TACRServerSession.GetInTransaction: Boolean;
begin
  Result := FHandle.InTransaction;
end; // GetInTransaction


//------------------------------------------------------------------------------
// start a transaction
//------------------------------------------------------------------------------
procedure TACRServerSession.StartTransaction;
begin
  FHandle.StartTransaction;
end; // StartTransaction


//------------------------------------------------------------------------------
// apply changes made by transaction
//------------------------------------------------------------------------------
procedure TACRServerSession.Commit(FlushFileBuffers: Boolean = True);
begin
  FHandle.Commit(FlushFileBuffers);
end; // Commit


//------------------------------------------------------------------------------
// cancel changes made by transaction
//------------------------------------------------------------------------------
procedure TACRServerSession.Rollback;
begin
  FHandle.Rollback;
end; // Rollback


//------------------------------------------------------------------------------
// flush changes
//------------------------------------------------------------------------------
procedure TACRServerSession.FlushFileBuffers;
begin
  FHandle.FlushFileBuffers;
end; // FlushFileBuffers


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerSession.Create(aServer: TComponent);
begin
  inherited Create;
  FServer := TACRServer(aServer);
  FHandle := nil;
  FLastCursorID := 0;
  FCursorList := TList.Create;
  FDatabase := nil;
  FServer.NetworkSettings.CopySettingsToConnectParams(FConnectParams);
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRServerSession.Destroy;
begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('ServerSession Destroy starting, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
  SetConnected(False);
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('ServerSession Destroy disconnected from database, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
  FCursorList.Free;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('ServerSession Destroy starting inherited, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
  inherited Destroy;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('ServerSession Destroy finished, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
end; // Destroy


//------------------------------------------------------------------------------
// call OnError event handler
//------------------------------------------------------------------------------
procedure TACRServerSession.DoOnError(ErrorCode: Integer; NativeError: Integer = -1; ErrorMessage: String = '');
begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog('Error on server session!');
aaWriteToLog('------------------------------------------------------------------');
aaWriteToLog('SessionID='+IntToStr(Integer(self.SessionID)));
aaWriteToLog('ErrorCode='+IntToStr(Integer(ErrorCode)));
aaWriteToLog('NativeError='+IntToStr(Integer(NativeError)));
aaWriteToLog('ErrorMessage:"'+ErrorMessage+'"');
aaWriteToLog('GetTickCount = '+IntToStr(aaGetTickCount));
aaWriteToLog('==================================================================');
{$ENDIF}
  if (FServer <> nil) then
   FServer.DoOnError(ErrorCode,NativeError,ErrorMessage);
end; // DoOnError


//------------------------------------------------------------------------------
// Send command error occured - session must be destroyed
//------------------------------------------------------------------------------
procedure TACRServerSession.DoCloseSessionOnNetworkError;
begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server DoCloseSessionOnNetworkError starting, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
  FServer.FConnectionManager.TerminateSession(self);
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Server DoCloseSessionOnNetworkError finish, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
end; // DoCloseSessionOnNetworkError


//------------------------------------------------------------------------------
// execute received command
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteReceivedCommand;
begin
  try
    case FReceivedCommandHeader.Request of
     accrqConnectDatabase:                ExecuteConnectDatabase;
     accrqIsDatabaseExists:               ExecuteIsDatabaseExists;
     accrqGetTablesList:                  ExecuteGetTablesList;
     accrqIsTableExists:                  ExecuteIsTableExists;
     accrqGetFormatVersion:               ExecuteGetFormatVersion;
     accrqGetTotalPageCount:              ExecuteGetTotalPageCount;
     accrqGetFreePageCount:               ExecuteGetFreePageCount;
     accrqIsDatabaseEncrypted:            ExecuteIsDatabaseEncrypted;
     accrqIsDatabaseEncryptedByPassword:  ExecuteIsDatabaseEncryptedByPassword;
     accrqIsCryptoParamsValid:            ExecuteIsCryptoParamsValid;
     accrqIsInTransaction:                ExecuteIsInTransaction;
     accrqStartTransaction:               ExecuteStartTransaction;
     accrqCommit:                         ExecuteCommit;
     accrqRollback:                       ExecuteRollback;
     accrqCreateTable:                    ExecuteCreateTable;
     accrqDeleteTable:                    ExecuteDeleteTable;
     accrqEmptyTable:                     ExecuteEmptyTable;
     accrqRenameTable:                    ExecuteRenameTable;
     accrqRenameField:                    ExecuteRenameField;
     accrqOpenTable:                      ExecuteOpenTable;
     accrqCloseTable:                     ExecuteCloseTable;
     accrqApplyDefaultValues:             ExecuteApplyDefaultValues;
     accrqAddIndex:                       ExecuteAddIndex;
     accrqDeleteIndex:                    ExecuteDeleteIndex;
     accrqDeleteAllIndexes:               ExecuteDeleteAllIndexes;
     accrqGetRecordCount:                 ExecuteGetRecordCount;
     accrqGetRecordBuffer:                ExecuteGetRecordBuffer;
     accrqSetRecNo:                       ExecuteSetRecNo;
     accrqGetRecNo:                       ExecuteGetRecNo;
     accrqInternalEdit:                   ExecuteInternalEdit;
     accrqInternalCancel:                 ExecuteInternalCancel;
     accrqInternalPost:                   ExecuteInternalPost;
     accrqInternalDelete:                 ExecuteInternalDelete;
     accrqActivateFilters:                ExecuteActivateFilters;
     accrqDeactivateFilters:              ExecuteDeactivateFilters;
     accrqLocate:                         ExecuteLocate;
     accrqFindKey:                        ExecuteFindKey;
     accrqResetRange:                     ExecuteResetRange;
     accrqApplyRange:                     ExecuteApplyRange;
     accrqReadBLOBValue:                  ExecuteReadBLOBValue;
     accrqExecSQL:                        ExecuteExecSQL;
     accrqIsRecordExists:                 ExecuteIsRecordExists;
    end; // Request
  finally
   FCommandReceived := False;
   FReceivedCommandDataStream.Size := 0;
  end;
end; // ExecuteReceivedCommand


//------------------------------------------------------------------------------
// receive data from network and move it to ReceivedCommandHeader and ReceivedCommandDataStream
//------------------------------------------------------------------------------
procedure TACRServerSession.ReceiveData(Buffer: PChar; BufferSize: Integer);
begin
  try
   inherited ReceiveData(Buffer,BufferSize);
{$IFDEF DEBUG_LOG_COMMUNICATION}
aaWriteToLog(#13#10+'vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv');
aaWriteToLog('S> ServerSession is starting to execute a received command...');
aaWriteToLog('S> SessionID = '+IntToStr(SessionID)+', ServerID = '+IntToStr(FServer.ServerID));
aaWriteToLog('S> Request = '+IntToStr(Integer(FReceivedCommandHeader.Request)));
aaWriteBufferToLog(Buffer,BufferSize);
aaWriteToLog('S> Execute Start Time = '+aaGetCurrentTimeAsString);
{$ENDIF}
  finally
   if (Buffer <> nil) and (BufferSize > 0) then
    MemoryManager.FreeAndNilMem(Buffer);
  end;
  if (FCommandReceived) then
    ExecuteReceivedCommand;
{$IFDEF DEBUG_LOG_COMMUNICATION}
aaWriteToLog('S> Execute End Time = '+aaGetCurrentTimeAsString);
aaWriteToLog('S> End of Request = '+IntToStr(Integer(FReceivedCommandHeader.Request)));
aaWriteToLog('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^'+#13#10);
{$ENDIF}
end; // ReceiveData


//------------------------------------------------------------------------------
// send buffer via established connection using connection manager
//------------------------------------------------------------------------------
procedure TACRServerSession.SendBuffer(var Buffer: PChar; BufferSize: Integer);
begin
{$IFDEF DEBUG_LOG_COMMUNICATION}
aaWriteToLog(#13#10+'vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv');
aaWriteToLog('S> ServerSession is starting to send a reply...');
aaWriteToLog('S> SessionID = '+IntToStr(SessionID)+', ServerID = '+IntToStr(FServer.ServerID));
aaWriteToLog('S> Request = '+IntToStr(Integer(FSentCommandHeader.Request)));
aaWriteToLog('S> Reply = '+IntToStr(Integer(FSentCommandHeader.Reply)));
aaWriteBufferToLog(Buffer,BufferSize);
aaWriteToLog('S> Send Start Time = '+aaGetCurrentTimeAsString);
{$ENDIF}
  FServer.ConnectionManager.SendBuffer(Self,Buffer,BufferSize);
{$IFDEF DEBUG_LOG_COMMUNICATION}
aaWriteToLog('S> Send End Time = '+aaGetCurrentTimeAsString);
aaWriteToLog('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^'+#13#10);
{$ENDIF}
end; // SendBuffer


//------------------------------------------------------------------------------
// send custom message
//------------------------------------------------------------------------------
procedure TACRServerSession.SendMessage(Buffer: PChar; BufferSize: Integer);
begin
  FServer.ConnectionManager.SendMessage(Self,Buffer,BufferSize);
end; // SendMessage


//------------------------------------------------------------------------------
// receive custom message from client
//------------------------------------------------------------------------------
procedure TACRServerSession.ReceiveMessage(Buffer: PChar; BufferSize: Integer);
begin
  FServer.ReceiveMessage(Self,Buffer,BufferSize);
end; // ReceiveMessage


//------------------------------------------------------------------------------
// return client info
//------------------------------------------------------------------------------
function TACRServerSession.GetClientInfo: TACRClientInfo;
var Host,Application: String;
    Port:             Integer;
begin
  FServer.ConnectionManager.GetClientInfo(Self,Host,Port,Application);
  Result.Host := Host;
  Result.Port := Port;
  Result.Application := Application;
  Result.DatabaseName := Self.DatabaseName;
  Result.DatabaseFileName := Self.DatabaseFileName;
  Result.SessionID := Self.SessionID;
end; // GetClientInfo


//------------------------------------------------------------------------------
// execute ConnectDatabase request
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteConnectDatabase;
var Len,DBNumber:       Integer;
    DBName:             String;
    b:                  ByteBool;
begin
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10866);
    SetLength(DBName,Len);
    LoadDataFromStream(PChar(@DBName[1])^,Len,FReceivedCommandDataStream,10867);
    LoadCryptoParamsFromStream(FCryptoParams,FReceivedCommandDataStream,11316);

  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  FDatabaseName := DBName;
  DBNumber := FServer.FindDatabaseByName(DBName);
  if (DBNumber < 0) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseDoesNotExists,
                     Format(ErrorL_CS_ErrorDatabaseDoesNotExists,[DBName]));
   end
  else
   begin
    DatabaseFileName := FServer.DatabaseFileNames[DBNumber];
    try
      SetConnected(True);
      FSentCommandHeader.Request := accrqConnectDatabase;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      SaveDataToStream(FOptions,SizeOf(FOptions),FSentCommandDataStream,10869);
      b := FHandle.Exclusive;
      SaveDataToStream(b,SizeOf(b),FSentCommandDataStream,11067);
    except
    on e: Exception do
     begin
      SendErrorMessage(ACR_CS_ErrorDatabaseFileCannotBeOpen,
                       Format(ErrorL_CS_ErrorDatabaseFileCannotBeOpen,[e.Message]));
      Exit;
     end;
    end;
    SendCommand;
   end;
end; // ExecuteConnectDatabase


//------------------------------------------------------------------------------
// execute IsDatabaseExists
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteIsDatabaseExists;
var Len:    Integer;
    DBName: String;
begin
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10875);
    SetLength(DBName,Len);
    LoadDataFromStream(PChar(@DBName[1])^,Len,FReceivedCommandDataStream,10876);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  FSentCommandHeader.Request := accrqIsDatabaseExists;
  SentCommandDataStream.Size := 0;
  if (FServer.FindDatabaseByName(DBName) < 0) then
   FSentCommandHeader.Reply := accrplNo
  else
   FSentCommandHeader.Reply := accrplYes;
  SendCommand;
end; // ExecuteIsDatabaseExists


//------------------------------------------------------------------------------
// execute GetTablesList
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteGetTablesList;
var Len,i,Count: Integer;
    s:           String;
    List:        TStringList;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  FSentCommandHeader.Request := accrqGetTablesList;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  SentCommandDataStream.Size := 0;
  List := TStringList.Create;
  try
   try
    GetTablesList(List);
    Count := List.Count;
    SaveDataToStream(Count,SizeOf(Count),FSentCommandDataStream,10881);
    for i := 0 to Count - 1 do
     begin
      s := List.Strings[i];
      Len := Length(s);
      SaveDataToStream(Len,SizeOf(Len),SentCommandDataStream,10882);
      SaveDataToStream(PChar(@s[1])^,Len,SentCommandDataStream,10883);
     end;
   except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorGetTablesList,
                        Format(ErrorL_CS_ErrorGetTablesList,[e.Message]));
       Exit;
     end;
   end;
   SendCommand;
  finally
    List.Free;
  end;
end; // ExecuteGetTablesList


//------------------------------------------------------------------------------
// execute IsTableExists
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteIsTableExists;
var Len:        Integer;
    TableName:  String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10886);
    SetLength(TableName,Len);
    LoadDataFromStream(PChar(@TableName[1])^,Len,FReceivedCommandDataStream,10887);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  FSentCommandHeader.Request := accrqIsTableExists;
  SentCommandDataStream.Size := 0;
  if (TableExists(TableName)) then
   FSentCommandHeader.Reply := accrplYes
  else
   FSentCommandHeader.Reply := accrplNo;
  SendCommand;
end; // ExecuteIsTableExists


//------------------------------------------------------------------------------
// execute GetFormatVersion
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteGetFormatVersion;
var Version: Double;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  Version := GetFormatVersion;
  FSentCommandHeader.Request := accrqGetFormatVersion;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  FSentCommandDataStream.Size := 0;
  SaveDataToStream(Version,SizeOf(Version),FSentCommandDataStream,10898);
  SendCommand;
end; // ExecuteGetFormatVersion


//------------------------------------------------------------------------------
// execute GetTotalPageCount
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteGetTotalPageCount;
var PageCount: Integer;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  PageCount := GetTotalPageCount;
  FSentCommandHeader.Request := accrqGetTotalPageCount;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  FSentCommandDataStream.Size := 0;
  SaveDataToStream(PageCount,SizeOf(PageCount),FSentCommandDataStream,10899);
  SendCommand;
end; // ExecuteGetTotalPageCount


//------------------------------------------------------------------------------
// execute GetFreePageCount
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteGetFreePageCount;
var PageCount: Integer;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  PageCount := GetFreePageCount;
  FSentCommandHeader.Request := accrqGetFreePageCount;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  FSentCommandDataStream.Size := 0;
  SaveDataToStream(PageCount,SizeOf(PageCount),FSentCommandDataStream,10900);
  SendCommand;
end; // ExecuteGetFreePageCount


//------------------------------------------------------------------------------
// execute IsDatabaseEncrypted
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteIsDatabaseEncrypted;
var Len:        Integer;
    DBName:     String;
    Res:        Boolean;
    DBNumber:   Integer;
begin
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10902);
    SetLength(DBName,Len);
    LoadDataFromStream(PChar(@DBName[1])^,Len,FReceivedCommandDataStream,10903);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  if (FConnected) then
    Res := IsDatabaseEncrypted
  else
   begin
    DBNumber := FServer.FindDatabaseByName(DBName);
    if (DBNumber < 0) then
     begin
      SendErrorMessage(ACR_CS_ErrorDatabaseDoesNotExists,
                       Format(ErrorL_CS_ErrorDatabaseDoesNotExists,[DBName]));
      Exit;
     end;
    try
      FHandle := TACRLocalSession.Create;
      try
        FHandle.DatabaseName := GetTemporaryName('ACRServerDBName'+'_'+IntToStr(SessionID));
        FHandle.DatabaseFileName := FServer.DatabaseFileNames[DBNumber];
        Res := FHandle.IsDatabaseEncrypted;
      finally
        FHandle.Free;
      end;
    except
      SendErrorMessage(ACR_CS_ErrorDatabaseDoesNotExists,
                       Format(ErrorL_CS_ErrorDatabaseDoesNotExists,[DBName]));
      Exit;
    end;
   end; // not connected
  FSentCommandHeader.Request := accrqIsDatabaseEncrypted;
  if (Res) then
   FSentCommandHeader.Reply := accrplYes
  else
   FSentCommandHeader.Reply := accrplNo;
  SentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteIsDatabaseEncrypted


//------------------------------------------------------------------------------
// execute IsDatabaseEncryptedByPassword
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteIsDatabaseEncryptedByPassword;
var Len:        Integer;
    DBName:     String;
    Res:        Boolean;
    DBNumber:   Integer;
begin
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10904);
    SetLength(DBName,Len);
    LoadDataFromStream(PChar(@DBName[1])^,Len,FReceivedCommandDataStream,10905);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  if (FConnected) then
    Res := IsDatabaseEncryptedByPassword
  else
   begin
    DBNumber := FServer.FindDatabaseByName(DBName);
    if (DBNumber < 0) then
     begin
      SendErrorMessage(ACR_CS_ErrorDatabaseDoesNotExists,
                       Format(ErrorL_CS_ErrorDatabaseDoesNotExists,[DBName]));
      Exit;
     end;
    try
      FHandle := TACRLocalSession.Create;
      try
        FHandle.DatabaseName := GetTemporaryName('ACRServerDBName'+'_'+IntToStr(SessionID));
        FHandle.DatabaseFileName := FServer.DatabaseFileNames[DBNumber];
        Res := FHandle.IsDatabaseEncryptedByPassword;
      finally
        FHandle.Free;
      end;
    except
      SendErrorMessage(ACR_CS_ErrorDatabaseDoesNotExists,
                       Format(ErrorL_CS_ErrorDatabaseDoesNotExists,[DBName]));
      Exit;
    end;
   end; // not connected
  FSentCommandHeader.Request := accrqIsDatabaseEncryptedByPassword;
  if (Res) then
   FSentCommandHeader.Reply := accrplYes
  else
   FSentCommandHeader.Reply := accrplNo;
  SentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteIsDatabaseEncryptedByPassword


//------------------------------------------------------------------------------
// execute IsCryptoParamsValid
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteIsCryptoParamsValid;
var Len:        Integer;
    DBName:     String;
    Res:        Boolean;
    DBNumber:   Integer;
begin
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,11317);
    SetLength(DBName,Len);
    LoadDataFromStream(PChar(@DBName[1])^,Len,FReceivedCommandDataStream,11318);
    LoadCryptoParamsFromStream(FCryptoParams,FReceivedCommandDataStream,11313);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;

  if (FConnected) then
    Res := IsCryptoParamsValid
  else
   begin
    DBNumber := FServer.FindDatabaseByName(DBName);
    if (DBNumber < 0) then
     begin
      SendErrorMessage(ACR_CS_ErrorDatabaseDoesNotExists,
                       Format(ErrorL_CS_ErrorDatabaseDoesNotExists,[DBName]));
      Exit;
     end;
    try
      FHandle := TACRLocalSession.Create;
      try
        FHandle.DatabaseName := GetTemporaryName('ACRServerDBName'+'_'+IntToStr(SessionID));
        FHandle.DatabaseFileName := FServer.DatabaseFileNames[DBNumber];
        Res := IsCryptoParamsValid;
      finally
        FHandle.Free;
      end;
    except
      SendErrorMessage(ACR_CS_ErrorDatabaseDoesNotExists,
                       Format(ErrorL_CS_ErrorDatabaseDoesNotExists,[DBName]));
      Exit;
    end;
   end; // not connected

  FSentCommandHeader.Request := accrqIsCryptoParamsValid;
  if (Res) then
   FSentCommandHeader.Reply := accrplYes
  else
   FSentCommandHeader.Reply := accrplNo;
  SentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteIsCryptoParamsValid


//------------------------------------------------------------------------------
// execute IsInTransaction
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteIsInTransaction;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  FSentCommandHeader.Request := accrqIsInTransaction;
  if (GetInTransaction) then
    FSentCommandHeader.Reply := accrplYes
  else
    FSentCommandHeader.Reply := accrplNo;
  FSentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteIsInTransaction


//------------------------------------------------------------------------------
// execute StartTransaction
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteStartTransaction;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    StartTransaction;
  except
    SendErrorMessage(ACR_CS_ErrorStartTransactionFailed,
                     Format(ErrorL_CS_ErrorStartTransactionFailed,[FDatabaseName]));
    Exit;
  end;
  FSentCommandHeader.Request := accrqIsInTransaction;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  FSentCommandDataStream.Size := 0;
  SendCommand; 
end; // ExecuteStartTransaction 
	
	
//------------------------------------------------------------------------------ 
// execute Commit 
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteCommit;
var b: ByteBool; 
begin
  if (not Connected) then 
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected, 
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit; 
   end;
  try 
    LoadDataFromStream(b,SizeOf(b),FReceivedCommandDataStream,10914); 
  except
    on e: Exception do 
     begin 
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams, 
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message])); 
       Exit; 
     end;
  end;
  try
    Commit(Boolean(b));
  except
    SendErrorMessage(ACR_CS_ErrorCommitFailed, 
                     Format(ErrorL_CS_ErrorCommitFailed,[FDatabaseName]));
    Exit; 
  end;
  FSentCommandHeader.Request := accrqIsInTransaction; 
  FSentCommandHeader.Reply := accrplOperationSucceed; 
  FSentCommandDataStream.Size := 0;
  SendCommand; 
end; // ExecuteCommit
 
 
//------------------------------------------------------------------------------
// execute Rollback
//------------------------------------------------------------------------------ 
procedure TACRServerSession.ExecuteRollback; 
begin 
  if (not Connected) then 
   begin 
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected, 
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName])); 
    Exit;
   end;
  try 
    Rollback;
  except
    SendErrorMessage(ACR_CS_ErrorRollbackFailed, 
                     Format(ErrorL_CS_ErrorRollbackFailed,[FDatabaseName]));
    Exit;
  end; 
  FSentCommandHeader.Request := accrqIsInTransaction; 
  FSentCommandHeader.Reply := accrplOperationSucceed; 
  FSentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteRollback 
	
	
//------------------------------------------------------------------------------ 
// execute CreateTable 
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteCreateTable;
var TempCursor:     TACRLocalCursor;
    Len:            Integer; 
    TableName:      String; 
    FieldDefs:      TACRFieldDefs; 
    IndexDefs:      TACRIndexDefs; 
    ConstraintDefs: TACRConstraintDefs;
begin 
  if (not Connected) then 
   begin 
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end; 
  FieldDefs := TACRFieldDefs.Create;
  IndexDefs := TACRIndexDefs.Create;
  ConstraintDefs := TACRConstraintDefs.Create;
  try 
    try 
      LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10927);
      SetLength(TableName,Len);
      LoadDataFromStream(PChar(@TableName[1])^,Len,FReceivedCommandDataStream,10928);
      FieldDefs.LoadFromStream(FReceivedCommandDataStream);
      IndexDefs.LoadFromStream(FReceivedCommandDataStream);
      ConstraintDefs.LoadFromStream(FReceivedCommandDataStream);
    except 
      on e: Exception do
       begin
         SendErrorMessage(ACR_CS_ErrorReadingRequestParams, 
                          Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
         Exit;
       end; 
    end; 
    try 
      TempCursor := TACRLocalCursor.Create; 
      try
        TempCursor.Session := FHandle; 
        TempCursor.Temporary := False; 
        TempCursor.InMemory := False;
        TempCursor.TableName := TableName;
        TempCursor.ReadOnly := False;
        TempCursor.Exclusive := True;
        TempCursor.CreateTable(FieldDefs,IndexDefs,ConstraintDefs); 
      finally 
        TempCursor.Free;
      end;
    except 
      on e: Exception do 
       begin 
         SendErrorMessage(ACR_CS_ErrorCreateTableFailed,
                          Format(ErrorL_CS_ErrorCreateTableFailed,[e.Message]));
         Exit;
       end;
    end;
    FSentCommandHeader.Request := accrqCreateTable; 
    FSentCommandHeader.Reply := accrplOperationSucceed;
    FSentCommandDataStream.Size := 0; 
    SendCommand;
  finally
    ConstraintDefs.Free;
    IndexDefs.Free;
    FieldDefs.Free;
  end;
end; // ExecuteCreateTable


//------------------------------------------------------------------------------
// execute DeleteTable
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteDeleteTable;
var TempCursor:     TACRLocalCursor;
    Len:            Integer;
    TableName:      String;
    b:              ByteBool;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10945);
    SetLength(TableName,Len);
    LoadDataFromStream(PChar(@TableName[1])^,Len,FReceivedCommandDataStream,10946);
    LoadDataFromStream(b,SizeOf(b),FReceivedCommandDataStream,11480);

  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    TempCursor := TACRLocalCursor.Create;
    try
      TempCursor.Session := FHandle;
      TempCursor.Temporary := False;
      TempCursor.InMemory := False;
      TempCursor.TableName := TableName;
      TempCursor.ReadOnly := False;
      TempCursor.Exclusive := True;
      TempCursor.DeleteTable(b);
    finally
      TempCursor.Free;
    end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorDeleteTableFailed,
                        Format(ErrorL_CS_ErrorDeleteTableFailed,[e.Message]));
       Exit;
     end;
  end;
  FSentCommandHeader.Request := accrqDeleteTable;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  FSentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteDeleteTable


//------------------------------------------------------------------------------
// execute EmptyTable
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteEmptyTable;
var TempCursor:     TACRLocalCursor;
    Len:            Integer;
    TableName:      String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10947);
    SetLength(TableName,Len);
    LoadDataFromStream(PChar(@TableName[1])^,Len,FReceivedCommandDataStream,10948);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    TempCursor := TACRLocalCursor.Create;
    try
      TempCursor.Session := FHandle;
      TempCursor.Temporary := False;
      TempCursor.InMemory := False;
      TempCursor.TableName := TableName;
      TempCursor.ReadOnly := False;
      TempCursor.Exclusive := True;
      TempCursor.EmptyTable;
    finally
      TempCursor.Free;
    end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorEmptyTableFailed,
                        Format(ErrorL_CS_ErrorEmptyTableFailed,[e.Message]));
       Exit;
     end;
  end;
  FSentCommandHeader.Request := accrqEmptyTable;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  FSentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteEmptyTable


//------------------------------------------------------------------------------
// execute RenameTable
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteRenameTable;
var TempCursor:     TACRLocalCursor;
    Len:            Integer;
    TableName:      String;
    NewTableName:   String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10949);
    SetLength(TableName,Len);
    LoadDataFromStream(PChar(@TableName[1])^,Len,FReceivedCommandDataStream,10950);
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10951);
    SetLength(NewTableName,Len);
    LoadDataFromStream(PChar(@NewTableName[1])^,Len,FReceivedCommandDataStream,10952);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    TempCursor := TACRLocalCursor.Create;
    try
      TempCursor.Session := FHandle;
      TempCursor.Temporary := False;
      TempCursor.InMemory := False;
      TempCursor.TableName := TableName;
      TempCursor.ReadOnly := False;
      TempCursor.Exclusive := True;
      TempCursor.RenameTable(NewTableName);
    finally
      TempCursor.Free;
    end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorRenameTableFailed,
                        Format(ErrorL_CS_ErrorRenameTableFailed,[e.Message]));
       Exit;
     end;
  end;
  FSentCommandHeader.Request := accrqRenameTable;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  FSentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteRenameTable


//------------------------------------------------------------------------------
// execute RenameField
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteRenameField;
var TempCursor:     TACRLocalCursor;
    Len:            Integer;
    TableName:      String;
    FieldName:      String;
    NewFieldName:   String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10953);
    SetLength(TableName,Len);
    LoadDataFromStream(PChar(@TableName[1])^,Len,FReceivedCommandDataStream,10954);
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10955);
    SetLength(FieldName,Len);
    LoadDataFromStream(PChar(@FieldName[1])^,Len,FReceivedCommandDataStream,10956);
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10957);
    SetLength(NewFieldName,Len);
    LoadDataFromStream(PChar(@NewFieldName[1])^,Len,FReceivedCommandDataStream,10958);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    TempCursor := TACRLocalCursor.Create;
    try
      TempCursor.Session := FHandle;
      TempCursor.Temporary := False;
      TempCursor.InMemory := False;
      TempCursor.TableName := TableName;
      TempCursor.ReadOnly := False;
      TempCursor.Exclusive := True;
      TempCursor.RenameField(FieldName,NewFieldName);
    finally
      TempCursor.Free;
    end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorRenameFieldFailed,
                        Format(ErrorL_CS_ErrorRenameFieldFailed,[e.Message]));
       Exit;
     end;
  end;
  FSentCommandHeader.Request := accrqRenameField;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  FSentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteRenameField


//------------------------------------------------------------------------------
// execute OpenTable
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteOpenTable;
var LocalCursor:  TACRLocalCursor;
    TableName:    String;
    bExclusive:   ByteBool;
    Len:          Integer;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10965);
    SetLength(TableName,Len);
    LoadDataFromStream(PChar(@TableName[1])^,Len,FReceivedCommandDataStream,10966);
    LoadDataFromStream(bExclusive,SizeOf(bExclusive),FReceivedCommandDataStream,10967);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    LocalCursor := TACRLocalCursor.Create;
    try
      LocalCursor.Session := FHandle;
      LocalCursor.Temporary := False;
      LocalCursor.InMemory := False;
      LocalCursor.TableName := TableName;
      LocalCursor.ReadOnly := False;
      LocalCursor.Exclusive := bExclusive;
      LocalCursor.OpenTableByFieldDefs(nil,nil,nil);
      LocalCursor.InternalInitFieldDefs;
      Inc(FLastCursorID);
      // prepare reply
      FSentCommandHeader.Request := accrqOpenTable;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      CreateServerCursor(LocalCursor,nil);
    except
      LocalCursor.Free;
      raise;
    end;
    // send reply
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorOpenTableFailed,
                        Format(ErrorL_CS_ErrorOpenTableFailed,[e.Message]));
       Exit;
     end;
  end;
  SendCommand;
end; // ExecuteOpenTable


//------------------------------------------------------------------------------
// execute CloseTable
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteCloseTable;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,10969);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FCursorList.Remove(ServerCursor);
      ServerCursor.Free;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorCloseTableFailed,
                        Format(ErrorL_CS_ErrorCloseTableFailed,[msg]));
      Exit;
     end;
    // prepare reply
    FSentCommandHeader.Request := accrqCloseTable;
    FSentCommandHeader.Reply := accrplOperationSucceed;
    FSentCommandDataStream.Size := 0;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorCloseTableFailed,
                        Format(ErrorL_CS_ErrorCloseTableFailed,[e.Message]));
       Exit;
     end;
  end;
  SendCommand;
end; // ExecuteCloseTable


//------------------------------------------------------------------------------
// execute ApplyDefaultValues
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteApplyDefaultValues;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,10992);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqApplyDefaultValues;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.ApplyDefaultValues;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorApplyDefaultValuesFailed,
                        Format(ErrorL_CS_ErrorApplyDefaultValuesFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorApplyDefaultValuesFailed,
                        Format(ErrorL_CS_ErrorApplyDefaultValuesFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteApplyDefaultValues


//------------------------------------------------------------------------------
// execute AddIndex
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteAddIndex;
var
    TableName:    String;
    Len:          Integer;
    IndexDef:     TACRIndexDef;
    LocalCursor:  TACRLocalCursor;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  IndexDef := TACRIndexDef.Create;
  try
    try
      LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10995);
      SetLength(TableName,Len);
      LoadDataFromStream(PChar(@TableName[1])^,Len,FReceivedCommandDataStream,10996);
      IndexDef.LoadFromStream(FReceivedCommandDataStream);
    except
      on e: Exception do
       begin
         SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                          Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
         Exit;
       end;
    end;
    try
      LocalCursor := TACRLocalCursor.Create;
      try
        LocalCursor.Session := FHandle;
        LocalCursor.TableName := TableName;
        LocalCursor.Exclusive := True;
        LocalCursor.ReadOnly := False;
        LocalCursor.InMemory := False;
        LocalCursor.Temporary := False;
        LocalCursor.AddIndex(IndexDef);
      finally
        LocalCursor.Free;
      end;
    except
      on e: Exception do
       begin
         SendErrorMessage(ACR_CS_ErrorAddIndexFailed,
                          Format(ErrorL_CS_ErrorAddIndexFailed,[e.Message]));
         Exit;
       end;
    end;
  finally
    IndexDef.Free;
  end;
  FSentCommandHeader.Request := accrqAddIndex;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  FSentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteAddIndex


//------------------------------------------------------------------------------
// execute DeleteIndex
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteDeleteIndex;
var
    TableName:    String;
    IndexName:    String;
    Len:          Integer;
    LocalCursor:  TACRLocalCursor;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10997);
    SetLength(TableName,Len);
    LoadDataFromStream(PChar(@TableName[1])^,Len,FReceivedCommandDataStream,10998);
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10999);
    SetLength(IndexName,Len);
    LoadDataFromStream(PChar(@IndexName[1])^,Len,FReceivedCommandDataStream,11000);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    LocalCursor := TACRLocalCursor.Create;
    try
      LocalCursor.Session := FHandle;
      LocalCursor.TableName := TableName;
      LocalCursor.Exclusive := True;
      LocalCursor.ReadOnly := False;
      LocalCursor.InMemory := False;
      LocalCursor.Temporary := False;
      LocalCursor.DeleteIndex(IndexName);
    finally
      LocalCursor.Free;
    end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorDeleteIndexFailed,
                        Format(ErrorL_CS_ErrorDeleteIndexFailed,[e.Message]));
       Exit;
     end;
  end;
  FSentCommandHeader.Request := accrqDeleteIndex;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  FSentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteDeleteIndex


//------------------------------------------------------------------------------
// execute DeleteAllIndexes
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteDeleteAllIndexes;
var
    TableName:    String;
    Len:          Integer;
    LocalCursor:  TACRLocalCursor;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,11001);
    SetLength(TableName,Len);
    LoadDataFromStream(PChar(@TableName[1])^,Len,FReceivedCommandDataStream,11002);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    LocalCursor := TACRLocalCursor.Create;
    try
      LocalCursor.Session := FHandle;
      LocalCursor.TableName := TableName;
      LocalCursor.Exclusive := True;
      LocalCursor.ReadOnly := False;
      LocalCursor.InMemory := False;
      LocalCursor.Temporary := False;
      LocalCursor.DeleteAllIndexes;
    finally
      LocalCursor.Free;
    end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorDeleteAllIndexesFailed,
                        Format(ErrorL_CS_ErrorDeleteAllIndexesFailed,[e.Message]));
       Exit;
     end;
  end;
  FSentCommandHeader.Request := accrqDeleteAllIndexes;
  FSentCommandHeader.Reply := accrplOperationSucceed;
  FSentCommandDataStream.Size := 0;
  SendCommand;
end; // ExecuteDeleteAllIndexes


//------------------------------------------------------------------------------
// execute GetRecordCount
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteGetRecordCount;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11003);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqGetRecordCount;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.GetRecordCount;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorGetRecordCountFailed,
                        Format(ErrorL_CS_ErrorGetRecordCountFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorGetRecordCountFailed,
                        Format(ErrorL_CS_ErrorGetRecordCountFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteGetRecordCount


//------------------------------------------------------------------------------
// execute GetRecordBuffer
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteGetRecordBuffer;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11004);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqGetRecordBuffer;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.GetRecordBuffer;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorGetRecordBufferFailed,
                        Format(ErrorL_CS_ErrorGetRecordBufferFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorGetRecordBufferFailed,
                        Format(ErrorL_CS_ErrorGetRecordBufferFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteGetRecordBuffer


//------------------------------------------------------------------------------
// execute SetRecNo
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteSetRecNo;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11016);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqSetRecNo;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.SetRecNo;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorSetRecNoFailed,
                        Format(ErrorL_CS_ErrorSetRecNoFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorSetRecNoFailed,
                        Format(ErrorL_CS_ErrorSetRecNoFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteSetRecNo


//------------------------------------------------------------------------------
// execute GetRecNo
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteGetRecNo;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11019);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqGetRecNo;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.GetRecNo;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorGetRecNoFailed,
                        Format(ErrorL_CS_ErrorGetRecNoFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorGetRecNoFailed,
                        Format(ErrorL_CS_ErrorGetRecNoFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteSetRecNo


//------------------------------------------------------------------------------
// execute InternalEdit
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteInternalEdit;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11044);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqInternalEdit;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.InternalEdit;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorInternalEditFailed,
                        Format(ErrorL_CS_ErrorInternalEditFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorInternalEditFailed,
                        Format(ErrorL_CS_ErrorInternalEditFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteInternalEdit


//------------------------------------------------------------------------------
// execute InternalCancel
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteInternalCancel;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11045);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqInternalCancel;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.InternalCancel;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorInternalCancelFailed,
                        Format(ErrorL_CS_ErrorInternalCancelFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorInternalCancelFailed,
                        Format(ErrorL_CS_ErrorInternalCancelFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteInternalCancel


//------------------------------------------------------------------------------
// execute InternalPost
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteInternalPost;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11046);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqInternalPost;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.InternalPost;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorInternalPostFailed,
                        Format(ErrorL_CS_ErrorInternalPostFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorInternalPostFailed,
                        Format(ErrorL_CS_ErrorInternalPostFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteInternalPost


//------------------------------------------------------------------------------
// execute InternalDelete
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteInternalDelete;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11047);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqInternalDelete;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.InternalDelete;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorInternalDeleteFailed,
                        Format(ErrorL_CS_ErrorInternalDeleteFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorInternalDeleteFailed,
                        Format(ErrorL_CS_ErrorInternalDeleteFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteInternalDelete


//------------------------------------------------------------------------------
// execute ActivateFilters
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteActivateFilters;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11047);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqActivateFilters;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.ActivateFilters;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorActivateFiltersFailed,
                        Format(ErrorL_CS_ErrorActivateFiltersFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorActivateFiltersFailed,
                        Format(ErrorL_CS_ErrorActivateFiltersFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteActivateFilters


//------------------------------------------------------------------------------
// execute DeactivateFilters
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteDeactivateFilters;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11047);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqDeactivateFilters;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.DeactivateFilters;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorDeactivateFiltersFailed,
                        Format(ErrorL_CS_ErrorDeactivateFiltersFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorDeactivateFiltersFailed,
                        Format(ErrorL_CS_ErrorDeactivateFiltersFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteDeactivateFilters


//------------------------------------------------------------------------------
// execute Locate
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteLocate;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11047);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqLocate;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.Locate;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorLocateFailed,
                        Format(ErrorL_CS_ErrorLocateFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorLocateFailed,
                        Format(ErrorL_CS_ErrorLocateFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteLocate


//------------------------------------------------------------------------------
// execute FindKey
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteFindKey;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11047);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqFindKey;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.FindKey;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorFindKeyFailed,
                        Format(ErrorL_CS_ErrorFindKeyFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorFindKeyFailed,
                        Format(ErrorL_CS_ErrorFindKeyFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteFindKey


//------------------------------------------------------------------------------
// execute ResetRange
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteResetRange;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11047);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqResetRange;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.ResetRange(False);
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorResetRangeFailed,
                        Format(ErrorL_CS_ErrorResetRangeFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorResetRangeFailed,
                        Format(ErrorL_CS_ErrorResetRangeFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteResetRange


//------------------------------------------------------------------------------
// execute ApplyRange
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteApplyRange;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11047);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqApplyRange;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.ApplyRange;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorApplyRangeFailed,
                        Format(ErrorL_CS_ErrorApplyRangeFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorApplyRangeFailed,
                        Format(ErrorL_CS_ErrorApplyRangeFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteApplyRange


//------------------------------------------------------------------------------
// execute ReadBLOBValue
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteReadBLOBValue;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11047);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqReadBLOBValue;
      FSentCommandHeader.Reply := accrplOperationSucceed;
      FSentCommandDataStream.Size := 0;
      ServerCursor.ReadBLOBValue;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorReadBLOBValueFailed,
                        Format(ErrorL_CS_ErrorReadBLOBValueFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadBLOBValueFailed,
                        Format(ErrorL_CS_ErrorReadBLOBValueFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteReadBLOBValue


//------------------------------------------------------------------------------
// execute ExecSQL
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteExecSQL;
var
    FQuery:                   TACRQuery;
    LocalCursor:              TACRLocalCursor;
    QueryText:                String;
    Len:                      Integer;
    b:                        ByteBool;
    RowsAffected:             Int64;
    bAbort:                   Boolean;
    ClientInfo:               TACRClientInfo;
    TempParams:               TACRSQLParams;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  FQuery := TACRQuery.Create(nil);
  FQuery.DatabaseName := FHandle.DatabaseName;
  try
    LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,11178);
    SetLength(QueryText,Len);
    LoadDataFromStream(PChar(@QueryText[1])^,Len,FReceivedCommandDataStream,11179);
    FQuery.SQL.Text := QueryText;
    // load RequestLive
    LoadDataFromStream(b,SizeOf(b),FReceivedCommandDataStream,11180);
    FQuery.RequestLive := b;
    // load ReadOnly
    LoadDataFromStream(b,SizeOf(b),FReceivedCommandDataStream,11181);
    FQuery.ReadOnly := b;
    TACRQuery(FQuery).Prepare;
    ReceiveParams(FQuery);
  except
    on e: Exception do
     begin
      FQuery.Free;
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ClientInfo := GetClientInfo;
    if (Assigned(FServer.OnSQL)) then
     begin
      TempParams := TACRSQLParams.Create;
      try
        TempParams.Assign(FQuery.StmtHandle.SQLParams);
        FServer.DoOnSQL(ClientInfo,TempParams,QueryText,bAbort);
        if (bAbort) then
         raise EACRException.Create(11249,ErrorLSQLCommandRejectedByServer)
        else
         begin
          FQuery.SQL.Text := QueryText;
          TACRQuery(FQuery).Prepare;
          FQuery.StmtHandle.SQLParams.Assign(TempParams);
         end;
      finally
        TempParams.Free;
      end;
     end;
    LocalCursor := TACRLocalCursor(FQuery.StmtHandle.OpenQuery);
    FSentCommandHeader.Request := accrqExecSQL;
    FSentCommandHeader.Reply := accrplOperationSucceed;
    FSentCommandDataStream.Size := 0;
    RowsAffected := FQuery.StmtHandle.RowsAffected;
    b := (LocalCursor <> nil);
    SaveDataToStream(RowsAffected,SizeOf(RowsAffected),FSentCommandDataStream,11182);
    // if cursor returned
    SaveDataToStream(b,SizeOf(b),FSentCommandDataStream,11183);
    if (b) then
     begin
      b := not FQuery.StmtHandle.ReadOnly;
      SaveDataToStream(b,SizeOf(b),FSentCommandDataStream,11187);
      if (b) then
       begin
        SendLiveQuery(FQuery,LocalCursor);
       end
      else
       begin
         SendNotLiveQuery(FQuery,LocalCursor);
         FQuery.Free;
       end;
     end
    else
     FQuery.Free;
  except
    on e: Exception do
     begin
       FQuery.Free;
       SendErrorMessage(ACR_CS_ErrorExecutingSQLScript,
                        Format(ErrorL_CS_ErrorExecutingSQLScript,[e.Message]));
       Exit;
     end;
  end;
  SendCommand;
end; // ExecuteExecSQL


//------------------------------------------------------------------------------
// execute IsRecordExists
//------------------------------------------------------------------------------
procedure TACRServerSession.ExecuteIsRecordExists;
var
    ServerCursor: TACRServerCursor;
    CursorID:     TACRObjectID;
    msg:          String;
begin
  if (not Connected) then
   begin
    SendErrorMessage(ACR_CS_ErrorDatabaseIsNotConnected,
                     Format(ErrorL_CS_ErrorDatabaseIsNotConnected,[FDatabaseName]));
    Exit;
   end;
  try
    LoadDataFromStream(CursorID,SizeOf(CursorID),FReceivedCommandDataStream,11285);
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorReadingRequestParams,
                        Format(ErrorL_CS_ErrorReadingRequestParams,[e.Message]));
       Exit;
     end;
  end;
  try
    ServerCursor := FindServerCursor(CursorID);
    if (ServerCursor <> nil) then
     begin
      FSentCommandHeader.Request := accrqIsRecordExists;
      FSentCommandHeader.Reply := accrplNo;
      FSentCommandDataStream.Size := 0;
      ServerCursor.IsRecordExists;
     end
    else
     begin
      msg := Format(ErrorLCannotFindServerCursor,[CursorID]);
      SendErrorMessage(ACR_CS_ErrorIsRecordExistsFailed,
                        Format(ErrorL_CS_ErrorIsRecordExistsFailed,[msg]));
      Exit;
     end;
  except
    on e: Exception do
     begin
       SendErrorMessage(ACR_CS_ErrorIsRecordExistsFailed,
                        Format(ErrorL_CS_ErrorIsRecordExistsFailed,[e.Message]));
       Exit;
     end;
  end;
end; // ExecuteIsRecordExists


////////////////////////////////////////////////////////////////////////////////
//
// TACRServerCursor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// send record buffer
//------------------------------------------------------------------------------
procedure TACRServerCursor.SendRecordBuffer(const ErrorCode: Integer; RecordBuffer: TACRRecordBuffer);
begin
  SaveDataToStream(RecordBuffer^,FLocalCursor.RecordBufferSize,TACRServerSession(FSession).FSentCommandDataStream,ErrorCode);
end; // SendRecordBuffer


//------------------------------------------------------------------------------
// receive record buffer
//------------------------------------------------------------------------------
procedure TACRServerCursor.ReceiveRecordBuffer(const ErrorCode: Integer; RecordBuffer: TACRRecordBuffer);
begin
  LoadDataFromStream(RecordBuffer^,FLocalCursor.RecordBufferSize,TACRServerSession(FSession).FReceivedCommandDataStream,ErrorCode);
end; // ReceiveRecordBuffer


//------------------------------------------------------------------------------
// receive modified BLOB values
//------------------------------------------------------------------------------
procedure TACRServerCursor.ReceiveModifiedBLOBValues(ToInsert: Boolean);
var i, FieldNumber, Count: Integer;
    Size:                  Int64;
    LocalBLOBStream:       TACRStream;
begin
  LoadDataFromStream(Count,SizeOf(Count),TACRServerSession(FSession).FReceivedCommandDataStream,11228);
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: receive modified blob values, Count = '+IntToStr(Count));
{$ENDIF}
  for i := 0 to Count - 1 do
   begin
    LoadDataFromStream(FieldNumber,SizeOf(FieldNumber),TACRServerSession(FSession).FReceivedCommandDataStream,11229);
    LoadDataFromStream(Size,SizeOf(Size),TACRServerSession(FSession).FReceivedCommandDataStream,11230);
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: receive modified blob values - opening blob stream, FieldNumber = '+IntToStr(FieldNumber)+
', Size = '+IntToStr(Size));
{$ENDIF}
    LocalBLOBStream := FLocalCursor.InternalCreateBlobStream(ToInsert,FieldNumber,bomWrite);
    LocalBLOBStream.Modified := True;
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: receive modified blob values - blob stream opened ok');
{$ENDIF}
    try
     if (Size > 0) then
      begin
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: receive modified blob values - loading data, Start Position = '+
 IntToStr(TACRServerSession(FSession).FReceivedCommandDataStream.Position)+
 ', Received Stream Size = '+IntToStr(TACRServerSession(FSession).FReceivedCommandDataStream.Size));
{$ENDIF}
       LocalBLOBStream.LoadFromStreamWithPosition(
        TACRServerSession(FSession).FReceivedCommandDataStream,
        TACRServerSession(FSession).FReceivedCommandDataStream.Position,Size);
       TACRServerSession(FSession).FReceivedCommandDataStream.Position :=
        TACRServerSession(FSession).FReceivedCommandDataStream.Position + Size; 
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: receive modified blob values - loading data ok, End Position = '+
 IntToStr(TACRServerSession(FSession).FReceivedCommandDataStream.Position)+
 ', LocalBLOBStream.Size = '+IntToStr(LocalBLOBStream.Size)
);
{$ENDIF}
      end;
    finally
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: receive modified blob values - closing blob ... ');
{$ENDIF}
      FLocalCursor.InternalCloseBLOB(FieldNumber);
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: receive modified blob values - blob closed ok');
{$ENDIF}
    end;
   end;
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: receive modified blob values - ok');
{$ENDIF}
end; // ReceiveModifiedBLOBValues


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerCursor.Create(aCursor: TACRLocalCursor; aCursorID: TACRObjectID; aSession: TACRServerSession; aQuery: TACRQuery);
begin
  FLocalCursor := aCursor;
  FCursorID := aCursorID;
  FSession := aSession;
  FQuery := aQuery;
  FForcedDestroy := False;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRServerCursor.Destroy;
begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
  aaWriteToLog('TACRServerCursor.Destroy starting .... '+
               'Session = '+IntToStr(Integer(Session))+
               ', FLocalCursor = '+IntToStr(Integer(FLocalCursor))+
               ', FQuery = '+IntToStr(Integer(FQuery))+#13#10
                );
  if (FSession <> nil) then
   aaWriteToLog('TACRServerCursor.Destroy starting ... SessionID = '+IntToStr(FSession.SessionID));
 {$ENDIF}
  ResetRange(True);
 {$IFDEF DEBUG_LOG_COMMUNICATION}
  aaWriteToLog('TACRServerCursor.Destroy ResetRange ok! ');
 {$ENDIF}
  if (FQuery <> nil) then
   begin
    if (Query.Handle <> nil) then
     if (Query.Handle is TACRLocalCursor) then
      begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
  aaWriteToLog('TACRServerCursor.Destroy  setting handle.DoNotUnlockTable... ');
 {$ENDIF}
       TACRLocalCursor(Query.Handle).DoNotUnlockTable := FForcedDestroy;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
  aaWriteToLog('TACRServerCursor.Destroy  setting handle.DoNotUnlockTable... OK');
 {$ENDIF}
      end;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
  aaWriteToLog('TACRServerCursor.Destroy  destroying Query... ');
 {$ENDIF}
    FQuery.Free;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
  aaWriteToLog('TACRServerCursor.Destroy  destroying Query... OK');
 {$ENDIF}
   end
  else
   begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
  aaWriteToLog('TACRServerCursor.Destroy setting LocalCursor.DoNotUnlockTable ... ');
  aaWriteToLog('TACRServerCursor.Destroy TableName = '+FLocalCursor.TableName);
 {$ENDIF}
    FLocalCursor.DoNotUnlockTable := FForcedDestroy;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
  aaWriteToLog('TACRServerCursor.Destroy  destroying LocalCursor... ');
 {$ENDIF}
    FLocalCursor.Free;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
  aaWriteToLog('TACRServerCursor.Destroy  destroying LocalCursor... OK');
 {$ENDIF}
   end;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
  aaWriteToLog('TACRServerCursor.Destroy starting inherited');
 {$ENDIF}
  inherited;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
  aaWriteToLog('TACRServerCursor.Destroy  finished');
 {$ENDIF}
end; // Destroy


//------------------------------------------------------------------------------
// apply default values
//------------------------------------------------------------------------------
procedure TACRServerCursor.ApplyDefaultValues;
begin
  FCurrentRecordBuffer := FLocalCursor.AllocateRecordBuffer;
  try
    // read source buffer
    ReceiveRecordBuffer(10993,FCurrentRecordBuffer);
    // update buffer and send it
    FLocalCursor.InternalInitRecord(FCurrentRecordBuffer,True);
    SendRecordBuffer(10994,FCurrentRecordBuffer);
    FSession.SendCommand;
  finally
    FLocalCursor.FreeRecordBuffer(FCurrentRecordBuffer);
  end;
end; // ApplyDefaultValues


//------------------------------------------------------------------------------
// apply default values
//------------------------------------------------------------------------------
procedure TACRServerCursor.GetRecordCount;
var RecordNo: TACRRecordNo;
begin
  RecordNo := FLocalCursor.GetRecordCount;
  SaveDataToStream(RecordNo,SizeOf(RecordNo),FSession.FSentCommandDataStream,11005);
  FSession.SendCommand;
end; // GetRecordCount


//------------------------------------------------------------------------------
// get record buffer
//------------------------------------------------------------------------------
procedure TACRServerCursor.GetRecordBuffer;
var b:                Byte;
    bool:             ByteBool;
    GetRecordMode:    TACRGetRecordMode;
    GetRecordResult:  TACRGetRecordResult;
    RecordID:         TACRRecordID;
    IndexID:          TACRObjectID;
begin
  LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11006);
  GetRecordMode := TACRGetRecordMode(b);
  LoadDataFromStream(bool,SizeOf(bool),FSession.FReceivedCommandDataStream,11007);
  LocalCursor.FirstPosition := bool;
  LoadDataFromStream(bool,SizeOf(bool),FSession.FReceivedCommandDataStream,11008);
  LocalCursor.LastPosition := bool;
  LoadDataFromStream(RecordID,SizeOf(RecordID),FSession.FReceivedCommandDataStream,11009);
  LocalCursor.CurrentRecordID := RecordID;
  LoadDataFromStream(IndexID,SizeOf(IndexID),FSession.FReceivedCommandDataStream,11090);
  LocalCursor.IndexID := IndexID;
  FCurrentRecordBuffer := FLocalCursor.AllocateRecordBuffer;
  try
    LocalCursor.CurrentRecordBuffer := FCurrentRecordBuffer;
    // update buffer and send it
    GetRecordResult := FLocalCursor.GetRecordBuffer(GetRecordMode);
    b := Byte(GetRecordResult);
    SaveDataToStream(b,SizeOf(b),FSession.FSentCommandDataStream,11010);
    bool := LocalCursor.FirstPosition;
    SaveDataToStream(bool,SizeOf(bool),FSession.FSentCommandDataStream,11134);
    bool := LocalCursor.LastPosition;
    SaveDataToStream(bool,SizeOf(bool),FSession.FSentCommandDataStream,11135);
    RecordID := LocalCursor.CurrentRecordID;
    SaveDataToStream(RecordID,SizeOf(RecordID),FSession.FSentCommandDataStream,11136);
    if (GetRecordResult = grrOK) then
      SendRecordBuffer(11011,FCurrentRecordBuffer);
    FSession.SendCommand;
  finally
    FLocalCursor.FreeRecordBuffer(FCurrentRecordBuffer);
  end;
end; // GetRecordBuffer


//------------------------------------------------------------------------------
// set RecNo
//------------------------------------------------------------------------------
procedure TACRServerCursor.SetRecNo;
var Value:    Int64;
    RecordID: TACRRecordID;
    IndexID:  TACRObjectID;
begin
  LoadDataFromStream(Value,SizeOf(Value),FSession.FReceivedCommandDataStream,11020);
  LoadDataFromStream(IndexID,SizeOf(IndexID),FSession.FReceivedCommandDataStream,11091);
  LocalCursor.IndexID := IndexID;
  LocalCursor.SetRecNo(Value);
  RecordID := LocalCursor.CurrentRecordID;
  SaveDataToStream(RecordID,SizeOf(RecordID),FSession.FSentCommandDataStream,11021);
  FSession.SendCommand;
end; // SetRecNo


//------------------------------------------------------------------------------
// get RecNo
//------------------------------------------------------------------------------
procedure TACRServerCursor.GetRecNo;
var Value:    Int64;
    RecordID: TACRRecordID;
    IndexID:  TACRObjectID;
begin
  LoadDataFromStream(RecordID,SizeOf(RecordID),FSession.FReceivedCommandDataStream,11022);
  LoadDataFromStream(IndexID,SizeOf(IndexID),FSession.FReceivedCommandDataStream,11092);
  LocalCursor.IndexID := IndexID;
  LocalCursor.FirstPosition := False;
  LocalCursor.LastPosition := False;
  LocalCursor.CurrentRecordID := RecordID;
  Value := LocalCursor.GetRecNo;
  SaveDataToStream(Value,SizeOf(Value),FSession.FSentCommandDataStream,11023);
  FSession.SendCommand;
end; // GetRecNo


//------------------------------------------------------------------------------
// edit record
//------------------------------------------------------------------------------
procedure TACRServerCursor.InternalEdit;
var
    RecordID: TACRRecordID;
begin
  LoadDataFromStream(RecordID,SizeOf(RecordID),FSession.FReceivedCommandDataStream,11048);
  LocalCursor.FirstPosition := False;
  LocalCursor.LastPosition := False;
  LocalCursor.CurrentRecordID := RecordID;
  LocalCursor.InternalEdit;
  FSession.SendCommand;
end; // InternalEdit


//------------------------------------------------------------------------------
// cancel record
//------------------------------------------------------------------------------
procedure TACRServerCursor.InternalCancel;
var
    RecordID: TACRRecordID;
    b:        ByteBool;
begin
  LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11050);
  LoadDataFromStream(RecordID,SizeOf(RecordID),FSession.FReceivedCommandDataStream,11051);
  LocalCursor.FirstPosition := False;
  LocalCursor.LastPosition := False;
  LocalCursor.CurrentRecordID := RecordID;
  LocalCursor.InternalCancel(b);
  FSession.SendCommand;
end; // InternalCancel


//------------------------------------------------------------------------------
// post record
//------------------------------------------------------------------------------
procedure TACRServerCursor.InternalPost;
var
    RecordID: TACRRecordID;
    b:        ByteBool;
    ToInsert: Boolean;
    IndexID:  TACRObjectID;
begin
    LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11052);
    ToInsert := b;
    LoadDataFromStream(RecordID,SizeOf(RecordID),FSession.FReceivedCommandDataStream,11053);
    LocalCursor.CurrentRecordID := RecordID;
    LoadDataFromStream(IndexID,SizeOf(IndexID),FSession.FReceivedCommandDataStream,11097);
    LocalCursor.IndexID := IndexID;
    FCurrentRecordBuffer := LocalCursor.AllocateRecordBuffer;
    FEditRecordBuffer := LocalCursor.AllocateRecordBuffer;
    try
     try
      ReceiveRecordBuffer(11054,FCurrentRecordBuffer);
      LocalCursor.CurrentRecordBuffer := FCurrentRecordBuffer;
      LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11055);
      if (b) then
       LocalCursor.EditRecordBuffer := nil
      else
       begin
        ReceiveRecordBuffer(11056,FEditRecordBuffer);
        LocalCursor.EditRecordBuffer := FEditRecordBuffer;
       end;
      ReceiveModifiedBLOBValues(ToInsert);
      LocalCursor.InternalPost(ToInsert);
      if (LocalCursor.ErrorCode <> ACR_ERR_OK) then
       begin
        LocalCursor.ClearBLOBStreams(False);
        raise EACRException.Create(11234,ErrorLPostFailed,[LocalCursor.ErrorMessage]);
       end;
      b := LocalCursor.FirstPosition;
      SaveDataToStream(b,SizeOf(b),FSession.FSentCommandDataStream,11057);
      b := LocalCursor.LastPosition;
      SaveDataToStream(b,SizeOf(b),FSession.FSentCommandDataStream,11058);
      RecordID := LocalCursor.CurrentRecordID;
      SaveDataToStream(RecordID,SizeOf(RecordID),FSession.FSentCommandDataStream,11059);
      SendRecordBuffer(11060,FCurrentRecordBuffer);
     except
      on E: Exception do
       begin
        FSession.SendErrorMessage(ACR_CS_ErrorInternalPostFailed,
                         Format(ErrorL_CS_ErrorInternalPostFailed,[e.Message]));
        Exit;
       end;
     end;
     FSession.SendCommand;
    finally
      LocalCursor.FreeRecordBuffer(FCurrentRecordBuffer);
      LocalCursor.FreeRecordBuffer(FEditRecordBuffer);
    end;
end; // InternalPost


//------------------------------------------------------------------------------
// delete record
//------------------------------------------------------------------------------
procedure TACRServerCursor.InternalDelete;
var
    RecordID: TACRRecordID;
    b:        ByteBool;
    IndexID:  TACRObjectID;
begin
  LoadDataFromStream(RecordID,SizeOf(RecordID),FSession.FReceivedCommandDataStream,11061);
  LoadDataFromStream(IndexID,SizeOf(IndexID),FSession.FReceivedCommandDataStream,11098);
  LocalCursor.IndexID := IndexID;
  LocalCursor.FirstPosition := False;
  LocalCursor.LastPosition := False;
  LocalCursor.CurrentRecordID := RecordID;
  FCurrentRecordBuffer := LocalCursor.AllocateRecordBuffer;
  try
    LocalCursor.CurrentRecordBuffer := FCurrentRecordBuffer;
    Move(RecordID,PChar(FCurrentRecordBuffer + LocalCursor.BookmarkOffset)^,sizeof(TACRRecordID));
    LocalCursor.InternalDelete;
    b := LocalCursor.FirstPosition;
    SaveDataToStream(b,SizeOf(b),FSession.FSentCommandDataStream,11062);
    b := LocalCursor.LastPosition;
    SaveDataToStream(b,SizeOf(b),FSession.FSentCommandDataStream,11063);
    RecordID := LocalCursor.CurrentRecordID;
    SaveDataToStream(RecordID,SizeOf(RecordID),FSession.FSentCommandDataStream,11064);
    SendRecordBuffer(11065,FCurrentRecordBuffer);
    FSession.SendCommand;
  finally
    LocalCursor.FreeRecordBuffer(FCurrentRecordBuffer);
  end;
end; // InternalDelete


//------------------------------------------------------------------------------
// Activate filters
//------------------------------------------------------------------------------
procedure TACRServerCursor.ActivateFilters;
var
    Len:              Integer;
    FilterText:       String;
    b:                ByteBool;
    PartialKey:       Boolean;
    CaseInsensitive:  Boolean;
begin
  LoadDataFromStream(Len,SizeOf(Len),FSession.FReceivedCommandDataStream,11077);
  SetLength(FilterText,Len);
  LoadDataFromStream(PChar(@FilterText[1])^,Len,FSession.FReceivedCommandDataStream,11078);
  LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11079);
  CaseInsensitive := b;
  LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11080);
  PartialKey := b;
  LocalCursor.ActivateFilters(FilterText,CaseInsensitive,PartialKey);
  FSession.SendCommand;
end; // ActivateFilters


//------------------------------------------------------------------------------
// Deactivate filters
//------------------------------------------------------------------------------
procedure TACRServerCursor.DeactivateFilters;
begin
  LocalCursor.DeactivateFilters;
  FSession.SendCommand;
end; // DeactivateFilters


//------------------------------------------------------------------------------
// Locate record
//------------------------------------------------------------------------------
procedure TACRServerCursor.Locate;
var
    RecordID:         TACRRecordID;
    b:                ByteBool;
    IndexID:          TACRObjectID;
    Len,i,ArrLen:     Integer;
    KeyFields:        String;
    CaseInsensitive:  Boolean;
    PartialKey:       Boolean;
    v:                TACRVariant;
    KeyValues:        Variant;
begin
  LoadDataFromStream(IndexID,SizeOf(IndexID),FSession.FReceivedCommandDataStream,11114);
  LoadDataFromStream(Len,SizeOf(Len),FSession.FReceivedCommandDataStream,11115);
  SetLength(KeyFields,Len);
  LoadDataFromStream(PChar(@KeyFields[1])^,Len,FSession.FReceivedCommandDataStream,11116);
  LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11117);
  CaseInsensitive := b;
  LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11118);
  PartialKey := b;
  LoadDataFromStream(ArrLen,SizeOf(ArrLen),FSession.FReceivedCommandDataStream,11119);
  v := TACRVariant.Create;
  try
    if (ArrLen = 1) then
     begin
      v.LoadFromStream(FSession.FReceivedCommandDataStream);
      KeyValues := v.AsVariant;
     end
    else
     begin
      KeyValues := VarArrayCreate([0, (ArrLen-1)], varVariant);
      for i := 0 to ArrLen-1 do
       begin
        v.LoadFromStream(FSession.FReceivedCommandDataStream);
        {$IFDEF D6H}
        VarArrayPut(KeyValues,v.AsVariant,i);
        {$ELSE}
        KeyValues[i] := v.AsVariant;
        {$ENDIF}
       end;
     end;
  finally
    v.Free;
  end;
  LocalCursor.IndexID := IndexID;
  FCurrentRecordBuffer := FLocalCursor.AllocateRecordBuffer;
  try
    LocalCursor.CurrentRecordBuffer := FCurrentRecordBuffer;
    b := LocalCursor.Locate(KeyFields,KeyValues,CaseInsensitive,PartialKey);
    if (b) then
     begin
      FSession.FSentCommandHeader.Reply := accrplYes;
      RecordID := LocalCursor.CurrentRecordID;
      SaveDataToStream(RecordID,SizeOf(RecordID),FSession.FSentCommandDataStream,11120);
      SendRecordBuffer(11121,FCurrentRecordBuffer);
     end
    else
     begin
      FSession.FSentCommandHeader.Reply := accrplNo;
     end;
  finally
    FLocalCursor.FreeRecordBuffer(FCurrentRecordBuffer);
  end;
  FSession.SendCommand;
end; // Locate


//------------------------------------------------------------------------------
// Find record by key
//------------------------------------------------------------------------------
procedure TACRServerCursor.FindKey;
var
    RecordID:         TACRRecordID;
    b:                ByteBool;
    IndexID:          TACRObjectID;
    KeyFieldCount:    Integer;
    sc:               Byte;
    SearchCondition:  TACRSearchCondition;
begin
  LoadDataFromStream(IndexID,SizeOf(IndexID),FSession.FReceivedCommandDataStream,11124);
  LoadDataFromStream(sc,SizeOf(sc),FSession.FReceivedCommandDataStream,11125);
  SearchCondition := TACRSearchCondition(sc);
  LoadDataFromStream(KeyFieldCount,SizeOf(KeyFieldCount),FSession.FReceivedCommandDataStream,11126);
  FCurrentRecordBuffer := FLocalCursor.AllocateRecordBuffer;
  try
    FLocalCursor.IndexID := IndexID;
    FLocalCursor.KeyBuffer := FCurrentRecordBuffer;
    FLocalCursor.KeyFieldCount := KeyFieldCount;
    ReceiveRecordBuffer(11127,FCurrentRecordBuffer);
    b := FLocalCursor.FindKey(SearchCondition);
    if (b) then
     begin
      FSession.FSentCommandHeader.Reply := accrplYes;
      RecordID := LocalCursor.CurrentRecordID;
      SaveDataToStream(RecordID,SizeOf(RecordID),FSession.FSentCommandDataStream,11128);
     end
    else
     FSession.FSentCommandHeader.Reply := accrplNo;
  finally
    FLocalCursor.FreeRecordBuffer(FCurrentRecordBuffer);
  end;
  FSession.SendCommand;
end; // FindKey


//------------------------------------------------------------------------------
// Reset range
//------------------------------------------------------------------------------
procedure TACRServerCursor.ResetRange(DoNotSendReply: Boolean);
var StartBuffer,EndBuffer: TACRRecordBuffer;
begin
  StartBuffer := FLocalCursor.RangeStartBuffer;
  EndBuffer := FLocalCursor.RangeEndBuffer;
  if (StartBuffer <> nil) then
    FLocalCursor.FreeRecordBuffer(StartBuffer);
  if (EndBuffer <> nil) then
    FLocalCursor.FreeRecordBuffer(EndBuffer);
  FLocalCursor.ResetRange;
  if (not DoNotSendReply) then
    FSession.SendCommand;
end; // ResetRange


//------------------------------------------------------------------------------
// Apply range
//------------------------------------------------------------------------------
procedure TACRServerCursor.ApplyRange;
var
  StartBuffer, EndBuffer: TACRRecordBuffer;
  StartKeyFieldCount:     Integer;
  EndKeyFieldCount:       Integer;
  StartExclusive:         Boolean;
  EndExclusive:           Boolean;
  b:                      ByteBool;
  IndexID:                TACRObjectID;
begin
  ResetRange; // free old buffers
  LoadDataFromStream(IndexID,SizeOf(IndexID),FSession.FReceivedCommandDataStream,11130);
  LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11131);
  StartExclusive := b;
  LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11132);
  EndExclusive := b;
  LoadDataFromStream(StartKeyFieldCount,SizeOf(StartKeyFieldCount),FSession.FReceivedCommandDataStream,11133);
  LoadDataFromStream(EndKeyFieldCount,SizeOf(EndKeyFieldCount),FSession.FReceivedCommandDataStream,11134);
  LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11135);
  if (b) then
   begin
    StartBuffer := nil;
   end
  else
   begin
    StartBuffer := FLocalCursor.AllocateRecordBuffer;
    ReceiveRecordBuffer(11136,StartBuffer);
   end;
  LoadDataFromStream(b,SizeOf(b),FSession.FReceivedCommandDataStream,11137);
  if (b) then
   begin
    EndBuffer := nil;
   end
  else
   begin
    EndBuffer := FLocalCursor.AllocateRecordBuffer;
    ReceiveRecordBuffer(11138,EndBuffer);
   end;
  FLocalCursor.ApplyRange(StartBuffer,EndBuffer,StartKeyFieldCount,EndKeyFieldCount,StartExclusive,EndExclusive);
  FSession.SendCommand;
end; // ApplyRange


//------------------------------------------------------------------------------
// read blob value
//------------------------------------------------------------------------------
procedure TACRServerCursor.ReadBLOBValue;
var FieldNumber:      Integer;
    LocalBLOBStream:  TACRStream;
    Size:             Int64;
begin
  try
    LoadDataFromStream(FieldNumber,SizeOf(FieldNumber),FSession.FReceivedCommandDataStream,11217);
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: read blob value, fieldNo = '+IntToStr(FieldNumber)+
  ', pageno = '+IntToStr(FLocalCursor.CurrentRecordID.PageNo)+
  ', pageitemno = '+IntToStr(FLocalCursor.CurrentRecordID.PageNo));
{$ENDIF}
    FCurrentRecordBuffer := FLocalCursor.AllocateRecordBuffer;
    try
      FLocalCursor.CurrentRecordBuffer := FCurrentRecordBuffer;
      if (FLocalCursor.GetRecordBuffer(grmCurrent) <> grrOK) then
       raise EACRException.Create(11231,ErrorLLoadingRecord,
              [FLocalCursor.TableName,FLocalcursor.Session.SessionID,
              FLocalCursor.CurrentRecordID.PageNo,FLocalCursor.CurrentRecordID.PageItemNo]);
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: read blob value - record loaded ok');
{$ENDIF}
      LocalBLOBStream := FLocalCursor.InternalCreateBLOBStream(False,FieldNumber,bomRead);
      if (LocalBLOBStream = nil) then
        raise EACRException.Create(11218,ErrorLNilPointer);
      try
        Size := LocalBLOBStream.Size;
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: read blob value - blob stream created, size = '+IntToStr(Size));
{$ENDIF}
        SaveDataToStream(Size,SizeOf(Size),FSession.FSentCommandDataStream,11219);
        if (Size > 0) then
         begin
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: read blob value - saving stream ... ');
{$ENDIF}
           LocalBLOBStream.SaveToStream(FSession.FSentCommandDataStream);
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: read blob value - saving stream ... ok');
{$ENDIF}
         end;
      finally
        FLocalCursor.InternalCloseBLOB(FieldNumber);
      end;
    finally
      FLocalCursor.FreeRecordBuffer(FCurrentRecordBuffer);
    end;
  except
   on e: Exception do
    begin
      FSession.SendErrorMessage(ACR_CS_ErrorReadBLOBValueFailed,
                       Format(ErrorL_CS_ErrorReadBLOBValueFailed,[e.Message]));
      Exit;
    end;
  end;
  FSession.SendCommand;
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Server: read blob value - ok');
{$ENDIF}
end; // ReadBLOBValue


//------------------------------------------------------------------------------
// return true if current record exists
//------------------------------------------------------------------------------
procedure TACRServerCursor.IsRecordExists;
var
    RecordID:         TACRRecordID;
begin
  FSession.FSentCommandHeader.Reply := accrplNo;
  LoadDataFromStream(RecordID,SizeOf(RecordID),FSession.FReceivedCommandDataStream,11284);
  LocalCursor.CurrentRecordID := RecordID;
  if (LocalCursor.IsRecordExists) then
   FSession.FSentCommandHeader.Reply := accrplYes;
  FSession.SendCommand;
end; // IsRecordExists


////////////////////////////////////////////////////////////////////////////////
//
// General functions
//
////////////////////////////////////////////////////////////////////////////////


function InstallingService: Boolean;
begin
 Result := FindCmdLineSwitch('INSTALL',['-','\','/'],True) or
           FindCmdLineSwitch('UNINSTALL',['-','\','/'],True);
end;

function InteractiveService: Boolean;
begin
 Result := (not FindCmdLineSwitch('NOINTERACT',['-','\','/'],True));
end;


initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRServer> initialized');
{$ENDIF}

end.
