//------------------------------------------------------------------------------
//
// Client classes
//
//------------------------------------------------------------------------------

unit ACRClient;

interface

{$I ACRVer.inc}

uses SysUtils, Classes, Db,
{$IFDEF D6H}
     Variants,
{$ELSE}
     ACRD4Routines,
{$ENDIF}

// Accuracer units

{$IFDEF DEBUG_LOG}
     ACRDebug,
{$ENDIF}
     ACRExcept,
     ACRVariant,
     ACRConnection,
     ACRCommunication,
     ACRCompression,
     ACRLocalEngine,
     ACRBase,
     ACRMemory,
     ACRTypes,
     ACRSQLProcessor,
     ACRConst;

type

////////////////////////////////////////////////////////////////////////////////
//
// TACRClientSession
//
////////////////////////////////////////////////////////////////////////////////

  TACROnClientReceiveMesaage = procedure (Buffer: PChar; Size: Integer) of object;
  TACROnDisconnectRemoteDatasets = procedure (ServerShutdown: Boolean) of object;
  TACROnErrorEvent =  procedure (
                       const ErrorCode:    Integer;
                       const NativeError:  Integer;
                       const ErrorMessage: String
                       ) of object;

  TACRClientSession = class (TACRNetworkSession)
   private
    FOnReceiveMessage:            TACROnClientReceiveMesaage;
    FConnecting:                  Boolean;
    FOnDisconnectRemoteDatasets:  TACROnDisconnectRemoteDatasets;
    FOnError:                     TACROnErrorEvent;
    FDisconnected:                Boolean;
   protected
    // get error message and raises an excpetion
    procedure CheckReceivedReply(CorrectReply: TACRCommunicationReply);
    // db connected?
    function GetConnected: Boolean; override;
    // connect / disconnect
    procedure SetConnected(Value: Boolean); override;
   public
    // call OnError event handler
    procedure DoOnError(ErrorCode: Integer; NativeError: Integer = -1; ErrorMessage: String = ''); override;
    // check if database exists
    function GetDatabaseExists: Boolean; override;
    // fills list with table names from this database
    procedure GetTablesList(List: TStrings); override;
    // return true if table exists
    function TableExists(TableName: String): Boolean; override;
    // flush file buffers
    procedure FlushFileBuffers; override;
    // return database format version
    function GetFormatVersion: Double; override;
    // return total number of pages
    function GetTotalPageCount : Integer; override;
    // return number of free pages
    function GetFreePageCount : Integer; override;
    // return true if database is encrypted
    function IsDatabaseEncrypted: Boolean; override;
    // return true if database is encrypted by password or by key
    function IsDatabaseEncryptedByPassword: Boolean; override;
    // return true if CryptoParams are valid
    function IsCryptoParamsValid: Boolean; override;
    // makes Exe database from edb file
    procedure MakeExeDatabase(ExeFileName, ExeDatabaseFileName: string); override;
    // removes database file from executable database file
    procedure RemoveDatabaseFromExe; override;
    // returns true if this file is an Accuracer database
    function IsAccuracerDatabaseFile: Boolean; override;
    //------------------------ Transactions ------------------------------------
   protected
    // retrun true if database has active transaction
    function GetInTransaction: Boolean; override;
   public
    constructor Create;
    procedure DoCloseSessionOnNetworkError; override;
    // start a transaction
    procedure StartTransaction; override;
    // apply changes made by transaction
    procedure Commit(FlushFileBuffers: Boolean = True); override;
    // cancel changes made by transaction
    procedure Rollback; override;
   public
    procedure OnDisconnect; override;
    // send command specified by SentCommandHeader [ optionally SentCommandDataStream ]
    function SendCommand: Boolean; override;
    // send buffer via established connection using connection manager
    procedure SendBuffer(var Buffer: PChar; BufferSize: Integer); override;
    // wait while command will not be received
    function ReceiveCommand: Boolean;
    // send custom message
    procedure SendMessage(Buffer: PChar; BufferSize: Integer); override;
    // receive custom message
    procedure ReceiveMessage(Buffer: PChar; BufferSize: Integer); override;
   public
    property OnReceiveMessage: TACROnClientReceiveMesaage read FOnReceiveMessage write FOnReceiveMessage;
    property OnDisconnectRemoteDatasets: TACROnDisconnectRemoteDatasets read FOnDisconnectRemoteDatasets
              write FOnDisconnectRemoteDatasets;
    property OnError: TACROnErrorEvent read FOnError write FOnError;          
  end; // TACRClientSession


////////////////////////////////////////////////////////////////////////////////
//
// TACRClientCursor
//
////////////////////////////////////////////////////////////////////////////////

  TACRClientCursor = class (TACRCursor)
   private
    FCursorID:  TACRObjectID;
    FBLOBCache: TList;
   protected
    procedure SendRecordBuffer(const ErrorCode: Integer; RecordBuffer: TACRRecordBuffer);
    procedure ReceiveRecordBuffer(const ErrorCode: Integer; RecordBuffer: TACRRecordBuffer);
   public
    procedure ReceiveFieldNoReferences;
    procedure ReceiveServerCursor;
    constructor Create;
    destructor Destroy; override;
    // create table
    procedure CreateTable(
                          FieldDefs:      TACRFieldDefs;
                          IndexDefs:      TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                         ); override;
    procedure DeleteTable(Cascade: Boolean = False); override;
    procedure EmptyTable; override;
    procedure AddForeignKey(ConstraintDef: TACRConstraintDefForeignKey); override;
    procedure DeleteConstraint(Name: String; Cascade: Boolean; FKPartialDelete: Boolean); override;
    procedure RenameTable(NewTableName: String); override;
    procedure RenameField(FieldName, NewFieldName: String); override;
    function RepairTable(var Log: String; NewSession: Pointer = nil): Boolean; override;
    procedure LoadTableFromStream(Stream: TStream); override;
    procedure SaveTableToStream(
                        Stream:               TStream;
                        CompressionAlgorithm: TACRCompressionAlgorithm = acaNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      ); override;
    function ExportTableToSQL(
                              ExportStructure:      Boolean;
                              AddDropTableCommand:  Boolean;
                              ExportIndexes:        Boolean;
                              AddDropIndexCommand:  Boolean;
                              ExportData:           Boolean;
                              ExportBLOBFields:     Boolean;
                              UseBracketsForNames:  Boolean
                            ): String; override;
    procedure InternalInitFieldDefs; override;
    procedure OpenTableByFieldDefs(
                          FieldDefs:      TACRFieldDefs;
                          IndexDefs:      TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                       ); override;
    procedure CloseTable; override;
    // initialize record buffer
    procedure InternalInitRecord(RecordBuffer: TACRRecordBuffer;
                                 InsertMode: Boolean); override;
    procedure AddIndex(IndexDef: TACRIndexDef); override;
    procedure DeleteIndex(Name: String); override;
    procedure DeleteAllIndexes; override;
    // return index name of the index or '' if not found
    function FindIndex(FieldNamesList, AscDescList,
              CaseSensitivityList: TStringList): String; override;
    // return index name
    function FindOrCreateIndex(FieldNamesList,
                AscDescList, CaseSensitivityList: TStringList): String; override;

    //---------------------------------------------------------------------------
    // navigation & bookmark methods
    //---------------------------------------------------------------------------

    function GetRecordCount: TACRRecordNo; override;
    // get record
    function GetRecordBuffer(
              GetRecordMode:  TACRGetRecordMode
              ): TACRGetRecordResult; override;
    // go to record
    procedure SetRecNo(Value: TACRRecordNo); override;
    function GetRecNo: TACRRecordNo; override;

    //---------------------------------------------------------------------------
    // insert, edit, post, delete methods
    //---------------------------------------------------------------------------

    // insert record
    procedure InternalInsert; override;
    // edit record
    procedure InternalEdit; override;
    // cancels updates
    procedure InternalCancel(ToInsert: Boolean); override;
    // insert or update record
    procedure InternalPost(ToInsert: Boolean); override;
    // delete record
    procedure InternalDelete; override;
    procedure DeleteVisibleRecords; override;
    procedure UpdateVisibleRecords(FieldNames:   TStringList;
                                   values:       array of TACRVariant;
                                   SkipFKCheck:  Boolean = False
                                   ); override;

    //---------------------------------------------------------------------------
    // search & filter methods
    //---------------------------------------------------------------------------

    // disable record bitmap
    procedure DisableRecordBitmap; override;
    // set distinct
    procedure ApplyDistinct(FieldNamesList, AscDescList, CaseSensitivityList: TStringList); override;
    // apply projection
    procedure ApplyProjection(FieldNamesList, AliasList: TStringList); override;
    procedure ActivateFilters(
                              FilterText:      String;
                              CaseInsensitive: Boolean;
                              PartialKey:      Boolean
                            ); override;
    procedure DeactivateFilters; override;
    function Locate(
                    const KeyFields: String;
                    const KeyValues: Variant;
                    CaseInsensitive: Boolean;
                    PartialKey:      Boolean
                   ): Boolean; override;
    function FindKey(SearchCondition: TACRSearchCondition): Boolean; override;
    procedure ResetRange; override;
    procedure ApplyRange(
                          StartBuffer, EndBuffer: TACRRecordBuffer;
                          StartKeyFieldCount:     Integer;
                          EndKeyFieldCount:       Integer;
                          StartExclusive:         Boolean;
                          EndExclusive:           Boolean
                        ); override;
    // set SQL Filter
    procedure SetSQLFilter(FilterExpr: TObject); override;

    //---------------------------------------------------------------------------
    // BLOB methods
    //---------------------------------------------------------------------------

    function InternalCreateBlobStream(
              ToInsert: Boolean;
              FieldNo:  Integer;
              OpenMode: TACRBLOBOpenMode
              ):TACRStream; override;
    procedure InternalCloseBLOB(FieldNo: Integer); override;
    // clear blob streams
    procedure ClearBLOBStreams(WriteOnly: Boolean = False); override;
    procedure SendModifiedBLOBValues;

    function LastAutoincValue(FieldNo: Integer): Int64; override;
    procedure SetLastAutoincValue(Value: Int64; FieldNo: Integer); override;

    procedure LockTableForRead; override;
    procedure UnlockTableForRead; override;

    procedure LockTableS(ErrorCode: Integer); override;
    procedure UnlockTableS(ErrorCode: Integer); override;
    procedure LockTableIRW(ErrorCode: Integer); override;
    procedure UnlockTableIRW(ErrorCode: Integer); override;
    procedure LockTableRW(ErrorCode: Integer); override;
    procedure UnlockTableRW(ErrorCode: Integer); override;

    function TryToLockTableS: Boolean; override;
    function TryToUnlockTableS: Boolean; override;
    function TryToLockTableIRW: Boolean; override;
    function TryToUnlockTableIRW: Boolean; override;
    function TryToLockTableRW: Boolean; override;
    function TryToUnlockTableRW: Boolean; override;

    // return true if current record exists
    function IsRecordExists: Boolean; override;

  end; // TACRClientCursor


////////////////////////////////////////////////////////////////////////////////
//
// TACRClientSQLProcessor
//
////////////////////////////////////////////////////////////////////////////////


  TACRClientSQLProcessor = class (TACRSQLProcessor)
   private
    FSession: TACRClientSession;
   protected
    procedure SendParams;
    function ReceiveLiveQuery: TACRCursor;
    function ReceiveNotLiveQuery: TACRCursor;
   public
    constructor Create(Query: TDataSet; aSession: TACRBaseSession); overload;
    destructor Destroy; override;
    function OpenQuery: TACRCursor; override;
   public
    property Session: TACRClientSession read FSession;
  end;

var ClientConnectionManager: TACRClientConnectionManager;

implementation

 uses ACRMain;

////////////////////////////////////////////////////////////////////////////////
//
// TACRClientSession
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// get error message and raises an excpetion
//------------------------------------------------------------------------------
procedure TACRClientSession.CheckReceivedReply(CorrectReply: TACRCommunicationReply);
var Len:          Integer;
    ErrorMessage: String;
    ErrorCode:    Integer;
begin
  if (FReceivedCommandHeader.Reply <> CorrectReply) then
   begin
    ErrorMessage := ErrorL_CS_UnknownError;
    ErrorCode := ACR_CS_UnknownError;
    if (FReceivedCommandDataStream.Size > (SizeOf(Len) + SizeOf(ErrorCode))) then
     begin
      LoadDataFromStream(ErrorCode,SizeOf(ErrorCode),FReceivedCommandDataStream,10860);
      LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10860);
      SetLength(ErrorMessage,Len);
      LoadDataFromStream(PChar(@ErrorMessage[1])^,Len,FReceivedCommandDataStream,10861);
      FReceivedCommandDataStream.Size := 0;
     end;
    raise EACRException.Create(ErrorCode,ErrorLRemoteCommandFailed,
          [Integer(FReceivedCommandHeader.Reply),ErrorMessage]);
   end;
end; // CheckReceivedReply


//------------------------------------------------------------------------------
// db connected?
//------------------------------------------------------------------------------
function TACRClientSession.GetConnected: Boolean;
begin
  Result := FConnected;
end; // GetConnected


//------------------------------------------------------------------------------
// connect / disconnect
//------------------------------------------------------------------------------
procedure TACRClientSession.SetConnected(Value: Boolean);
var Len: Integer;
    b:   ByteBool;
begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client start SetConnected SessionID = '+IntToStr(FSessionID)+
 ', Value = '+BoolToStr(Value,True)+
  ', FConnected = '+BoolToStr(FConnected,True));
 {$ENDIF}
  if (FConnected <> Value) then
    begin
      if (Value) then
       begin
        FConnecting := True;
        try
          try
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client Connect SessionID = '+IntToStr(FSessionID)+
 'starting sonnect ....');
 {$ENDIF}
            ClientConnectionManager.Connect(Self);
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client Connect SessionID = '+IntToStr(FSessionID)+
 'starting sonnect .... OK');
 {$ENDIF}
          except
            on e: EACRException do
              DoOnError(ACR_CS_ErrorConnectingClientSession,e.NativeError,
                         Format(ErrorL_CS_ErrorConnectingClientSession,
                               [FConnectParams.RemoteHost,FConnectParams.RemotePort,FConnectParams.ServerID,e.Message]));
            on e: Exception do
              DoOnError(ACR_CS_ErrorConnectingClientSession,-1,
                         Format(ErrorL_CS_ErrorConnectingClientSession,
                               [FConnectParams.RemoteHost,FConnectParams.RemotePort,FConnectParams.ServerID,e.Message]))
            else
              DoOnError(ACR_CS_ErrorConnectingClientSession,-1,
                         Format(ErrorL_CS_ErrorConnectingClientSession,
                               [FConnectParams.RemoteHost,FConnectParams.RemotePort,FConnectParams.ServerID,'']));
            raise;
          end;
          FConnected := Value;
          FDisconnected := False;
          try
            // connecting to a remote database
            FSentCommandHeader.Request := accrqConnectDatabase;
            FSentCommandHeader.Reply := accrplNULL;
            FSentCommandDataStream.Size := 0;
            Len := Length(DatabaseName);
            SaveDataToStream(Len,SizeOf(Len),SentCommandDataStream,10856);
            SaveDataToStream(PChar(@DatabaseName[1])^,Len,SentCommandDataStream,10857);
            SaveCryptoParamsToStream(FCryptoParams,SentCommandDataStream,11315);
   {$IFDEF DEBUG_LOG_COMMUNICATION}
   aaWriteToLog('Client Connect SessionID = '+IntToStr(FSessionID)+
   'starting send command  ....');
   {$ENDIF}

            if (not SendCommand) then
             raise EACRException.Create(11512,ErrorLCannotConnectToServerSendCommandFailed);
   {$IFDEF DEBUG_LOG_COMMUNICATION}
   aaWriteToLog('Client Connect SessionID = '+IntToStr(FSessionID)+
   'starting receive command  ....');
   {$ENDIF}
            if (not ReceiveCommand) then
             raise EACRException.Create(11513,ErrorLCannotConnectToServerReceiveCommandFailed);
   {$IFDEF DEBUG_LOG_COMMUNICATION}
   aaWriteToLog('Client Connect SessionID = '+IntToStr(FSessionID)+
   'starting checking received command, reply = '+IntToStr(Integer(FReceivedCommandHeader.Reply)));
   {$ENDIF}
            CheckReceivedReply(accrplOperationSucceed);
   {$IFDEF DEBUG_LOG_COMMUNICATION}
   aaWriteToLog('Client Connect SessionID = '+IntToStr(FSessionID)+
   'starting checking received command ok, reply = '+IntToStr(Integer(FReceivedCommandHeader.Reply)));
   {$ENDIF}
            LoadDataFromStream(FOptions,SizeOf(FOptions),ReceivedCommandDataStream,10871);
            LoadDataFromStream(b,SizeOf(b),ReceivedCommandDataStream,11066);
            FExclusive := b;
            ReceivedCommandDataStream.Size := 0;
   {$IFDEF DEBUG_LOG_COMMUNICATION}
   aaWriteToLog('Client Connect SessionID = '+IntToStr(FSessionID)+
   'connection started ok!');
   {$ENDIF}
          except
            FConnected := False;
            FDisconnected := True;
            FConnecting := False;
            raise;
          end;
        finally
          FConnecting := False;
        end;
       end
      else
       begin
        // disconnecting a remote database
        FDisconnected := True;
        FConnected := Value;
        FSentCommandDataStream.Size := 0;
        FReceivedCommandDataStream.Size := 0;
        try
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client  SessionID = '+IntToStr(FSessionID)+
 'starting dicsonnect ....');
 {$ENDIF}
          ClientConnectionManager.Disconnect(Self);
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client starting disconnect SessionID = '+IntToStr(FSessionID)+
 'starting dicsonnect .... OK');
 {$ENDIF}
        except
          on e: EACRException do
            DoOnError(ACR_CS_ErrorDisconnectingClientSession,e.NativeError,
                       Format(ErrorL_CS_ErrorDisconnectingClientSession,
                             [FSessionID,e.Message]));
          on e: Exception do
            DoOnError(ACR_CS_ErrorDisconnectingClientSession,-1,
                       Format(ErrorL_CS_ErrorDisconnectingClientSession,
                             [FSessionID,e.Message]))
          else
            DoOnError(ACR_CS_ErrorDisconnectingClientSession,-1,
                       Format(ErrorL_CS_ErrorDisconnectingClientSession,
                             [FSessionID,'']));
        end;
       end;
    end;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client finish SetConnected SessionID = '+IntToStr(FSessionID)+
 ', Value = '+BoolToStr(Value,True)+
  ', FConnected = '+BoolToStr(FConnected,True));
 {$ENDIF}
end; // SetConnected


//------------------------------------------------------------------------------
// call OnError event handler
//------------------------------------------------------------------------------
procedure TACRClientSession.DoOnError(ErrorCode: Integer; NativeError: Integer = -1; ErrorMessage: String = '');
begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog('Error on client session!');
aaWriteToLog('------------------------------------------------------------------');
aaWriteToLog('SessionID='+IntToStr(Integer(self.SessionID)));
aaWriteToLog('ErrorCode='+IntToStr(Integer(ErrorCode)));
aaWriteToLog('NativeError='+IntToStr(Integer(NativeError)));
aaWriteToLog('ErrorMessage: "'+ErrorMessage+'"');
aaWriteToLog('GetTickCount = '+IntToStr(aaGetTickCount));
aaWriteToLog('==================================================================');
{$ENDIF}
  if (Assigned(FOnError)) then
   FOnError(ErrorCode,NativeError,ErrorMessage);
end; // DoOnError


//------------------------------------------------------------------------------
// check if database exists
//------------------------------------------------------------------------------
function TACRClientSession.GetDatabaseExists: Boolean;
var bCon: Boolean;
    Len:  Integer;
begin
  Result := False;
  bCon := GetConnected;
  try
    if (not bCon) then
     ClientConnectionManager.Connect(Self,False,False);
    try
      FDisconnected := False;
      // send IsDatabaseExists request
      FSentCommandHeader.Request := accrqIsDatabaseExists;
      FSentCommandHeader.Reply := accrplNULL;
      Len := Length(DatabaseName);
      SaveDataToStream(Len,SizeOf(Len),SentCommandDataStream,10873);
      SaveDataToStream(PChar(@DatabaseName[1])^,Len,SentCommandDataStream,10874);
      if (not SendCommand) then
       raise EACRException.Create(11571,ErrorLCannotSendCommandToServer);
      if (not ReceiveCommand) then
       raise EACRException.Create(11572,ErrorLCannotReceiveCommandFromServer);
      Result := (FReceivedCommandHeader.Reply = accrplYes);
    finally
     if (not bCon) then
      ClientConnectionManager.Disconnect(Self);
     FDisconnected := not bCon;
    end;
  except
  end;
end; // GetDatabaseExists


//------------------------------------------------------------------------------
// get tables list
//------------------------------------------------------------------------------
procedure TACRClientSession.GetTablesList(List: TStrings);
var i,Len,Count: Integer;
    s:           String;
begin
  // send GetTablesList request
  List.Clear;
  SentCommandDataStream.Size := 0;
  FSentCommandHeader.Request := accrqGetTablesList;
  FSentCommandHeader.Reply := accrplNULL;
  if (SendCommand) then
   if (ReceiveCommand) then
    begin
      CheckReceivedReply(accrplOperationSucceed);
      // loading tables list
      LoadDataFromStream(Count,SizeOf(Count),FReceivedCommandDataStream,10878);
      for i := 0 to Count - 1 do
       begin
        LoadDataFromStream(Len,SizeOf(Len),FReceivedCommandDataStream,10879);
        SetLength(s,Len);
        LoadDataFromStream(PChar(@s[1])^,Len,FReceivedCommandDataStream,10880);
        List.Add(s);
       end;
      FReceivedCommandDataStream.Size := 0;
    end;
end; // GetTablesList


//------------------------------------------------------------------------------
// return true if table exists
//------------------------------------------------------------------------------
function TACRClientSession.TableExists(TableName: String): Boolean;
var Len: Integer;
begin
  Result := False;
  SentCommandDataStream.Size := 0;
  FSentCommandHeader.Request := accrqIsTableExists;
  FSentCommandHeader.Reply := accrplNULL;
  Len := Length(TableName);
  SaveDataToStream(Len,SizeOf(Len),SentCommandDataStream,10884);
  SaveDataToStream(PChar(@TableName[1])^,Len,SentCommandDataStream,10885);
  if (SendCommand) then
   if (ReceiveCommand) then
    Result := (FReceivedCommandHeader.Reply = accrplYes);
end; // TableExists


//------------------------------------------------------------------------------
// flush file buffers
//------------------------------------------------------------------------------
procedure TACRClientSession.FlushFileBuffers;
begin
{ TODO : implement it }
{
  SentCommandDataStream.Size := 0;
  FSentCommandHeader.Request := accrqFlushFileBuffers;
  FSentCommandHeader.Reply := accrplNULL;
  if (SendCommand) then
   if (ReceiveCommand) then
    CheckReceivedReply(accrplOperationSucceed);
}
end; // FlushFileBuffers


//------------------------------------------------------------------------------
// return database format version
//------------------------------------------------------------------------------
function TACRClientSession.GetFormatVersion: Double;
begin
  Result := ACRVersion;
  if (GetConnected) then
   begin
    SentCommandDataStream.Size := 0;
    FSentCommandHeader.Request := accrqGetFormatVersion;
    FSentCommandHeader.Reply := accrplNULL;
    if (SendCommand) then
     if (ReceiveCommand) then
      begin
       if (FReceivedCommandHeader.Reply = accrplOperationSucceed) then
        LoadDataFromStream(Result,SizeOf(Result),FReceivedCommandDataStream,10891);
       FReceivedCommandDataStream.Size := 0;
      end;
   end;
end; // GetFormatVersion


//------------------------------------------------------------------------------
// return total number of pages
//------------------------------------------------------------------------------
function TACRClientSession.GetTotalPageCount: Integer;
begin
  Result := 0;
  if (GetConnected) then
   begin
    SentCommandDataStream.Size := 0;
    FSentCommandHeader.Request := accrqGetTotalPageCount;
    FSentCommandHeader.Reply := accrplNULL;
    if (SendCommand) then
     if (ReceiveCommand) then
      begin
        if (FReceivedCommandHeader.Reply = accrplOperationSucceed) then
         LoadDataFromStream(Result,SizeOf(Result),FReceivedCommandDataStream,10892);
        FReceivedCommandDataStream.Size := 0;
      end;
   end;
end; // GetTotalPageCount


//------------------------------------------------------------------------------
// return number of free pages
//------------------------------------------------------------------------------
function TACRClientSession.GetFreePageCount: Integer;
begin
  Result := 0;
  if (GetConnected) then
   begin
    SentCommandDataStream.Size := 0;
    FSentCommandHeader.Request := accrqGetFreePageCount;
    FSentCommandHeader.Reply := accrplNULL;
    if (SendCommand) then
     if (ReceiveCommand) then
      begin
        if (FReceivedCommandHeader.Reply = accrplOperationSucceed) then
         LoadDataFromStream(Result,SizeOf(Result),FReceivedCommandDataStream,10893);
        FReceivedCommandDataStream.Size := 0;
      end;
   end;
end; // GetFreePageCount


//------------------------------------------------------------------------------
// return true if database is encrypted
//------------------------------------------------------------------------------
function TACRClientSession.IsDatabaseEncrypted: Boolean;
var bCon: Boolean;
    Len:  Integer;
begin
  Result := False;
  bCon := GetConnected;
  if (not bCon) then
     ClientConnectionManager.Connect(Self,False,False);
  try
    FDisconnected := False;
    // send request
    FSentCommandHeader.Request := accrqIsDatabaseEncrypted;
    FSentCommandHeader.Reply := accrplNULL;
    SentCommandDataStream.Size := 0;
    Len := Length(DatabaseName);
    SaveDataToStream(Len,SizeOf(Len),SentCommandDataStream,10894);
    SaveDataToStream(PChar(@DatabaseName[1])^,Len,SentCommandDataStream,10895);
    if (not SendCommand) then
     raise EACRException.Create(11573,ErrorLCannotSendCommandToServer);
    if (not ReceiveCommand) then
     raise EACRException.Create(11574,ErrorLCannotReceiveCommandFromServer);
    Result := (FReceivedCommandHeader.Reply = accrplYes);
  finally
   if (not bCon) then
    ClientConnectionManager.Disconnect(Self);
   FDisconnected := not bCon;
  end;
end; // IsDatabaseEncrypted


//------------------------------------------------------------------------------
// return true if database is encrypted by password or by key
//------------------------------------------------------------------------------
function TACRClientSession.IsDatabaseEncryptedByPassword: Boolean;
var bCon: Boolean;
    Len:  Integer;
begin
  Result := False;
  bCon := GetConnected;
  if (not bCon) then
    ClientConnectionManager.Connect(Self,False,False);
  try
    FDisconnected := False;
    // send request
    FSentCommandHeader.Request := accrqIsDatabaseEncryptedByPassword;
    FSentCommandHeader.Reply := accrplNULL;
    SentCommandDataStream.Size := 0;
    Len := Length(DatabaseName);
    SaveDataToStream(Len,SizeOf(Len),SentCommandDataStream,10896);
    SaveDataToStream(PChar(@DatabaseName[1])^,Len,SentCommandDataStream,10897);
    if (not SendCommand) then
     raise EACRException.Create(11575,ErrorLCannotSendCommandToServer);
    if (not ReceiveCommand) then
     raise EACRException.Create(11576,ErrorLCannotReceiveCommandFromServer);
    Result := (FReceivedCommandHeader.Reply = accrplYes);
  finally
   if (not bCon) then
    ClientConnectionManager.Disconnect(Self);
   FDisconnected := not bCon;
  end;
end; // IsDatabaseEncryptedByPassword


//------------------------------------------------------------------------------
// return true if CryptoParams are valid
//------------------------------------------------------------------------------
function TACRClientSession.IsCryptoParamsValid: Boolean;
var bCon: Boolean;
    Len:  Integer;
begin
  Result := False;
  bCon := GetConnected;
  if (not bCon) then
     ClientConnectionManager.Connect(Self,False,False);
  try
    FDisconnected := False;
    // send request
    FSentCommandHeader.Request := accrqIsCryptoParamsValid;
    FSentCommandHeader.Reply := accrplNULL;
    SentCommandDataStream.Size := 0;
    Len := Length(DatabaseName);
    SaveDataToStream(Len,SizeOf(Len),SentCommandDataStream,10906);
    SaveDataToStream(PChar(@DatabaseName[1])^,Len,SentCommandDataStream,10907);
    SaveCryptoParamsToStream(FCryptoParams,SentCommandDataStream,11314);
    if (not SendCommand) then
     raise EACRException.Create(11577,ErrorLCannotSendCommandToServer);
    if (not ReceiveCommand) then
     raise EACRException.Create(11578,ErrorLCannotReceiveCommandFromServer);
    Result := (FReceivedCommandHeader.Reply = accrplYes);
  finally
   if (not bCon) then
    ClientConnectionManager.Disconnect(Self);
   FDisconnected := not bCon;
  end;
end; // IsCryptoParamsValid


//------------------------------------------------------------------------------
// makes Exe database from edb file
//------------------------------------------------------------------------------
procedure TACRClientSession.MakeExeDatabase(ExeFileName, ExeDatabaseFileName: string);
begin
  raise EACRException.Create(11288,ErrorLOperationIsNotSupported);
end; // MakeExeDatabase


//------------------------------------------------------------------------------
// removes database file from executable database file
//------------------------------------------------------------------------------
procedure TACRClientSession.RemoveDatabaseFromExe;
begin
  raise EACRException.Create(11298,ErrorLOperationIsNotSupported);
end; // RemoveDatabaseFromExe


//------------------------------------------------------------------------------
// returns true if this file is an Accuracer database
//------------------------------------------------------------------------------
function TACRClientSession.IsAccuracerDatabaseFile: Boolean;
begin
  Result := False;
end; // IsAccuracerDatabaseFile


//------------------------------------------------------------------------------
// retrun true if database has active transaction
//------------------------------------------------------------------------------
function TACRClientSession.GetInTransaction: Boolean;
begin
  Result := False;
  // send request
  FSentCommandHeader.Request := accrqIsInTransaction;
  FSentCommandHeader.Reply := accrplNULL;
  SentCommandDataStream.Size := 0;
  if (SendCommand) then
   if (ReceiveCommand) then
     Result := (FReceivedCommandHeader.Reply = accrplYes);
end; // GetInTransaction


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientSession.Create;
begin
  inherited;
  FOnReceiveMessage := nil;
  FConnecting := False;
  FDisconnected := True;
end; // Create


//------------------------------------------------------------------------------
// do close session on network error
//------------------------------------------------------------------------------
procedure TACRClientSession.DoCloseSessionOnNetworkError;
begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client DoCloseSessionOnNetworkError starting, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
  if (FConnected) then
   begin
    {$IFDEF DEBUG_LOG_COMMUNICATION}
    aaWriteToLog('Client DoCloseSessionOnNetworkError starting disconnect session, SessionID = '+IntToStr(FSessionID));
    {$ENDIF}
    SetConnected(False);
    {$IFDEF DEBUG_LOG_COMMUNICATION}
    aaWriteToLog('Client DoCloseSessionOnNetworkError starting disconnect datasets, SessionID = '+IntToStr(FSessionID));
    {$ENDIF}
    if (Assigned(FOnDisconnectRemoteDatasets)) then
      FOnDisconnectRemoteDatasets(False);
    {$IFDEF DEBUG_LOG_COMMUNICATION}
    aaWriteToLog('Client DoCloseSessionOnNetworkError datasets disconnected, SessionID = '+IntToStr(FSessionID));
    {$ENDIF}
   end;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client DoCloseSessionOnNetworkError finished, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
end; // DoCloseSessionOnNetworkError


//------------------------------------------------------------------------------
// start a transaction
//------------------------------------------------------------------------------
procedure TACRClientSession.StartTransaction;
begin
  // send request
  FSentCommandHeader.Request := accrqStartTransaction;
  FSentCommandHeader.Reply := accrplNULL;
  SentCommandDataStream.Size := 0;
  if (SendCommand) then
   if (ReceiveCommand) then
    CheckReceivedReply(accrplOperationSucceed);
end; // StartTransaction


//------------------------------------------------------------------------------
// apply changes made by transaction
//------------------------------------------------------------------------------
procedure TACRClientSession.Commit(FlushFileBuffers: Boolean = True);
var b: ByteBool;
begin
  // send request
  FSentCommandHeader.Request := accrqCommit;
  FSentCommandHeader.Reply := accrplNULL;
  SentCommandDataStream.Size := 0;
  b := FlushFileBuffers;
  SaveDataToStream(b,SizeOf(b),FSentCommandDataStream,10913);
  if (SendCommand) then
   if (ReceiveCommand) then
    CheckReceivedReply(accrplOperationSucceed);
end; // Commit


//------------------------------------------------------------------------------
// cancel changes made by transaction
//------------------------------------------------------------------------------
procedure TACRClientSession.Rollback;
begin
  // send request
  FSentCommandHeader.Request := accrqRollback;
  FSentCommandHeader.Reply := accrplNULL;
  SentCommandDataStream.Size := 0;
  if (SendCommand) then
   if (ReceiveCommand) then
    CheckReceivedReply(accrplOperationSucceed);
end; // Rollback


//------------------------------------------------------------------------------
// disconnect occurs when server is shutting down or network socket error occurs
//------------------------------------------------------------------------------
procedure TACRClientSession.OnDisconnect;
begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client OnDisconnect starting, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
  if (FConnecting) then
   raise EACRException.Create(11335,ErrorLCannotConnetToRemoteServer,[FConnectParams.RemoteHost,FConnectParams.RemotePort,FConnectParams.LocalPort,FConnectParams.ServerID]);
  if (FConnected) then
   begin
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client OnDisconnect starting closing session, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
    FDisconnected := True;
    FConnected := False;
    FSentCommandDataStream.Size := 0;
    FReceivedCommandDataStream.Size := 0;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client OnDisconnect starting disonnect datasets, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
    if (Assigned(FOnDisconnectRemoteDatasets)) then
      FOnDisconnectRemoteDatasets(True);
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client OnDisconnect datasets disonnected, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
   end;
 {$IFDEF DEBUG_LOG_COMMUNICATION}
 aaWriteToLog('Client OnDisconnect ok, SessionID = '+IntToStr(FSessionID));
 {$ENDIF}
end; // OnDisconnect


//------------------------------------------------------------------------------
// send command specified by SentCommandHeader [ optionally SentCommandDataStream ]
//------------------------------------------------------------------------------
function TACRClientSession.SendCommand: Boolean;
begin
  Result := False;
  if (not FDisconnected) then
   Result := inherited SendCommand;
end; // SendCommand


//------------------------------------------------------------------------------
// send buffer via established connection using connection manager
//------------------------------------------------------------------------------
procedure TACRClientSession.SendBuffer(var Buffer: PChar; BufferSize: Integer);
begin
{$IFDEF DEBUG_LOG_COMMUNICATION}
aaWriteToLog('C> ClientSession is starting to send a request...');
aaWriteToLog('C> SessionID = '+IntToStr(SessionID)+', ServerID = '+IntToStr(FConnectParams.ServerID));
aaWriteToLog('C> Request = '+IntToStr(Integer(FSentCommandHeader.Request)));
aaWriteBufferToLog(Buffer,BufferSize);
aaWriteToLog('C> Send Start Time = '+aaGetCurrentTimeAsString);
{$ENDIF}
ClientConnectionManager.SendBuffer(Self,Buffer,BufferSize);
{$IFDEF DEBUG_LOG_COMMUNICATION}
aaWriteToLog('C> Send End Time = '+aaGetCurrentTimeAsString);
{$ENDIF}
end; // SendBuffer


//------------------------------------------------------------------------------
// wait while command will not be received
//------------------------------------------------------------------------------
function TACRClientSession.ReceiveCommand: Boolean;
var Buffer:     PChar;
    BufferSize: Integer;
begin
  Result := False;
  if (FDisconnected) then
   Exit;
  FCommandReceived := False;
{$IFDEF DEBUG_LOG_COMMUNICATION}
aaWriteToLog('C> ClientSession is waiting for reply...');
aaWriteToLog('C> SessionID = '+IntToStr(SessionID)+', ServerID = '+IntToStr(FConnectParams.ServerID));
aaWriteToLog('C> Receive Start Time = '+aaGetCurrentTimeAsString);
{$ENDIF}
  try
    ClientConnectionManager.ReceiveBuffer(Self,Buffer,BufferSize);
    try
      ReceiveData(Buffer,BufferSize);
{$IFDEF DEBUG_LOG_COMMUNICATION}
aaWriteBufferToLog(Buffer,BufferSize);
aaWriteToLog('C> Request = '+IntToStr(Integer(FReceivedCommandHeader.Request)));
aaWriteToLog('C> Reply = '+IntToStr(Integer(FReceivedCommandHeader.Reply)));
aaWriteToLog('C> Receive End Time = '+aaGetCurrentTimeAsString);
{$ENDIF}
    finally
     if (Buffer <> nil) and (BufferSize > 0) then
      MemoryManager.FreeAndNilMem(Buffer);
    end;
    Result := True;
  except
    on e: EACRException do
     begin
      DoOnError(ACR_CS_ErrorReceiveCommandFailed,e.NativeError,
                 Format(ErrorL_CS_ErrorReceiveCommandFailed,
                       [FSessionID,e.Message]));
      DoCloseSessionOnNetworkError;
     end;
    on e: Exception do
     begin
      DoOnError(ACR_CS_ErrorReceiveCommandFailed,-1,
                 Format(ErrorL_CS_ErrorReceiveCommandFailed,
                       [FSessionID,e.Message]));
      DoCloseSessionOnNetworkError;
     end
    else
     begin
      DoOnError(ACR_CS_ErrorReceiveCommandFailed,-1,
                 Format(ErrorL_CS_ErrorReceiveCommandFailed,
                       [FSessionID,'']));
      DoCloseSessionOnNetworkError;
     end;
  end;

//aaWriteToLog('1 TACRClientSession.ReceiveCommand, BufferSize = '+IntToStr(BufferSize));
//  repeat
//  until (FCommandReceived);
end; // ReceiveCommand


//------------------------------------------------------------------------------
// send custom message
//------------------------------------------------------------------------------
procedure TACRClientSession.SendMessage(Buffer: PChar; BufferSize: Integer);
begin
  ClientConnectionManager.SendMessage(Self,Buffer,BufferSize);
end; // SendMessage


//------------------------------------------------------------------------------
// receive custom message
//------------------------------------------------------------------------------
procedure TACRClientSession.ReceiveMessage(Buffer: PChar; BufferSize: Integer);
begin
  if (Assigned(FOnReceiveMessage)) then
   FOnReceiveMessage(Buffer,BufferSize);
end; // ReceiveMessage


////////////////////////////////////////////////////////////////////////////////
//
// TACRClientCursor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// send record buffer
//------------------------------------------------------------------------------
procedure TACRClientCursor.SendRecordBuffer(const ErrorCode: Integer; RecordBuffer: TACRRecordBuffer);
begin
  SaveDataToStream(RecordBuffer^,RecordBufferSize,TACRClientSession(FSession).FSentCommandDataStream,ErrorCode);
end; // SendRecordBuffer


//------------------------------------------------------------------------------
// receive record buffer
//------------------------------------------------------------------------------
procedure TACRClientCursor.ReceiveRecordBuffer(const ErrorCode: Integer; RecordBuffer: TACRRecordBuffer);
begin
  LoadDataFromStream(RecordBuffer^,RecordBufferSize,TACRClientSession(FSession).FReceivedCommandDataStream,ErrorCode);
end; // ReceiveRecordBuffer


//------------------------------------------------------------------------------
// Receive VisisbleFieldDefs.FieldNoReference
//------------------------------------------------------------------------------
procedure TACRClientCursor.ReceiveFieldNoReferences;
var i, num: Integer;
begin
  for i := 0 to FVisibleFieldDefs.Count - 1 do
   begin
    LoadDataFromStream(num,SizeOf(num),TACRClientSession(FSession).FReceivedCommandDataStream,11186);
    FVisibleFieldDefs.Items[i].FieldNoReference := num;
   end;
end; // ReceiveFieldNoReferences


//------------------------------------------------------------------------------
// Receive cursor information from server
//------------------------------------------------------------------------------
procedure TACRClientCursor.ReceiveServerCursor;
begin
  FCursorID := INVALID_OBJECT_ID;
  FirstPosition := True;
  LastPosition := False;
  // loading ursorID, field defs, visible field defs, index defs, constraint defs
  LoadDataFromStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FReceivedCommandDataStream,10963);
  if (FFieldDefs <> nil) then
   FFieldDefs.Free;
  FFieldDefs := TACRFieldDefs.Create;
  if (FVisibleFieldDefs <> nil) then
   FVisibleFieldDefs.Free;
  FVisibleFieldDefs := TACRFieldDefs.Create;
  if (FIndexDefs <> nil) then
   FIndexDefs.Free;
  FIndexDefs := TACRIndexDefs.Create;
  if (FConstraintDefs <> nil) then
   FConstraintDefs.Free;
  FConstraintDefs := TACRConstraintDefs.Create;
  FFieldDefs.LoadFromStream(TACRClientSession(FSession).FReceivedCommandDataStream);
  FVisibleFieldDefs.LoadFromStream(TACRClientSession(FSession).FReceivedCommandDataStream);
  FIndexDefs.LoadFromStream(TACRClientSession(FSession).FReceivedCommandDataStream);
  FConstraintDefs.LoadFromStream(TACRClientSession(FSession).FReceivedCommandDataStream);
  // create blob lists
  if (FBLOBStreams <> nil) then
    FBLOBStreams.Free;
  FBLOBStreams := TList.Create;
  if (FBLOBCache <> nil) then
    FBLOBCache.Free;
  FBLOBCache := TList.Create;
  FIsOpen := True;
end; // ReceiveServerCursor


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientCursor.Create;
begin
  CurrentRecordBuffer := nil;
  FRandomOrder := False;
  FFieldDefs := nil;
  FVisibleFieldDefs := nil;
  FIndexDefs := nil;
  IsOpen := False;
  FBLOBStreams := nil;
  FBLOBCache := nil;
  FIndexName := '';
  FIndexID := INVALID_OBJECT_ID;
  FilterExpression := nil;
  SQLFilterExpression := nil;
  FilterRecord := nil;
  FSettingProjection := False;
  DistinctFieldCount := 0;
  RecordBitmap := nil;
  FCursorID := INVALID_OBJECT_ID;
  FCreateTableStarted := False;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRClientCursor.Destroy;
begin
  CloseTable;
  inherited;
end; // Destroy


//------------------------------------------------------------------------------
// create table
//------------------------------------------------------------------------------
procedure TACRClientCursor.CreateTable
                         (
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                          );
var Len: Integer;
begin
  // send request
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqCreateTable;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  // prepare params
  Len := Length(FTableName);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10929);
  SaveDataToStream(PChar(@FTableName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10930);
  FieldDefs.SaveToStream(TACRClientSession(FSession).FSentCommandDataStream);
  IndexDefs.SaveToStream(TACRClientSession(FSession).FSentCommandDataStream);
  ConstraintDefs.SaveToStream(TACRClientSession(FSession).FSentCommandDataStream);
  // send command
  if (not TACRClientSession(FSession).SendCommand) then
   raise EACRException.Create(11515,ErrorLCannotSendCommandToServer);
  // receive and check the reply
  if (not TACRClientSession(FSession).ReceiveCommand) then
   raise EACRException.Create(11516,ErrorLCannotReceiveCommandFromServer);
  TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end;// CreateTable


//------------------------------------------------------------------------------
// delete table
//------------------------------------------------------------------------------
procedure TACRClientCursor.DeleteTable(Cascade: Boolean);
var Len: Integer;
    b:   ByteBool;
begin
  // send request
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqDeleteTable;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  // prepare params
  Len := Length(FTableName);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10931);
  SaveDataToStream(PChar(@FTableName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10932);
  b := Cascade;
  SaveDataToStream(b,sizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11479);
  // send command
  if (not TACRClientSession(FSession).SendCommand) then
   raise EACRException.Create(11517,ErrorLCannotSendCommandToServer);
  // receive and check the reply
  if (not TACRClientSession(FSession).ReceiveCommand) then
   raise EACRException.Create(11518,ErrorLCannotReceiveCommandFromServer);
  TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end; // DeleteTable;


//------------------------------------------------------------------------------
// empty table
//------------------------------------------------------------------------------
procedure TACRClientCursor.EmptyTable;
var Len: Integer;
begin
  // send request
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqEmptyTable;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  // prepare params
  Len := Length(FTableName);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10933);
  SaveDataToStream(PChar(@FTableName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10934);
  // send command
  if (not TACRClientSession(FSession).SendCommand) then
   raise EACRException.Create(11519,ErrorLCannotSendCommandToServer);
  // receive and check the reply
  if (not TACRClientSession(FSession).ReceiveCommand) then
   raise EACRException.Create(11520,ErrorLCannotReceiveCommandFromServer);
  TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end; // EmptyTable


//------------------------------------------------------------------------------
// add foreign key
//------------------------------------------------------------------------------
procedure TACRClientCursor.AddForeignKey(ConstraintDef: TACRConstraintDefForeignKey);
begin
  raise EACRException.Create(11585,ErrorLOperationIsNotSupported);
end; // AddForeignKey


//------------------------------------------------------------------------------
// delete constraint (FK,FKAction or PK) and write changes to MetaData file
//------------------------------------------------------------------------------
procedure TACRClientCursor.DeleteConstraint(Name: String; Cascade: Boolean; FKPartialDelete: Boolean);
begin
  raise EACRException.Create(11485,ErrorLOperationIsNotSupported);
end; // DeleteConstraint


//------------------------------------------------------------------------------
// rename table
//------------------------------------------------------------------------------
procedure TACRClientCursor.RenameTable(NewTableName: String);
var Len: Integer;
begin 
  // send request 
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqRenameTable;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0; 
  // prepare params 
  Len := Length(FTableName);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10935);
  SaveDataToStream(PChar(@FTableName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10936); 
  Len := Length(NewTableName);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10937); 
  SaveDataToStream(PChar(@NewTableName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10938);
  // send command 
  if (not TACRClientSession(FSession).SendCommand) then
   raise EACRException.Create(11520,ErrorLCannotSendCommandToServer); 
  // receive and check the reply
  if (not TACRClientSession(FSession).ReceiveCommand) then 
   raise EACRException.Create(11521,ErrorLCannotReceiveCommandFromServer); 
  TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
  FTableName := NewTableName; 
end; // RenameTable 
 
 
//------------------------------------------------------------------------------ 
// Rename Field
//------------------------------------------------------------------------------
procedure TACRClientCursor.RenameField(FieldName, NewFieldName: String);
var Len: Integer;
begin
  // send request 
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqRenameField;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL; 
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  // prepare params 
  Len := Length(FTableName); 
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10939);
  SaveDataToStream(PChar(@FTableName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10940); 
  Len := Length(FieldName);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10941); 
  SaveDataToStream(PChar(@FieldName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10942); 
  Len := Length(NewFieldName);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10943);
  SaveDataToStream(PChar(@NewFieldName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10944); 
  // send command 
  if (not TACRClientSession(FSession).SendCommand) then 
   raise EACRException.Create(11522,ErrorLCannotSendCommandToServer); 
  // receive and check the reply 
  if (not TACRClientSession(FSession).ReceiveCommand) then 
   raise EACRException.Create(11523,ErrorLCannotReceiveCommandFromServer); 
  TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end;//RenameField
 
	
//------------------------------------------------------------------------------
// repair table 
//------------------------------------------------------------------------------
function TACRClientCursor.RepairTable(var Log: String; NewSession: Pointer = nil): Boolean;
begin 
  raise EACRException.Create(11510,ErrorLOperationIsNotSupported); 
{ TODO : implement it } 
end; // RepairTable
	
 
//------------------------------------------------------------------------------
// load table
//------------------------------------------------------------------------------ 
procedure TACRClientCursor.LoadTableFromStream( 
                    Stream:               TStream
                   );
begin
  raise EACRException.Create(10925,ErrorLCannotLoadRemoteTable); 
end; // LoadTableToStream 
 
 
//------------------------------------------------------------------------------
// save table 
//------------------------------------------------------------------------------ 
procedure TACRClientCursor.SaveTableToStream( 
                    Stream:               TStream;
                    CompressionAlgorithm: TACRCompressionAlgorithm = acaNone;
                    CompressionMode:      Byte = 0;
                    BlockSize:            Integer = 0 
                  );
begin
  raise EACRException.Create(10926,ErrorLCannotSaveRemoteTable);
end; // SaveTableToStream 
 
	
//------------------------------------------------------------------------------
// export table to SQL
//------------------------------------------------------------------------------
function TACRClientCursor.ExportTableToSQL(
                            ExportStructure:      Boolean;
                            AddDropTableCommand:  Boolean; 
                            ExportIndexes:        Boolean;
                            AddDropIndexCommand:  Boolean;
                            ExportData:           Boolean; 
                            ExportBLOBFields:     Boolean;
                            UseBracketsForNames:  Boolean
                        ): String; 
begin 
{ TODO : implement it!!! } 
  Result := ''; 
end; // ExportTableToSQL
 
 
//------------------------------------------------------------------------------
// open table
//------------------------------------------------------------------------------
procedure TACRClientCursor.InternalInitFieldDefs;
begin 
  FFieldDefs.RecalcFieldOffsets; 
  FieldValuesOffset := FFieldDefs[0].MemoryOffset;
  BookmarkOffset := FFieldDefs[FFieldDefs.Count-1].MemoryOffset +
                   FFieldDefs[FFieldDefs.Count-1].MemoryDataSize; 
  KeyOffset := BookmarkOffset; 
  KeyBufferSize := BookmarkOffset + sizeof(TACRKeyBuffer); 
  RecordSize := BookmarkOffset + SizeOf(TACRBookmarkInfo);
  RecordBufferSize := RecordSize;
  CalculatedFieldsOffset := RecordSize;
end; // InternalInitFieldDefs
	
 
//------------------------------------------------------------------------------
// open table 
//------------------------------------------------------------------------------
procedure TACRClientCursor.OpenTableByFieldDefs(
                                    FieldDefs: TACRFieldDefs;
                                    IndexDefs: TACRIndexDefs;
                                    ConstraintDefs: TACRConstraintDefs
                                   );
var Len,i:    Integer;
    b:        ByteBool;
begin
  FIsOpen := False;
  // send request
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqOpenTable;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  // prepare params
  Len := Length(FTableName);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10960);
  SaveDataToStream(PChar(@FTableName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10961);
  b := FExclusive;
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,10962);
  // send command
  if (not TACRClientSession(FSession).SendCommand) then
   raise EACRException.Create(11524,ErrorLCannotSendCommandToServer);
  // receive and check the reply
  if (not TACRClientSession(FSession).ReceiveCommand) then
   raise EACRException.Create(11525,ErrorLCannotReceiveCommandFromServer);
  TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
  ReceiveServerCursor;
  for i := 0 to VisibleFieldDefs.Count-1 do
   FVisibleFieldDefs[i].FieldNoReference := i;
  FIsOpen := True;
  TACRClientSession(FSession).FReceivedCommandDataStream.Size := 0;
end;// OpenTable


//------------------------------------------------------------------------------
// close table
//------------------------------------------------------------------------------
procedure TACRClientCursor.CloseTable;
begin
  if (FIsOpen) then
   begin
    if (FFieldDefs <> nil) then
     FFieldDefs.Free;
    FFieldDefs := nil;
    if (FVisibleFieldDefs <> nil) then
     FVisibleFieldDefs.Free;
    FVisibleFieldDefs := nil;
    if (FIndexDefs <> nil) then
     FIndexDefs.Free;
    FIndexDefs := nil;
    if (FConstraintDefs <> nil) then
     FConstraintDefs.Free;
    FConstraintDefs := nil;
    ClearBLOBStreams(False);
    if (FBLOBStreams <> nil) then
      FBLOBStreams.Free;
    FBLOBStreams := nil;
    if (FBLOBCache <> nil) then
      FBLOBCache.Free;
    FBLOBCache := nil;
    FIsOpen := False;
    TACRClientSession(FSession).FSentCommandHeader.Request := accrqCloseTable;
    TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
    TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
    SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,10964);
    FCursorID := INVALID_OBJECT_ID;
    // send command
    if (TACRClientSession(FSession).SendCommand) then
      // receive and check the reply
      if (TACRClientSession(FSession).ReceiveCommand) then
        TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
   end;
end; // CloseTable


//------------------------------------------------------------------------------
// initialize record buffer
//------------------------------------------------------------------------------
procedure TACRClientCursor.InternalInitRecord(RecordBuffer: TACRRecordBuffer; InsertMode: Boolean);
var x,y,i:            Integer;
begin
  if ((InsertMode) and (FFieldDefs.AutoIncFieldsExists)) then
    begin
      // autoincrement fields exists - send query to update them and get values
      TACRClientSession(FSession).FSentCommandHeader.Request := accrqApplyDefaultValues;
      TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
      TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
      SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,10970);
      SendRecordBuffer(10971,RecordBuffer);
      // send command
      if (TACRClientSession(FSession).SendCommand) then
      // receive and check the reply
       if (TACRClientSession(FSession).ReceiveCommand) then
        begin
          TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
          ReceiveRecordBuffer(10972,RecordBuffer);
        end;
    end
   else
    begin
      // set null flags for all fields
      y := FFieldDefs.Count;
      x := y div 8;
      if (y mod 8 > 0) then
      Inc(x);
      FillChar(RecordBuffer^,x,$FF);
      // autoincrement fields does not exists - apply default values
      if ((InsertMode) and (FFieldDefs.DefaultValuesExists)) then
       for i := 0 to FFieldDefs.Count - 1 do
         if (not FFieldDefs[i].DefaultValue.IsNull) then
           begin
            FFieldDefs[i].DefaultValue.CopyDataToAddress(RecordBuffer +
              FFieldDefs[i].MemoryOffset);
            SetNullFlag(False, i, RecordBuffer);
           end;
    end; // autoincrement fields does not exists - apply default values
end; // InternalInitRecord


//------------------------------------------------------------------------------
// create new index
//------------------------------------------------------------------------------
procedure TACRClientCursor.AddIndex(IndexDef: TACRIndexDef);
var Len: Integer;
begin
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqAddIndex;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  Len := Length(FTableName);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10973);
  SaveDataToStream(PChar(@FTableName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10977);
  IndexDef.SaveToStream(TACRClientSession(FSession).FSentCommandDataStream);
  // send command
  if (not TACRClientSession(FSession).SendCommand) then
   raise EACRException.Create(11528,ErrorLCannotSendCommandToServer);
  // receive and check the reply
  if (not TACRClientSession(FSession).ReceiveCommand) then
   raise EACRException.Create(11529,ErrorLCannotReceiveCommandFromServer);
  TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end; // AddIndex


//------------------------------------------------------------------------------
// delete index
//------------------------------------------------------------------------------
procedure TACRClientCursor.DeleteIndex(Name: String);
var Len: Integer;
begin
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqDeleteIndex;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  Len := Length(FTableName);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10978);
  SaveDataToStream(PChar(@FTableName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10979);
  Len := Length(Name);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10975);
  SaveDataToStream(PChar(@Name[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10976);
  // send command
  if (not TACRClientSession(FSession).SendCommand) then
   raise EACRException.Create(11530,ErrorLCannotSendCommandToServer);
  // receive and check the reply
  if (not TACRClientSession(FSession).ReceiveCommand) then
   raise EACRException.Create(11531,ErrorLCannotReceiveCommandFromServer);
  TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end; // DeleteIndex


//------------------------------------------------------------------------------
// delete all indexes
//------------------------------------------------------------------------------
procedure TACRClientCursor.DeleteAllIndexes;
var Len: Integer;
begin
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqDeleteAllIndexes;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  Len := Length(FTableName);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,10980);
  SaveDataToStream(PChar(@FTableName[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,10981);
  // send command
  if (not TACRClientSession(FSession).SendCommand) then
   raise EACRException.Create(11532,ErrorLCannotSendCommandToServer);
  // receive and check the reply
  if (not TACRClientSession(FSession).ReceiveCommand) then
   raise EACRException.Create(11533,ErrorLCannotReceiveCommandFromServer);
  TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end; // DeleteAllIndexes


//------------------------------------------------------------------------------
// return index name of the index or '' if not found
//------------------------------------------------------------------------------
function TACRClientCursor.FindIndex(FieldNamesList,
            AscDescList, CaseSensitivityList: TStringList): String;
begin
  raise EACRException.Create(11383,ErrorLOperationIsNotSupported);
end; // FindIndex


//------------------------------------------------------------------------------
// return index name - for SQL only
//------------------------------------------------------------------------------
function TACRClientCursor.FindOrCreateIndex(FieldNamesList,
            AscDescList, CaseSensitivityList: TStringList): String;
begin
  raise EACRException.Create(10982,ErrorLOperationIsNotSupported);
end; // FindOrCreateIndex


//------------------------------------------------------------------------------
// return record count
//------------------------------------------------------------------------------
function TACRClientCursor.GetRecordCount: TACRRecordNo;
begin
  Result := -1;
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqGetRecordCount;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,10983);
  // send command
  if (TACRClientSession(FSession).SendCommand) then
   // receive and check the reply
   if (TACRClientSession(FSession).ReceiveCommand) then
    begin
      TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
      LoadDataFromStream(Result,SizeOf(Result),TACRClientSession(FSession).FReceivedCommandDataStream,10984);
    end;
end; // GetRecordCount


//------------------------------------------------------------------------------
// get record
//------------------------------------------------------------------------------
function TACRClientCursor.GetRecordBuffer(
          GetRecordMode:  TACRGetRecordMode
          ): TACRGetRecordResult;
var b:    Byte;
    bool: ByteBool;
begin
  Result := grrError;
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqGetRecordBuffer;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,10985);
  b := Byte(GetRecordMode);
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,10986);
  bool := FirstPosition;
  SaveDataToStream(bool,SizeOf(bool),TACRClientSession(FSession).FSentCommandDataStream,10987);
  bool := LastPosition;
  SaveDataToStream(bool,SizeOf(bool),TACRClientSession(FSession).FSentCommandDataStream,10988);
  SaveDataToStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FSentCommandDataStream,10989);
  SaveDataToStream(FIndexID,SizeOf(IndexID),TACRClientSession(FSession).FSentCommandDataStream,11089);
  // send command
  if (TACRClientSession(FSession).SendCommand) then
   // receive and check the reply
   if (TACRClientSession(FSession).ReceiveCommand) then
    begin
      TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
      LoadDataFromStream(b,SizeOf(b),TACRClientSession(FSession).FReceivedCommandDataStream,10990);
      Result := TACRGetRecordResult(b);
      LoadDataFromStream(bool,SizeOf(bool),TACRClientSession(FSession).FReceivedCommandDataStream,11131);
      FirstPosition := bool;
      LoadDataFromStream(bool,SizeOf(bool),TACRClientSession(FSession).FReceivedCommandDataStream,11132);
      LastPosition := bool;
      LoadDataFromStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FReceivedCommandDataStream,11133);
      if (Result = grrOK) then
        ReceiveRecordBuffer(10991,CurrentRecordBuffer);
    end;
end; // GetRecordBuffer


//------------------------------------------------------------------------------
// go to record
//------------------------------------------------------------------------------
procedure TACRClientCursor.SetRecNo(Value: Int64);
begin
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqSetRecNo;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11017);
  SaveDataToStream(Value,SizeOf(Value),TACRClientSession(FSession).FSentCommandDataStream,11012);
  SaveDataToStream(FIndexID,SizeOf(IndexID),TACRClientSession(FSession).FSentCommandDataStream,11093);
  // send command
  if (TACRClientSession(FSession).SendCommand) then
   // receive and check the reply
   if (TACRClientSession(FSession).ReceiveCommand) then
    begin
      TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
      FirstPosition := False;
      LastPosition := False;
      LoadDataFromStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FReceivedCommandDataStream,11013);
    end;
end; // SetRecNo


//------------------------------------------------------------------------------
// return current record number
//------------------------------------------------------------------------------
function TACRClientCursor.GetRecNo: Int64;
begin
  Result := -1;
  if ((not FirstPosition) and (not LastPosition)) then
   begin
    TACRClientSession(FSession).FSentCommandHeader.Request := accrqGetRecNo;
    TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
    TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
    SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11018);
    SaveDataToStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FSentCommandDataStream,11014);
    SaveDataToStream(FIndexID,SizeOf(IndexID),TACRClientSession(FSession).FSentCommandDataStream,11094);
    // send command
    if (TACRClientSession(FSession).SendCommand) then
     // receive and check the reply
     if (TACRClientSession(FSession).ReceiveCommand) then
      begin
        TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
        LoadDataFromStream(Result,SizeOf(Result),TACRClientSession(FSession).FReceivedCommandDataStream,11015);
      end;
   end;
end; // GetRecNo


//------------------------------------------------------------------------------
// insert record
//------------------------------------------------------------------------------
procedure TACRClientCursor.InternalInsert;
begin
end; // InternalInsert


//------------------------------------------------------------------------------
// edit record
//------------------------------------------------------------------------------
procedure TACRClientCursor.InternalEdit;
begin
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqInternalEdit;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11024);
  SaveDataToStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FSentCommandDataStream,11025);
  // send command
  if (TACRClientSession(FSession).SendCommand) then
   // receive and check the reply
   if (TACRClientSession(FSession).ReceiveCommand) then
    TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end; // InternalEdit


//------------------------------------------------------------------------------
// cancels updates
//------------------------------------------------------------------------------
procedure TACRClientCursor.InternalCancel(ToInsert: Boolean);
var
    b:        ByteBool;
begin
  ClearBLOBStreams(True);
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqInternalCancel;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11026);
  b := ToInsert;
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11049);
  SaveDataToStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FSentCommandDataStream,11027);
  // send command
  if (TACRClientSession(FSession).SendCommand) then
   // receive and check the reply
   if (TACRClientSession(FSession).ReceiveCommand) then
    TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end; // InternalCancel


//------------------------------------------------------------------------------
// update record
//------------------------------------------------------------------------------
procedure TACRClientCursor.InternalPost(ToInsert: Boolean);
var
    b:        ByteBool;
    Bookmark: PACRBookmarkInfo;
begin
  if (CurrentRecordBuffer = nil) then
   raise EACRException.Create(11028,ErrorLNilPointer);
  if (not ToInsert) then
   Move(PChar(FCurrentRecordBuffer + BookmarkOffset)^,
    CurrentRecordID,sizeof(TACRRecordID));
  ClearBLOBStreams(True);
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqInternalPost;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11026);
  b := ToInsert;
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11029);
  SaveDataToStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FSentCommandDataStream,11030);
  SaveDataToStream(FIndexID,SizeOf(IndexID),TACRClientSession(FSession).FSentCommandDataStream,11095);
  SendRecordBuffer(11031,CurrentRecordBuffer);
  b := (EditRecordBuffer = nil);
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11032);
  if (not b) then
    SendRecordBuffer(11033,EditRecordBuffer);
  SendModifiedBLOBValues;  
  // send command
  if (TACRClientSession(FSession).SendCommand) then
    // receive and check the reply
    if (TACRClientSession(FSession).ReceiveCommand) then
     begin
      TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
      LoadDataFromStream(b,SizeOf(b),TACRClientSession(FSession).FReceivedCommandDataStream,11034);
      FirstPosition := b;
      LoadDataFromStream(b,SizeOf(b),TACRClientSession(FSession).FReceivedCommandDataStream,11035);
      LastPosition := b;
      LoadDataFromStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FReceivedCommandDataStream,11036);
      ReceiveRecordBuffer(11037,CurrentRecordBuffer);
      if (ToInsert) then
       begin
         // write bookmark info to new record buffer
         Bookmark := PACRBookmarkInfo(CurrentRecordBuffer + BookmarkOffset);
         Bookmark^.BookmarkData := CurrentRecordID;
         Bookmark^.BookmarkFlag := abfInserted;
       end;
     end;
end; // InternalPost


//------------------------------------------------------------------------------
// delete record
//------------------------------------------------------------------------------
procedure TACRClientCursor.InternalDelete;
var
    b:        ByteBool;
begin
  Move(PChar(FCurrentRecordBuffer + BookmarkOffset)^,
    CurrentRecordID,sizeof(TACRRecordID));
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqInternalDelete;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11038);
  SaveDataToStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FSentCommandDataStream,11039);
  SaveDataToStream(FIndexID,SizeOf(IndexID),TACRClientSession(FSession).FSentCommandDataStream,11096);
  // send command
  if (TACRClientSession(FSession).SendCommand) then
    // receive and check the reply
    if (TACRClientSession(FSession).ReceiveCommand) then
     begin
      TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
      LoadDataFromStream(b,SizeOf(b),TACRClientSession(FSession).FReceivedCommandDataStream,11040);
      FirstPosition := b;
      LoadDataFromStream(b,SizeOf(b),TACRClientSession(FSession).FReceivedCommandDataStream,11041);
      LastPosition := b;
      LoadDataFromStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FReceivedCommandDataStream,11042);
      ReceiveRecordBuffer(11043,CurrentRecordBuffer);
     end;
end; // InternalDelete


//------------------------------------------------------------------------------
// delete all visible records
//------------------------------------------------------------------------------
procedure TACRClientCursor.DeleteVisibleRecords;
begin
  raise EACRException.Create(11303,ErrorLOperationIsNotSupported);
end; // DeleteVisibleRecords


//------------------------------------------------------------------------------
// update visible records
//------------------------------------------------------------------------------
procedure TACRClientCursor.UpdateVisibleRecords(FieldNames:   TStringList;
                                                values:       array of TACRVariant;
                                                SkipFKCheck:  Boolean = False
                                                );
begin
  raise EACRException.Create(11459,ErrorLOperationIsNotSupported);
end; // UpdateVisibleRecords


//------------------------------------------------------------------------------
// disable record bitmap
//------------------------------------------------------------------------------
procedure TACRClientCursor.DisableRecordBitmap;
begin
// do nothing
end; // DisableRecordBitmap


//------------------------------------------------------------------------------
// set distinct
//------------------------------------------------------------------------------
procedure TACRClientCursor.ApplyDistinct(FieldNamesList, AscDescList, CaseSensitivityList: TStringList);
begin
  raise EACRException.Create(11068,ErrorLOperationIsNotSupported);
end; // ApplyDistinct


//------------------------------------------------------------------------------
// apply projection
//------------------------------------------------------------------------------
procedure TACRClientCursor.ApplyProjection(FieldNamesList, AliasList: TStringList);
begin
  raise EACRException.Create(11069,ErrorLOperationIsNotSupported);
end; // ApplyProjection


//------------------------------------------------------------------------------
// Activate filters
//------------------------------------------------------------------------------
procedure TACRClientCursor.ActivateFilters(
                          FilterText:      String;
                          CaseInsensitive: Boolean;
                          PartialKey:      Boolean
                        );
var Len:  Integer;
    b:    ByteBool;
begin
  // send request
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqActivateFilters;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  // prepare params
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11075);
  Len := Length(FilterText);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,11071);
  SaveDataToStream(PChar(@FilterText[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,11072);
  b := CaseInsensitive;
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11073);
  b := PartialKey;
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11074);
  // send command
  if (TACRClientSession(FSession).SendCommand) then
   // receive and check the reply
   if (TACRClientSession(FSession).ReceiveCommand) then
    TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end; // ActivateFilters


//------------------------------------------------------------------------------
// Deactivate filters
//------------------------------------------------------------------------------
procedure TACRClientCursor.DeactivateFilters;
begin
  // send request
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqDeactivateFilters;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11076);
  // send command
  if (TACRClientSession(FSession).SendCommand) then
    // receive and check the reply
    if ( TACRClientSession(FSession).ReceiveCommand) then
      TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end; // DeactivateFilters


//------------------------------------------------------------------------------
// Locate
//------------------------------------------------------------------------------
function TACRClientCursor.Locate(
                const KeyFields: String;
                const KeyValues: Variant;
                CaseInsensitive: Boolean;
                PartialKey:      Boolean
               ): Boolean;
var
    i, ArrLen, Len: Integer;
    v:              TACRVariant;
    b:              ByteBool;
begin
  // send request
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqLocate;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11081);
  SaveDataToStream(FIndexID,SizeOf(IndexID),TACRClientSession(FSession).FSentCommandDataStream,11099);
  Len := Length(KeyFields);
  SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,11082);
  SaveDataToStream(PChar(@KeyFields[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,11083);
  b := CaseInsensitive;
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11084);
  b := PartialKey;
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11085);
  // Determinate KeyValues count (can be array)
  if  (VarType(KeyValues) and varArray ) <> 0 then
    ArrLen := VarArrayHighBound(KeyValues,1) - VarArrayLowBound(KeyValues,1) + 1
  else
    ArrLen := 1;
  SaveDataToStream(ArrLen,SizeOf(ArrLen),TACRClientSession(FSession).FSentCommandDataStream,11086);
  v := TACRVariant.Create;
  try
   if (ArrLen = 1) then
    begin
     v.AsVariant := KeyValues;
     v.SaveToStream(TACRClientSession(FSession).FSentCommandDataStream);
    end
   else
    begin
     for i := 0 to ArrLen-1 do
      begin
       v.AsVariant := VarArrayGet(KeyValues, i);
       v.SaveToStream(TACRClientSession(FSession).FSentCommandDataStream);
      end;
    end;
  finally
    v.Free;
  end;
  Result := False;
  // send command
  if (TACRClientSession(FSession).SendCommand) then
    // receive and check the reply
    if (TACRClientSession(FSession).ReceiveCommand) then
      begin
        Result := (TACRClientSession(FSession).ReceivedCommandHeader.Reply = accrplYes);
        if (Result) then
         begin
          LoadDataFromStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FReceivedCommandDataStream,11122);
          ReceiveRecordBuffer(11123,FCurrentRecordBuffer);
          FirstPosition := False;
          LastPosition := False;
         end;
      end;
end; // Locate


//------------------------------------------------------------------------------
// FindKey
//------------------------------------------------------------------------------
function TACRClientCursor.FindKey(SearchCondition: TACRSearchCondition): Boolean;
var
    sc:             Byte;
begin
  // send request
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqFindKey;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11087);
  SaveDataToStream(FIndexID,SizeOf(IndexID),TACRClientSession(FSession).FSentCommandDataStream,11100);
  sc := Byte(SearchCondition);
  SaveDataToStream(sc,SizeOf(sc),TACRClientSession(FSession).FSentCommandDataStream,11088);
  SaveDataToStream(FKeyFieldCount,SizeOf(FKeyFieldCount),TACRClientSession(FSession).FSentCommandDataStream,11101);
  SendRecordBuffer(11102,FKeyBuffer);
  Result := False;
  // send command
  if (TACRClientSession(FSession).SendCommand) then
    // receive and check the reply
    if (TACRClientSession(FSession).ReceiveCommand) then
     begin
      Result := (TACRClientSession(FSession).ReceivedCommandHeader.Reply = accrplYes);
      if (Result) then
       begin
        LoadDataFromStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FReceivedCommandDataStream,11129);
        FirstPosition := False;
        LastPosition := False;
       end;
     end;
end; // FindKey


//------------------------------------------------------------------------------
// reset range
//------------------------------------------------------------------------------
procedure TACRClientCursor.ResetRange;
begin
  // send request
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqResetRange;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11103);
  // send command
  if (TACRClientSession(FSession).SendCommand) then
    // receive and check the reply
    if (TACRClientSession(FSession).ReceiveCommand) then
      TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end; // ResetRange


//------------------------------------------------------------------------------
// apply range
//------------------------------------------------------------------------------
procedure TACRClientCursor.ApplyRange(
                          StartBuffer, EndBuffer: TACRRecordBuffer;
                          StartKeyFieldCount:     Integer;
                          EndKeyFieldCount:       Integer;
                          StartExclusive:         Boolean;
                          EndExclusive:           Boolean
                        );
var b:  ByteBool;
begin
  // send request
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqApplyRange;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11104);
  SaveDataToStream(FIndexID,SizeOf(IndexID),TACRClientSession(FSession).FSentCommandDataStream,11105);
  b := StartExclusive;
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11106);
  b := EndExclusive;
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11107);
  SaveDataToStream(StartKeyFieldCount,SizeOf(StartKeyFieldCount),TACRClientSession(FSession).FSentCommandDataStream,11108);
  SaveDataToStream(EndKeyFieldCount,SizeOf(EndKeyFieldCount),TACRClientSession(FSession).FSentCommandDataStream,11109);
  b := (StartBuffer = nil);
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11112);
  if (not b) then
    SendRecordBuffer(11110,StartBuffer);
  b := (EndBuffer = nil);
  SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11113);
  if (not b) then
    SendRecordBuffer(11111,EndBuffer);
  // send command
  if (TACRClientSession(FSession).SendCommand) then
    // receive and check the reply
    if (TACRClientSession(FSession).ReceiveCommand) then
      TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
end; // ApplyRange


//------------------------------------------------------------------------------
// set SQL Filter
//------------------------------------------------------------------------------
procedure TACRClientCursor.SetSQLFilter(FilterExpr: TObject);
begin
  raise EACRException.Create(11070,ErrorLOperationIsNotSupported);
end; // SetSQLFilter


//------------------------------------------------------------------------------
// create blob stream
//------------------------------------------------------------------------------
function TACRClientCursor.InternalCreateBlobStream(
          ToInsert: Boolean;
          FieldNo:  Integer;
          OpenMode: TACRBLOBOpenMode
          ):TACRStream;
var FieldNumber:      Integer;
    i:                Integer;
    LocalBLOBStream:  TACRLocalBLOBStream;
    TempStream:       TACRTemporaryStream;
    Size:             Int64;
begin
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Client: create client blob stream 0, FieldNo = '+IntToStr(FieldNo)+', openmode = '+IntToStr(Integer(OpenMode)));
{$ENDIF}
 Size := 0;
 if (FieldNo >= FVisibleFieldDefs.Count) then
  raise EACRException.Create(11208,ErrorLInvalidFieldNumber,
    [FieldNo,FVisibleFieldDefs.Count]);
 if (FBLOBStreams = nil) then
  raise EACRException.Create(11209,ErrorLNilPointer);
 FieldNumber := FVisibleFieldDefs.Items[FieldNo].FieldNoReference;

{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Client: create client blob stream 01, FieldNumber = '+IntToStr(FieldNumber));
{$ENDIF}
 // find existing blob stream
 if ((OpenMode = bomReadWrite) or (OpenMode = bomWrite)) then
  begin
    for i := 0 to FBLOBStreams.Count -1 do
     begin
      LocalBLOBStream := TACRLocalBLOBStream(FBLOBStreams.Items[i]);
      if (LocalBLOBStream = nil) then
        raise EACRException.Create(11210,ErrorLNilPointer);
      if (LocalBLOBStream.FieldNo = FieldNumber) then
       if ((LocalBLOBStream.OpenMode = bomReadWrite) or
           (LocalBLOBStream.OpenMode = bomWrite)) then
        raise EACRException.Create(11211,ErrorLBLOBFieldAlreadyOpened,[FieldNumber,
          FFieldDefs.Items[FieldNumber].Name]);
     end;
    for i := 0 to FBLOBCache.Count -1 do
     begin
      LocalBLOBStream := TACRLocalBLOBStream(FBLOBStreams.Items[i]);
      if (LocalBLOBStream = nil) then
        raise EACRException.Create(11213,ErrorLNilPointer);
      // if BLOB already was modified - reopen it
      if (LocalBLOBStream.FieldNo = FieldNumber) then
       begin
        FBLOBCache.Remove(LocalBLOBStream);
        LocalBLOBStream.OpenMode := OpenMode;
        if (OpenMode = bomWrite) then
         LocalBLOBStream.TemporaryStream.Size := 0;
        FBLOBStreams.Add(LocalBLOBStream);
        Result := LocalBLOBStream;
        Exit;
       end;
     end;
  end;
  // create temporary stream
  TempStream := TACRTemporaryStream.Create;
  LocalBLOBStream := TACRLocalBLOBStream.Create(TempStream,Self,OpenMode,FieldNumber);
  LocalBLOBStream.DoNotFreeCompressedStream := True;
  Result := LocalBLOBStream;
  FBLOBStreams.Add(Result);
  LocalBLOBStream.OpenMode := OpenMode;
  LocalBLOBStream.Modified := (OpenMode = bomWrite);
  if (not LocalBLOBStream.Modified) then
   begin
    // get blob stream from server
    // send request
    TACRClientSession(FSession).FSentCommandHeader.Request := accrqReadBLOBValue;
    TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
    TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
    SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11214);
    SaveDataToStream(FieldNumber,SizeOf(FieldNumber),TACRClientSession(FSession).FSentCommandDataStream,11215);
    // send command
    if (TACRClientSession(FSession).SendCommand) then
      // receive and check the reply
      if (TACRClientSession(FSession).ReceiveCommand) then
        begin
          {$IFDEF DEBUG_LOG_CLIENT_BLOB}
          aaWriteToLog('Client: create client blob stream 1');
          {$ENDIF}
          try
                TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
          {$IFDEF DEBUG_LOG_CLIENT_BLOB}
          aaWriteToLog('Client: create client blob stream 2');
          {$ENDIF}
          except
          {$IFDEF DEBUG_LOG_CLIENT_BLOB}
          aaWriteToLog('Client: create client blob stream - error');
          {$ENDIF}
            LocalBLOBStream.Free;
            raise;
          end;
          LoadDataFromStream(Size,SizeOf(Size),TACRClientSession(FSession).FReceivedCommandDataStream,11216);
        end;
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Client: create client blob stream 3, size = '+IntToStr(Size));
{$ENDIF}
    if (Size > 0) then
     TempStream.LoadFromStreamWithPosition(
        TACRClientSession(FSession).FReceivedCommandDataStream,
        TACRClientSession(FSession).FReceivedCommandDataStream.Position,Size);
   end; // get blob stream from server
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Client: create client blob stream  - ok');
{$ENDIF}
end; // InternalCreateBlobStream


//------------------------------------------------------------------------------
// close blob field
//------------------------------------------------------------------------------
procedure TACRClientCursor.InternalCloseBLOB(FieldNo: Integer);
var i,j:               Integer;
    FieldNumber:       Integer;
    LocalBLOBStream:   TACRLocalBLOBStream;
begin
 if (FieldNo >= FVisibleFieldDefs.Count) then
  raise EACRException.Create(11220,ErrorLInvalidFieldNumber,
    [FieldNo,FVisibleFieldDefs.Count]);
 if (FBLOBStreams = nil) then
  raise EACRException.Create(11221,ErrorLNilPointer);
 FieldNumber := FVisibleFieldDefs.Items[FieldNo].FieldNoReference;
 i := 0;
 while (i < FBLOBStreams.Count) do
  begin
   LocalBLOBStream := TACRLocalBLOBStream(FBLOBStreams.Items[i]);
   if (LocalBLOBStream = nil) then
      raise EACRException.Create(11222,ErrorLNilPointer);
   if (LocalBLOBStream.FieldNo = FieldNumber) then
    begin
     if (LocalBLOBStream.Modified and ((LocalBLOBStream.OpenMode = bomWrite) or
        (LocalBLOBStream.OpenMode = bomReadWrite))) then
      begin
       // move changed blob value to cache
       j := FBLOBCache.IndexOf(LocalBLOBStream);
       if (j < 0) then
        FBLOBCache.Add(LocalBLOBStream);
      end
     else
       LocalBLOBStream.Free;
     if (LocalBLOBStream.UserBLOBStream <> nil) then
          LocalBLOBStream.UserBLOBStream.Free;
     FBLOBStreams.Delete(i);
     Dec(i);
    end;
   Inc(i);
  end;// scan all opened blob streams and close them if necessary
end; // InternalCloseBLOB


//------------------------------------------------------------------------------
// clear blob streams
//------------------------------------------------------------------------------
procedure TACRClientCursor.ClearBLOBStreams(WriteOnly: Boolean = False);
var i:                  Integer;
    LocalBLOBStream:    TACRLocalBLOBStream;
begin
 i := 0;
 if (FBLOBStreams <> nil) then
   while (i < FBLOBStreams.Count) and (FBLOBStreams.Count > 0) do
    begin
     LocalBLOBStream := TACRLocalBLOBStream(FBLOBStreams.Items[i]);
     if (LocalBLOBStream  = nil) then
      raise EACRException.Create(11223,ErrorLNilPointer);
     if ((Not WriteOnly) or
         ((WriteOnly) and
          ((LocalBLOBStream.OpenMode = bomWrite) or
           (LocalBLOBStream.OpenMode = bomReadWrite)))) then
       begin
        if (LocalBLOBStream.UserBLOBStream <> nil) then
          LocalBLOBStream.UserBLOBStream.Free
        else
         begin
          LocalBLOBStream.Free;
          FBLOBStreams.Delete(i);
         end;
        Dec(i);
       end;
     Inc(i);
    end;
  if (not WriteOnly) then
   if (FBLOBCache <> nil) then
    begin
     for i := 0 to FBLOBCache.Count - 1 do
      begin
       LocalBLOBStream := TACRLocalBLOBStream(FBLOBCache.Items[i]);
       LocalBLOBStream.Free;
      end;
     FBLOBCache.Clear;
    end;
end; // ClearBLOBStreams


//------------------------------------------------------------------------------
// send modified blob values to server
//------------------------------------------------------------------------------
procedure TACRClientCursor.SendModifiedBLOBValues;
var i, FieldNumber, Count: Integer;
    Size:                  Int64;
    LocalBLOBStream:       TACRLocalBLOBStream;
begin
  Count := FBLOBCache.Count;
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Client: send modified blob values, Count = '+IntToStr(Count));
{$ENDIF}
  SaveDataToStream(Count,SizeOf(Count),TACRClientSession(FSession).FSentCommandDataStream,11224);
  for i := 0 to Count - 1 do
   begin
    LocalBLOBStream := TACRLocalBLOBStream(FBLOBCache.Items[i]);
    if (LocalBLOBStream  = nil) then
      raise EACRException.Create(11225,ErrorLNilPointer);
    FieldNumber := LocalBLOBStream.FieldNo;
    Size := LocalBLOBStream.Size;
    LocalBLOBStream.Position := 0;
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Client: send modified blob values ... FieldNumber = '+IntToStr(FieldNumber)+
', Size = '+IntToStr(Size)+', Position = '+IntToStr(LocalBLOBStream.Position)+
', openMode = '+IntToStr(Integer(LocalBLOBStream.OpenMode))+
', Modified = '+IntToStr(Integer(LocalBLOBStream.Modified)));
{$ENDIF}
    SaveDataToStream(FieldNumber,SizeOf(FieldNumber),TACRClientSession(FSession).FSentCommandDataStream,11226);
    SaveDataToStream(Size,SizeOf(Size),TACRClientSession(FSession).FSentCommandDataStream,11227);
    if (Size > 0) then
     begin
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Client: send modified blob values ... Saving stream, Start Position = '+IntToStr(TACRClientSession(FSession).FSentCommandDataStream.Position)+
', LocalBLOBStream.Position = '+IntToStr(LocalBLOBStream.Position));
{$ENDIF}
      LocalBLOBStream.SaveToStream(TACRClientSession(FSession).FSentCommandDataStream);
      TACRClientSession(FSession).FSentCommandDataStream.Position :=
        TACRClientSession(FSession).FSentCommandDataStream.Position + Size;
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Client: send modified blob values ... Saving stream ok, End Position = '+IntToStr(TACRClientSession(FSession).FSentCommandDataStream.Position)+
', LocalBLOBStream.Position = '+IntToStr(LocalBLOBStream.Position));
{$ENDIF}
     end;
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Client: send modified blob values ... closing stream ...');
{$ENDIF}
    LocalBLOBStream.DoNotFreeCompressedStream := True;
    LocalBLOBStream.Free;
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('Client: send modified blob values ... closing stream ... OK');
{$ENDIF}
   end;
  FBLOBCache.Clear;
{$IFDEF DEBUG_LOG_CLIENT_BLOB}
aaWriteToLog('send modified blob values ok');
{$ENDIF}
end; // SendModifiedBLOBValues


//------------------------------------------------------------------------------
// LastAutoincValue
//------------------------------------------------------------------------------
function TACRClientCursor.LastAutoincValue(FieldNo: Integer): Int64;
begin
  raise EACRException.Create(11137,ErrorLOperationIsNotSupported);
end; // LastAutoincValue


//------------------------------------------------------------------------------
// SetLastAutoincValue
//------------------------------------------------------------------------------
procedure TACRClientCursor.SetLastAutoincValue(Value: Int64; FieldNo: Integer);
begin
  raise EACRException.Create(11138,ErrorLOperationIsNotSupported);
end; // SetLastAutoincValue


//------------------------------------------------------------------------------
// lock table for read
//------------------------------------------------------------------------------
procedure TACRClientCursor.LockTableForRead;
begin
{ TODO : implement it }
  raise EACRException.Create(11568,ErrorLOperationIsNotSupported);
end; // LockTableForRead;


//------------------------------------------------------------------------------
// unlock table for read
//------------------------------------------------------------------------------
procedure TACRClientCursor.UnlockTableForRead;
begin
{ TODO : implement it }
  raise EACRException.Create(11569,ErrorLOperationIsNotSupported);
end; // UnlockTableForRead;


//------------------------------------------------------------------------------
// LockTableS
//------------------------------------------------------------------------------
procedure TACRClientCursor.LockTableS(ErrorCode: Integer);
begin
  raise EACRException.Create(11139,ErrorLOperationIsNotSupported);
end; // LockTableS


//------------------------------------------------------------------------------
// UnlockTableS
//------------------------------------------------------------------------------
procedure TACRClientCursor.UnlockTableS(ErrorCode: Integer);
begin
  raise EACRException.Create(11140,ErrorLOperationIsNotSupported);
end; // UnlockTableS


//------------------------------------------------------------------------------
// LockTableIRW
//------------------------------------------------------------------------------
procedure TACRClientCursor.LockTableIRW(ErrorCode: Integer);
begin
  raise EACRException.Create(11141,ErrorLOperationIsNotSupported);
end; // LockTableIRW


//------------------------------------------------------------------------------
// UnlockTableIRW
//------------------------------------------------------------------------------
procedure TACRClientCursor.UnlockTableIRW(ErrorCode: Integer);
begin
  raise EACRException.Create(11142,ErrorLOperationIsNotSupported);
end; // UnlockTableIRW


//------------------------------------------------------------------------------
// LockTableRW
//------------------------------------------------------------------------------
procedure TACRClientCursor.LockTableRW(ErrorCode: Integer);
begin
  raise EACRException.Create(11143,ErrorLOperationIsNotSupported);
end; // LockTableRW


//------------------------------------------------------------------------------
// UnlockTableRW
//------------------------------------------------------------------------------
procedure TACRClientCursor.UnlockTableRW(ErrorCode: Integer);
begin
  raise EACRException.Create(11144,ErrorLOperationIsNotSupported);
end; // UnlockTableRW


//------------------------------------------------------------------------------
// TryToLockTableS
//------------------------------------------------------------------------------
function TACRClientCursor.TryToLockTableS: Boolean;
begin
  raise EACRException.Create(11145,ErrorLOperationIsNotSupported);
end; // TryToLockTableS


//------------------------------------------------------------------------------
// TryToUnlockTableS
//------------------------------------------------------------------------------
function TACRClientCursor.TryToUnlockTableS: Boolean;
begin
  raise EACRException.Create(11146,ErrorLOperationIsNotSupported);
end; // TryToUnlockTableS


//------------------------------------------------------------------------------
// TryToLockTableIRW
//------------------------------------------------------------------------------
function TACRClientCursor.TryToLockTableIRW: Boolean;
begin
  raise EACRException.Create(11147,ErrorLOperationIsNotSupported);
end; // TryToLockTableIRW


//------------------------------------------------------------------------------
// TryToUnlockTableIRW
//------------------------------------------------------------------------------
function TACRClientCursor.TryToUnlockTableIRW: Boolean;
begin
  raise EACRException.Create(11148,ErrorLOperationIsNotSupported);
end; // TryToUnlockTableIRW


//------------------------------------------------------------------------------
// TryToLockTableRW
//------------------------------------------------------------------------------
function TACRClientCursor.TryToLockTableRW: Boolean;
begin
  raise EACRException.Create(11149,ErrorLOperationIsNotSupported);
end; // TryToLockTableRW


//------------------------------------------------------------------------------
// TryToUnlockTableRW
//------------------------------------------------------------------------------
function TACRClientCursor.TryToUnlockTableRW: Boolean;
begin
  raise EACRException.Create(11150,ErrorLOperationIsNotSupported);
end; // TryToUnlockTableRW


//------------------------------------------------------------------------------
// return true if current record exists
//------------------------------------------------------------------------------
function TACRClientCursor.IsRecordExists: Boolean;
begin
  Result := False;
  TACRClientSession(FSession).FSentCommandHeader.Request := accrqIsRecordExists;
  TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
  TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
  SaveDataToStream(FCursorID,SizeOf(FCursorID),TACRClientSession(FSession).FSentCommandDataStream,11281);
  SaveDataToStream(CurrentRecordID,SizeOf(CurrentRecordID),TACRClientSession(FSession).FSentCommandDataStream,11282);
  // send command
  if (TACRClientSession(FSession).SendCommand) then
   // receive and check the reply
   if (TACRClientSession(FSession).ReceiveCommand) then
    Result := (TACRClientSession(FSession).FReceivedCommandHeader.Reply = accrplYes);
end; // IsRecordExists


////////////////////////////////////////////////////////////////////////////////
//
// TACRClientSQLProcessor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// send params
//------------------------------------------------------------------------------
procedure TACRClientSQLProcessor.SendParams;
begin
  FSQLParams.SaveToStream(FSession.FSentCommandDataStream);
end; // SendParams


//------------------------------------------------------------------------------
// receive live query
//------------------------------------------------------------------------------
function TACRClientSQLProcessor.ReceiveLiveQuery: TACRCursor;
begin
  Result := TACRClientCursor.Create;
  Result.TableName := GetTemporaryName(ACRTemporaryTableName);
  Result.Session := FSession;
  TACRClientCursor(Result).ReceiveServerCursor;
  TACRClientCursor(Result).ReceiveFieldNoReferences;
end; // ReceiveLiveQuery


//------------------------------------------------------------------------------
// receive not live query
//------------------------------------------------------------------------------
function TACRClientSQLProcessor.ReceiveNotLiveQuery: TACRCursor;
var TempSession: TACRLocalSession;
begin
  TempSession := TACRLocalSession.Create;
  TempSession.DatabaseName := ACRTemporaryDatabaseName;
  TempSession.Connected := True;
  Result := TACRLocalCursor.Create;
  Result.Temporary := True;
  Result.Session := TempSession;
  Result.LoadTableFromStream(FSession.FReceivedCommandDataStream);
  Result.OpenTableByFieldDefs(nil,nil,nil);
  Result.InternalInitFieldDefs;
end; // ReceiveNotLiveQuery


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientSQLProcessor.Create(Query: TDataSet; aSession: TACRBaseSession);
begin
  inherited Create(Query);
  FSession := TACRClientSession(aSession);
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRClientSQLProcessor.Destroy;
begin
  inherited;
end; // Destroy


//------------------------------------------------------------------------------
// Open query
//------------------------------------------------------------------------------
function TACRClientSQLProcessor.OpenQuery: TACRCursor;
var b:                ByteBool;
    len:              Integer;
begin
  Result := inherited OpenQuery;
  Len := Length(TACRQuery(FACRQuery).SQL.Text);
  if (Len > 0) then
   begin
    // send request
    TACRClientSession(FSession).FSentCommandHeader.Request := accrqExecSQL;
    TACRClientSession(FSession).FSentCommandHeader.Reply := accrplNULL;
    TACRClientSession(FSession).FSentCommandDataStream.Size := 0;
    // save query text
    SaveDataToStream(Len,SizeOf(Len),TACRClientSession(FSession).FSentCommandDataStream,11170);
    SaveDataToStream(PChar(@TACRQuery(FACRQuery).SQL.Text[1])^,Len,TACRClientSession(FSession).FSentCommandDataStream,11171);
    b := TACRQuery(FACRQuery).RequestLive;
    SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11173);
    b := TACRQuery(FACRQuery).ReadOnly;
    SaveDataToStream(b,SizeOf(b),TACRClientSession(FSession).FSentCommandDataStream,11174);
    SendParams;
    // send command
    if (TACRClientSession(FSession).SendCommand) then
     // receive and check the reply
     if (TACRClientSession(FSession).ReceiveCommand) then
      begin
        TACRClientSession(FSession).CheckReceivedReply(accrplOperationSucceed);
        // get number of rows affected by query
        LoadDataFromStream(FRowsAffected,SizeOf(FRowsAffected),TACRClientSession(FSession).FReceivedCommandDataStream,11175);
        // get cursor if necessary
        LoadDataFromStream(b,SizeOf(b),TACRClientSession(FSession).FReceivedCommandDataStream,11176);
        if (b) then
         begin
          // receive cursor info if live query or receive records if not live query
          LoadDataFromStream(b,SizeOf(b),TACRClientSession(FSession).FReceivedCommandDataStream,11177);
          if (b) then
            Result := ReceiveLiveQuery
          else
            Result := ReceiveNotLiveQuery;
          Result.ReadOnly := not b;
          Result.IsClientCursor := True;
          FReadOnly := Result.ReadOnly;
         end;
        TACRClientSession(FSession).FReceivedCommandDataStream.Size := 0;
      end;
   end
  else
   raise EACRException.Create(11172,ErrorLNoQueryText);
end; // OpenQuery


initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRClient> initializing...');
{$ENDIF}

 ClientConnectionManager := TACRClientConnectionManager.Create;

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRClient> initialized');
{$ENDIF}

finalization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRClient> finalizing...');
{$ENDIF}

 ClientConnectionManager.Free;
 ClientConnectionManager := nil;


{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRClient> finalized');
{$ENDIF}

end.
