unit ACRPage;

{$I ACRVer.inc}

interface

uses Classes, SysUtils,
{$IFDEF MSWINDOWS}
     Windows,
{$ENDIF}
{$IFDEF LINUX}
     Libc,
{$ENDIF}
     Math,

// Accuracer units
     ACRCriticalSection,
     {$IFDEF DEBUG_LOG}
     ACRDebug,
     {$ENDIF}
     ACRExcept,
     ACRMemory,
     ACRCompression,
     ACRConst,
     ACRTypes;

type

  TACRPageManager = class;

////////////////////////////////////////////////////////////////////////////////
//
// TACRPage
//
////////////////////////////////////////////////////////////////////////////////

  TACRPage = class(TObject)
   private
     LPageManager:          TACRPageManager;
     FPageNo:               TACRPageNo;
     FPageBuffer:           TACRPageBuffer;
     FOwnBuffer:            Boolean;
     FIsDirty:              Boolean;
     FUseCount:             Integer;
     FParentList:           TList;
     FSessionID:            TACRSessionID;
     FState:                Cardinal;
     FPageHeaderSize:       Integer;
     FPageSize:             Integer;

     function GetPageHeader: PACRDiskPageHeader;
     function GetPageHeaderSize: Integer;
     function GetPageData: TACRPageBuffer;
     function GetPageDataSize: Integer;
     function GetPageSize: Integer;

     procedure InitHeader;

   public
     constructor Create(PageManager: TACRPageManager);
     destructor Destroy; override;
     procedure ClearPageBuffer;
     procedure AllocPageBuffer;
     procedure FreeAndNilPageBuffer;
     procedure EnlargePageBuffer(NewSize: Integer);
     procedure Synchronize;

     property PageNo: TACRPageNo read FPageNo write FPageNo;
     property PageSize: Integer read GetPageSize;
     property PageBuffer: TACRPageBuffer read FPageBuffer write FPageBuffer;
     property OwnBuffer: Boolean read FOwnBuffer write FOwnBuffer;
     property IsDirty: Boolean read FIsDirty write FIsDirty;
     property UseCount: Integer read FUseCount write FUseCount;
     property PageManager: TACRPageManager read LPageManager write LPageManager;
     property PageHeader: PACRDiskPageHeader read GetPageHeader;
     property PageHeaderSize: Integer read GetPageHeaderSize;
     property PageData: TACRPageBuffer read GetPageData;
     property PageDataSize: Integer read GetPageDataSize;
     property ParentList: TList read FParentList write FParentList;
     property SessionID: TACRSessionID read FSessionID write FSessionID;
     property State: Cardinal read FState write FState;
end;// TACRPage


////////////////////////////////////////////////////////////////////////////////
//
// TACRPageController
//
////////////////////////////////////////////////////////////////////////////////

  TACRPageController = class(TObject)
   private
    LPage:  TACRPage;
   protected
     procedure SetPageNo(Value: TACRPageNo);
     function GetPageNo: TACRPageNo;
     function GetPageSize: Integer;
     procedure SetPageBuffer(Value: TACRPageBuffer);
     function GetPageBuffer: TACRPageBuffer;
     procedure SetOwnBuffer(Value: Boolean);
     function GetOwnBuffer: Boolean;
     procedure SetIsDirty(Value: Boolean);
     function GetIsDirty: Boolean;
     procedure SetUseCount(Value: Integer);
     function GetUseCount: Integer;
     function GetPageManager: TACRPageManager;
   public
     constructor Create(Page: TACRPage);
     procedure EnlargePageBuffer(NewSize: Integer);

     property Page: TACRPage read LPage;
     property PageNo: TACRPageNo read GetPageNo write SetPageNo;
     property PageSize: Integer read GetPageSize;
     property PageBuffer: TACRPageBuffer read GetPageBuffer write SetPageBuffer;
     property OwnBuffer: Boolean read GetOwnBuffer write SetOwnBuffer;
     property IsDirty: Boolean read GetIsDirty write SetIsDirty;
     property UseCount: Integer read GetUseCount write SetUseCount;
     property PageManager: TACRPageManager read GetPageManager;
  end;// TACRPageController


////////////////////////////////////////////////////////////////////////////////
//
// TACRPageManager
//
////////////////////////////////////////////////////////////////////////////////

  TACRPageManager = class (TObject)
   private
     FCSect:               TRTLCriticalSection;
     FCacheCapacity:       Integer;
     FSharedPages:         TList;  // pages shared between all sessions
     FSessions:            array of TACRSessionPageInfo;

   protected
     FPageSize:            Word;
     FPageHeaderSize:      Word;
     FPageDataSize:        Word;
     FPageCount:           TACRPageNo;
     FExclusive:           Boolean;

     procedure Lock;
     procedure Unlock;
     function AddNewSession(SessionID: TACRSessionID): Integer;
     function FindSession(SessionID: TACRSessionID; var SessionNo: Integer): Boolean;
     function FindPage(Pages: TList; PageNo: TACRPageNo; var Page: TACRPage): Boolean;
     procedure RemovePageIfExists(Pages: TList; PageNo: TACRPageNo);

     function GetPageCount: TACRPageNo; virtual;
   public
     function UpdateState: Cardinal; virtual;
     function GetState: Cardinal; virtual;
     procedure LoadFromStream(Stream: TStream); virtual;
     procedure SaveToStream(Stream: TStream); virtual;
     procedure InternalAddPage(aPage: TACRPage); virtual; abstract;
     procedure InternalRemovePage(PageNo: TACRPageNo); virtual; abstract;
     procedure InternalReadPage(aPage: TACRPage); virtual; abstract;
     procedure InternalWritePage(aPage: TACRPage); virtual; abstract;
     constructor Create;
     destructor Destroy; override;
     function AddPage(SessionID: TACRSessionID): TACRPage;
     procedure RemovePage(SessionID: TACRSessionID; PageNo: TACRPageNo); overload;
     function GetPage(SessionID:                  TACRSessionID;
                      PageNo:                     TACRPageNo;
                      IsSynchronizationRequired:  TACRIsPageSynchronizationRequired = nil
                      ): TACRPage;
     procedure PutPage(aPage: TACRPage);
     procedure ApplyChanges(SessionID: TACRSessionID);
     procedure CancelChanges(SessionID: TACRSessionID);
     procedure FlushFileBuffers; virtual;

     property PageSize: Word read FPageSize;
     property PageHeaderSize: Word read FPageHeaderSize;
     property PageDataSize: Word read FPageDataSize;
     property PageCount: TACRPageNo read GetPageCount;
     property Exclusive: Boolean read FExclusive write FExclusive;
  end; // TACRPageManager


////////////////////////////////////////////////////////////////////////////////
//
// TACRMemoryPageManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRMemoryPageManager = class (TACRPageManager)
   protected
     FAllocatedPageMap:    TACRBitsArray;
     FAllocatedPageCount:  TACRPageNo;
     FAllocateBy:          Integer;
     LastAllocatedPageNo:  TACRPageNo;
     FDataPtrs:            array of PChar;

   public
     procedure LoadFromStream(Stream: TStream); override;
     procedure SaveToStream(Stream: TStream); override;
     procedure InternalAddPage(aPage: TACRPage); override;
     procedure InternalRemovePage(PageNo: TACRPageNo); override;
     procedure InternalReadPage(aPage: TACRPage); override;
     procedure InternalWritePage(aPage: TACRPage); override;

     constructor Create;
     destructor Destroy; override;
  end; // TACRMemoryPageManager


////////////////////////////////////////////////////////////////////////////////
//
// TACRTemporaryPageManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRTemporaryPageManager = class (TACRPageManager)
   protected
     FAllocatedPageMap:    TACRBitsArray;
     FAllocatedPageCount:  TACRPageNo;
     FTempPageFile:        TACRFileStream;
     FMaxMemoryPageCount:  TACRPageNo;
     FMemoryPageManager:   TACRPageManager;

   public
     procedure InternalAddPage(aPage: TACRPage); override;
     procedure InternalRemovePage(PageNo: TACRPageNo); override;
     procedure InternalReadPage(aPage: TACRPage); override;
     procedure InternalWritePage(aPage: TACRPage); override;

     constructor Create;
     destructor Destroy; override;
     procedure LoadFromStream(Stream: TStream); override;
     procedure SaveToStream(Stream: TStream); override;
  end; // TACRMemoryPageManager



implementation

uses ACRBTree, ACRBaseEngine;

////////////////////////////////////////////////////////////////////////////////
//
// TACRPage
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// GetPageHEader
//------------------------------------------------------------------------------
function TACRPage.GetPageHeader: PACRDiskPageHeader;
begin
  Result := PACRDiskPageHeader(FPageBuffer);
end; // GetPageHeader


//------------------------------------------------------------------------------
// GetPageData
//------------------------------------------------------------------------------
function TACRPage.GetPageHeaderSize: Integer;
begin
  Result := FPageHeaderSize;
end;//GetPageHeaderSize


//------------------------------------------------------------------------------
// GetPageData
//------------------------------------------------------------------------------
function TACRPage.GetPageData: TACRPageBuffer;
begin
  Result := TACRPageBuffer(FPageBuffer + PageHeaderSize);
end;//GetPageData


//------------------------------------------------------------------------------
// GetPageDataSize
//------------------------------------------------------------------------------
function TACRPage.GetPageDataSize: Integer;
begin
  Result := PageSize - PageHeaderSize;
end;//GetPageDataSize


//------------------------------------------------------------------------------
// GetPageSize
//------------------------------------------------------------------------------
function TACRPage.GetPageSize: Integer;
begin
  Result := FPageSize;
end;// GetPageSize


//------------------------------------------------------------------------------
// InitHeader
//------------------------------------------------------------------------------
procedure TACRPage.InitHeader;
begin
  PageHeader.Signature := ACRDiskPageSignature;
  PageHeader.State := 0;
  PageHeader.PageType := ACRPageTypeIDUnknown;
  PageHeader.NextPageNo := INVALID_PAGE_NO;
  PageHeader.CRC32 := 0;
  PageHeader.CRCType := 0;
  PageHeader.HashType := 0;
  PageHeader.Cipherype := 0;
  PageHeader.MACType := 0;
  PageHeader.ObjectID := INVALID_OBJECT_ID;
  PageHeader.RecordID.PageNo := INVALID_PAGE_NO;
  PageHeader.RecordID.PageItemNo := 0;
end;//InitHeader


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRPage.Create(PageManager: TACRPageManager);
begin
  LPageManager := PageManager;
  FPageBuffer := nil;
  FPageNo := INVALID_PAGE_NO;
  FOwnBuffer := True;
  FIsDirty := False;
  FUseCount := 0;
  FPageHeaderSize := PageManager.PageHeaderSize;
  FPageSize := PageManager.PageSize;
end;// Create


//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------
destructor TACRPage.Destroy;
begin
  if (FOwnBuffer) then
    FreeAndNilPageBuffer;
end;// Destroy


//------------------------------------------------------------------------------
// ClearPageBuffer
//------------------------------------------------------------------------------
procedure TACRPage.ClearPageBuffer;
begin
  FillChar(FPageBuffer^,PageSize,00);
end;// ClearPageBuffer


//------------------------------------------------------------------------------
// AllocPageBuffer
//------------------------------------------------------------------------------
procedure TACRPage.AllocPageBuffer;
begin
  FPageBuffer := MemoryManager.AllocMem(PageSize);
  InitHeader;
end;// AllocPageBuffer


//------------------------------------------------------------------------------
// FreeAndNilPageBuffer
//------------------------------------------------------------------------------
procedure TACRPage.FreeAndNilPageBuffer;
begin
  if (FPageBuffer <> nil) then
   MemoryManager.FreeAndNilMem(FPageBuffer);
end;// FreeAndNilPageBuffer


//------------------------------------------------------------------------------
// EnlargePageBuffer
//------------------------------------------------------------------------------
procedure TACRPage.EnlargePageBuffer(NewSize: Integer);
var
  NewBuffer: PChar;
begin
  if (FPageBuffer = nil) then
   raise EACRException.Create(20038, ErrorAInvalidPageBuffer);
  if (NewSize < PageSize) then
   raise EACRException.Create(20039, ErrorAInvalidPageModification);

  if (FOwnBuffer) then
    MemoryManager.ReallocMem(FPageBuffer, NewSize)
  else
   begin
     NewBuffer := MemoryManager.AllocMem(NewSize);
     Move(FPageBuffer^, NewBuffer^, PageSize);
     FPageBuffer := NewBuffer;
     FOwnBuffer := True;
   end;
end;// EnlargePageBuffer


//------------------------------------------------------------------------------
// Synchronize
//------------------------------------------------------------------------------
procedure TACRPage.Synchronize;
var NewState: Cardinal;
begin
  NewState := LPageManager.GetState;
  if (NewState <> FState) then
   begin
    LPageManager.InternalReadPage(Self);
    FState := NewState;
   end;
end;// Synchronize



////////////////////////////////////////////////////////////////////////////////
//
// TACRPageController
//
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// SetPageNo
//------------------------------------------------------------------------------
procedure TACRPageController.SetPageNo(Value: TACRPageNo);
begin
  LPage.PageNo := Value;
end;// SetPageNo


//------------------------------------------------------------------------------
// GetPageNo
//------------------------------------------------------------------------------
function TACRPageController.GetPageNo: TACRPageNo;
begin
  Result := LPage.PageNo;
end;// GetPageNo


//------------------------------------------------------------------------------
// GetPageSize
//------------------------------------------------------------------------------
function TACRPageController.GetPageSize: Integer;
begin
  Result := LPage.PageSize;
end;// GetPageSize


//------------------------------------------------------------------------------
// SetPageBuffer
//------------------------------------------------------------------------------
procedure TACRPageController.SetPageBuffer(Value: TACRPageBuffer);
begin
  LPage.PageBuffer := Value;
end;// SetPageBuffer


//------------------------------------------------------------------------------
// GetPageBuffer
//------------------------------------------------------------------------------
function TACRPageController.GetPageBuffer: TACRPageBuffer;
begin
  Result := LPage.PageBuffer;
end;// GetPageBuffer


//------------------------------------------------------------------------------
// SetOwnBuffer
//------------------------------------------------------------------------------
procedure TACRPageController.SetOwnBuffer(Value: Boolean);
begin
  LPage.OwnBuffer := Value;
end;// SetOwnBuffer


//------------------------------------------------------------------------------
// GetOwnBuffer
//------------------------------------------------------------------------------
function TACRPageController.GetOwnBuffer: Boolean;
begin
  Result := LPage.OwnBuffer;
end;// GetOwnBuffer


//------------------------------------------------------------------------------
// SetIsDirty
//------------------------------------------------------------------------------
procedure TACRPageController.SetIsDirty(Value: Boolean);
begin
  LPage.IsDirty := Value;
end;// SetIsDirty


//------------------------------------------------------------------------------
// GetIsDirty
//------------------------------------------------------------------------------
function TACRPageController.GetIsDirty: Boolean;
begin
  Result := LPage.IsDirty;
end;// GetIsDirty


//------------------------------------------------------------------------------
// SetUseCount
//------------------------------------------------------------------------------
procedure TACRPageController.SetUseCount(Value: Integer);
begin
  LPage.UseCount := Value;
end;// SetUseCount


//------------------------------------------------------------------------------
// GetUseCount
//------------------------------------------------------------------------------
function TACRPageController.GetUseCount: Integer;
begin
  Result := LPage.UseCount;
end;// GetUseCount


//------------------------------------------------------------------------------
// GetPageManager
//------------------------------------------------------------------------------
function TACRPageController.GetPageManager: TACRPageManager;
begin
  Result := LPage.PageManager;
end;// GetPageManager


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRPageController.Create(Page: TACRPage);
begin
  LPage := Page;
end;// Create


//------------------------------------------------------------------------------
// EnlargePageBuffer
//------------------------------------------------------------------------------
procedure TACRPageController.EnlargePageBuffer(NewSize: Integer);
begin
  LPage.EnlargePageBuffer(NewSize);
end;// EnlargePageBuffer



////////////////////////////////////////////////////////////////////////////////
//
// TACRPageManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// lock
//------------------------------------------------------------------------------
procedure TACRPageManager.Lock;
begin
  EnterCriticalSection(FCSect);
end;// Lock


//------------------------------------------------------------------------------
// unlock
//------------------------------------------------------------------------------
procedure TACRPageManager.Unlock;
begin
  LeaveCriticalSection(FCSect);
end;// Unlock


//------------------------------------------------------------------------------
// AddNewSession
//------------------------------------------------------------------------------
function TACRPageManager.AddNewSession(SessionID: TACRSessionID): Integer;
var
  n: Integer;
begin
  n := Length(FSessions);
  SetLength(FSessions, n+1);

  FSessions[n].SessionID := SessionID;
  FSessions[n].DirtyPages := TList.Create;
  FSessions[n].DirtyPages.Capacity := FCacheCapacity;
  FSessions[n].AddedPageNumbers := TACRPagesArray.Create;
  FSessions[n].RemovedPageNumbers := TACRPagesArray.Create;
  Result := n;
end;// AddNewSession


//------------------------------------------------------------------------------
// FindSession
//------------------------------------------------------------------------------
function TACRPageManager.FindSession(SessionID: TACRSessionID; var SessionNo: Integer): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Length(FSessions)-1 do
    if (FSessions[i].SessionID = SessionID) then
     begin
       Result := True;
       SessionNo := i;
       break; 
     end; 
end;// FindSession
	
 
//------------------------------------------------------------------------------ 
// FindPage
//------------------------------------------------------------------------------
function TACRPageManager.FindPage(Pages: TList; PageNo: TACRPageNo; var Page: TACRPage): Boolean; 
var
  i: Integer; 
begin
  Page := nil; 
  Result := False;
  if (Pages <> nil) then 
    for i := 0 to Pages.Count-1 do
     if (TACRPage(Pages[i]).PageNo = PageNo) then 
      begin 
        Page := Pages[i];
        Result := True; 
        break; 
      end; 
end;// FindPage 
 
	
//------------------------------------------------------------------------------
// return page count
//------------------------------------------------------------------------------
function TACRPageManager.GetPageCount: TACRPageNo;
begin 
  Result := FPageCount;
end; // GetPageCount 
	
//------------------------------------------------------------------------------ 
// RemovePageIfExists 
//------------------------------------------------------------------------------
procedure TACRPageManager.RemovePageIfExists(Pages: TList; PageNo: TACRPageNo); 
var
  Page: TACRPage; 
begin 
  if (FindPage(Pages, PageNo, Page)) then
   begin
     Pages.Remove(Page); 
     Page.Free; 
   end; 
end;// RemovePageIfExists 
 
 
//------------------------------------------------------------------------------ 
// update state and return new state
//------------------------------------------------------------------------------
function TACRPageManager.UpdateState: Cardinal; 
begin
 Result := 0;
end; // UpdateState 
	
	
//------------------------------------------------------------------------------ 
// return state 
//------------------------------------------------------------------------------ 
function TACRPageManager.GetState: Cardinal;
begin
 Result := 0; 
end; // GetState
	
 
//------------------------------------------------------------------------------ 
// LoadFromStream
//------------------------------------------------------------------------------
procedure TACRPageManager.LoadFromStream(Stream: TStream);
begin 
; 
end; // LoadFromStream 
 
	
//------------------------------------------------------------------------------ 
// SaveToStream 
//------------------------------------------------------------------------------ 
procedure TACRPageManager.SaveToStream(Stream: TStream);
begin
;
end; // SaveToStream 
	
	
//------------------------------------------------------------------------------
// constructor 
//------------------------------------------------------------------------------ 
constructor TACRPageManager.Create;
begin
  InitializeCriticalSection(FCSect);
  FCacheCapacity := ACRDefaultPageMangerCacheCapactiy;
  FSharedPages := TList.Create;
  FSharedPages.Capacity := FCacheCapacity;
  FExclusive := False; 
end;// Create
	
 
//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------ 
destructor TACRPageManager.Destroy; 
var 
  i: Integer; 
begin
  for i := 0 to Length(FSessions)-1 do 
   begin 
    if (FSessions[i].DirtyPages.Count > 0) then
      raise EACRException.Create(20097, ErrorAUncommitedPages, [Self.ClassName]);
    FSessions[i].DirtyPages.Free;
    FSessions[i].AddedPageNumbers.Free;
    FSessions[i].RemovedPageNumbers.Free; 
   end; 
  for i := 0 to FSharedPages.Count-1 do
    TACRPage(FSharedPages[i]).Free;
 
  FSharedPages.Free; 
  DeleteCriticalSection(FCSect); 
end;// Destroy
	
	
//------------------------------------------------------------------------------
// AddPage
//------------------------------------------------------------------------------ 
function TACRPageManager.AddPage(SessionID: TACRSessionID): TACRPage;
var 
  SessionNo: Integer;
begin
  Lock;
  try
    if (not FindSession(SessionID, SessionNo)) then
      SessionNo := AddNewSession(SessionID);
    Result := TACRPage.Create(Self);
    InternalAddPage(Result);
    if (Result.PageNo = INVALID_PAGE_NO) then
      raise EACRException.Create(20064, ErrorAInvalidPageNo);
    Result.UseCount := Result.UseCount + 1;
    Result.IsDirty := True;
    Result.ParentList := FSessions[SessionNo].DirtyPages;
    Result.SessionID := SessionID;
    FSessions[SessionNo].DirtyPages.Add(Result);
    FSessions[SessionNo].AddedPageNumbers.Append(Result.PageNo);
  finally
    Unlock;
  end;
end;// AddPage


//------------------------------------------------------------------------------
// RemovePage
//------------------------------------------------------------------------------
procedure TACRPageManager.RemovePage(SessionID: TACRSessionID; PageNo: TACRPageNo);
var
  SessionNo: Integer;
begin
  Lock;
  try
    if (not FindSession(SessionID, SessionNo)) then
      SessionNo := AddNewSession(SessionID);
    FSessions[SessionNo].RemovedPageNumbers.Append(PageNo);
  finally
    Unlock;
  end;
end;// RemovePage


//------------------------------------------------------------------------------
// GetPage
//------------------------------------------------------------------------------
function TACRPageManager.GetPage(
                    SessionID:                  TACRSessionID;
                    PageNo:                     TACRPageNo;
                    IsSynchronizationRequired:  TACRIsPageSynchronizationRequired
                                 ): TACRPage;
var
  SessionNo: Integer;
begin
  Lock;
  try
    if (not FindSession(SessionID, SessionNo)) then
      SessionNo := AddNewSession(SessionID);
    if (not (FindPage(FSessions[SessionNo].DirtyPages, PageNo, Result))) then
     begin
      if (FindPage(FSharedPages, PageNo, Result)) then
       begin
        if (not FExclusive) then
         begin
           if (@IsSynchronizationRequired = nil) then
            Result.Synchronize
           else
            if (IsSynchronizationRequired(Result.State)) then
              Result.Synchronize;
         end;
       end
      else
       begin
        Result := TACRPage.Create(Self);
        Result.PageNo := PageNo;
        Result.SessionID := SessionID;
        FSharedPages.Add(Result);
        InternalReadPage(Result);
        Result.State := GetState;
       end;
      Result.ParentList := FSharedPages;
     end;
    // re-read full page, if was read only header?
//    if ((Result.UseHeaderOnly) and (not ReadHeaderOnly)) then
//     InternalReadPage(Result, ReadHeaderOnly);

    if ((Result.UseCount > 0) and (Result.SessionID <> SessionID)) then
      raise EACRException.Create(20090, ErrorAPageIsUsedByOtherSession);
    Result.SessionID := SessionID;
    Result.UseCount := TACRPage(Result).UseCount + 1;
  finally
    Unlock;
  end;
end;// GetPage


//------------------------------------------------------------------------------
// PutPage
//------------------------------------------------------------------------------
procedure TACRPageManager.PutPage(aPage: TACRPage);
var
  i: Integer;
  SessionNo: Integer;
begin
 if (aPage.PageNo = INVALID_PAGE_NO) then
  raise EACRException.Create(20065, ErrorAInvalidPageNo);
 Lock;
  try
     begin
      if (aPage.ParentList.IndexOf(aPage) = -1) then
       raise EACRException.Create(20040, ErrorAUsedPageWasReleased);
      aPage.UseCount := aPage.UseCount - 1;
      if (aPage.UseCount < 0) then raise
       EACRException.Create(10665,ErrorLInvalidUseCount,
        [aPage.UseCount,aPage.PageNo]);
      // apply LRU cache (last items are more recently used) and dirty page move
      if ((aPage.UseCount = 0) and (aPage.ParentList = FSharedPages)) then
       begin
        FSharedPages.Remove(aPage);
        if (aPage.IsDirty) then
         begin
           if (not FindSession(aPage.SessionID, SessionNo)) then
             SessionNo := AddNewSession(aPage.SessionID);
           FSessions[SessionNo].DirtyPages.Add(aPage);
           aPage.ParentList := FSessions[SessionNo].DirtyPages;
         end
        else
          FSharedPages.Add(aPage);
       end;
      if (aPage.ParentList = FSharedPages) then
        // if cache overflow remove first unused pages
        if (FSharedPages.Count > FCacheCapacity) then
         begin
           i := 0;
           while (i < FSharedPages.Count) do
            if (TACRPage(FSharedPages[i]).UseCount = 0) then
             begin
               TACRPage(FSharedPages[i]).Free;
               FSharedPages.Remove(FSharedPages[i]);
               if (FSharedPages.Count <= FCacheCapacity) then
                break;
             end
            else
             Inc(i);
         end;
     end;
  finally
    Unlock;
  end;
end;// PutPage


//------------------------------------------------------------------------------
// ApplyChanges
//------------------------------------------------------------------------------
procedure TACRPageManager.ApplyChanges(SessionID: TACRSessionID);
var
  SessionNo: Integer;
  i:         Integer;
  NewState:  Cardinal;
  aPage:     TACRPage;
  PageNo:    TACRPageNo;
begin
  Lock;
  try
    if (FindSession(SessionID, SessionNo)) then
      begin
        // save dirty pages
        NewState := UpdateState;
        for i := 0 to FSessions[SessionNo].DirtyPages.Count-1 do
         begin
           aPage := TACRPage(FSessions[SessionNo].DirtyPages[i]);
           if (not aPage.IsDirty) then
             raise EACRException.Create(10822, ErrorLPageIsNotDirty, [aPage.PageNo]);
           if (aPage.UseCount > 0) then
             raise EACRException.Create(10823, ErrorLCannotRemoveUsedPage, [aPage.PageNo,aPage.UseCount]);
           InternalWritePage(aPage);
           aPage.IsDirty := False;
           aPage.State := NewState;
           // move dirty page to shared pages list
           FSharedPages.Add(aPage);
         end;
        FSessions[SessionNo].DirtyPages.Clear;

        // free removed pages
        for i := 0 to FSessions[SessionNo].RemovedPageNumbers.ItemCount-1 do
         begin
           PageNo := FSessions[SessionNo].RemovedPageNumbers.Items[i];
           InternalRemovePage(PageNo);
           RemovePageIfExists(FSessions[SessionNo].DirtyPages,PageNo);
           RemovePageIfExists(FSharedPages,PageNo);
         end;

        FSessions[SessionNo].RemovedPageNumbers.SetSize(0);
        FSessions[SessionNo].AddedPageNumbers.SetSize(0);
      end;
  finally
    Unlock;
  end;
end;// ApplyChanges


//------------------------------------------------------------------------------
// CancelChanges
//------------------------------------------------------------------------------
procedure TACRPageManager.CancelChanges(SessionID: TACRSessionID);
var
  SessionNo: Integer;
  i:         Integer;
  aPage:     TACRPage;
begin
  Lock;
  try
    if (FindSession(SessionID, SessionNo)) then
      begin
        // remove added pages
        for i := 0 to FSessions[SessionNo].AddedPageNumbers.ItemCount-1 do
         InternalRemovePage(FSessions[SessionNo].AddedPageNumbers.Items[i]);
        FSessions[SessionNo].AddedPageNumbers.SetSize(0);

        // free all dirty pages
        while (FSessions[SessionNo].DirtyPages.Count > 0) do
         begin
           aPage := TACRPage(FSessions[SessionNo].DirtyPages[0]);
           if (not aPage.IsDirty) then
             raise EACRException.Create(10820, ErrorLPageIsNotDirty, [aPage.PageNo]);
           if (aPage.UseCount > 0) then
             raise EACRException.Create(10821, ErrorLCannotRemoveUsedPage, [aPage.PageNo,aPage.UseCount]);
           RemovePageIfExists(FSessions[SessionNo].DirtyPages,aPage.PageNo);
         end;

        // do not delete removed pages
        FSessions[SessionNo].RemovedPageNumbers.SetSize(0);
      end;
  finally
    Unlock;
  end;
end;// CancelChanges


//------------------------------------------------------------------------------
// flush file buffer
//------------------------------------------------------------------------------
procedure TACRPageManager.FlushFileBuffers;
begin
;
end; //


////////////////////////////////////////////////////////////////////////////////
//
// TACRMemoryPageManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRMemoryPageManager.LoadFromStream(Stream: TStream);
var i: Integer;
begin
  LoadDataFromStream(FAllocatedPageCount,Sizeof(FAllocatedPageCount),Stream,10369);
  LoadDataFromStream(LastAllocatedPageNo,Sizeof(LastAllocatedPageNo),Stream,10370);
  LoadDataFromStream(FAllocateBy,Sizeof(FAllocateBy),Stream,10371);
  LoadDataFromStream(FPageSize,Sizeof(FPageSize),Stream,10372);
  LoadDataFromStream(FPageHeaderSize,Sizeof(FPageHeaderSize),Stream,10373);
  LoadDataFromStream(FPageCount,Sizeof(FPageCount),Stream,10374);
  FAllocatedPageMap.LoadFromStream(Stream);
  SetLength(FDataPtrs,FAllocatedPageCount);
  for i := 0 to Integer(FAllocatedPageCount) - 1 do
   begin
     FDataPtrs[i] := MemoryManager.AllocMem(FPageSize);
     LoadDataFromStream(FDataPtrs[i]^,FPageSize,Stream,10375);
   end;
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRMemoryPageManager.SaveToStream(Stream: TStream);
var i: Integer;
begin
  SaveDataToStream(FAllocatedPageCount,Sizeof(FAllocatedPageCount),Stream,10362);
  SaveDataToStream(LastAllocatedPageNo,Sizeof(LastAllocatedPageNo),Stream,10363);
  SaveDataToStream(FAllocateBy,Sizeof(FAllocateBy),Stream,10364);
  SaveDataToStream(FPageSize,Sizeof(FPageSize),Stream,10365);
  SaveDataToStream(FPageHeaderSize,Sizeof(FPageHeaderSize),Stream,10366);
  SaveDataToStream(FPageCount,Sizeof(FPageCount),Stream,10367);
  FAllocatedPageMap.SaveToStream(Stream);
  for i := 0 to Integer(FAllocatedPageCount) - 1 do
    SaveDataToStream(FDataPtrs[i]^,FPageSize,Stream,10368);
end; // SaveToStream


//------------------------------------------------------------------------------
// InternalAddPage
//------------------------------------------------------------------------------
procedure TACRMemoryPageManager.InternalAddPage(aPage: TACRPage);
var
  PageNo:    Integer;
  OldLength: Integer;
  i:         Integer;
begin
  if (not FAllocatedPageMap.Find(False, PageNo)) then
   begin
    if (FAllocatedPageCount <> 0) then
     PageNo := LastAllocatedPageNo+1
    else
     PageNo := 0;
    if (PageNo >= FAllocatedPageCount) then
      begin
        Inc(FAllocatedPageCount, FAllocateBy);
        FAllocatedPageMap.Size := FAllocatedPageCount;
      end;
   end;
  Inc(FPageCount);
  if (PageNo < FPageCount -1) then
   if (FAllocatedPageMap.GetBit(PageNo)) then
    raise EACRException.Create(20037, ErrorANotReleasedPageIsAllocated);
  FAllocatedPageMap.SetBit(PageNo, True);

  if (Length(FDataPtrs) <= PageNo) then
   begin
    OldLength := Length(FDataPtrs);
    SetLength(FDataPtrs, max(PageNo+1, OldLength+FAllocateBy));
    for i := OldLength to Length(FDataPtrs)-1 do
     FDataPtrs[i] := MemoryManager.AllocMem(FPageSize);
   end;
  if (PageNo > LastAllocatedPageNo) then
    LastAllocatedPageNo := PageNo;

  aPage.PageNo := PageNo;
  aPage.OwnBuffer := False;
  aPage.PageBuffer := FDataPtrs[PageNo];
end;// InternalAddPage


//------------------------------------------------------------------------------
// InternalRemovePage
//------------------------------------------------------------------------------
procedure TACRMemoryPageManager.InternalRemovePage(PageNo: TACRPageNo);
begin
  Dec(FPageCount);
  FAllocatedPageMap.SetBit(PageNo, False);
end;// InternalRemovePage


//------------------------------------------------------------------------------
// InternalReadPage
//------------------------------------------------------------------------------
procedure TACRMemoryPageManager.InternalReadPage(aPage: TACRPage);
begin
  if (not FAllocatedPageMap.GetBit(aPage.PageNo)) then
   raise EACRException.Create(10392, ErrorANotReleasedPageIsAllocated);
  if (aPage.OwnBuffer) then
   aPage.FreeAndNilPageBuffer;
  aPage.OwnBuffer := False;
  aPage.PageBuffer := FDataPtrs[aPage.PageNo];
end;// InternalReadPage


//------------------------------------------------------------------------------
// InternalWritePage
//------------------------------------------------------------------------------
procedure TACRMemoryPageManager.InternalWritePage(aPage: TACRPage);
begin
  if (not FAllocatedPageMap.GetBit(aPage.PageNo)) then
   raise EACRException.Create(10393, ErrorANotReleasedPageIsAllocated);
  if (aPage.OwnBuffer) then
   Move(aPage.PageBuffer^, FDataPtrs[aPage.PageNo]^, FPageSize);
end;// InternalWritePage


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRMemoryPageManager.Create;
begin
  inherited Create;
  FAllocatedPageMap := TACRBitsArray.Create;
  SetLength(FDataPtrs, 0);
  FAllocatedPageCount := 0;
  FPageCount := 0;
  FPageHeaderSize := 0;
  FPageSize := ACRDefaultMemoryPageSize;
  FAllocateBy := 50;
end;// Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRMemoryPageManager.Destroy;
var
  i: Integer;
begin
  if (FPageCount > 0) then
   raise EACRException.Create(20023, ErrorAIndexPagesNotReleased);
  for i := 0 to Length(FDataPtrs)-1 do
   MemoryManager.FreeAndNilMem(FDataPtrs[i]);
  FAllocatedPageMap.Free;
  inherited Destroy;
end;// Destroy



////////////////////////////////////////////////////////////////////////////////
//
// TACRTemporaryPageManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// InternalAddPage
//------------------------------------------------------------------------------
procedure TACRTemporaryPageManager.InternalAddPage(aPage: TACRPage);
var
  PageNo: Integer;
begin
  PageNo := FAllocatedPageCount;
  if (PageNo < FMaxMemoryPageCount) then
   begin
    FMemoryPageManager.InternalAddPage(aPage);
    if (aPage.PageNo >= FAllocatedPageCount) then
     begin
      FAllocatedPageCount := aPage.PageNo +1;
      FAllocatedPageMap.Size := FAllocatedPageCount;
     end;
    FAllocatedPageMap.SetBit(aPage.PageNo, True);
    Inc(FPageCount);
   end
  else
   begin
    Inc(FAllocatedPageCount);
    FAllocatedPageMap.Size := FAllocatedPageCount;
    Inc(FPageCount);
    FAllocatedPageMap.SetBit(PageNo, True);
    if (FTempPageFile = nil) then
     FTempPageFile := TACRFileStream.Create(GetTempFileName,fmCreate);
    FTempPageFile.Size := FAllocatedPageCount * FPageSize;
    aPage.PageNo := PageNo;
    if ((not aPage.OwnBuffer) or (aPage.PageBuffer = nil)) then
     begin
       aPage.AllocPageBuffer;
       aPage.OwnBuffer := True;
     end;
   end;
end;// InternalAddPage


//------------------------------------------------------------------------------
// InternalRemovePage
//------------------------------------------------------------------------------
procedure TACRTemporaryPageManager.InternalRemovePage(PageNo: TACRPageNo);
begin
  Dec(FPageCount);
  FAllocatedPageMap.SetBit(PageNo, False);
  if (PageNo < FMaxMemoryPageCount) then
   FMemoryPageManager.InternalRemovePage(PageNo)
end;// InternalRemovePage


//------------------------------------------------------------------------------
// InternalReadPage
//------------------------------------------------------------------------------
procedure TACRTemporaryPageManager.InternalReadPage(aPage: TACRPage);
begin
  if (not FAllocatedPageMap.GetBit(aPage.PageNo)) then
   raise EACRException.Create(10396, ErrorANotReleasedPageIsAllocated);
  if (aPage.PageNo < FMaxMemoryPageCount) then
   FMemoryPageManager.InternalReadPage(aPage)
  else
   begin
    if ((not aPage.OwnBuffer) or (aPage.PageBuffer = nil)) then
     begin
      aPage.AllocPageBuffer;
      aPage.OwnBuffer := True;
     end;
    FTempPageFile.Position := aPage.PageNo * FPageSize;
    FTempPageFile.ReadBuffer(aPage.PageBuffer^, FPageSize);
   end;
end;// InternalReadPage


//------------------------------------------------------------------------------
// InternalWritePage
//------------------------------------------------------------------------------
procedure TACRTemporaryPageManager.InternalWritePage(aPage: TACRPage);
begin
  if (not FAllocatedPageMap.GetBit(aPage.PageNo)) then
   raise EACRException.Create(10397, ErrorANotReleasedPageIsAllocated);
  if (aPage.PageNo < FMaxMemoryPageCount) then
   FMemoryPageManager.InternalWritePage(aPage)
  else
   begin
    if (FTempPageFile = nil) then
     FTempPageFile := TACRFileStream.Create(GetTempFileName,fmCreate);
    FTempPageFile.Position := aPage.PageNo * FPageSize;
    FTempPageFile.WriteBuffer(aPage.PageBuffer^, FPageSize);
   end;
end;// InternalWritePage


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRTemporaryPageManager.Create;
begin
  inherited Create;
  FAllocatedPageMap := TACRBitsArray.Create;
  FAllocatedPageCount := 0;
  FPageCount := 0;
  FPageHeaderSize := 0;
  FPageSize := ACRDefaultTemporaryPageSize;
  FMaxMemoryPageCount := ((MemoryManager.GetFreeMemorySize div 10) div
                          FPageSize);
  FTempPageFile := nil;
  FMemoryPageManager := TACRMemoryPageManager.Create;
end;// Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRTemporaryPageManager.Destroy;
var FileName: String;
begin
  if (FPageCount > 0) then
   raise EACRException.Create(20036, ErrorAIndexPagesNotReleased);
  FAllocatedPageMap.Free;
  if (FTempPageFile <> nil) then
   begin
    FileName := FTempPageFile.FileName;
    FTempPageFile.Free;
    SysUtils.DeleteFile(FileName);
   end;
  FMemoryPageManager.Free;
  inherited Destroy;
end;// Destroy


//------------------------------------------------------------------------------
// LoadFromStream
//------------------------------------------------------------------------------
procedure TACRTemporaryPageManager.LoadFromStream(Stream: TStream);
var Size: Int64;
begin
  FMemoryPageManager.LoadFromStream(Stream);
  FAllocatedPageMap.LoadFromStream(Stream);
  LoadDataFromStream(FAllocatedPageCount,Sizeof(FAllocatedPageCount),Stream,11202);
  LoadDataFromStream(FMaxMemoryPageCount,Sizeof(FMaxMemoryPageCount),Stream,11203);
  LoadDataFromStream(FPageSize,Sizeof(FPageSize),Stream,11204);
  LoadDataFromStream(FPageHeaderSize,Sizeof(FPageHeaderSize),Stream,11205);
  LoadDataFromStream(FPageCount,Sizeof(FPageCount),Stream,11206);
  LoadDataFromStream(Size,Sizeof(Size),Stream,11207);

  if (Size > 0) then
   begin
    if (FTempPageFile = nil) then
     FTempPageFile := TACRFileStream.Create(GetTempFileName,fmCreate)
    else
     FTempPageFile.Size := 0;
    FTempPageFile.LoadFromStreamWithPosition(Stream,Stream.Position,Size)
   end;
end; // LoadFromStream


//------------------------------------------------------------------------------
// SaveToStream
//------------------------------------------------------------------------------
procedure TACRTemporaryPageManager.SaveToStream(Stream: TStream);
var Size: Int64;
begin
  FMemoryPageManager.SaveToStream(Stream);
  FAllocatedPageMap.SaveToStream(Stream);
  SaveDataToStream(FAllocatedPageCount,Sizeof(FAllocatedPageCount),Stream,11196);
  SaveDataToStream(FMaxMemoryPageCount,Sizeof(FMaxMemoryPageCount),Stream,111197);
  SaveDataToStream(FPageSize,Sizeof(FPageSize),Stream,11198);
  SaveDataToStream(FPageHeaderSize,Sizeof(FPageHeaderSize),Stream,11199);
  SaveDataToStream(FPageCount,Sizeof(FPageCount),Stream,11200);
  if (FTempPageFile = nil) then
   Size := 0
  else
   Size := FTempPageFile.Size;
  SaveDataToStream(Size,Sizeof(Size),Stream,11201);
  if (Size > 0) then
   FTempPageFile.SaveToStream(Stream);
end; // SaveToStream


initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRPage> initialized');
{$ENDIF}

end.
