unit ACRRelationalAlgebra;

interface

{$I ACRVer.inc}
{$DEFINE NEW_JOIN_ALGORITHM}

uses
{$IFDEF MSWINDOWS}
  Windows,
{$ENDIF}
  Classes, DB, SysUtils, Math,

// Accuracer units

 {$IFDEF DEBUG_LOG}
 ACRDebug,
 {$ENDIF}
 ACRConst,
 ACRConverts,
 ACRExcept,
 ACRBase,
 ACRTypes,
 ACRVariant
 ;

type

 TACRFieldLink = record
  FieldName:                String;   // field name
  DisplayName:              String;   // result field name
  OriginalFieldName:        String;   // original filed name for joins
  TableName:                String;   // name of the source table of the field
  TableAlias:               String;   // alias of the source table of the field
  FieldType:                TACRAdvancedFieldType;
  FieldPrecision:           Integer;
  FieldSize:                Integer;
  BLOBCompressionAlgorithm: Byte;
  BLOBCompressionMode:      Byte;
  BLOBBlockSize:            Integer;
  AO:                       Pointer;  // was TACRAO type, CB4 bug fix
  Dataset:                  TDataset; // dataset
  IsHidden:                 Boolean;
  FieldNo:                  Integer;  // field number in AO or FieldNo
  IsExpression:             Boolean;  // Expression or field?
  IsAggregate:              Boolean;  // Expression is aggregate (contains agg. functions)?
  Expr:                     TObject;  // TACRExpression
 end;

 // fields (expressions) list in select
 TACRSelectListItem = record
  TableName:    String;   // 'table1.' | 't1'
  AllFields:    Boolean;  // 'table1.*' ?
  FieldName:    String;   // field1
  IsExpression: Boolean;  // field or expr?
  ValueExpr:    TObject;  // TACRExpression
  Pseudonym:    String;   // field1 as f1
 end;

 // array of fields
 TACRFields = class
 public
   Items:       array of TACRSelectListItem; // fields
   ItemCount:   Integer;                     // length
   // creates
   constructor Create;
   // adds item to the end
   procedure Append(var Item: TACRSelectListItem);
 end;


 // base class for relational algebra operations
 TACRAO = class (TObject)
 public
  FIsRootAO:                  Boolean;
  FIsAOTable:                 Boolean;
  FIsAOGroupBy:               Boolean;
  FFilterExpr:                TObject; // TACRExpression
  FTopRowCount:               Int64;
  FFirstRowNo:                Int64;
 protected
  FRequestLive:               Boolean;
  FResultInMemory:            Boolean; // for SELECT INTO optimization
  FResultTableName:           String; // for SELECT INTO optimization
  FResultDatabaseName:        String; // for SELECT INTO optimization
  FTableName:                 String;
  FTableAlias:                String;
  FIsMaterialized:            Boolean;
  FResultDataset:             TDataset; // result dataset
  FResultFieldsOrder:         TACRIntegerArray;
  FFieldCount:                Integer;
  FLeftAONull:                Boolean;
  FRightAONull:               Boolean;
  FDistinctApplied:           Boolean;
  FDistinctFields:            String;
  FResultIndexFieldsList:             TStringList;
  FResultIndexAscDescFieldsList:      TStringList;
  FResultIndexCaseInsFieldsList:      TStringList;
  FResultIndexName:                   String;
  FResultIndexByAllFields:            Boolean; // for Union childs
  FExpressionsExists:                 Boolean;
  FAggregateExpressionsExists:        Boolean;
  FActivateIndexAfterMaterialize:     Boolean;
  FAllFieldsNull:                     Boolean;
 protected
  FFieldLinks:                array of TACRFieldLink;
 public
  FLeftAO,FRightAO:           TACRAO;
 protected
 procedure InternalCreate(
                      LeftAO:     TACRAO = nil;
                      RightAO:    TACRAO = nil;
                      TableName:  String = '';
                      TableAlias: String = ''
                      );
  // navigating
  procedure InternalFirst; virtual;
  procedure InternalNext; virtual;
  function InternalGetEof: Boolean; virtual;
  function InternalGetRecordCount: Integer; virtual;
  procedure First; virtual;
  procedure Next; virtual;
  function GetEof: Boolean; virtual;
  function GetRecordCount: Integer; virtual;
  // sets names to FieldLinks list and renames duplicate names
  procedure SetFieldNames;
  // materialization routines
  procedure CreateIndexForMaterialize;
  procedure CreateTableForMaterializeFillFieldDefs(
                                            FieldList:  TStringList;
                                            AliasList:  TStringList
                                               );
  procedure CreateTableForMaterialize(
                                            FieldList:  TStringList;
                                            AliasList:  TStringList
                                      );
  procedure FillTableForMaterialize;
  procedure FinalizeMaterialize(
                                            FieldList:  TStringList;
                                            AliasList:  TStringList
                                );

  // materializes AO
  procedure DoMaterialize; virtual;
 public
  destructor Destroy; override;
  // gets all result records
  procedure Execute(IsRootAO: Boolean = False); virtual;

  // sets filter
  procedure SetFilter(FilterExpr: TObject);
  // for SELECT INTO optimization
  procedure SetResultTable(InMemory: Boolean; TableName: String; DatabaseName: String);
  // sets Top row count
  procedure SetTopRowCount(FirstRowNo, TopRowCount: Integer);
  // applies distinct
  procedure ApplyDistinct(DistinctFields: String);
  // sets projection for other TACRAO
  procedure SetResultFields(var FieldRefs: array of TACRSelectListItem;
                            bDistinct: Boolean); virtual;
  // mapping function - return number of found fields and found field No
  // also optionally unhides fields in AO
  function FieldExists(
                  FieldName, TableName: String;
                  Unhide:               Boolean;
                  FieldNumbers:         TACRIntegerArray = nil;
                  UnhideChildrenOnly:   Boolean = False
                      ): Integer; virtual;
  function GetFieldName(FieldNo: Integer; ApplyOrderBy: Boolean = False): string;
  function GetFieldNameByOrderByColumnNumber(ColumnNumber: Integer): String;
  // return value (copy or direct pointer)
  procedure GetFieldValue(
                        Value:      TACRVariant;
                        FieldNo:        Integer;
                        bCopy:          Boolean = False;
                        AccessToHidden: Boolean = False
                        );
  function GetFieldType(FieldNo: Integer): TACRAdvancedFieldType; overload;
  function GetFieldSize(FieldNo: Integer): Integer;
  function GetFieldPrecision(FieldNo: Integer): Integer;
  procedure LockTable; virtual;
  procedure UnlockTable; virtual;
  // sets index
 private
  procedure CreateResultIndexLists(ToClear: Boolean = True);
  procedure FreeResultIndexLists;
  procedure SetResultIndexByAllFields;
 public
  procedure SetIndex(IndexFieldNames, DescFields, CaseInsensitiveFields, IndexName: String;
                     ActivateIndexAfterMaterialize: Boolean);
  property IsMaterialized: Boolean read FIsMaterialized;
  property FieldCount: Integer read FFieldCount;
  property RecordCount: Integer read GetRecordCount;
  property ResultDataset: TDataset read FResultDataset;
  property Eof: Boolean read GetEof;
  property RequestLive: Boolean read FRequestLive;
  property ResultIndexByAllFields: Boolean read FResultIndexByAllFields
            write FResultIndexByAllFields; // for Union childs
 end;

 // table
 TACRAOTable = class (TACRAO)
 private
  FSourceDataset: TDataset;
  FTableLocked:   Boolean;
 protected
  // materializes AO
  procedure DoMaterialize; override;
 public
  constructor Create(
                      DatabaseName: string;
                      SessionName:  string;
                      TableName:    string;
                      TableAlias:   string;
                      bInMemory:    Boolean = false;
                      bRequestLive: Boolean = false;
                      Password:     String = ''
                      );
  destructor Destroy; override;
  // sets projection
  procedure SetResultFields(var FieldRefs: array of TACRSelectListItem;
                      bDistinct: Boolean); override;
  procedure LockTable; override;
  procedure UnlockTable; override;
 public
  property IsMaterialized;
  property FieldCount;
  property RecordCount;
  property ResultDataset;
  property Eof;
 end;


 // joins and dekart
 TACRAOJoin = class (TACRAO)
 private
  FDekart:    Boolean;
  FOuterJoin: Boolean;
  FInnerJoin: Boolean;
  FFirstTimeCalled:   Boolean; // true if Next called First time
  FEof:               Boolean; // Eof is set

  FJoinType:    TACRJoinType;
  FFields1:     TACRIntegerArray;
  FFields2:     TACRIntegerArray;

  // algorithm 2 (fast)
  FFieldNames1:                   TStringList;
  FFieldNames2:                   TStringList;
  FTempFilterExpr:                 TObject;
  FScanTable:                     TDataset;
  FFilterTable:                   TDataset;
  FScanLeft:                      Boolean;
  FConstList:                     TList; // const nodes for filter
  FSearchingCorrespondingRecords: Boolean; // ScanTable
  FScanningRestRecords:           Boolean; // for full outer join
  FFilterApplied:                 Boolean;
  FScannedRecords:                TACRRecordIDArray;
  FFilterTableEmpty:              Boolean;

  // algorithm 1
  // inner / outer joins
  FCompareResult:     TACRCompareResult;
  FBookmark:          Pointer;
  FEqualStarted:      Boolean; // true if equal values in both AO

 protected
  procedure FixResultFieldNames;

  // algorithm #1: scannng both tables with active indexes on join fields
  // records are called Equal if all their join attributes are equal
  procedure CompareRecords;
  {$IFNDEF NEW_JOIN_ALGORITHM}
  procedure Init1(LeftFields,RightFields: String);
  procedure InternalFirst1;
  procedure InternalNext1;
  function InternalGetEof1: Boolean;
  {$ENDIF}

  // algorithm #2: scannng first table with filtering second table
  function IsIndexExists(Table: TDataset; FieldNames: TStringList): Boolean;
  procedure ChooseScanTable;
  procedure BuildFilterExpression;
  procedure PrepareFilterForCurrentRecord;
  function ApplyFilterAndCheckIfRecordsFound: Boolean;
  procedure ClearFilter;
  procedure SetLeftNull;
  procedure SetRightNull;
  procedure SwitchToScanningRest;
  procedure StoreCurrentRecordID;
  {$IFDEF NEW_JOIN_ALGORITHM}
  procedure Init2;
  procedure InternalFirst2;
  procedure InternalNext2;
  function InternalGetEof2: Boolean;
  {$ENDIF}


  procedure InternalFirst; override;
  procedure InternalNext; override;
  function InternalGetEof: Boolean; override;
  function InternalGetRecordCount: Integer; override;
 public
  constructor Create(
                      LeftChild:  TACRAO;
                      RightChild: TACRAO;
                      JoinType:   TACRJoinType;
                      IsNatural:  Boolean = False;
                      FieldList1: TACRFields = nil; // join fields
                      FieldList2: TACRFields = nil // field1 = field2
                      );
  destructor Destroy; override;

 public
  property IsMaterialized;
  property FieldCount;
  property RecordCount;
  property ResultDataset;
  property Eof;
  property OuterJoin: Boolean read FOuterJoin;
  property JoinType: TACRJoinType read FJoinType;
 end; // TACRAOJoin


 TACRAOUnion = class (TACRAO)
 private
  FUnionType:         TACRUnionType;
  FFields1:           TACRIntegerArray;
  FFields2:           TACRIntegerArray;
  FFirstTimeCalled:   Boolean; // true if Next called First time
  FShowLeft:          Boolean; // if then leftAO records will be added otherwise right
  FSearchAO:          TACRAO;
  FScanAO:            TACRAO;
 protected
  // records are called Equal if all their join attributes are equal
  // return true if search ao has record like in scan ao
  function CompareRecords: Boolean;
  procedure SearchRecord;
  procedure ShowLeftAO;
  procedure ShowRightAO;
  procedure InternalFirst; override;
  procedure InternalNext; override;
  function InternalGetEof: Boolean; override;
  function InternalGetRecordCount: Integer; override;
 public
  constructor Create(
                      LeftChild:  TACRAO;
                      RightChild: TACRAO;
                      UnionType:   TACRUnionType;
                      IsCorresponding:  Boolean = False;
                      bDistinct: Boolean = True;
                      FieldList: TACRFields = nil // corresponding fields
                      );
  destructor Destroy; override;
 public
  property IsMaterialized;
  property FieldCount;
  property RecordCount;
  property ResultDataset;
  property Eof;
 end; // TACRAOUnion - union, intersect, except

 // not used any more!!!
 TACRAOTableExpr = class (TACRAO)

 protected
  procedure InternalFirst; override;
  procedure InternalNext; override;
  function InternalGetEof: Boolean; override;
  function InternalGetRecordCount: Integer; override;
 public
  constructor Create(
                     Child: TACRAO
                    );
 public
  property IsMaterialized;
  property FieldCount;
  property RecordCount;
  property ResultDataset;
  property Eof;
 end; // not used any more!!!

 // table expression
 TACRAOGroupBy = class (TACRAO)
 protected
  FAllFields:         Boolean;
  FFields:            TACRIntegerArray;
  FValues:            array of TACRVariant;
  FEof:               Boolean;
 protected
  // return true if new group is started from current record
  function CompareRecords: Boolean;
  procedure Init;
  procedure Accumulate;
  procedure ScanGroup;
  procedure InternalFirst; override;
  procedure InternalNext; override;
  function InternalGetEof: Boolean; override;
  function InternalGetRecordCount: Integer; override;
 public
  // sets projection
  procedure SetResultFields(var FieldRefs: array of TACRSelectListItem;
                      bDistinct: Boolean); override;
  constructor Create(
                     Child: TACRAO;
                     FieldList: TACRFields = nil // corresponding fields
                    );
  destructor Destroy; override;
 public
  property IsMaterialized;
  property FieldCount;
  property RecordCount;
  property ResultDataset;
  property Eof;
 end;

 // table expression
 TACRAOSQLTopRowCount = class (TACRAO)
 protected
  procedure InternalFirst; override;
  procedure InternalNext; override;
  function InternalGetEof: Boolean; override;
  function InternalGetRecordCount: Integer; override;
 public
  constructor Create(
                     Child: TACRAO
                    );
 public
  property IsMaterialized;
  property FieldCount;
  property RecordCount;
  property ResultDataset;
  property Eof;
 end;


 procedure ConvertIndexFieldNamesToLists(
                FieldNames, DescFields, CaseInsFields:  String;
                FieldList, AscDescList, CaseInsList:    TStringList
                );
 procedure ConvertListsToIndexFieldNames(
                var FieldNames:     String;
                var DescNames:     String;
                var CaseInsNames:  String;
                FieldList, AscDescList, CaseInsList:    TStringList
                );

implementation

uses ACRMain,ACRExpressions,ACRBaseEngine;


////////////////////////////////////////////////////////////////////////////////
//
// TACRFields
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// creates
//------------------------------------------------------------------------------
constructor TACRFields.Create;
begin
 ItemCount := 0;
end;// Create


//------------------------------------------------------------------------------
// add item to the end
//------------------------------------------------------------------------------
procedure TACRFields.Append(var Item: TACRSelectListItem);
begin
  Inc(ItemCount);
  SetLength(Items, ItemCount);
  Items[ItemCount-1] := Item;
end;// Append


////////////////////////////////////////////////////////////////////////////////
//
// TACRAO
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
procedure TACRAO.InternalCreate(
                      LeftAO:     TACRAO = nil;
                      RightAO:    TACRAO = nil;
                      TableName:  String = '';
                      TableAlias: String = ''
                      );
begin
  FResultDataset := TACRTable.Create(nil);
  FActivateIndexAfterMaterialize := False;
  FResultIndexByAllFields := False;
  FAllFieldsNull := False;
  FResultIndexName := '';
  FRequestLive := False;
  FIsRootAO := False;
  FIsAOTable := False;
  FIsAOGroupBy := False;
  FTableName := TableName;
  FTableAlias := TableAlias;
  FLeftAO := LeftAO;
  FRightAO := RightAO;
  FResultFieldsOrder := TACRIntegerArray.Create(0,1,100);
  FIsMaterialized := False;
  FFieldCount := 0;
  FLeftAONull := False;
  FRightAONull := False;
  FDistinctApplied := False;
  FDistinctFields := '';
  FFilterExpr := nil;
  FTopRowCount := -1;
  FFirstRowNo := -1;
  FExpressionsExists := False;
  FAggregateExpressionsExists := False;
  FResultIndexFieldsList := nil;
  FResultIndexAscDescFieldsList := nil;
  FResultIndexCaseInsFieldsList := nil;
  FResultTableName := '';
  FResultDatabaseName := '';
  FResultInMemory := False;
end; // InternalCreate


//------------------------------------------------------------------------------
// go to first record
//------------------------------------------------------------------------------
procedure TACRAO.InternalFirst;
begin
;
end; // First


//------------------------------------------------------------------------------
// go to next record
//------------------------------------------------------------------------------
procedure TACRAO.InternalNext;
begin
;
end; // Next


//------------------------------------------------------------------------------
// return true if cursor points to the last record
//------------------------------------------------------------------------------
function TACRAO.InternalGetEof: Boolean;
begin
 Result := False;
end; // InternalGetEof


//------------------------------------------------------------------------------
// return number of records
//------------------------------------------------------------------------------
function TACRAO.InternalGetRecordCount: Integer;
begin
 Result := 0;
 if (FIsMaterialized) then
   Result := FResultDataset.RecordCount;
end; // InternalGetRecordCount


//------------------------------------------------------------------------------
// go to first record
//------------------------------------------------------------------------------
procedure TACRAO.First;
begin
  if (FIsMaterialized) then
    FResultDataset.First
  else
    InternalFirst;
end; // First


//------------------------------------------------------------------------------
// go to next record
//------------------------------------------------------------------------------
procedure TACRAO.Next;
begin
  if (FIsMaterialized) then
    FResultDataset.Next
  else
    InternalNext;
end; // Next


//------------------------------------------------------------------------------
// return true if cursor points to the last record
//------------------------------------------------------------------------------
function TACRAO.GetEof: Boolean;
begin
  if (FIsMaterialized) then
    Result := FResultDataset.Eof
  else
  Result := InternalGetEof;
end; // GetEof


//------------------------------------------------------------------------------
// return number of records
//------------------------------------------------------------------------------
function TACRAO.GetRecordCount: Integer;
begin
  if (FIsMaterialized) then
    Result := FResultDataset.RecordCount
  else
    Result := InternalGetRecordCount;
end; // GetRecordCount


//------------------------------------------------------------------------------
// sets names to FieldLinks list and renames duplicate names
//------------------------------------------------------------------------------
procedure TACRAO.SetFieldNames;
var i,j,k:    Integer;
    name,s:   String;
    bOk:      Boolean;
begin
  for i := 0 to FFieldCount-1 do
    begin
      FFieldLinks[i].FieldName := GetFieldName(i);
      // renaming fields without name (calculated, expressions...)
      if (FFieldLinks[i].FieldName = '') then
        FFieldLinks[i].FieldName := GetTemporaryName(ACRExpressionFieldName);
    end;
  // remove duplicates
  for i := 1 to FFieldCount-1 do
    begin
      k := 1;
      name := FFieldLinks[i].FieldName;
      repeat
        s := AnsiUpperCase(name);
        bOk := true;
        for j := 0 to i-1 do
          if (s = AnsiUpperCase(FFieldLinks[j].FieldName)) then
            begin
              bOk := false;
              break;
            end;
        if (bOk) then
          Break;
        if (k = 1) then
          name := name +'_' + IntToStr(k)
        else
          name := name + IntToStr(k);
        inc(k);
      until bOk;
      FFieldLinks[i].FieldName := name;
  end;
end;


//------------------------------------------------------------------------------
// create table
//------------------------------------------------------------------------------
procedure TACRAO.CreateIndexForMaterialize;
var IndexDef:                            TIndexDef;
    FieldNames, DescNames, CaseInsNames: String;
    IndexName:                           String;
begin
  ConvertListsToIndexFieldNames(FieldNames,DescNames,CaseInsNames,
    FResultIndexFieldsList,FResultIndexAscDescFieldsList,
    FResultIndexCaseInsFieldsList);
  repeat
    IndexName := GetTemporaryName(ACRTemporaryTableIndexName);
  until (TACRTable(FResultDataset).IndexDefs.IndexOf(IndexName) = -1);
  IndexDef := TACRTable(FResultDataset).IndexDefs.AddIndexDef;
  IndexDef.Name := IndexName;
  IndexDef.Fields := FieldNames;
  IndexDef.DescFields := DescNames;
  IndexDef.CaseInsFields := CaseInsNames;
  IndexDef.Options := [];
end; // CreateIndexForMaterialize


//------------------------------------------------------------------------------
// fill field defs
//------------------------------------------------------------------------------
procedure TACRAO.CreateTableForMaterializeFillFieldDefs(
                                              FieldList:  TStringList;
                                              AliasList:  TStringList
                                           );
var i,j:        Integer;
    FieldDef:   TACRAdvFieldDef;
    name:       String;
begin
 // result table is a table with applied projection (output fields list)
 if (FResultFieldsOrder.ItemCount <= 0) then
  begin
   // no projection
   // not result table
   for i := 0 to FFieldCount-1 do
    begin
      // field names will be 'Field'+n
      if (FFieldLinks[i].IsHidden) then continue;

      FieldList.Add(FFieldLinks[i].FieldName);
      AliasList.Add('');
      FieldDef := TACRTable(FResultDataset).AdvFieldDefs.AddFieldDef;
      FieldDef.Name := FFieldLinks[i].FieldName;
      if (IsAutoincFieldType(FFieldLinks[i].FieldType)) then
       FieldDef.DataType := BaseFieldTypeToAdvancedFieldType(AdvancedFieldTypeToBaseFieldType(FFieldLinks[i].FieldType))
      else
       FieldDef.DataType := FFieldLinks[i].FieldType;
      FieldDef.Size := FFieldLinks[i].FieldSize;
      FieldDef.BLOBCompressionAlgorithm :=
        TCompressionAlgorithm(FFieldLinks[i].BLOBCompressionAlgorithm);
      FieldDef.BLOBCompressionMode := FFieldLinks[i].BLOBCompressionMode;
      FieldDef.BLOBBlockSize := FFieldLinks[i].BLOBBlockSize;
    end; // not result table
  end
 else
  begin
   // projection is set
   // prepare visible fields
   for i := 0 to FFieldCount-1 do
    begin
      if (FFieldLinks[i].IsHidden) then continue;
      name := FFieldLinks[i].FieldName;
      j := 1;
      while (TACRTable(FResultDataset).AdvFieldDefs.Find(name) <> nil) do
       begin
        name := FFieldLinks[i].FieldName+'_'+IntToStr(j);
        Inc(j);
       end;
      FFieldLinks[i].FieldName := name;
      FieldDef := TACRTable(FResultDataset).AdvFieldDefs.AddFieldDef;
      FieldDef.Name := name;
//      if (FFieldLinks[i].FieldType = aftAutoInc) then
//       FieldDef.DataType := aftInteger
//      else
       FieldDef.DataType := FFieldLinks[i].FieldType;
      FieldDef.Size := FFieldLinks[i].FieldSize;
      FieldDef.BLOBCompressionAlgorithm :=
        TCompressionAlgorithm(FFieldLinks[i].BLOBCompressionAlgorithm);
      FieldDef.BLOBCompressionMode := FFieldLinks[i].BLOBCompressionMode;
      FieldDef.BLOBBlockSize := FFieldLinks[i].BLOBBlockSize;
    end;
    // perpare projection
   for i := 0 to FResultFieldsOrder.ItemCount-1 do
    begin
     j := FResultFieldsOrder.Items[i];
     if (FFieldLinks[j].IsHidden) then
      raise EACRException.Create(10312,ErrorLCannotAccessHiddenField,[FTableName,
              FFieldLinks[j].FieldName,FFieldLinks[j].DisplayName,j]);
     if (FFieldLinks[j].DisplayName = '') then
      raise EACRException.Create(11499,ErrorLEmptyDisplayName,
              [FTableName,FFieldLinks[j].FieldName,j]);
     FieldList.Add(FFieldLinks[j].FieldName);
     AliasList.Add(FFieldLinks[j].DisplayName);
    end;
  end; // result table
end; // CreateTableForMaterializeFillFields


//------------------------------------------------------------------------------
// create table
//------------------------------------------------------------------------------
procedure TACRAO.CreateTableForMaterialize(
                                            FieldList:  TStringList;
                                            AliasList:  TStringList
                                          );

var 
    IndexName:  String;
begin
   // prepare field defs
   TACRTable(FResultDataset).FieldDefs.Clear;
   TACRTable(FResultDataset).IndexDefs.Clear;
   TACRTable(FResultDataset).AdvFieldDefs.Clear;
   TACRTable(FResultDataset).AdvIndexDefs.Clear;
   CreateTableForMaterializeFillFieldDefs(FieldList,AliasList);
   // create table
   if (FResultTableName <> '') then
    begin
     TACRTable(FResultDataset).DatabaseName := FResultDatabaseName;
     TACRTable(FResultDataset).InMemory := FResultInMemory;
     TACRTable(FResultDataset).Temporary := False;
     TACRTable(FResultDataset).TableName := FResultTableName;
    end // result table
   else
    begin
     TACRTable(FResultDataset).InMemory := False;
     TACRTable(FResultDataset).Temporary := True;
     repeat
      TACRTable(FResultDataset).TableName := GetTemporaryName(ACRTemporaryTableName);
     until (not TACRTable(FResultDataset).Exists);
    end; // temporary table
   if (FResultIndexFieldsList <> nil) then
     CreateIndexForMaterialize;
   TACRTable(FResultDataset).CreateTable;
   TACRTable(FResultDataset).Open;
   if (FDistinctApplied) then
    begin
     if (TACRTable(FResultDataset).IndexDefs.Count <> 1) then
      raise EACRException.Create(11388,ErrorLInvalidIndexCount,[TACRTable(FResultDataset).IndexDefs.Count]);
     IndexName := TACRTable(FResultDataset).IndexDefs[0].Name;
     TACRTable(FResultDataset).IndexName := IndexName;
    end;
end; // CreateTableForMaterialize


//------------------------------------------------------------------------------
// fill table
//------------------------------------------------------------------------------
procedure TACRAO.FillTableForMaterialize;
var i,j,k,d,n,dc:     Integer;
    Value:            TACRVariant;
    FieldNumbersSrc:  array of Integer;
    FieldNumbersDest: array of Integer;
    bOK,Res:          Boolean;
begin
 FieldNumbersSrc := nil;
 FieldNumbersDest := nil;
 if (FDistinctApplied) then
  begin
   dc := FResultIndexFieldsList.Count;
   SetLength(FieldNumbersSrc,dc);
   SetLength(FieldNumbersDest,dc);
   for d := 0 to dc-1 do
    begin
     n := 0;
     for i := 0 to FFieldCount-1 do
      if (not FFieldLinks[i].IsHidden) then
       begin
         if (AnsiUpperCase(FFieldLinks[i].FieldName) =
              AnsiUpperCase(FResultIndexFieldsList[d])) then
          begin
           FieldNumbersSrc[d] := i;
           FieldNumbersDest[d] := n;
           break;
          end;
         inc(n);
       end;
    end;
  end
 else
  dc := 0;
 value := TACRVariant.Create;
 try
   // filling table with data
   First;
   k := 0;
   while Not Eof do
    begin
     inc(k);
     // check TOP n?
      // TOP n?
      if (FTopRowCount > -1) then
      begin
       if (TACRDataset(FResultDataset).RecordCount >= FTopRowCount) then
        break;
      end;
      // TOP first_row
      if (FFirstRowNo > -1)  then
      begin
       if (k < FFirstRowNo) then
        begin
         Next;
         continue;
        end;
      end;

     if (FFilterExpr <> nil) then
      begin
       // check filter
       Res := TACRExpression(FFilterExpr).GetResult;
       if (not Res) then
        begin
         Next;
         continue;
        end;
      end;
     // check distinct
     if (FDistinctApplied) then
      begin
       if (TACRTable(FResultDataset).RecordCount > 0) then
        begin
         TACRTable(FResultDataset).SetKey;
         for d := 0 to dc-1 do
          begin
           // searching for the distinct field
           n := FieldNumbersSrc[d];
           // read current record
           GetFieldValue(value,n,false,true);
           n := FieldNumbersDest[d];
           TACRTable(FResultDataset).SetFieldValue(value,n,False);
          end;//for
         bOK := TACRTable(FResultDataset).GotoKey;
         if (bOK) then
          begin
           // skip this record
           Next;
           Continue;
          end;
        end;
      end;

     TACRTable(FResultDataset).Insert;

       // not result table
       j := 0;
       for i := 0 to FFieldCount-1 do
        begin
         if (FFieldLinks[i].IsHidden) then continue;
         // get field value
         GetFieldValue(value,i,false,true);
         // set value
         TACRDataset(FResultDataset).SetFieldValue(value,j,False);
         value.SetNull;
         inc(j);
        end; // not result table
     // insert record
     TACRTable(FResultDataset).Post;
     // go to next record
     Next;
    end; // enf of inserting records loop
   if (FTopRowCount > -1) then
    FTopRowCount := -1;
 finally
   value.Free;
   if (FDistinctApplied) then
    begin
     FieldNumbersSrc := nil;
     FieldNumbersDest := nil;
    end;
 end;
end; // FillTableForMaterialize


//------------------------------------------------------------------------------
// finalize materialize
//------------------------------------------------------------------------------
procedure TACRAO.FinalizeMaterialize(
                                            FieldList:  TStringList;
                                            AliasList:  TStringList
                                    );
var i,j:      Integer;
   IndexName: String;
begin
 if ((FResultIndexFieldsList <> nil) and (FActivateIndexAfterMaterialize)) then
  begin
   if (TACRTable(FResultDataset).IndexDefs.Count <> 1) then
    raise EACRException.Create(11385,ErrorLInvalidIndexCount,[TACRTable(FResultDataset).IndexDefs.Count]);
   IndexName := TACRTable(FResultDataset).IndexDefs[0].Name;
   TACRTable(FResultDataset).IndexName := IndexName;
  end;
 if (FResultFieldsOrder.ItemCount > 0) then
  TACRTable(FResultDataset).ApplyProjection(FieldList,AliasList);
 // move field links to the result dataset
 if (FResultFieldsOrder.ItemCount <= 0) then
  begin
   j := 0;
   // not result table
   for i := 0 to FFieldCount-1 do
    begin
     if (FFieldLinks[i].IsExpression) then
      if (FFieldLinks[i].Expr <> nil) then
        TACRExpression(FFieldLinks[i].Expr).Free;
     // field names will be 'Field'+n
     if (FFieldLinks[i].IsHidden) then continue;
     FFieldLinks[i].AO := nil;
     FFieldLinks[i].IsExpression := False;
     FFieldLinks[i].IsAggregate := False;
     FFieldLinks[i].Expr := nil;
     FFieldLinks[i].Dataset := FResultDataset;
     FFieldLinks[i].FieldNo := j;
     inc(j);
    end; // not result table
  end
 else
  begin
   for i := 0 to FFieldCount-1 do
     if (FFieldLinks[i].IsExpression) then
      if (FFieldLinks[i].Expr <> nil) then
        TACRExpression(FFieldLinks[i].Expr).Free;
  end; // result table

 FIsMaterialized := True;
 //---------------------- destroy child operations -----------------------------
 if (Self is TACRAOJoin) then
  TACRAOJoin(Self).ClearFilter;
 if (FLeftAO <> nil) then
  FLeftAO.Free;
 if (FRightAO <> nil) then
  FRightAO.Free;
 FRightAO := nil;
 FLeftAO := nil;
end; // FinalizeMaterialize


//------------------------------------------------------------------------------
// materializes AO
//------------------------------------------------------------------------------
procedure TACRAO.DoMaterialize;
var FieldList,AliasList: TStringList;
begin
 if (not FIsMaterialized) then
  begin
   FieldList := TStringList.Create;
   AliasList := TStringList.Create;
   try
     CreateTableForMaterialize(FieldList,AliasList);
     FillTableForMaterialize;
     FinalizeMaterialize(FieldList,AliasList);
   finally
     FieldList.Free;
     AliasList.Free;
   end;
  end;
end; // DoMaterialize


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRAO.Destroy;
begin
  FFieldLinks := nil;
  if (FResultDataset <> nil) then
    FResultDataset.Free;
  if (FResultFieldsOrder <> nil) then
    FResultFieldsOrder.Free;
  if (FLeftAO <> nil) then
    FLeftAO.Free;
  if (FRightAO <> nil) then
    FRightAO.Free;
  if (FFilterExpr <> nil) then
    TACRExpression(FFilterExpr).Free;
  FreeResultIndexLists;  
  inherited Destroy;
end; // Destroy


//------------------------------------------------------------------------------
// getting all result records
//------------------------------------------------------------------------------
procedure TACRAO.Execute(IsRootAO : Boolean = false);
begin
  FIsRootAO := IsRootAO;
  if (FLeftAO <> nil) then
    FLeftAO.Execute;
  if (FRightAO <> nil) then
    FRightAO.Execute;

 try
    // filter
  if (FFilterExpr <> nil) then
    TACRExpression(FFilterExpr).AssignAO(Self);
  if (FResultIndexByAllFields) then
   SetResultIndexByAllFields;
  DoMaterialize;
 finally
  if (FResultIndexFieldsList <> nil) then
    FreeResultIndexLists;
 end;
end;// Execute


//------------------------------------------------------------------------------
// set filter
//------------------------------------------------------------------------------
procedure TACRAO.SetFilter(FilterExpr: TObject);
begin
 FFilterExpr := FilterExpr;
end; // SetFilter


//------------------------------------------------------------------------------
// for SELECT INTO optimization
//------------------------------------------------------------------------------
procedure TACRAO.SetResultTable(InMemory: Boolean; TableName: String; DatabaseName: String);
begin
  FResultInMemory := InMemory;
  FResultTableName := TableName;
  FResultDatabaseName := DatabaseName;
end; // SetResultTable


//------------------------------------------------------------------------------
// sets Top row count
//------------------------------------------------------------------------------
procedure TACRAO.SetTopRowCount(FirstRowNo, TopRowCount: integer);
begin
  FTopRowCount := TopRowCount;
  if (FirstRowNo < 0) then
   FFirstRowNo := 1
  else
   FFirstRowNo := FirstRowNo;
end;// SetTopRowCount


//------------------------------------------------------------------------------
// applies distinct
//------------------------------------------------------------------------------
procedure TACRAO.ApplyDistinct(DistinctFields: String);
var i,j: Integer;
    DistinctFieldList: TStringList;
begin
  DistinctFieldList := TStringList.Create;
  try
    FillFieldNames(DistinctFieldList,DistinctFields);
    FDistinctApplied := (DistinctFieldList.Count > 0);
    if (FDistinctApplied) then
     begin
      FDistinctFields := DistinctFields;
      CreateResultIndexLists(False);
      for i := 0 to DistinctFieldList.Count - 1 do
       begin
        j := FResultIndexFieldsList.IndexOf(AnsiUpperCase(DistinctFieldList[i]));
        if (j < 0) then
         begin
          FResultIndexFieldsList.Add(DistinctFieldList[i]);
          FResultIndexAscDescFieldsList.Add(ACR_ASC);
          FResultIndexCaseInsFieldsList.Add(ACR_CASE);
         end;
       end;
     end;
  finally
    DistinctFieldList.Free;
  end;
end; // ApplyDistinct


//------------------------------------------------------------------------------
// sets projection for other TACRAO
//------------------------------------------------------------------------------
procedure TACRAO.SetResultFields(var FieldRefs: array of TACRSelectListItem;
                                  bDistinct: Boolean);


var i,j,k,x,res: integer;
    fname,tname: string;
    aDistinctFields: String;
    tempArray: TACRIntegerArray;

procedure AddField(FieldNo: Integer);
begin
  if (FResultFieldsOrder.IsValueExists(FieldNo)) then
   begin
    Inc(FFieldCount);
    SetLength(FFieldLinks,FFieldCount);
    FFieldLinks[FFieldCount-1] := FFieldLinks[FieldNo];
    FieldNo := FFieldCount-1;
   end;
  FResultFieldsOrder.Append(FieldNo);
  if (FieldRefs[i].Pseudonym <> '') then
   // set pseudonym
    FFieldLinks[FieldNo].DisplayName := FieldRefs[i].Pseudonym
  else
   // use name like it was written in query
   if (fname <> '*') then
    begin
     if (AnsiUpperCase(FFieldLinks[FieldNo].FieldName) <> AnsiUpperCase(fname)) then
       FFieldLinks[FieldNo].DisplayName := FFieldLinks[FieldNo].FieldName
     else
       FFieldLinks[FieldNo].DisplayName := fname;
    end;
end; // AddField

begin
 aDistinctFields := '';
 j := Length(FieldRefs);
 if (j <= 0) then
  begin
   FieldExists('*','',true,FResultFieldsOrder);
  end
 else
 begin
  tempArray := TACRIntegerArray.Create;
  try
    for i := 0 to j-1 do
     begin
      if (FieldRefs[i].AllFields) then
       fname := '*'
      else
       fname := FieldRefs[i].FieldName;
      tname := FieldRefs[i].TableName;
      if (FieldRefs[i].IsExpression) then
       begin
        inc(FFieldCount);
        SetLength(FFieldLinks,FFieldCount);
        FFieldLinks[FFieldCount-1].Expr := FieldRefs[i].ValueExpr;
        TACRExpression(FFieldLinks[FFieldCount-1].Expr).AssignAO(self);
        // create new field for this expression
        if (FieldRefs[i].Pseudonym <> '') then
         // pseudonim specified
         FFieldLinks[FFieldCount-1].FieldName := FieldRefs[i].Pseudonym
        else
         // random name
         FFieldLinks[FFieldCount-1].FieldName := GetTemporaryName(ACRExpressionFieldName);
        FFieldLinks[FFieldCount-1].DisplayName := FFieldLinks[FFieldCount-1].FieldName;
        FFieldLinks[FFieldCount-1].OriginalFieldName := FFieldLinks[FFieldCount-1].FieldName;
        FFieldLinks[FFieldCount-1].TableName := '';
        FFieldLinks[FFieldCount-1].TableAlias := '';
        // get field type
        FFieldLinks[FFieldCount-1].FieldType :=
          TACRExpression(FieldRefs[i].ValueExpr).GetDataType;
        // get field size
        FFieldLinks[FFieldCount-1].FieldSize :=
          TACRExpression(FieldRefs[i].ValueExpr).GetDataSize;
        // get field precision
        FFieldLinks[FFieldCount-1].FieldPrecision :=
          TACRExpression(FieldRefs[i].ValueExpr).GetPrecision;
        // AO
        FFieldLinks[FFieldCount-1].AO := self;
        FFieldLinks[FFieldCount-1].Dataset := nil;
        FFieldLinks[FFieldCount-1].IsHidden := false;
        FFieldLinks[FFieldCount-1].IsExpression := true;
        FFieldLinks[FFieldCount-1].IsAggregate :=
          TACRExpression(FieldRefs[i].ValueExpr).IsAggregated;
        FExpressionsExists := True;
        if (not FAggregateExpressionsExists) then
         FAggregateExpressionsExists := FFieldLinks[FFieldCount-1].IsAggregate;
        FResultFieldsOrder.Append(FFieldCount-1);
        continue;
       end; // Expression
      tempArray.SetSize(0);
      res := FieldExists(fname,tname,true,TempArray);
      if (res = 0) then
        raise EACRException.Create(10305, ErrorLCannotFindField, [fname]);
      if (fname = '*') then
       for x := 0 to TempArray.ItemCount - 1 do
         AddField(TempArray.Items[x])
      else
       AddField(TempArray.Items[0]);

      // FResultFieldsOrder
{
      if (res = 1) then
       begin
        x := FResultFieldsOrder.Items[FResultFieldsOrder.ItemCount-1];
       if (FieldRefs[i].Pseudonym <> '') then
        // set pseudonym
         FFieldLinks[x].DisplayName := FieldRefs[i].Pseudonym
       else
        // use name like it was written in query
        if (fname <> '*') then
         begin
          if (AnsiUpperCase(FFieldLinks[x].DisplayName) <> AnsiUpperCase(fname)) then
            FFieldLinks[x].DisplayName := FFieldLinks[x].FieldName
          else
            FFieldLinks[x].DisplayName := fname;
         end;
       end // res = 0
      else
      if (res = 0) then
        raise EACRException.Create(10305, ErrorLCannotFindField, [fname])
  // result dataset
      else
       if (res > 1) then
        if (fname <> '*') then
         begin
          for x := 0 to res-2 do
           FResultFieldsOrder.Delete(FResultFieldsOrder.ItemCount-1);
          x := FResultFieldsOrder.Items[FResultFieldsOrder.ItemCount-1];
          if (FieldRefs[i].Pseudonym <> '') then
           // set pseudonym
           FFieldLinks[x].DisplayName := FieldRefs[i].Pseudonym
          else
           // use name like it was written in query
           if (fname <> '*') then
            FFieldLinks[x].DisplayName := fname;
         end; // DUPLICATE NAMES IN SELECT LIST REMOVED
}
     end;
  finally
    tempArray.Free;
  end;

 end;
 for i := 0 to FResultFieldsOrder.ItemCount - 1 do
  begin
   k := FResultFieldsOrder.Items[i];
   if (bDistinct) then
    begin
     if (aDistinctFields = '') then
      aDistinctFields := FFieldLinks[k].FieldName
     else
      aDistinctFields := aDistinctFields + ';' + FFieldLinks[k].FieldName;
    end;

   if (FFieldLinks[k].DisplayName = '') then
    FFieldLinks[k].DisplayName :=
      FFieldLinks[k].FieldName;
  end;

  if (bDistinct) then
   ApplyDistinct(aDistinctFields);
  if (Self is TACRAOJoin) then
   TACRAOJoin(Self).FixResultFieldNames;
end;// SetResultFields


//------------------------------------------------------------------------------
// mapping function - return number of found fields and found field No
// also optionally unhides fields in AO
//------------------------------------------------------------------------------
function TACRAO.FieldExists(
                  FieldName, TableName: string;
                  Unhide: Boolean;
                  FieldNumbers: TACRIntegerArray = nil;
                  UnhideChildrenOnly:   Boolean = False
                  ): Integer;
var i,j,k: integer;
    fname: string;
    tempArray: TACRIntegerArray;
begin
 result := 0;
 if (FIsAOTable) then
   if (TableName <> '') then
    if (AnsiUpperCase(TableName) <> AnsiUpperCase(FTableName)) and
      (AnsiUpperCase(TableName) <> AnsiUpperCase(FTableAlias)) then
     Exit; // another table

 // all fields
   if (FieldName = '*') then
    begin
     // unhide or field numbers
     tempArray := TACRIntegerArray.Create(0,1,100);
     if (FLeftAO <> nil) then
       FLeftAO.FieldExists(FieldName,TableName,Unhide, tempArray);

     for i := 0 to FFieldCount-1 do
      if (FIsAOTable) or
         ((AnsiUpperCase(TableName) = AnsiUpperCase(FTableName)) and
         (FFieldLinks[i].IsExpression)) or
         ((FLeftAO <> nil) and
            (FFieldLinks[i].AO = FLeftAO) and
            (tempArray.IsValueExists(FFieldLinks[i].FieldNo))) then
       begin
        inc(result);
        if (FieldNumbers <> nil) then
         FieldNumbers.Append(i);
        // group by does not needs in unhiding
        // it will cause hidden fields to be appeared in result dataset
        if (Unhide) and (not UnhideChildrenOnly) then
         if (not FIsAOGroupBy) then
          FFieldLinks[i].IsHidden := false;
       end;
     tempArray.SetSize(0);

     if (FRightAO <> nil) then
      begin
       FRightAO.FieldExists(FieldName,TableName,Unhide, tempArray);
       for i := 0 to FFieldCount-1 do
       if ((FRightAO <> nil) and
            (FFieldLinks[i].AO = FRightAO) and
            (tempArray.IsValueExists(FFieldLinks[i].FieldNo))) then
         begin
          inc(result);
          if (FieldNumbers <> nil) then
           FieldNumbers.Append(i);
          // group by does not needs in unhiding
          // it will cause hidden fields to be appeared in result dataset
          if (Unhide) and (not UnhideChildrenOnly) then
           FFieldLinks[i].IsHidden := false;
         end;
      end; // RightAO is not materialized
   tempArray.Free;
   Exit;
  end; // all fields

   fname := AnsiUpperCase(FieldName);
   // not TACRAOTable
   tempArray := nil;
   if (FieldNumbers <> nil) then
    tempArray := TACRIntegerArray.Create(0,1,100);
   if (FLeftAO <> nil) then
    begin
     result := result + FLeftAO.FieldExists(FieldName,TableName,Unhide,tempArray);
     if (FieldNumbers <> nil) then
      begin
       for i := 0 to tempArray.ItemCount-1 do
        begin
         k := tempArray.Items[i];
         for j := 0 to FFieldCount-1 do
          if (FFieldLinks[j].AO = FLeftAO) then
           if (FFieldLinks[j].FieldNo = k) then
            begin
             // return FieldNo in current FFieldLinks
             FieldNumbers.Append(j);
             // group by does not needs in unhiding
             // it will cause hidden fields to be appeared in result dataset
             if (Unhide) and (not UnhideChildrenOnly) then
              FFieldLinks[j].IsHidden := false;
            end;
        end;
       tempArray.SetSize(0); // empty temp array
      end; // fieldNumbers exists
    end; // LeftAO
   if (FRightAO <> nil) then
    begin
     result := result + FRightAO.FieldExists(FieldName,TableName,Unhide,tempArray);
     if (FieldNumbers <> nil) then
      begin
       for i := 0 to tempArray.ItemCount-1 do
        begin
         k := tempArray.Items[i];
         for j := 0 to FFieldCount-1 do
          if (FFieldLinks[j].AO = FRightAO) then
           if (FFieldLinks[j].FieldNo = k) then
            begin
             // return FieldNo in current FFieldLinks
             FieldNumbers.Append(j);
             // group by does not needs in unhiding
             // it will cause hidden fields to be appeared in result dataset
             if (Unhide) and (not UnhideChildrenOnly) then
              FFieldLinks[j].IsHidden := false;
            end;
        end;
        tempArray.SetSize(0);
      end; // fieldNumbers exists
    end; // RightAO
   if (tempArray <> nil) then
    tempArray.Free;

   // find field
   for i := 0 to FFieldCount-1 do
    if (AnsiUpperCase(FFieldLinks[i].FieldName) = fname) or
       (AnsiUpperCase(FFieldLinks[i].DisplayName) = fname) then
     if (
     (TableName = '') or
         (AnsiUpperCase(TableName) = AnsiUpperCase(FFieldLinks[i].TableName)) or
         (AnsiUpperCase(TableName) = AnsiUpperCase(FFieldLinks[i].TableAlias))) then
        begin
         if (FieldNumbers <> nil) then
          if (not FieldNumbers.IsValueExists(i)) then
           begin
            FieldNumbers.Append(i);
            inc(result);
           end;
         // group by does not needs in unhiding
         // it will cause hidden fields to be appeared in result dataset
         if (Unhide) and (not UnhideChildrenOnly) then
          FFieldLinks[i].IsHidden := false;
         break;
        end;
end;// FieldExists


//------------------------------------------------------------------------------
// return FieldName from AOTable for specified internal field
//------------------------------------------------------------------------------
function TACRAO.GetFieldName(FieldNo: Integer; ApplyOrderBy: Boolean = False): string;
begin
 if (FieldNo < 0) or (FieldNo >= FieldCount) then
  raise EACRException.Create(10306,ErrorLInvalidFieldNumber,[FieldNo,FieldCount]);
 Result := '';
 if (ApplyOrderBy) then
  begin
   // apply order by
   if (FFieldLinks[FieldNo].Dataset <> nil) then
    begin
     if (FFieldLinks[FieldNo].Dataset = FResultDataset) then
      Result := FFieldLinks[FieldNo].FieldName
     else
      Result := FFieldLinks[FieldNo].Dataset.Fields[FFieldLinks[FieldNo].FieldNo].FieldName;
    end
   else
    begin
      Result := FFieldLinks[FieldNo].FieldName;
    end;
  end
 else
  begin
   // set field names for join
   if (FFieldLinks[FieldNo].Dataset <> nil) then
    begin
     if (FFieldLinks[FieldNo].Dataset = FResultDataset) then
      Result := FFieldLinks[FieldNo].FieldName
     else
      Result := FFieldLinks[FieldNo].Dataset.Fields[FFieldLinks[FieldNo].FieldNo].FieldName
    end
   else
    Result := TACRAO(FFieldLinks[FieldNo].AO).GetFieldName(FFieldLinks[FieldNo].FieldNo)
  end;
end; // GetFieldName


//------------------------------------------------------------------------------
// return field name by column number specified in ORDER BY
//------------------------------------------------------------------------------
function TACRAO.GetFieldNameByOrderByColumnNumber(ColumnNumber: Integer): String;
var i: Integer;
begin
 if (((ColumnNumber-1) >= FResultFieldsOrder.ItemCount) or (ColumnNumber <= 0)) then
  raise EACRException.Create(11395,ErrorLInvalidOrderByColumnNumber,[ColumnNumber]);
 i := FResultFieldsOrder.Items[ColumnNumber-1];
 Result := FFieldLinks[i].FieldName;
end; // GetFieldNameByOrderByColumnNumber


//------------------------------------------------------------------------------
// return field value
//------------------------------------------------------------------------------
procedure TACRAO.GetFieldValue(
                        Value:          TACRVariant;
                        FieldNo:        Integer;
                        bCopy:          Boolean = False;
                        AccessToHidden: Boolean = False
                              );
var
  TmpValue: TACRVariant;
begin
  if ((FieldNo < 0) or (FieldNo >= FieldCount)) then
    raise EACRException.Create(10307, ErrorLInvalidFieldNumber,
                               [FieldNo,FieldCount]);

  Value.SetNull;
  if (FAllFieldsNull and (not FFieldLinks[FieldNo].IsAggregate)) then
   Exit;
  // AO = Dataset = nil
  if ((FFieldLinks[FieldNo].Dataset = nil) and (FFieldLinks[FieldNo].AO = nil)) then
    raise EACRException.Create(10309, ErrorLInvalidFieldLink, [FTableName,FieldNo]);

 // hiden field
 if not AccessToHidden then
   if (FFieldLinks[FieldNo].IsHidden) then
     raise EACRException.Create(10310,ErrorLCannotAccessHiddenField,
                    [FTableName,FFieldLinks[FieldNo].FieldName,
                     FFieldLinks[FieldNo].DisplayName, FieldNo]);

 if (FFieldLinks[FieldNo].IsExpression) then
   begin
     TmpValue := TACRExpression(FFieldLinks[FieldNo].Expr).GetValue;
     if (TmpValue <> nil) then
       Value.Assign(TmpValue, False);
   end // Expression
 else
   if (FFieldLinks[FieldNo].Dataset <> nil) then
     TACRTable(FFieldLinks[FieldNo].Dataset).GetFieldValue(value,
                                                 FFieldLinks[FieldNo].FieldNo,False,bCopy)
 else
   begin
     if (FLeftAONull and (FFieldLinks[FieldNo].AO = FLeftAO)) or
        (FRightAONull and (FFieldLinks[FieldNo].AO = FRightAO)) then
       begin
         // return null value
         Exit;
       end;
     TACRAO(FFieldLinks[FieldNo].AO).GetFieldValue(value,
                          FFieldLinks[FieldNo].FieldNo, bCopy, AccessToHidden);
   end;
end; // GetFieldValue


//------------------------------------------------------------------------------
// get field type
//------------------------------------------------------------------------------
function TACRAO.GetFieldType(FieldNo: Integer): TACRAdvancedFieldType;
begin
  if ((FieldNo < 0) or (FieldNo >= FieldCount)) then
    raise EACRException.Create(10376, ErrorLInvalidFieldNumber,
                               [FieldNo,FieldCount]);
  Result := FFieldLinks[FieldNo].FieldType;
end; // GetFieldType


//------------------------------------------------------------------------------
// get field size
//------------------------------------------------------------------------------
function TACRAO.GetFieldSize(FieldNo: Integer): Integer;
begin
  if ((FieldNo < 0) or (FieldNo >= FieldCount)) then
    raise EACRException.Create(10377, ErrorLInvalidFieldNumber,
                               [FieldNo,FieldCount]);
  Result := FFieldLinks[FieldNo].FieldSize;
end; // GetFieldSize


//------------------------------------------------------------------------------
// get field precision
//------------------------------------------------------------------------------
function TACRAO.GetFieldPrecision(FieldNo: Integer): Integer;
begin
  if ((FieldNo < 0) or (FieldNo >= FieldCount)) then
    raise EACRException.Create(10378, ErrorLInvalidFieldNumber,
                               [FieldNo,FieldCount]);
  Result := FFieldLinks[FieldNo].FieldPrecision;
end; // GetFieldPrecision


//------------------------------------------------------------------------------
// lock table
//------------------------------------------------------------------------------
procedure TACRAO.LockTable;
begin
 if (FLeftAO <> nil) then
  FLeftAO.LockTable;
 if (FRightAO <> nil) then
  FRightAO.LockTable;
end; // LockTable


//------------------------------------------------------------------------------
// unlock table
//------------------------------------------------------------------------------
procedure TACRAO.UnlockTable;
begin
 if (FLeftAO <> nil) then
  FLeftAO.UnlockTable;
 if (FRightAO <> nil) then
  FRightAO.UnlockTable;
end; // LockTable


//------------------------------------------------------------------------------
// create result index lists
//------------------------------------------------------------------------------
procedure TACRAO.CreateResultIndexLists(ToClear: Boolean = True);
begin
 if (FResultIndexFieldsList = nil) then
   FResultIndexFieldsList := TStringList.Create
 else
  if (ToClear) then
   FResultIndexFieldsList.Clear;
 if (FResultIndexAscDescFieldsList = nil) then
   FResultIndexAscDescFieldsList := TStringList.Create
 else
  if (ToClear) then
   FResultIndexAscDescFieldsList.Clear;
 if (FResultIndexCaseInsFieldsList = nil) then
   FResultIndexCaseInsFieldsList := TStringList.Create
 else
  if (ToClear) then
   FResultIndexCaseInsFieldsList.Clear;
end; // CreateResultIndexLists


//------------------------------------------------------------------------------
// free and nil result index lists
//------------------------------------------------------------------------------
procedure TACRAO.FreeResultIndexLists;
begin
 if (FResultIndexFieldsList <> nil) then
   FResultIndexFieldsList.Free;
 if (FResultIndexAscDescFieldsList <> nil) then
   FResultIndexAscDescFieldsList.Free;
 if (FResultIndexCaseInsFieldsList <> nil) then
   FResultIndexCaseInsFieldsList.Free;
 FResultIndexFieldsList := nil;
 FResultIndexAscDescFieldsList := nil;
 FResultIndexCaseInsFieldsList := nil;
end; // FreeResultIndexLists


//------------------------------------------------------------------------------
// SetResultIndexByAllFields
//------------------------------------------------------------------------------
procedure TACRAO.SetResultIndexByAllFields;
var i: Integer;
    FieldNames: String;
begin
 if (FResultIndexFieldsList <> nil) then
  FreeResultIndexLists;
 FieldNames := '';
 for i := 0 to FFieldCount - 1 do
  if (not FFieldLinks[i].IsHidden) then
   begin
    if (FieldNames = '') then
     FieldNames := FFieldLinks[i].FieldName
    else
     FieldNames := FieldNames + ';'+ FFieldLinks[i].FieldName;
   end;
  SetIndex(FieldNames,'','','',True); 
end; // SetResultIndexByAllFields


//------------------------------------------------------------------------------
// sets index
//------------------------------------------------------------------------------
procedure TACRAO.SetIndex(IndexFieldNames, DescFields, CaseInsensitiveFields, IndexName: String;
                          ActivateIndexAfterMaterialize: Boolean);
begin
 FreeResultIndexLists;
 FResultIndexName := IndexName;
 FActivateIndexAfterMaterialize := ActivateIndexAfterMaterialize;
 CreateResultIndexLists(True);
 ConvertIndexFieldNamesToLists(
      IndexFieldNames,DescFields,CaseInsensitiveFields,
      FResultIndexFieldsList,FResultIndexAscDescFieldsList,
      FResultIndexCaseInsFieldsList);
 if (FDistinctApplied) then
  ApplyDistinct(FDistinctFields);
end; // SetIndex


////////////////////////////////////////////////////////////////////////////////
//
// TACRAOTable
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// materializes AO
//------------------------------------------------------------------------------
procedure TACRAOTable.DoMaterialize;
var s:          String;
    value:      TACRVariant;
    values:     array of TACRVariant;
    dFields:    array of Integer;
    dFields1:   array of Integer;
    i,dc,j:     Integer;
    bOK:        Boolean;
    Distinct1:  Boolean;
    FieldList:  TStringList;
    AliasList:  TStringList;
    Res:        Boolean;


function IsMaterializationRequired: Boolean;
var IndexName: String;
begin
 Result := (FIsRootAO and (not FRequestLive));
 if (FFieldCount > FSourceDataset.FieldCount) then
  Result := True;
 if (FDistinctApplied) then
  Result := True;
 if (FResultIndexFieldsList <> nil) then
   begin
     IndexName := TACRTable(FSourceDataset).FindIndex(FResultIndexFieldsList,
                                           FResultIndexAscDescFieldsList,
                                           FResultIndexCaseInsFieldsList);
     TACRTable(FSourceDataset).IndexName := IndexName;
     if (not Result) then
      if (TACRTable(FSourceDataset).IndexName = '') then
       Result := True;
   end;
end; // IsMaterializationRequired

procedure ActivateIndex;
var IndexName: String;
begin
 if (FResultIndexFieldsList <> nil) then
  begin
   if (TACRTable(FResultDataset).IndexDefs.Count <> 1) then
    raise EACRException.Create(11386,ErrorLInvalidIndexCount,[TACRTable(FResultDataset).IndexDefs.Count]);
   IndexName := TACRTable(FResultDataset).IndexDefs[0].Name;
   TACRTable(FResultDataset).IndexName := IndexName;
  end;
end; // ActivateIndex

procedure InitializeDistinct;
var i,d,j: Integer;
begin
 dc := FResultIndexFieldsList.Count;
 SetLength(values,dc);
 SetLength(dFields,dc);
 SetLength(dFields1,dc);
 for i := 0 to dc -1 do
  begin
   values[i] := TACRVariant.Create;
  end;
 d := 0;
 while (d < dc) do
  begin
   j := 0;
   for i := 0 to FFieldCount-1 do
    begin
      if (AnsiUpperCase(FFieldLinks[i].FieldName) =
            AnsiUpperCase(FResultIndexFieldsList[d])) then
         begin
          dFields[d] := i;
          dFields1[d] := j;
          Inc(d);
          break;
         end;
      if (not FFieldLinks[i].IsHidden) then
       Inc(j);
    end;
  end;
 Distinct1 := (TACRTable(FSourceDataset).IndexName <> '');
end; // InitializeDistinct

procedure FinalizeDistinct;
var i: Integer;
begin
 dc := FResultIndexFieldsList.Count;
 for i := 0 to dc -1 do
  begin
   values[i].Free;
  end;
 values := nil;
 dFields := nil;
 dFields1 := nil;
end; // FinalizeDistinct

procedure CreateTemporaryTable;
begin
 TACRTable(FResultDataset).Close;
 TACRTable(FResultDataset).Temporary := True;
 repeat
  TACRTable(FResultDataset).TableName := GetTemporaryName(ACRTemporaryTableName);
 until (not TACRTable(FResultDataset).Exists);
 TACRTable(FResultDataset).FieldDefs.Clear;
 CreateTableForMaterializeFillFieldDefs(FieldList,AliasList);
 TACRTable(FResultDataset).IndexDefs.Clear;
 TACRTable(FResultDataset).AdvIndexDefs.Clear;
 if (FResultIndexFieldsList <> nil) then
  if (TACRTable(FSourceDataset).IndexName = '') then
   CreateIndexForMaterialize;
 TACRTable(FResultDataset).CreateTable;
 TACRTable(FResultDataset).Open;
end; // CreateTemporaryTable

// source table has proper index
function IsRecordMatchesDistinct1: Boolean;
var d,n: Integer;
begin
 Result := False;
 for d := 0 to dc-1 do
  begin
   // searching for the distinct field
   n := dFields[d];
   if (n < 0) then
    raise EACRException.Create(10323,ErrorLCannotFindField,
      [FResultIndexFieldsList[d]]);
   // read current record
   GetFieldValue(value,n,False,False);
   if (TACRTable(FResultDataset).RecordCount <= 0) then
    Result := True
   else
    if (not Result) then
     if (value.Compare(values[d],True,False,False) <> cmprEqual) then
      Result := True;
   values[d].Assign(value,True);
  end;//for
end; // IsRecordMatchesDistinct1

// result table has proper index
function IsRecordMatchesDistinct2: Boolean;
var d,n,k: Integer;
begin
 if (TACRTable(FResultDataset).RecordCount <= 0) then
  Result := True
 else
  begin
   TACRTable(FResultDataset).SetKey;
   for d := 0 to dc-1 do
    begin
     // searching for the distinct field
     n := dFields[d];
     k := dFields1[d];
     if (n < 0) then
      raise EACRException.Create(11387,ErrorLCannotFindField,
        [FResultIndexFieldsList[d]]);
     // read current record
     GetFieldValue(value,n,false,true);
     TACRTable(FResultDataset).SetFieldValue(value,k,False);
    end;//for
   Result := not TACRTable(FResultDataset).GotoKey;
  end;
end; // IsRecordMatchesDistinct2


function IsRecordMatchesDistinct: Boolean;
begin
 if (Distinct1) then
  Result := IsRecordMatchesDistinct1
 else
  Result := IsRecordMatchesDistinct2;
end; // IsRecordMatchesDistinct

procedure FinalizeMaterialize(Live: Boolean);
var i,j: Integer;
begin
 // not result table
 j := 0;
 for i := 0 to FFieldCount-1 do
  begin
   // field names will be 'Field'+n
   FFieldLinks[i].AO := nil;
   FFieldLinks[i].Dataset := FResultDataset;
   if (FFieldLinks[i].IsExpression) then
    if (FFieldLinks[i].Expr <> nil) then
      TACRExpression(FFieldLinks[i].Expr).Free;
   FFieldLinks[i].Expr := nil;
   FFieldLinks[i].IsExpression := False;
   FFieldLinks[i].IsAggregate := False;
   if (not Live) then
    FFieldLinks[i].FieldNo := j;
   if (not FFieldLinks[i].IsHidden) then
    Inc(j);
  end; // not result table
 if (FTableLocked) then
  begin
   TACRTable(FSourceDataset).UnlockTable;
   FTableLocked := False;
  end;
 if (not Live) then
  begin
   FSourceDataset.Free;
   TACRExpression(FFilterExpr).Free;
  end;
 FFilterExpr := nil;
end; // fnalizes materialization

begin
 if (FIsMaterialized) then
  Exit;
  FieldList := TStringList.Create;
  AliasList := TStringList.Create;
  try
    if (IsMaterializationRequired) then
     begin
       FRequestLive := False;
       FResultDataset := TACRTable.Create(nil);
       if (FDistinctApplied) then
        InitializeDistinct;
       try
         CreateTemporaryTable;
         if (FDistinctApplied) and (not Distinct1) then
           ActivateIndex;
         FSourceDataset.First;
         value := TACRVariant.Create;
         try
           while (not FSourceDataset.Eof) do
            begin
             if (FFilterExpr <> nil) then
              begin
               // check filter
               Res := TACRExpression(FFilterExpr).GetResult;
               if (not Res) then
                begin
                 FSourceDataset.Next;
                 continue;
                end;
              end;
             // check distinct
             if (FDistinctApplied) then
              begin
               bOK := IsRecordMatchesDistinct;
               if (not bOK) then
                begin
                 // skip this record
                 FSourceDataset.Next;
                 Continue;
                end;
              end; // distinct
             j := 0;
             TACRTable(FResultDataset).Insert;
             // not result table
             for i := 0 to FFieldCount-1 do
              begin
               if (FFieldLinks[i].IsHidden) then
                continue;
               // get field value
               GetFieldValue(value,i,False,False);
//               TACRDataset(FSourceDataset).GetFieldValue(value,i,False,False);
               // set value
               TACRDataset(FResultDataset).SetFieldValue(value,j,False);
               Inc(j);
               value.SetNull;
              end; // not result table
             // insert record
             TACRTable(FResultDataset).Post;
             // go to next record
             FSourceDataset.Next;
            end;
         finally
          if (FResultFieldsOrder.ItemCount > 0) then
           TACRTable(FResultDataset).ApplyProjection(FieldList,AliasList);
          value.Free;
          if (FDistinctApplied) then
            FinalizeDistinct
          else
           if ((FResultIndexFieldsList <> nil) and (FActivateIndexAfterMaterialize)) then
            if (TACRTable(FSourceDataset).IndexName = '') then
             ActivateIndex;
         end;
         if (s <> '') then
          raise EACRException.Create(10851,ErrorLCannotCopyRecords,[s]);
         FinalizeMaterialize(False);
       except
        FResultDataset.Free;
        FResultDataset := FSourceDataset;
        raise;
       end;
     end
    else
     begin
      TACRDataset(FSourceDataset).SetSQLFilter(FFilterExpr);
      if (FResultFieldsOrder.ItemCount > 0) then
       begin
         for i := 0 to FResultFieldsOrder.ItemCount-1 do
          begin
           j := FResultFieldsOrder.Items[i];
           if (FFieldLinks[j].IsHidden) then
            raise EACRException.Create(11498,ErrorLCannotAccessHiddenField,[FTableName,
                    FFieldLinks[j].FieldName,FFieldLinks[j].DisplayName,j]);
           if (FFieldLinks[j].DisplayName = '') then
            raise EACRException.Create(11500,ErrorLEmptyDisplayName,
                    [FTableName,FFieldLinks[j].FieldName,j]);
           FieldList.Add(FFieldLinks[j].FieldName);
           AliasList.Add(FFieldLinks[j].DisplayName);
          end;
        FResultDataset.First;
        TACRTable(FResultDataset).ApplyProjection(FieldList,AliasList);
       end; // just set a projection
      FinalizeMaterialize(True);
     end; // Live query
  finally
   FIsMaterialized := True;
   FieldList.Free;
   AliasList.Free;
  end;
end; // DoMaterialize


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRAOTable.Create(
                      DatabaseName: string;
                      SessionName:  string;
                      TableName:    string;
                      TableAlias:   string;
                      bInMemory:    Boolean = false;
                      bRequestLive: Boolean = false;
                      Password:     String = ''
                      );
var i: integer;
begin
 InternalCreate(nil,nil,TableName,TableAlias);
 FSourceDataset := FResultDataset;
 FTableLocked := False;
 FRequestLive := bRequestLive;
// FRequestLive := True;
 FIsMaterialized := False;
 FIsAOTable := true;
 TACRTable(FResultDataset).DatabaseName := DatabaseName; 
 TACRTable(FResultDataset).SessionName := SessionName; 
 TACRTable(FResultDataset).TableName := TableName;
 TACRTable(FResultDataset).InMemory := bInMemory;
 if (bInMemory) then 
  if (not TACRTable(FResultDataset).Exists) then 
   raise EACRException.Create(11505,ErrorLTableDoesNotExist,[TableName]);
 try
  FResultDataset.Open; 
 except
  raise EACRException.Create(10303,ErrorLCannotOpenTable,[TableName,Word(bInMemory)]); 
 end;
 FFieldCount := FResultDataset.FieldCount; 
 SetLength(FFieldLinks,FFieldCount);
 for i := 0 to FFieldCount-1 do 
  begin
   FFieldLinks[i].AO := nil; 
   FFieldLinks[i].Dataset := FResultDataset; 
   FFieldLinks[i].FieldNo := i;
   FFieldLinks[i].IsHidden := True; 
   FFieldLinks[i].IsExpression := False; 
   FFieldLinks[i].IsAggregate  := False; 
   FFieldLinks[i].FieldName := TACRTable(FResultDataset).AdvFieldDefs.Items[i].Name; 
   FFieldLinks[i].DisplayName := ''; 
   FFieldLinks[i].TableName := FTableName;
   FFieldLinks[i].TableAlias := FTableAlias;
   FFieldLinks[i].FieldType := TACRTable(FResultDataset).AdvFieldDefs.Items[i].DataType;
   FFieldLinks[i].FieldSize := TACRTable(FResultDataset).AdvFieldDefs.Items[i].Size;
   FFieldLinks[i].BLOBCompressionAlgorithm :=
    Byte(TACRTable(FResultDataset).AdvFieldDefs.Items[i].BLOBCompressionAlgorithm); 
   FFieldLinks[i].BLOBCompressionMode := TACRTable(FResultDataset).AdvFieldDefs.Items[i].BLOBCompressionMode;
   FFieldLinks[i].BLOBBlockSize := TACRTable(FResultDataset).AdvFieldDefs.Items[i].BLOBBlockSize; 
  end;
end; // Create 
 
	
//------------------------------------------------------------------------------ 
// destroy
//------------------------------------------------------------------------------ 
destructor TACRAOTable.Destroy; 
begin
  if (FTableLocked) then
   TACRTable(FResultDataset).UnlockTable; 
  inherited; 
end; // Destroy 
 
 
//------------------------------------------------------------------------------ 
// sets projection for TACRAOTable 
//------------------------------------------------------------------------------
procedure TACRAOTable.SetResultFields(var FieldRefs: array of TACRSelectListItem;
          bDistinct: Boolean); 
{
var i,j: integer;
    tableName: string; 
    tableAlias:string;
    tName:     string;
}    
begin 
 inherited SetResultFields(FieldRefs,bDistinct); 
{
 j := Length(FieldRefs);
 if (j <= 0) then 
  begin
   inherited SetResultFields(FieldRefs,bDistinct);
   Exit; 
  end; 
 tableName := AnsiUpperCase(FTableName);
 tableAlias := AnsiUpperCase(FTableAlias);
 for i := 0 to j-1 do
  begin 
   tName := AnsiUpperCase(FieldRefs[i].TableName); 
   if (tName <> '') then 
    if (tableName <> tName) and (tableAlias <> tName) then 
     raise EACRException.Create(30327,ErrorLInvalidTableNameInReference,
              [tableName,tableAlias,tName]); 
   if (FieldRefs[i].AllFields) then 
     begin 
      // no projection
//      FieldList.Free;
//      AliasList.Free;
      Exit; 
     end;
   FieldList.Add(FieldRefs[i].FieldName);
   AliasList.Add(FieldRefs[i].Pseudonym);
  end; 
 inherited SetResultFields(FieldRefs,bDistinct); 
 }
end; // SetResultsFields
	
	
//------------------------------------------------------------------------------
// lock table
//------------------------------------------------------------------------------ 
procedure TACRAOTable.LockTable;
begin
  if (not FTableLocked) then 
   if ((not TACRTable(FResultDataset).InMemory) and
       (not TACRTable(FResultDataset).Temporary)) then
    begin 
     TACRTable(FResultDataset).LockTable; 
     FTableLocked := True; 
    end; 
end; // LockTable
 
 
//------------------------------------------------------------------------------
// unlock table
//------------------------------------------------------------------------------
procedure TACRAOTable.UnlockTable;
begin 
  if (FTableLocked) then 
   if ((not TACRTable(FResultDataset).InMemory) and
       (not TACRTable(FResultDataset).Temporary)) then
    begin 
     TACRTable(FResultDataset).UnlockTable; 
     FTableLocked := False; 
    end;
end; // LockTable
	
	
////////////////////////////////////////////////////////////////////////////////
// 
// TACRAOJoin
// 
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// fix result fields that were renamed (_x) if it is possible
//------------------------------------------------------------------------------
procedure TACRAOJoin.FixResultFieldNames;
var i,j:    Integer;
    bFound: Boolean;
begin
 for i := 0 to FFieldCount - 1 do
  if (not FFieldLinks[i].IsHidden) then
   if (FFieldLinks[i].DisplayName <> '') then
    if (FFieldLinks[i].FieldName <> FFieldLinks[i].OriginalFieldName) then
     if (FFieldLinks[i].DisplayName = FFieldLinks[i].FieldName) then
      begin
        bFound := False;
        for j := 0 to FFieldCount - 1 do
         if (j <> i) then
          if (not FFieldLinks[j].IsHidden) then
           if (FFieldLinks[j].DisplayName <> '') then
            if (AnsiUpperCase(FFieldLinks[i].OriginalFieldName) =
               AnsiUpperCase(FFieldLinks[j].OriginalFieldName)) then
             begin
              bFound := True;
              break;
             end;
        if (not bFound) then
         FFieldLinks[i].DisplayName := FFieldLinks[i].OriginalFieldName;
      end;
end; // FixResultFieldNames


//------------------------------------------------------------------------------
// records are called Equal if all their join attributes are equal
//------------------------------------------------------------------------------
procedure TACRAOJoin.CompareRecords;
var i,j,k:            Integer;
    value1, value2:   TACRVariant;
begin
 FLeftAONull := False;
 FRightAONull := False;
 value1 := TACRVariant.Create;
 value2 := TACRVariant.Create;
 try
   for i := 0 to FFields1.ItemCount - 1 do
    begin
     // get first value
     j := FFields1.Items[i];
     GetFieldValue(value1,j);

     // get second value
     k := FFields2.Items[i];
     GetFieldValue(value2,k);

     // compare values
     try
      FCompareResult := value1.Compare(value2,False,False,False);
     except
      on e: EACRException do
       begin
         raise EACRException.Create(10328,ErrorLCompareFieldValues,
          [FFieldLinks[j].FieldName,FFieldLinks[k].FieldName,e.Message]);
       end;
      else
       raise;
     end;
     if (FCompareResult <> cmprEqual) then
       break;
    end;
 finally
   value1.Free;
   value2.Free;
 end; // try .. finally
end; // CompareRecords


{$IFNDEF NEW_JOIN_ALGORITHM}
//------------------------------------------------------------------------------
// init
//------------------------------------------------------------------------------
procedure TACRAOJoin.Init1(LeftFields,RightFields: String);
begin
 FLeftAO.SetIndex(LeftFields,'','','');
 FRightAO.SetIndex(RightFields,'','','');
end; // Init1


//------------------------------------------------------------------------------
// first
//------------------------------------------------------------------------------
procedure TACRAOJoin.InternalFirst1;
begin
 // inner ot outer join
 FEqualStarted := false;
 FLeftAO.First;
 FRightAO.First;
 if (FOuterJoin) then
  begin
   if (FLeftAO.Eof and FRightAO.Eof) then
    FEof := true
   else
   if ((FJoinType = ajtLeftOuter) and (FLeftAO.Eof)) then
    FEof := true
   else
   if ((FJoinType = ajtRightOuter) and (FRightAO.Eof)) then
    FEof := true;
  end
 else
  begin
   if (FLeftAO.Eof or FRightAO.Eof) then
    FEof := true;
  end;
 if (not FEof) then
  begin
   FEqualStarted := false;
   FFirstTimeCalled := true;
   InternalNext;
  end;
end; // InternalFirst1


//------------------------------------------------------------------------------
// next
//------------------------------------------------------------------------------
procedure TACRAOJoin.InternalNext1;
var FOldBookmark:       Pointer;
    bStopAtFirstMoving: Boolean;
begin
 bStopAtFirstMoving := false;
 // inner or outer
 // called by first or after materializing current record
 if (FFirstTimeCalled) then
  begin
   FFirstTimeCalled := false;
   if (FRightAO.Eof) then
    begin
     FRightAONull := True;
     FLeftAONull := False;
     Exit;
    end;
   if (FLeftAO.Eof) then
    begin
     FRightAONull := False;
     FLeftAONull := True;
     Exit;
    end;
   CompareRecords;
   if (FCompareResult = cmprEqual) then
    begin
     FEqualStarted := true;
     if (FBookmark <> nil) then
      FRightAO.ResultDataset.FreeBookmark(FBookmark);
     FBookmark := FRightAO.ResultDataset.GetBookmark;
     Exit; // record is valid - all ok
    end;
   if (FOuterJoin) then
    begin
     if ((FJoinType = ajtFullOuter) or (FJoinType = ajtLeftOuter))
        and
        ((FCompareResult = cmprLower) or
        (FCompareResult = cmprLeftNull)) then
      begin
       FLeftAONull := false;
       FRightAONull := true;
       Exit; // record is valid - all ok
      end;
     if ((FJoinType = ajtFullOuter) or (FJoinType = ajtRightOuter))
        and
        ((FCompareResult = cmprGreater) or
        (FCompareResult = cmprBothNull) or
        (FCompareResult = cmprRightNull)) then
      begin
       FLeftAONull := true;
       FRightAONull := false;
       Exit; // record is valid - all ok
      end;
     bStopAtFirstMoving := true;
    end; // outer join
  end; // called by first or after materializing current record

 if ((FLeftAO.Eof or FRightAO.Eof) and (not EOF)) then
  begin
   if (FRightAO.Eof) then
    begin
     FLeftAO.Next;
     FRightAONull := True;
     FLeftAONull := False;
     Exit;
    end;
   if (FLeftAO.Eof) then
    begin
     FRightAO.Next;
     FRightAONull := False;
     FLeftAONull := True;
     Exit;
    end;
  end; // left or right AO is at eof

 // if equal records were found at previous call
 if (FEqualStarted) then
  begin
   FRightAO.Next; // FRightAO.Eof will be true only if we were on last record
   if (FRightAO.Eof) then
    begin
      if (FLeftAO.Eof) then
        Exit // Eof reached
      else
       begin
        FLeftAO.Next;
        if (FLeftAO.Eof) then
         Exit; // Eof reached
        CompareRecords;
        if (FCompareResult = cmprEqual) then
         begin
          FRightAO.ResultDataset.GotoBookmark(FBookmark);
          Exit; // record is valid - all ok
         end;
        // we are at the last RightAO record
        // and have some records at LeftAO that
        // were not processed
        if ((FJoinType = ajtLeftOuter) or (FJoinType = ajtFullOuter)) then
         begin
          FEqualStarted := false;
          FLeftAONull := false;
          FRightAONull := true;
          Exit; // record is valid - all ok
         end;
        // there is no records left
        FEof := true;
        Exit;
       end; // left AO moved to next record successfully
    end; // right AO at the last record
   // check new record
   CompareRecords;
   if (FCompareResult = cmprEqual) then
    Exit; // record is valid - all ok
   // rightAO record is greater than leftAO
   FLeftAO.Next;
   if (FLeftAO.Eof) then
    begin
      if ((FJoinType = ajtRightOuter) or (FJoinType = ajtFullOuter)) then
       begin
        FEqualStarted := false;
        FLeftAONull := true;
        FRightAONull := false;
        Exit; // record is valid - all ok
       end;
      // there is no records left
      FEof := true;
      Exit;
    end;
   // get bookmark
   FOldBookmark := FRightAO.FResultDataset.GetBookmark;
   try
     // move right AO to the first equal record
     // check if left AO has records some equal records that were not joined
     // with right AO records
     FRightAO.FResultDataset.GotoBookmark(FBookmark);
     CompareRecords;
     if (FCompareResult = cmprEqual) then
      Exit; // record is valid - all ok
     // we have successfully joined equal records - now let's go forward
     // return to previous position (both ao after last previously "equal" records)
     FRightAO.FResultDataset.GotoBookmark(FOldBookmark);
   finally
     FRightAO.FResultDataset.FreeBookmark(FOldBookmark);
   end;
   CompareRecords;
   if (FCompareResult = cmprEqual) then
    begin
     // new found records
     if (FBookmark <> nil) then
      FRightAO.ResultDataset.FreeBookmark(FBookmark);
     FBookmark := FRightAO.FResultDataset.GetBookmark;
     Exit; // record is valid - all ok
    end;
   // now we have not equal records
   FEqualStarted := false;
   if (FOuterJoin) then
    begin
     if ((FJoinType = ajtFullOuter) or (FJoinType = ajtLeftOuter))
        and
        ((FCompareResult = cmprLower) or
//          (FCompareResult = cmprBothNull) or
        (FCompareResult = cmprLeftNull)) then
      begin
        begin
         FLeftAONull := false;
         FRightAONull := true;
        end;
       Exit; // record is valid - all ok
      end;
     if ((FJoinType = ajtFullOuter) or (FJoinType = ajtRightOuter))
        and
        ((FCompareResult = cmprGreater) or
        (FCompareResult = cmprBothNull) or
        (FCompareResult = cmprRightNull)) then
      begin
       FLeftAONull := true;
       FRightAONull := false;
       Exit; // record is valid - all ok
      end;
     bStopAtFirstMoving := true;
    end; // outer join
  end; // equal records were found at previous call
 // searching for next valid record
 while (not Eof) do
  begin
   if (FCompareResult = cmprEqual)  then
    raise EACRException.Create(10329,ErrorLEqualRecordsInJoins);
   if (FLeftAO.Eof) and
      ((FCompareResult = cmprLower) or
       (FCompareResult = cmprLeftNull)) then
    begin
     // there is no right records
     if (FInnerJoin or (FJoinType = ajtLeftOuter)) then
      begin
       FEof := true;
       Exit; // there is no more joined records
      end;
     if (not bStopAtFirstMoving) then
      begin
        FRightAO.Next;
       // if Eof - exiting
       if (FRightAO.Eof) then
         Exit;
      end;
     FLeftAONull := true;
     FRightAONull := false;
     // only not equal records left
     Exit; // record is valid - all ok
    end // LeftAO.Eof
   else
   if (FRightAO.Eof) and
      ((FCompareResult = cmprGreater) or
       (FCompareResult = cmprBothNull) or
       (FCompareResult = cmprRightNull)) then
    begin
     // there is no right records
     if (FInnerJoin or (FJoinType = ajtRightOuter)) then
      begin
       FEof := true;
       Exit; // there is no more records
      end;
     if (not bStopAtFirstMoving) then
      begin
        FLeftAO.Next;
       // if Eof - exiting
       if (FLeftAO.Eof) then
         Exit;
      end;
     FLeftAONull := false;
     FRightAONull := true;
     // only not equal records left
     Exit; // record is valid - all ok
    end // RightAO.Eof
   else
   if ((FCompareResult = cmprLower) or
       (FCompareResult = cmprLeftNull)) then
    begin
     if (bStopAtFirstMoving) then
      if ((FJoinType = ajtLeftOuter) or (FJoinType = ajtFullOuter)) then
       begin
        FLeftAONull := false;
        FRightAONull := true;
        Exit; // record is valid - all ok
       end; // Lower or LeftNull again

     FLeftAO.Next;
     // we are at last record
     if (FLeftAO.Eof) then
      begin
       // next found record will be at result dataset with left part = null
       if ((FJoinType = ajtLeftOuter) or (FJoinType = ajtFullOuter)) then
        bStopAtFirstMoving := true;
       continue;
      end;

     CompareRecords;
     if (FCompareResult = cmprEqual) then
      begin
       FEqualStarted := true;
       if (FBookmark <> nil) then
        FRightAO.ResultDataset.FreeBookmark(FBookmark);
       FBookmark := FRightAO.FResultDataset.GetBookmark;
       Exit; // record is valid - all ok
      end;
     if (FCompareResult = cmprLower) or
        (FCompareResult = cmprLeftNull) then
      if ((FJoinType = ajtLeftOuter) or (FJoinType = ajtFullOuter)) then
       begin
        FLeftAONull := false;
        FRightAONull := true;
        Exit; // record is valid - all ok
       end; // Lower or LeftNull again
    if (FCompareResult = cmprGreater) then
     bStopAtFirstMoving := true;
//       continue; // check Eof and search again
    end // Lower or LeftNull
   else
    if ((FCompareResult = cmprGreater) or
       (FCompareResult = cmprBothNull) or
       (FCompareResult = cmprRightNull)) then
    begin
     if (bStopAtFirstMoving) then
      if ((FJoinType = ajtRightOuter) or (FJoinType = ajtFullOuter)) then
       begin
        FLeftAONull := true;
        FRightAONull := false;
        Exit; // record is valid - all ok
       end; // Greater, BothNull or RightNull again

     FRightAO.Next;
     // we are at last record
     if (FRightAO.Eof) then
      begin
       // next found record will be at result dataset with right part = null
       if ((FJoinType = ajtRightOuter) or (FJoinType = ajtFullOuter)) then
         bStopAtFirstMoving := true;
       continue;
      end;


     CompareRecords;
     if (FCompareResult = cmprEqual) then
      begin
       FEqualStarted := true;
       if (FBookmark <> nil) then
        FRightAO.ResultDataset.FreeBookmark(FBookmark);
       FBookmark := FRightAO.FResultDataset.GetBookmark;
       Exit; // record is valid - all ok
      end;
    if ((FCompareResult = cmprGreater) or
       (FCompareResult = cmprBothNull) or
       (FCompareResult = cmprRightNull)) then
      if ((FJoinType = ajtRightOuter) or (FJoinType = ajtFullOuter)) then
       begin
        FLeftAONull := true;
        FRightAONull := false;
        Exit; // record is valid - all ok
       end; // Greater, BothNull or RightNull again
    if ((FCompareResult = cmprLower) or
        (FCompareResult = cmprLeftNull)) then
     bStopAtFirstMoving := true;
//       continue; // check Eof and search again
    end // Greater, BothNull or RightNull
  end; // while not Eof - searching for valid records
end; // InternalNext1


//------------------------------------------------------------------------------
// EOF
//------------------------------------------------------------------------------
function TACRAOJoin.InternalGetEof1: Boolean;
begin
 if (FEof) then
  begin
   Result := True;
  end
 else
  begin
   Result := FLeftAO.Eof and FRightAO.Eof;
   if (Result) then
     if (FBookmark <> nil) then
       begin
        FRightAO.ResultDataset.FreeBookmark(FBookmark);
        FBookmark := nil;
       end;
  end;
end; // InternalGetEof1
{$ENDIF}


function TACRAOJoin.IsIndexExists(Table: TDataset; FieldNames: TStringList): Boolean;
var i,j,k:     Integer;
    s,name:    String;
    bString:   Boolean;
begin
  Result := False;
  for i := 0 to TACRTable(Table).IndexDefs.Count - 1 do
   for j := 0 to FieldNames.Count - 1 do
    begin
     name := AnsiUpperCase(FieldNames[j]);
     bString := IsStringFieldType(TACRTable(Table).AdvFieldDefs.Find(name).DataType);
     s := AnsiUpperCase(TACRTable(Table).IndexDefs[i].Fields);
     k := Pos(';',s);
     if (k > 0) then
      begin
       s := Copy(s,1,k-1);
      end;
     if (name = s) then
      begin
       Result := True;
       if (bString) then
        begin
         if (ixCaseInsensitive in TACRTable(Table).IndexDefs[i].Options) then
          Result := False;
         if (Result) then
          if (Pos(name,AnsiUpperCase(TACRTable(Table).IndexDefs[i].CaseInsFields)) > 0) then
           Result := False;
        end;
       if (Result) then
        Exit;
      end; // index for this field found
    end;
end; // IsIndexExists


//------------------------------------------------------------------------------
// Choose Scan Table
//------------------------------------------------------------------------------
procedure TACRAOJoin.ChooseScanTable;
var bIndexExists1, bIndexExists2: Boolean;
    RecCount1, RecCount2:         Int64;
begin
 if (FJoinType = ajtLeftOuter) then
  begin
   FScanLeft := True;
  end // left
 else
 if (FJoinType = ajtRightOuter) then
  begin
   FScanLeft := False;
  end // right
 else
  begin
   bIndexExists1 := IsIndexExists(FLeftAO.ResultDataset, FFieldNames1);
   bIndexExists2 := IsIndexExists(FRightAO.ResultDataset, FFieldNames2);
   if (bIndexExists1 <> bIndexExists2) then
    begin
     FScanLeft := bIndexExists2;
    end
   else
    begin
     RecCount1 := TACRTable(FLeftAO.ResultDataset).Handle.GetRecordCount;
     RecCount2 := TACRTable(FRightAO.ResultDataset).Handle.GetRecordCount;
     if ((not TACRTable(FLeftAO.ResultDataset).InMemory) and
         (not TACRTable(FLeftAO.ResultDataset).Temporary)) then
      if (TACRTable(FLeftAO.ResultDataset).AdvFieldDefs.IsVarcharExists) then
       RecCount1 := RecCount1 * ACRVarcharSlowDownRate;
     if ((not TACRTable(FRightAO.ResultDataset).InMemory) and
         (not TACRTable(FRightAO.ResultDataset).Temporary)) then
      if (TACRTable(FRightAO.ResultDataset).AdvFieldDefs.IsVarcharExists) then
       RecCount2 := RecCount2 * ACRVarcharSlowDownRate;
     FScanLeft := (RecCount1 <= RecCount2);
    end;
  end; // inner or full
 if (FScanLeft) then
  begin
   FScanTable := FLeftAO.ResultDataset;
   FFilterTable := FRightAO.ResultDataset;
  end
 else
  begin
   FScanTable := FRightAO.ResultDataset;
   FFilterTable := FLeftAO.ResultDataset;
  end;
end; // ChooseScanTable


//------------------------------------------------------------------------------
// Build Filter Expression
//------------------------------------------------------------------------------
procedure TACRAOJoin.BuildFilterExpression;
var Expr:     TACRExpression;
    i:        Integer;
    count:    Integer;
    Cursor:   TACRCursor;
    RootNode: TACRExprNode;


  function AddComparison(FieldName: String): TACRExprNode;
  var NodeField, NodeConst: TACRExprNode;
  begin
   NodeField := TACRExprNodeField.Create(Cursor,FieldName,TACRTable(FFilterTable).TableName);
   NodeConst := TACRExprNodeConst.Create(False,False);
   FConstList.Add(NodeConst);
   Result := TACRExprNodeComparison.Create(doEQ,NodeField,NodeConst,False,False);
  end;

begin
 RootNode := TACRExprNodeBoolean.Create(doAND,False,False);
 Expr := TACRExpression.Create(RootNode);
 FTempFilterExpr := Expr;
 count := FFields1.ItemCount;
 Cursor := TACRTable(FFilterTable).Handle;
 for i := 0 to count -1 do
  begin
   if (FScanLeft) then
    Expr.AddNode(AddComparison(FFieldNames2.Strings[i]))
   else
    Expr.AddNode(AddComparison(FFieldNames1.Strings[i]));
  end;
end; // BuildFilterExpression


//------------------------------------------------------------------------------
// prepare filter for current record
//------------------------------------------------------------------------------
procedure TACRAOJoin.PrepareFilterForCurrentRecord;
var
    i,FieldNo:  Integer;
    v:          TACRVariant;
begin
 for i := 0 to FConstList.Count - 1 do
  begin
   v := TACRExprNodeConst(FConstList[i]).GetDataValue;
   if (FScanLeft) then
    FieldNo := FFields1.Items[i]
   else
    FieldNo := FFields2.Items[i];
//TACRTable(FScanTable).
   GetFieldValue(v,FieldNo,True,False);
  end;
end; // PrepareFilterForCurrentRecord


//------------------------------------------------------------------------------
// apply filter and return true if records were found
//------------------------------------------------------------------------------
function TACRAOJoin.ApplyFilterAndCheckIfRecordsFound: Boolean;
var Cursor: TACRCursor;
begin
 Result := not FFilterTableEmpty;
 if (Result) then
  begin
   PrepareFilterForCurrentRecord;
   Cursor := TACRTable(FFilterTable).Handle;
   Cursor.FilterExpression := FTempFilterExpr;
   FFilterApplied := True;
   Cursor.DisableRecordBitmap;
   FFilterTable.First;
   Result := (FFilterTable.RecordCount > 0);
   if (Result) then
    begin
     FLeftAONull := False;
     FRightAONull := False;
    end;
  end;
end; // ApplyFilter


//------------------------------------------------------------------------------
// clear filter
//------------------------------------------------------------------------------
procedure TACRAOJoin.ClearFilter;
var Cursor: TACRCursor;
begin
 if (FFilterApplied) then
  begin
   Cursor := TACRTable(FFilterTable).Handle;
   Cursor.DisableRecordBitmap;
   Cursor.FilterExpression := nil;
   FFilterTable.First;
   FFilterApplied := False;
  end;
end; // ClearFilter


//------------------------------------------------------------------------------
// Left table is null
//------------------------------------------------------------------------------
procedure TACRAOJoin.SetLeftNull;
begin
  FRightAONull := False;
  FLeftAONull := True;
end; // SetLeftNull


//------------------------------------------------------------------------------
// Right table is null
//------------------------------------------------------------------------------
procedure TACRAOJoin.SetRightNull;
begin
  FRightAONull := True;
  FLeftAONull := False;
end; // SetRightNull


//------------------------------------------------------------------------------
// for full join: switch to scanning rest records in FFilterTable (not found FScanTable)
//------------------------------------------------------------------------------
procedure TACRAOJoin.SwitchToScanningRest;
begin
  ClearFilter;
  FSearchingCorrespondingRecords := False;
  FScanningRestRecords := True;
  if (FScanLeft) then
   SetLeftNull
  else
   SetRightNull;
  FScannedRecords.Sort;
end; // SwitchToScanningRest


//------------------------------------------------------------------------------
// store current record id
//------------------------------------------------------------------------------
procedure TACRAOJoin.StoreCurrentRecordID;
var RecordID: TACRRecordID;
begin
  RecordID := TACRTable(FFilterTable).GetCurrentRecordID;
  FScannedRecords.Append(RecordID);
end; // StoreCurrentRecordID


{$IFDEF NEW_JOIN_ALGORITHM}
//------------------------------------------------------------------------------
// init
//------------------------------------------------------------------------------
procedure TACRAOJoin.Init2;
begin
  FLeftAONull := False;
  FRightAONull := False;
  ChooseScanTable;
  BuildFilterExpression;
  FFilterApplied := False;
  TACRTable(FFilterTable).Handle.RandomOrder := True;
  FEof := False;
  if (FJoinType = ajtFullOuter) then
   FScannedRecords := TACRRecordIDArray.Create(0,FScanTable.RecordCount div 10,FScanTable.RecordCount);
end; // Init2


//------------------------------------------------------------------------------
// first
//------------------------------------------------------------------------------
procedure TACRAOJoin.InternalFirst2;
begin
 Init2;
 FScanTable.First;
 FFilterTable.First;
 FSearchingCorrespondingRecords := True;
 FScanningRestRecords := False;
 FFilterTableEmpty := (FFilterTable.RecordCount <= 0);
 if (FJoinType = ajtFullOuter) then
   if (FScanTable.Eof) then
     SwitchToScanningRest;
 if ((not FScanningRestRecords) and (not InternalGetEof)) then
  begin
   if (ApplyFilterAndCheckIfRecordsFound) then
    begin
     if (FJoinType = ajtFullOuter) then
      StoreCurrentRecordID;
    end
   else
    begin
     if (FJoinType = ajtInner) then
      InternalNext2
     else
      begin
       if (FScanLeft) then
        SetRightNull
       else
        SetLeftNull;
      end; // outer join
    end;
  end; // scanning scan table
end; // InternalFirst2


//------------------------------------------------------------------------------
// next
//------------------------------------------------------------------------------
procedure TACRAOJoin.InternalNext2;
begin
 repeat
  if (FScanningRestRecords) then
   begin
    FFilterTable.Next;
    if (not FFilterTable.Eof) then
     if (FScannedRecords.FindRecordByID(TACRTable(FFilterTable).GetCurrentRecordID) < 0) then
      Break;
   end // scanning rest records in full outer join
  else
  if (FFilterApplied) then
   begin
    FFilterTable.Next;
    if (not FFilterTable.Eof) then
     begin
      if (FJoinType = ajtFullOuter) then
       StoreCurrentRecordID;
      Break;
     end;
    ClearFilter;
   end // scanning corresponding records
  else
   begin
    FScanTable.Next;
    if (FScanTable.Eof) then
     begin
      if (FJoinType = ajtFullOuter) then
       begin
        SwitchToScanningRest;
        if (not FFilterTable.Eof) then
         if (FScannedRecords.FindRecordByID(TACRTable(FFilterTable).GetCurrentRecordID) < 0) then
          Break;
        Continue;
       end
      else
       Break;
     end;
    if (ApplyFilterAndCheckIfRecordsFound) then
     begin
      if (FJoinType = ajtFullOuter) then
       StoreCurrentRecordID;
      Break;
     end
    else
     if (FJoinType <> ajtInner) then
      begin
       if (FScanLeft) then
        SetRightNull
       else
        SetLeftNull;
       Break;
      end; // outer join
   end; // scanning FScanTable
 until (Eof);
end; // InternalNext2


//------------------------------------------------------------------------------
// EOF
//------------------------------------------------------------------------------
function TACRAOJoin.InternalGetEof2: Boolean;
begin
 if (FJoinType = ajtFullOuter) then
  Result := (FFilterTable.Eof and FScanningRestRecords)
 else
 if ((FJoinType = ajtInner) and FFilterTableEmpty) then
  Result := True
 else
  Result := FScanTable.Eof;
end; // InternalGetEof2
{$ENDIF}


//------------------------------------------------------------------------------
// go to first record
//------------------------------------------------------------------------------
procedure TACRAOJoin.InternalFirst;
begin
 FEof := False;
 if (FDekart) then
  begin
   FLeftAO.First;
   FRightAO.First;
   if (FLeftAO.Eof or FRightAO.Eof) then
    FEof := true;
  end
 else
  begin
   {$IFDEF NEW_JOIN_ALGORITHM}
    InternalFirst2;
   {$ELSE}
    InternalFirst1;
   {$ENDIF}
  end; // inner ot outer join
end; // First


//------------------------------------------------------------------------------
// go to next record
//------------------------------------------------------------------------------
procedure TACRAOJoin.InternalNext;
begin
 if (FDekart) then
  begin
   FRightAO.Next;
   if (FRightAO.Eof) then
    begin
     FLeftAO.Next;
     if (not FLeftAO.Eof) then
      FRightAO.First;
    end // Eof
  end // Dekart
 else
  begin
   {$IFDEF NEW_JOIN_ALGORITHM}
    InternalNext2;
   {$ELSE}
    if (Eof) then Exit;
    InternalNext1;
   {$ENDIF}
  end; // inner or outer
end; // InternalNext


//------------------------------------------------------------------------------
// return true if cursor points to the last record
//------------------------------------------------------------------------------
function TACRAOJoin.InternalGetEof: Boolean;
begin
 if (FDekart) then
  Result := FLeftAO.Eof or FRightAO.Eof
 else
  begin
   {$IFDEF NEW_JOIN_ALGORITHM}
    Result := InternalGetEOF2;
   {$ELSE}
    Result := InternalGetEOF1;
   {$ENDIF}
  end;
end; // InternalGetEof


//------------------------------------------------------------------------------
// return number of records
//------------------------------------------------------------------------------
function TACRAOJoin.InternalGetRecordCount: Integer;
begin
 Result := 0;
 if (FIsMaterialized) then
  Result := FResultDataset.RecordCount
 else
 begin
   if (FDekart) then
    Result := FLeftAO.RecordCount * FRightAO.RecordCount;
 end;
end; // GetRecordCount


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRAOJoin.Create(
                      LeftChild:  TACRAO;
                      RightChild: TACRAO;
                      JoinType:   TACRJoinType;
                      IsNatural:  Boolean = False;
                      FieldList1: TACRFields = nil; // join fields
                      FieldList2: TACRFields = nil // field1 = field2
                      );
var i,j,k,l1,l2:            integer;
    LeftFields,RightFields: string;

 procedure FillIndexNames;
 var s: String;
 begin
   if (FFieldLinks[j].AO = FLeftAO) then
    begin
      s := FLeftAO.FFieldLinks[FFieldLinks[j].FieldNo].FieldName;
      FFieldNames1.Add(s);
      if (LeftFields = '') then
       LeftFields := s
      else
       LeftFields := LeftFields + ';' + s;
    end // LeftAO
   else
    begin
      s := FRightAO.FFieldLinks[FFieldLinks[j].FieldNo].FieldName;
      FFieldNames2.Add(s);
      if (RightFields = '') then
       RightFields := s
      else
       RightFields := RightFields + ';' + s;
    end; // RightAO
 end;


begin
 InternalCreate(LeftChild,RightChild);
 FFieldNames1 := nil;
 FFieldNames2 := nil;
 FConstList := nil;
 FBookmark := nil;
 FScannedRecords := nil;
 FTempFilterExpr := nil;
 FJoinType := JoinType;
 FInnerJoin := (FJoinType = ajtInner);
 FDekart := (FJoinType = ajtCross);
 FOuterJoin := (FJoinType = ajtLeftOuter) or
               (FJoinType = ajtRightOuter) or
               (FJoinType = ajtFullOuter);
 l1 := 0;
 l2 := 0;
 if (not FDekart) and (not IsNatural) then
  begin
   l1 := FieldList1.ItemCount;
   l2 := FieldList2.ItemCount;
   if (l1 <> l2) then
     raise EACRException.Create(10330,ErrorLDifferentListsLength,[l1,l2]);
  end;
 FEof := false;
 FEqualStarted := false;
 FFirstTimeCalled := true;
 FFields1 :=  TACRIntegerArray.Create(0,1,FLeftAO.FieldCount);
 FFields2 := TACRIntegerArray.Create(0,1,FRightAO.FieldCount);
 // creating field links for none-union joins
 FFieldCount := FLeftAO.FieldCount + FRightAO.FieldCount;
 SetLength(FFieldLinks,FFieldCount);
 // linking left child
 j := FLeftAO.FieldCount-1;
 for i := 0 to j do
  begin
   FFieldLinks[i].FieldNo := i;
   FFieldLinks[i].Dataset := nil;
   FFieldLinks[i].AO := FLeftAO;
   FFieldLinks[i].FieldName := FLeftAO.FFieldLinks[i].FieldName;
   FFieldLinks[i].OriginalFieldName := FFieldLinks[i].FieldName;
   FFieldLinks[i].DisplayName := '';
   FFieldLinks[i].TableName := FLeftAO.FFieldLinks[i].TableName;
   FFieldLinks[i].TableAlias := FLeftAO.FFieldLinks[i].TableAlias;
   FFieldLinks[i].FieldType := FLeftAO.FFieldLinks[i].FieldType;
   FFieldLinks[i].FieldSize := FLeftAO.FFieldLinks[i].FieldSize;
   FFieldLinks[i].IsHidden := FLeftAO.FFieldLinks[i].IsHidden;
   FFieldLinks[i].IsExpression := FLeftAO.FFieldLinks[i].IsExpression;
   FFieldLinks[i].IsAggregate := FLeftAO.FFieldLinks[i].IsAggregate;
   FFieldLinks[i].BLOBCompressionAlgorithm := FLeftAO.FFieldLinks[i].BLOBCompressionAlgorithm;
   FFieldLinks[i].BLOBCompressionMode := FLeftAO.FFieldLinks[i].BLOBCompressionMode;
   FFieldLinks[i].BLOBBlockSize := FLeftAO.FFieldLinks[i].BLOBBlockSize;
  end;
 inc(j);
 // linking right child
 for i := 0 to FRightAO.FieldCount-1 do
  begin
   FFieldLinks[j+i].FieldNo := i;
   FFieldLinks[j+i].Dataset := nil;
   FFieldLinks[j+i].AO := FRightAO;
   FFieldLinks[j+i].FieldName := FRightAO.FFieldLinks[i].FieldName;
   FFieldLinks[j+i].OriginalFieldName := FFieldLinks[j+i].FieldName;
   FFieldLinks[j+i].DisplayName := '';
   FFieldLinks[j+i].TableName := FRightAO.FFieldLinks[i].TableName;
   FFieldLinks[j+i].TableAlias := FRightAO.FFieldLinks[i].TableAlias;
   FFieldLinks[j+i].FieldType := FRightAO.FFieldLinks[i].FieldType;
   FFieldLinks[j+i].FieldSize := FRightAO.FFieldLinks[i].FieldSize;
   FFieldLinks[j+i].IsHidden := FRightAO.FFieldLinks[i].IsHidden;
   FFieldLinks[j+i].IsExpression := FRightAO.FFieldLinks[i].IsExpression;
   FFieldLinks[j+i].IsAggregate := FRightAO.FFieldLinks[i].IsAggregate;
   FFieldLinks[j+i].BLOBCompressionAlgorithm := FRightAO.FFieldLinks[i].BLOBCompressionAlgorithm;
   FFieldLinks[j+i].BLOBCompressionMode := FRightAO.FFieldLinks[i].BLOBCompressionMode;
   FFieldLinks[j+i].BLOBBlockSize := FRightAO.FFieldLinks[i].BLOBBlockSize;
   if (IsNatural) then
    for k := 0 to j-1 do
     if (AnsiUpperCase(FFieldLinks[j+i].FieldName) =
          AnsiUpperCase(FFieldLinks[k].FieldName)) then
      begin
       FFields1.Append(k);
       FFields2.Append(j+i);
       FFieldLinks[k].IsHidden := false;
       FFieldLinks[j+i].IsHidden := false;
       FLeftAO.FieldExists(FFieldLinks[k].FieldName,'',True);
       FRightAO.FieldExists(FFieldLinks[j+i].FieldName,'',True);
      end; // natural join
  end;
 if (FInnerJoin or FOuterJoin)  then
  begin
   if (not IsNatural) then
    begin
     // creating join fields lists
     for i := 0 to l1-1 do
      begin
       // join fields should not be hidden
       j := FieldExists(
        FieldList1.Items[i].FieldName,
        FieldList1.Items[i].TableName,True,FFields1);
       if (j <> 1) then
         raise EACRException.Create(10331,ErrorLInvalidJoinField,
          [FieldList1.Items[i].TableName,FieldList1.Items[i].FieldName,i,j]);
      end;
     for i := 0 to l2-1 do
      begin
       j := FieldExists(
        FieldList2.Items[i].FieldName,
        FieldList2.Items[i].TableName,True,FFields2);
       if (j <> 1) then
         raise EACRException.Create(10332,ErrorLInvalidJoinField,
          [FieldList2.Items[i].TableName,FieldList2.Items[i].FieldName,i,j]);
      end;
    end; // not natural

   if (FFields1.ItemCount <> FFields2.ItemCount) then
    raise EACRException.Create(10332,ErrorLDifferentListsLength,[FFields1.ItemCount,FFields2.ItemCount]);
   if (FFields1.ItemCount <= 0) then
    raise EACRException.Create(10333,ErrorLEmptyFieldList,[FFields1.ItemCount]);
   LeftFields := '';
   RightFields := '';
   FFieldNames1 := TStringList.Create;
   FFieldNames2 := TStringList.Create;
   FConstList := TList.Create;
   // filling Left and Right fields
   for i := 0 to FFields1.ItemCount-1 do
    begin
     j := FFields1.Items[i];
     k := FFields2.Items[i];
     if (FFieldLinks[j].AO = FFieldLinks[k].AO) then
      raise EACRException.Create(10334,ErrorLBothJoinAOEqual,[FFieldLinks[j].FieldName,
        FFieldLinks[k].FieldName,j,k])
     else
      if (FFieldLinks[j].AO = FRightAO) then
       begin
        // swap left / right join fields
        FFields1.Items[i] := k;
        FFields2.Items[i] := j;
        j := FFields1.Items[i];
        k := FFields2.Items[i];
       end;
     FillIndexNames;
     j := k;
     FillIndexNames;
    end;
    {$IFNDEF NEW_JOIN_ALGORITHM}
    Init1(LeftFields,RightFields);
    {$ENDIF}
  end; // inner or outer joins
  SetFieldNames;
  FIsMaterialized := False;
end; // create;


//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------
destructor TACRAOJoin.Destroy;
begin
 if (FFields1 <> nil) then
   FFields1.Free;
 if (FFields2 <> nil) then
  FFields2.Free;
 if (FFieldNames1 <> nil) then
   FFieldNames1.Free;
 if (FFieldNames2 <> nil) then
  FFieldNames2.Free;
 if (FConstList <> nil) then
  FConstList.Free;
 if (FTempFilterExpr <> nil) then
  FTempFilterExpr.Free;
 if (FScannedRecords <> nil) then
  FScannedRecords.Free;
 inherited Destroy;
end; // destroy


////////////////////////////////////////////////////////////////////////////////
//
// TACRAOUnion
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// records are called Equal if all their join attributes are equal
// return true if search ao has record like in scan ao
//------------------------------------------------------------------------------
function TACRAOUnion.CompareRecords: Boolean;
var i,j,k:   integer;
    value: TACRVariant;
begin
 if (FSearchAO.RecordCount <= 0) then
  Result := False
 else
  begin
   TACRTable(FSearchAO.FResultDataset).SetKey;
   value := TACRVariant.Create;
   try
    for i := 0 to FFields1.ItemCount - 1 do
     begin
      j := FFields1.Items[i];
      k := FFields2.Items[i];
      if (FScanAO = FLeftAO) then
       begin
        TACRDataset(FLeftAO.FResultDataset).GetFieldValue(value,j,False);
        if (not value.IsNull) then
         TACRDataset(FRightAO.FResultDataset).SetFieldValue(value,k,False);
       end
      else
       begin
        TACRDataset(FRightAO.FResultDataset).GetFieldValue(value,k,False);
        if (not value.IsNull) then
         TACRDataset(FLeftAO.FResultDataset).SetFieldValue(value,j,False);
       end;
     end;
   finally
     value.Free;
   end;
   Result := TACRTable(FSearchAO.FResultDataset).GotoKey;
  end;
end; // CompareRecords


//------------------------------------------------------------------------------
// search next record starting from current position
//------------------------------------------------------------------------------
procedure TACRAOUnion.SearchRecord;
var b: Boolean;

 function IsRecordSuitable: Boolean;
 begin
  Result := ((FUnionType = autExcept) and (not b)) or
            ((FUnionType = autIntersect) and b);
 end; // IsRecordSuitable

begin
  b := CompareRecords;
  while ((not Eof) and (not IsRecordSuitable)) do
   begin
    FScanAO.Next;
    if (not Eof) then
     b := CompareRecords;
   end;
end; // SearchRecord


//------------------------------------------------------------------------------
// show recrod from LeftAO
//------------------------------------------------------------------------------
procedure TACRAOUnion.ShowLeftAO;
var i,k: Integer;
begin
 k := 0;
 for i := 0 to FFieldCount-1 do
  begin
   FFieldLinks[i].Dataset := FLeftAO.FResultDataset;
   FFieldLinks[i].FieldNo := FFields1.Items[k];
   inc(k);
  end;
end; // ShowLeftAO


//------------------------------------------------------------------------------
// show recrod from RightAO
//------------------------------------------------------------------------------
procedure TACRAOUnion.ShowRightAO;
var i,k: Integer;
begin
 k := 0;
 for i := 0 to FFieldCount-1 do
  begin
   FFieldLinks[i].Dataset := FRightAO.FResultDataset;
   FFieldLinks[i].FieldNo := FFields2.Items[k];
   inc(k);
  end;
end; // ShowRightAO


//------------------------------------------------------------------------------
// first
//------------------------------------------------------------------------------
procedure TACRAOUnion.InternalFirst;
begin
 FLeftAO.First;
 FRightAO.First;
 if (Eof) then Exit;
 if (FUnionType <> autUnion) then
  SearchRecord;
end; // InternalFirst


//------------------------------------------------------------------------------
// next
//------------------------------------------------------------------------------
procedure TACRAOUnion.InternalNext;
begin
 if (Eof) then Exit;
 if (FUnionType = autUnion) then
  begin
   if (not FShowLeft) then
    FRightAO.Next
   else
    FLeftAO.Next;
   if (FShowLeft and (FLeftAO.Eof) and (not FRightAO.Eof)) then
    begin
     ShowRightAO;
     FShowLeft := false;
     // switch to the right AO
    end;
  end
 else
  begin
   FScanAO.Next;
   if (not Eof) then
    SearchRecord;
  end;
end; // InternalNext



//------------------------------------------------------------------------------
// Eof
//------------------------------------------------------------------------------
function TACRAOUnion.InternalGetEof: Boolean;
begin
 case FUnionType of
  autUnion: Result := FLeftAO.Eof and FRightAO.Eof;
  autExcept, autIntersect: Result := FScanAO.Eof;
 end;
end; // InternalGetEof


//------------------------------------------------------------------------------
// return recordcount
//------------------------------------------------------------------------------
function TACRAOUnion.InternalGetRecordCount: Integer;
begin
 result := 0;
 if (FIsMaterialized) then
  result := FResultDataset.RecordCount;
end; // InternalGetRecordCount


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRAOUnion.Create(
                      LeftChild:  TACRAO;
                      RightChild: TACRAO;
                      UnionType:   TACRUnionType;
                      IsCorresponding:  Boolean = False;
                      bDistinct: Boolean = True;
                      FieldList: TACRFields = nil // corresponding fields
                      );
var LeftIndexFieldNames: string;

procedure AddFieldToFieldLists(LeftIndex: integer; RightIndex: integer);
var name: string;
begin
      if ((TACRTable(FLeftAO.FResultDataset).AdvFieldDefs[LeftIndex].DataType =
          TACRTable(FRightAO.FResultDataset).AdvFieldDefs[RightIndex].DataType) or
          (
            (TACRTable(FLeftAO.FResultDataset).AdvFieldDefs[LeftIndex].DataType in [aftInteger,aftAutoInc]) and
            (TACRTable(FRightAO.FResultDataset).AdvFieldDefs[RightIndex].DataType in [aftInteger,aftAutoInc])
          )) then
       if (((UnionType = autUnion) and (bDistinct)) or
           IsConvertableFieldType(TACRTable(FLeftAO.FResultDataset).
            AdvFieldDefs[LeftIndex].DataType)) then
        begin
         inc(FFieldCount);
         SetLength(FFieldLinks,FFieldCount);
         name := FLeftAO.FResultDataset.Fields[LeftIndex].FieldName;
         FFieldLinks[FFieldCount-1].FieldName := name;
         FFieldLinks[FFieldCount-1].DisplayName := name;
         FFieldLinks[FFieldCount-1].TableName := FLeftAO.FFieldLinks[LeftIndex].TableName;
         FFieldLinks[FFieldCount-1].TableAlias := FLeftAO.FFieldLinks[LeftIndex].TableAlias;
         FFieldLinks[FFieldCount-1].FieldType := TACRTable(FLeftAO.FResultDataset).AdvFieldDefs[LeftIndex].DataType;
         FFieldLinks[FFieldCount-1].FieldSize := Max(TACRTable(FLeftAO.FResultDataset).AdvFieldDefs[LeftIndex].Size,
                                                     TACRTable(FRightAO.FResultDataset).AdvFieldDefs[RightIndex].Size);

         FFieldLinks[FFieldCount-1].IsHidden := false;
         FFieldLinks[FFieldCount-1].AO := nil;
         FFieldLinks[FFieldCount-1].Dataset := FLeftAO.FResultDataset;
         FFieldLinks[FFieldCount-1].FieldNo := LeftIndex;
         FFieldLinks[FFieldCount-1].IsExpression := False;
         FFieldLinks[FFieldCount-1].IsAggregate := False;

         FFields1.Append(LeftIndex);
         FFields2.Append(RightIndex);
         if (LeftIndexFieldNames = '') then
          LeftIndexFieldNames := name
         else
          LeftIndexFieldNames := LeftIndexFieldNames + ';' +  name;
        end;
end; // AddFieldToFieldLists


var i,j,k,n,f1,f2: integer;
    name: string;
    bNoFields: Boolean;
begin
 InternalCreate(LeftChild,RightChild);
 FIsMaterialized := False;
 FFirstTimeCalled := false;
 FLeftAONull := false;
 FRightAONull := false;
 FShowLeft := true;
 FUnionType := UnionType;
 bNoFields := false;
 if (FieldList = nil) then
  bNoFields := true;

 if (UnionType = autExcept) then
  begin
//   FRightAO.SetIndex(RightIndexFieldNames,'','','',True);
   FRightAO.ResultIndexByAllFields := True;
   FSearchAO := FRightAO;
   FScanAO := FLeftAO;
  end
 else
 if (UnionType = autIntersect) then
  begin
   if (not FLeftAO.FIsAOTable) then
    begin
     FSearchAO := FLeftAO;
     FScanAO := FRightAO;
     FLeftAO.ResultIndexByAllFields := True;
    end
   else
    begin
     FSearchAO := FRightAO;
     FScanAO := FLeftAO;
     FRightAO.ResultIndexByAllFields := True;
    end;
  end;

 FLeftAO.LockTable;
 FLeftAO.Execute;
 FRightAO.LockTable;
 FRightAO.Execute;


 LeftIndexFieldNames := '';
 FFields1 :=  TACRIntegerArray.Create(0,1,FLeftAO.FieldCount);
 FFields2 := TACRIntegerArray.Create(0,1,FRightAO.FieldCount);
 if bNoFields then
  begin
   // scanning all fields in FLeftAO
    for i := 0 to FLeftAO.FResultDataset.FieldCount-1 do
    begin
     k := -1;
     if (IsCorresponding) then
      begin
       name := AnsiUpperCase(FLeftAO.FResultDataset.Fields[i].FieldName);
       for j := 0 to FRightAO.FResultDataset.FieldCount-1 do
        begin
         if (name = AnsiUpperCase(FRightAO.FResultDataset.Fields[j].FieldName)) then
          begin
           k := j;
           break;
          end;
        end;
      end // searching for corresponding field
     else
      begin
       if (i < FRightAO.FResultDataset.FieldCount) then
        k := i;
      end; // corresponding fields by number
     if (k >= 0) then
      AddFieldToFieldLists(i,k);
    end; // scanning all fields in FLeftAO
  end // no corresponding fields were specified
 else
  begin
   // scanning FieldList
   n := FieldList.ItemCount;
   if (n <= 0) then
    raise EACRException.Create(10337,ErrorLEmptyFieldList,[n]);
   for i := 0 to n-1 do
    begin
     // store field name
     name := AnsiUpperCase(FieldList.Items[i].FieldName);
     // scanning LeftAO
     f1 := -1;
     for j := 0 to FLeftAO.FResultDataset.FieldCount-1 do
      begin
       if (name = AnsiUpperCase(FLeftAO.FResultDataset.Fields[j].FieldName)) then
        begin
         f1 := j;
         break;
        end;
      end; // scanning LeftAO
     // scanning RightAO
     f2 := -1;
     for j := 0 to FRightAO.FResultDataset.FieldCount-1 do
      begin
       if (name = AnsiUpperCase(FRightAO.FResultDataset.Fields[j].FieldName)) then
        begin
         f2 := j;
         break;
        end;
      end; // scanning RightAO
     if (f1 >= 0) and (f2 >= 0) then
      begin
        AddFieldToFieldLists(f1,f2);
      end; // adding field to the union field list
    end; // scanning FieldList
  end; // corresponding fields were specified
 if (FFields1.ItemCount <> FFields2.ItemCount) then
  raise EACRException.Create(10338,ErrorLDifferentListsLength,
        [FFields1.ItemCount,FFields2.ItemCount]);
 if (FFields1.ItemCount <= 0) then
  raise EACRException.Create(10339,ErrorLEmptyFieldList,[FFields1.ItemCount]);
 if (bDistinct) then
   Self.ApplyDistinct(LeftIndexFieldNames);
end; // Create


//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------
destructor TACRAOUnion.Destroy;
begin
 if (FFields1 <> nil) then
   FFields1.Free;
 if (FFields2 <> nil) then
   FFields2.Free;
 inherited Destroy;
end; // destroy


////////////////////////////////////////////////////////////////////////////////
//
// TACRAOTableExpr
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// first
//------------------------------------------------------------------------------
procedure TACRAOTableExpr.InternalFirst;
begin
 FLeftAO.First;
end; // InternalFirst


//------------------------------------------------------------------------------
// next
//------------------------------------------------------------------------------
procedure TACRAOTableExpr.InternalNext;
begin
 FLeftAO.Next;
end; // InternalNext


//------------------------------------------------------------------------------
// EOF
//------------------------------------------------------------------------------
function TACRAOTableExpr.InternalGetEof: Boolean;
begin
 result := FLeftAO.EOF;
end; // InternalGetEof


//------------------------------------------------------------------------------
// Record count
//------------------------------------------------------------------------------
function TACRAOTableExpr.InternalGetRecordCount: Integer;
begin
 result := FLeftAO.RecordCount;
end; // InternalGetRecordCount


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRAOTableExpr.Create(
                     Child: TACRAO
                    );
var i: integer;
begin
 InternalCreate(Child);
 FIsMaterialized := false;
 FFieldCount := FLeftAO.FieldCount;
 SetLength(FFieldLinks,FFieldCount);
 for i := 0 to FFieldCount-1 do
  begin
   FFieldLinks[i].AO := FLeftAO;
   FFieldLinks[i].Dataset := nil;
   FFieldLinks[i].FieldNo := i;
   FFieldLinks[i].IsHidden := True;
   FFieldLinks[i].FieldName := FLeftAO.FFieldLinks[i].FieldName;
   FFieldLinks[i].FieldType := FLeftAO.FFieldLinks[i].FieldType;
   FFieldLinks[i].TableName := FLeftAO.FFieldLinks[i].TableName;
   FFieldLinks[i].TableAlias := FLeftAO.FFieldLinks[i].TableAlias;
   FFieldLinks[i].FieldSize := FLeftAO.FFieldLinks[i].FieldSize;
   FFieldLinks[i].FieldPrecision := FLeftAO.FFieldLinks[i].FieldPrecision;
   FFieldLinks[i].IsExpression := FLeftAO.FFieldLinks[i].IsExpression;
   FFieldLinks[i].IsAggregate := FLeftAO.FFieldLinks[i].IsAggregate;
   FFieldLinks[i].BLOBCompressionAlgorithm := FLeftAO.FFieldLinks[i].BLOBCompressionAlgorithm;
   FFieldLinks[i].BLOBCompressionMode := FLeftAO.FFieldLinks[i].BLOBCompressionMode;
   FFieldLinks[i].BLOBBlockSize := FLeftAO.FFieldLinks[i].BLOBBlockSize;
  end;
end; // Create


////////////////////////////////////////////////////////////////////////////////
//
// TACRAOGroupBy
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// return true if new group is started from current record
//------------------------------------------------------------------------------
function TACRAOGroupBy.CompareRecords: Boolean;
var i,j:    Integer;
    value1: TACRVariant;
begin
 Result := False;
 value1 := TACRVariant.Create;
 try
   for i := 0 to FFields.ItemCount - 1 do
    begin
     // get Current value
     j := FFields.Items[i];
     FLeftAO.GetFieldValue(value1,j);
     // compare values
     if (not Result) then
      Result := (value1.Compare(FValues[i],False,False,False) <> cmprEqual);
     FValues[i].Assign(value1,True); 
    end;
 finally
   value1.Free;
 end;
end; // CompareRecords


//------------------------------------------------------------------------------
// initialize expressions
//------------------------------------------------------------------------------
procedure TACRAOGroupBy.Init;
var i: integer;
begin
 if (FAggregateExpressionsExists) then
  for i := 0 to FFieldCount-1 do
   if ((FFieldLinks[i].IsAggregate) and (FFieldLinks[i].Expr <> nil)) then
     TACRExpression(FFieldLinks[i].Expr).Init;
 if (FFilterExpr <> nil) then
  if (TACRExpression(FFilterExpr).IsAggregated) then
   TACRExpression(FFilterExpr).Init;
end; // initialize expressions


//------------------------------------------------------------------------------
// accumulate aggregate expressions
//------------------------------------------------------------------------------
procedure TACRAOGroupBy.Accumulate;
var i: integer;
begin
 if (FAggregateExpressionsExists) then
  for i := 0 to FFieldCount-1 do
   if ((FFieldLinks[i].IsAggregate) and (FFieldLinks[i].Expr <> nil)) then
     TACRExpression(FFieldLinks[i].Expr).Accumulate;
 if (FFilterExpr <> nil) then
  if (TACRExpression(FFilterExpr).IsAggregated) then
   TACRExpression(FFilterExpr).Accumulate;
end; // accumulate aggregate expressions


//------------------------------------------------------------------------------
// scan group
//------------------------------------------------------------------------------
procedure TACRAOGroupBy.ScanGroup;
var
    bCompare: Boolean;
begin
 Init;
 CompareRecords;
 Accumulate;
 FLeftAO.Next;
 while (not FLeftAO.Eof) do
  begin
   bCompare := CompareRecords;
   if (bCompare) then
    begin
     FLeftAO.FResultDataset.Prior;
     break;
    end;
   Accumulate;
   FLeftAO.Next;
  end;
end; // ScanGroup


//------------------------------------------------------------------------------
// first
//------------------------------------------------------------------------------
procedure TACRAOGroupBy.InternalFirst;
begin
 FLeftAO.First;
 FEof := False;
 if (FLeftAO.Eof) then
  begin
   if (FAggregateExpressionsExists) then
    begin
     FAllFieldsNull := True;
     Init;
    end
   else
    FEof := True
  end
 else
  begin
   if (FAllFields) then
    begin
     Init;
     while (not FLeftAO.Eof) do
      begin
       Accumulate;
       FLeftAO.Next;
      end;
    end
   else
    ScanGroup;
  end;
end; // InternalFirst


//------------------------------------------------------------------------------
// next
//------------------------------------------------------------------------------
procedure TACRAOGroupBy.InternalNext;
begin
 if (FAllFields) then
   FEof := True
 else
  begin
   FLeftAO.Next;
   FEof := FLeftAO.Eof;
   if (not FEof) then
    ScanGroup;
  end;
end; // InternalNext


//------------------------------------------------------------------------------
// EOF
//------------------------------------------------------------------------------
function TACRAOGroupBy.InternalGetEof: Boolean;
begin
 Result := FEof;
end; // InternalGetEof


//------------------------------------------------------------------------------
// Record count
//------------------------------------------------------------------------------
function TACRAOGroupBy.InternalGetRecordCount: Integer;
begin
 Result := FLeftAO.RecordCount;
end; // InternalGetRecordCount


//------------------------------------------------------------------------------
// sets projection for TACRAOTable
//------------------------------------------------------------------------------
procedure TACRAOGroupBy.SetResultFields(var FieldRefs: array of TACRSelectListItem;
          bDistinct: Boolean);
var i,j,k,x:      integer;
    fname,tname:  string;
    fno:          TACRIntegerArray;
begin
 j := Length(FieldRefs);
 if (j <= 0) then
  begin
   inherited SetResultFields(FieldRefs,False);
   Exit;
  end;
 inherited SetResultFields(FieldRefs,False);
 // check for invalid fields in select list
 fno := TACRIntegerArray.Create(0,1,1);
 for i := 0 to j-1 do
  begin
   if (FieldRefs[i].IsExpression) then continue;
   fname := FieldRefs[i].FieldName;
   tname := FieldRefs[i].TableName;
   fno.SetSize(0);
   if (FieldExists(fname,tname,False,fno) <= 0) then
    begin
     fno.Free;
     raise EACRException.Create(10341,ErrorLCannotFindFieldInTable,[fname,tname]);
    end;
   for k := 0 to fno.ItemCount-1 do
    begin
     x := fno.Items[k];
     if (not FFields.IsValueExists(x)) then
      begin
       fno.Free;
       raise EACRException.Create(10342,ErrorLFieldDoesNotIncludedInGroupByList,
        [tname,fname,x,FFieldLinks[x].FieldName]);
      end;
    end;
  end;
 fno.Free;
end; // SetResultFields


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRAOGroupBy.Create(
                     Child: TACRAO;
                     FieldList: TACRFields = nil // corresponding fields
                    );
var i,j,k: integer;
    GroupByFields: string;
begin
 InternalCreate(Child);
 try
   FAllFields := (FieldList = nil);
   if (not FAllFields) then
    if (FieldList.ItemCount = 0) then
     FAllFields := true;
   FIsAOGroupBy := true;

   FIsMaterialized := false;
   FFields := TACRIntegerArray.Create(0,1,FLeftAO.FieldCount);
   if (not FAllFields) then
    begin
     SetLength(FValues,FieldList.ItemCount);
     for i := 0 to FieldList.ItemCount-1 do
      FValues[i] := TACRVariant.Create;
    end;
   try
     FFieldCount := FLeftAO.FieldCount;
     SetLength(FFieldLinks,FFieldCount);
     for i := 0 to FFieldCount-1 do
      begin
       FFieldLinks[i].AO := FLeftAO;
       FFieldLinks[i].Dataset := nil;
       FFieldLinks[i].FieldNo := i;
       FFieldLinks[i].IsHidden := True;
       FFieldLinks[i].FieldName := FLeftAO.FFieldLinks[i].FieldName;
       FFieldLinks[i].FieldType := FLeftAO.FFieldLinks[i].FieldType;
       FFieldLinks[i].FieldSize := FLeftAO.FFieldLinks[i].FieldSize;
       FFieldLinks[i].TableName := FLeftAO.FFieldLinks[i].TableName;
       FFieldLinks[i].TableAlias := FLeftAO.FFieldLinks[i].TableAlias;
       FFieldLinks[i].IsExpression := FLeftAO.FFieldLinks[i].IsExpression;
       FFieldLinks[i].IsAggregate := FLeftAO.FFieldLinks[i].IsAggregate;
       FFieldLinks[i].BLOBCompressionAlgorithm := FLeftAO.FFieldLinks[i].BLOBCompressionAlgorithm;
       FFieldLinks[i].BLOBCompressionMode := FLeftAO.FFieldLinks[i].BLOBCompressionMode;
       FFieldLinks[i].BLOBBlockSize := FLeftAO.FFieldLinks[i].BLOBBlockSize;
      end;
     if (not FAllFields) then
      begin
       for i := 0 to FieldList.ItemCount-1 do
        begin
         // unhide Group BY field in children and move it to FFields array
         j := FieldExists(
               FieldList.Items[i].FieldName,
               FieldList.Items[i].TableName,True,FFields);
          if (j <> 1) then
               raise EACRException.Create(10343,ErrorLInvalidJoinField,
                [FieldList.Items[i].TableName,FieldList.Items[i].FieldName,i,j]);
         k := FFields.Items[FFields.ItemCount-1];
         if (GroupByFields = '') then
          GroupByFields := FFieldLinks[k].FieldName
         else
          GroupByFields := GroupByFields + ';' + FFieldLinks[k].FieldName;
        end;
       FLeftAO.SetIndex(GroupByFields,'','','',True);
      end; // not all fields
   except
    FFields.Free;
    FFields := nil;
    raise;
   end;
  except
   FResultDataset.Free;
   FResultDataset := nil;
   FResultFieldsOrder.Free;
   FResultFieldsOrder := nil;
   raise;
  end;
end; // Create


//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------
destructor TACRAOGroupBy.Destroy;
var i: Integer;
begin
 if (FFields <> nil) then
  FFields.Free;
 if (not FAllFields) then
  begin
   for i := Low(FValues) to High(FValues) do
    FValues[i].Free;
   FValues := nil;
  end;
 inherited Destroy;
end; // destroy


////////////////////////////////////////////////////////////////////////////////
//
// TACRAOSQLTopRowCount
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// go to first record
//------------------------------------------------------------------------------
procedure TACRAOSQLTopRowCount.InternalFirst;
begin
  FLeftAO.First;
end;// InternalFirst


//------------------------------------------------------------------------------
// go to next record
//------------------------------------------------------------------------------
procedure TACRAOSQLTopRowCount.InternalNext;
begin
  FLeftAO.Next;
end;// InternalNext


//------------------------------------------------------------------------------
// return true if EOF
//------------------------------------------------------------------------------
function TACRAOSQLTopRowCount.InternalGetEof: Boolean;
begin
  Result := FLeftAO.GetEof;
end;// InternalGetEof


//------------------------------------------------------------------------------
// return record count
//------------------------------------------------------------------------------
function TACRAOSQLTopRowCount.InternalGetRecordCount: Integer;
begin
  Result := FLeftAO.GetRecordCount;
end;// InternalGetRecordCount


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRAOSQLTopRowCount.Create(
                   Child: TACRAO
                  );
var i: integer;
begin
 InternalCreate(Child);
 FIsMaterialized := false;
 FFieldCount := FLeftAO.FieldCount;
 SetLength(FFieldLinks,FFieldCount);
 for i := 0 to FFieldCount-1 do
  begin
   FFieldLinks[i].AO := FLeftAO;
   FFieldLinks[i].Dataset := nil;
   FFieldLinks[i].FieldNo := i;
   FFieldLinks[i].IsHidden := True;
   FFieldLinks[i].FieldName := FLeftAO.FFieldLinks[i].FieldName;
   FFieldLinks[i].TableName := FLeftAO.FFieldLinks[i].TableName;
   FFieldLinks[i].TableAlias := FLeftAO.FFieldLinks[i].TableAlias;
   FFieldLinks[i].FieldType := FLeftAO.FFieldLinks[i].FieldType;
   FFieldLinks[i].FieldSize := FLeftAO.FFieldLinks[i].FieldSize;
   FFieldLinks[i].FieldPrecision := FLeftAO.FFieldLinks[i].FieldPrecision;
   FFieldLinks[i].IsExpression := FLeftAO.FFieldLinks[i].IsExpression;
   FFieldLinks[i].IsAggregate := FLeftAO.FFieldLinks[i].IsAggregate;
   FFieldLinks[i].BLOBCompressionAlgorithm := FLeftAO.FFieldLinks[i].BLOBCompressionAlgorithm;
   FFieldLinks[i].BLOBCompressionMode := FLeftAO.FFieldLinks[i].BLOBCompressionMode;
   FFieldLinks[i].BLOBBlockSize := FLeftAO.FFieldLinks[i].BLOBBlockSize;
  end;
end;// Create


//------------------------------------------------------------------------------
// convert FieldNames, DescFields, CaseInsFields
// to FieldList, AscDescList, CaseInsList
//------------------------------------------------------------------------------
procedure ConvertIndexFieldNamesToLists(
                FieldNames, DescFields, CaseInsFields:  String;
                FieldList, AscDescList, CaseInsList:    TStringList
                );
var sl:   TStringList;
    i,j:  Integer;
begin
  sl := TStringList.Create;
  try
  FillFieldNames(FieldList,FieldNames);
  FillFieldNames(sl,DescFields);
  for i := 0 to FieldList.Count-1 do
   begin
     j := sl.IndexOf(AnsiUpperCase(FieldList[i]));
     if (j >= 0) then
      AscDescList.Add(ACR_DESC)
     else
      AscDescList.Add(ACR_ASC);
   end;
  sl.Clear;
  FillFieldNames(sl,CaseInsFields);
  for i := 0 to FieldList.Count-1 do
   begin
     j := sl.IndexOf(AnsiUpperCase(FieldList[i]));
     if (j >= 0) then
      CaseInsList.Add(ACR_NO_CASE)
     else
      CaseInsList.Add(ACR_CASE);
   end;
  finally
    sl.Free;
  end;
end; // ConvertIndexFieldNames


//------------------------------------------------------------------------------
// convert FieldList, AscDescList, CaseInsList
// to FieldNames, DescFields, CaseInsFields
//------------------------------------------------------------------------------
procedure ConvertListsToIndexFieldNames(
                var FieldNames:     String;
                var DescNames:     String;
                var CaseInsNames:  String;
                FieldList, AscDescList, CaseInsList: TStringList
                );
var i: Integer;
begin
  FieldNames := '';
  DescNames := '';
  CaseInsNames := '';
  for i := 0 to FieldList.Count - 1 do
   begin
    if (FieldNames = '') then
      FieldNames := FieldList[i]
    else
      FieldNames := FieldNames + ';' + FieldList[i];
    if (AscDescList[i] = ACR_DESC) then
     begin
       if (DescNames = '') then
        DescNames := FieldList[i]
       else
        DescNames := DescNames + ';' + FieldList[i];
     end;
    if (CaseInsList[i] = ACR_NO_CASE) then
     begin
       if (CaseInsNames = '') then
        CaseInsNames := FieldList[i]
       else
        CaseInsNames := CaseInsNames + ';' + FieldList[i];
     end;
   end;
end; // ConvertListsToIndexFieldNames


// dummy procedure
function GetTempIndexName: String;
var t: TACRTable;
    fl,al,cl: TStringList;
begin
 Result := t.FindOrCreateIndex(fl,al,cl);
end; // dummy function

initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRRelationalAlgebra> initialized');
{$ENDIF}


end.
