unit ACRSQLProcessor;

{$I ACRVer.inc}

interface

uses Classes, SysUtils, DB,

     {$IFDEF DEBUG_LOG}
     ACRDebug,
     {$ENDIF}
     ACRLexer,
     ACRBase,
     ACRRelationalAlgebra,
     ACRTypes,
     ACRExpressions,
     ACRVariant,
     ACRCompression,
     ACRConst,
     ACRConverts,
     ACRExcept;

 type



////////////////////////////////////////////////////////////////////////////////
//
// TACRTableReference
//
////////////////////////////////////////////////////////////////////////////////

  // reference to the "table" in FROM clause
  TACRTableReference = class (TObject)
   public
    TableType:        TACRTableType;     // Table | JoinedTable | SubQuery
    SessionName:      string;             // session name
    DatabaseName:     string;             // database1
    TableName:        string;             // table1
    Pseudonym:        string;             // table1 as t1
    Password:         string;             // table1 PASSWORD 'password'
    InMemory:         Boolean;            // [MEMORY]
    NaturalJoin:      Boolean;            // Natural join?
    JoinType:         TACRJoinType;      // inner | left | ...
    LeftTable:        TACRTableReference; // left table in join
    RightTable:       TACRTableReference; // right table in join
    UsingFields:      TStringList;         // join column list
    SearchCondition:  TACRExpression;     // ON '(t1.Field1 = t2.Field2)'

    // creates
    constructor Create;
    // destroys
    destructor Destroy; override;
    // makes join (left and right node-table)
    procedure MakeJoin(RightNode: TACRTableReference; JType: TACRJoinType;
                     IsNatural: Boolean; Fields: TStringList;
                     OnCondition: TACRExpression);
  end;//TACRTableReference


////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLCommand
// base class for TACRSQLSelect, TACRSQLUpdate, ...
//
////////////////////////////////////////////////////////////////////////////////

   // base class for TACRSQLSelect, TACRSQLUpdate, ...
  TACRSQLCommand = class (TObject)
   protected
    LLex:         TACRLexer;   // lexer with expression to parse
    Token:        TToken;      // current token
    LQuery:       TDataSet;
    RowsAffected: Integer;
    IntoMemory:   Boolean;     // into memory table?
   private
    // parses list of columns (without table name): field1, field2, ...
    procedure ParseColumnList(var Fields: TStringList);
    // parses list of fields:  table.field1, field2, ..
    procedure ParseFieldList(var Fields: TACRFields);

    // get current token
    function GetCurrentToken: Boolean;
    // get next token
    function GetNextToken: Boolean; overload;
    // get token and lokks at next token with check for token type restrictions
    function GetNextToken(PermittedTypes: TTokenTypes;
                          NativeErrorCode: integer;
                          ErrorText: String): Boolean; overload;

   public
    // creates object
    constructor Create(Lexer: TACRLexer; Query: TDataSet);
    // parses query
    procedure Parse; virtual;
    // executes query
    procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); virtual; abstract;
    // get result cursor
    function GetResultCursor: TACRCursor; virtual;
    // gets result dataset
    function GetResultDataset: TDataset; virtual;
  end;//TACRSQLCommand



////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLCursorCommand
// base class for SQL command with cursor
//
////////////////////////////////////////////////////////////////////////////////


 TACRSQLCursorCommand = class (TACRSQLCommand)
  protected
   RootAO:                TACRAO;   // top level AO
   ResultDataset:         TDataset;  // query component
   IntoDatabase:          string;    // select into <database>.<table>
   IntoTable:             string;    // select into <database>.<table>
   FRequestLive:          Boolean;
   //IntoMemory:            Boolean;   // select into memory table?

  protected
   procedure LockAllTables; virtual; abstract;
   procedure UnlockAllTables; virtual; abstract;

  public
   // creates object
   constructor Create(Lexer: TACRLexer; Query: TDataSet);
   // destroys object
   destructor Destroy; override;
   // executes query
   //procedure ExecSQL(query: TDataset; IsRoot: Boolean = False); override;
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                    ); override;
   // builds AO tree
   function BuildAOTree(query: TDataset): TACRAO; virtual; abstract;

   // get result cursor
   function GetResultCursor: TACRCursor; override;
   // get result dataset
   function GetResultDataset: TDataset; override;
 end;//TACRSQLCursorCommand



////////////////////////////////////////////////////////////////////////////////
//
// TACRLocalSQLProcessor 
//
////////////////////////////////////////////////////////////////////////////////


  TACRLocalSQLProcessor = class (TACRSQLProcessor)
   private
    FQuery:         TACRSQLCommand;

   private

    // add SELECT query object
    procedure AddSelectQuery(Lexer: TACRLexer);
    // add INSERT query object
    procedure AddInsertQuery(Lexer: TACRLexer);
    // add UPDATE query object
    procedure AddUpdateQuery(Lexer: TACRLexer);
    // add DELETE query object
    procedure AddDeleteQuery(Lexer: TACRLexer);
    // DDL:
    // add CREATE TABLE query object
    procedure AddCreateTableQuery(Lexer: TACRLexer);
    // add DROP TABLE query object
    procedure AddDropTableQuery(Lexer: TACRLexer);
    // add ALTER TABLE
    procedure AddAlterTableQuery(Lexer: TACRLexer);
    // add RENAME TABLE
    procedure AddRenameTableQuery(Lexer: TACRLexer);
    // add CREATE INDEX
    procedure AddCreateIndex(Lexer: TACRLexer);
    // add DROP INDEX
    procedure AddDropIndexQuery(Lexer: TACRLexer);
    // add START TRANSACTION
    procedure AddStartTransactionQuery(Lexer: TACRLexer);
    // add COMMIT
    procedure AddCommitQuery(Lexer: TACRLexer);
    // add ROLLBACK
    procedure AddRollbackQuery(Lexer: TACRLexer);

   public
    constructor Create(Query: TDataSet);
    destructor Destroy; override;

    function OpenQuery: TACRCursor; override;
  end; // TACRLocalSQLProcessor



////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLSelect
//
////////////////////////////////////////////////////////////////////////////////


 TACRSQLSelect = class (TACRSQLCursorCommand)
  protected
   Distinct:              Boolean; // ALL | DISTINCT
   TopRowCount:           Integer; // TOP (-1 | n)
   FirstRowNo:            Integer; // TOP row count [, first row]
   AllFields:             Boolean; // Select *
   SelectList:            array of TACRSelectListItem; // fields list
   SelectListCount:       integer; // count of array elements
   FromTables:            array of TACRTableReference; // From clause
   FromTablesCount:       integer; // count of array elements
   SearchCondition:       TACRExpression;  // WHERE clause
   GroupByFields:         TACRFields; // GROUP BY field1, f2, ...
   HavingCondition:       TACRExpression;  // HAVING clause
   OrderBySpecs:          array of TACRSortSpecification;
   OrderBySpecsCount:     integer;
   OrderByIndex:          String; // indexName
   FTableList:            TList;

   // parse SELECT token
   function ParseSelectToken: Boolean;
   // DISTINCT | ALL ?
   function ParseSetQuantifier: Boolean;
   // TOP n ?
   function ParseTopOperator: Boolean;
   // table1.* | Table1.Field1 | Fields1 AS F1
   function ParseSelectSubList: Boolean;
   // * | <select sublist>
   function ParseSelectList: Boolean;
   // INTO <target>
   function ParseInto: Boolean;
   // ON <join condition>
   function ParseJoinCondition(var SearchCondition: TACRExpression): Boolean;
   // USING <join columns>
   function ParseNamedColumnsJoin(var Fields: TStringList): Boolean;
   // CROSS JOIN | INNER JOIN | ...
   function ParseJoin(var tblRef: TACRTableReference): Boolean;
   // <table name> [ [ AS ] <correlation name> ...
   function ParseTableReference(tblRef: TACRTableReference=nil): Boolean;
   // FROM ...
   function ParseFromClause: Boolean;
   // WHERE ...
   function ParseWhereClause: Boolean;
   // GROUP BY ...
   function ParseGroupByClause: Boolean;
   // HAVING ...
   function ParseHavingClause: Boolean;
   // <sort key> [ <collate clause> ] [ <ordering specification> ]
   function ParseSortSpecification: Boolean;
   // ORDER BY
   function ParseOrderByClause: Boolean;

   // gets default database name
   function GetDefaultDatabaseName: string;
   // creates and adjusts table AO
   function CreateTableAO(var TableRef: TACRTableReference): TACRAO;
   // creates and adjusts joined table AO
   function CreateJoinedTableAO(var TableRef: TACRTableReference): TACRAO;
   // creates and adjusts AO
   function CreateAO(var TableRef: TACRTableReference): TACRAO;
   // builds one-table AO
   function BuildOneTableTree: TACRAO;
   // builds multi-tables AO tree
   function BuildMultiTablesTree: TACRAO;
   // applies Order By clause
   procedure ApplyOrderBy(AO: TACRAO);

  protected
   procedure LockAllTables; override;
   procedure UnlockAllTables; override;
   procedure GetTableList(TableList: TList);
   procedure FreeTableList(TableList: TList);
   function LockTableListS(TableList: TList): Boolean;
   procedure UnlockTableListS(TableList: TList);
  public
   // creates object
   constructor Create(Lexer: TACRLexer; Query: TDataSet);
   // destroys object
   destructor Destroy; override;
   // parses query
   procedure Parse; override;
   // builds AO tree
   function BuildAOTree(query: TDataset): TACRAO; override;
 end;//TACRSQLSelect



////////////////////////////////////////////////////////////////////////////////
//
// TACRQueryExprNode
//
////////////////////////////////////////////////////////////////////////////////

 // select | union | except | ...
 TACRQueryExprNode = class (TObject)
  NodeType:         TACRQueryExprType;  // node is: select | union | except | ...
  Left:             TACRQueryExprNode;  // left node in union, except, ...
  Right:            TACRQueryExprNode;  // right table in union, except, ...
  All:              Boolean;            // [ALL] specified?
  Corresponding:    Boolean;            // [CORRESPONDING] specified?
  CorrespondingFields: TStringList;     // column list
  SelectCommand:    TACRSQLSelect;      // SELECT command (if NodeType is select)

  // creates
  constructor Create; overload;
  // creates copy
  constructor Create(Src: TACRQueryExprNode); overload;
  // destroys
  destructor Destroy; override;
  // adds new node to the tree
  procedure AddNode(NewType: TACRQueryExprType; RightNode: TACRQueryExprNode;
                    bAll, bCorresponding: Boolean; ColumnList: TStringList=nil);
 end;



////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLUnion
// UNION SQL command
//
////////////////////////////////////////////////////////////////////////////////


 TACRSQLUnion = class (TACRSQLCursorCommand)
  protected
   FRootNode: TACRQueryExprNode;         // root node in unions,excepts tree

   // parses [ALL]
   function ParseAll: Boolean;

   // parses [ <corresponding spec> ]
   function ParseCorrespondingSpec(var ColumnsList: TStringList): Boolean;

   // parses SELECT ...
   function ParseQuerySpecification: TACRQueryExprNode;

   // parses <query specification> | <table value constructor>  | <explicit table>
   function ParseSimpleTable: TACRQueryExprNode;

   // parses <simple table> |
   // <left paren> <non-join query expression> <right paren>
   function ParseNonJoinQueryPrimary: TACRQueryExprNode;

   // parses <non-join query primary> |
   // <query term> INTERSECT [ ALL ] [ <corresponding spec> ] <query primary>
   function ParseNonJoinQueryTerm: TACRQueryExprNode;

   // parses <non-join query term> |
   // <query expression> UNION  [ ALL ] [ <corresponding spec> ] <query term> |
   // <query expression> EXCEPT [ ALL ] [ <corresponding spec> ] <query term>
   function ParseNonJoinQueryExpression: TACRQueryExprNode;

   // parses <non-join query expression>  | <joined table>
   function ParseQueryExpression: TACRQueryExprNode;

   // builds AO
   function BuildAO(Node: TACRQueryExprNode): TACRAO;
   procedure LockAllTables; override;
   procedure UnlockAllTables; override;

  public
   // creates object
   constructor Create(Lexer: TACRLexer; Query: TDataSet);
   // destroys object
   destructor Destroy; override;
   // parses query
   procedure Parse; override;
   // builds AO tree
   function BuildAOTree(query: TDataset): TACRAO; override;

 end;//TACRSQLUnion



////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLInsert
//
////////////////////////////////////////////////////////////////////////////////


 TACRSQLInsert = class (TACRSQLCommand)
  protected
   TableName:             string;  // Table Name
   Password:              string;  // table password
   FieldNames:            TStringList;
   FieldValues:           TList;
   InternalSelecter:      TACRSQLSelect;
  protected
   procedure ParseValuesList;
  public
   // creates object
   constructor Create(Lexer: TACRLexer; Query: TDataSet);
   // destroys object
   destructor Destroy; override;
   // parses query
   procedure Parse; override;
   // executes query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override;
 end;//TACRSQLInsert



////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLUpdate
//
////////////////////////////////////////////////////////////////////////////////


 TACRSQLUpdate = class (TACRSQLCommand)
  protected
   TableName:             string;  // Table Name
   Password:              string;  // table password
   FieldNames:            TStringList;
   FieldValues:           TList;
   InternalSelecter:      TACRSQLSelect;
   //procedure SetFieldValue(field: TField; value: String);
  public
   // creates object
   constructor Create(Lexer: TACRLexer; Query: TDataSet);
   // parses query
   procedure Parse; override;
   // executes query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override;
 end;//TACRSQLUpdate



////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLDelete
//
////////////////////////////////////////////////////////////////////////////////


 TACRSQLDelete = class (TACRSQLCommand)
  protected
   InternalSelecter:      TACRSQLSelect;
   TableName:             string;  // Table Name
   FullDelete: boolean;
  public
   // creates object
   constructor Create(Lexer: TACRLexer; Query: TDataSet);
   // parses query
   procedure Parse; override;
   // executes query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override;
 end;//TACRSQLDelete



////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLTableManipulation
//
////////////////////////////////////////////////////////////////////////////////


 // FieldDef element
 TSQLFieldDef = class (TObject)
  public
   FieldName:                String;       // field name or pseudonym
   
   // FieldType
   newFieldType:             Boolean;
      FieldType:             TACRAdvancedFieldType;

   // field length ex: String(255)
   newLength:                Boolean;
      Length:                Integer;

   // is Requared field (NOT NULL)
   newRequired:              Boolean;
      Required:              Boolean;

   // Autoinc settings
   newAutoincIncrement:      Boolean;
      AutoincIncrement:      Int64;

   newAutoincLastValue:      Boolean;
      AutoincLastValue:      Int64;

   newAutoincMinValue:       Boolean;
      AutoincMinValue:       Int64;

   newAutoincMaxValue:       Boolean;
      AutoincMaxValue:       Int64;

   newAutoincCycled:         Boolean;
      AutoincCycled:         Boolean;

   // Blob settings
   newBlobBlockSize:         Boolean;
      BlobBlockSize:         Integer; // 1..

   newBlobCompressionAlgorithm: Boolean;
      BlobCompressionAlgorithm: TACRCompressionAlgorithm;// {NONE | ZLIB | BZIP | PPM}

   newBlobCompressionMode:   Boolean;
      BlobCompressionMode:   Byte; //     {0 .. 9}

   //Default Value
   newDefaultValue:          Boolean;
      DefaultValue:          TACRVariant;

   newMinValue:              Boolean;
      MinValue:              TACRVariant;

   newMaxValue:              Boolean;
      MaxValue:              TACRVariant;

   newPrimaryKey:            Boolean;
      PrimaryKey:            Boolean;

   newUnique:                Boolean;
      Unique:                Boolean;

  public
   constructor Create;
   destructor Destroy; override;
 end;

 
 TSQLFieldDefs = class (TObject)
  protected
   List: TList;
   function GetDef(Index: Integer): TSQLFieldDef;
  public
   constructor Create;
   destructor Destroy; override;
   function AddCreated: TSQLFieldDef;
   function GetCount: Integer;
  public
   property Items[Index: Integer]: TSQLFieldDef read GetDef; default;
   property Count: Integer read GetCount;
 end;


 TACRIndexField = record
   FieldName: string;       // field name or pseudonym
   desc:      boolean;      // desc/asc
   nocase:    boolean;      // ncase/case sensitive
 end;

 TACRDDLTableManipulation = class (TACRSQLCommand)
  protected
   FForeignKeyDefs:       TObject;
   Password:              String;  // table password
   TableName:             String;  // Table Name
   SQLFieldDefs:          TSQLFieldDefs; // SQLFieldDefs
   PrimaryIndexName:      String;
   //PrimaryKey:            TStringList;
   PrimaryKeyFields:      array of TACRIndexField;
   UniqueFields:          TStringList;

   // table element list
   procedure ParseTableElementList;
   // parse TableName token
   procedure ParseTableNameToken;
   // Fill table column type into Structure
   procedure FillColumnType(var FieldDef:TSQLFieldDef);
   // Fill column requared value into Structure
   procedure ParseColumnRequared(var FieldDef:TSQLFieldDef);
   // parse PASSWORD 'aaa'
   procedure ParsePassword;
   // parse DEFAULT {const | NULL}
   function ParseDefaultValue(var FieldDef:TSQLFieldDef): Boolean;
   // parse MINVALUE value
   function ParseMinValue(var FieldDef:TSQLFieldDef): Boolean;
   // parse MAXVALUE value
   function ParseMaxValue(var FieldDef:TSQLFieldDef): Boolean;
   // parse fiald ... PRIMARY KEY
   function ParseFieldPrimaryKey(var FieldDef:TSQLFieldDef): Boolean;
   // parse fiald ... UNIQUE
   function ParseFieldUnique(var FieldDef:TSQLFieldDef): Boolean;

   // parse Primary Key
   function ParsePrimaryKey: boolean;
   // parse foreign key
   function ParseForeignKey: Boolean;
   // go to the Next token and Parse Integer (Int64)
   function ParseInteger: Int64;
   // Create ACRTable object and fill ACRTable params
   function CreateInternalACRTable(query: TDataSet): TDataSet;
   // Fill AdvFieldDef
   procedure FillAdvFieldDef(AdvFieldDef: TFieldDef; SQLFieldDef: TSQLFieldDef);
   // Add Primary Key into ACRTable
   procedure AddPrimaryKey(T: TDataSet);
   // Add UNIQUE constraint and index into ACRTable
   procedure AddUnique(T: TDataSet);
   // Delete PrimaryKey
   procedure DeletePrimaryKey(T: TDataSet);
  public
   // creates object
   constructor Create(Lexer: TACRLexer; Query: TDataSet);
   // destroy
   destructor Destroy; override;
   // parses query
   procedure Parse; override;
   // executes query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override; abstract;
 end;//TACRDDLTableManipulation




////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLCreateTable
//
////////////////////////////////////////////////////////////////////////////////



 TACRDDLCreateTable = class (TACRDDLTableManipulation)
  protected
   // Create Table
   procedure CreateTable(query: TDataSet);
  public
   // parses query
   procedure Parse; override;
   // executes query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override;
 end;//TACRDDLCreateTable


////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLDropTable
//
////////////////////////////////////////////////////////////////////////////////


 TACRDDLDropTable = class (TACRDDLTableManipulation)
  protected
   FCascade: Boolean;
  protected
   // Drop Table
   procedure DropTable(query: TDataSet);
  public
   // parses query
   procedure Parse; override;
   // executes query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override;
 end;//TACRDDLDropTable



////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLAlterTable
//
////////////////////////////////////////////////////////////////////////////////


 // Types of alter table
 TAlterType = (atAdd,
               atDrop,
               atModify,
               atRenameColumn,
               atRenameTable,
               atDropConstraint,
               atAddConstraintPrimaryKey,
               atAddConstraintForeignKey
//               atAddConstraintUnique
               );

 TACRDDLAlterTable = class (TACRDDLTableManipulation)
  protected
   // AlterType token
   AlterType:                  TAlterType;
   NewPassword:                String;
   NewEncrypted:               Boolean;
   DropColumnNamesList:        TStringList;
   RenameColumnOldNamesList:   TStringList;
   RenameColumnNewNamesList:   TStringList;
   NewTableName:               String;
   DropConstraintName:         String;
   DropConstraintCascade:      Boolean;

   // parse AlterType token
   procedure ParseAlterTypeToken;
   // parse NewPassword token
   procedure ParseNewPasswordToken;
   // Parse RenameColumnList
   procedure ParseRenameColumnsList;
   // parse DROP CONSTRAINT
   procedure ParseDropConstraint;
   // Alter Table
   procedure DropColumn(T: TDataSet);
   procedure AddColumn(T: TDataSet);
   procedure Modify(T: TDataSet);
   procedure RenameColumn(T: TDataSet);
  public
   // creates object
   constructor Create(Lexer: TACRLexer; Query: TDataSet);
   // destroy
   destructor Destroy; override;
   // parse query
   procedure Parse; override;
   // execute query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override;
 end;//TACRDDLAlterTable


////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLRenameTable
//
////////////////////////////////////////////////////////////////////////////////

 TACRDDLRenameTable = class (TACRDDLAlterTable)
   // parse query
   procedure Parse; override;
 end;

////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLCreateIndex
//
////////////////////////////////////////////////////////////////////////////////

 TACRDDLCreateIndex = class (TACRSQLCommand)
  protected
   Unique:          Boolean;
   IfNotExists:     Boolean;
   IndexName:       String;
   TableName:       String;
   Password:        String;
   ACRIndexFields:  array of TACRIndexField; // Index Fields
  public
   // creates object
   constructor Create(Lexer: TACRLexer; Query: TDataSet);
   // parse query
   procedure Parse; override;
   // execute query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override;
 end;//TACRDDLCreateIndex



////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLDropIndex
//
////////////////////////////////////////////////////////////////////////////////


 TACRDDLDropIndex = class (TACRSQLCommand)
  protected
   IfExists:  Boolean;
   TableName: String;
   IndexName: String;
   Password:  String;
  public
   // creates object
   constructor Create(Lexer: TACRLexer; Query: TDataSet);
   // parse query
   procedure Parse; override;
   // execute query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override;
 end;//TACRDDLDropIndex


////////////////////////////////////////////////////////////////////////////////
//
// TACRStartTransaction
//
////////////////////////////////////////////////////////////////////////////////


 TACRStartTransaction = class (TACRSQLCommand)
  public
   // parse query
   procedure Parse; override;
   // execute query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override;
 end;//TACRStartTransaction


////////////////////////////////////////////////////////////////////////////////
//
// TACRCommit
//
////////////////////////////////////////////////////////////////////////////////


 TACRCommit = class (TACRSQLCommand)
  private
   FFlushFileBuffers: Boolean;
  public
   // parse query
   procedure Parse; override;
   // execute query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override;
 end;//TACRCommit


////////////////////////////////////////////////////////////////////////////////
//
// TACRRollback
//
////////////////////////////////////////////////////////////////////////////////


 TACRRollback = class (TACRSQLCommand)
  public
   // execute query
   procedure ExecSQL(
                      query: TDataset;
                      IsRoot,
                      RequestLive: boolean;
                      var ReadOnly: boolean
                     ); override;
 end;//TACRRollback


 // set field value from ACRVariant
 procedure SetACRVariantIntoField(Value: TACRVariant; Field: TField);



implementation


uses ACRMain;





////////////////////////////////////////////////////////////////////////////////
//
// TACRLocalSQLProcessor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// adds SELECT query object
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddSelectQuery(Lexer: TACRLexer);
begin
  FQuery := TACRSQLUnion.Create(Lexer, ACRQuery);
end;// AddSelectQuery


//------------------------------------------------------------------------------
// adds INSERT query object
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddInsertQuery(Lexer: TACRLexer);
begin
  FQuery := TACRSQLInsert.Create(Lexer, ACRQuery);
end;//AddInsertQuery


//------------------------------------------------------------------------------
// adds UPDATE query object
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddUpdateQuery(Lexer: TACRLexer);
begin
  FQuery := TACRSQLUpdate.Create(Lexer, ACRQuery);
end;//AddUpdateQuery


//------------------------------------------------------------------------------
// adds DELETE query object
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddDeleteQuery(Lexer: TACRLexer);
begin
  FQuery := TACRSQLDelete.Create(Lexer, ACRQuery);
end;//AddDeleteQuery


//------------------------------------------------------------------------------
// adds CREATE TABLE query object
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddCreateTableQuery(Lexer: TACRLexer);
begin
  FQuery := TACRDDLCreateTable.Create(Lexer, ACRQuery);
end;//AddCreateTableQuery


//------------------------------------------------------------------------------
// adds DROP TABLE query object
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddDropTableQuery(Lexer: TACRLexer);
begin
  FQuery := TACRDDLDropTable.Create(Lexer, ACRQuery);
end;//AddDropTableQuery


//------------------------------------------------------------------------------
// add ALTER TABLE query object
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddAlterTableQuery;
begin
  FQuery := TACRDDLAlterTable.Create(Lexer, ACRQuery);
end;//AddAlterTableQuery


//------------------------------------------------------------------------------
// add RENAME TABLE
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddRenameTableQuery(Lexer: TACRLexer);
begin
  FQuery := TACRDDLRenameTable.Create(Lexer, ACRQuery);
end;//AddRenameTableQuery


//------------------------------------------------------------------------------
// add CREATE INDEX query object
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddCreateIndex;
begin
  FQuery := TACRDDLCreateIndex.Create(Lexer, ACRQuery);
end;//AddCreateIndex


//------------------------------------------------------------------------------
// add DROP INDEX query object
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddDropIndexQuery;
begin
  FQuery := TACRDDLDropIndex.Create(Lexer, ACRQuery);
end;//AddDropIndexQuery


//------------------------------------------------------------------------------
// add START TRANSACTION
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddStartTransactionQuery(Lexer: TACRLexer);
begin
  FQuery := TACRStartTransaction.Create(Lexer, ACRQuery);
end; // AddStartTransactionQuery


//------------------------------------------------------------------------------
// add COMMIT
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddCommitQuery(Lexer: TACRLexer);
begin
  FQuery := TACRCommit.Create(Lexer, ACRQuery);
end; // AddCommitQuery


//------------------------------------------------------------------------------
// add ROLLBACK
//------------------------------------------------------------------------------
procedure TACRLocalSQLProcessor.AddRollbackQuery(Lexer: TACRLexer);
begin
  FQuery := TACRRollback.Create(Lexer, ACRQuery);
end; // AddRollbackQuery


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRLocalSQLProcessor.Create(Query: TDataSet);
begin
  inherited Create(Query);
  FQuery := nil;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRLocalSQLProcessor.Destroy;
begin
  if (FQuery <> nil) then
   FQuery.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// OpenQuery
//------------------------------------------------------------------------------
function TACRLocalSQLProcessor.OpenQuery: TACRCursor;
var
  Token:  TToken;
  Lexer:  TACRLexer;
begin
  Result := inherited OpenQuery;
  FQuery := nil;
  Lexer := TACRLexer.Create(SqlText, SQLParams);
  try
     // Parse and Execute queries from SQL script
       while Lexer.GetNextCommand do
        begin
          if (FQuery <> nil) then
           begin
            FQuery.Free;
            FQuery := nil;
           end;
         // Parse current command
         // look at first token
         if (not Lexer.GetCurrentToken(Token)) then
          raise EACRException.Create(30146, ErrorGBlankSqlCommand);

         if (Token.TokenType = tktReservedWord) then
           case Token.ReservedWord of
           rwSELECT:  // SELECT
                      AddSelectQuery(Lexer);
           rwINSERT:  // INSERT
                      AddInsertQuery(Lexer);
           rwUPDATE:  // UPDATE
                      AddUpdateQuery(Lexer);
           rwDELETE:  // DELETE
                      AddDeleteQuery(Lexer);
           rwCREATE:  // CREATE
                      begin
                        if Lexer.GetNextToken(Token) then
                          case Token.ReservedWord of
                            rwTABLE:  // CREATE TABLE
                                      AddCreateTableQuery(Lexer);
                            rwINDEX,
                            rwUNIQUE: // CREATE INDEX or UNIQUE INDEX
                                      AddCreateIndex(Lexer);
                            else raise EACRException.Create(30147, ErorrGObjectTypeKeywordExpected,
                                   [Token.Text, Token.LineNum, Token.ColumnNum]);
                          end
                        else raise EACRException.Create(30148, ErrorGUnexpectedEndOfCommand,
                                  [Token.LineNum, Token.ColumnNum]);
                      end;
            rwDROP:   // DROP
                      begin
                        if Lexer.GetNextToken(Token) then
                          case Token.ReservedWord of
                            rwTABLE:  // DROP TABLE
                                      AddDropTableQuery(Lexer);
                            rwINDEX:  // DROP INDEX
                                      AddDropIndexQuery(Lexer);
                            else raise EACRException.Create(30149, ErorrGObjectTypeKeywordExpected,
                                   [Token.Text, Token.LineNum, Token.ColumnNum]);
                          end
                        else raise EACRException.Create(30150, ErrorGUnexpectedEndOfCommand,
                                  [Token.LineNum, Token.ColumnNum]);
                      end;
            rwALTER:   // ALTER
                      begin
                        if Lexer.GetNextToken(Token) then
                          case Token.ReservedWord of
                            rwTable:  // ALTER TABLE
                                      AddAlterTableQuery(Lexer);
                            else raise EACRException.Create(30151, ErorrGObjectTypeKeywordExpected,
                                   [Token.Text, Token.LineNum, Token.ColumnNum]);

                          end
                        else raise EACRException.Create(30152, ErrorGUnexpectedEndOfCommand,
                                  [Token.LineNum, Token.ColumnNum]);
                      end;
            rwRENAME:   // RENAME
                      begin
                        if (not Lexer.GetNextToken(Token)) then
                          raise EACRException.Create(30353, ErrorGUnexpectedEndOfCommand,
                                                     [Token.LineNum, Token.ColumnNum]);
                        if IsReservedWord(Token, rwTABLE) then
                          Lexer.GetNextToken(Token);
                        // RENAME TABLE
                        AddRenameTableQuery(Lexer)
                      end;
           rwSTART:   // START TRANSACTION
                      AddStartTransactionQuery(Lexer);
           rwCOMMIT: // COMMIT
                      AddCommitQuery(Lexer);
           rwROLLBACK:  // START TRANSACTION
                      AddRollbackQuery(Lexer);
           else       // unsupported SQL or unexpected token
              raise EACRException.Create(30153, ErrorGSQLCommandExpected,
                                       [Token.Text, Token.LineNum, Token.ColumnNum]);
           end
         else
          // '( SELECT ...) ' ?
          if (Token.TokenType = tktLeftParenthesis) then
           AddSelectQuery(Lexer)
          else
           raise EACRException.Create(30154, ErrorGSQLCommandExpected,
                               [Token.Text, Token.LineNum, Token.ColumnNum]);
         // Execute query
         FQuery.ExecSQL(ACRQuery, True, FRequestLive, FReadonly);
         FRowsAffected := RowsAffected + FQuery.RowsAffected;
         Result := FQuery.GetResultCursor;
         if (Result <> nil) then
           Result.InternalFirst;
        end; // Parse and Execute queries
  finally
    Lexer.Free;
  end;
end;//OpenQuery


////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLSelect
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSQLSelect.Create(Lexer: TACRLexer; Query: TDataSet);
begin
  SelectListCount := 0;
  SetLength(SelectList, 0);
  FromTablesCount := 0;
  SetLength(FromTables, 0);
  OrderBySpecsCount := 0;
  SetLength(OrderBySpecs, 0);
  SearchCondition := nil;
  HavingCondition := nil;
  GroupByFields := TACRFields.Create;
  OrderByIndex := '';
  FTableList := TList.Create;
  inherited Create(Lexer, Query);
end;// Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRSQLSelect.Destroy;
var
  i: integer;
begin
  if (SearchCondition <> nil) then
   SearchCondition.Free;
  if (GroupByFields <> nil) then
   GroupByFields.Free;
  if (HavingCondition <> nil) then
   HavingCondition.Free;

  for i := 0 to FromTablesCount-1 do
   FromTables[i].Free;

  FTableList.Free;

  inherited;
end;// Destroy



//------------------------------------------------------------------------------
// builds AO tree
//------------------------------------------------------------------------------
function TACRSQLSelect.BuildAOTree(query: TDataset): TACRAO;
var
  i:                Integer;
  bGroupBy:         Boolean;
begin
  ResultDataset := query;

     // live result with 1 table?
     if ((FromTablesCount = 1) and (FromTables[0].TableType = attTable)) then
      Result := BuildOneTableTree
     else
      Result := BuildMultiTablesTree;
     try
      // create group by AO?
      bGroupBy := (GroupByFields.ItemCount > 0);
      // check for aggregates in select list
      if (not bGroupBy) then
       if (not AllFields) then
        begin
         for i := 0 to SelectListCount-1 do
          if (SelectList[i].ValueExpr <> nil) then
           if (TACRExpression(SelectList[i].ValueExpr).IsAggregated) then
            begin
             bGroupBy := True;
             break;
            end;
        end;

      // apply parts of WHERE clause if possible
      if (SearchCondition <> nil) then
         SearchCondition.ApplyFilterParts(Result,False);

      // create GroupByAO, apply having condition
      if (bGroupBy) then
       begin
        try
         Result := TACRAOGroupBy.Create(Result, GroupByFields);
        except
         // to avoid duplicated destroying
         Result := nil;
         for i := 0 to SelectListCount-1 do
          if (SelectList[i].ValueExpr <> nil) then
           begin
            SelectList[i].ValueExpr.Free;
            SelectList[i].ValueExpr := nil;
           end;

         raise;
        end;
       end;

      // TOP n?
      if (TopRowCount > -1) then
       begin
        if (Result.FIsAOTable) then
         Result := TACRAOSQLTopRowCount.Create(Result);
        Result.SetTopRowCount(FirstRowNo, TopRowCount);
       end;

      // sets projection
      Result.SetResultFields(SelectList,Distinct);

      // create GroupByAO, apply having condition
      if (bGroupBy) then
        if (HavingCondition <> nil) then
          HavingCondition.ApplyFilterParts(Result,True);

      // sort - order by
      ApplyOrderBy(Result);

      // apply pseudonames parts of WHERE clause if possible
      if (SearchCondition <> nil) then
       begin
         SearchCondition.ApplyFilterParts(Result,False);
         if (not SearchCondition.IsEmpty) then
           raise EACRException.Create(30144, ErorrGNotApplicableCondition);
       end;

     except
      if (Assigned(Result)) then
       Result.Free;
      raise;
     end;
end;//BuildAOTree


//------------------------------------------------------------------------------
// SELECT
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseSelectToken: Boolean;
begin
  Result := IsReservedWord(Token, rwSELECT);
  if (not Result) then
   // unsupported SQL or unexpected token
   raise EACRException.Create(30156, ErrorGSQLCommandExpected,
         [Token.Text, Token.LineNum, Token.ColumnNum]);
  GetNextToken;
end;//ParseSelectToken


//------------------------------------------------------------------------------
// DISTINCT | ALL ?
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseSetQuantifier: Boolean;
begin
 if (IsReservedWord(Token)) then
  begin
    if (IsReservedWord(Token, rwALL)) then
     begin
      Distinct := False;
      Result := True;
     end
    else
    if (IsReservedWord(Token, rwDISTINCT)) then
     begin
      Distinct := True;
      Result := True;
     end
    else // unrecognized reserved word - give up to handle
     begin
      Distinct := False; // default value
      Result := False; // handle this token by another handler
     end;
  end
 else // not reserved-word token give up handling
  begin
    Distinct := False; // default value
    Result := False; // handle this word by another handler
  end;

 // get next token
 if (Result) then
  GetNextToken;
end;// ParseSetQuantifier


//------------------------------------------------------------------------------
// TOP n ?
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseTopOperator: Boolean;
begin
  if (IsReservedWord(Token, rwTOP)) then
   begin
    // suppose 'TOP n'
    GetNextToken;
    // integer?
    if (Token.TokenType <> tktInt) then
     raise EACRException.Create(30158, ErrorGIntegerExpected,
                             [Token.Text, Token.LineNum, Token.ColumnNum]);
    TopRowCount := StrToInt(Token.Text);
    // get next token
    GetNextToken;
    // default value
    FirstRowNo := -1;
    // ','? => TOP n, first_row
    if (Token.Text = ',') then
     begin
       // get next token
       GetNextToken;
       // integer?
       if (Token.TokenType <> tktInt) then
         raise EACRException.Create(30159, ErrorGIntegerExpected,
                             [Token.Text, Token.LineNum, Token.ColumnNum]);
       FirstRowNo := StrToInt(Token.Text);
       // get next token
       GetNextToken;
     end;
    Result := True;
   end
  else // unrecognized reserved word - give up to handle
   begin
    TopRowCount := -1; // default value
    FirstRowNo := -1;
    Result := False; // handle this token by another handler
   end;
end;// ParseTopOperator


//------------------------------------------------------------------------------
// <derived column>  | <qualifier> <period> <asterisk>
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseSelectSubList: Boolean;
var
  FieldName, TableName, Pseudonym: string;
  AllFields: Boolean;
begin
 // handle field1 | table1.field1 | expr
  begin
   Result := True;

   // add new select list item
   inc(SelectListCount);
   SetLength(SelectList, SelectListCount);

   // parse field name | expr
   SelectList[SelectListCount-1].ValueExpr := TACRExpression.Create;

   TACRExpression(SelectList[SelectListCount-1].ValueExpr).Params :=
                                                     TACRQuery(LQuery).Params;
   TACRExpression(SelectList[SelectListCount-1].ValueExpr).InMemory :=
                                                     TACRQuery(LQuery).InMemory;
   TACRExpression(SelectList[SelectListCount-1].ValueExpr).SessionName :=
                                                     TACRQuery(LQuery).SessionName;
   TACRExpression(SelectList[SelectListCount-1].ValueExpr).DatabaseName :=
                                                     TACRQuery(LQuery).DatabaseName;

   TACRExpression(SelectList[SelectListCount-1].ValueExpr).ParseForValueExpression(LLex);

   if (TACRExpression(SelectList[SelectListCount-1].ValueExpr).IsEmpty) then
    begin
     SelectList[SelectListCount-1].ValueExpr.Free;
     raise EACRException.Create(30160, ErrorGExpressionExpected,
                             [Token.Text, Token.LineNum, Token.ColumnNum]);
    end;
   GetCurrentToken;

   // pseudonym is not specified yet
   Pseudonym := '';
   AllFields := False;
   // check pseudonym
   if (not AllFields) then
    // (Field1 AS F1) ?
          if (IsReservedWord(Token, rwAS)) then
           begin
             // get next token (F1?)
             GetNextToken([tktString, tktQuotedString, tktBracketedString],
                          30162, ErrorGFieldPseudonymExpected);
             // got pseudonym
             Pseudonym := Token.Text;
             // get next token
             GetNextToken;
            end
          else
          // (Field1 F1) ?
          if (Token.TokenType in
               [tktString, tktQuotedString, tktBracketedString]) then
            begin
             // got pseudonym
             Pseudonym := Token.Text;
             // look at next token
             GetNextToken;
            end;

   // expr or field?
   SelectList[SelectListCount-1].IsExpression :=
          not TACRExpression(SelectList[SelectListCount-1].ValueExpr).IsField;

   // field?
   if (not SelectList[SelectListCount-1].IsExpression) then
    begin
     // get field name, table name
     TACRExpression(SelectList[SelectListCount-1].ValueExpr).GetFieldInfo(TableName, FieldName);
     // store TableName
     SelectList[SelectListCount-1].TableName := TableName;
     // store ValueExpr (FieldName)
     SelectList[SelectListCount-1].FieldName := FieldName;
     if (FieldName = '*') then
      AllFields := True;
    end;

   // free non-expr objects (expr objects are freed by AO)
   if (not SelectList[SelectListCount-1].IsExpression) then
    begin
     TACRExpression(SelectList[SelectListCount-1].ValueExpr).Free;
     TACRExpression(SelectList[SelectListCount-1].ValueExpr) := nil;
    end;

   // store AllFields ('table1.*'?)
   SelectList[SelectListCount-1].AllFields := AllFields;
   // pseudonym
   SelectList[SelectListCount-1].Pseudonym := Pseudonym;
  end;
end;// ParseSelectSubList


//------------------------------------------------------------------------------
// * | <select sublist>
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseSelectList: Boolean;
begin
  if (Token.Text = Asterisk) then
   begin
    Result := True;
    GetNextToken;
    AllFields := True; // select all fields
   end
  else
   begin
    AllFields := False;
    Result := True;
     // get FieldName
    if (ParseSelectSubList) then
     // all field expressions
     repeat
        if (Token.TokenType <> tktComma) then
           break
        else
          // skip ',' token
          GetNextToken;
        // get next field expr
        if (not ParseSelectSubList) then
         raise EACRException.Create(30163, ErrorGFieldNameExpected,
               [Token.Text, Token.LineNum, Token.ColumnNum]);
     until False
    else
     raise EACRException.Create(30164, ErrorGFieldListExpected,
                        [Token.Text, Token.LineNum, Token.ColumnNum]);
   end;
end;//ParseSelectList


//------------------------------------------------------------------------------
// INTO <target>
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseInto: Boolean;
begin
 Result := False;
 IntoTable := '';
 IntoDatabase := '';
 IntoMemory := False;

 if (IsReservedWord(Token, rwINTO)) then
  begin
    // get table | database name
    GetNextToken([tktString, tktReservedWord, tktBracketedString, tktQuotedString],
                 30165, ErrorGTableNameExpected);
    if (IsReservedWord(Token, rwMEMORY)) then
     begin
      IntoMemory := True;
      // next token
      GetNextToken;
     end;

    IntoTable := Token.Text;
    // next token
    GetNextToken;
    // db.table?
    if (Token.TokenType = tktDot) then
     begin
      IntoDatabase := IntoTable;
      // get table name
      GetNextToken([tktString, tktBracketedString, tktQuotedString],
                   30166, ErrorGTableNameExpected);
      IntoTable := Token.Text;
      // next token
      GetNextToken;
     end;
    Result := True;
  end;
end;// ParseInto


//------------------------------------------------------------------------------
// ON <join condition>
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseJoinCondition(var SearchCondition: TACRExpression): Boolean;
begin
 Result := False;
 if (IsReservedWord(Token, rwON)) then
   begin
    Result := True;
    GetNextToken; // skip ON token
    //SearchCondition := TACRExpression.Create(LLex);
    SearchCondition := TACRExpression.Create;
    SearchCondition.Params := TACRQuery(LQuery).Params;
    SearchCondition.InMemory := TACRQuery(LQuery).InMemory;
    SearchCondition.DatabaseName := TACRQuery(LQuery).DatabaseName;
    SearchCondition.SessionName := TACRQuery(LQuery).SessionName;
    SearchCondition.ParseForBooleanExpression(LLex);
    if (not AllFields) then
     SearchCondition.ReplacePseudonyms(SelectList);
    // get current unhandled token
    GetCurrentToken;
   end;
end;// ParseJoinCondition


//------------------------------------------------------------------------------
// USING <join columns>
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseNamedColumnsJoin(var Fields: TStringList): Boolean;
begin
 Result := False;
 if (IsReservedWord(Token, rwUSING)) then
   begin
    Result := True;
    GetNextToken; // skip USING token

    // '(' ?
    if (Token.TokenType = tktLeftParenthesis) then
     begin
      // skip '('
      GetNextToken;

      // parse fields list
      ParseColumnList(Fields);

      // skip ')'
      if (Token.TokenType = tktRightParenthesis) then
         GetNextToken; // end of query is possible here
     end
    else
     raise EACRException.Create(30167, ErrorGOtherTokenExpected,
                   ['(', Token.Text, Token.LineNum, Token.ColumnNum]);
   end;
end;// ParseNamedColumnsJoin


//------------------------------------------------------------------------------
// CROSS JOIN | INNER JOIN | ...
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseJoin(var tblRef: TACRTableReference): Boolean;
var
 RightTblRef: TACRTableReference;
 JoinType: TACRJoinType;
 IsNatural: Boolean;
 Fields: TStringList;
 SearchCondition: TACRExpression;
begin
 Result := False;

 if (IsReservedWord(Token)) then
  repeat
     IsNatural := False; // default value
     JoinType := ajtInner; // default value
     // parse CROSS JOIN | [ NATURAL ] [ <join type> ] JOIN
     if (IsReservedWord(Token, rwCROSS)) then
      begin
          GetNextToken([tktReservedWord], 30168, ErrorGUnexpectedToken); // skip CROSS
          JoinType := ajtCross;
          Result := True;
      end
     else
      begin
       // [NATURAL]
       if (IsReservedWord(Token, rwNATURAL)) then
         begin
          IsNatural := True;
          GetNextToken([tktReservedWord], 30169, ErrorGUnexpectedToken);
          Result := True;
         end;
       // INNER?
       if (IsReservedWord(Token, rwINNER)) then
         begin
          JoinType := ajtInner;
          GetNextToken([tktReservedWord], 30170, ErrorGUnexpectedToken);
          Result := True;
         end
       else
        // OUTER join
        begin
         // LEFT
         if (IsReservedWord(Token, rwLEFT)) then
          begin
           JoinType := ajtLeftOuter;
           GetNextToken([tktReservedWord], 30171, ErrorGUnexpectedToken);
           Result := True;
          end
         else
         // RIGHT
         if (IsReservedWord(Token, rwRIGHT)) then
          begin
           JoinType := ajtRightOuter;
           GetNextToken([tktReservedWord], 30172, ErrorGUnexpectedToken);
           Result := True;
          end
         else
         // FULL
         if (IsReservedWord(Token, rwFULL)) then
          begin
           JoinType := ajtFullOuter;
           GetNextToken([tktReservedWord], 30173, ErrorGUnexpectedToken);
           Result := True;
          end;
         // [OUTER]
         if ((Result) and (IsReservedWord(Token, rwOUTER))) then
          begin
           GetNextToken([tktReservedWord], 30174, ErrorGUnexpectedToken);
          end;
        end;// outer join
      end; // non-cross join

     // next token - JOIN?
     if (not IsReservedWord(Token, rwJOIN)) then
      if (Result) then
       raise EACRException.Create(30175, ErrorGOtherTokenExpected,
                 ['JOIN', Token.Text, Token.LineNum, Token.ColumnNum])
      else
       exit; // no joins
     GetNextToken;

     // get right table ref
     RightTblRef := TACRTableReference.Create;
     if not ParseTableReference(RightTblRef) then
      raise EACRException.Create(30176, ErrorGTableNameExpected,
                       [Token.Text, Token.LineNum, Token.ColumnNum]);

     // <join condition> | <named columns join>
     Fields := nil;
     SearchCondition := nil;
     if not ParseJoinCondition(SearchCondition) then
       ParseNamedColumnsJoin(Fields);

     // update tree to make join
     tblRef.MakeJoin(RightTblRef, JoinType, IsNatural, Fields, SearchCondition)
  until (not(IsReservedWord(Token, rwCROSS) or
             IsReservedWord(Token, rwNATURAL) or
             IsReservedWord(Token, rwINNER) or
             IsReservedWord(Token, rwLEFT) or
             IsReservedWord(Token, rwRIGHT) or
             IsReservedWord(Token, rwFULL) or
             IsReservedWord(Token, rwOUTER) or
             IsReservedWord(Token, rwJOIN)));
end;// ParseJoin


//------------------------------------------------------------------------------
// <table name> [ [ AS ] <correlation name> ...
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseTableReference(tblRef: TACRTableReference): Boolean;
var
  DatabaseName, TableName, Pseudonym, Password: string;
  TableType: TACRTableType;
  InMemory: Boolean;
  tRef: TACRTableReference;
begin
 Result := False;
 // handle table1 | "database 1".table1
 if (Token.TokenType in [tktString, tktReservedWord, tktQuotedString,tktBracketedString]) then
  begin
   Result := True;
   if (IsReservedWord(Token, rwMEMORY)) then
    begin
      InMemory := True;
      GetNextToken;
    end
   else
    InMemory := TACRDataSet(LQuery).InMemory;
   DatabaseName := TACRDataSet(LQuery).DatabaseName;
   // get table name
   TableName := Token.Text;
   // pseudonym is not specified yet
   Pseudonym := '';
   // password too
   Password := '';
   // tabletype is ettTable by default
   TableType := attTable;
   // look at next token
   if (GetNextToken) then
    begin
     // db.table?
     if (Token.TokenType = tktDot) then
      begin
       DatabaseName := TableName;
       // get table name
       GetNextToken;
       TableName := Token.Text;
       // next token
       GetNextToken;
      end;
      // check pseudonym - correlation name (Table1 AS t1)
      if (IsReservedWord(Token, rwAS)) then
            begin
             // get next token (t1?)
             GetNextToken([tktString, tktQuotedString, tktBracketedString],
                           30177, ErrorGFieldPseudonymExpected);
             // got pseudonym
             Pseudonym := Token.Text;
             // get next token
             GetNextToken;
            end
      else
        if (Token.TokenType in [tktString, tktBracketedString]) then
            begin
             // got pseudonym
             Pseudonym := Token.Text;
             // get next token
             GetNextToken;
            end;
      // check password
      if (IsReservedWord(Token, rwPASSWORD)) then
       begin
        // get next token 'password'
        GetNextToken([tktQuotedString], 30178, ErrorGQuotedPasswordDataExpected);
        Password := Token.Text;
        // get next token
        GetNextToken;
       end;
    end;
   // first table in joins chain?
   if (tblRef = nil) then
    begin
     // add new tables list item
     inc(FromTablesCount);
     SetLength(FromTables, FromTablesCount);
     FromTables[FromTablesCount-1] := TACRTableReference.Create;
     tRef := FromTables[FromTablesCount-1];
    end
   else
    tRef := tblRef;

   // store TableType
   tRef.TableType := TableType;
   // store DatabaseName
   tRef.DatabaseName := DatabaseName;
   // store SessionName
   tRef.SessionName := TACRDataSet(LQuery).SessionName;
   // store TableName
   tRef.TableName := TableName;
   // store Pseudonym
   tRef.Pseudonym := Pseudonym;
   // store Password
   tRef.Password := Password;
   // store [MEMORY]
   tRef.InMemory := InMemory;
  end
 else
  // (<table reference>)
  if (Token.TokenType = tktLeftParenthesis) then
   begin
    // skip '('
    GetNextToken;
    // get table reference
    Result := ParseTableReference(tblRef);
    // get its pointer
    if (tblRef = nil) then
     tRef := FromTables[FromTablesCount-1]
    else
     tRef := tblRef;

    if not Result then
     raise EACRException.Create(30179, ErrorGFieldNameExpected,
                      [Token.Text, Token.LineNum, Token.ColumnNum]);
    // skip ')'
    if (Token.TokenType = tktRightParenthesis) then
       GetNextToken;
   end;

 // join?
 if (Result) then
   ParseJoin(tRef);
end;// ParseTableReference


//------------------------------------------------------------------------------
// FROM ...
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseFromClause: Boolean;
begin
 if (IsReservedWord(Token, rwFROM)) then
  begin
    // skip FROM token
    GetNextToken([tktString, tktReservedWord, tktQuotedString, tktBracketedString, tktLeftParenthesis],
                 30180, ErrorGTableNameExpected);
    Result := True;
    // get TableName
    if (ParseTableReference) then
     // get all tables
     repeat
        // end of sql command?
        if (Token.TokenType <> tktComma) then
          break
        else
          // skip ',' token
          GetNextToken([tktString, tktReservedWord, tktQuotedString, tktBracketedString, tktLeftParenthesis],
                       30181, ErrorGTableNameExpected);
       // get next table reference
       if (not ParseTableReference) then
        raise EACRException.Create(30182, ErrorGTableNameExpected,
                  [Token.Text, Token.LineNum, Token.ColumnNum]);
     until False
    else
      raise EACRException.Create(30183, ErrorGTableNameExpected,
                  [Token.Text, Token.LineNum, Token.ColumnNum]);
  end
 else
   raise EACRException.Create(30184, ErrorGOtherTokenExpected,
                ['FROM', Token.Text, Token.LineNum, Token.ColumnNum]);
end;// ParseFromClause


//------------------------------------------------------------------------------
// WHERE ...
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseWhereClause: Boolean;
begin
 Result := False;
 if (IsReservedWord(Token, rwINTO)) then
   raise EACRException.Create(11337, ErrorLTokenINTOShouldBeBeforeFROM,
                [Token.LineNum, Token.ColumnNum]);
 if (IsReservedWord(Token, rwWHERE)) then
  begin
    Result := True;
    // skip WHERE token
    GetNextToken;
    SearchCondition := TACRExpression.Create;
    SearchCondition.Params := TACRQuery(LQuery).Params;
    SearchCondition.InMemory := TACRQuery(LQuery).InMemory;
    SearchCondition.DatabaseName := TACRQuery(LQuery).DatabaseName;
    SearchCondition.SessionName := TACRQuery(LQuery).SessionName;
    SearchCondition.ParseForBooleanExpression(LLex);

    if (not AllFields) then
     SearchCondition.ReplacePseudonyms(SelectList);
    // get current (not handled) token
    GetCurrentToken;
  end;
end;// ParseWhereClause


//------------------------------------------------------------------------------
// GROUP BY ...
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseGroupByClause: Boolean;
begin
 if (IsReservedWord(Token, rwGROUP)) then
  begin
    Result := True;
    // skip GROUP token
    GetNextToken;
    // 'BY' ?
    if (not IsReservedWord(Token, rwBY)) then
     raise EACRException.Create(30185, ErrorGOtherTokenExpected,
                   ['BY', Token.Text, Token.LineNum, Token.ColumnNum]);
    // skip BY token
    GetNextToken;

    // parse fields list
    ParseFieldList(GroupByFields);
  end
 else
  Result := False;
end;// ParseGroupByClause


//------------------------------------------------------------------------------
// HAVING ...
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseHavingClause: Boolean;
begin
 Result := False;
 if (IsReservedWord(Token, rwHAVING)) then
  begin
    Result := True;
    // skip HAVING token
    GetNextToken;
    HavingCondition := TACRExpression.Create;
    HavingCondition.Params := TACRQuery(LQuery).Params;
    HavingCondition.InMemory := TACRQuery(LQuery).InMemory;
    HavingCondition.DatabaseName := TACRQuery(LQuery).DatabaseName;
    HavingCondition.SessionName := TACRQuery(LQuery).SessionName;
    HavingCondition.ParseForBooleanExpression(LLex);
    if (not AllFields) then
     HavingCondition.ReplacePseudonyms(SelectList);
    // get current (not handled) token
    GetCurrentToken;
  end;
end;// ParseHavingClause


//------------------------------------------------------------------------------
// <sort key> [ <collate clause> ] [ <ordering specification> ]
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseSortSpecification: Boolean;
begin
  Result := True;
  if (Token.TokenType in
       [tktInt, tktString, tktBracketedString, tktReservedWord, tktQuotedString]) then
   begin
     // new order spec
     Inc(OrderBySpecsCount);
     SetLength(OrderBySpecs, OrderBySpecsCount);
     if (Token.TokenType = tktInt) then
      OrderBySpecs[OrderBySpecsCount-1].ColumnNumber := StrToInt(Token.Text)
     else
      begin
       OrderBySpecs[OrderBySpecsCount-1].ColumnNumber := -1;
       OrderBySpecs[OrderBySpecsCount-1].ColumnName := Token.Text;
      end;
     OrderBySpecs[OrderBySpecsCount-1].CaseInsensitive := False;
     // get next token
     GetNextToken;
     // "." ?
     if (Token.TokenType = tktDot) then
      begin
       // previous token was table name
       OrderBySpecs[OrderBySpecsCount-1].TableName :=
                      OrderBySpecs[OrderBySpecsCount-1].ColumnName;
       // get next token - column name
       GetNextToken([tktString, tktQuotedString, tktBracketedString, tktReservedWord],
                    30186, ErrorGUnexpectedToken);
       OrderBySpecs[OrderBySpecsCount-1].ColumnName := Token.Text;
       // get next token
       GetNextToken;
      end;
     // ASC | DESC ?
     if (IsReservedWord(Token, rwASC)) then
      begin
       OrderBySpecs[OrderBySpecsCount-1].Descending := False;
       // skip ASC token
       GetNextToken;
       if (IsReservedWord(Token, rwNOCASE)) then
        begin
         OrderBySpecs[OrderBySpecsCount-1].CaseInsensitive := True;
         // skip NOCASE token
         GetNextToken;
        end;
      end
     else
     if (IsReservedWord(Token, rwDESC)) then
      begin
       OrderBySpecs[OrderBySpecsCount-1].Descending := True;
       // skip DESC token
       GetNextToken;
       if (IsReservedWord(Token, rwNOCASE)) then
        begin
         OrderBySpecs[OrderBySpecsCount-1].CaseInsensitive := True;
         // skip NOCASE token
         GetNextToken;
        end;
      end
     else
     if (IsReservedWord(Token, rwNOCASE)) then
      begin
       OrderBySpecs[OrderBySpecsCount-1].CaseInsensitive := True;
       // skip NOCASE token
       GetNextToken;
      end;
   end
  else
   raise EACRException.Create(30187, ErrorGFieldNameExpected,
                        [Token.Text, Token.LineNum, Token.ColumnNum]);
end;// ParseSortSpecification


//------------------------------------------------------------------------------
// ORDER BY
//------------------------------------------------------------------------------
function TACRSQLSelect.ParseOrderByClause: Boolean;
begin
 if (IsReservedWord(Token, rwORDER)) then
  begin
    Result := True;
    // skip ORDER token
    GetNextToken;
    // 'BY' ?
    if (not IsReservedWord(Token, rwBY)) then
     raise EACRException.Create(30188, ErrorGOtherTokenExpected,
                  ['BY', Token.Text, Token.LineNum, Token.ColumnNum]);
    // skip BY token
    GetNextToken;

    if (IsReservedWord(Token, rwINDEX)) then
     begin
       // get IndexName
       GetNextToken([tktString, tktQuotedString, tktBracketedString],
                     11244, ErrorGUnexpectedToken);
       OrderByIndex := Token.Text;
       // skip IndexName token
       GetNextToken;
     end
    else
     begin
      // parse sort specification list
      if (ParseSortSpecification) then
       // get all specs
       repeat
          // end of sql command?
          if (Token.TokenType <> tktComma) then
            break
          else
            // skip ',' token
            GetNextToken([tktInt, tktString, tktQuotedString, tktBracketedString, tktReservedWord],
                         30189, ErrorGUnexpectedToken);
         // get next spec
         if (not ParseSortSpecification) then
          raise EACRException.Create(30190, ErrorGSortSpecificationExpected,
                          [Token.Text, Token.LineNum, Token.ColumnNum]);

       until False
      else
        raise EACRException.Create(30191, ErrorGSortSpecificationExpected,
                          [Token.Text, Token.LineNum, Token.ColumnNum]);
     end;
  end
 else
  Result := False;
end;// ParseOrderByClause


//------------------------------------------------------------------------------
// parse query
//------------------------------------------------------------------------------
procedure TACRSQLSelect.Parse;
var
  State: integer;
  bOk: Boolean;
begin
   // get first token (for new query) or current token (for subquery)
   //bOk := LLex.GetCurrentToken(Token);
   bOk := GetCurrentToken;

   if (not bOk) then
      raise EACRException.Create(30155, ErrorGBlankSqlCommand);

   state := 0;
   repeat
      // states switch (depends on a current part of query)
      case state of
       0: // SELECT
          ParseSelectToken;
       1: // DISTINCT | ALL ?
          ParseSetQuantifier;
       2: // TOP n ?
          ParseTopOperator;
       3: // * | <select sublist>
          ParseSelectList;
       4: // <into ...>
          ParseInto;
       5: // <from clause>
          ParseFromClause;
       6: // <where clause>
          ParseWhereClause;
       7: // <group by clause>
          ParseGroupByClause;
       8: // <where clause>
          ParseHavingClause;
       9: // <order by clause>
          ParseOrderByClause;
       else
        break;
      end;// case
      inc(state);
   until (false);

end;// Parse


//------------------------------------------------------------------------------
// gets default database name
//------------------------------------------------------------------------------
function TACRSQLSelect.GetDefaultDatabaseName: string;
begin
  //raise EACRException.Create(30192, ErrorGNotImplementedYet);
  Result := TACRDataset(ResultDataset).DatabaseName;
end;// GetDefaultDatabaseName


//------------------------------------------------------------------------------
// creates and adjusts table AO
//------------------------------------------------------------------------------
function TACRSQLSelect.CreateTableAO(var TableRef: TACRTableReference): TACRAO;
var
  DatabaseName: string;
begin
  DatabaseName := TableRef.DatabaseName;
  if (DatabaseName = '') then
    DatabaseName := GetDefaultDatabaseName;

  Result := TACRAOTable.Create(DatabaseName,
                               TableRef.SessionName,
                               TableRef.TableName,
                               TableRef.Pseudonym,
                               TableRef.InMemory,
                               FRequestLive,
                               TableRef.Password);
end;// CreateTableAO


//------------------------------------------------------------------------------
// creates and adjusts joined table AO
//------------------------------------------------------------------------------
function TACRSQLSelect.CreateJoinedTableAO(var TableRef: TACRTableReference): TACRAO;
var
 LeftAO, RightAO: TACRAO;
 FieldList1, FieldList2: TACRFields;
 i: integer;
 Item1, Item2: TACRSelectListItem;

function RecursiveExtractJoinConditions(TableRef: TACRTableReference): integer;
begin
  if (TableRef.SearchCondition <> nil) then
     Result := TableRef.SearchCondition.ExtractJoinConditions(LeftAO, RightAO,
                                      FieldList1, FieldList2)
  else
   begin
     Result := 0;
     if (TableRef.TableType = attJoinedTable) then
      begin
       if (TableRef.LeftTable.SearchCondition <> nil) then
        Result := Result + TableRef.LeftTable.SearchCondition.ExtractJoinConditions(
                         LeftAO, RightAO, FieldList1, FieldList2);

       if (TableRef.RightTable.SearchCondition <> nil) then
        Result := Result + TableRef.RightTable.SearchCondition.ExtractJoinConditions(
                         LeftAO, RightAO, FieldList1, FieldList2);

      end;
   end;
end; // RecursiveExtractJoinConditions

begin
 // create joined tables
 LeftAO := CreateAO(TableRef.LeftTable);
 RightAO := CreateAO(TableRef.RightTable);

 // prepare column lists
 FieldList1 := nil;
 FieldList2 := nil;
 try
 // column lists required?
 if ((not TableRef.NaturalJoin) and
     (
      (TableRef.JoinType = ajtInner) or
      (TableRef.JoinType = ajtLeftOuter) or
      (TableRef.JoinType = ajtRightOuter) or
      (TableRef.JoinType = ajtFullOuter)
     )) then
  begin // build column lists
   // USING ?
   if (TableRef.UsingFields <> nil) then
    begin
      FieldList1 := TACRFields.Create;
      FieldList2 := TACRFields.Create;
      for i := 0 to TableRef.UsingFields.Count-1 do
       begin
        Item1.FieldName := TableRef.UsingFields.Strings[i];
        Item1.TableName := TableRef.LeftTable.TableName;
        Item2.FieldName := TableRef.UsingFields.Strings[i];
        Item2.TableName := TableRef.RightTable.TableName;
        FieldList1.Append(Item1);
        FieldList2.Append(Item2);
       end;
    end
   else
    begin
      FieldList1 := TACRFields.Create;
      FieldList2 := TACRFields.Create;
      // extract from ON clause
      RecursiveExtractJoinConditions(TableRef);
      // extract from WHERE clause
      if (SearchCondition <> nil) then
       SearchCondition.ExtractJoinConditions(LeftAO, RightAO,
                                       FieldList1, FieldList2);
    end
  end;

  // create join AO
  Result := TACRAOJoin.Create(LeftAO, RightAO, TableRef.JoinType,
                              TableRef.NaturalJoin,
                              FieldList1, FieldList2);
 // set filter
 if (Result <> nil) then
  if (TableRef.SearchCondition <> nil) then
    TableRef.SearchCondition.ApplyFilterParts(Result,False);

 finally
  if (FieldList1 <> nil) then
   FieldList1.Free;
  if (FieldList2 <> nil) then
   FieldList2.Free;
 end;
end;// CreateJoinedTableAO


//------------------------------------------------------------------------------
// creates and adjusts AO
//------------------------------------------------------------------------------
function TACRSQLSelect.CreateAO(var TableRef: TACRTableReference): TACRAO;
begin
 case TableRef.TableType of
  attTable:
    Result := CreateTableAO(TableRef);
  attJoinedTable:
    Result := CreateJoinedTableAO(TableRef);
  else
    Result := nil;
 end;
end;// CreateAO


//------------------------------------------------------------------------------
// builds one-table AO
//------------------------------------------------------------------------------
function TACRSQLSelect.BuildOneTableTree: TACRAO;
begin
  // create AO
  Result := CreateAO(FromTables[0]);
end;// BuildOneTableTree


//------------------------------------------------------------------------------
// builds multi-tables AO tree
//------------------------------------------------------------------------------
function TACRSQLSelect.BuildMultiTablesTree: TACRAO;
var
  i, JoinConditionCount: integer;
  RightAO: TACRAO;
  JoinType: TACRJoinType;
  FieldList1, FieldList2: TACRFields;
begin
 Result := CreateAO(FromTables[0]);
 // create other joins
 for i := 1 to FromTablesCount-1 do
  begin
   // init column lists
   FieldList1 := nil;
   FieldList2 := nil;
   try
    RightAO := CreateAO(FromTables[i]);
    // try to find join conditions in WHERE clause
    if (SearchCondition <> nil) then
     begin
       FieldList1 := TACRFields.Create;
       FieldList2 := TACRFields.Create;
       // try to extract conditions
       JoinConditionCount := SearchCondition.ExtractJoinConditions(Result, RightAO,
                                             FieldList1, FieldList2);
     end
    else
     JoinConditionCount := 0;

    if (JoinConditionCount = 0) then
     begin
      JoinType := ajtCross;
      if (FieldList1 <> nil) then
       FieldList1.Free;
      if (FieldList2 <> nil) then
       FieldList2.Free;
      FieldList1 := nil;
      FieldList2 := nil;
     end
    else
     JoinType := ajtInner;

    // create join AO
    Result := TACRAOJoin.Create(Result, RightAO, JoinType, False,
                                FieldList1, FieldList2);
   finally
     if (FieldList1 <> nil) then
      FieldList1.Free;
     if (FieldList2 <> nil) then
      FieldList2.Free;
   end;
  end;
end;// BuildMultiTablesTree


//------------------------------------------------------------------------------
// applies Order By clause
//------------------------------------------------------------------------------
procedure TACRSQLSelect.ApplyOrderBy(AO: TACRAO);
var
  i: integer;
  ColumnName, TableName: string;
  FieldNumber: TACRIntegerArray;
  FieldName: string;
  IndexFieldNames, DescFields, CaseInsFields: String;
begin
  if ((OrderBySpecsCount = 0) and (OrderByIndex = '')) then
   exit;
  FieldNumber := TACRIntegerArray.Create;
  try
    // for all sort specs
    for i := 0 to OrderBySpecsCount-1 do
     begin
      if (OrderBySpecs[i].ColumnNumber < 0) then
       ColumnName := OrderBySpecs[i].ColumnName
      else
       ColumnName := AO.GetFieldNameByOrderByColumnNumber(OrderBySpecs[i].ColumnNumber);
      TableName := OrderBySpecs[i].TableName;
      // get field No in AO
      FieldNumber.SetSize(0);
      AO.FieldExists(ColumnName, TableName,
                     not (AO is TACRAOGroupBy), FieldNumber);
      if (FieldNumber.ItemCount = 0) then
       raise EACRException.Create(30195, ErrorGColumnFromOrderByNotFound, [ColumnName]);
{
      if (FieldNumber.ItemCount > 1) then
       raise EACRException.Create(10412, ErrorLAmbiguousFieldReferenceInOrderBy,
          [TableName,ColumnName]);
}
      FieldName := AO.GetFieldName(FieldNumber.Items[0],True);
      // all index fields
      if (IndexFieldNames = '') then
       IndexFieldNames := FieldName
      else
       IndexFieldNames := IndexFieldNames + ';'+FieldName;
      // desc index fields
      if (OrderBySpecs[i].Descending) then
       if (DescFields = '') then
        DescFields := FieldName
       else
        DescFields := DescFields + ';'+FieldName;
      // desc index fields
      if (OrderBySpecs[i].CaseInsensitive) then
       if (CaseInsFields = '') then
        CaseInsFields := FieldName
       else
        CaseInsFields := CaseInsFields + ';'+FieldName;
     end;
     AO.SetIndex(IndexFieldNames, DescFields, CaseInsFields, OrderByIndex, True);
  finally
   FieldNumber.Free;
  end;
end;// ApplyOrderBy


//------------------------------------------------------------------------------
// lock all tables used by this query
//------------------------------------------------------------------------------
procedure TACRSQLSelect.LockAllTables;
begin
{
  GetTableList(FTableList);
  try
   if (not LockTableListS(FTableList)) then
    raise EACRException.Create(10682,ErrorLCannotLockTablesForSelect);
  except
   FreeTableList(FTableList);
   raise;
  end
}  
end; // LockAllTables


//------------------------------------------------------------------------------
// unlock all tables used by this query
//------------------------------------------------------------------------------
procedure TACRSQLSelect.UnlockAllTables;
begin
  try
   UnlockTableListS(FTableList);
  finally
   FreeTableList(FTableList);
  end;
end; // UnockAllTables


//------------------------------------------------------------------------------
// creates table list
//------------------------------------------------------------------------------
procedure TACRSQLSelect.GetTableList(TableList: TList);

 procedure AddTableToList(var TableRef: TACRTableReference; TableList: TList);
 var Table: TACRTable;
 begin
  if (TableRef <> nil) then
   case (TableRef.TableType) of
    attTable:
     begin
      if (not TableRef.InMemory) then
       begin
        Table := TACRTable.Create(nil);
        Table.DatabaseName := TableRef.DatabaseName;
        Table.InMemory := False;
        Table.TableName := TableRef.TableName;
        Table.SessionName := TableRef.SessionName;
//        Table.Password := TableRef.Password;
        TableList.Add(Table);
       end;
     end;
    attJoinedTable:
     begin
      AddTableToList(TableRef.LeftTable,TableList);
      AddTableToList(TableRef.RightTable,TableList);
     end;
   end;
 end;

var i,j: Integer;
begin
 j := -1;
{$IFDEF FILE_SERVER_VERSION}
  j := FromTablesCount - 1;
{$ENDIF}
{$IFDEF SERVER_VERSION}
  j := FromTablesCount - 1;
{$ENDIF}
  for i := 0 to j do
   AddTableToList(FromTables[i],TableList);
end; // GetTableList


//------------------------------------------------------------------------------
// free table list
//------------------------------------------------------------------------------
procedure TACRSQLSelect.FreeTableList(TableList: TList);
var
    Table: TACRTable;
begin
  while (TableList.Count > 0) do
   begin
    Table := TableList.Items[0];
    if (Table = nil) then
     raise EACRException.Create(10681,ErrorLNilPointer);
    TableList.Remove(Table);
    Table.Free;
   end;
end; // FreeTableList


//------------------------------------------------------------------------------
// Lock table list in S
//------------------------------------------------------------------------------
function TACRSQLSelect.LockTableListS(TableList: TList): Boolean;
{
var I:     Integer;
    Table: TACRTable;
}
begin
{
  Result := True;
  for I := 0 to TableList.Count - 1 do
   try
    Table := TableList.Items[i];
    if (Table = nil) then
     raise EACRException.Create(10684,ErrorLNilPointer);
    Table.Open;
    Table.LockTable;
   except
    Result := False;
   end;
}
end; // LockTableListS


//------------------------------------------------------------------------------
// unlock table list in S
//------------------------------------------------------------------------------
procedure TACRSQLSelect.UnlockTableListS(TableList: TList);
{
var I:      Integer;
    Table:  TACRTable;
}
begin
{
 for I := 0 to TableList.Count - 1 do
  begin
    Table := TableList.Items[i];
    if (Table = nil) then
      raise EACRException.Create(10683,ErrorLNilPointer);
    if (Table.Active) then
      Table.UnlockTable;
//     Table.Handle.TryToUnlockTableS;
  end;
}
end; // UnlockTableListS


////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLCommand
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSQLCommand.Create(Lexer: TACRLexer; Query: TDataSet);
begin
  {$IFDEF TRIAL_VERSION}
  {$IFNDEF TRIAL_VERSION_WITH_FULL_SQL}
   if ((not (Self is TACRSQLCursorCommand)) and
       (not (Self is TACRSQLSelect)) and (not (Self is TACRSQLUnion))) then
    begin
     acrtrshnm;
     raise EACRException.Create(10844,ErrorLCannotExecuteThisSQLStatementInTrialVersion);
    end;
  {$ENDIF}
  {$ENDIF}
  IntoMemory := False;
  LLex := Lexer;
  LQuery := Query;
  RowsAffected := 0;
  // parse with first token
  Parse;
end;//Create


//------------------------------------------------------------------------------
// parses query
//------------------------------------------------------------------------------
procedure TACRSQLCommand.Parse; 
begin
;
end; // Parse


//------------------------------------------------------------------------------
// parses list of columns (without table name): field1, field2, ...
//------------------------------------------------------------------------------
procedure TACRSQLCommand.ParseColumnList(var Fields: TStringList);

function ParseColumnName(var ColumnName: String): Boolean;
begin
 Result := False;
 if (Token.TokenType in [tktString, tktQuotedString, tktBracketedString]) then
  begin
   ColumnName := Token.Text;
   GetNextToken;
   Result := True;
  end;
end;

var
  s: string;

begin
      // get ColumnName
      if (ParseColumnName(s)) then
       // all columns
       repeat
        // create list if necessary
        if (Fields = nil) then
         Fields := TStringList.Create;
        // add column name
        Fields.Add(s);
        // ',' ?
        if (Token.TokenType <> tktComma) then
           break
        else
          // skip ',' token
          GetNextToken([tktString, tktQuotedString, tktBracketedString],10406,
            ErrorLUnexpectedToken);
        // get next field
        if (not ParseColumnName(s)) then
         raise EACRException.Create(10407,ErrorLFieldNameExpected,
          [Token.Text, Token.LineNum, Token.ColumnNum]);
       until False
      else
       raise EACRException.Create(10408, ErrorLFieldNameExpected,
        [Token.Text, Token.LineNum, Token.ColumnNum]);
end; // ParseColumnList


//------------------------------------------------------------------------------
// parses list of fields:  table.field1, field2, ..
//------------------------------------------------------------------------------
procedure TACRSQLCommand.ParseFieldList(var Fields: TACRFields);

function ParseFieldSpecification(var Fields: TACRFields): Boolean;
var
   item: TACRSelectListItem;
   TableName, FieldName: string;
   Expr: TACRExpression;
begin
 Result := True;
 // parse field as expression
 Expr := TACRExpression.Create;
 try
  Expr.InMemory := TACRQuery(LQuery).InMemory;
  Expr.DatabaseName := TACRQuery(LQuery).DatabaseName;
  Expr.SessionName := TACRQuery(LQuery).SessionName;
  Expr.ParseForValueExpression(LLex);
  // only field is allowed
  if (not Expr.IsField) then
   raise EACRException.Create(10409, ErrorLFieldNameExpected,
            [Token.Text, Token.LineNum, Token.ColumnNum]);
  GetCurrentToken;
  // extract field name, table name
  Expr.GetFieldInfo(TableName, FieldName);
  // add item
  item.TableName := TableName;
  item.FieldName := FieldName;
  item.IsExpression := False;
  item.AllFields := false;
  item.Pseudonym := '';
  item.ValueExpr := nil;
  Fields.Append(item);
 finally
  Expr.Free;
 end;

end;

begin
    // parse field specification list
    if (ParseFieldSpecification(Fields)) then
     // get all specs
     repeat
        // end of sql command?
        if (Token.TokenType <> tktComma) then
          break
        else
          // skip ',' token
          GetNextToken([tktString, tktQuotedString, tktBracketedString,
                        tktReservedWord],10410,ErrorLUnexpectedToken);
       // get next spec
       if (not ParseFieldSpecification(Fields)) then
        raise EACRException.Create(10411, ErrorLFieldNameExpected,
                  [Token.Text, Token.LineNum, Token.ColumnNum]);
     until False
    else
      raise EACRException.Create(10412, ErrorLFieldNameExpected,
          [Token.Text, Token.LineNum, Token.ColumnNum]);
end;// ParseFieldList


//------------------------------------------------------------------------------
// get current token
//------------------------------------------------------------------------------
function TACRSQLCommand.GetCurrentToken: Boolean;
begin
 Result := LLex.GetCurrentToken(Token);
end;// GetCurrentToken


//------------------------------------------------------------------------------
// get next token
//------------------------------------------------------------------------------
function TACRSQLCommand.GetNextToken: Boolean;
begin
 Result := LLex.GetNextToken(Token);
end;//GetNextToken


//------------------------------------------------------------------------------
// get next token
//------------------------------------------------------------------------------
function TACRSQLCommand.GetNextToken(PermittedTypes: TTokenTypes;
  NativeErrorCode: integer; ErrorText: String): Boolean;
begin
  // get next token
  Result := LLex.GetNextToken(Token);
  if (not Result) then
   if (NativeErrorCode <> -1) then
      raise EACRException.Create(30161, ErrorGUnexpectedEndOfCommand,
                                 [Token.LineNum, Token.ColumnNum]);
  // check token type
  if (Result) then
   if ((PermittedTypes <> []) and (NativeErrorCode <> -1)) then
    if not (Token.TokenType in PermittedTypes) then
      raise EACRException.Create(NativeErrorCode, ErrorText,
                                 [Token.Text, Token.LineNum, Token.ColumnNum])
      else
         Result := False;
end;//GetNextToken


//------------------------------------------------------------------------------
// get result cursor
//------------------------------------------------------------------------------
function TACRSQLCommand.GetResultCursor: TACRCursor;
begin
  Result := nil;
end;//GetResultCursor


//------------------------------------------------------------------------------
// gets result dataset
//------------------------------------------------------------------------------
function TACRSQLCommand.GetResultDataset: TDataset;
begin
  Result := nil;
end;// GetResultDataset



////////////////////////////////////////////////////////////////////////////////
//
// TACRQueryExprNode
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRQueryExprNode.Create;
begin
  NodeType := qetSelect;
  Left := nil;
  Right := nil;
  All := False;
  Corresponding := False;
  CorrespondingFields := nil;
  SelectCommand := nil;
end;// Create


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRQueryExprNode.Create(Src: TACRQueryExprNode);
begin
  NodeType := Src.NodeType;
  Left := Src.Left;
  Right := Src.Right;
  All := Src.All;
  Corresponding := Src.Corresponding;
  CorrespondingFields := Src.CorrespondingFields;
  SelectCommand := Src.SelectCommand;
end;// Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRQueryExprNode.Destroy;
begin
 if (Left <> nil) then
  Left.Free;
 if (Right <> nil) then
  Right.Free;

 if (SelectCommand <> nil) then
  SelectCommand.Free;
end;// Destroy

procedure TACRQueryExprNode.AddNode(NewType: TACRQueryExprType;
  RightNode: TACRQueryExprNode; bAll, bCorresponding: Boolean;
  ColumnList: TStringList);
var
  LeftNode: TACRQueryExprNode;
begin
  // copy current node to left
  LeftNode := TACRQueryExprNode.Create(Self);

  // set childs
  Left := LeftNode;
  Right := RightNode;

  // set current data as: union | except | ...
  SelectCommand := nil;
  NodeType := NewType;
  CorrespondingFields := ColumnList;
  All := bAll;
  Corresponding := bCorresponding;
end;// AddNode



////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLUnion
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSQLUnion.Create(Lexer: TACRLexer; Query: TDataSet);
begin
 FRootNode := nil; // not parsed yet
 inherited Create(Lexer, Query);
end;// Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRSQLUnion.Destroy;
begin
  if (FRootNode <> nil) then
    FRootNode.Free;
  inherited Destroy;
end;// Destroy


//------------------------------------------------------------------------------
// parses [ALL]
//------------------------------------------------------------------------------
function TACRSQLUnion.ParseAll: Boolean;
begin
  Result := IsReservedWord(Token, rwALL);
  // skip token
  if (Result) then
   GetNextToken;
end;//ParseAll


//------------------------------------------------------------------------------
// parses [ <corresponding spec> ]
//------------------------------------------------------------------------------
function TACRSQLUnion.ParseCorrespondingSpec(var ColumnsList: TStringList): Boolean;
begin
  ColumnsList := nil;
  Result := IsReservedWord(Token, rwCORRESPONDING);
  if (Result) then
   begin
    // skip token
    GetNextToken;
    // BY (<corresponding column list>)?
    if (IsReservedWord(Token, rwBY)) then
     begin
      // skip token
      GetNextToken;
      // '(' ?
      if (Token.TokenType = tktLeftParenthesis) then
       begin
        // skip '('
        GetNextToken;

        // parses fields list
        ParseColumnList(ColumnsList);

        // skip ')'
        if (Token.TokenType = tktRightParenthesis) then
          GetNextToken; // end of query is possible here
       end
      else
        raise EACRException.Create(30193, ErrorGOtherTokenExpected,
                    ['(', Token.Text, Token.LineNum, Token.ColumnNum]);
     end;
   end;
end;// ParseCorrespondingSpec


//------------------------------------------------------------------------------
// parses SELECT ...
//------------------------------------------------------------------------------
function TACRSQLUnion.ParseQuerySpecification: TACRQueryExprNode;
var
  SelectNode: TACRQueryExprNode;
begin
  if (IsReservedWord(Token, rwSELECT)) then
   begin
    SelectNode := TACRQueryExprNode.Create;
    SelectNode.NodeType := qetSelect;
    try
     SelectNode.SelectCommand := TACRSQLSelect.Create(LLex, LQuery);
     GetCurrentToken;
    except
     SelectNode.Free;
     raise;
    end;
    Result := SelectNode;
   end
  else
   raise EACRException.Create(30194, ErrorGOtherTokenExpected,
              ['SELECT', Token.Text, Token.LineNum, Token.ColumnNum]);
end;// ParseQuerySpecification


//------------------------------------------------------------------------------
// parses <query specification> | <table value constructor>  | <explicit table>
//------------------------------------------------------------------------------
function TACRSQLUnion.ParseSimpleTable: TACRQueryExprNode;
begin
  Result := ParseQuerySpecification;
end;// ParseSimpleTable


//------------------------------------------------------------------------------
// parses <simple table> |
// <left paren> <non-join query expression> <right paren>
//------------------------------------------------------------------------------
function TACRSQLUnion.ParseNonJoinQueryPrimary: TACRQueryExprNode;
begin
  // '(' ?
  if (Token.TokenType = tktLeftParenthesis) then
   begin
     // skip '('
     GetNextToken;

     // parses <query expression>
     Result := ParseQueryExpression;

     // skip ')'
     if (Token.TokenType = tktRightParenthesis) then
       GetNextToken; // end of query is possible here
   end
  else
   Result := ParseSimpleTable;
end;// ParseNonJoinQueryPrimary


//------------------------------------------------------------------------------
// parses <non-join query primary> |
// <query term> INTERSECT [ ALL ] [ <corresponding spec> ] <query primary>
//------------------------------------------------------------------------------
function TACRSQLUnion.ParseNonJoinQueryTerm: TACRQueryExprNode;
var
  RightNode: TACRQueryExprNode;
  bAll, bCorresponding: Boolean; // union or except
  ColumnList: TStringList; // corresponding spec
begin
  // parse <non-join query primary>
  Result := ParseNonJoinQueryPrimary;

  // intersects?
  while (IsReservedWord(Token, rwINTERSECT)) do
   begin
     // skip INTERSECT token
     GetNextToken; 
 
     // [ALL]
     bAll := ParseAll;
 
     // [ <corresponding spec> ] 
     bCorresponding := ParseCorrespondingSpec(ColumnList);
	
     // parse <non-join query primary> 
     RightNode := ParseNonJoinQueryPrimary;
 
     // make Intersect node
     Result.AddNode(qetIntersect, RightNode, bAll, bCorresponding, ColumnList); 
   end;
end;// ParseNonJoinQueryTerm 
	
 
//------------------------------------------------------------------------------ 
// parses <non-join query term> |
// <query expression> UNION  [ ALL ] [ <corresponding spec> ] <query term> | 
// <query expression> EXCEPT [ ALL ] [ <corresponding spec> ] <query term> 
//------------------------------------------------------------------------------ 
function TACRSQLUnion.ParseNonJoinQueryExpression: TACRQueryExprNode; 
var 
  RightNode: TACRQueryExprNode;
  bUnion, bAll, bCorresponding: Boolean; // union or except
  ColumnList: TStringList; // corresponding spec
  NewType: TACRQueryExprType;
begin
  // parse <non-join query term> 
  Result := ParseNonJoinQueryTerm;
 
  // unions | excepts?
  while (IsReservedWord(Token, rwUNION) or 
         IsReservedWord(Token, rwMINUS) or 
         IsReservedWord(Token, rwEXCEPT)) do
   begin 
     // union or except?
     bUnion := IsReservedWord(Token, rwUNION); 
     // skip UNION | EXCEPT token 
     GetNextToken;
     if (bUnion) then
      NewType := qetUnion 
     else 
      NewType := qetExcept; 
 
     // [ALL] 
     bAll := ParseAll; 
 
     // [ <corresponding spec> ]
     bCorresponding := ParseCorrespondingSpec(ColumnList);
 
     // parse <non-join query term>
     RightNode := ParseNonJoinQueryTerm;
 
     // make union | except node
     Result.AddNode(NewType, RightNode, bAll, bCorresponding, ColumnList);
   end; 
end;// ParseNonJoinQueryExpression 
 
	
//------------------------------------------------------------------------------
// parses <non-join query expression>  | <joined table> 
//------------------------------------------------------------------------------
function TACRSQLUnion.ParseQueryExpression: TACRQueryExprNode;
begin 
  Result := ParseNonJoinQueryExpression; 
end;// ParseQueryExpression
	
	
//------------------------------------------------------------------------------ 
// parses query 
//------------------------------------------------------------------------------ 
procedure TACRSQLUnion.Parse; 
begin
  GetCurrentToken; 
  FRootNode := ParseQueryExpression; 
  // store INTO parameters 
  if (FRootNode.NodeType = qetSelect) then
   begin
    IntoTable := FRootNode.SelectCommand.IntoTable;
    IntoDatabase := FRootNode.SelectCommand.IntoDatabase; 
    IntoMemory := FRootNode.SelectCommand.IntoMemory;
   end;
end;// Parse
 
 
//------------------------------------------------------------------------------
// builds AO
//------------------------------------------------------------------------------
function TACRSQLUnion.BuildAO(Node: TACRQueryExprNode): TACRAO;
var
  LeftAO, RightAO: TACRAO;
  UnionType: TACRUnionType; 
  CorrespondingFields: TACRFields;
  Item: TACRSelectListItem;
  i: integer; 
begin
 if (Node.NodeType = qetSelect) then
   Result := Node.SelectCommand.BuildAOTree(ResultDataset) 
 else 
  begin 
    // union | except | intersect 
    LeftAO := BuildAO(Node.Left);
    RightAO := BuildAO(Node.Right); 
 
    UnionType := autUnion;
    case (Node.NodeType) of
      qetUnion:
         UnionType := autUnion;
      qetExcept: 
         UnionType := autExcept; 
      qetIntersect:
         UnionType := autIntersect;
    end; 
 
    if (Node.CorrespondingFields <> nil) then 
     begin
      CorrespondingFields := TACRFields.Create;
      // fill fields list
      Item.TableName := '';
      Item.Pseudonym := '';
      Item.AllFields := False; 
      for i := 0 to Node.CorrespondingFields.Count-1 do
       begin 
        Item.FieldName := Node.CorrespondingFields.Strings[i];
        CorrespondingFields.Append(Item);
       end
     end
    else
     CorrespondingFields := nil;
    try
     // create AO
     Result := TACRAOUnion.Create(LeftAO, RightAO, UnionType,
                                  Node.Corresponding, not Node.All,
                                  CorrespondingFields);
    finally
     if (CorrespondingFields <> nil) then
      CorrespondingFields.Free;
    end;
  end;
end;// BuildAO


//------------------------------------------------------------------------------
// lock all tables used by this query
//------------------------------------------------------------------------------
procedure TACRSQLUnion.LockAllTables;
begin
 if (FRootNode.NodeType = qetSelect) then
   FRootNode.SelectCommand.LockAllTables
 else
  begin
    if (FRootNode.Left.NodeType = qetSelect) then
     FRootNode.Left.SelectCommand.LockAllTables;
    if (FRootNode.Right.NodeType = qetSelect) then
     FRootNode.Right.SelectCommand.LockAllTables;
  end;
end; // LockAllTables


//------------------------------------------------------------------------------
// unlock all tables used by this query
//------------------------------------------------------------------------------
procedure TACRSQLUnion.UnlockAllTables;
begin
 if (FRootNode.NodeType = qetSelect) then
   FRootNode.SelectCommand.UnlockAllTables
 else
  begin
    if (FRootNode.Left.NodeType = qetSelect) then
     FRootNode.Left.SelectCommand.UnlockAllTables;
    if (FRootNode.Right.NodeType = qetSelect) then
     FRootNode.Right.SelectCommand.UnlockAllTables;
  end;
end; // UnlockAllTables


//------------------------------------------------------------------------------
// builds AO tree
//------------------------------------------------------------------------------
function TACRSQLUnion.BuildAOTree(query: TDataset): TACRAO;
begin
  ResultDataset := query;
  Result := BuildAO(FRootNode);
end;// BuildAOTree




////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLCursorCommand
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSQLCursorCommand.Create(Lexer: TACRLexer; Query: TDataSet);
begin
  RootAO := nil;
  IntoTable := '';
  IntoDatabase := '';
  inherited Create(Lexer, Query);
  FRequestLive := TACRQuery(Query).RequestLive;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRSQLCursorCommand.Destroy;
begin
  if (RootAO <> nil) then
    RootAO.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// executes query
//------------------------------------------------------------------------------
procedure TACRSQLCursorCommand.ExecSQL(
                                        query: TDataset;
                                        IsRoot,
                                        RequestLive: boolean;
                                        var ReadOnly: boolean
                                      );
var
  dest:       TACRTable;
  log:        string;

begin
  RowsAffected := 0;
  FRequestLive := RequestLive;

  if (RootAO <> nil) then
    RootAO.Free;
  RootAO := BuildAOTree(query);
  try
    RootAO.LockTable;
  except
    RootAO.Free;
    RootAO := nil;
    raise;
  end;
  try
    if ((IntoTable <> '') and
       (not RootAO.FIsAOTable) and (IntoTable <> '')) then
     begin
      if (not IntoMemory) then
       if (IntoDatabase = '') then
        IntoDatabase := TACRDataset(query).DatabaseName;
      RootAO.SetResultTable(IntoMemory or TACRDataset(query).InMemory,IntoTable,IntoDatabase);
     end;


    RootAO.Execute(IsRoot);

    // specify read-only result?
    if (IsRoot) then
     begin
      ReadOnly := False;
      if (not RequestLive) then
        ReadOnly := True
      else
       begin
        // not table => not updatable result
        if ((not (RootAO is TACRAOTable)) or
            ((RootAO is TACRAOTable) and (not RootAO.RequestLive))) then
         ReadOnly := True;
       end;
       TACRDataSet(RootAO.ResultDataset).Handle.ReadOnly := ReadOnly;
     end;

    // select into table?

    if (RootAO.FIsAOTable and (IntoTable <> '')) then
     begin
      dest := nil;
      try
        dest:= TACRTable.Create(nil);
        dest.TableName := IntoTable;
        dest.InMemory := IntoMemory or TACRDataset(query).InMemory;
        if (dest.InMemory) then
          dest.DatabaseName := ACRMemoryDatabaseName
        else
         dest.DatabaseName := TACRDataset(query).DatabaseName;
        if (not dest.ImportTable(RootAO.ResultDataset,log)) then
          raise EACRException.Create(10403,ErrorLImportTableCannotCopyData,[log]);
      finally
       dest.Free;
      end;
     end;
  finally
    RootAO.UnlockTable;
  end;
end;// ExecSQL


//------------------------------------------------------------------------------
// get result cursor
//------------------------------------------------------------------------------
function TACRSQLCursorCommand.GetResultCursor: TACRCursor;
begin
  Result := TACRDataSet(RootAO.ResultDataset).Handle;
end;//GetResultCursor


//------------------------------------------------------------------------------
// get result dataset
//------------------------------------------------------------------------------
function TACRSQLCursorCommand.GetResultDataset: TDataset;
begin
 Result := RootAO.ResultDataset;
end;// GetResultDataset



////////////////////////////////////////////////////////////////////////////////
//
// TACRTableReference
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRTableReference.Create;
begin
  TableType := attTable;
  LeftTable := nil;
  RightTable := nil;
  SearchCondition := nil;
  UsingFields := nil;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRTableReference.Destroy;
begin
  if (LeftTable <> nil) then
    LeftTable.Free;
  if (RightTable <> nil) then
    RightTable.Free;
  if (UsingFields <> nil) then
    UsingFields.Free;
  if (SearchCondition <> nil) then
    SearchCondition.Free;
end;//Destroy


//------------------------------------------------------------------------------
// makes join (left and right node-table)
//------------------------------------------------------------------------------
procedure TACRTableReference.MakeJoin(RightNode: TACRTableReference; JType: TACRJoinType;
                 IsNatural: Boolean; Fields: TStringList;
                 OnCondition: TACRExpression);
var
 LeftNode: TACRTableReference;
begin
 // copy current node to left
 LeftNode := TACRTableReference.Create;
 LeftNode.TableType := TableType;
 LeftNode.DatabaseName := DatabaseName;
 LeftNode.SessionName := SessionName;
 LeftNode.TableName := TableName;
 LeftNode.Pseudonym := Pseudonym;
 LeftNode.Password := Password;
 LeftNode.InMemory := InMemory;
 LeftNode.NaturalJoin := NaturalJoin;
 LeftNode.JoinType := JoinType;
 LeftNode.LeftTable := LeftTable;
 LeftNode.RightTable := RightTable;
 LeftNode.UsingFields := UsingFields;
 LeftNode.SearchCondition := SearchCondition;

 // set childs
 LeftTable := LeftNode;
 RightTable := RightNode;

 // set current data as join
 TableType := attJoinedTable;
 JoinType := JType;
 NaturalJoin := IsNatural;
 UsingFields := Fields;
 SearchCondition := OnCondition;
end; // MakeJoin


////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLInsert
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSQLInsert.Create(Lexer: TACRLexer; Query: TDataSet);
begin
  FieldNames := TStringList.Create;
  FieldValues := TList.Create;
  InternalSelecter := nil;
  inherited Create(Lexer, Query);
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRSQLInsert.Destroy;
var
  i: integer;
begin
  Password:='';
  FieldNames.Free;
  for i:=0 to FieldValues.Count-1 do
    TACRExpression(FieldValues[i]).Free;
  FieldValues.Free;
  if Assigned(InternalSelecter) then
    InternalSelecter.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// ParseValuesList
//------------------------------------------------------------------------------
procedure TACRSQLInsert.ParseValuesList;
var
  Expr: TACRExpression;
begin
  // '(' ?
  if Token.TokenType <> tktLeftParenthesis then
    Exit;

  // list...
  repeat
    if (not GetNextToken) then
      raise EACRException.Create(30203, ErrorGUnexpectedEndOfCommand,
                                 [Token.LineNum, Token.ColumnNum]);

    Expr := TACRExpression.Create;
    Expr.Params := TACRQuery(LQuery).Params;
    Expr.InMemory := TACRQuery(LQuery).InMemory;
    Expr.DatabaseName := TACRQuery(LQuery).DatabaseName;
    Expr.SessionName := TACRQuery(LQuery).SessionName;
    Expr.ParseForValueExpression(LLex);
    GetCurrentToken;
    if Expr.IsEmpty then
      raise EACRException.Create(30357, ErrorGParseParameterError,
                                 [Token.Text, Token.LineNum, Token.ColumnNum]);
    FieldValues.Add(Expr);

    //GetNextToken([tktComma, tktRightParenthesis], 02074);
  until Token.TokenType = tktRightParenthesis;
  // ')'
end;//ParseValuesList


//------------------------------------------------------------------------------
// parses query
//------------------------------------------------------------------------------
procedure TACRSQLInsert.Parse;
begin
  if not GetCurrentToken then
    raise EACRException.Create(30204, ErrorGBlankSqlCommand);

  if (not GetNextToken) then
    raise EACRException.Create(30205, ErrorGUnexpectedEndOfCommand,
                               [Token.LineNum, Token.ColumnNum]);

  // INTO
  if not IsReservedWord(Token, rwINTO) then
    raise EACRException.Create(30206, ErrorGOtherTokenExpected,
                         ['INTO', Token.Text, Token.LineNum, Token.ColumnNum]);

  // MEMORY
  GetNextToken([tktReservedWord,tktString,tktQuotedString,tktBracketedString],
                   30207, ErrorGTableNameExpected);

  if IsReservedWord(Token, rwMEMORY) then
   begin
    IntoMemory := true;
    GetNextToken([tktReservedWord,tktString,tktQuotedString, tktBracketedString],
                 30208, ErrorGTableNameExpected);
   end;

  // table_name
  TableName := Token.Text;

  // PASSWORD ?
  if (not GetNextToken) then
    raise EACRException.Create(30209, ErrorGUnexpectedEndOfCommand,
                               [Token.LineNum, Token.ColumnNum]);

  if IsReservedWord(Token, rwPASSWORD) then
    begin
      GetNextToken([tktQuotedString], 30210, ErrorGQuotedPasswordDataExpected);
      Password := Token.Text;
      if (not GetNextToken) then
        raise EACRException.Create(30211, ErrorGUnexpectedEndOfCommand,
                                   [Token.LineNum, Token.ColumnNum]);
    end;
  // '(' ?
  if Token.TokenType = tktLeftParenthesis then
    begin
      // list...
      repeat
        GetNextToken;
        if not (Token.TokenType in [tktString,tktQuotedString,
                                    tktBracketedString,tktReservedWord]) then
          raise EACRException.Create(02156, ErrorGFieldNameExpected,
                                  [Token.Text, Token.LineNum, Token.ColumnNum]);

        //GetNextToken(02073);
        FieldNames.Add(Token.Text);

        GetNextToken([tktComma, tktRightParenthesis],
                     30212, ErrorGRightParenthesisOrCommaExpected);
      until Token.TokenType = tktRightParenthesis;
      // ')'
      if (not GetNextToken) then
        raise EACRException.Create(30213, ErrorGUnexpectedEndOfCommand,
                                   [Token.LineNum, Token.ColumnNum]);
    end;
  // 'VALUES' ?
  if Token.ReservedWord = rwVALUES then
    begin
      GetNextToken([tktLeftParenthesis], 30214, ErrorGLeftParenthesisExpected);
      ParseValuesList;
    end
  else
    if Token.ReservedWord = rwSELECT then
      InternalSelecter := TACRSQLSelect.Create(LLex, LQuery);
end;//Parse


//------------------------------------------------------------------------------
// executes query
//------------------------------------------------------------------------------
procedure TACRSQLInsert.ExecSQL(query: TDataset; IsRoot, RequestLive: boolean; var ReadOnly: boolean);
var
   T:         TACRTable;
   i:         Integer;
   ds:        TDataSet;
   s:         String;
   FieldNo:   Integer;
   Data:      TACRVariant;
   RecCount:  Integer;
begin
  RowsAffected := 0;
  T := TACRTable.Create(query.Owner);
  try
    T.DatabaseName := (query as TACRQuery).DatabaseName;
    T.SessionName := (query as TACRQuery).SessionName;
    T.TableName := TableName;
    //T.Password := Password;
    T.InMemory := IntoMemory or TACRDataset(query).InMemory;
    t.LogComponent := TACRDataset(query);
    //if T.InMemory then
    //  T.DatabaseName := ACRMemoryDatabaseName;

    T.Open;
    if InternalSelecter = nil then
      begin
        try
            T.Insert;
            for i:=0 to FieldValues.Count-1 do
              begin
                // Get Field No
                FieldNo := i;
                if (FieldNames.Count = FieldValues.Count) then
                 if (FieldNames.Count > 0) then
                  if (FieldNames[i] <> '') then
                   FieldNo := T.Fields.IndexOf(T.Fields.FieldByName(FieldNames[i]));

                Data := TACRExpression(FieldValues[i]).GetValue;
                if (Data <> nil) then
                  SetACRVariantIntoField(Data, T.Fields.Fields[FieldNo])
                else
                  raise EACRException.Create(30356, ErrorGParameterValueUndefined,
                        [FieldNames[i], AftToStr(FieldTypeToACRAdvFieldType(
                                 T.Fields.FieldByName(FieldNames[i]).DataType))]);
              end;
              T.Post;
            except
              T.Cancel;
              raise;
            end;
        Inc(RowsAffected);
      end
    else
      begin
        InternalSelecter.ExecSQL(query, IsRoot, False, ReadOnly);
        ds := InternalSelecter.GetResultDataset;
        RecCount := T.RecordCount;
        s := CopyDatasets(ds,T,True,tbopCopy);
        if (s <> '') then
         raise EACRException.Create(11350,ErrorLInsertSelectFailed,[T.TableName,s]);
        RowsAffected := T.RecordCount - RecCount;
      end; // insert from select
    T.Close;
  finally
    T.Free;
  end;
end;//ExecSQL



////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLUpdate
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSQLUpdate.Create(Lexer: TACRLexer; Query: TDataSet);
begin
  InternalSelecter := nil;
  FieldNames  := TStringList.Create;
  FieldValues := TList.Create;
  TableName := '';
  inherited Create(Lexer, Query);
end;//Create


//------------------------------------------------------------------------------
// parses query
//------------------------------------------------------------------------------
procedure TACRSQLUpdate.ExecSQL(query: TDataset; IsRoot, RequestLive: boolean; var ReadOnly: boolean);
var
   i:           Integer;
   ds:          TDataSet;
   oldRecCount: Integer;
   values:      array of TACRVariant;
   ConstsOnly:  Boolean;
   AdvFieldDef: TACRAdvFieldDef;

 procedure UpdateConstsOnly;
 var i: Integer;
 begin
    SetLength(values,FieldValues.Count);
    for i:=0 to FieldValues.Count-1 do
     begin
      values[i] := TACRVariant.Create;
      values[i].Assign(TACRExpression(FieldValues[i]).GetValue,True);
      TACRExpression(FieldValues[i]).Free;
     end;
    try
     oldRecCount := ds.RecordCount;
     TACRTable(ds).UpdateVisibleRecords(FieldNames,values);
     Inc(RowsAffected,oldRecCount)
    finally
     for i:=0 to FieldValues.Count-1 do
       values[i].Free;
    end;
 end; // UpdateConstsOnly

 procedure UpdateGeneral;
 var i: Integer;
 begin
  try
    ds.First;
    while (not ds.Eof) and (ds.RecordCount > 0) do
      begin
        ds.Edit;
        for i:=0 to FieldNames.Count-1 do
          SetACRVariantIntoField(
                                  TACRExpression(FieldValues[i]).GetValue,
                                  ds.Fields.FieldByName(FieldNames[i])
                                );
        oldRecCount := ds.RecordCount;
        ds.post;
        Inc(RowsAffected);
        if (oldRecCount = ds.RecordCount) then
          ds.Next;
      end;
  finally
    for i:=0 to FieldValues.Count-1 do
     TACRExpression(FieldValues[i]).Free;
  end;
 end; // UpdateGeneral


begin
  RowsAffected := 0;
  ConstsOnly := True;
  InternalSelecter.ExecSQL(query, IsRoot, True, ReadOnly);
  ds := InternalSelecter.GetResultDataset;
  TACRDataSet(ds).Handle.LockTableIRW(10677);
  try
    TACRDataset(ds).LogComponent := TACRDataset(query);
    if (TACRDataset(ds).IsBeforeUpdateRecordAssigned or
        TACRDataset(ds).IsAfterUpdateRecordAssigned) then
     ConstsOnly := False;
    for i:=0 to FieldValues.Count-1 do
     begin
      TACRExpression(FieldValues[i]).AssignCursor(InternalSelecter.GetResultCursor);
      if (ConstsOnly) then
       begin
        if (not TACRExpression(FieldValues[i]).IsConst) then
         ConstsOnly := False;
        if (ConstsOnly) then
         begin
          AdvFieldDef := TACRDataSet(ds).AdvFieldDefs.Find(FieldNames[i]);
          if (AdvFieldDef = nil) then
           raise EACRException.Create(11590,ErrorLCannotFindFieldInTable,[FieldNames[i],TACRTable(ds).TableName]);
          ConstsOnly := not IsBLOBFieldType(AdvFieldDef.DataType);
         end;
       end;
     end;
    if (ConstsOnly) then
     UpdateConstsOnly
    else
     UpdateGeneral;
  finally
   TACRDataSet(ds).Handle.UnlockTableIRW(10678);
   if Assigned(InternalSelecter) then
    InternalSelecter.Free;
   FieldNames.Free;
   FieldValues.Free;
  end;
end;//ExecSQL


//------------------------------------------------------------------------------
// parses query
//------------------------------------------------------------------------------
procedure TACRSQLUpdate.Parse;
var
  oldPos, newPos, i,j: Integer;
  Expr: TACRExpression;
begin
  if not GetCurrentToken then
    raise EACRException.Create(30216, ErrorGBlankSqlCommand);

  GetNextToken([tktReservedWord,tktString,tktQuotedString,tktBracketedString],
               30217, ErrorGTableNameExpected);

  // MEMORY
  if IsReservedWord(Token, rwMEMORY) then
   begin
    IntoMemory := true;
    GetNextToken([tktReservedWord,tktString,tktQuotedString,tktBracketedString],
                 30218, ErrorGTableNameExpected);
   end;
  // table_name
  TableName := Token.Text;

  // get next token
  GetNextToken;

  // PASSWORD ?
  if IsReservedWord(Token, rwPASSWORD) then
    begin
      GetNextToken([tktQuotedString],
                   30219, ErrorGQuotedPasswordDataExpected);
      Password := Token.Text;
      GetNextToken;
    end;

  // SET
  if not IsReservedWord(Token, rwSET) then
    raise EACRException.Create(30220, ErrorGOtherTokenExpected,
             ['SET', Token.Text, Token.LineNum, Token.ColumnNum]);

  GetNextToken;

  // field_name = value, ...
  repeat
    // Field_name
    FieldNames.Add(Token.Text);
    GetNextToken;
    if Token.Text <> '=' then
      raise EACRException.Create(30221, ErrorGOtherTokenExpected,
                   ['=', Token.Text, Token.LineNum, Token.ColumnNum]);
    GetNextToken;

    Expr := TACRExpression.Create;
    Expr.Params := TACRQuery(LQuery).Params;
    Expr.InMemory := TACRQuery(LQuery).InMemory;
    Expr.DatabaseName := TACRQuery(LQuery).DatabaseName;
    Expr.SessionName := TACRQuery(LQuery).SessionName;
    Expr.ParseForValueExpression(LLex);
    FieldValues.Add(Expr);

    if not GetCurrentToken then
      // end of command
      break
    else
      begin
        // ',' or 'WHERE'
        if Token.ReservedWord = rwWHERE then
          break;

        if Token.TokenType <> tktComma then
            raise EACRException.Create(30222, ErrorGOtherTokenExpected,
                     [',', Token.Text, Token.LineNum, Token.ColumnNum]);

        // get next field_name
        if (not GetNextToken) then
            raise EACRException.Create(30223, ErrorGUnexpectedEndOfCommand,
                                       [Token.LineNum, Token.ColumnNum]);
      end;
  until false;

 // sql: select f1, f2 from table_name where .......
 //      update table_name set a=b, c=d where .......
  with  LLex.CurrentCommand do
    begin
      // Free Params
      for i:=0 to CurrentTokenNo-1 do
        if Tokens[i].ParamValue <> nil then
          begin
            Tokens[i].ParamValue.Free;
            Tokens[i].ParamValue := nil;
          end;

      oldPos := CurrentTokenNo;
      newPos := 4; // select * from tablename where
      if IntoMemory then Inc(newPos);
      //  Shift Tokens (where clause) ?
      if IsReservedWord(Token, rwWHERE) then
        for i:=oldPos to length(Tokens)-1 do
          Tokens[newPos+i-oldPos] := Tokens[i];
      SetLength(Tokens, length(Tokens)-(oldPos-newPos));
      NumTokens := length(Tokens);
      CurrentTokenNo:=0;

      Tokens[0].TokenType := tktReservedWord;
      Tokens[0].ReservedWord := rwSELECT;
      Tokens[0].Text := 'select';

      Tokens[1].TokenType := tktNone;
      Tokens[1].ReservedWord := rwNone;
      Tokens[1].Text:=Asterisk;

      Tokens[2].TokenType := tktReservedWord;
      Tokens[2].ReservedWord := rwFROM;
      Tokens[2].Text:='from';

      j := 3;
      if IntoMemory then
       begin
        Tokens[3].TokenType := tktReservedWord;
        Tokens[3].ReservedWord := rwMEMORY;
        Tokens[3].Text:=ACRMemoryDatabaseName;
        inc(j);
       end;

      Tokens[j].TokenType := tktString;
      Tokens[j].ReservedWord := rwNone;
      Tokens[j].Text:=TableName;

      InternalSelecter := TACRSQLSelect.Create(LLex, LQuery);
    end;
end;//Parse



////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLDelete
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSQLDelete.Create(Lexer: TACRLexer; Query: TDataSet);
begin
  InternalSelecter := nil;
  FullDelete := false;
  TableName := '';
  inherited Create(Lexer, Query);
end;//Create


//------------------------------------------------------------------------------
// Execute SQL statement
//------------------------------------------------------------------------------
procedure TACRSQLDelete.ExecSQL(query: TDataset; IsRoot, RequestLive: boolean; var ReadOnly: boolean);
var
  ds:       TDataSet;

  procedure FastDelete;
  begin
   RowsAffected := ds.RecordCount;
   TACRDataSet(ds).DeleteVisibleRecords;
  end;

  procedure GeneralDelete;
  begin
    TACRDataset(ds).LogComponent := TACRDataset(query);
    RowsAffected := 0;
    TACRDataSet(ds).Handle.LockTableIRW(10679);
    try
      ds.First;
      while not ds.Eof do
        begin
          ds.Delete;
          Inc(RowsAffected);
        end;
    finally
     TACRDataSet(ds).Handle.UnlockTableIRW(10680);
    end;
  end;

begin
  InternalSelecter.ExecSQL(query, IsRoot, True, ReadOnly);
  try
   ds := InternalSelecter.GetResultDataset;
   if (TACRDataset(ds).IsBeforeDeleteRecordAssigned or
       TACRDataset(ds).IsAfterDeleteRecordAssigned) then
    GeneralDelete
   else
    FastDelete;
  finally
   if Assigned(InternalSelecter) then
     InternalSelecter.Free;
  end;
end;//ExecSQL


//------------------------------------------------------------------------------
// parses query
//------------------------------------------------------------------------------
procedure TACRSQLDelete.Parse;
var
   i: integer;
begin
  // CONVERT DELETE TO SELECT
  with  LLex.CurrentCommand do
    begin
      SetLength(Tokens, length(Tokens)+1);
      inc(NumTokens);
      for i:=length(Tokens)-1 downto 1 do
        begin
          Tokens[i] := Tokens[i-1];
        end;
      Tokens[0].TokenType := tktReservedWord;
      Tokens[0].ReservedWord := rwSELECT;
      Tokens[0].Text := 'select';
      Tokens[1].TokenType := tktNone;
      Tokens[1].ReservedWord := rwNone;
      Tokens[1].Text := Asterisk;
    end;
  if not GetCurrentToken then
    raise EACRException.Create(30224, ErrorGUnexpectedEndOfCommand,
                               [Token.LineNum, Token.ColumnNum]);

  InternalSelecter := TACRSQLSelect.Create(LLex, LQuery);

end;//Parse



////////////////////////////////////////////////////////////////////////////////
//
// TSQLFieldDef
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TSQLFieldDef.Create;
begin
  FieldName := '';
  
  newFieldType := False;
  FieldType := aftUnknown;

  newLength := False;
  Length := 0;

  newRequired := False;

  DefaultValue := TACRVariant.Create;
  MinValue := TACRVariant.Create;
  MaxValue := TACRVariant.Create;

  // Blob settings
  newBlobblocksize := False;
  newBlobCompressionAlgorithm := False;
  newBlobCompressionMode := False;

  //Autoinc
  newAutoincIncrement := False;
  newAutoincLastValue := False;
  newAutoincMinValue := False;
  newAutoincMaxValue := False;
  newAutoincCycled := False;

  // Default Value
  newDefaultValue := False;
  // Min
  newMinValue := False;
  // Max
  newMaxValue := False;

  // PK
  newPrimaryKey := False;
  // Unique
  newUnique := False;

end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TSQLFieldDef.Destroy;
begin
  DefaultValue.Free;
  MinValue.Free;
  MaxValue.Free;
end;//Destroy



////////////////////////////////////////////////////////////////////////////////
//
// TSQLFieldDefs
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TSQLFieldDefs.Create;
begin
  List := TList.Create;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TSQLFieldDefs.Destroy;
var i: Integer;
begin
  for i:=0 to List.Count-1 do
    Items[i].Free;
  List.Free;
end;//Destroy


//------------------------------------------------------------------------------
// Create SQLFieldDef and add it to List
//------------------------------------------------------------------------------
function TSQLFieldDefs.AddCreated: TSQLFieldDef;
begin
  Result := TSQLFieldDef.Create;
  List.Add(Result);
end;//AddCreated


//------------------------------------------------------------------------------
// Get Count
//------------------------------------------------------------------------------
function TSQLFieldDefs.GetCount: Integer;
begin
  Result := List.Count;
end;//GetCount


//------------------------------------------------------------------------------
// Get Def
//------------------------------------------------------------------------------
function TSQLFieldDefs.GetDef(Index: Integer): TSQLFieldDef;
begin
  Result := TSQLFieldDef(List[Index]);
end;//GetDef




////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLTableManipulation
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRDDLTableManipulation.Create(Lexer: TACRLexer; Query: TDataSet);
begin
  Password := '';
  TableName:='';             // Table Name
  SQLFieldDefs := TSQLFieldDefs.Create;
  SetLength(PrimaryKeyFields, 0);
  UniqueFields := TStringList.Create;
  PrimaryIndexName := '';
  FForeignKeyDefs := nil;
  inherited Create(Lexer, Query);
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRDDLTableManipulation.Destroy;
begin
  SQLFieldDefs.Free;
  UniqueFields.Free;
  if (FForeignKeyDefs <> nil) then
   FForeignKeyDefs.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// table element list
//------------------------------------------------------------------------------
procedure TACRDDLTableManipulation.ParseTableElementList;
var
  FieldDef: TSQLFieldDef;
begin
  if (Token.TokenType <> tktLeftParenthesis) then
    raise EACRException.Create(30226, ErrorGOtherTokenExpected,
            ['(', Token.Text, Token.LineNum, Token.ColumnNum]);

  // Columns Loop
  repeat
    // end of create table sql ')' ?
    if not LLex.GetCurrentToken(Token) then
      Break;
    if (Token.TokenType = tktRightParenthesis) then
      break;
    GetNextToken([tktString, tktQuotedString, tktBracketedString, tktReservedWord],
                 30227, ErrorGFieldNameExpected);

    // end of column list?
{    if IsReservedWord(Token, rwPASSWORD) then
     begin
      ParsePassword;
      continue;
     end else
    if IsReservedWord(Token, rwBLOB_COMPRESSION_LEVEL) then
     begin
      ParseBlobCompressionLevel;
      continue;
     end else
    if IsReservedWord(Token, rwBLOB_BLOCK_SIZE) then
     begin
      ParseBlobBlockSize;
      continue;
     end else
}
    if ParsePrimaryKey then continue;
    if ParseForeignKey then continue;

    // Field Params

    FieldDef := SQLFieldDefs.AddCreated;


    // column name
    FieldDef.FieldName := Token.Text;

    GetNextToken([tktString, tktBracketedString, tktReservedWord],
                 30228, ErrorGFieldTypeExpected);


    // Fill Column Type
    FillColumnType(FieldDef);

    // until ',' or ')' or 'any_word' after FIELDNAME FIELDTYPE ...
    while not (Token.TokenType in [tktComma,tktRightParenthesis]) do
     begin

      // NULL or NOT NULL
      if Token.ReservedWord in [rwNOT,rwNULL] then
        // Fill Column Required value
        ParseColumnRequared(FieldDef)
      else

      // BlobBlockSize
      if IsReservedWord(Token, rwBLOBBLOCKSIZE) then
       begin
        FieldDef.newBlobBlockSize := true;
        FieldDef.BlobBlockSize := ParseInteger;
       end
      else

      // BlobCompressionMode
      if IsReservedWord(Token, rwBLOBCOMPRESSIONMODE) then
       begin
        FieldDef.newBlobCompressionMode := true;
        FieldDef.BlobCompressionMode := ParseInteger;
       end
      else

      // BlobCompressionAlgorithm
      if IsReservedWord(Token, rwBLOBCOMPRESSIONALGORITHM) then
       begin
        // {NONE | ZLIB | BZIP | PPM}
        GetNextToken;
        FieldDef.BlobCompressionAlgorithm := GetCompressionAlgorithm(Token.Text);
        FieldDef.newBlobCompressionAlgorithm := true;
       end
      // DEFAULT
      else if ParseDefaultValue(FieldDef) then Continue
      // MINVALUE
      else if ParseMinValue(FieldDef) then Continue
      // MAXVALUE
      else if ParseMaxValue(FieldDef) then Continue
      // PRIMARY KEY
      else if ParseFieldPrimaryKey(FieldDef) then Continue
      // UNIQUE
      else if ParseFieldUnique(FieldDef) then Continue;


      GetNextToken([tktComma, tktRightParenthesis,
                    tktString, tktQuotedString, tktBracketedString,
                    tktReservedWord],
                    30229, ErrorGRightParenthesisOrCommaExpected);


     end;//while
  until False;//Columns Loop
end;//ParseTableElementList


//------------------------------------------------------------------------------
// parse TableName token
//------------------------------------------------------------------------------
procedure TACRDDLTableManipulation.ParseTableNameToken;
begin
  if not (Token.TokenType in [tktString, tktReservedWord, tktBracketedString]) then
    raise EACRException.Create(30230, ErrorGTableNameExpected,
                        [Token.Text, Token.LineNum, Token.ColumnNum]);

  // MEMORY
  if IsReservedWord(Token, rwMEMORY) then
   begin
    IntoMemory := true;
    GetNextToken([tktReservedWord,tktString,tktQuotedString,tktBracketedString],
                  30231, ErrorGTableNameExpected);
   end;
  // table_name
  TableName := Token.Text;
  GetNextToken;
end;//ParseTableNameToken


//------------------------------------------------------------------------------
// Fill table column type into Structure
//------------------------------------------------------------------------------
procedure TACRDDLTableManipulation.FillColumnType(var FieldDef: TSQLFieldDef);
begin
  // column type
  FieldDef.FieldType := GetFieldType(Token.Text);
  FieldDef.newFieldType := True;

  // Unknown type ?
  if FieldDef.FieldType = aftUnknown then
    raise EACRException.Create(30232, ErrorGUnknownFieldType,
                 [Token.Text, Token.LineNum, Token.ColumnNum]);

  // sizeble type ? ex: String(255)
  if (IsStringFieldType(FieldDef.FieldType) or
      IsBytesFieldType(FieldDef.FieldType)) then
    begin
     // '('
     GetNextToken([tktLeftParenthesis],
                  30233, ErrorGLeftParenthesisExpected);
     // Int
     GetNextToken([tktInt], 30235, ErrorGDecimalConstantExpected);

     FieldDef.Length := StrToInt(Token.Text);
     FieldDef.newLength := True;
     
     // ')'
     GetNextToken([tktRightParenthesis],
                  30236, ErrorGRightParenthesisOrCommaExpected);
    end
  else
    FieldDef.Length := 0;

  if IsAutoincFieldType(FieldDef.FieldType) then
   if (GetNextToken) then
    // '('
    if Token.TokenType = tktLeftParenthesis then
     begin
      repeat
       if (not GetNextToken) then
        raise EACRException.Create(30290, ErrorGUnexpectedEndOfCommand,
                                  [Token.LineNum, Token.ColumnNum]);
       // ')'
       if (Token.TokenType = tktRightParenthesis) then
         break;

       // ','
       if (Token.TokenType = tktComma) then
         GetNextToken;


       case Token.ReservedWord of
         rwINCREMENT:
           begin
             FieldDef.AutoincIncrement := ParseInteger;
             FieldDef.newAutoincIncrement := True;
           end;
         rwINITIALVALUE:
           begin
             FieldDef.AutoincLastValue := ParseInteger;
             FieldDef.newAutoincLastValue := True;
           end;
         rwMINVALUE:
           begin
             FieldDef.AutoincMinValue := ParseInteger;
             FieldDef.newAutoincMinValue := True;
           end;
         rwMAXVALUE:
           begin
             FieldDef.AutoincMaxValue := ParseInteger;
             FieldDef.newAutoincMaxValue := True;
           end;
         rwCYCLED:
           begin
             FieldDef.AutoincCycled := True;
             FieldDef.newAutoincCycled := True;
           end;
         rwNOMINVALUE:
           begin
             FieldDef.AutoincMinValue := 0;//Low(Int64);
             FieldDef.newAutoincMinValue := True;
           end;
         rwNOMAXVALUE:
           begin
             FieldDef.AutoincMaxValue := High(Int64);
             FieldDef.newAutoincMaxValue := True;
           end;
         rwNOCYCLED:
           begin
             FieldDef.AutoincCycled := False;
             FieldDef.newAutoincCycled := True;
           end;
         else
           begin
             if (FieldDef.FieldType = aftAutoinc) then
              begin
               FieldDef.FieldType := GetFieldType(Token.Text);
               // Convert to Autoinc type
               case FieldDef.FieldType of
                aftShortint:  FieldDef.FieldType := aftAutoIncShortint;
                aftSmallint:  FieldDef.FieldType := aftAutoIncSmallint;
                aftInteger:   FieldDef.FieldType := aftAutoInc;
                aftLargeint:  FieldDef.FieldType := aftAutoIncLargeint;
                aftByte:      FieldDef.FieldType := aftAutoIncByte;
                aftWord:      FieldDef.FieldType := aftAutoIncWord;
                aftCardinal:  FieldDef.FieldType := aftAutoIncCardinal;

                aftAutoInc,   // check for correct types
                aftAutoIncShortint,
                aftAutoIncSmallint,
                aftAutoIncInteger,
                aftAutoIncLargeint,
                aftAutoIncByte,
                aftAutoIncWord,
                aftAutoIncCardinal: ;
                else
                  raise EACRException.Create(30295, ErrorGUnsupportedAutoincDataType,
                                  [Token.Text, Token.LineNum, Token.ColumnNum]);
               end;//case

               if FieldDef.FieldType = aftUnknown then
                begin
                 FieldDef.FieldType := aftAutoinc;
                 raise EACRException.Create(30293, ErrorGUnexpectedToken,
                                  [Token.Text, Token.LineNum, Token.ColumnNum]);
                end
              end
             else
              raise EACRException.Create(30294, ErrorGUnexpectedToken,
                               [Token.Text, Token.LineNum, Token.ColumnNum]);

           end;//if
       end;// case
       
      until False;
      GetNextToken;
     end
    else
   else
  else
   GetNextToken;
end;//FillColumnType


//------------------------------------------------------------------------------
// Fill column requared value into Structure
//------------------------------------------------------------------------------
procedure TACRDDLTableManipulation.ParseColumnRequared(var FieldDef: TSQLFieldDef);
begin
  if IsReservedWord(Token, rwNOT) then
    begin
      GetNextToken([tktReservedWord], 30237, ErrorGUnexpectedEndOfCommand);
      // NOT NULL
      if IsReservedWord(Token, rwNULL) then
        begin
          FieldDef.newRequired := True;
          FieldDef.Required := True;
        end
      else
        raise EACRException.Create(02011, ErrorGNullKeywordExpected,
                                 [Token.Text, Token.LineNum, Token.ColumnNum]);
    end
  else
    // NULL
    if IsReservedWord(Token, rwNULL) then
      begin
        FieldDef.newRequired := True;
        FieldDef.Required := False;
      end;
end;//ParseColumnRequared


//------------------------------------------------------------------------------
// Fill BlobBlockSize value into Structure
//------------------------------------------------------------------------------
//procedure TACRDDLTableManipulation.FillBlobBlockSize(var FieldDef: TSQLFieldDef);
//begin
//  GetNextToken([tktInt], 30296, ErrorGIntegerExpected);
//  FieldDef.newBlobBlockSize := True;
//  FieldDef.BlobBlockSize := ParseInteger;
//end;//FillBlobBlockSize


//------------------------------------------------------------------------------
// parse PASSWORD 'aaa'
//------------------------------------------------------------------------------
procedure TACRDDLTableManipulation.ParsePassword;
begin
  if (Token.ReservedWord = rwPASSWORD) then
    begin
      GetNextToken([tktQuotedString], 30238, ErrorGQuotedPasswordDataExpected);
      Password := Token.Text;
      GetNextToken([tktComma, tktRightParenthesis],
                    30239, ErrorGRightParenthesisOrCommaExpected);
    end;
end;//ParsePassword


//------------------------------------------------------------------------------
// parse DEFAULT {const | NULL}
//------------------------------------------------------------------------------
function TACRDDLTableManipulation.ParseDefaultValue(var FieldDef:TSQLFieldDef): Boolean;
var expr: TACRExpression;
begin
  Result := False;
  if (Token.ReservedWord = rwDEFAULT) then
    begin
      GetNextToken([], 30297, ErrorGUnexpectedEndOfCommand);

      if FieldDef.newDefaultValue then
        raise EACRException.Create(30299, ErrorGDefaultValueReDeclared, [FieldDef.FieldName]);

      // Set Value
      FieldDef.newDefaultValue := True;
      if IsReservedWord(Token, rwNULL) then
       begin
        // DEFAULT = NULL ?
        FieldDef.DefaultValue.SetNull;
        GetNextToken;
       end
      else
        begin
          expr := TACRExpression.Create;
          try
            expr.ParseForValueExpression(LLex);
            FieldDef.DefaultValue.Assign(expr.GetValue(True));
            FieldDef.DefaultValue.Cast(AdvancedFieldTypeToBaseFieldType(FieldDef.FieldType));
          finally
            LLex.GetCurrentToken(Token);
            expr.Free;
          end;
        end;
      Result := True;
    end;
end;//ParseDefaultValue


//------------------------------------------------------------------------------
// parse MINVALUE value
//------------------------------------------------------------------------------
function TACRDDLTableManipulation.ParseMinValue(var FieldDef: TSQLFieldDef): Boolean;
begin
  Result := False;
  if (IsReservedWord(Token, rwMINVALUE)) then
    begin
      GetNextToken([], 30300, ErrorGUnexpectedEndOfCommand);
      FieldDef.MinValue.AsString := Token.Text;
      FieldDef.MinValue.Cast(bftSignedInt64);
      FieldDef.newMinValue := True;
      Result := True;
      GetNextToken;
    end
  else
  if (IsReservedWord(Token, rwNOMINVALUE)) then
    begin
      FieldDef.newMinValue := True;
      FieldDef.MinValue.SetNull;
      Result := True;
      GetNextToken;
    end
end;//ParseMinValue


//------------------------------------------------------------------------------
// parse MaxVALUE value
//------------------------------------------------------------------------------
function TACRDDLTableManipulation.ParseMaxValue(var FieldDef: TSQLFieldDef): Boolean;
begin
  Result := False;
  if (IsReservedWord(Token, rwMaxVALUE)) then
    begin
      GetNextToken([], 30300, ErrorGUnexpectedEndOfCommand);
      FieldDef.MaxValue.AsString := Token.Text;
      FieldDef.MaxValue.Cast(bftSignedInt64);
      FieldDef.newMaxValue := True;
      Result := True;
      GetNextToken;
    end
  else
  if (IsReservedWord(Token, rwNOMaxVALUE)) then
    begin
      FieldDef.newMaxValue := True;
      FieldDef.MaxValue.SetNull;
      Result := True;
      GetNextToken;
    end

end;//ParseMaxValue


//------------------------------------------------------------------------------
// parse fiald ... PRIMARY KEY
//------------------------------------------------------------------------------
function TACRDDLTableManipulation.ParseFieldPrimaryKey(var FieldDef: TSQLFieldDef): Boolean;
var
  TokenNo, i: Integer;
  desc, nocase: Boolean;
begin
  // PRIMARY [KEY] [ ASC | DESC ]  [ CASE | NOCASE ]
  Result := False;
  // PRIMARY
  if (Token.ReservedWord = rwPRIMARY) then
    begin
      TokenNo := LLex.GetCurrentTokenNo;
      GetNextToken;
      // [KEY] ?
      if (Token.ReservedWord <> rwKEY) then
       LLex.SetCurrentTokenNo(TokenNo, Token);

      desc := false;
      nocase := false;
      // ASC, CASE ?
      repeat
        GetNextToken([], 30333, ErrorGUnexpectedEndOfCommand);
        case Token.ReservedWord of
          rwASC:    desc:=false;
          rwDESC:   desc:=true;
          rwCASE:   nocase:=false;
          rwNOCASE: nocase:=true;
          else
            break;
        end;
      until false;


      FieldDef.newPrimaryKey := True;
      FieldDef.PrimaryKey := True;
      
      // add field to primary index
      i := Length(PrimaryKeyFields);
      SetLength(PrimaryKeyFields, i + 1);
      PrimaryKeyFields[i].FieldName := FieldDef.FieldName;
      PrimaryKeyFields[i].desc := desc;
      PrimaryKeyFields[i].nocase := nocase;
      Result := True;
    end;
end;//ParseFieldPrimaryKey


//------------------------------------------------------------------------------
// parse fiald ... UNIQUE
//------------------------------------------------------------------------------
function TACRDDLTableManipulation.ParseFieldUnique(var FieldDef: TSQLFieldDef): Boolean;
begin
  Result := False;
  // UNIQUE
  if (Token.ReservedWord = rwUNIQUE) then
    begin
      FieldDef.newUnique := True;
      FieldDef.Unique := True;
      UniqueFields.Add(FieldDef.FieldName);
      Result := True;
      GetNextToken;
    end;
end;//ParseFieldUnique


//------------------------------------------------------------------------------
// parse Primary Key
//------------------------------------------------------------------------------
function TACRDDLTableManipulation.ParsePrimaryKey: boolean;
var
  desc, nocase: Boolean;
  FieldName: String;
  i: Integer;
begin
  // PRIMARY [KEY] [indexname] (column1 [asc|desc] [case|nocase] [, column2...])
  Result := false;
  // PRIMARY
  if (Token.ReservedWord = rwPRIMARY) then
    begin
      GetNextToken;
      // KEY
      if (Token.ReservedWord = rwKEY) then
      GetNextToken;

      // indexname
      if (Token.TokenType <> tktLeftParenthesis) then
        begin
          PrimaryIndexName := Token.Text;
          GetNextToken;
        end;

      // '('
      if (Token.TokenType <> tktLeftParenthesis) then
        raise EACRException.Create(30327, ErrorGLeftParenthesisExpected,
                                 [Token.Text, Token.LineNum, Token.ColumnNum]);

      repeat
        GetNextToken;
        // FieldName
        if Token.TokenType in [tktString, tktBracketedString, tktReservedWord] then
          FieldName := Token.Text
        else
          raise EACRException.Create(30244, ErrorGFieldNameExpected,
                                  [Token.Text, Token.LineNum, Token.ColumnNum]);

        GetNextToken;

        desc := false;
        nocase := false;
        // ASC, CASE ?
        repeat
            case Token.ReservedWord of
                rwASC:    desc:=false;
                rwDESC:   desc:=true;
                rwCASE:   nocase:=false;
                rwNOCASE: nocase:=true;
            else
              break;
            end;
            GetNextToken([], 30333, ErrorGUnexpectedEndOfCommand);
        until false;


        // add field to primary index
        i := Length(PrimaryKeyFields);
        SetLength(PrimaryKeyFields, i + 1);
        PrimaryKeyFields[i].FieldName := FieldName;
        PrimaryKeyFields[i].desc := desc;
        PrimaryKeyFields[i].nocase := nocase;


        // ')' or ','
        if not (Token.TokenType in [tktComma, tktRightParenthesis]) then
          raise EACRException.Create(30245, ErrorGRightParenthesisOrCommaExpected,
                                  [Token.Text, Token.LineNum, Token.ColumnNum]);
      until (Token.TokenType = tktRightParenthesis);

      // get next token
      GetNextToken;
      Result := true;
    end;
end;//ParsePrimaryKey


//------------------------------------------------------------------------------
// parse Foreign Key
//------------------------------------------------------------------------------
function TACRDDLTableManipulation.ParseForeignKey: boolean;
var
    ForeignKeyDef: TACRForeignKeyDef;
    bUpdate:          Boolean;
    bDelete:          Boolean;
    bUpdateCompleted: Boolean;

 procedure ParseAction(Update: Boolean);
 begin
  if (IsReservedWord(Token,rwCASCADE)) then
   begin
    if (Update) then
     ForeignKeyDef.UpdateAction := fkaCascade
    else
     ForeignKeyDef.DeleteAction := fkaCascade;
   end
  else
  if (IsReservedWord(Token,rwSET)) then
   begin
    GetNextToken;
    if (IsReservedWord(Token,rwNULL)) then
     begin
      if (Update) then
       ForeignKeyDef.UpdateAction := fkaSetNull
      else
       ForeignKeyDef.DeleteAction := fkaSetNull;
     end
    else
    if (IsReservedWord(Token,rwDEFAULT)) then
     begin
      if (Update) then
       ForeignKeyDef.UpdateAction := fkaSetDefault
      else
       ForeignKeyDef.DeleteAction := fkaSetDefault;
     end
    else
     raise EACRException.Create(11474,ErrorGOtherTokenExpected,
      [GetReservedWord(rwNULL) + ' or '+GetReservedWord(rwDEFAULT),
       Token.Text,Token.LineNum,Token.ColumnNum]);
   end
  else
  if (IsReservedWord(Token,rwNO)) then
   begin
    GetNextToken;
    if (IsReservedWord(Token,rwACTION)) then
     begin
      if (Update) then
       ForeignKeyDef.UpdateAction := fkaNoAction
      else
       ForeignKeyDef.DeleteAction := fkaNoAction;
     end
    else
     raise EACRException.Create(11475,ErrorGOtherTokenExpected,
      [GetReservedWord(rwACTION) ,
       Token.Text,Token.LineNum,Token.ColumnNum]);
   end
  else
   raise EACRException.Create(11476,ErrorGOtherTokenExpected,
     [GetReservedWord(rwCASCADE) + ' or '+GetReservedWord(rwSET)+' or '+
      GetReservedWord(rwNO),
      Token.Text,Token.LineNum,Token.ColumnNum]);
  // get next token
  GetNextToken;
 end; // ParseAction


begin
  // FOREIGN KEY [keyname] (column1 [, column2...]) REFERENCES tablename
  // [MATCH FULL | MATCH PARTIAL]
  // [ON DELETE {CASCADE | SET NULL | SET DEFAULT | NO ACTION}]
  // [ON UPDATE {CASCADE | SET NULL | SET DEFAULT | NO ACTION}]
  Result := false;
  // FOREIGN
  if (IsReservedWord(Token,rwFOREIGN)) then
    begin
      GetNextToken;
      // KEY
      if (IsReservedWord(Token,rwKEY)) then
       GetNextToken
      else
       raise EACRException.Create(11467,ErrorGOtherTokenExpected,
        [GetReservedWord(rwKEY),Token.Text,Token.LineNum,Token.ColumnNum]);
      if (FForeignKeyDefs = nil) then
       FForeignKeyDefs := TACRForeignKeyDefs.Create;
      ForeignKeyDef := TACRForeignKeyDefs(FForeignKeyDefs).AddForeignKeyDef;

      // foreign key name
      if (Token.TokenType <> tktLeftParenthesis) then
        begin
          ForeignKeyDef.Name := Token.Text;
          GetNextToken;
        end;

      // '('
      if (Token.TokenType <> tktLeftParenthesis) then
        raise EACRException.Create(11468, ErrorGLeftParenthesisExpected,
                                 [Token.Text, Token.LineNum, Token.ColumnNum]);

      repeat
        GetNextToken;
        // FieldName
        if Token.TokenType in [tktString, tktBracketedString] then
         begin
          if (ForeignKeyDef.Columns = '') then
           ForeignKeyDef.Columns := Token.Text
          else
           ForeignKeyDef.Columns := ForeignKeyDef.Columns + SemiColon + Token.Text;
         end
        else
          raise EACRException.Create(11469, ErrorGFieldNameExpected,
                                  [Token.Text, Token.LineNum, Token.ColumnNum]);
        GetNextToken;
        // ')' or ','
        if not (Token.TokenType in [tktComma, tktRightParenthesis]) then
          raise EACRException.Create(11470, ErrorGRightParenthesisOrCommaExpected,
                                  [Token.Text, Token.LineNum, Token.ColumnNum]);
      until (Token.TokenType = tktRightParenthesis);

      // get next token
      GetNextToken;

      if (IsReservedWord(Token,rwREFERENCES)) then
       begin
        GetNextToken;
        if Token.TokenType in [tktString, tktBracketedString] then
         begin
          ForeignKeyDef.ReferencedTableName := Token.Text;
          GetNextToken;
         end
        else
         raise EACRException.Create(11478, ErrorGTableNameExpected,
                         [Token.Text, Token.LineNum, Token.ColumnNum]);
       end
      else
       raise EACRException.Create(11477,ErrorGOtherTokenExpected,
         [GetReservedWord(rwREFERENCES),Token.Text,Token.LineNum,Token.ColumnNum]);

      if (IsReservedWord(Token,rwMATCH)) then
       begin
        GetNextToken;
        if (IsReservedWord(Token,rwFULL)) then
         ForeignKeyDef.MatchType := fkmtFull
        else
        if (IsReservedWord(Token,rwPARTIAL)) then
         ForeignKeyDef.MatchType := fkmtPartial
        else
         raise EACRException.Create(11471,ErrorGOtherTokenExpected,
          [GetReservedWord(rwFULL)+' or '+ GetReservedWord(rwPARTIAL),Token.Text,Token.LineNum,Token.ColumnNum]);
        GetNextToken;
       end;

      bDelete := False;
      bUpdate := False;
      if (IsReservedWord(Token,rwON)) then
       begin
        GetNextToken;
        if (IsReservedWord(Token,rwUPDATE)) then
         bUpdate := True
        else
        if (IsReservedWord(Token,rwDELETE)) then
         bDelete := True
        else
         raise EACRException.Create(11472,ErrorGOtherTokenExpected,
          [GetReservedWord(rwUPDATE) + ' or '+GetReservedWord(rwDELETE),
          Token.Text,Token.LineNum,Token.ColumnNum]);
        GetNextToken;
        ParseAction(bUpdate);
        bUpdateCompleted := bUpdate;
       end;

      if (IsReservedWord(Token,rwON)) then
       begin
        GetNextToken;
        if ((not bUpdate) and IsReservedWord(Token,rwUPDATE)) then
         bUpdate := True
        else
        if ((not bDelete) and IsReservedWord(Token,rwDELETE)) then
         bDelete := True
        else
         raise EACRException.Create(11473,ErrorGOtherTokenExpected,
          [GetReservedWord(rwUPDATE) + ' or '+GetReservedWord(rwDELETE),
          Token.Text,Token.LineNum,Token.ColumnNum]);
        GetNextToken;
        ParseAction(not bUpdateCompleted);
       end;
      Result := true;
    end; // Foreign Key
end;//ParseForeignKey


//------------------------------------------------------------------------------
// go to the Next token and Parse Integer (Int64)
//------------------------------------------------------------------------------
function TACRDDLTableManipulation.ParseInteger: Int64;
begin
  GetNextToken([tktInt], 30292, ErrorGIntegerExpected);
  try
    Result := StrToInt64(Token.Text);
  except
    raise EACRException.Create(30291, ErrorGIntegerExpected,
                               [Token.Text, Token.LineNum, Token.ColumnNum]);
  end;
  //GetNextToken;
end;//ParseInteger


//------------------------------------------------------------------------------
// Create ACRTable object and fill ACRTable params
//------------------------------------------------------------------------------
function TACRDDLTableManipulation.CreateInternalACRTable(query: TDataSet): TDataSet;
var
  i: Integer;
  t: TACRTable;
begin
  t := TACRTable.Create(query.Owner);
  Result := t;

  t.DatabaseName := TACRQuery(query).DatabaseName;
  t.SessionName := TACRQuery(query).SessionName;

  // Table Name
  t.TableName := Self.TableName;

  // Memory
  t.InMemory := IntoMemory or TACRQuery(query).InMemory;
  //if InMemory then DatabaseName := ACRMemoryDatabaseName;


  t.AdvFieldDefs.Clear;
  for i:=0 to SQLFieldDefs.Count-1 do
    FillAdvFieldDef(
                    TFieldDef( t.AdvFieldDefs.AddFieldDef ),
                    SQLFieldDefs[i]
                   );

  //FillPrimaryKey(Result);

end;//CreateInternalACRTable


//------------------------------------------------------------------------------
// Add Primary Key into ACRTable
//------------------------------------------------------------------------------
procedure TACRDDLTableManipulation.FillAdvFieldDef(
                                        AdvFieldDef: TFieldDef;
                                        SQLFieldDef: TSQLFieldDef
                                                  );
var
  fd: TACRAdvFieldDef;
begin
  fd := TACRAdvFieldDef(AdvFieldDef);

  fd.Name := SQLFieldDef.FieldName;

  // DataType
  if SQLFieldDef.newFieldType then
    fd.DataType := SQLFieldDef.FieldType;

  // Size
  if SQLFieldDef.newLength then
    fd.Size := SQLFieldDef.Length;

  // Required
  if SQLFieldDef.newRequired then
    fd.Required := SQLFieldDef.Required;

  // Autoinc settings
  if SQLFieldDef.newAutoincIncrement then
    fd.AutoincIncrement := SQLFieldDef.AutoincIncrement;
  if SQLFieldDef.newAutoincLastValue then
    fd.AutoincInitialValue := SQLFieldDef.AutoincLastValue;
  if SQLFieldDef.newAutoincMinValue then
    fd.AutoincMinValue  := SQLFieldDef.AutoincMinValue;
  if SQLFieldDef.newAutoincMaxValue then
    fd.AutoincMaxValue  := SQLFieldDef.AutoincMaxValue;
  if SQLFieldDef.newAutoincCycled then
    fd.AutoincCycled    := SQLFieldDef.AutoincCycled;

  // Blob settings
  if SQLFieldDef.newBlobBlockSize then
    fd.BlobBlockSize := SQLFieldDef.BlobBlockSize;
  if SQLFieldDef.newBlobCompressionAlgorithm then
    fd.BlobCompressionAlgorithm := TCompressionAlgorithm(SQLFieldDef.BlobCompressionAlgorithm);
  if SQLFieldDef.newBlobCompressionMode then
    fd.BlobCompressionMode := SQLFieldDef.BlobCompressionMode;

  // Default Value
  if SQLFieldDef.newDefaultValue then
    fd.DefaultValue.Assign(SQLFieldDef.DefaultValue);

  // MinValue
  if SQLFieldDef.newMinValue then
    fd.MinValue.Assign(SQLFieldDef.MinValue);
  // MaxValue
  if SQLFieldDef.newMaxValue then
    fd.MaxValue.Assign(SQLFieldDef.MaxValue);

end;//FillAdvFieldDef


//------------------------------------------------------------------------------
// Add Primary Key into ACRTable
//------------------------------------------------------------------------------
procedure TACRDDLTableManipulation.AddPrimaryKey(T: TDataSet);
var
  i:      Integer;
  s,d:    String;
  pkName: String;
begin
  if (Length(PrimaryKeyFields) <> 0) then
    begin
      d:='';
      s:='';
      pkName := AutoNameConstraintPKPreffix;
      for i:=0 to Length(PrimaryKeyFields) - 1 do
        begin
         pkName := pkName + AutoNameSymbol + PrimaryKeyFields[i].FieldName;
         s := s + d + PrimaryKeyFields[i].FieldName;
         d := ',';
        end;
      //IndexDefs.Clear;
      if (PrimaryIndexName <> '') then
       pkName := PrimaryIndexName
      else
       repeat
        pkName := GetTemporaryName(pkName);
       until (TACRTable(T).IndexDefs.IndexOf(pkName) < 0);
      TACRTable(T).IndexDefs.Add(pkName, s, [ixPrimary]);
      TACRTable(T).RestructureIndexDefs.Add(pkName, s, [ixPrimary]);
    end;
end;//AddPrimaryKey


//------------------------------------------------------------------------------
// Add UNIQUE constraint and index into ACRTable
//------------------------------------------------------------------------------
procedure TACRDDLTableManipulation.AddUnique(T: TDataSet);
var
  i:      Integer;
  uName: String;
begin
  for i:=0 to UniqueFields.Count-1 do
    begin
     repeat
      uName := GetTemporaryName(AutoNameConstraintUniquePreffix +
                                AutoNameSymbol +
                                UniqueFields[i]);
     until (TACRTable(T).IndexDefs.IndexOf(uName) < 0);
     TACRTable(T).IndexDefs.Add(uName, UniqueFields[i], [ixUnique]);
     TACRTable(T).RestructureIndexDefs.Add(uName, UniqueFields[i], [ixUnique]);
    end;
end;//AddUnique


//------------------------------------------------------------------------------
// Delete PrimaryKey
//------------------------------------------------------------------------------
procedure TACRDDLTableManipulation.DeletePrimaryKey(T: TDataSet);
var
  i: integer;
  AT: TACRTable;

  procedure DeleteIndexDef(idxDefs: TIndexDefs; index: integer);
  {$IFNDEF D5H}
  var
    i: integer;
    indDefs: TIndexDefs;
  {$ENDIF}

  begin
  {$IFDEF D5H}
    idxDefs.Delete(index);
  {$ELSE}
    indDefs := TIndexDefs.Create(AT);
    try
      for i:=0 to idxDefs.Count-1 do
        begin
          if i <> index then
            with indDefs[i] do
              indDefs.Add(idxDefs[i].Name, idxDefs[i].Fields, idxDefs[i].Options);
        end;
      idxDefs.assign(indDefs);
    finally
      indDefs.Free;
    end
  {$ENDIF}
  end;

begin
  AT := TACRTable(T);

  if AT.IndexDefs.Count <> 0 then
    begin

      for i:=0 to AT.IndexDefs.Count-1 do
        if ixPrimary in AT.IndexDefs[i].Options then
          begin
           DeleteIndexDef(AT.IndexDefs, i);
           break;
          end;

      for i:=0 to AT.RestructureIndexDefs.Count-1 do
        if ixPrimary in AT.RestructureIndexDefs[i].Options then
          begin
            DeleteIndexDef(AT.RestructureIndexDefs, i);
            break;
          end;

    end;//if

end;//DeletePrimaryKey


//------------------------------------------------------------------------------
// parses query
//------------------------------------------------------------------------------
procedure TACRDDLTableManipulation.Parse;
begin
  // check for not expected token
  if (LLex.LookNextToken(Token)) then
     raise EACRException.Create(30246, ErrorGUnexpectedToken,
                                [Token.Text, Token.LineNum, Token.ColumnNum]);
end;//Parse




////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLCreateTable
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create Table
//------------------------------------------------------------------------------
procedure TACRDDLCreateTable.CreateTable(query: TDataSet);
var
  T: TACRTable;
begin
  T := CreateInternalACRTable(query) as TACRTable;
  try
    // If table already exists - raise
    if T.Exists then
      raise EACRException.Create(30247, ErrorGTableAlreadyExists, [T.TableName]);

    try
      AddPrimaryKey(T);
      AddUnique(T);
      if (FForeignKeyDefs <> nil) then
       t.ForeignKeyDefs.Assign(TACRForeignKeyDefs(FForeignKeyDefs));
      T.CreateTable;
    except
      on e:Exception do
        raise EACRException.Create(30248, ErrorGErrorCreatingTable, [e.Message]);
    end;

  finally
    T.Free;
  end
end;//CreateTable


//------------------------------------------------------------------------------
// parses query
//------------------------------------------------------------------------------
procedure TACRDDLCreateTable.Parse;
begin
  if (not GetNextToken) then
    raise EACRException.Create(30249, ErrorGBlankSqlCommand);

  ParseTableNameToken;
  ParseTableElementList;

  inherited;
end;//Parse


//------------------------------------------------------------------------------
// executes query
//------------------------------------------------------------------------------
procedure TACRDDLCreateTable.ExecSQL(query: TDataset; IsRoot, RequestLive: boolean; var ReadOnly: boolean);
begin
  CreateTable(query);
end;//ExecSQL



////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLDropTable
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Drop Table
//------------------------------------------------------------------------------
procedure TACRDDLDropTable.DropTable(query: TDataSet);
var
  T: TACRTable;
begin
  T := CreateInternalACRTable(query) as TACRTable;
  try

    try
      T.Exclusive := True;
      if (T.Exists) then
       T.DeleteTable(FCascade);
    except
      on e:Exception do
        raise EACRException.Create(30250, ErrorGDroppingTable, [e.Message]);
    end;
    
  finally
    T.Free;
  end
end;//DropTable


//------------------------------------------------------------------------------
// parse query
//------------------------------------------------------------------------------
procedure TACRDDLDropTable.Parse;
begin
  if (not GetNextToken) then
    raise EACRException.Create(30251,ErrorGBlankSQLCommand);
  ParseTableNameToken;
  FCascade := False;
  if (IsReservedWord(Token,rwCASCADE)) then
   begin
    FCascade := True;
    GetNextToken;
   end
  else
  if (IsReservedWord(Token,rwRESTRICT)) then
   GetNextToken;
  inherited;
end;//Parse


//------------------------------------------------------------------------------
// executes query
//------------------------------------------------------------------------------
procedure TACRDDLDropTable.ExecSQL(query: TDataset; IsRoot, RequestLive: boolean; var ReadOnly: boolean);
begin
  DropTable(query);
end;//ExecSQL


////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLAlterTable
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRDDLAlterTable.Create(Lexer: TACRLexer; Query: TDataSet);
begin
  NewPassword := '';
  NewEncrypted := false;
  DropColumnNamesList := TStringList.Create;
  RenameColumnOldNamesList := TStringList.Create;
  RenameColumnNewNamesList := TStringList.Create;
  NewTableName := '';
  inherited Create(Lexer, Query);
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRDDLAlterTable.Destroy;
begin
  DropColumnNamesList.Free;
  RenameColumnOldNamesList.Free;
  RenameColumnNewNamesList.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// parse AlterType token
//------------------------------------------------------------------------------
procedure TACRDDLAlterTable.ParseAlterTypeToken;
begin
  if Token.TokenType = tktReservedWord then
    case Token.ReservedWord of
     rwADD:
      begin
       // ADD [COLUMN] | ADD PRIMARY KEY | ADD FOREIGN KEY | ADD UNIQUE
       GetNextToken;
       if (IsReservedWord(Token,rwCOLUMN)) then
        begin
          AlterType := atAdd;
          GetNextToken;
        end
       else
       if (IsReservedWord(Token,rwPRIMARY)) then
        begin
          AlterType := atAddConstraintPrimaryKey;
        end
       else
       if (IsReservedWord(Token,rwFOREIGN)) then
        begin
          AlterType := atAddConstraintForeignKey;
        end
{
       else
       if (IsReservedWord(Token,rwUNIQUE)) then
        begin
          AlterType := atAddConstraintUnique;
          GetNextToken;
        end
}        
       else
        AlterType := atAdd; // Column
      end;
     rwDROP:
       begin
         GetNextToken;
         if (IsReservedWord(Token,rwCOLUMN)) then
          begin
           AlterType := atDrop;
           GetNextToken;
          end
         else
         if (IsReservedWord(Token,rwCONSTRAINT)) then
          begin
           AlterType := atDropConstraint;
           GetNextToken;
          end
         else
          AlterType := atDrop;
       end;
     // MODIFY <column_name> <column difinition>
     rwMODIFY:
       begin
         AlterType := atModify;
         GetNextToken;
       end;
     // ALTER [COLUMN] <column_name> <column difinition>
     rwALTER:
       begin
         AlterType := atModify;
         GetNextToken;
         if (IsReservedWord(Token,rwCOLUMN)) then
          GetNextToken;
       end;
     rwRENAME:
       begin
         // RENAME TO | RENAME [COLUMN] columnname TO
         GetNextToken;
         if IsReservedWord(Token, rwTO) then
          begin
           AlterType := atRenameTable;
           GetNextToken;
          end
         else
           begin
             AlterType := atRenameColumn;
             if IsReservedWord(Token, rwCOLUMN) then
               GetNextToken;
           end;
       end;
     else
       raise EACRException.Create(30252, ErrorGAddOrDropOrModifyKeywordExpected,
                                 [Token.Text, Token.LineNum, Token.ColumnNum]);
    end//case
  else
    raise EACRException.Create(30253, ErrorGAddOrDropOrModifyKeywordExpected,
                            [Token.Text, Token.LineNum, Token.ColumnNum]);
end;//ParseAlterTypeToken


//------------------------------------------------------------------------------
// parse NewPassword token
//------------------------------------------------------------------------------
procedure TACRDDLAlterTable.ParseNewPasswordToken;
begin
  GetNextToken([tktReservedWord], 30255, ErrorGQuotedPasswordDataExpected);
  if Token.ReservedWord <> rwPASSWORD then exit;
  GetNextToken([tktReservedWord], 30256, ErrorGQuotedPasswordDataExpected);
  NewPassword := Token.Text;
  NewEncrypted := true;
  GetNextToken([tktComma, tktRightParenthesis],
               30257, ErrorGRightParenthesisOrCommaExpected);
end;//ParseNewPasswordToken


//------------------------------------------------------------------------------
// Alter Table DropColumn
//------------------------------------------------------------------------------
procedure TACRDDLAlterTable.ParseRenameColumnsList;
begin
  // OldColumnName
  RenameColumnOldNamesList.Add(Token.Text);
  GetNextToken;

  // TO
  if not IsReservedWord(Token, rwTO) then
    raise EACRException.Create(30352, ErrorGOtherTokenExpected,
                          ['TO', Token.Text, Token.LineNum, Token.ColumnNum]);
  GetNextToken;

  // NewColumnName
  RenameColumnNewNamesList.Add(Token.Text);
  GetNextToken;
end;//ParseRenameColumnsList


//------------------------------------------------------------------------------
// parse DROP CONSTRAINT
//------------------------------------------------------------------------------
procedure TACRDDLAlterTable.ParseDropConstraint;
begin
  DropConstraintCascade := False;
  if (Token.TokenType in [tktString,tktBracketedString,tktQuotedString]) then
   begin
    DropConstraintName := Token.Text;
    if (GetNextToken) then
     begin
      if (IsReservedWord(Token,rwRESTRICT)) then
       GetNextToken
      else
      if (IsReservedWord(Token,rwCASCADE)) then
       begin
        DropConstraintCascade := True;
        GetNextToken;
       end
      else
       raise EACRException.Create(11496,ErrorGOtherTokenExpected,
         [GetReservedWord(rwCASCADE) + ' or '+GetReservedWord(rwRESTRICT),
          Token.Text,Token.LineNum,Token.ColumnNum]);
     end;
   end
  else
   raise EACRException.Create(11495, ErrorLConstraintNameExpected,
                                  [Token.Text, Token.LineNum, Token.ColumnNum]);
end; // ParseDropConstraint


//------------------------------------------------------------------------------
// Alter Table DropColumn
//------------------------------------------------------------------------------
procedure TACRDDLAlterTable.DropColumn(T: TDataSet);
var i: Integer;
begin
  for i:=0 to DropColumnNamesList.Count-1 do
    TACRTable(T).RestructureFieldDefs.DeleteFieldDef(DropColumnNamesList[i]);
end;//DropColumn


//------------------------------------------------------------------------------
// Alter Table AddColumn
//------------------------------------------------------------------------------
procedure TACRDDLAlterTable.AddColumn(T: TDataSet);
var
  i: Integer;
  fd: TACRAdvFieldDef;
begin
    for i:=0 to SQLFieldDefs.Count-1 do
      begin
        fd := TACRTable(T).RestructureFieldDefs.Find(SQLFieldDefs[i].FieldName);
        if fd <> nil then
          raise EACRException.Create(30351, ErrorGFieldAlreadyExists,
                                     [SQLFieldDefs[i].FieldName]);
        // Add Field
        fd := TACRTable(T).RestructureFieldDefs.AddFieldDef;

        // Fill Field
        FillAdvFieldDef(TFieldDef(fd), SQLFieldDefs[i]);

      end;
//  DeletePrimaryKey(T);
//  AddPrimaryKey(T);
end;//AddColumn


//------------------------------------------------------------------------------
// Alter Table Modify
//------------------------------------------------------------------------------
procedure TACRDDLAlterTable.Modify(T: TDataSet);
var
  i: Integer;
  fd: TACRAdvFieldDef;
begin
    for i:=0 to SQLFieldDefs.Count-1 do
      begin
        fd := TACRTable(T).RestructureFieldDefs.Find(SQLFieldDefs[i].FieldName);
        if fd = nil then
          raise EACRException.Create(30259, ErrorGCannotFindField,
                                     [SQLFieldDefs[i].FieldName]);

        // Apply Changes
        FillAdvFieldDef(TFieldDef(fd), SQLFieldDefs[i]);

      end;
//  DeletePrimaryKey(T);
//  AddPrimaryKey(T);
end;//Modify


//------------------------------------------------------------------------------
// Rename Column
//------------------------------------------------------------------------------
procedure TACRDDLAlterTable.RenameColumn(T: TDataSet);
var  i: Integer;
begin
  for i:=0 to  RenameColumnOldNamesList.Count-1 do
   TACRTable(T).RenameField(RenameColumnOldNamesList[i],
                            RenameColumnNewNamesList[i]);
end;//RenameColumn


//------------------------------------------------------------------------------
// parse query
//------------------------------------------------------------------------------
procedure TACRDDLAlterTable.Parse;
begin
  if (not GetNextToken) then
    raise EACRException.Create(30261, ErrorGBlankSqlCommand);

  // Parse Table Name
  ParseTableNameToken;

  // Parse Alter Table Type
  ParseAlterTypeToken;

  case AlterType of
   atAdd:
    ParseTableElementList;
   atModify:
    ParseTableElementList;
   atDrop:
    begin
     if (Token.TokenType = tktLeftParenthesis) then
      begin
       GetNextToken;
       ParseColumnList(DropColumnNamesList);
       GetNextToken;
      end
     else
      ParseColumnList(DropColumnNamesList);
    end;
   atRenameColumn:
    ParseRenameColumnsList;
   atRenameTable:
    NewTableName := Token.Text;
   atAddConstraintPrimaryKey:
    ParsePrimaryKey;
{
   atAddConstraintUnique:
    ParseTableElementList;
}
   atAddConstraintForeignKey:
    ParseForeignKey;
   atDropConstraint:
    ParseDropConstraint;
   else
    raise EACRException.Create(30350, ErrorGNotImplementedYet);
  end;
end;//Parse


//------------------------------------------------------------------------------
// executes query
//------------------------------------------------------------------------------
procedure TACRDDLAlterTable.ExecSQL(query: TDataset; IsRoot, RequestLive: boolean; var ReadOnly: boolean);
var
  T: TACRTable;
  //compr: TCompressionLevel;
  i:   Integer;
  Log: String;
begin
  T := TACRTable.Create(query.Owner);
  try
    try
      t.DatabaseName := TACRQuery(query).DatabaseName;
      t.SessionName := TACRQuery(query).SessionName;
      // Table Name
      t.TableName := Self.TableName;
      // Memory
      t.InMemory := IntoMemory or TACRDataset(query).InMemory;
      t.Exclusive := True;

      // get RestructureFieldDefs
      t.Open;
      t.Close;

      case AlterType of
        atDrop:
          DropColumn(T);
        atAdd:
         begin
          AddColumn(T);
          AddUnique(T);
         end;
        atModify:
          Modify(T);
        atRenameColumn:
          RenameColumn(T);
        atRenameTable:
          T.RenameTable(NewTableName);
        atAddConstraintPrimaryKey:
          AddPrimaryKey(T);
        atAddConstraintForeignKey:
         for i := 0 to TACRForeignKeyDefs(FForeignKeyDefs).Count-1 do
          T.AddForeignKey(TACRForeignKeyDefs(FForeignKeyDefs).Items[i]);
//          if (FForeignKeyDefs <> nil) then
//           T.RestructureForeignKeyDefs.Assign(TACRForeignKeyDefs(FForeignKeyDefs));
        atDropConstraint:
          T.DeleteConstraint(DropConstraintName,DropConstraintCascade);
      end;

      if (AlterType in [atDrop, atAdd, atModify, atAddConstraintPrimaryKey]) then
        // RestructureTable
       if not (t.RestructureTable(Log)) then
        raise EACRException.Create(30347 ,ErrorGRestructureTableError, [Log]);

    except
      on e:Exception do
        raise EACRException.Create(30265, ErrorGErrorAlteringTable,[e.Message]);
    end
  finally
    T.Free;
  end
end;//ExecSQL




////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLRenameTable
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// parse query
//------------------------------------------------------------------------------
procedure TACRDDLRenameTable.Parse;
begin
  if (not GetCurrentToken) then
    raise EACRException.Create(30354, ErrorGBlankSqlCommand);

  AlterType := atRenameTable;

  // Parse Table Name
  ParseTableNameToken;

  // TO
  if not IsReservedWord(Token, rwTO) then
    raise EACRException.Create(30355, ErrorGOtherTokenExpected,
                          ['TO', Token.Text, Token.LineNum, Token.ColumnNum]);
  GetNextToken;

  // New Table Name
  NewTableName := Token.Text;

end;//Parse



////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLCreateIndex
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRDDLCreateIndex.Create(Lexer: TACRLexer; Query: TDataSet);
begin
  Unique := False;
  IfNotExists := False;
  SetLength(ACRIndexFields, 0);
  inherited Create(Lexer, Query);
end;//Create


//------------------------------------------------------------------------------
// parse query
//------------------------------------------------------------------------------
procedure TACRDDLCreateIndex.Parse;
var TempToken: TToken;
begin
  if (not GetCurrentToken) then
    raise EACRException.Create(30266, ErrorGBlankSQLCommand);

  // UNIQUE
  if IsReservedWord(Token, rwUNIQUE) then
    begin
      Unique:=true;
      GetNextToken([tktReservedWord], 30267, ErrorGIndexKeywordExpected);
    end;

  // INDEX
  if not IsReservedWord(Token, rwINDEX) then
    raise EACRException.Create(30268, ErrorGIndexKeywordExpected,
             [Token.Text, Token.LineNum, Token.ColumnNum]);

  LLex.LookNextToken(TempToken);
  if (IsReservedWord(TempToken, rwIF)) then
    begin
     GetNextToken;
     GetNextToken([tktReservedWord], 11236, ErrorLSQLNotKeywordExpected);
     if (not IsReservedWord(Token, rwNOT)) then
      raise EACRException.Create(11237, ErrorLSQLNotKeywordExpected,
               [Token.Text, Token.LineNum, Token.ColumnNum]);
     GetNextToken([tktReservedWord], 11238, ErrorLSQLExistsKeywordExpected);
     if (not IsReservedWord(Token, rwEXISTS)) then
      raise EACRException.Create(11239, ErrorLSQLExistsKeywordExpected,
               [Token.Text, Token.LineNum, Token.ColumnNum]);
     IfNotExists := True;
    end;

  // Index_name
  GetNextToken([tktReservedWord,tktString,tktQuotedString,tktBracketedString],
               30269, ErrorGIndexNameExpected);
  IndexName := Token.Text;

  // ON
  GetNextToken([tktReservedWord], 30270, ErrorGOnKeywordExpected);
  if (not IsReservedWord(Token, rwON)) then
    raise EACRException.Create(30271, ErrorGOnKeywordExpected,
                 [Token.Text, Token.LineNum, Token.ColumnNum]);
                 
  // table_name
  GetNextToken([tktReservedWord,tktString,tktQuotedString,tktBracketedString],
               30272, ErrorGTableNameExpected);
  IntoMemory := False;
  if (IsReservedWord(Token,rwMEMORY)) then
   begin
    IntoMemory := True;
    GetNextToken([tktReservedWord,tktString,tktQuotedString,tktBracketedString],
               11301, ErrorGTableNameExpected);
   end;
  // table_name
  TableName := Token.Text;
  
  // '('
  GetNextToken([tktLeftParenthesis], 30273, ErrorGLeftParenthesisExpected);
  // column list...
  repeat
    GetNextToken([tktReservedWord,tktString,tktQuotedString,tktBracketedString],
                 30274, ErrorGFieldNameExpected);
    // Password
    if IsReservedWord(Token, rwPassword) then
      begin
        GetNextToken([tktQuotedString], 30275, ErrorGQuotedPasswordDataExpected);
        Password := Token.Text;
        GetNextToken([tktComma, tktRightParenthesis],
                     30276, ErrorGRightParenthesisOrCommaExpected);
      end
    else
      begin
        SetLength(ACRIndexFields, Length(ACRIndexFields)+1);
        with ACRIndexFields[Length(ACRIndexFields)-1] do
          begin
            // Column name
            FieldName := Token.Text;
            // default ASC
            desc := false;
            // default CASE
            nocase := false;
            repeat
              GetNextToken([], 30277, ErrorGUnexpectedEndOfCommand);
              case Token.ReservedWord of
                rwASC:    desc:=false;
                rwDESC:   desc:=true;
                rwCASE:   nocase:=false;
                rwNOCASE: nocase:=true;
              end;
            until Token.TokenType in [tktComma, tktRightParenthesis];
          end;
      end;
  until Token.TokenType = tktRightParenthesis;

//  // check for not expected token
//  if (LLex.LookNextToken(Token)) then
//    raise ETblException.Create(02050,
//                [Token.Text, Token.LineNum, Token.ColumnNum]);
end;//Parse


//------------------------------------------------------------------------------
// execute query
//------------------------------------------------------------------------------
procedure TACRDDLCreateIndex.ExecSQL(query: TDataset; IsRoot, RequestLive: boolean; var ReadOnly: boolean);
var
   T:                                 TACRTable;
   Fields, DescFields, CaseInsFields: String;
   delim1,delim2,delim3:              String;
   i:                                 Integer;
   Options:                           TIndexOptions;
   bCancel:                           Boolean;
begin
   Fields:=''; DescFields:=''; CaseInsFields:='';
   delim1:=''; delim2:=''; delim3:='';
   for i:=0 to Length(ACRIndexFields)-1 do
     begin
       Fields := Fields + delim1 + ACRIndexFields[i].FieldName;
       delim1 := ';';
       if ACRIndexFields[i].desc then
         begin
           DescFields := DescFields + delim2 + ACRIndexFields[i].FieldName;
           delim2 := ';';
         end;
       if ACRIndexFields[i].nocase then
         begin
           CaseInsFields := CaseInsFields + delim3 + ACRIndexFields[i].FieldName;
           delim3 := ';';
         end;
     end;
   // Index Options
   if Unique then
     Options:=[ixUnique]
   else
     Options:=[];


   // Creating
   T := TACRTable.Create(query.Owner);
   try
     T.DatabaseName := (query as TACRQuery).DatabaseName;
     T.SessionName := (query as TACRQuery).SessionName;
     T.TableName := TableName;
     T.InMemory := IntoMemory;
     T.Exclusive := True;
     //T.Password := Password;
     T.Open;
     bCancel := IfNotExists and (T.IndexDefs.IndexOf(IndexName) >= 0);
     if (not bCancel) then
       T.AddIndex(IndexName, Fields, Options, DescFields, CaseInsFields);
     T.Close;
   finally
     T.Free;
   end;
end;//ExecSQL




////////////////////////////////////////////////////////////////////////////////
//
// TACRDDLDropIndex
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRDDLDropIndex.Create(Lexer: TACRLexer; Query: TDataSet);
begin
  Password := '';
  IfExists := False;
  inherited Create(Lexer, Query);
end;//Create


//------------------------------------------------------------------------------
// parse query
//------------------------------------------------------------------------------
procedure TACRDDLDropIndex.Parse;
begin
  if (not GetNextToken) then
    raise EACRException.Create(30278, ErrorGUnexpectedEndOfCommand,
                               [Token.LineNum, Token.ColumnNum]);

  if (IsReservedWord(Token,rwIF)) then
   begin
    GetNextToken([tktReservedWord], 11241, ErrorLSQLExistsKeywordExpected);
    if (not IsReservedWord(Token, rwEXISTS)) then
      raise EACRException.Create(11242, ErrorLSQLExistsKeywordExpected,
               [Token.Text, Token.LineNum, Token.ColumnNum]);
    IfExists := True;
    if (not GetNextToken) then
      raise EACRException.Create(11243, ErrorGUnexpectedEndOfCommand,
                                 [Token.LineNum, Token.ColumnNum]);
   end;

  IntoMemory := False;
  if (IsReservedWord(Token,rwMEMORY)) then
   begin
    IntoMemory := True;
    GetNextToken([tktReservedWord,tktString,tktQuotedString,tktBracketedString],
               11302, ErrorGTableNameExpected);
   end;
  // table_name
  TableName:=Token.Text;
  if (not GetNextToken) then
    raise EACRException.Create(30279, ErrorGUnexpectedEndOfCommand,
                               [Token.LineNum, Token.ColumnNum]);

  // '.'
  if (Token.TokenType <> tktDot) then
    raise EACRException.Create(30280, ErrorGOtherTokenExpected,
                             ['.', Token.Text, Token.LineNum, Token.ColumnNum]);

  // index_name
  if (not GetNextToken) then
    raise EACRException.Create(30281, ErrorGUnexpectedEndOfCommand,
                               [Token.LineNum, Token.ColumnNum]);
  IndexName:=Token.Text;

  // password
  if GetNextToken then
    if IsReservedWord(Token, rwPassword) then
      begin
        GetNextToken([tktQuotedString], 30282, ErrorGQuotedPasswordDataExpected);
        Password := Token.Text;
      end;

  // check for not expected token
//  if (FLex.LookNextToken(Token)) then
//    raise ETblException.Create(02061,
//                [Token.Text, Token.LineNum, Token.ColumnNum]);
end;//Parse


//------------------------------------------------------------------------------
// execute query
//------------------------------------------------------------------------------
procedure TACRDDLDropIndex.ExecSQL(query: TDataset; IsRoot, RequestLive: boolean; var ReadOnly: boolean);
var
   T:       TACRTable;
   bCancel: Boolean;
begin
  T := TACRTable.Create(query.Owner);
  try
    T.DatabaseName := (query as TACRQuery).DatabaseName;
    T.SessionName := (query as TACRQuery).SessionName;
    T.TableName := TableName;
    T.InMemory := IntoMemory;
    T.Exclusive := True;
    T.Open;
    bCancel := IfExists and (T.IndexDefs.IndexOf(IndexName) < 0);
    if (not bCancel) then
      T.DeleteIndex(IndexName);
    T.Close;
  finally
    T.Free;
  end;
end;//ExecSQL


////////////////////////////////////////////////////////////////////////////////
//
// TACRStartTransaction
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// parse query
//------------------------------------------------------------------------------
procedure TACRStartTransaction.Parse;
begin
  if (not GetNextToken) then
    raise EACRException.Create(10845, ErrorGUnexpectedEndOfCommand,
                               [Token.LineNum, Token.ColumnNum]);
  if (not IsReservedWord(Token, rwTransaction)) then
    raise EACRException.Create(10846, ErrorGSQLCommandExpected,
                               [Token.Text,Token.LineNum, Token.ColumnNum]);
end; // Parse


//------------------------------------------------------------------------------
// execute query
//------------------------------------------------------------------------------
procedure TACRStartTransaction.ExecSQL(
                  query: TDataset;
                  IsRoot,
                  RequestLive: boolean;
                  var ReadOnly: boolean
                 );
begin
{$IFNDEF SQLMEMTABLE}
  TACRDataSet(query).Database.StartTransaction;
{$ENDIF}
end; // ExecSQL




////////////////////////////////////////////////////////////////////////////////
//
// TACRCommit
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// parse query
//------------------------------------------------------------------------------
procedure TACRCommit.Parse;
begin
  FFLushFileBuffers := True;
  if (GetNextToken) then
   if (IsReservedWord(Token, rwNOFLUSH)) then
    FFlushFileBuffers := False;
end; // Parse


//------------------------------------------------------------------------------
// execute query
//------------------------------------------------------------------------------
procedure TACRCommit.ExecSQL(
                  query: TDataset;
                  IsRoot,
                  RequestLive: boolean;
                  var ReadOnly: boolean
                 );
begin
{$IFNDEF SQLMEMTABLE}
  TACRDataSet(query).Database.Commit(FFlushFileBuffers);
{$ENDIF}
end; // ExecSQL


////////////////////////////////////////////////////////////////////////////////
//
// TACRRollback
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// execute query
//------------------------------------------------------------------------------
procedure TACRRollback.ExecSQL(
                  query: TDataset;
                  IsRoot,
                  RequestLive: boolean;
                  var ReadOnly: boolean
                 );
begin
{$IFNDEF SQLMEMTABLE}
  TACRDataSet(query).Database.Rollback;
{$ENDIF}
end; // ExecSQL


//------------------------------------------------------------------------------
// set field value from ACRVariant
//------------------------------------------------------------------------------
procedure SetACRVariantIntoField(Value: TACRVariant; Field: TField);
var
  s:  String;
  bs: TStream;
begin
  if Value = nil then
    raise EACRException.Create(30359, ErrorGValueIsNull);

  if (Value.IsNull) then
    Field.Clear
  else
    case Field.DataType of
      ftBlob,
      ftGraphic:
        begin
          bs := Field.DataSet.CreateBlobStream(Field,bmWrite);
          try
            bs.WriteBuffer(Value.PData^,Value.DataSize);
          finally
            bs.Free;
          end;
        end;
      ftMemo,
      ftFmtMemo:
        begin
          if (Value.DataSize <= 1) then
           Field.Clear
          else
           begin
            SetLength(s, Value.DataSize-1);
            Value.CopyDataToAddress(Pchar(s));
            Field.AsString := s;
           end;
        end;
      ftLargeint:
        begin
          TLargeintField(Field).Value := Value.AsInt64;
        end;
      ftDate:
        begin
         Field.AsDateTime := Value.AsTDate;
        end;
      ftTime:
        begin
         Field.AsDateTime := Value.AsTTime;
        end;
      ftDateTime:
        begin
         Field.AsDateTime := Value.AsTDateTime;
        end;
{$IFDEF D6H}
      ftTimeStamp:
        begin
          TSQLTimeStampField(Field).AsDateTime := Value.AsTDateTime;
        end;
{$ENDIF}
      else
        Field.Value := Value.AsVariant;
    end;
end;//SetACRVariantIntoField


initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRSQLProcessor> initialized');
{$ENDIF}

end.
