unit ACRMemEngine;

{$I ACRVer.inc}

interface

uses SysUtils, Classes, Math,
{$IFDEF MSWINDOWS}
     Windows,
{$ENDIF}
{$IFDEF LINUX}
     Types,
     Libc,
{$ENDIF}

// Accuracer units

     {$IFDEF DEBUG_LOG}
     ACRDebug,
     {$ENDIF}
     ACRCriticalSection,
     ACRExcept,
     ACRBase,
     ACRBaseEngine,
     ACRBTree,
     ACRPage,
     ACRMemory,
     ACRCompression,
     ACRTypes,
     ACRExpressions,
     ACRConverts,
     ACRVariant,
     ACRConst;

type

////////////////////////////////////////////////////////////////////////////////
//
// TACRMemoryDatabaseData
//
////////////////////////////////////////////////////////////////////////////////


  TACRMemoryDatabaseData = class (TACRDatabaseData)
   public
    constructor Create;
    destructor Destroy; override;
    // create table data
    function CreateTableData(Cursor: TACRCursor): TACRTableData; override;
    procedure ConnectSession(Session: TACRBaseSession); override;
  end; // TACRMemoryDatabaseData


////////////////////////////////////////////////////////////////////////////////
//
// TACRMemoryRecordManager
//
////////////////////////////////////////////////////////////////////////////////

  TACRLoadingRecordsInfo = record
    pFlag:    PByte;    // pointer to last byte in FDeletedFlagBuffer
    BitNo:    Integer;  // (0 - 7) number of last bit found in pFlag
    ByteNo:   Integer;  // (0 - n) number of byte in PFlag relative to the beginning of FDeletedFlagBuffer
  end;

  TACRMemoryRecordManager = class (TACRBaseRecordManager)
   private
    LTableData:             TACRTableData;
    // record buffers
    FRecordBuffer:          PChar;
    // sequence of bits: 1 - for deleted record, 0 - for existing record
    FDeleteFlagBuffer:      PChar;
    FAllocatedRecordCount:  Integer;
    FAllocRecordsBy:        Integer;
    FFirstRecordID:         Integer;
    FLastRecordID:          Integer;
    FBitsTable:             array [0..255] of byte;
    FNoRecordsDeleted:      ByteBool;
    FLoadingRecordInfo:     TACRLoadingRecordsInfo;

   protected
    function GetBitmapSize: TACRRecordNo;
   private 
    // find record id by record position
    function FindRecord(RecordNo: TACRRecordNo): TACRRecordNo;
    // returns position of the record specified by record id,
    // or INVALID_ID8 if the record with this id does not exists
    function GetTablePositionByRecordID(
              RecordNo:           TACRRecordNo
                                        ): TACRRecordNo;
    // return result for attempt of getting record relatively to first position
    // and set RecordID to new record ID
    function GetRecordFromFirstPosition(
            GetRecordMode: TACRGetRecordMode;
            var RecordID: Integer
                                       ): TACRGetRecordResult;
    // return result for attempt of getting record relatively to last position
    // and set RecordID to new record ID
    function GetRecordFromLastPosition(
            GetRecordMode: TACRGetRecordMode;
            var RecordID: Integer
                                      ): TACRGetRecordResult;
    // return result for attempt of getting record relatively any position
    // and set RecordID to new record ID
    function GetRecordFromAnyPosition(
            GetRecordMode: TACRGetRecordMode;
            var RecordID: Integer
                                      ): TACRGetRecordResult;
   protected
    procedure SetRecordCount(NewRecordCount: Integer);
   public
    constructor Create(
                        aTableData:           TACRTableData;
                        RecordBufferSize:     Integer;
                        AllocRecordsBy:       Integer
                      );
    destructor Destroy; override;

    procedure Empty(SessionID: TACRSessionID = INVALID_SESSION_ID); override;

    procedure RewriteBLOBValues(SourceRecordBuffer, DestRecordBuffer: TACRRecordBuffer);

    // add record and return its number
    function AddRecord(
                       RecordBuffer:  TACRRecordBuffer;
                       var RecordID:  TACRRecordID;
                       SessionID:     TACRSessionID = INVALID_SESSION_ID
                      ): Boolean; override;
    // update record, return true if record was updated, false if record was deleted
    function UpdateRecord(
                          RecordBuffer: TACRRecordBuffer;
                          RecordID:     TACRRecordID;
                          SessionID:    TACRSessionID = INVALID_SESSION_ID
                         ): Boolean; override;
    // delete record, return true if record was deleted, false if record was deleted earlier
    function DeleteRecord(
                          var RecordID: TACRRecordID;
                          SessionID:    TACRSessionID = INVALID_SESSION_ID
                         ): Boolean; override;
    // return true if record exists
    function IsRecordExists(var RecordID: TACRRecordID; SessionID: TACRSessionID = INVALID_SESSION_ID): Boolean; override;
    procedure GetRecordBuffer(var NavigationInfo: TACRNavigationInfo); override;
    // return record no
    function GetApproximateRecNo(RecordID: TACRRecordID; SessionID: TACRSessionID): TACRRecordNo; override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
    // add loaded record
    procedure AddLoadedRecord(RecordBuffer: TACRRecordBuffer; var RecordPos: Integer); override;
  end; // TACRMemoryRecordManager


////////////////////////////////////////////////////////////////////////////////
//
// TACRMemoryTableData
//
////////////////////////////////////////////////////////////////////////////////


  TACRMemoryTableData = class (TACRTableData)
   private
    FLoadedRecordCount: Integer;
    FAllocRecordsBy:    Integer;
    FPageManager:       TACRPageManager;

   protected
    function GetPageManager: TACRPageManager; override;
    procedure CreateRecordManager(TablePFSMapFileStartPageNo: TACRPageNo = INVALID_PAGE_NO); override;
    procedure CreateIndexManager(IndexDefs: TACRIndexDefs); override;
    procedure LoadTableHeader(
              Stream:             TStream;
              var BLOBDescriptor: TACRBLOBDescriptor
                              );
    procedure SaveTableHeader(
              Stream:             TStream;
              var BLOBDescriptor: TACRBLOBDescriptor
                              );
    procedure LoadBLOBDataFromStream(
            RecordBuffer: TACRRecordBuffer;
            FieldNo:      Integer;
            Stream:       TStream
                                 );
    procedure LoadRecordFromStream(
            RecordBuffer: TACRRecordBuffer;
            Stream:       TStream
                                 );
    procedure PrepareRecordBufferForSave(
            RecordBuffer:             TACRRecordBuffer;
            BLOBDescriptorList:       TList;
            BLOBDataList:             TList;
            BLOBPosition:             Int64
                                                );
    procedure SaveRecordToStream(
            RecordBuffer: TACRRecordBuffer;
            Stream:       TStream
                                 );
    procedure LoadRecordManager(Stream: TStream);
    procedure SaveRecordManager(Stream: TStream);
    function GetBitmapSize(SessionID: TACRSessionID): TACRRecordNo; override;
    // return filter bitmap rec no by record id
    function GetBitmapRecNoByRecordID(RecordID: TACRRecordID): TACRRecordNo; override;
    // return filter bitmap rec no by record id
    function GetRecordIDByBitmapRecNo(RecordNo: TACRRecordNo): TACRRecordID; override;
    procedure InternalEmptyTable;
   public
    constructor Create(
                        aDatabaseData: TACRDatabaseData;
                        AllocRecordsBy: Integer
                      );
    destructor Destroy; override;

    procedure CreateTable(
                          Cursor: TACRCursor;
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                         ); override;
    procedure DeleteTable(Session: TACRBaseSession; Cascade: Boolean; DesignMode: Boolean = False); override;
    procedure DeleteConstraint(Cursor: TACRCursor; Name: String; Cascade: Boolean; FKPartialDelete: Boolean); override;
    procedure EmptyTable(Cursor: TACRCursor); override;
    procedure RenameTable(Cursor: TACRCursor; NewTableName: String); override;
    procedure LoadTableFromStream(
                        Stream: TStream
                       ); override;
    procedure SaveTableToStream(
                        Stream: TStream;
                        CompressionAlgorithm: TACRCompressionAlgorithm = acaNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      ); override;
    procedure OpenTable(Cursor: TACRCursor); override;
    procedure CloseTable(Cursor: TACRCursor); override;
    procedure AddIndex(IndexDef: TACRIndexDef; Cursor: TACRCursor); override;

    // return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
    function CompareRecordID(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer; override;

    function InsertRecord(var Cursor: TACRCursor): Boolean; override;
    function DeleteRecord(Cursor: TACRCursor): Boolean; override;
    function UpdateRecord(Cursor: TACRCursor): Boolean; override;
    procedure DeleteVisibleRecords(Cursor: TACRCursor); override;
    procedure UpdateVisibleRecords(
                                    Cursor:       TACRCursor;
                                    FieldNames:   TStringList;
                                    values:       array of TACRVariant;
                                    SkipFKCheck:  Boolean = False
                                  ); override;
    function InternalGetRecordCount: TACRRecordNo; override;
    // move cursor to specified position and set current record id in cursor
    procedure InternalSetRecNo(Cursor: TACRCursor; RecNo: TACRRecordNo); override;
    // get current record position from cursor
    function InternalGetRecNo(Cursor: TACRCursor): TACRRecordNo; override;

    function LastAutoincValue(FieldNo: Integer; Session: TACRBaseSession): Int64; override;
    procedure SetLastAutoincValue(Value: Int64; FieldNo: Integer; Cursor: TACRCursor); override;
    procedure LockTableForRead(aSessionID: TACRSessionID); override;
    procedure UnlockTableForRead(aSessionID: TACRSessionID); override;

   public
    property SequenceManager: TACRBaseSequenceManager read FSequenceManager write FSequenceManager;
  end; // TACRMemoryTableData



implementation

uses

// Accuracer units

  ACRLocalEngine;

////////////////////////////////////////////////////////////////////////////////
//
// TACRMemoryDatabaseData
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRMemoryDatabaseData.Create;
begin
 inherited Create;
 FPageManager := nil;
 DatabaseName := ACRMemoryDatabaseName;
end;// Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRMemoryDatabaseData.Destroy;
begin
 while (FTableDataList.Count > 0) do
  begin
    if (FTableDataList.Items[0] = nil) then
     raise EACRException.Create(10258,ErrorLNilPointer);
    TACRTableData(FTableDataList.Items[0]).DeleteTable(nil,True,False);
  end;
 inherited Destroy;
end;// Destroy


//------------------------------------------------------------------------------
// create table data
//------------------------------------------------------------------------------
function TACRMemoryDatabaseData.CreateTableData(Cursor: TACRCursor): TACRTableData;
begin
 Result := TACRMemoryTableData.Create(Self,Cursor.MemoryTableAllocBy);
 Result.TableName := Cursor.TableName;
 FTableDataList.Add(Result);
end;


//------------------------------------------------------------------------------
// ConnectSession
//------------------------------------------------------------------------------
procedure TACRMemoryDatabaseData.ConnectSession(Session: TACRBaseSession);
begin
  Session.SessionID := 0;
  inherited ConnectSession(Session);
end;// ConnectSession




////////////////////////////////////////////////////////////////////////////////
//
// TACRMemoryRecordManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// return bitmap size
//------------------------------------------------------------------------------
function TACRMemoryRecordManager.GetBitmapSize: TACRRecordNo;
begin
//  Result := FAllocatedRecordCount;
  Result := FLastRecordID + 1;
end; // GetBitmapSize


//------------------------------------------------------------------------------
// find record position in buffer by record number
//------------------------------------------------------------------------------
function TACRMemoryRecordManager.FindRecord(RecordNo: TACRRecordNo): TACRRecordNo;
var i,l:        Integer;
    b,k:        Byte;
    n:          TACRRecordNo;
begin
 Result := INVALID_ID8;
 if (RecordNo < 0) or (RecordNo > FRecordCount) then
  Exit;

 if (FNoRecordsDeleted) then
  begin
   Result := RecordNo;
   Exit;
  end;

 i := 0; // byte number
 n := 0; // bits count
 while (n <= RecordNo) do
  begin
   b := PByte(FDeleteFlagBuffer + i)^;
   if (n + FBitsTable[b] > RecordNo) then
    break;
   Inc(n,FBitsTable[b]);
   Inc(i);
  end;
 Result := i * 8;
 if (Result > FAllocatedRecordCount) then
  raise EACRException.Create(10457,ErrorLInvalidBitNo,[RecordNo,FAllocatedRecordCount]);
 b := PByte(FDeleteFlagBuffer + i)^;
 l := 7;
 if (i = (FAllocatedRecordCount div 8)) then
  l := (FAllocatedRecordCount mod 8)-1;
 for k := 0 to l do
  begin
   if ((b and (1 shl k)) = 0) then Inc(n);
   if (n > RecordNo) then Break;
   Inc(Result);
  end;
end;


//------------------------------------------------------------------------------
// returns position of the record specified by record id,
// or INVALID_ID8 if the record with this id does not exists
//------------------------------------------------------------------------------
function TACRMemoryRecordManager.GetTablePositionByRecordID(
              RecordNo:            TACRRecordNo
                ): TACRRecordNo;
var n,i,j:    Integer;
    b,k:      Byte;
begin
 Result := INVALID_ID8;
 if (RecordNo >= FAllocatedRecordCount) then
  Exit;
 if (FNoRecordsDeleted) then
  begin
   Result := RecordNo+1;
   Exit;
  end;
 // number of byte with flags
 i := RecordNo div 8;
 Result := 0; // bits count
 if (i > 0) then
   for j := 0 to i-1 do
     for k := 7 downto 0 do
      begin
       b := PByte(FDeleteFlagBuffer + j)^;
       if ((b and (1 shl k)) = 0) then Inc(Result);
      end;
 // scan last byte
 b := PByte(FDeleteFlagBuffer + i)^;
 n := RecordNo mod 8;
 for k := 0 to n do
   if ((b and (1 shl k)) = 0) then Inc(Result);
end;


//------------------------------------------------------------------------------
// return result for attempt of getting record relatively to first position
//------------------------------------------------------------------------------
function TACRMemoryRecordManager.GetRecordFromFirstPosition(
            GetRecordMode: TACRGetRecordMode;
            var RecordID: Integer
                                                           ): TACRGetRecordResult;
begin
 Result := grrError;
 case GetRecordMode of
  grmPrior:
   begin
    Result := grrBOF;
   end;
  grmCurrent:
   begin
    Result := grrError;
   end;
  grmNext:
   begin
    RecordID := FFirstRecordID;
    Result := grrOK;
   end;
 end; // GetRecordMode
end; // GetRecordFromFirstPosition


//------------------------------------------------------------------------------
// return result for attempt of getting record relatively to last position
//------------------------------------------------------------------------------
function TACRMemoryRecordManager.GetRecordFromLastPosition(
            GetRecordMode: TACRGetRecordMode;
            var RecordID: Integer
                                                          ): TACRGetRecordResult;
begin
 Result := grrError;
 case GetRecordMode of
  grmPrior:
   begin
    RecordID := FLastRecordID;
    Result := grrOK;
   end;
  grmCurrent:
   begin
    Result := grrError;
   end;
  grmNext:
   begin
    Result := grrEOF;
   end;
 end; // GetRecordMode
end; // GetRecordFromLastPosition


//------------------------------------------------------------------------------
// return result for attempt of getting record relatively any position
// and set RecordID to new record ID
//------------------------------------------------------------------------------
function TACRMemoryRecordManager.GetRecordFromAnyPosition(
            GetRecordMode: TACRGetRecordMode;
            var RecordID: Integer
                                      ): TACRGetRecordResult;
begin
 Result := grrError;
 case GetRecordMode of
  grmPrior:
   begin
    Result := grrBOF;
    if (RecordID > 0) then
     begin
      Dec(RecordID);
      while (RecordID >= FFirstRecordID) do
       begin
        // record found
        if (not CheckNullFlag(RecordID,FDeleteFlagBuffer)) then
         begin
          Result := grrOK;
          break;
         end;
        Dec(RecordID);
       end; // scan prior record
     end; // RecordID > 0
   end;
  grmCurrent:
   begin
    if ((RecordID > FLastRecordID) or (RecordID < FFirstRecordID)) then
     // current record does not exist
     Result := grrError
    else
    if (CheckNullFlag(RecordID,FDeleteFlagBuffer)) then
     // current record is deleted
     Result := grrError
    else
     Result := grrOK;
   end;
  grmNext:
   begin
    Result := grrEOF;
    if (RecordID < FAllocatedRecordCount-1) then
     begin
      Inc(RecordID);
      while ((RecordID <= FLastRecordID) and (RecordID < FAllocatedRecordCount)) do
       begin
        // record found
        if (not CheckNullFlag(RecordID,FDeleteFlagBuffer)) then
         begin
          Result := grrOK;
          break;
         end;
        Inc(RecordID);
       end; // scan prior record
     end;
   end;
 end; // GetRecordMode
end; // GetRecordFromAnyPosition


//------------------------------------------------------------------------------
// set record count
//------------------------------------------------------------------------------
procedure TACRMemoryRecordManager.SetRecordCount(NewRecordCount: Integer);
var i: Integer;
begin
 if (NewRecordCount = 0) then
  MemoryManager.FreeAndNilMem(FRecordBuffer)
 else
 if (FAllocatedRecordCount = 0) then
   FRecordBuffer := MemoryManager.GetMem(NewRecordCount * FRecordBufferSize)
 else
   MemoryManager.ReallocMem(FRecordBuffer,NewRecordCount * FRecordBufferSize,False);
 i := (NewRecordCount div 8) + Integer((NewRecordCount mod 8) > 0);
 if (i = 0) then
  MemoryManager.FreeAndNilMem(FDeleteFlagBuffer)
 else
 if (FAllocatedRecordCount = 0) then
  FDeleteFlagBuffer := MemoryManager.GetMem(i)
 else
  MemoryManager.ReallocMem(Pointer(FDeleteFlagBuffer),i);
 FAllocatedRecordCount := NewRecordCount;
end; // SetRecordCount


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRMemoryRecordManager.Create(
                        aTableData:           TACRTableData;
                        RecordBufferSize:     Integer;
                        AllocRecordsBy:       Integer
                      );
var
    i,c: Byte;
begin
 if (RecordBufferSize = 0) then
  raise EACRException.Create(10011,ErrorLInvalidRecordSize);
 InitializeCriticalSection(FCSect);
 LTableData := aTableData;
 FRecordBuffer := nil;
 FDeleteFlagBuffer := nil;
 FRecordCount := 0;
 FAllocatedRecordCount := 0;
 FFirstRecordID := 0;
 FLastRecordID := 0;
 FRecordBufferSize := RecordBufferSize;
 FAllocRecordsBy := AllocRecordsBy;
 FNoRecordsDeleted := True;
 for i := 0 to 255 do
  begin
   c := 0;
   if ((i and 1) = 0) then Inc(c);
   if ((i and 2) = 0) then Inc(c);
   if ((i and 4) = 0) then Inc(c);
   if ((i and 8) = 0) then Inc(c);
   if ((i and 16) = 0) then Inc(c);
   if ((i and 32) = 0) then Inc(c);
   if ((i and 64) = 0) then Inc(c);
   if ((i and 128) = 0) then Inc(c);
   FBitsTable[i] := c;
  end;
end; // Create;


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRMemoryRecordManager.Destroy;
begin
 Empty;
 DeleteCriticalSection(FCSect);
 inherited Destroy;
end; // Destroy


//------------------------------------------------------------------------------
// Empty
//------------------------------------------------------------------------------
procedure TACRMemoryRecordManager.Empty(SessionID: TACRSessionID);
begin
 FAllocatedRecordCount := 0;
 FRecordCount := 0;
 if (FRecordBuffer <> nil) then
   MemoryManager.FreeAndNilMem(FRecordBuffer);
 if (FDeleteFlagBuffer <> nil) then
   MemoryManager.FreeAndNilMem(FDeleteFlagBuffer);
 FFirstRecordID := 0;
 FLastRecordID := 0;
end; // Empty


//------------------------------------------------------------------------------
// rewrite BLOB values
//------------------------------------------------------------------------------
procedure TACRMemoryRecordManager.RewriteBLOBValues(
                            SourceRecordBuffer,
                            DestRecordBuffer:   TACRRecordBuffer);
var i: Integer;
begin
 if (LTableData.BLOBFieldsPresent) then
  for i := 0 to LTableData.FieldManager.FieldDefs.Count - 1 do
   if (IsBLOBFieldType(LTableData.FieldManager.FieldDefs[i].BaseFieldType)) then
    begin
      if (not CheckNullFlag(i,DestRecordBuffer)) then
       begin
         if (LTableData.IsBLOBModified(SourceRecordBuffer,
              LTableData.FieldManager.FieldDefs[i])) or
            (CheckNullFlag(i,SourceRecordBuffer)) then
          begin
           LTableData.ClearBLOBFieldInRecordBuffer(DestRecordBuffer,i);
           LTableData.SetBLOBModified(False,SourceRecordBuffer,
            LTableData.FieldManager.FieldDefs[i]);
          end;
       end // old value is not null
      else
       begin
           LTableData.SetBLOBModified(False,SourceRecordBuffer,
            LTableData.FieldManager.FieldDefs[i]);
       end; // old value is null
    end;
end; // RewriteBLOBValues


//------------------------------------------------------------------------------
// add record and return its number
//------------------------------------------------------------------------------
function TACRMemoryRecordManager.AddRecord(
                       RecordBuffer:  TACRRecordBuffer;
                       var RecordID:  TACRRecordID;
                       SessionID:     TACRSessionID = INVALID_SESSION_ID
                      ): Boolean;
var OldRecordCount, i: Integer;
begin
 try
   // add to the end of buffer
//   RecordID.PageNo := FRecordCount;
   if (FRecordCount = 0) then
    RecordID.PageNo := 0
   else
    RecordID.PageNo := FLastRecordID+1;
   RecordID.PageItemNo := 0;
   if (FAllocatedRecordCount <= RecordID.PageNo) then
    begin
     // increase record buffer size
     OldRecordCount := FAllocatedRecordCount;
     Inc(FAllocatedRecordCount,FAllocRecordsBy);
     i := FAllocatedRecordCount * FRecordBufferSize;
     if (OldRecordCount = 0) then
      FRecordBuffer := MemoryManager.GetMem(i)
     else
      MemoryManager.ReallocMem(Pointer(FRecordBuffer),i);
     i := FAllocatedRecordCount div 8;
     if (FAllocatedRecordCount mod 8 > 0) then
      Inc(i);
     if (OldRecordCount = 0) then
      FDeleteFlagBuffer := MemoryManager.GetMem(i)
     else
      MemoryManager.ReallocMem(Pointer(FDeleteFlagBuffer),i);
    end; // increase record buffer size
    Move(RecordBuffer^,PChar(FRecordBuffer + RecordID.PageNo * FRecordBufferSize)^,
          FRecordBufferSize);
    SetNullFlag(False,RecordID.PageNo,FDeleteFlagBuffer);
    Inc(FRecordCount);
    if (RecordID.PageNo < FFirstRecordID) then
     FFirstRecordID := RecordID.PageNo;
    if (RecordID.PageNo > FLastRecordID) then
     FLastRecordID := RecordID.PageNo;
    Result := True;
   except
  Result := False;
 end; // try
end; // AddRecord


//------------------------------------------------------------------------------
// update record, return true if record was updated, false if record was deleted
//------------------------------------------------------------------------------
function TACRMemoryRecordManager.UpdateRecord(
                          RecordBuffer: TACRRecordBuffer;
                          RecordID:     TACRRecordID;
                          SessionID:    TACRSessionID = INVALID_SESSION_ID
                         ): Boolean;
begin
 Result := False;
 if (not ((RecordID.PageNo < FFirstRecordID) or (RecordID.PageNo > FLastRecordID))) then
  try
   RewriteBLOBValues(RecordBuffer,PChar(FRecordBuffer + RecordID.PageNo * FRecordBufferSize));
   Move(RecordBuffer^,PChar(FRecordBuffer + RecordID.PageNo * FRecordBufferSize)^,
          FRecordBufferSize);
   Result := True;
  except
   Result := False;
  end;
end; // UpdateRecord


//------------------------------------------------------------------------------
// delete record, return true if record was deleted, false if record was deleted earlier
//------------------------------------------------------------------------------
function TACRMemoryRecordManager.DeleteRecord(
                          var RecordID: TACRRecordID;
                          SessionID:    TACRSessionID = INVALID_SESSION_ID
                         ): Boolean;
begin
 Result := False;
 RecordID.PageItemNo := 0;
 if (not (FRecordCount <= 0)) and
    (not ((RecordID.PageNo < FFirstRecordID) or (RecordID.PageNo > FLastRecordID))) and
    (not (CheckNullFlag(RecordID.PageNo,FDeleteFlagBuffer))) then
  try
   // set delete flag
   SetNullFlag(True,RecordID.PageNo,FDeleteFlagBuffer);
   Dec(FRecordCount);
   if (FRecordCount = 0) then
    begin
     // no records
     FFirstRecordID := 0;
     FLastRecordID := 0;
    end
   else
    begin
     // update first and last records ID
     if (RecordID.PageNo = FFirstRecordID) then
      GetRecordFromAnyPosition(grmNext,FFirstRecordID);
     if (RecordID.PageNo = FLastRecordID) then
      begin
       GetRecordFromAnyPosition(grmPrior,FLastRecordID);
       RecordID.PageNo := FLastRecordID;
      end;
    end;
   FNoRecordsDeleted := False;
   Result := True;
 except
  Result := False;
 end;
end; // DeleteRecord


//------------------------------------------------------------------------------
// return true if record exists
//------------------------------------------------------------------------------
function TACRMemoryRecordManager.IsRecordExists(var RecordID: TACRRecordID; SessionID: TACRSessionID): Boolean;
begin
  if ((FRecordCount <= 0) or (RecordID.PageNo >= FAllocatedRecordCount)) then
   Result := False
  else
   Result := not CheckNullFlag(RecordID.PageNo,FDeleteFlagBuffer);
end; // IsRecordExists


//------------------------------------------------------------------------------
// get record
//------------------------------------------------------------------------------
procedure TACRMemoryRecordManager.GetRecordBuffer(var NavigationInfo: TACRNavigationInfo);
begin
 if (FRecordCount = 0) then
  begin
   NavigationInfo.GetRecordResult := grrEOF;
   Exit;
  end;
 // get record relatively to the first position
 if (NavigationInfo.FirstPosition) then
  begin
   NavigationInfo.GetRecordResult := GetRecordFromFirstPosition(
                                      NavigationInfo.GetRecordMode,
                                      NavigationInfo.RecordID.PageNo
                                                               );
   if (NavigationInfo.GetRecordResult = grrOK) then
    NavigationInfo.FirstPosition := False;
  end
 else
 // get record relatively to the last position
 if (NavigationInfo.LastPosition) then
  begin
   NavigationInfo.GetRecordResult := GetRecordFromLastPosition(
                                      NavigationInfo.GetRecordMode,
                                      NavigationInfo.RecordID.PageNo
                                                               );
   if (NavigationInfo.GetRecordResult = grrOK) then
    NavigationInfo.LastPosition := False;
  end
 else
  NavigationInfo.GetRecordResult := GetRecordFromAnyPosition(
                                      NavigationInfo.GetRecordMode,
                                      NavigationInfo.RecordID.PageNo
                                                               );
 NavigationInfo.RecordID.PageItemNo := 0;                                                              
 if (NavigationInfo.GetRecordResult = grrOK) then
  begin
   try
     Move(
          PChar(FRecordBuffer + NavigationInfo.RecordID.PageNo * FRecordBufferSize)^,
          NavigationInfo.RecordBuffer^,
          FRecordBufferSize
         );
     NavigationInfo.GetRecordResult := grrOK;
   except
    raise EACRException.Create(10019,ErrorLRetreivingRecord);
   end;
  end; // record retrieved successfully
end; // GetRecordBuffer


//------------------------------------------------------------------------------
// return record no
//------------------------------------------------------------------------------
function TACRMemoryRecordManager.GetApproximateRecNo(RecordID: TACRRecordID; SessionID: TACRSessionID): TACRRecordNo;
begin
 if (FRecordCount = 0) or (RecordID.PageNo < FFirstRecordID) or (RecordID.PageNo > FLastRecordID) then
  Result := -1
 else
 if (FRecordCount = 1) then
  Result := 1
 else
  Result :=  Round((RecordID.PageNo - FFirstRecordID) / (FRecordCount) *
                   (FLastRecordID - FFirstRecordID)) + 1;
end; // GetApproximateRecNo


//------------------------------------------------------------------------------
// Load from stream
//------------------------------------------------------------------------------
procedure TACRMemoryRecordManager.LoadFromStream(Stream: TStream);
var i: Integer;
begin
 LoadDataFromStream(i,Sizeof(i),Stream,10449);
 SetRecordCount(i);
 i := (FAllocatedRecordCount div 8) + Integer((FAllocatedRecordCount mod 8) > 0);
 LoadDataFromStream(FNoRecordsDeleted,Sizeof(FNoRecordsDeleted),Stream,10450);
 LoadDataFromStream(FFirstRecordID,Sizeof(FFirstRecordID),Stream,10451);
 LoadDataFromStream(FLastRecordID,Sizeof(FLastRecordID),Stream,10452);
 if (FAllocatedRecordCount > 0) then
   LoadDataFromStream(FDeleteFlagBuffer^,i,Stream,10453);
end; // LoadFromStream


//------------------------------------------------------------------------------
// Save from stream
//------------------------------------------------------------------------------
procedure TACRMemoryRecordManager.SaveToStream(Stream: TStream);
var i: Integer;
begin
 i := (FAllocatedRecordCount div 8) + Integer((FAllocatedRecordCount mod 8) > 0);
 SaveDataToStream(FAllocatedRecordCount,Sizeof(FAllocatedRecordCount),Stream,10444);
 SaveDataToStream(FNoRecordsDeleted,Sizeof(FNoRecordsDeleted),Stream,10445);
 SaveDataToStream(FFirstRecordID,Sizeof(FFirstRecordID),Stream,10446);
 SaveDataToStream(FLastRecordID,Sizeof(FLastRecordID),Stream,10447);
 if (FAllocatedRecordCount > 0) then
  SaveDataToStream(FDeleteFlagBuffer^,i,Stream,10448);
end; // SaveToStream


//------------------------------------------------------------------------------
// add loaded record
//------------------------------------------------------------------------------
procedure TACRMemoryRecordManager.AddLoadedRecord(RecordBuffer: TACRRecordBuffer; var RecordPos: Integer);

 procedure GoToNextBit;
 begin
   Inc(RecordPos);
   if (not FNoRecordsDeleted) then
     if (FLoadingRecordInfo.BitNo < 7) then
      Inc(FLoadingRecordInfo.BitNo)
     else
      begin
       FLoadingRecordInfo.BitNo := 0;
       Inc(FLoadingRecordInfo.ByteNo);
       Inc(FLoadingRecordInfo.pFlag);
       if (FLoadingRecordInfo.ByteNo >= (FAllocatedRecordCount div 8)) then
         raise EACRException.Create(10454,ErrorLInvalidRecordNo,[FLoadingRecordInfo.ByteNo]);
      end;
   if (RecordPos >= FAllocatedRecordCount) then
     raise EACRException.Create(10847,ErrorLInvalidRecordNo,[FAllocatedRecordCount]);
 end;

 procedure SearchForNextRecord;
 var bOK: Boolean;
 begin
  repeat
   bOK := (((FLoadingRecordInfo.pFlag^ shr FLoadingRecordInfo.BitNo) and $01) = 0);
   if (not bOK) then
    GoToNextBit;
  until (bOK);
 end; // SearchForNextRecord

begin
 Inc(FRecordCount);

 if (RecordPos = INVALID_ID4) then
  begin
    if (FNoRecordsDeleted) then
     RecordPos := 0
    else
     begin
        // init FLoadingRecordsInfo
        FLoadingRecordInfo.pFlag := PByte(FDeleteFlagBuffer);
        FLoadingRecordInfo.BitNo := 0;
        FLoadingRecordInfo.ByteNo := 0;
        RecordPos := 0; 
     end; 
  end
 else
  GoToNextBit; 
 if (not FNoRecordsDeleted) then 
   SearchForNextRecord;
	
 Move(RecordBuffer^,PChar(FRecordBuffer + RecordPos * FRecordBufferSize)^,FRecordBufferSize); 
end; // AddLoadedRecord
 
	
//////////////////////////////////////////////////////////////////////////////// 
//
// TACRMemoryTableData 
//
//////////////////////////////////////////////////////////////////////////////// 
 
	
//------------------------------------------------------------------------------ 
// GetPageManager 
//------------------------------------------------------------------------------ 
function TACRMemoryTableData.GetPageManager: TACRPageManager; 
begin 
  Result := FPageManager;
end;// GetPageManager
	
	
//------------------------------------------------------------------------------
// create RecordManager 
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.CreateRecordManager(TablePFSMapFileStartPageNo: TACRPageNo); 
begin
 if (FRecordManager <> nil) then 
  FRecordManager.Free; 
 FRecordManager := TACRMemoryRecordManager.Create(
                      Self, 
                      GetRecordBufferSize,
                      FAllocRecordsBy 
                                                 ); 
end;// CreateRecordManager
	
 
//------------------------------------------------------------------------------ 
// create index manager 
//------------------------------------------------------------------------------ 
procedure TACRMemoryTableData.CreateIndexManager(IndexDefs: TACRIndexDefs); 
begin 
 if (FIndexManager <> nil) then 
  FIndexManager.Free;
 FillDefsByObjectId(IndexDefs);
 FIndexManager := TACRBaseIndexManager.Create(Self); 
 FIndexManager.CreateIndexDefs(IndexDefs);
end;// CreateIndexManager
 
	
//------------------------------------------------------------------------------
// load table header 
//------------------------------------------------------------------------------ 
procedure TACRMemoryTableData.LoadTableHeader( 
              Stream:             TStream;
              var BLOBDescriptor: TACRBLOBDescriptor
                                             ); 
var FileHeader: TACRMemoryTableFileHeader;
    DatabaseID: TACRSequenceValue;
begin 
  LoadDataFromStream(FileHeader,sizeof(FileHeader),Stream,10160); 
  if (FileHeader.Signature <> ACRSignature) then
    raise EACRException.Create(10161,ErrorLInvalidSignature,
      [FileHeader.Signature,ACRSignature]);
  if (FileHeader.Version > ACRVersion + 0.0001) then 
    raise EACRException.Create(10162,ErrorLInvalidVersion, 
      [FileHeader.Version,ACRVersion]); 
  if (FileHeader.Version < 1.0001) then 
    raise EACRException.Create(10443,ErrorLInvalidVersion,
      [FileHeader.Version,ACRVersion]); 
  if (FileHeader.Version < 1.010001) then 
    raise EACRException.Create(10454,ErrorLInvalidVersion, 
      [FileHeader.Version,ACRVersion]);
  if (FileHeader.Version < 1.020001) then
    raise EACRException.Create(10455,ErrorLInvalidVersion,
      [FileHeader.Version,ACRVersion]); 
  SetLength(FTableName,FileHeader.NameLength);
  LoadDataFromStream(PChar(@FTableName[1])^,FileHeader.NameLength,Stream,10163);
  LoadDataFromStream(DatabaseID,sizeof(DatabaseID),Stream,10417);
  if (FDatabaseData.ObjectIdSequence.LastValue < DatabaseID) then 
    FDatabaseData.ObjectIdSequence.LastValue := DatabaseID; 
  BLOBDescriptor.StartPosition := Stream.Position;
  BLOBDescriptor.BlockSize := FileHeader.BlockSize;
  BLOBDescriptor.NumBlocks := FileHeader.NumBlocks;
  BLOBDescriptor.UncompressedSize := FileHeader.UncompressedSize;
  BLOBDescriptor.CompressionAlgorithm := FileHeader.CompressionAlgorithm;
  BLOBDescriptor.CompressionMode := FileHeader.CompressionMode;
  FLoadedRecordCount := FileHeader.RecordCount; 
end; // LoadTableHeader
	
 
//------------------------------------------------------------------------------
// save table header
//------------------------------------------------------------------------------ 
procedure TACRMemoryTableData.SaveTableHeader( 
              Stream:             TStream; 
              var BLOBDescriptor: TACRBLOBDescriptor 
              );
var FileHeader: TACRMemoryTableFileHeader; 
    DatabaseID: TACRSequenceValue; 
begin
  FileHeader.Signature := ACRSignature;
  FileHeader.Version := ACRVersion;
  FileHeader.RecordCount := FRecordManager.GetRecordCount;
  FileHeader.UncompressedSize := BLOBDescriptor.UncompressedSize; 
  FileHeader.NumBlocks := BLOBDescriptor.NumBlocks; 
  FileHeader.BlockSize := BLOBDescriptor.BlockSize;
  FileHeader.CompressionAlgorithm := BLOBDescriptor.CompressionAlgorithm;
  FileHeader.CompressionMode := BLOBDescriptor.CompressionMode; 
  FileHeader.NameLength := Length(FTableName); 
  SaveDataToStream(FileHeader,sizeof(FileHeader),Stream,10156); 
  SaveDataToStream(PChar(@FTableName[1])^,FileHeader.NameLength,Stream,10157);
  DatabaseID := FDatabaseData.ObjectIdSequence.LastValue;
  SaveDataToStream(DatabaseID,sizeof(DatabaseID),Stream,10416);
end; // SaveTableHeader
	
 
//------------------------------------------------------------------------------
// load blob data 
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.LoadBLOBDataFromStream(
        RecordBuffer: TACRRecordBuffer;
        FieldNo:      Integer;
        Stream:       TStream
                             );
var Buffer:                 PChar;
    FBLOBPosition:          Int64;
    Offset,Offset2:         Integer;
    BLOBPartialDescriptor:  TACRPartialTemporaryBLOBDescriptor;
begin
  Offset := FFieldManager.FieldDefs[FieldNo].MemoryOffset;
  Move(PChar(RecordBuffer + Offset)^,FBLOBPosition,sizeof(FBLOBPosition));
  // load disk blob descriptor
  SetStreamPosition(Stream,FBLOBPosition,10202);
  LoadDataFromStream(BLOBPartialDescriptor,sizeof(BLOBPartialDescriptor),
    Stream,10203);
  Offset2 := sizeof(TACRPartialBLOBDescriptor);
  Buffer := MemoryManager.GetMem(BLOBPartialDescriptor.CompressedSize + Offset2);
  PACRPartialBLOBDescriptor(Buffer)^.NumBlocks :=
    BLOBPartialDescriptor.NumBlocks;
  PACRPartialBLOBDescriptor(Buffer)^.UncompressedSize :=
    BLOBPartialDescriptor.UncompressedSize;
  // load compresse blob data
  LoadDataFromStream(PChar(Buffer + Offset2)^,
    BLOBPartialDescriptor.CompressedSize,Stream,10204);
  Move(Buffer,PChar(RecordBuffer + Offset)^,sizeof(Buffer));
end; // LoadBLOBDataFromStream


//------------------------------------------------------------------------------
// load record
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.LoadRecordFromStream(
            RecordBuffer: TACRRecordBuffer;
            Stream:       TStream
                                                  );
var i:                        Integer;
begin
  LoadDataFromStream(RecordBuffer^,GetRecordBufferSize,Stream,10201);
  if (FBLOBFieldsPresent) then
   begin
     for i := 0 to FFieldManager.FieldDefs.Count - 1 do
      if (IsBLOBFieldType(FFieldManager.FieldDefs[i].BaseFieldType)) then
        if (not CheckNullFlag(i,RecordBuffer)) then
         begin
          // load blob
          LoadBLOBDataFromStream(RecordBuffer,i,Stream);
         end; // blob field not null
   end;
end; // LoadRecord


//------------------------------------------------------------------------------
// prepare record buffer for save
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.PrepareRecordBufferForSave(
            RecordBuffer:             TACRRecordBuffer;
            BLOBDescriptorList:       TList;
            BLOBDataList:             TList;
            BLOBPosition:             Int64
                                                );
var FBLOBPosition:            Int64;
    i:                        Integer;
    BLOBDataSize:             Integer;
    Offset:                   Integer;
    Buffer:                   PChar;
    PBLOBPartialDescriptor:   PACRPartialTemporaryBLOBDescriptor;
begin
 FBLOBPosition := BLOBPosition;
 for i := 0 to FFieldManager.FieldDefs.Count - 1 do
  if (IsBLOBFieldType(FFieldManager.FieldDefs[i].BaseFieldType)) then
    if (not CheckNullFlag(i,RecordBuffer)) then
     begin
      // save blob field data
      Offset := FFieldManager.FieldDefs[i].MemoryOffset;
      Move(PChar(RecordBuffer + Offset)^,Buffer,Sizeof(Buffer));
      if (Buffer = nil) then
        raise EACRException.Create(10196,ErrorLNilPointer);
      // calculate size of blob data
      BLOBDataSize := MemoryManager.GetMemoryBufferSize(Buffer) -
        sizeof(TACRPartialBLOBDescriptor) +
        sizeof(TACRPartialTemporaryBLOBDescriptor);
      PBLOBPartialDescriptor :=
        MemoryManager.GetMem(sizeof(TACRPartialTemporaryBLOBDescriptor));
      PBLOBPartialDescriptor^.NumBlocks :=
        PACRPartialBLOBDescriptor(Buffer)^.NumBlocks;
      PBLOBPartialDescriptor^.UncompressedSize :=
        PACRPartialBLOBDescriptor(Buffer)^.UncompressedSize;
      PBLOBPartialDescriptor^.CompressedSize := BLOBDataSize -
        sizeof(TACRPartialTemporaryBLOBDescriptor);
      BLOBDataList.Add(Buffer);
      BLOBDescriptorList.Add(PBLOBPartialDescriptor);
      // store offset to blob data instead of pointer to memory buffer
      Move(FBLOBPosition,PChar(RecordBuffer + Offset)^,Sizeof(FBLOBPosition));
      Inc(FBLOBPosition,BLOBDataSize);
     end; // not null blob field
end;


//------------------------------------------------------------------------------
// save record
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.SaveRecordToStream(
            RecordBuffer: TACRRecordBuffer;
            Stream:       TStream
                                                );
var i:                        Integer;
    BLOBDataSize:             Integer;
    Buffer:                   PChar;
    PBLOBPartialDescriptor:   PACRPartialTemporaryBLOBDescriptor;
    BLOBDescriptorList:       TList;
    BLOBDataList:             TList;
begin
 if (FBLOBFieldsPresent) then
  begin
   BLOBDescriptorList := TList.Create;
   BLOBDataList := TList.Create;
   try
     PrepareRecordBufferForSave(RecordBuffer,BLOBDescriptorList,
        BLOBDataList,Int64(Stream.Position + Int64(GetRecordBufferSize)));
     SaveDataToStream(RecordBuffer^,GetRecordBufferSize,Stream,10197);
     // save blob fields
     for i := 0 to BLOBDataList.Count - 1 do
      begin
       PBLOBPartialDescriptor := BLOBDescriptorList.Items[i];
       SaveDataToStream(PBLOBPartialDescriptor^,
        sizeof(TACRPartialTemporaryBLOBDescriptor),Stream,10198);
       Buffer := BLOBDataList.Items[i];
       BLOBDataSize := PBLOBPartialDescriptor^.CompressedSize;
       SaveDataToStream(PChar(Buffer + sizeof(TACRPartialBLOBDescriptor))^,BLOBDataSize,Stream,10199);
       MemoryManager.FreeAndNilMem(PBLOBPartialDescriptor);
      end; // not null blob field
   finally
    BLOBDescriptorList.Free;
    BLOBDataList.Free;
   end;
  end // blob fields
 else
  // no blob fields
  SaveDataToStream(RecordBuffer^,GetRecordBufferSize,Stream,10200);
end; // SaveRecord


//------------------------------------------------------------------------------
// load all records
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.LoadRecordManager(Stream: TStream);
var RecordBuffer:     TACRRecordBuffer;
    i,RecPos:         Integer;
begin
  if (FLoadedRecordCount > 0) then
   begin
    FRecordManager.LoadFromStream(Stream);
    RecordBuffer := MemoryManager.GetMem(FRecordManager.RecordBufferSize);
    try
      RecPos := INVALID_ID4;
      for i := 0 to FLoadedRecordCount - 1 do
       begin
        LoadRecordFromStream(RecordBuffer,Stream);
        FRecordManager.AddLoadedRecord(RecordBuffer,RecPos);
       end;
    finally
      MemoryManager.FreeAndNilMem(RecordBuffer);
    end;
   end;
end; // LoadRecordManager


//------------------------------------------------------------------------------
// save all records
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.SaveRecordManager(Stream: TStream);
var RecordBuffer:     TACRRecordBuffer;
    NavigationInfo:   TACRNavigationInfo;
begin
  if (FRecordManager.GetRecordCount > 0) then
   begin
     FRecordManager.SaveToStream(Stream);
     RecordBuffer := MemoryManager.GetMem(FRecordManager.RecordBufferSize);
     NavigationInfo.LastPosition := False;
     NavigationInfo.FirstPosition := True;
     NavigationInfo.RecordBuffer := RecordBuffer;
     NavigationInfo.GetRecordMode := grmNext;
     try
       FRecordManager.GetRecordBuffer(NavigationInfo);
       if (NavigationInfo.GetRecordResult = grrOK) then
        begin
         NavigationInfo.FirstPosition := False;
         repeat
          // save record buffer and blob fields
          SaveRecordToStream(RecordBuffer,Stream);
          FRecordManager.GetRecordBuffer(NavigationInfo);
         until (NavigationInfo.GetRecordResult <> grrOK);
        end;
     finally
      MemoryManager.FreeAndNilMem(RecordBuffer);
     end;
   end;
end; // SaveRecordManager


//------------------------------------------------------------------------------
// Return bitmap size
//------------------------------------------------------------------------------
function TACRMemoryTableData.GetBitmapSize(SessionID: TACRSessionID): TACRRecordNo;
begin
  Result := TACRMemoryRecordManager(FRecordManager).GetBitmapSize;
end; // GetBitmapSize


//------------------------------------------------------------------------------
// return filter bitmap rec no by record id
//------------------------------------------------------------------------------
function TACRMemoryTableData.GetBitmapRecNoByRecordID(RecordID: TACRRecordID): TACRRecordNo;
begin
  Result := RecordID.PageNo;
end; // GetBitmapRecNoByRecordID


//------------------------------------------------------------------------------
// return filter bitmap rec no by record id
//------------------------------------------------------------------------------
function TACRMemoryTableData.GetRecordIDByBitmapRecNo(RecordNo: TACRRecordNo): TACRRecordID;
begin
  Result.PageNo := RecordNo;
  Result.PageItemNo := 0;
end; // GetRecordIDByBitmapRecNo


//------------------------------------------------------------------------------
// empty table
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.InternalEmptyTable;
var
    NavigationInfo:   TACRNavigationInfo;
begin
 if (FRecordManager <> nil) then
   if (FBLOBFieldsPresent and (FRecordManager.GetRecordCount > 0)) then
    begin
     NavigationInfo.LastPosition := False;
     NavigationInfo.FirstPosition := True;
     NavigationInfo.RecordBuffer := MemoryManager.GetMem(FRecordManager.RecordBufferSize);
     NavigationInfo.GetRecordMode := grmNext;
     try
       FRecordManager.GetRecordBuffer(NavigationInfo);
       if (NavigationInfo.GetRecordResult = grrOk) then
        begin
         NavigationInfo.FirstPosition := False;
         repeat
          ClearBLOBFieldsInRecordBuffer(NavigationInfo.RecordBuffer);
          FRecordManager.GetRecordBuffer(NavigationInfo);
         until (NavigationInfo.GetRecordResult <> grrOK);
        end;
     finally
      MemoryManager.FreeAndNilMem(NavigationInfo.RecordBuffer);
     end;
    end;
 if (FRecordManager <> nil) then
  FRecordManager.Empty;
 FPageManager.ApplyChanges(SYSTEM_SESSION_ID);
end; // InternalEmptyTable


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRMemoryTableData.Create(
                        aDatabaseData: TACRDatabaseData;
                        AllocRecordsBy: Integer
                      );
begin
 inherited Create(aDatabaseData);
 FPageManager := TACRMemoryPageManager.Create;
 FAllocRecordsBy := AllocRecordsBy;
 FInMemory := True;
 FTemporary := False;
end; // Create;


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRMemoryTableData.Destroy;
begin
 EmptyTable(nil);
 if (FSequenceManager <> nil) then
  FSequenceManager.Free;
 FSequenceManager := nil;
 DeleteAllIndexes(nil);
 inherited Destroy;
 FPageManager.ApplyChanges(SYSTEM_SESSION_ID);
 FPageManager.Free;
end; // Destroy


//------------------------------------------------------------------------------
// create table
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.CreateTable(
                          Cursor: TACRCursor;
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                                         );
begin

{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRMemoryTableData.CreateTable start');
 aaWriteToLog('TACRMemoryTableData.CreateTable FieldDefs.Count = '+IntToStr(FieldDefs.Count));
{$ENDIF}
 if (FieldDefs.Count <= 0) then
  raise EACRException.Create(10025,ErrorLNoFields);
 TableName := Cursor.TableName;
 CheckFieldDefinitions(FieldDefs);
 CheckIndexDefinitions(IndexDefs);
 CheckConstraintDefinitions(ConstraintDefs);
 CreateSequenceManager;
 CreateFieldManager(FieldDefs);
 CreateIndexManager(IndexDefs);
 CreateConstraintManager(ConstraintDefs);
 BuildSequences;
 if (FConstraintManager.ConstraintDefs.ForeignKeysExists) then
  CheckForeignKeyDefinitionsAndCreateForeignKeyActions(Cursor);
 CreateRecordManager;
 InitCursor(Cursor);
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRMemoryTableData.CreateTable finish');
{$ENDIF}
end;// CreateTable


//------------------------------------------------------------------------------
// delete table
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.DeleteTable(Session: TACRBaseSession; Cascade: Boolean; DesignMode: Boolean = False);
begin
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRMemoryTableData.DeleteTable start, name = '+FTableName);
{$ENDIF}
 inherited DeleteTable(Session,Cascade,DesignMode);
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRMemoryTableData.DeleteTable 1, name = '+FTableName);
{$ENDIF}
 if (FIndexManager.IndexDefs.Count > 0) then
  begin
   if (Session <> nil) then
    FSessionID := Session.SessionID
   else
    FSessionID := SYSTEM_SESSION_ID;
   FIndexManager.DropAllIndexes(FSessionID);
   FIndexManager.PageManager.ApplyChanges(FSessionID);
  end;
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRMemoryTableData.DeleteTable 2, name = '+FTableName+ ' cursors count ='+IntToStr(FCursorList.Count));
{$ENDIF}
 if ((not DesignMode) or (FCursorList.Count = 0)) then
  begin
  {$IFDEF DEBUG_TRACE_DATASET}
   aaWriteToLog('TACRMemoryTableData.DeleteTable 3');
  {$ENDIF}
   Free;
  {$IFDEF DEBUG_TRACE_DATASET}
   aaWriteToLog('TACRMemoryTableData.DeleteTable 4');
  {$ENDIF}
  end;
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRMemoryTableData.DeleteTable finish');
{$ENDIF}
end; // DeleteTable


//------------------------------------------------------------------------------
// delete constraint (FK,FKAction or PK) and write changes to MetaData file
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.DeleteConstraint(
                          Cursor:           TACRCursor;
                          Name:             String;
                          Cascade:          Boolean;
                          FKPartialDelete:  Boolean
                          );
begin
  Lock;
  try
    inherited DeleteConstraint(Cursor,Name,Cascade,FKPartialDelete);
    FPageManager.ApplyChanges(Cursor.Session.SessionID);
  finally
    Unlock;
  end;
end; // DeleteConstraint


//------------------------------------------------------------------------------
// empty table
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.EmptyTable(Cursor: TACRCursor);
begin
 if (FConstraintManager.ConstraintDefs.ForeignKeysActionsDeleteExists) then
  raise EACRException.Create(11494,ErrorLCannotEmptyTableWithForeignKeyActionsDelete,[FTableName]);
 inherited EmptyTable(Cursor);
 InternalEmptyTable;
 if (Cursor <> nil) then
  FPageManager.ApplyChanges(Cursor.Session.SessionID);
end; // EmptyTable


//------------------------------------------------------------------------------
// rename table
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.RenameTable(Cursor: TACRCursor; NewTableName: String);
begin
 if (FDatabaseData.TableExists(NewTableName)) then
  raise EACRException.Create(11491,ErrorLCannotRenameTableAlreadyExists,[FTableName, NewTableName]);
 inherited RenameTable(Cursor,NewTableName);
end; // RenameTable


//------------------------------------------------------------------------------
// load table from stream
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.LoadTableFromStream(
                        Stream: TStream
                       );
var
    TempStream:       TStream;
    BLOBDescriptor:   TACRBLOBDescriptor;
    i:                Integer;
    Buffer,OutBuf:    PChar;
begin
 Lock;
// EmptyTable(nil);
 try
  inherited;
  CreateSequenceManager;
  if (FFieldManager <> nil) then
   FFieldManager.Free;
  FFieldManager := TACRBaseFieldManager.Create(Self,FSequenceManager);
  if (FConstraintManager = nil) then
    FConstraintManager := TACRBaseConstraintManager.Create(Self);
  LoadTableHeader(Stream,BLOBDescriptor);
  // create compressed stream
  if (BLOBDescriptor.CompressionAlgorithm = Byte(acaNone)) then
   TempStream := Stream
  else
   begin
    Buffer := MemoryManager.GetMem(BLOBDescriptor.BlockSize);
    try
     Stream.ReadBuffer(Buffer^,BLOBDescriptor.BlockSize);
     ACRInternalDecompressBuffer(
          TACRCompressionAlgorithm(BLOBDescriptor.CompressionAlgorithm),
          Buffer,BLOBDescriptor.BlockSize,OutBuf,i);
    finally
      MemoryManager.FreeAndNilMem(Buffer);
    end;
    TempStream := TACRMemoryStream.Create;
    TempStream.WriteBuffer(OutBuf^,i);
    TempStream.Position := 0;
    FreeMem(OutBuf);
   end;
  try
   LoadSequencesFromStream(TempStream);
   FFieldManager.LoadFromStream(TempStream);
   FFieldManager.FieldDefs.RecalcFieldOffsets;
   FConstraintManager.LoadFromStream(TempStream);
   if (FIndexManager <> nil) then
    FIndexManager.Free;
   FIndexManager := TACRBaseIndexManager.Create(Self);
   FIndexManager.LoadFromStream(TempStream);
   FBLOBFieldsPresent := False;
   for i := 0 to FFieldManager.FieldDefs.Count - 1 do
    if (IsBLOBFieldType(FFieldManager.FieldDefs[i].BaseFieldType)) then
     begin
      FBLOBFieldsPresent := True;
      break;
     end;
   if (FRecordManager = nil) then
    CreateRecordManager;
   LoadRecordManager(TempStream);

  finally
   if (BLOBDescriptor.CompressionAlgorithm <> Byte(acaNone)) then
    TempStream.Free;
  end;
 finally
  Unlock;
 end;
end; // LoadTable


//------------------------------------------------------------------------------
// save table to stream
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.SaveTableToStream(
                        Stream: TStream;
                        CompressionAlgorithm: TACRCompressionAlgorithm;
                        CompressionMode:      Byte;
                        BlockSize:            Integer
                      );
var OldPos,StartPos:  Int64;
    TempStream:       TStream;
    BLOBDescriptor:   TACRBLOBDescriptor;
    Buffer:           PChar;
    OutSize:          Integer;
begin
 Lock;
 try
  inherited;
  StartPos := Stream.Position;
  BLOBDescriptor.CompressionAlgorithm := Byte(CompressionAlgorithm);
  if (CompressionMode = 0) then
   CompressionMode := 1;
  BLOBDescriptor.CompressionMode := CompressionMode;
  BLOBDescriptor.BlockSize := BlockSize;
  // save table header
  // create compressed stream
  if (CompressionAlgorithm = acaNone) then
   TempStream := Stream
  else
   TempStream := TACRMemoryStream.Create;
  try
   SaveTableHeader(Stream,BLOBDescriptor);
   SaveSequencesToStream(TempStream);
   // save other objects
   FFieldManager.SaveToStream(TempStream);
   FConstraintManager.SaveToStream(TempStream);
   FIndexManager.SaveToStream(TempStream);
   SaveRecordManager(TempStream);
   if (CompressionAlgorithm <> acaNone) then
    begin
     ACRInternalCompressBuffer(CompressionAlgorithm,CompressionMode,
      TACRMemoryStream(TempStream).Buffer,TempStream.Size,Buffer,OutSize);
     BLOBDescriptor.BlockSize := OutSize;
     BLOBDescriptor.UncompressedSize := TempStream.Size;
     Stream.WriteBuffer(Buffer^,OutSize);
     FreeMem(Buffer);
    end;
  finally
    if (CompressionAlgorithm <> acaNone) then
     TempStream.Free;
    // save table header second time - store params of compressed blob stream
    OldPos := Stream.Position;
    Stream.Position := StartPos;
    if (Stream.Position <> StartPos) then
      raise EACRException.Create(10155,ErrorLCannotSetPosition,
        [StartPos,OldPos,Stream.Position]);
    SaveTableHeader(Stream,BLOBDescriptor);
    Stream.Position := OldPos;
  end;
 finally
  Unlock;
 end;
end; // SaveTable


//------------------------------------------------------------------------------
// open table
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.OpenTable(Cursor: TACRCursor);
begin
 TableName := Cursor.TableName;
 inherited OpenTable(Cursor);
end;// OpenTable


//------------------------------------------------------------------------------
// close table
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.CloseTable(Cursor: TACRCursor);
begin
 inherited CloseTable(Cursor);
end;// CloseTable


//------------------------------------------------------------------------------
// add index
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.AddIndex(IndexDef: TACRIndexDef; Cursor: TACRCursor);
begin
 Lock;
 try
   inherited AddIndex(IndexDef,Cursor);
   FPageManager.ApplyChanges(Cursor.Session.SessionID);
 finally
  Unlock;
 end;
end; // AddIndex


//------------------------------------------------------------------------------
// return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
//------------------------------------------------------------------------------
function TACRMemoryTableData.CompareRecordID(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer;
begin
 if (RecordID1.PageNo = RecordID2.PageNo) then
  Result := 0
 else
 if (RecordID1.PageNo > RecordID2.PageNo) then
  Result := 1
 else
  Result := -1; 
end; // CompareRecordID


//------------------------------------------------------------------------------
// insert record
//------------------------------------------------------------------------------
function TACRMemoryTableData.InsertRecord(var Cursor: TACRCursor): Boolean;
var
    Pos:           Pointer;
begin
 Result := False;
 Lock;
 try
  if (FRecordManager = nil) then
   raise EACRException.Create(10010,ErrorLNilPointer);
  if (Cursor.CurrentRecordBuffer = nil) then
   raise EACRException.Create(10018,ErrorLNilPointer);
  // add record to first empty space
  SetBLOBValuesModified(False,Cursor);
  Pos := Cursor.SavePosition;
  try
    Result := FRecordManager.AddRecord(Cursor.CurrentRecordBuffer,Cursor.CurrentRecordID);
    // current record id will be the same as position as we insert in first empty record
    Cursor.CurrentRecordID.PageItemNo := 0;
    if (Result) then
     begin
       Cursor.FirstPosition := False;
       Cursor.LastPosition := False;
       if (TACRRecordBitmap(Cursor.RecordBitmap).Active) then
        UpdateRecordBitmapAfterInsertRecord(Cursor,Pos);
       if (FIndexManager.IndexDefs.Count > 0) then
        FIndexManager.InsertRecord(Cursor);
       if (FIndexManager.IndexDefs.Count > 0) then
         FPageManager.ApplyChanges(Cursor.Session.SessionID);
     end;
  finally
    Cursor.FreePosition(Pos);
  end;
 finally
  Unlock;
 end;
end; // InsertRecord;


//------------------------------------------------------------------------------
// delete record
//------------------------------------------------------------------------------
function TACRMemoryTableData.DeleteRecord(Cursor: TACRCursor): Boolean;
var
  RecordID:    TACRRecordID;
  NewRecordID: TACRRecordID;
  Buffer:      TACRRecordBuffer;
  TempBuffer:  TACRRecordBuffer;
  OldPos:      Pointer;
begin
 Result := False;
 Lock;
 try
  if (FRecordManager = nil) then
   raise EACRException.Create(10013,ErrorLNilPointer);
  RecordID := Cursor.CurrentRecordID;
  Buffer := Cursor.CurrentRecordBuffer;
  if (FConstraintManager.ConstraintDefs.ForeignKeysActionsDeleteExists) then
   ExecuteForeignKeyActions(Cursor,False);
  // clear blob fields
  ClearBLOBFieldsInRecordBuffer(Buffer);
  TempBuffer := Cursor.AllocateRecordBuffer;
  OldPos := Cursor.SavePosition;
  try
   Cursor.CurrentRecordBuffer := TempBuffer;
   if (GetRecordBuffer(Cursor, grmNext) = grrOK) then
     NewRecordID := Cursor.CurrentRecordID
   else
    begin
      Cursor.RestorePosition(OldPos);
      if (GetRecordBuffer(Cursor, grmPrior) = grrOK) then
        NewRecordID := Cursor.CurrentRecordID;
    end;
   Cursor.RestorePosition(OldPos);
   Cursor.CurrentRecordBuffer := Buffer;
   Cursor.CurrentRecordID := RecordID;
  finally
   Cursor.FreePosition(OldPos);
   Cursor.FreeRecordBuffer(TempBuffer);
  end;
  FIndexManager.DeleteRecord(Cursor);
  Result := FRecordManager.DeleteRecord(RecordID);
  if (Result) then
   begin
    if (TACRRecordBitmap(Cursor.RecordBitmap).Active) then
      TACRRecordBitmap(Cursor.RecordBitmap).HideRecord(Cursor.CurrentRecordID);
    if (Cursor.GetRecordCount = 0) then
     begin
      Cursor.FirstPosition := True;
      Cursor.LastPosition := False;
     end
    else
     begin
      Cursor.FirstPosition := False;
      Cursor.LastPosition := False;
      Cursor.CurrentRecordID := NewRecordID;
     end;
    if (FIndexManager.IndexDefs.Count > 0) then
       FPageManager.ApplyChanges(Cursor.Session.SessionID);
   end;

 finally
  Unlock;
 end;
end; // DeleteRecord


//------------------------------------------------------------------------------
// update record
//------------------------------------------------------------------------------
function TACRMemoryTableData.UpdateRecord(Cursor: TACRCursor): Boolean;
begin
 Result := False;
 Lock;
 try
  if (FRecordManager = nil) then
   raise EACRException.Create(10014,ErrorLNilPointer);
  if (Cursor.CurrentRecordBuffer = nil) then
   raise EACRException.Create(10017,ErrorLNilPointer);
  if (FConstraintManager.ConstraintDefs.ForeignKeysActionsUpdateExists) then
   ExecuteForeignKeyActions(Cursor,True);
  FIndexManager.UpdateRecord(Cursor);
  Result := FRecordManager.UpdateRecord(Cursor.CurrentRecordBuffer,
                              Cursor.CurrentRecordID);
  SetBLOBValuesModified(False,Cursor);
  if (Result) then
   begin
     if (FIndexManager.IndexDefs.Count > 0) then
       FPageManager.ApplyChanges(Cursor.Session.SessionID);
     if (TACRRecordBitmap(Cursor.RecordBitmap).Active) then
       UpdateRecordBitmapAfterUpdateRecord(Cursor);
   end;
 finally
  Unlock;
 end;
end; // UpdateRecord


//------------------------------------------------------------------------------
// delete all visible records
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.DeleteVisibleRecords(Cursor: TACRCursor);
var VisibleRecordCount: TACRRecordNo;
    Filtered:           Boolean;
begin
  Lock;
  try
    if (FRecordManager = nil) then
     raise EACRException.Create(11312,ErrorLNilPointer);
    if (Cursor = nil) then
     raise EACRException.Create(11509,ErrorLNilPointer);
    FSessionID := Cursor.Session.SessionID;
    FCachedRecordCount := FRecordManager.GetRecordCount;
    Filtered := TACRRecordBitmap(Cursor.RecordBitmap).Active;
    if (Filtered) then
     VisibleRecordCount := TACRRecordBitmap(Cursor.RecordBitmap).GetRecordCount
    else
     VisibleRecordCount := FCachedRecordCount;
    if (VisibleRecordCount > 0) then
     begin
       if (FCachedRecordCount < VisibleRecordCount) then
        raise EACRException.Create(11311,ErrorLInvalidRecordCount,[VisibleRecordCount,FCachedRecordCount])
       else
       if ((FCachedRecordCount = VisibleRecordCount) and
          (not FConstraintManager.ConstraintDefs.ForeignKeysActionsDeleteExists)) then
        begin
          InternalEmptyTable;
          EmptyAllIndexes(Cursor.Session.SessionID);
        end
       else
        begin
         inherited DeleteVisibleRecords(Cursor);
        end;
      if (FIndexManager.IndexDefs.Count > 0) then
       FPageManager.ApplyChanges(Cursor.Session.SessionID);
     end;
    Cursor.FirstPosition := True;
    Cursor.LastPosition := False;
  finally
    Unlock;
  end;
end; // DeleteVisibleRecords


//------------------------------------------------------------------------------
// update visible records
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.UpdateVisibleRecords(
                                Cursor:       TACRCursor;
                                FieldNames:   TStringList;
                                values:       array of TACRVariant;
                                SkipFKCheck:  Boolean = False
                                );
begin
  Lock;
  try
    FSessionID := Cursor.Session.SessionID;
    if (FRecordManager = nil) then
     raise EACRException.Create(11450,ErrorLNilPointer);
    inherited UpdateVisibleRecords(Cursor,FieldNames,values,SkipFKCheck);
    if (FIndexManager.IndexDefs.Count > 0) then
     FPageManager.ApplyChanges(Cursor.Session.SessionID);
  finally
    Unlock;
  end;
end; // UpdateVisibleRecords


//------------------------------------------------------------------------------
// return record count
//------------------------------------------------------------------------------
function TACRMemoryTableData.InternalGetRecordCount: TACRRecordNo;
begin
 Result := 0;
 Lock;
 try
  if (FRecordManager = nil) then
   raise EACRException.Create(10009,ErrorLNilPointer);
  Result := FRecordManager.GetRecordCount;
 finally
  Unlock;
 end;
end; // GetRecordCount


//------------------------------------------------------------------------------
// move cursor to specified position and set current record id in cursor
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.InternalSetRecNo(Cursor: TACRCursor; RecNo: TACRRecordNo);
var RecordID: TACRRecordNo;
begin
 Lock;
 try
  if (FRecordManager = nil) then
   raise EACRException.Create(10074,ErrorLNilPointer);
  RecordID := TACRMemoryRecordManager(FRecordManager).FindRecord(RecNo-1);
  if (RecordID <> INVALID_ID8) then
   begin
    Cursor.CurrentRecordID.PageNo := RecordID;
    Cursor.CurrentRecordID.PageItemNo := 0;
   end;
 finally
  Unlock;
 end;
end; // SetRecNo


//------------------------------------------------------------------------------
// get current record position from cursor
//------------------------------------------------------------------------------
function TACRMemoryTableData.InternalGetRecNo(Cursor: TACRCursor): TACRRecordNo;
begin
 Result := -1;
 Lock;
 try
  if (FRecordManager = nil) then
   raise EACRException.Create(10033,ErrorLNilPointer);
  Result := TACRMemoryRecordManager(FRecordManager).
                GetTablePositionByRecordID(Cursor.CurrentRecordID.PageNo);
  if (Result = INVALID_ID8) then
   Result := -1;
//  else
//   Inc(Result);
 finally
  Unlock;
 end;
end; // GetRecNo


//------------------------------------------------------------------------------
// LastAutoincValue
//------------------------------------------------------------------------------
function TACRMemoryTableData.LastAutoincValue(FieldNo: Integer; Session: TACRBaseSession): Int64;
var
  id:           TACRObjectID;
begin
  if (FieldNo >= FFieldManager.FieldDefs.Count) then
    raise EACRException.Create(30335, ErrorGFieldWithObjectIdNotFound, [FieldNo]);
  id := FFieldManager.FieldDefs[FieldNo].SequenceDefObjectId;

  Result := FSequenceManager.GetLastVal(Session, id);
end;//LastAutoincValue


//------------------------------------------------------------------------------
// SetLastAutoincValue
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.SetLastAutoincValue(Value: Int64; FieldNo: Integer; Cursor: TACRCursor);
var
  id:           TACRObjectID;
begin
  if (FieldNo >= FFieldManager.FieldDefs.Count) then
    raise EACRException.Create(10575, ErrorGFieldWithObjectIdNotFound, [FieldNo]);

  id := FFieldManager.FieldDefs[FieldNo].SequenceDefObjectId;
  TACRSequenceDef(FSequenceManager.SequenceDefs.GetDefByObjectId(id)).LastValue :=
   Value;
end; // SetLastAutoincValue


//------------------------------------------------------------------------------
// lock table for read and free data for fast navigation
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.LockTableForRead(aSessionID: TACRSessionID);
begin
 Lock;
 try
   inherited;
 finally
   Unlock;
 end;
end; // LockTableForRead


//------------------------------------------------------------------------------
// unlock table for read and free data for fast navigation
//------------------------------------------------------------------------------
procedure TACRMemoryTableData.UnlockTableForRead(aSessionID: TACRSessionID);
begin
 Lock;
 try
   inherited;
 finally
   Unlock;
 end;
end; // UnlockTableForRead


initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRMemEngine> initialized');
{$ENDIF}

end.
