unit ACRDiskEngine;

interface

{$I ACRVer.inc}

uses SysUtils, Classes,
{$IFDEF MSWINDOWS}
     Windows,
{$ENDIF}
{$IFDEF LINUX}
     Types,
     Libc,
{$ENDIF}
     Math,

// Accuracer units

 {$IFDEF DEBUG_LOG}
     ACRDebug,
 {$ENDIF}
{$IFDEF LINUX}
     ACRLinux,
{$ENDIF}
{$IFNDEF D5H}
     ACRD4Routines,
{$ENDIF}
     ACRCriticalSection,
     ACRCrypto,
     ACRExcept,
     ACRBase,
     ACRBaseEngine,
     ACRExpressions,
     ACRMemory,
     ACRPage,
     ACRCompression,
     ACRTypes,
     ACRConverts,
     ACRVariant,
     ACRConst;

type

 // forward decls
 TACRTableMetadataFile = class;
 TACRTableMostUpdatedFile = class;
 TACRTablePFSMapFile = class;
 TACRTableLocksFile = class;
 TACRDiskPageManager = class;
 TACRDiskTableData = class;
 TACRSystemDirectory = class;
 TACRActiveSessionsFile = class;
 TACRTableListFile = class;
 TACRTableLocksManager = class;
 TACRInternalLockedDBFile = class;

////////////////////////////////////////////////////////////////////////////////
//
// TACRDiskDatabaseData
//
////////////////////////////////////////////////////////////////////////////////


  TACRDiskDatabaseData = class (TACRDatabaseData)
   private
    FSystemDir:           TACRSystemDirectory;
    FActiveSessionsFile:  TACRActiveSessionsFile;
    FTableListFile:       TACRTableListFile;
    FLockParams:          TACRLockParams;
    FOptions:             TACROptions;
    FCryptoParams:        TACRCryptoParams;
    FTableLocksManager:   TACRTableLocksManager;

    procedure OpenDatabase(Session: TACRBaseSession);
    procedure CloseDatabase(Session: TACRBaseSession);
    function GetNewSessionID: TACRSessionID;
   public
    constructor Create;
    destructor Destroy; override;
    // database operations
    procedure CreateDatabase(Session: TACRBaseSession); override;
    procedure ConnectSession(Session: TACRBaseSession); override;
    procedure DisconnectSession(Session: TACRBaseSession); override;
    procedure FreeIfNoSessionsConnected; override;
    // find or create table data
    function CreateTableData(Cursor: TACRCursor): TACRTableData; override;
    // database operations
    procedure GetTablesList(List: TStrings); override;
    function GetTableItemIfExists(TableName: String; var TableItem: TACRTableListItem): Boolean;
    function TableExists(TableName: String): Boolean; override;
    // lock tables (virtual object)
    function InternalLockTables: Boolean;
    // unlock tables (virtual object)
    function InternalUnlockTables: Boolean;
    // flush file buffers
    procedure FlushFileBuffers; override;
    // return database format version
    function GetFormatVersion(Session: TACRBaseSession): Double; override;
    // return total number of pages
    function GetTotalPageCount(Session: TACRBaseSession) : Integer; override;
    // return number of free pages
    function GetFreePageCount(Session: TACRBaseSession) : Integer; override;
    // return true if database is encrypted
    function IsDatabaseEncrypted(Session: TACRBaseSession): Boolean; override;
    // return true if database is encrypted by password or by key
    function IsDatabaseEncryptedByPassword(Session: TACRBaseSession): Boolean; override;
    // makes Exe database from edb file
    procedure MakeExeDatabase(Session: TACRBaseSession; ExeFileName, ExeDatabaseFileName: string); override;
    // removes database file from executable database file
    procedure RemoveDatabaseFromExe(Session: TACRBaseSession); override;
    // returns true if this file is an Accuracer database
    function IsAccuracerDatabaseFile(Session: TACRBaseSession): Boolean; override;
    // return true if CryptoParams are valid
    function IsCryptoParamsValid(Session: TACRBaseSession): Boolean; override;
    // fill TableLockList with lock info for the tables opened by this session
    procedure GetTableLocksListForSession(
                                              SessionID:      TACRSessionID;
                                              TableLockList:  TList
                                          ); override;
    // if result is true and PTableLockInfo is not nil - increment corresponding field of this record
    function TryToLockTable(
                        TableName:            String;
                        SessionID:            TACRSessionID;
                        LockType:             TACRLockType;
                        PTableLockInfo:       PACRTableLockInfo = nil): Boolean; override;
    // decrements number of locks by this session and removes all locks if their number = 0
    // if PTableLockInfo = nil - all table locks byt this session will be removed
    function UnlockTableAll(
                            TableName:      String;
                            SessionID:      TACRSessionID;
                            PTableLockInfo: PACRTableLockInfo = nil): Boolean; override;
    procedure RemoveAllLocks(SessionID: TACRSessionID); override;
   public
    property LockParams: TACRLockParams read FLockParams write FLockParams;
    property Options: TACROptions read FOptions write FOptions;
    property CryptoParams: TACRCryptoParams read FCryptoParams write FCryptoParams;
    property TableLocksManager: TACRTableLocksManager read FTableLocksManager;
  end; // TACRDiskDatabaseData


////////////////////////////////////////////////////////////////////////////////
//
// TACRTablePageMapsManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRTablePageMapsManager = class(TObject)
   private
    LTableData:               TACRDiskTableData;
    LPageManager:             TACRDiskPageManager;
    FPFSItemsPerPage:         Word;
    FPageTypeOffset:          Word;
    FTablePFSMapFile:         TACRTablePFSMapFile;
    FRandomSearchRetryCount:  Integer;

   private
    procedure SetPageType(
                SessionID:      TACRSessionID;
                PageNo:         TACRPageNo;
                TablePageType:  TACRTablePageType
                         );
    procedure SetPageRecordCount(
                SessionID:        TACRSessionID;
                PageNo:           TACRPageNo;
                PageRecordCount:  Word;
                PageType:         TACRTablePageType
                         );
    function InternalGetPageType(
                PFSPage:        TACRPage;
                RelativePageNo: TACRPageNo): TACRTablePageType;

    function GetPageRecordCount(
                SessionID:      TACRSessionID;
                PageNo:         TACRPageNo
                         ): Word;

    // find page in PFS page map
    function FindPageInPFSPage(
                      PageData:             PChar;
                      MaxRelativePageNo:    Integer;
                      TablePageType:        TACRTablePageType;
                      Size:                 Word;
                      out RelativePageNo:   Integer;
                      out PageRecordCount:  Word
                              ): Boolean;
   public
    constructor Create(
                       aTableData:                TACRTableData;
                       aRandomSearchRetryCount:   Integer;
                       IsSynchronizationRequired: TACRIsPageSynchronizationRequired
                      );
    destructor Destroy; override;
    // open TablePFSMapFile
    procedure OpenTablePageMapsManager(StartPageNo: TACRPageNo; OpenExclusive: Boolean);
    function AddPage(
                      SessionID:      TACRSessionID;
                      TablePageType:  TACRTablePageType
                    ): TACRPage;
    procedure RemovePage(
                          SessionID:      TACRSessionID;
                          PageNo:         TACRPageNo;
                          TablePageType:  TACRTablePageType
                          );
    // try to find a page for adding a small amount of data (< 1 page)
    // if failed performs add page
    function GetPageForAddingNewData(
                      SessionID:            TACRSessionID;
                      TablePageType:        TACRTablePageType;
                      Size:                 Word;
                      out PageRecordCount:  Word
                                    ): TACRPage;

    function GetFirstRowBeginPageNo(SessionID: TACRSessionID): TACRPageNo;
    function GetLastRowBeginPageNo(SessionID: TACRSessionID): TACRPageNo;
    function GetNextRowBeginPageNo(SessionID: TACRSessionID; PageNo: TACRPageNo): TACRPageNo;
    function GetPriorRowBeginPageNo(SessionID: TACRSessionID; PageNo: TACRPageNo): TACRPageNo;

    function GetRecNoByRecordID(SessionID: TACRSessionID; RecordID: TACRRecordID): TACRRecordNo;
    function GetRecordIDByRecNo(SessionID: TACRSessionID; RecNo: TACRRecordNo): TACRRecordID;
    function GetRecordCount: TACRRecordNo;

    procedure LoadFile(SessionID: TACRSessionID);
    procedure SaveFile(SessionID: TACRSessionID);
    procedure Empty(SessionID: TACRSessionID);

    function SaveState: Pointer;
    procedure RestoreState(State: Pointer);
    procedure FreeState(var State: Pointer);

    procedure GetPageTypeAndRecordCount(
                SessionID:           TACRSessionID;
                PageNo:              TACRPageNo;
                out TablePageType:   TACRTablePageType;
                out PageRecordCount: Word);
   public
    property PFSItemsPerPage: Word read FPFSItemsPerPage;
  end;// TACRTablePageMapsManager



////////////////////////////////////////////////////////////////////////////////
//
// TACRDiskRecordManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRDiskRecordManager = class (TACRBaseRecordManager)
   private
    FDiskRecordBuffer:     TACRRecordBuffer;
    FTablePageMapsManager: TACRTablePageMapsManager;
    LPageManager:          TACRDiskPageManager;
    LTableData:            TACRDiskTableData;
    FDiskRecordBufferSize: Integer;
    FLargeRows:            Boolean;
    FRecordsPerPage:       Integer;

    function LoadDataItemsMapFromPage(
            ErrorCode:            Integer;
            Page:                 TACRPage;
            var ItemCount:        Word
                                      ): TACRDataItemsMap;

    procedure SaveDataItemsMapToPage(
            Page:                 TACRPage;
            ItemsMap:             TACRDataItemsMap;
            ItemCount:            Word
                                      );

    function GetFirstRecordOnPage(Page: TACRPage): TACRPageItemID;
    function GetLastRecordOnPage(Page: TACRPage): TACRPageItemID;
    function GetNextRecordOnPage(Page: TACRPage; PageItemNo: TACRPageItemID): TACRPageItemID;
    function GetPriorRecordOnPage(Page: TACRPage; PageItemNo: TACRPageItemID): TACRPageItemID;

    // return result for attempt of getting record relatively to first position
    // and set RecordID to new record ID
    function GetRecordFromFirstPosition(
            SessionID:            TACRSessionID;
            GetRecordMode:        TACRGetRecordMode;
            var RecordID:         TACRRecordID
                                       ): TACRGetRecordResult;
    // return result for attempt of getting record relatively to last position
    // and set RecordID to new record ID
    function GetRecordFromLastPosition(
            SessionID:            TACRSessionID;
            GetRecordMode:        TACRGetRecordMode;
            var RecordID:         TACRRecordID
                                      ): TACRGetRecordResult;
    // return result for attempt of getting record relatively any position
    // and set RecordID to new record ID
    function GetRecordFromAnyPosition(
            SessionID:            TACRSessionID;
            GetRecordMode:        TACRGetRecordMode;
            var RecordID:         TACRRecordID
                                      ): TACRGetRecordResult;

    // convert data in the memory to disk record buffer
    // including varchars and blobs
    procedure GetDiskRecordBufferFromRecordBuffer(
                    SessionID:     TACRSessionID;
                    RecordBuffer:  TACRRecordBuffer
                                                );

    // copy data from disk record buffer to memory record buffer
    // including varchars and blobs
    procedure CopyDataFromDiskRecordBufferToMemoryRecordBuffer(
                SessionID:        TACRSessionID;
                RecordBuffer:     TACRRecordBuffer);

    procedure InternalWriteBLOBOrVarcharValue(
                  SessionID:        TACRSessionID;
                  BLOBDescriptor:   TACRPartialTemporaryBLOBDescriptor;
                  PData:            PChar;
                  IsVarchar:        Boolean;
                  out RecordID:     TACRRecordID
                                             );

    procedure WriteBLOBOrVarcharValue(
                  SessionID:        TACRSessionID;
                  FieldDef:         TACRFieldDef;
                  RecordBuffer:     TACRRecordBuffer
                                      );

    procedure InternalReadBLOBOrVarcharValue(
                  SessionID:            TACRSessionID;
                  RecordID:             TACRRecordID;
                  IsVarchar:            Boolean;
                  var BLOBDescriptor:   TACRPartialTemporaryBLOBDescriptor;
                  var PData:            PChar;
                  out DataSize:         Integer
                  );

    procedure ReadBLOBOrVarcharValue(
                  SessionID:        TACRSessionID;
                  FieldDef:         TACRFieldDef;
                  RecordBuffer:     TACRRecordBuffer
                                      );

    procedure DeleteBLOBOrVarcharValue(
                  SessionID:        TACRSessionID;
                  FieldDef:         TACRFieldDef
                                      );

    // load record from disk into new allocated buffer
    procedure LoadDiskRecord(
                  SessionID:        TACRSessionID;
                  RecordID:         TACRRecordID;
                  DiskRecordBuffer: TACRRecordBuffer
                           );

    // load record from disk into new allocated buffer
    procedure InternalLoadDiskRecordLarge(
                  SessionID:        TACRSessionID;
                  RecordID:         TACRRecordID;
                  DiskRecordBuffer: TACRRecordBuffer
                           );

    // load record from disk into new allocated buffer
    procedure InternalLoadDiskRecordSmall(
                  SessionID:        TACRSessionID;
                  RecordID:         TACRRecordID;
                  DiskRecordBuffer: TACRRecordBuffer
                           );

    // return true if record exists
    function InternalIsRecordExistsLarge(
                  SessionID:        TACRSessionID;
                  var RecordID:     TACRRecordID
                           ): Boolean;

    // return true if record exists
    function InternalIsRecordExistsSmall(
                  SessionID:        TACRSessionID;
                  var RecordID:     TACRRecordID
                           ): Boolean;

    // add record and return its number (RecordSize > PageDataSize div 2)
    procedure InternalAddRecordLarge(
                       RecordBuffer:  TACRRecordBuffer;
                       var RecordID:  TACRRecordID;
                       SessionID:     TACRSessionID = INVALID_SESSION_ID
                      );

    // add record and return its number (RecordSize <= PageDataSize div 2)
    procedure InternalAddRecordSmall(
                       RecordBuffer:  TACRRecordBuffer;
                       var RecordID:  TACRRecordID;
                       SessionID:     TACRSessionID = INVALID_SESSION_ID
                      );
    // delete record (RecordSize <= PageDataSize div 2)
    procedure InternalDeleteRecordSmall(
                          var RecordID: TACRRecordID;
                          SessionID:    TACRSessionID = INVALID_SESSION_ID
                         );
    // delete record (RecordSize > PageDataSize div 2)
    procedure InternalDeleteRecordLarge(
                          var RecordID: TACRRecordID;
                          SessionID:    TACRSessionID = INVALID_SESSION_ID
                         );
    // update record (RecordSize > PageDataSize div 2)
    procedure InternalUpdateRecordLarge(
                       RecordBuffer:  TACRRecordBuffer;
                       var RecordID:  TACRRecordID;
                       SessionID:     TACRSessionID = INVALID_SESSION_ID
                      );
    // update record (RecordSize <= PageDataSize div 2)
    procedure InternalUpdateRecordSmall(
                       RecordBuffer:  TACRRecordBuffer;
                       var RecordID:  TACRRecordID;
                       SessionID:     TACRSessionID = INVALID_SESSION_ID
                      );
   public
    constructor Create(
                        TablePFSMapFileStartPageNo: TACRPageNo;
                        RecordBufferSize:           Integer;
                        DiskRecordBufferSize:       Integer;
                        aPageManager:               TACRDiskPageManager;
                        aTableData:                 TACRDiskTableData
                      );
    destructor Destroy; override;
    function GetRecordCount: TACRRecordNo; override;
    // empty table (delete all records)
    procedure Empty(SessionID: TACRSessionID = INVALID_SESSION_ID); override;
    // add record and return its number
    function AddRecord(
                       RecordBuffer:  TACRRecordBuffer;
                       var RecordID:  TACRRecordID;
                       SessionID:     TACRSessionID = INVALID_SESSION_ID
                      ): Boolean; override;
    // update record, return true if record was updated, false if record was deleted
    function UpdateRecord(
                          RecordBuffer: TACRRecordBuffer;
                          RecordID:     TACRRecordID;
                          SessionID:    TACRSessionID = INVALID_SESSION_ID
                         ): Boolean; override;
    // delete record, return true if record was deleted, false if record was deleted earlier
    function DeleteRecord(
                          var RecordID: TACRRecordID;
                          SessionID:    TACRSessionID = INVALID_SESSION_ID
                         ): Boolean; override;
    // return true if record exists
    function IsRecordExists(var RecordID: TACRRecordID; SessionID: TACRSessionID = INVALID_SESSION_ID): Boolean; override;
    procedure GetRecordBuffer(var NavigationInfo: TACRNavigationInfo); override;
    // return record no
    function GetApproximateRecNo(RecordID: TACRRecordID; SessionID: TACRSessionID): TACRRecordNo; override;
    function GetRecNoByRecordID(RecordID: TACRRecordID; SessionID: TACRSessionID): TACRRecordNo;
    function GetRecordIDByRecNo(RecNo: TACRRecordNo; SessionID: TACRSessionID): TACRRecordID;
    procedure ReadTablePFSMapFile(SessionID: TACRSessionID);
    procedure WriteTablePFSMapFile(SessionID: TACRSessionID);

    function SaveState: Pointer;
    procedure RestoreState(State: Pointer);
    procedure FreeState(var State: Pointer);

   public
    property RecordsPerPage: Integer read FRecordsPerPage;
  end;// TACRDiskRecordManager


////////////////////////////////////////////////////////////////////////////////
//
// TACRDiskTableData
//
////////////////////////////////////////////////////////////////////////////////


  TACRDiskTableData = class (TACRTableData)
   private
    FRecordID:                     TACRRecordID;
    FLockParams:                   TACRLockParams;
    FRandomSearchRetryCount:       Integer;
    FIsTableOpened:                Boolean;
    //  FTableID is in FTableItem
    FTableItem:                    TACRTableListItem;
    FExclusive:                    Boolean;
    FInTransaction:                Boolean;
    FSavedPFSMapState:             Pointer;
    LPageManager:                  TACRDiskPageManager;
    LDiskDatabaseData:             TACRDiskDatabaseData;
    // Table metadata files
    FTableMetadataFile:            TACRTableMetadataFile;
    FTableMostUpdatedFile:         TACRTableMostUpdatedFile;


   protected
    function GetTableID: TACRObjectID; override;
    function GetPageManager: TACRPageManager; override;
    procedure CreateRecordManager(TablePFSMapFileStartPageNo: TACRPageNo = INVALID_PAGE_NO); override;
   private
    //----------------------- locking methods ----------------------------------
    function InternalLockTableX(const CurrentRetryNo: Integer): Boolean;
    function InternalUnlockTableX: Boolean;
    function InternalLockTableIS(const CurrentRetryNo: Integer): Boolean;
    function InternalUnlockTableIS: Boolean;
    function InternalLockTableS(const CurrentRetryNo: Integer): Boolean;
    function InternalUnlockTableS: Boolean;
    function InternalLockTableIRW(const CurrentRetryNo: Integer): Boolean;
    function InternalUnlockTableIRW: Boolean;
    function InternalLockTableRW(const CurrentRetryNo: Integer): Boolean;
    function InternalUnlockTableRW: Boolean;
    function LockRecordU(const CurrentRetryNo: Integer): Boolean;
    function UnlockRecordU: Boolean;
    //-------------------- end of locking methods ------------------------------

    procedure SetSessionID(SessionID: TACRSessionID);

    procedure ReadMostUpdatedData(SessionID: TACRSessionID);
    procedure WriteMostUpdatedData(SessionID: TACRSessionID);

    procedure ReadTablePFSMapFile(SessionID: TACRSessionID);
    procedure WriteTablePFSMapFile(SessionID: TACRSessionID);
    procedure SaveTablePFSMapFileState;
    procedure RestoreTablePFSMapFileState;
    procedure FreeTablePFSMapFileState;

    procedure ReadTableMetadata(
                SessionID:        TACRSessionID;
                aStartPageNo:     TACRPageNo);
    procedure WriteTableMetadata(
                SessionID:        TACRSessionID;
                aStartPageNo:     TACRPageNo);
    procedure InternalDeleteTable(Session: TACRBaseSession; Cascade: Boolean; LockTables: Boolean = True);
   protected
    procedure WriteMetadataChanges(SessionID: TACRSessionID); override;
   public
    constructor Create(aDiskDatabaseData: TACRDiskDatabaseData);
    destructor Destroy; override;

    function GetRecordSize: Integer;
    // table operations
    procedure CreateTable(
                          Cursor:         TACRCursor;
                          FieldDefs:      TACRFieldDefs;
                          IndexDefs:      TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                         ); override;
    procedure DeleteTable(Session: TACRBaseSession; Cascade: Boolean; DesignMode: Boolean = False); override;
    procedure AddForeignKey(Cursor: TACRCursor; ConstraintDef: TACRConstraintDefForeignKey); override;
    procedure DeleteConstraint(Cursor: TACRCursor; Name: String; Cascade: Boolean; FKPartialDelete: Boolean); override;
    procedure RenameReferenceTableName(Cursor: TACRCursor; OldName, NewName: String); override;

    procedure EmptyTable(Cursor: TACRCursor); override;
    procedure RenameTable(Cursor: TACRCursor; NewTableName: String); override;
   private
    procedure InternalOpenTable;
    procedure InternalCloseTable;
   public
    procedure OpenTable(Cursor: TACRCursor); override;
    procedure CloseTable(Cursor: TACRCursor); override;

    // Rename Field by Field Index in FieldDefs
    procedure RenameField(Cursor: TACRCursor; FieldName, NewFieldName: String); override;

    procedure AddIndex(IndexDef: TACRIndexDef; Cursor: TACRCursor); override;
    procedure DeleteIndex(IndexID: TACRObjectID; Cursor: TACRCursor); override;
    procedure DeleteAllIndexes(Cursor: TACRCursor); override;

    // return true if record was found and is visible by cursor
    // prepares params for FindRecordByScan and calls it
    function FindRecord(
                        Cursor:                 TACRCursor;
                        SearchExpression:       TACRExpression; // locate
                        KeyCondition:           TACRScanSearchCondition; // find key
                        // find first or last record in order specified by index settings
                        Restart:                Boolean;
                        // go from first record to last or vice versa
                        GoForward:              Boolean;
                        // if record was found then return record id
                        ResultRecordID:         PACRRecordID;
                        // if specified - fill record bitmap
                        RecordBitmap:           TACRRecordBitmap = nil;
                        InternalCall:           Boolean = False;
                        StopAtFirstFoundRecord: Boolean = False
                       ): Boolean; override;
    function IsRecordExists(Cursor: TACRCursor): Boolean; override;
    // return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
    function CompareRecordID(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer; override;
    // read record with Distinct, SQLFilter, SQLTopRowCount, Filter, Range, OnFilterRecord
    function GetRecordBuffer(
                              Cursor:         TACRCursor;
                              GetRecordMode:  TACRGetRecordMode
                            ): TACRGetRecordResult; override;

    // check constraint conditions
    procedure CheckConstraints(
                                Cursor:          TACRCursor;
                                SessionID:       TACRSessionID;
                                NewRecordBuffer: TACRRecordBuffer;
                                OldRecordBuffer: TACRRecordBuffer;
                                ToInsert:        Boolean;
                                CurrentRecordID: TACRRecordID;
                                SkipFKCheck:     Boolean = False
                              ); override;
    procedure ApplyDefaultValuesToRecordBuffer(
                      Cursor:       TACRCursor;
                      RecordBuffer: TACRRecordBuffer;
                      InsertMode:   Boolean
                                              ); override;
    function InsertRecord(var Cursor: TACRCursor): Boolean; override;
    function DeleteRecord(Cursor: TACRCursor): Boolean; override;
    function UpdateRecord(Cursor: TACRCursor): Boolean; override;
    procedure EditRecord(Cursor: TACRCursor); override;
    procedure CancelRecord(Cursor: TACRCursor; ToInsert: Boolean); override;
    procedure DeleteVisibleRecords(Cursor: TACRCursor); override;
    procedure UpdateVisibleRecords(
                                    Cursor:       TACRCursor;
                                    FieldNames:   TStringList;
                                    values:       array of TACRVariant;
                                    SkipFKCheck:  Boolean = False
                                   ); override;
    function InternalCreateBlobStream(
              Cursor:   TACRCursor;
              ToInsert: Boolean;
              FieldNo:  Integer;
              OpenMode: TACRBLOBOpenMode
              ): TACRStream; override;
    procedure ClearBLOBFieldInRecordBuffer(
              var RecordBuffer: TACRRecordBuffer;
              FieldNo:    Integer
              ); override;

    // record operations
    function GetRecordCount(Cursor: TACRCursor; InternalCall: Boolean = False): TACRRecordNo; override;
    procedure SetRecNo(Cursor: TACRCursor; RecNo: TACRRecordNo); override;
    function GetRecNo(Cursor: TACRCursor): TACRRecordNo; override;
    function InternalGetRecordCount: TACRRecordNo; override;
    // move cursor to specified position and set current record id in cursor
    procedure InternalSetRecNo(Cursor: TACRCursor; RecNo: TACRRecordNo); override;
    // get current record position from cursor
    function InternalGetRecNo(Cursor: TACRCursor): TACRRecordNo; override;
    // return size of the bitmap
    function GetBitmapSize(Session: TACRSessionID): TACRRecordNo; override;
    // return filter bitmap rec no by record id
    function GetBitmapRecNoByRecordID(RecordID: TACRRecordID): TACRRecordNo; override;
    // return filter bitmap rec no by record id
    function GetRecordIDByBitmapRecNo(RecordNo: TACRRecordNo): TACRRecordID; override;

    function LastAutoincValue(FieldNo: Integer; Session: TACRBaseSession): Int64; override;
    procedure SetLastAutoincValue(Value: Int64; FieldNo: Integer; Cursor: TACRCursor); override;

    procedure LockTableForRead(aSessionID: TACRSessionID); override;
    procedure UnlockTableForRead(aSessionID: TACRSessionID); override;

    function TryToLockTableS: Boolean; override;
    function TryToUnlockTableS: Boolean; override;
    function TryToLockTableIRW: Boolean; override;
    function TryToUnlockTableIRW: Boolean; override;
    function TryToLockTableRW: Boolean; override;
    function TryToUnlockTableRW: Boolean; override;
    function TryToLockTableX: Boolean; override;
    function TryToUnlockTableX: Boolean; override;
    function TryToLockTableIS: Boolean; override;
    function TryToUnlockTableIS: Boolean; override;

    procedure StartTransaction(aSessionID: TACRSessionID); override;
    procedure Commit(aSessionID: TACRSessionID); override;
    procedure Rollback(aSessionID: TACRSessionID); override;

    function IsSynchronizationRequired(PageState: Cardinal): Boolean; override;

   public
    property SessionID: TACRSessionID read FSessionID write SetSessionID;
    property LockParams: TACRLockParams read FLockParams write FLockParams;
//    property CryptoParams: TACRCryptoParams read FCryptoParams write FCryptoParams;
    property Exclusive: Boolean read FExclusive;
  end; // TACRDiskTableData

  
////////////////////////////////////////////////////////////////////////////////
//
// TACRDatabaseFile
//
////////////////////////////////////////////////////////////////////////////////

  TACRDatabaseFile = class(TObject)
   private
    FFileName:  String;

    FAccessMode:  TACRAccessMode;
    FShareMode:   TACRShareMode;
    FAttrFlags:   DWORD;

{$IFDEF MSWINDOWS}
    FHandle:      THandle;
{$ENDIF}
{$IFDEF LINUX}
    FHandle:      Integer;
    FSize:        Int64;
{$ENDIF}

    FIsOpened:    Boolean;
   private
    function ConvertAccessMode(am: TACRAccessMode): DWORD;
    function ConvertShareMode(sm: TACRShareMode): DWORD;

    // if file closed then raise
    procedure CheckOpened(OperationName: String);
    // if file opened then raise
    procedure CheckClosed(OperationName: String);
    // Seek
    procedure Seek(Offset: Int64; ErrorCode: Integer = 0);

    procedure SetPosition(Offset: Int64);
    function  GetPosition: Int64;

    procedure SetSize(NewSize: Int64);
    function  GetSize: Int64;
   public
    // Constructor
    constructor Create;
    // Destructor
    destructor Destroy; override;

    // Create and Open File
    procedure CreateAndOpenFile(FileName: String);
    // Delete File
    procedure DeleteFile(FileName: String);
    // Rename Closed File
    procedure RenameFile(OldFileName, NewFileName: String);

    // Open File
    procedure OpenFile(FileName: String; var AccessMode: TACRAccessMode; var ShareMode: TACRShareMode);
    // Close File
    procedure CloseFile;

    // Read Buffer
    procedure ReadBuffer(var Buffer; const Count: Int64; const Pos: Int64; ErrorCode: Integer);
    // Write Buffer
    procedure WriteBuffer(const Buffer; const Count: Int64; const Pos: Int64; ErrorCode: Integer);
    // Flush Buffers
    procedure FlushFileBuffers;

    // Lock Byte (return TRUE if success)
    function LockByte(Offset: Int64; Count: Integer = 1): Boolean;
    // Unlock Byte
    function UnlockByte(Offset: Int64; Count: Integer = 1): Boolean;
    // return TRUE if byte Locked
    function IsByteLocked(Offset: Int64): Boolean;
    // return FALSE if any byte of region is locked
    function IsRegionLocked(Offset: Int64; Count: Integer): Boolean;
   public
    property FileName: String read FFileName;
    property Size: Int64 read GetSize write SetSize;
    property IsOpened: Boolean read FIsOpened;
    property Position: Int64 read GetPosition write SetPosition;
  end;// TACRDatabaseFile



////////////////////////////////////////////////////////////////////////////////
//
// TACRDatabaseFreeSpaceManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRFreeSpaceManagerPageType = (fsmPFS,fsmGAM,fsmSGAM);
{$IFDEF FREE_SPACE_MANAGER_ON}
  TACRDatabaseFreeSpaceManager = class(TObject)
   private
    LPageManager: TACRDiskPageManager;
    FLockParams:   TACRLockParams;
    PFS:        PChar; // Page Free Space bits
    GAM:        PChar; // Global Allocation Map bits
    SGAM:       PChar; // Shared Global Allocation Map bits
    PagePerExt: Integer;
    PFSCount: TACRPageNo;
    GAMCount: TACRPageNo;
   public
    constructor Create(
                        aPageManager: TACRDiskPageManager;
                        aLockParams:  TACRLockParams
                      );
    destructor Destroy; override;
    procedure CreateFreeSpaceManager;
   private
    procedure LoadPage(IPageNo: Integer; IPageType: TACRFreeSpaceManagerPageType);
    procedure SavePage(IPageNo: Integer; IPageType: TACRFreeSpaceManagerPageType);
    function AppendNewPage (CurGAM: integer; PFSPageNo, EmptyExt : TACRPageNo;
                                                 PFSNeed : Boolean): TACRPageNo;
    procedure CheckShrinking;
   public
    function GetPage(DesiredStartPageNo: TACRPageNo = INVALID_PAGE_NO): TACRPageNo;
    procedure FreePage(PageNo: TACRPageNo);
   public
    function GetFreePageCount: TACRPageNo;
{
    procedure GetPages(
           PageCount:          TACRPageNo;
           var Pages:          TACRPagesArray;
           bUniform:           Boolean;
           DesiredStartPageNo: TACRPageNo
                     ); virtual; abstract;
    procedure FreePages(Pages: TACRPagesArray); virtual; abstract;
}
   public
    property LockParams: TACRLockParams read FLockParams write FLockParams;
  end;// TACRDatabaseFreeSpaceManager
{$ELSE}
  TACRDatabaseFreeSpaceManager = class(TObject)
   private
    LPageManager: TACRDiskPageManager;
    FLockParams:   TACRLockParams;
   public
    constructor Create(
                        aPageManager: TACRDiskPageManager;
                        aLockParams:  TACRLockParams
                      );
    procedure CreateFreeSpaceManager;
    function GetPage(DesiredStartPageNo: TACRPageNo = INVALID_PAGE_NO): TACRPageNo;
    procedure FreePage(PageNo: TACRPageNo);

   public
    property LockParams: TACRLockParams read FLockParams write FLockParams;
  end;// TACRDatabaseFreeSpaceManager
{$ENDIF}


////////////////////////////////////////////////////////////////////////////////
//
// TACRTableLocksManager
//
////////////////////////////////////////////////////////////////////////////////

 TACRTableLockItem = packed record
   TableName:         String;
   TableLocksFile:    TACRTableLocksFile;
   SessionLocksList:  TList;
 end;
 PACRTableLockItem = ^TACRTableLockItem;

  TACRTableLocksManager = class (TObject)
   private
    FCSect:             TRTLCriticalSection;
    LDiskDatabaseData:  TACRDiskDatabaseData;
    FTableLockItemList: TList;
    FInMemory:          Boolean;
   private
    // lock
    procedure Lock;
    // unlock
    procedure Unlock;
   protected
    procedure InitTableLockInfo(var TableLockInfo: TACRTableLockInfo);
    function IsLockInfoEmpty(var TableLockInfo: TACRTableLockInfo): Boolean;
    function GetTableLockCount(var TableLockInfo: TACRTableLockInfo; LockType: TACRLockType): Integer;
    procedure IncTableLockCount(var TableLockInfo: TACRTableLockInfo; LockType: TACRLockType);
    procedure DecTableLockCount(var TableLockInfo: TACRTableLockInfo; LockType: TACRLockType; Decrement: Integer = 1);
    procedure ClearTableLockCount(var TableLockInfo: TACRTableLockInfo; LockType: TACRLockType);

    function AddSession(PTableLockItem: PACRTableLockItem; SessionID: TACRSessionID): Integer;
    function FindSession(PTableLockItem: PACRTableLockItem; SessionID: TACRSessionID): Integer;
    function FindOrCreateSession(PTableLockItem: PACRTableLockItem; SessionID: TACRSessionID): Integer;
    procedure RemoveSession(PTableLockItem: PACRTableLockItem; SessionItem: PACRSessionLockItem);
    function CreateTableLockItem(TableName: String; SessionID: TACRSessionID): Integer;
    procedure DestroyTableLockItem(PTableLockItem: PACRTableLockItem);
    function FindTableLockItem(TableName: String): Integer;
    function FindOrCreateTableLockItem(TableName: String; SessionID: TACRSessionID): Integer;
    // perform locking of the byte in physical database file if other lock applied
    // by other applications / threads allows it
    function InternalLockTable(
                        const CurrentRetryNo: Integer;
                        var TableItem:        TACRTableListItem;
                        PTableLockItem:       PACRTableLockItem;
                        SessionID:            TACRSessionID;
                        LockType:             TACRLockType;
                        PTableLockInfo:       PACRTableLockInfo = nil): Boolean;
    // return true if lock specified by LockType is compatible with session lock item
    function IsTableLockAllowed(
                        PSessionLockItem: PACRSessionLockItem;
                        LockType:         TACRLockType): Boolean;
    // perform unlocking of the byte in physical database file
    function InternalUnlockTable(
                        PTableLockItem: PACRTableLockItem;
                        SessionID:      TACRSessionID;
                        LockType:       TACRLockType
                        ): Boolean;
   public
    constructor Create(aDiskDatabaseData: TACRDiskDatabaseData);
    destructor Destroy; override;

    // if result is true and PTableLockInfo is not nil - increment corresponding field of this record
    function LockTable(
                        const CurrentRetryNo: Integer;
                        var TableItem:        TACRTableListItem;
                        TableName:            String;
                        SessionID:            TACRSessionID;
                        LockType:             TACRLockType;
                        PTableLockInfo:       PACRTableLockInfo = nil): Boolean;

    // if result is true and PTableLockInfo is not nil - decrement corresponding field of this record
    // if ForceUnlock is true - unlocks all locks set by this session
    function UnlockTable(
                          TableName:      String;
                          SessionID:      TACRSessionID;
                          LockType:       TACRLockType;
                          PTableLockInfo: PACRTableLockInfo = nil;
                          ForceUnlock:    Boolean = False
                         ): Boolean;
    // decrements number of locks by this session and removes all locks if their number = 0
    // if PTableLockInfo = nil - all table locks by this session will be removed
    function UnlockTableAll(
                            TableName:      String;
                            SessionID:      TACRSessionID;
                            PTableLockInfo: PACRTableLockInfo = nil): Boolean;
    function LockRecord(
                        TableName:      String;
                        SessionID:  TACRSessionID;
                        RecordID:   TACRRecordID
                       ): Boolean;
    function UnlockRecord(
                        TableName:      String;
                        SessionID:  TACRSessionID;
                        RecordID:   TACRRecordID
                         ): Boolean;
    function IsRecordLocked(
                        TableName:      String;
                        SessionID:  TACRSessionID;
                        RecordID:   TACRRecordID
                         ): Boolean;
    // fill TableLockList with lock info for the tables opened by this session
    procedure GetTableLocksListForSession(
                                              SessionID:      TACRSessionID;
                                              TableLockList:  TList
                                          );
    procedure RemoveAllLocksForSession(SessionID: TACRSessionID);
    {$IFDEF DEBUG_TRACE_TABLE_LOCKS}
    procedure WriteCurrentLocksToLog(Caption: String = '');
    {$ENDIF}
   public
    property InMemory: Boolean read FInMemory write FInMemory;
  end; // TACRTableLocksManager


////////////////////////////////////////////////////////////////////////////////
//
// TACRDiskPageManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRDiskPageManager = class (TACRPageManager)
   private
    FDatabaseFreeSpaceManager:  TACRDatabaseFreeSpaceManager;
    FDatabaseFile:              TACRDatabaseFile;
    FDBHeader:                  TACRDBHeader;
    FLockedBytes:               TACRLockedBytes;
    FOffsetToFirstPage:         Int64;
    FOffsetToLockedBytes:       Int64;
    FOffsetToDBHeader:          Int64;
    FLockParams:                TACRLockParams;
    FCryptoInfo:                TACRCryptoInfo;
    FReadOnly:                  Boolean;
   private
    procedure InitDBHeader;
    procedure LoadDBHeader;
    procedure SaveDBHeader;
    function GetIsOpened: Boolean;
    function GetPageOffset(PageNo: TACRPageNo): Int64;
   protected
    function GetPageCount: TACRPageNo; override;
   public
    function UpdateState: Cardinal; override;
    function GetState: Cardinal; override;
    procedure ReadPageRegion(var Buffer; PageNo: TACRPageNo; Offset, Count: Integer);
    procedure WritePageRegion(const Buffer; PageNo: TACRPageNo; Offset, Count: Integer);

    function DirectAddPage: TACRPageNo;
    procedure InternalAddPage(aPage: TACRPage); override;
    procedure InternalRemovePage(PageNo: TACRPageNo); override;
    procedure InternalReadPage(aPage: TACRPage); override;
    procedure InternalWritePage(aPage: TACRPage); override;

    // lock Free Space Manager byte
    function LockFreeSpaceManager: Boolean;
    // unlock Free Space Manager byte
    function UnlockFreeSpaceManager: Boolean;
    // lock Tables byte
    function InternalLockTables: Boolean;
    // unlock Tables byte
    function InternalUnlockTables: Boolean;
    // Lock Byte (return TRUE if success)
    function LockPageByte(PageNo: TACRPageNo; Offset: Word): Boolean;
    // Unlock Byte
    function UnlockPageByte(PageNo: TACRPageNo; Offset: Word): Boolean;
    // return True if byte is locked
    function IsPageByteLocked(PageNo: TACRPageNo; Offset: Word): Boolean;
    // return True if any byte of region is locked
    function IsPageRegionLocked(PageNo: TACRPageNo; Offset: Word; Count: Word): Boolean;
   public
    constructor Create(aLockParams: TACRLockParams);
    destructor Destroy; override;
    procedure CreateAndOpenDatabase(
                              DatabaseFileName: String;
                              PageSize:         Word = ACRDefaultPageSize;
                              ExtentPageCount:  Word = ACRDefaultExtentPageCount
                              );
    procedure DeleteDatabase(DatabaseFileName: String);
    procedure RenameDatabase(
                                             DatabaseFileName,
                                             NewDatabaseName: String
                            );
    function GetExists(DatabaseFileName: String): Boolean;
    procedure OpenDatabase(
                DatabaseFileName:     String;
                var aReadOnly:        Boolean;
                aExclusive:           Boolean = False;
                DoNotCheckPassword:   Boolean = False
                          );
    procedure CloseDatabase;
    procedure FlushFileBuffers; override;
    // extend file by number of pages specified by PageCount

    procedure ExtendFile(PageCount: TACRPageNo);
    procedure SetPageCount(LastPageNo: TACRPageNo);
    function IsDBHeaderValid(Offset: Int64): Boolean;
    function GetOffsetToDBHeader: Int64;
    // returns true if this file is an Accuracer database
    function IsAccuracerDatabaseFile(DatabaseFileName: String): Boolean;
    procedure RemoveDatabaseFromExe(DatabaseFileName: String);
   public
    property DatabaseFile: TACRDatabaseFile read FDatabaseFile;
    property DBHeader: TACRDBHeader read FDBHeader;
    property LockedBytes: TACRLockedBytes read FLockedBytes;
    property IsOpened: Boolean read GetIsOpened;
    property OffsetToFirstPage: Int64 read FOffsetToFirstPage;
    property OffsetToLockedBytes: Int64 read FOffsetToLockedBytes;
    property OffsetToDBHeader: Int64 read FOffsetToDBHeader;
    property LockParams: TACRLockParams read FLockParams write FLockParams;
    property ReadOnly: Boolean read FReadOnly;
    property CryptoInfo: TACRCryptoInfo read FCryptoInfo write FCryptoInfo;
  end; // TACRDiskPageManager


////////////////////////////////////////////////////////////////////////////////
//
// TACRInternalDBFile
//
////////////////////////////////////////////////////////////////////////////////

  TACRInternalLockedDBFile = class(TObject)
   private
    LPageManager: TACRDiskPageManager;
    FStartPageNo: TACRPageNo;
    FPageTypeID:  TACRPageTypeID;
    FSize:        Int64;
    FPages:       TACRIntegerArray;
   private
    // raise if file is Closed
    procedure CheckFileOpened(OperationName: String);
    function GetPageOffset(Position: Int64): Word;
    function GetPageNo(Position: Int64): Integer;
   public
    // Constructor
    constructor Create(PageManager: TACRDiskPageManager; PageTypeID: TACRPageTypeID);
    // Destructor
    destructor Destroy; override;

    // Create file
    procedure CreateFile(FileSize: Integer; ObjectID: TACRObjectID = ACRObjectIDUnknown);
    // Delete file. And free all file pages
    procedure DeleteFile(StartPageNo: TACRPageNo);
    // Open file (reading file header)
    procedure OpenFile(StartPageNo: TACRPageNo);

    // direct (without sessions and transactions) Read buffer from file
    procedure DirectReadBuffer(var Buffer; const Count: Integer; const Position: Integer);
    // direct (without sessions and transactions) Write buffer from file
    procedure DirectWriteBuffer(const Buffer; const Count: Integer; const Position: Integer);

    function LockFile: Boolean;
    function UnlockFile: Boolean;
    function IsFileLocked: Boolean;
    function LockByte(ByteNo: Integer): Boolean;
    function UnlockByte(ByteNo: Integer): Boolean;
    function IsByteLocked(ByteNo: Integer): Boolean;
    function IsRegionLocked(const Position: Integer; const Count: Integer): Boolean;
   public
    property PageManager: TACRDiskPageManager read LPageManager;
    property StartPageNo: TACRPageNo read FStartPageNo;
    property Size: Int64 read FSize;
  end;// TACRInternalLockedDBFile


  TACRInternalDBFile = class(TObject)
   private
    LPageManager:               TACRDiskPageManager;
    FStartPageNo:               TACRPageNo;
    FPageTypeID:                TACRPageTypeID;
    FExclusive:                 Boolean;
    FIsLoaded:                  Boolean;
    FIsSynchronizationRequired: TACRIsPageSynchronizationRequired;
   protected
    FFileData:    TACRFileData;
   private
    procedure InitFile;
    // raise if file is Closed
    procedure CheckFileOpened(OperationName: String);
   public
    // Constructor
    constructor Create(PageManager:               TACRDiskPageManager;
                       PageTypeID:                TACRPageTypeID;
                       IsSynchronizationRequired: TACRIsPageSynchronizationRequired = nil);
    // Destructor
    destructor Destroy; override;

    procedure ResizeFileData(NewSize: Integer);
    // Create file
    procedure CreateFile(SessionID: TACRSessionID; ObjectID: TACRObjectID = ACRObjectIDUnknown);
    // Delete file. And free all file pages
    procedure DeleteFile(SessionID: TACRSessionID; StartPageNo: TACRPageNo = INVALID_PAGE_NO);
    // Open file (reading file header)
    procedure OpenFile(StartPageNo: TACRPageNo; OpenExclusive: Boolean = False);

    procedure ReadFile(SessionID: TACRSessionID);
    procedure WriteFile(SessionID: TACRSessionID);
    procedure EmptyFile(SessionID: TACRSessionID);
   public
    property PageManager: TACRDiskPageManager read LPageManager;
    property StartPageNo: TACRPageNo read FStartPageNo;
    property Exclusive: Boolean read FExclusive;
    property FileData: TACRFileData read FFileData write FFileData;
  end;// TACRInternalDBFile



////////////////////////////////////////////////////////////////////////////////
//
// TACRActiveSessionsFile
//
////////////////////////////////////////////////////////////////////////////////

  TACRActiveSessionsFile = class(TObject)
   private
    LPageManager:     TACRDiskPageManager;
    FHandle:          TACRInternalLockedDBFile;
    FLockParams:      TACRLockParams;

   public
    constructor Create(
                        PageManager: TACRPageManager;
                        aLockParams: TACRLockParams
                       );
    destructor Destroy; override;
    function CreateFile(aMaxSessionCount: Integer): TACRPageNo;
    procedure OpenFile(aStartPageNo: TACRPageNo);
    procedure CloseFile;
    function Connect: TACRSessionID;
    procedure Disconnect(SessionID: TACRSessionID);
   public
    property LockParams: TACRLockParams read FLockParams write FLockParams;
  end;// TACRActiveSessionsFile



////////////////////////////////////////////////////////////////////////////////
//
// TACRTableListFile
//
////////////////////////////////////////////////////////////////////////////////

  TACRTableListFile = class(TObject)
   private
    LPageManager: TACRDiskPageManager;
    FHandle:      TACRInternalDBFile;
    FTableList: array of TACRTableListItem;
    procedure Load;
    procedure Save;
   public
    function GetTableItem(TableName: String): PACRTableListItem;
    function GetTableItemIfExists(TableName: String; var TableItem: TACRTableListItem): Boolean;
    constructor Create(PageManager: TACRPageManager);
    destructor Destroy; override;

    function CreateFile: TACRPageNo;
    procedure OpenFile(aStartPageNo: TACRPageNo);

    function IsTableExist(TableName: String): Boolean;
    function CreateTable(TableItem: TACRTableListItem):TACRTableID;
    procedure DeleteTable(TableName: String);
    procedure RenameTable(OldTableName, NewTableName: String);
    procedure GetTablesList(List: TStrings);
    function GetNextTableID: TACRObjectID;
   public
    property Items[Name: String]: PACRTableListItem read GetTableItem;
  end;// TACRTableListFile



////////////////////////////////////////////////////////////////////////////////
//
// TACRTableMetadataFile
//
////////////////////////////////////////////////////////////////////////////////

  TACRTableMetadataFile = class(TObject)
   private
    LPageManager:     TACRDiskPageManager;
    FHandle:          TACRInternalDBFile;
   public
    constructor Create(
                       PageManager: TACRPageManager
                      );
    destructor Destroy; override;
   public
    // Create File
    function CreateFile(
              SessionID:        TACRSessionID;
              ObjectID:         TACRObjectID
                        ): TACRPageNo;

    // Delete File
    procedure DeleteFile(
              SessionID:        TACRSessionID;
              aStartPageNo: TACRPageNo
                        );
    procedure EmptyFile(SessionID: TACRSessionID);
    // Read Buffer
    procedure ReadFile(
              SessionID:        TACRSessionID;
              aStartPageNo:     TACRPageNo;
              ms:               TACRStream
                        );
    // Write Buffer
    procedure WriteFile(
              SessionID:        TACRSessionID;
              aStartPageNo:     TACRPageNo;
              ms:               TACRStream
                        );
  end;// TACRTableMetadataFile



////////////////////////////////////////////////////////////////////////////////
//
// TACRTableMostUpdatedFile
//
////////////////////////////////////////////////////////////////////////////////

  TACRTableMostUpdatedFile = class(TObject)
   private
    LPageManager:     TACRDiskPageManager;
    FHandle:          TACRInternalDBFile;
 public
    constructor Create(
                       PageManager:               TACRPageManager;
                       IsSynchronizationRequired: TACRIsPageSynchronizationRequired
                      );
    destructor Destroy; override;
   public
    // Create File
    function CreateFile(
              SessionID:        TACRSessionID;
              ObjectID:         TACRObjectID
                        ): TACRPageNo;

    procedure OpenFile(
              aStartPageNo:   TACRPageNo;
              OpenExclusive:  Boolean
              );
    // Delete File
    procedure DeleteFile(
              SessionID:    TACRSessionID;
              aStartPageNo: TACRPageNo
                        );
    // Read Buffer
    procedure ReadFile(SessionID: TACRSessionID);
    // Write Buffer
    procedure WriteFile(SessionID: TACRSessionID);
  end;// TACRTableMostUpdatedFile



////////////////////////////////////////////////////////////////////////////////
//
// TACRTableLocksFile
//
////////////////////////////////////////////////////////////////////////////////

  TACRTableLocksFile = class(TACRInternalDBFile)
   private
    LPageManager:     TACRDiskPageManager;
    FHandle:          TACRInternalLockedDBFile;
    FMaxSessionCount: Integer;
    FLockParams:      TACRLockParams;
    FLockDelayOn:     Boolean;
    FLockDelayTime:   Cardinal;

    function GetStartPageNo: TACRPageNo;
    function GetMaxSessionCount: Integer;
    function GetByteNo(
                        SessionID:    TACRSessionID;
                        LockType:     TACRLockType;
                        UseRecordID:  Boolean = False
                      ): Integer;
   public
    procedure OpenFile(aStartPageNo: TACRPageNo);
    constructor Create(
                       PageManager: TACRPageManager;
                       aLockParams: TACRLockParams
                      );
    destructor Destroy; override;
    function CreateFile(aMaxSessionCount: Integer; ObjectID: TACRObjectID): TACRPageNo;
    procedure DeleteFile(
                          SessionID:        TACRSessionID;
                          aStartPageNo: TACRPageNo
                         );

    //------------------- new locking methods for TableLocksManager ------------------

    function LockTableIfLockAllowed(
                          const CurrentRetryNo: Integer;
                          SessionID: TACRSessionID;
                          LockType: TACRLockType): Boolean;
    function UnlockTableIfPossible(SessionID: TACRSessionID; LockType: TACRLockType): Boolean;

    function LockRecordIfPossible(SessionID: TACRSessionID; RecordID: TACRRecordID): Boolean;
    function UnlockRecordIfPossible(SessionID: TACRSessionID; RecordID: TACRRecordID): Boolean;
    function CheckIfRecordIsLocked(SessionID: TACRSessionID; RecordID: TACRRecordID): Boolean;

   public
    property StartPageNo: TACRPageNo read GetStartPageNo;
    property MaxSessionCount: Integer read FMaxSessionCount;
    property LockParams: TACRLockParams read FLockParams write FLockParams;
  end;// TACRTableLocksFile



////////////////////////////////////////////////////////////////////////////////
//
// TACRTablePFSMapFile
//
////////////////////////////////////////////////////////////////////////////////
  TACRTablePFSMapState = record
    PageCount:             Integer;
    TablePFSPageCount:     Integer;
    TablePFSPageMapItems:  array of TACRTablePFSPageMapItem;
  end;
  PACRTablePFSMapState = ^TACRTablePFSMapState;

  TACRTablePFSMapFile = class(TObject)
   private
    LPageManager:          TACRDiskPageManager;
    FHandle:               TACRInternalDBFile;
   protected 
    PageCount:             Integer;
    TablePFSPageCount:     Integer;
    TablePFSPageMapItems:  array of TACRTablePFSPageMapItem;

   public
    procedure LoadFile(SessionID: TACRSessionID);
    procedure SaveFile(SessionID: TACRSessionID);
    procedure OpenFile(aStartPageNo: TACRPageNo; OpenExclusive: Boolean);
    constructor Create(
                       aPageManager:              TACRPageManager;
                       IsSynchronizationRequired: TACRIsPageSynchronizationRequired
                      );
    destructor Destroy; override;
    function CreateFile(ObjectID: TACRObjectID): TACRPageNo;
    procedure DeleteFile(
                          SessionID:        TACRSessionID;
                          aStartPageNo: TACRPageNo
                         );
    procedure EmptyFile(SessionID: TACRSessionID);

    procedure AddPage(SessionID: TACRSessionID; PageNo: TACRPageNo);
    function GetTableRecordCount: TACRRecordNo;

    function SaveState: Pointer;
    procedure RestoreState(State: Pointer);
    procedure FreeState(var State: Pointer);
  end;// TACRTablePFSMapFile



////////////////////////////////////////////////////////////////////////////////
//
// TACRSystemDirectory
//
////////////////////////////////////////////////////////////////////////////////


  TACRSystemDirectory = class(TObject)
   private
    LPageManager: TACRDiskPageManager;
    FHandle:      TACRInternalDBFile;

    FFileList:    array of TACRSystemDirectoryListItem;
   public
    constructor Create(PageManager: TACRDiskPageManager);
    destructor Destroy; override;
    procedure CreateDirectory;
    procedure LoadDirectory;
    procedure SaveDirectory;
    procedure CreateFile(FileType: TACRDBFileType; FirstPageNo: TACRPageNo);
    function GetFileFirstPageNo(FileType: TACRDBFileType): TACRPageNo;
   public
    property PageManager: TACRDiskPageManager read LPageManager;
  end;// TACRSystemDirectory


type
  TBooleanFunctionForTimeOutCall = function: Boolean of object;
  TBooleanFunctionForTimeOutCallWithRetryNo = function(const CurrentRetryNo: Integer): Boolean of object;

  // convert PageTypeID to String (for exceptions messages)
  function PageTypeToStr(PageType: TACRPageTypeID): String;
  function TryUsingTimeOut(
                           Func:        TBooleanFunctionForTimeOutCall;
                           LockParams:  TACRLockParams
                          ): Boolean;
 function TryUsingTimeOutForTableLocks(
                           Func:        TBooleanFunctionForTimeOutCallWithRetryNo;
                           LockParams:  TACRLockParams
                         ): Boolean;


implementation

uses ACRLocalEngine;


////////////////////////////////////////////////////////////////////////////////
//
// TACRDiskDatabaseData
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// OpenDatabase
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.OpenDatabase(Session: TACRBaseSession);
var ReadOnly: Boolean;
begin
  FLockParams := Session.LockParams;
  FDatabaseName := Session.DatabaseFileName;
  FCryptoParams := Session.CryptoParams;
  if (FPageManager <> nil) then
   raise EACRException.Create(10498,ErrorLDatabaseFileIsInUse,[FDatabaseName]);
  FPageManager := TACRDiskPageManager.Create(FLockParams);
  try
    TACRDiskPageManager(FPageManager).CryptoInfo := FCryptoParams;
    ReadOnly := Session.ReadOnly;
    TACRDiskPageManager(FPageManager).OpenDatabase(
        Session.DatabaseFileName,ReadOnly,Session.Exclusive);
    Session.ReadOnly := ReadOnly;
    FCryptoParams := TACRDiskPageManager(FPageManager).CryptoInfo;
    if (FCryptoParams.CryptoAlgorithm = ACR_Cipher_None) then
      FCryptoParams.CryptoMode := ACR_Cipher_Mode_CTS;
    FSystemDir := TACRSystemDirectory.Create(TACRDiskPageManager(FPageManager));
    FSystemDir.LoadDirectory;
    FActiveSessionsFile := TACRActiveSessionsFile.Create(FPageManager,FLockParams);
    FActiveSessionsFile.OpenFile(FSystemDir.GetFileFirstPageNo(dbftActiveSessionsList));
    FTableListFile := TACRTableListFile.Create(TACRDiskPageManager(FPageManager));
    FTableListFile.OpenFile(FSystemDir.GetFileFirstPageNo(dbftTablesList));
    FTableLocksManager := TACRTableLocksManager.Create(Self);
    FOptions.MaxSessionCount := FActiveSessionsFile.FHandle.Size;
    FOptions.PageSize := TACRDiskPageManager(FPageManager).FDBHeader.PageSize;
    FOptions.ExtentPageCount := TACRDiskPageManager(FPageManager).FDBHeader.ExtentPageCount;
    FOptions.RandomSearchRetryCount := Session.Options.RandomSearchRetryCount;
  except
    if (FSystemDir <> nil) then
     FreeAndNil(FSystemDir);
    if (FActiveSessionsFile <> nil) then
     FreeAndNil(FActiveSessionsFile);
    if (FTableListFile <> nil) then
     FreeAndNil(FTableListFile);
    FreeAndNil(FPageManager);
    raise;
  end;
end;// OpenDatabase


//------------------------------------------------------------------------------
// CloseDatabase
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.CloseDatabase(Session: TACRBaseSession);
var i: Integer;
begin
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskDatabaseData.CloseDatabase starting - free FTableLocksManager...'+
              ', SessionID = '+IntToStr(Session.SessionID)+
              ', DatabaseName = '+IntToStr(Integer(FDatabaseName))+#13#10);
 {$ENDIF}
  if (FTableLocksManager <> nil) then
   FreeAndNil(FTableLocksManager);
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskDatabaseData.CloseDatabase closing active sessions file...'+
              ', SessionID = '+IntToStr(Session.SessionID)+
              ', DatabaseName = '+IntToStr(Integer(FDatabaseName))+#13#10);
 {$ENDIF}
  if (FActiveSessionsFile <> nil) then
   begin
    FActiveSessionsFile.Free;
    FActiveSessionsFile := nil;
   end;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskDatabaseData.CloseDatabase closing table list file...'+
              ', SessionID = '+IntToStr(Session.SessionID)+
              ', DatabaseName = '+IntToStr(Integer(FDatabaseName))+#13#10);
 {$ENDIF}
  if (FTableListFile <> nil) then
   begin
    FTableListFile.Free;
    FTableListFile := nil;
   end;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskDatabaseData.CloseDatabase closing table system dir...'+
              ', SessionID = '+IntToStr(Session.SessionID)+
              ', DatabaseName = '+IntToStr(Integer(FDatabaseName))+#13#10);
 {$ENDIF}
  if (FSystemDir <> nil) then
   begin
    FSystemDir.Free;
    FSystemDir := nil;
   end;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskDatabaseData.CloseDatabase dropping temporary indexes...'+
              ', FTableDataList.Count = '+IntToStr(FTableDataList.Count)+
              ', SessionID = '+IntToStr(Session.SessionID)+
              ', DatabaseName = '+IntToStr(Integer(FDatabaseName))+#13#10);
 {$ENDIF}
  // drop temporary indexes
  for i := 0 to FTableDataList.Count - 1 do
   if (FTableDataList.Items[i] <> nil) then
    if (TACRTableData(FTableDataList.Items[i]).IndexManager <> nil) then
     begin
      TACRTableData(FTableDataList.Items[i]).IndexManager.DropTemporaryIndexes(Session.SessionID);
     end;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskDatabaseData.CloseDatabase closing page manager'+
              ', SessionID = '+IntToStr(Session.SessionID)+
              ', DatabaseName = '+IntToStr(Integer(FDatabaseName))+#13#10);
 {$ENDIF}
  if (FPageManager <> nil) then
   begin
    TACRDiskPageManager(FPageManager).CloseDatabase;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskDatabaseData.CloseDatabase destroying page manager'+
              ', SessionID = '+IntToStr(Session.SessionID)+
              ', DatabaseName = '+IntToStr(Integer(FDatabaseName))+#13#10);
 {$ENDIF}
    FPageManager.Free;
    FPageManager := nil;
   end;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskDatabaseData.CloseDatabase finished'+
              ', SessionID = '+IntToStr(Session.SessionID)+
              ', DatabaseName = '+IntToStr(Integer(FDatabaseName))+#13#10);
 {$ENDIF}
end;// CloseDatabase


//------------------------------------------------------------------------------
// returns new session ID
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.GetNewSessionID: TACRSessionID;
var i: Integer;
begin
  Lock;
  try
    Result := 1;
    i := 0;
    while (i < FSessionList.Count) do
     begin
      if (TACRBaseSession(FSessionList.Items[i]).SessionID = Result) then
       begin
        Inc(Result);
        i := 0;
        continue;
       end
      else
       Inc(i); 
     end;
  finally
    Unlock;
  end;
end; // GetNewSessionID


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRDiskDatabaseData.Create;
begin
  inherited Create;
  FPageManager := nil;
  FSystemDir := nil;
  FActiveSessionsFile := nil;
  FTableListFile := nil;
  FTableLocksManager := nil;
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRDiskDatabaseData.Destroy;
var
  DBDatas: TList;
  i: integer;
begin
  if (FPageManager <> nil) then
   raise EACRException.Create(10499,ErrorLDatabaseFileIsNotClosed);
  DBDatas := DBDataList.LockList;
  try
    for i:=0 to DBDatas.Count-1 do
     if (DBDatas.Items[i] = Self) then
      begin
       DBDatas.Delete(i);
       break;
      end;
  finally
    DBDataList.UnlockList;
  end;
  if (Length(FCryptoParams.Password) > 0) then
   FillChar(FCryptoParams.Password[1],Length(FCryptoParams.Password),$00);
  FCryptoParams.Password := '';
  FillChar(FCryptoParams,SizeOf(FCryptoParams),$00);
  inherited Destroy;
end;// Destroy


//------------------------------------------------------------------------------
// CreateDatabase
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.CreateDatabase(Session: TACRBaseSession);
begin
  FLockParams := Session.LockParams;
  FCryptoParams := Session.CryptoParams;
  FOptions := Session.Options;
  FPageManager := TACRDiskPageManager.Create(FLockParams);
  try
    TACRDiskPageManager(FPageManager).CryptoInfo := FCryptoParams;
    TACRDiskPageManager(FPageManager).CreateAndOpenDatabase(DatabaseName,
                                  FOptions.PageSize,FOptions.ExtentPageCount);
    FSystemDir := TACRSystemDirectory.Create(TACRDiskPageManager(FPageManager));
    FActiveSessionsFile := TACRActiveSessionsFile.Create(FPageManager,FLockParams);
    FTableListFile := TACRTableListFile.Create(TACRDiskPageManager(FPageManager));
    try
      FSystemDir.CreateDirectory;
      FSystemDir.CreateFile(dbftActiveSessionsList,
                            FActiveSessionsFile.CreateFile(Options.MaxSessionCount));
      FSystemDir.CreateFile(dbftTablesList, FTableListFile.CreateFile);
      FSystemDir.SaveDirectory;
      FPageManager.ApplyChanges(SYSTEM_SESSION_ID);
      TACRDiskPageManager(FPageManager).FlushFileBuffers;
    finally
      FTableListFile.Free;
      FTableListFile := nil;
      FActiveSessionsFile.Free;
      FActiveSessionsFile := nil;
      FSystemDir.Free;
      FSystemDir := nil;
    end;
  finally
    TACRDiskPageManager(FPageManager).CloseDatabase;
    FPageManager.Free;
    FPageManager := nil;
  end;
end;// CreateDatabase


//------------------------------------------------------------------------------
// ConnectSession
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.ConnectSession(Session: TACRBaseSession);
var opt:        TACROptions;
    MaxCount:   Integer;
    bFirstOpen: Boolean;
begin
  Lock;
  try
    bFirstOpen := (FPageManager = nil);
    if (bFirstOpen) then
     OpenDatabase(Session);
    try
      {$IFDEF FILE_SERVER_VERSION}
      if (not FPageManager.Exclusive) then
       Session.SessionID := FActiveSessionsFile.Connect
      else
       Session.SessionID := GetNewSessionID;
      {$ELSE}
      Session.SessionID := GetNewSessionID;
      {$ENDIF}
      opt.RandomSearchRetryCount := Session.Options.RandomSearchRetryCount;
      {$IFDEF FILE_SERVER_VERSION}
      if (not FPageManager.Exclusive) then
        MaxCount := Integer(FActiveSessionsFile.FHandle.Size)
      else
        MaxCount := ACRMaxSingleUserConnections;
      if (MaxCount > ACRMaxSessionCount) then
       MaxCount := ACRMaxSessionCount;
      {$ELSE}
      MaxCount := ACRMaxSingleUserConnections;
      {$ENDIF}
      opt.PageSize := FPageManager.PageSize;
      opt.ExtentPageCount := TACRDiskPageManager(FPageManager).FDBHeader.ExtentPageCount;
      opt.MaxSessionCount := Integer(FActiveSessionsFile.FHandle.Size);
      opt.PageSize := TACRDiskPageManager(FPageManager).FDBHeader.PageSize;
      Session.Options := opt;
      Session.CryptoParams := FCryptoParams;
      if (FSessionList.Count >= MaxCount) then
       raise EACRException.Create(10843,ErrorLMaximumSessionCountExceeded,[MaxCount]);
      if (Session.SessionID >= MaxCount) then
       raise EACRException.Create(11334,ErrorLMaximumSessionCountExceeded,[MaxCount]);
      FSessionList.Add(Session);
    except
     if (bFirstOpen) then
      CloseDatabase(Session);
     raise;
    end;
  finally
    Unlock;
  end;
end;// ConnectSession


//------------------------------------------------------------------------------
// DisconnectSession
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.DisconnectSession(Session: TACRBaseSession);
begin
  Lock;
  try
    FSessionList.Remove(Session);
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskDatabaseData.DisconnectSession disconnecting from active sessions file...'+
              ', SessionID = '+IntToStr(Session.SessionID)+
              ', DatabaseName = '+IntToStr(Integer(FDatabaseName))+#13#10);
 {$ENDIF}
    {$IFDEF FILE_SERVER_VERSION}
    if (FPageManager <> nil) then
     if (not FPageManager.Exclusive) then
      FActiveSessionsFile.Disconnect(Session.SessionID);
    {$ENDIF}
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskDatabaseData.DisconnectSession closing database if needed...'+
              ', FSessionList.Count = '+IntToStr(FSessionList.Count)+
              ', SessionID = '+IntToStr(Session.SessionID)+
              ', DatabaseName = '+IntToStr(Integer(FDatabaseName))+#13#10);
 {$ENDIF}
    if (FSessionList.Count <= 0) then
      CloseDatabase(Session);
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskDatabaseData.DisconnectSession closing database if needed...OK'+
              ', FSessionList.Count = '+IntToStr(FSessionList.Count)+
              ', SessionID = '+IntToStr(Session.SessionID)+
              ', DatabaseName = '+IntToStr(Integer(FDatabaseName))+#13#10);
 {$ENDIF}
  finally
    Unlock;
  end;
end;// DisconnectSession


//------------------------------------------------------------------------------
// TACRDiskDatabaseData
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.FreeIfNoSessionsConnected;
var
  DestroyIt: Boolean;
begin
  Lock;
  try
   DestroyIt := (FSessionList.Count = 0);
  finally
   Unlock;
   if (DestroyIt) then
     Free;
  end;
end;// FreeIfNoSessionsConnected


//------------------------------------------------------------------------------
// FindOrCreateTableData
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.CreateTableData(Cursor: TACRCursor): TACRTableData;
begin
 Result := TACRDiskTableData.Create(Self);
 Result.TableName := Cursor.TableName;
 FTableDataList.Add(Result);
end;// FindOrCreateTableData


//------------------------------------------------------------------------------
// get list of all tables in the database file
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.GetTablesList(List: TStrings);
begin
  Lock;
  try
   if (not TryUsingTimeOut(InternalLockTables,FLockParams)) then
    raise EACRException.Create(10612,ErrorLCannotLockTableS,[FDatabaseName]);
   try
    FTableListFile.GetTablesList(List);
   finally
     if (not InternalUnlockTables) then
      raise EACRException.Create(10613,ErrorLCannotUnlockTables,[FDatabaseName]);
   end;
  finally
   Unlock;
  end;
end; // GetTablesList


//------------------------------------------------------------------------------
// return table item if exists
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.GetTableItemIfExists(
                TableName: String; var TableItem: TACRTableListItem): Boolean;
begin
 Result := FTableListFile.GetTableItemIfExists(TableName,TableItem);
end; // GetTableItemIfExists


//------------------------------------------------------------------------------
// table exists
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.TableExists(TableName: String): Boolean;
begin
  Lock;
  try
   if (not TryUsingTimeOut(InternalLockTables,FLockParams)) then
    raise EACRException.Create(10614,ErrorLCannotLockTableS,[FDatabaseName]);
   try
    Result := FTableListFile.IsTableExist(TableName);
   finally
     if (not InternalUnlockTables) then
      raise EACRException.Create(10615,ErrorLCannotUnlockTables,[FDatabaseName]);
   end;
  finally
   Unlock;
  end;
end; // TableExists


//------------------------------------------------------------------------------
// Lock All Tables
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.InternalLockTables: Boolean;
begin
  Result := TACRDiskPageManager(FPageManager).InternalLockTables;
end; // InternalLockTables


//------------------------------------------------------------------------------
// Unlock All Tables
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.InternalUnlockTables: Boolean;
begin
  Result := TACRDiskPageManager(FPageManager).InternalUnlockTables;
end; // InternalUnlockTables


//------------------------------------------------------------------------------
// flush file buffers
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.FlushFileBuffers;
begin
  TACRDiskPageManager(FPageManager).FlushFileBuffers;
end; // FlushFileBuffers


//------------------------------------------------------------------------------
// return database format version
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.GetFormatVersion(Session: TACRBaseSession): Double;
var PageManagerNil: Boolean;
    ReadOnly:       Boolean;
begin
  PageManagerNil := (FPageManager = nil);
  if (PageManagerNil) then
   begin
    FPageManager := TACRDiskPageManager.Create(FLockParams);
    ReadOnly := True;
    TACRDiskPageManager(FPageManager).OpenDatabase(
        Session.DatabaseFileName,ReadOnly,False,True);
   end;
  try
   TACRDiskPageManager(FPageManager).LoadDBHeader;
   Result := TACRDiskPageManager(FPageManager).DBHeader.Version;
  finally
   if (PageManagerNil) then
    FreeAndNil(FPageManager);
  end;
end;// GetFormatVersion


//------------------------------------------------------------------------------
// get count of free pages
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.GetFreePageCount(Session: TACRBaseSession): Integer;
var PageManagerNil: Boolean;
begin
  PageManagerNil := (FPageManager = nil);
  try
   if (not PageManagerNil) then
    Result := TACRDiskPageManager(FPageManager).FDatabaseFreeSpaceManager.GetFreePageCount;
  finally
   if (PageManagerNil) then
    FreeAndNil(FPageManager);
  end;
end;// GetFreePageCount


//------------------------------------------------------------------------------
// get total count of pages
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.GetTotalPageCount(Session: TACRBaseSession): Integer;
var PageManagerNil: Boolean;
    ReadOnly:       Boolean;
begin
  PageManagerNil := (FPageManager = nil);
  if (PageManagerNil) then
   begin
    FPageManager := TACRDiskPageManager.Create(FLockParams);
    ReadOnly := True;
    TACRDiskPageManager(FPageManager).OpenDatabase(
        Session.DatabaseFileName,ReadOnly,False,True);
   end;
  try
   TACRDiskPageManager(FPageManager).LoadDBHeader;
   Result := TACRDiskPageManager(FPageManager).DBHeader.TotalPageCount;
  finally
   if (PageManagerNil) then
    FreeAndNil(FPageManager);
  end;
end;// GetTotalPageCount


//------------------------------------------------------------------------------
// return true if database is encrypted
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.IsDatabaseEncrypted(Session: TACRBaseSession): Boolean;
var PageManagerNil: Boolean;
    ReadOnly:       Boolean;
begin
  PageManagerNil := (FPageManager = nil);
  if (PageManagerNil) then
   begin
    FPageManager := TACRDiskPageManager.Create(FLockParams);
    try
      ReadOnly := True;
      TACRDiskPageManager(FPageManager).OpenDatabase(
          Session.DatabaseFileName,ReadOnly,False,True);
    except
      Result := False;
      Exit;
    end;
   end;
  try
   TACRDiskPageManager(FPageManager).LoadDBHeader;
   Result := (TACRDiskPageManager(FPageManager).DBHeader.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None);
  finally
   if (PageManagerNil) then
    FreeAndNil(FPageManager);
  end;
end; // IsDatabaseEncrypted


//------------------------------------------------------------------------------
// return true if database is encrypted by password or by key
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.IsDatabaseEncryptedByPassword(Session: TACRBaseSession): Boolean;
var PageManagerNil: Boolean;
    ReadOnly:       Boolean;
begin
  PageManagerNil := (FPageManager = nil);
  if (PageManagerNil) then
   begin
    FPageManager := TACRDiskPageManager.Create(FLockParams);
    try
      ReadOnly := True;
      TACRDiskPageManager(FPageManager).OpenDatabase(
          Session.DatabaseFileName,ReadOnly,False,True);
    except
      Result := False;
      Exit;
    end;
   end;
  try
   TACRDiskPageManager(FPageManager).LoadDBHeader;
   Result := (TACRDiskPageManager(FPageManager).DBHeader.CryptoHeader.CryptoAskPassword = 1);
  finally
   if (PageManagerNil) then
    FreeAndNil(FPageManager);
  end;
end; // IsDatabaseEncryptedByPassword


//------------------------------------------------------------------------------
// makes Exe database from edb file
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.MakeExeDatabase(Session: TACRBaseSession; ExeFileName, ExeDatabaseFileName: string);
var StubStream, ExeDatabaseStream, DatabaseStream: TACRFileStream;
begin
 // open stub
 ExeDatabaseStream := TACRFileStream.Create(ExeDatabaseFileName,fmCreate);
 try
   StubStream := TACRFileStream.Create(ExeFileName, fmOpenRead or fmShareDenyNone);
   try
     StubStream.SaveToStream(ExeDatabaseStream);
   finally
     StubStream.Free;
   end;
   ExeDatabaseStream.Position := ExeDatabaseStream.Size;
   DatabaseStream := TACRFileStream.Create(Session.DatabaseFileName, fmOpenRead or fmShareExclusive);
   try
     DatabaseStream.SaveToStream(ExeDatabaseStream);
   finally
     DatabaseStream.Free;
   end;
 finally
   ExeDatabaseStream.Free;
 end;
end; // MakeExeDatabase


//------------------------------------------------------------------------------
// removes database file from executable database file
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.RemoveDatabaseFromExe(Session: TACRBaseSession);
var PageManagerNil: Boolean;
begin
  PageManagerNil := (FPageManager = nil);
  if (PageManagerNil) then
    FPageManager := TACRDiskPageManager.Create(FLockParams);
  try
    TACRDiskPageManager(FPageManager).RemoveDatabaseFromExe(Session.DatabaseFileName);
  finally
   if (PageManagerNil) then
    FreeAndNil(FPageManager);
  end;
end; // RemoveDatabaseFromExe


//------------------------------------------------------------------------------
// returns true if this file is an Accuracer database
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.IsAccuracerDatabaseFile(Session: TACRBaseSession): Boolean;
var PageManagerNil: Boolean;
begin
  PageManagerNil := (FPageManager = nil);
  if (PageManagerNil) then
    FPageManager := TACRDiskPageManager.Create(FLockParams);
  try
   if (PageManagerNil) then
    begin
     Result := TACRDiskPageManager(FPageManager).GetExists(Session.DatabaseFileName);
     if (Result) then
      Result := TACRDiskPageManager(FPageManager).IsAccuracerDatabaseFile(Session.DatabaseFileName);
    end
   else
    Result := True; // file already opened
  finally
   if (PageManagerNil) then
    FreeAndNil(FPageManager);
  end;
end; // IsAccuracerDatabaseFile


//------------------------------------------------------------------------------
// return true if CryptoParams are valid
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.IsCryptoParamsValid(Session: TACRBaseSession): Boolean;
var PageManagerNil: Boolean;
    ReadOnly:       Boolean;
begin
  PageManagerNil := (FPageManager = nil);
  if (PageManagerNil) then
   begin
    FPageManager := TACRDiskPageManager.Create(FLockParams);
    try
      ReadOnly := True;
      TACRDiskPageManager(FPageManager).OpenDatabase(
          Session.DatabaseFileName,ReadOnly,False,True);
    except
      Result := True;
      Exit;
    end;
    FCryptoParams := TACRDiskPageManager(FPageManager).CryptoInfo;
    FCryptoParams.KeyInfo := Session.CryptoParams.KeyInfo;
    FCryptoParams.Password := Session.CryptoParams.Password;
    FCryptoParams.UseInitVector := Session.CryptoParams.UseInitVector;
    FCryptoParams.InitVector := Session.CryptoParams.InitVector;
    Session.CryptoParams := FCryptoParams;
   end;
  try
   TACRDiskPageManager(FPageManager).LoadDBHeader;
   if (TACRDiskPageManager(FPageManager).DBHeader.CryptoHeader.CryptoAlgorithm = ACR_Cipher_None) then
    Result := True
   else
    begin
      Result := ACRIsKeyValid(TACRDiskPageManager(FPageManager).DBHeader.CryptoHeader,
                           Session.CryptoParams);
    end;
  finally
   if (PageManagerNil) then
    FreeAndNil(FPageManager);
  end;
end; // IsCryptoParamsValid


//------------------------------------------------------------------------------
// fill TableLockList with lock info for the tables opened by this session
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.GetTableLocksListForSession(
                                          SessionID:      TACRSessionID;
                                          TableLockList:  TList
                                      );
begin
  FTableLocksManager.GetTableLocksListForSession(SessionID,TableLockList);
end; // GetTableLocksListForSession




//------------------------------------------------------------------------------
// if result is true and PTableLockInfo is not nil - increment corresponding field of this record
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.TryToLockTable(
                    TableName:            String;
                    SessionID:            TACRSessionID;
                    LockType:             TACRLockType;
                    PTableLockInfo:       PACRTableLockInfo = nil): Boolean;
var
  i:                Integer;
  TableItem:        TACRTableListItem;
begin
  Result := GetTableItemIfExists(TableName,TableItem);
  if (Result) then
   begin
    for i := 1 to FLockParams.RetryCount do
     begin
      Result := FTableLocksManager.LockTable(i,TableItem,TableName,
                  SessionID,LockType,PTableLockInfo);
      if Result then
         break
      else
        Sleep(FLockParams.Delay);
     end;
   end;
end; // LockTable


//------------------------------------------------------------------------------
// decrements number of locks by this session and removes all locks if their number = 0
// if PTableLockInfo = nil - all table locks byt this session will be removed
//------------------------------------------------------------------------------
function TACRDiskDatabaseData.UnlockTableAll(
                        TableName:      String;
                        SessionID:      TACRSessionID;
                        PTableLockInfo: PACRTableLockInfo = nil): Boolean;
begin
 Result := FTableLocksManager.UnlockTableAll(TableName,SessionID,PTableLockInfo);
end; // UnlockTableAll


//------------------------------------------------------------------------------
// remove all locks of the session - called by disconnect of server session
//------------------------------------------------------------------------------
procedure TACRDiskDatabaseData.RemoveAllLocks(SessionID: TACRSessionID); 
begin
  if (FTableLocksManager <> nil) then
   FTableLocksManager.RemoveAllLocksForSession(SessionID);
end; // RemoveAllLocks


////////////////////////////////////////////////////////////////////////////////
//
// TACRTablePageMapsManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// set page type
//------------------------------------------------------------------------------
procedure TACRTablePageMapsManager.SetPageType(
                  SessionID:          TACRSessionID;
                  PageNo:             TACRPageNo;
                  TablePageType:      TACRTablePageType);
var PFSPage:    TACRPage;
    PFSPageNo:  TACRPageNo;
    aPageNo:    TACRPageNo;
    OldType:    pByte;
begin
  // mark new page in accordance with page type
  PFSPageNo := PageNo div FPFSItemsPerPage;
  if (PFSPageNo >= FTablePFSMapFile.TablePFSPageCount) then
   raise EACRException.Create(10512,ErrorLInvalidPageCount,[PFSPageNo,FTablePFSMapFile.TablePFSPageCount]);
  PFSPageNo := FTablePFSMapFile.TablePFSPageMapItems[PFSPageNo].PageNo;
  PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
  PFSPage.IsDirty := True;
  aPageNo := PageNo mod FPFSItemsPerPage;
  // each page type size = 2 bits
  OldType := PByte(PFSPage.PageData + FPageTypeOffset + (aPageNo div 4));
  case aPageNo mod 4 of
   0: OldType^ := (OldType^ and $FC) or (Byte(TablePageType) and $03);
   1: OldType^ := (OldType^ and $F3) or ((Byte(TablePageType) and $03) shl 2);
   2: OldType^ := (OldType^ and $CF) or ((Byte(TablePageType) and $03) shl 4);
   3: OldType^ := (OldType^ and $3F) or ((Byte(TablePageType) and $03) shl 6);
  end;
  LPageManager.PutPage(PFSPage);
end; // SetPageType


//------------------------------------------------------------------------------
// set PageRecordCount or PageFreeSpace
//------------------------------------------------------------------------------
procedure TACRTablePageMapsManager.SetPageRecordCount(
                SessionID:        TACRSessionID;
                PageNo:           TACRPageNo;
                PageRecordCount:  Word;
                PageType:         TACRTablePageType);
var PFSPage:            TACRPage;
    PFSItemNo:          Integer;
    PFSPageNo:          TACRPageNo;
    aPageNo:            TACRPageNo;
    Delta:              Integer;
    PPageSize:          PWord;
begin
  // mark new page in accordance with page type
  PFSItemNo := PageNo div FPFSItemsPerPage;
  if (PFSItemNo >= FTablePFSMapFile.TablePFSPageCount) then
   raise EACRException.Create(10513,ErrorLInvalidPageCount,[PFSItemNo,FTablePFSMapFile.TablePFSPageCount]);
  PFSPageNo := FTablePFSMapFile.TablePFSPageMapItems[PFSItemNo].PageNo;
  PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
  PFSPage.IsDirty := True;
  aPageNo := PageNo mod FPFSItemsPerPage;
  PPageSize := PWord(PFSPage.PageData + aPageNo * SizeOf(Word));
  // if page was row_start then calculate new page record count
  if (PageType = tptRowStart) then
   begin
    Delta := 0;
    if (PPageSize^ = INVALID_PAGE_RECORD_NO) then
     begin
      if (PageRecordCount <> INVALID_PAGE_RECORD_NO) then
       Delta := PageRecordCount;
     end
    else
     begin
      if (PageRecordCount <> INVALID_PAGE_RECORD_NO) then
       Delta := (Integer(PageRecordCount) - Integer(PPageSize^))
      else
       Delta := -Integer(PPageSize^);
     end;
    Inc(FTablePFSMapFile.TablePFSPageMapItems[PFSItemNo].PageRecordCount,Delta);
   end;
  Move(PageRecordCount,PPageSize^,SizeOf(Word));
  LPageManager.PutPage(PFSPage);
end; // SetPageRecordCount


//------------------------------------------------------------------------------
// return PageType
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.InternalGetPageType(
            PFSPage:        TACRPage;
            RelativePageNo: TACRPageNo): TACRTablePageType;
var
    b: byte;
begin
  b := PByte(PFSPage.PageData + FPageTypeOffset + (RelativePageNo div 4))^;
  case RelativePageNo mod 4 of
   0: Result := TACRTablePageType(b and $03);
   1: Result := TACRTablePageType((b shr 2) and $03);
   2: Result := TACRTablePageType((b shr 4) and $03);
   3: Result := TACRTablePageType((b shr 6) and $03);
  end;
end; // InternalGetPageType


//------------------------------------------------------------------------------
// return PageRecordCount or PageFreeSpace
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.GetPageRecordCount(
                  SessionID:          TACRSessionID;
                  PageNo:             TACRPageNo): Word;
var PFSPage:    TACRPage;
    PFSPageNo:  TACRPageNo;
    aPageNo:    TACRPageNo;
begin
  // mark new page in accordance with page type
  PFSPageNo := PageNo div FPFSItemsPerPage;
  if (PFSPageNo >= FTablePFSMapFile.TablePFSPageCount) then
   raise EACRException.Create(10515,ErrorLInvalidPageCount,[PFSPageNo,FTablePFSMapFile.TablePFSPageCount]);
  PFSPageNo := FTablePFSMapFile.TablePFSPageMapItems[PFSPageNo].PageNo;
  PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
  aPageNo := PageNo mod FPFSItemsPerPage;
  Move(PWord(PFSPage.PageData + aPageNo * SizeOf(Word))^,Result,SizeOf(Word));
  LPageManager.PutPage(PFSPage);
end; // GetPageRecordCount


//------------------------------------------------------------------------------
// find page in PFS page map
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.FindPageInPFSPage(
                  PageData:             PChar;
                  MaxRelativePageNo:    Integer;
                  TablePageType:        TACRTablePageType;
                  Size:                 Word;
                  out RelativePageNo:   Integer;
                  out PageRecordCount:  Word
                          ): Boolean;
var i,imod4:    Integer;
    PPageSize:  PWord;
    bRowStart:  Boolean;
    bOK:        Boolean;
    PType:      PByte;
    pt0,pt1,
    pt2,pt3:    Byte;
begin
  i := 0;
  imod4 := 0;
  pt0 := Byte(TablePageType);
  pt1 := Byte(TablePageType) shl 2;
  pt2 := Byte(TablePageType) shl 4;
  pt3 := Byte(TablePageType) shl 6;
  PPageSize := PWord(PageData);
  PType := PByte(PageData + FPageTypeOffset);
  bRowStart := (TablePageType = tptRowStart);
  Result := False;
  while (i <= MaxRelativePageNo) do
   begin
    if (PPageSize^ <> INVALID_PAGE_RECORD_NO) then
     begin
      if (bRowStart) then
        bOK := (PPageSize^ < Size)
      else
        bOK := (PPageSize^ >= Size);
      if (bOK) then
       begin
        case imod4 of
        0: bOK := (((PType^ and $03) xor pt0) = 0);
        1: bOK := (((PType^ and $0C) xor pt1) = 0);
        2: bOK := (((PType^ and $30) xor pt2) = 0);
        3: bOK := (((PType^ and $C0) xor pt3) = 0);
        end;
        if (bOK) then
         begin
          Result := True;
          RelativePageNo := i;
          PageRecordCount := PPageSize^;
          break;
         end;
       end;
     end;
    Inc(PPageSize);
    Inc(i);
    Inc(imod4);
    if (imod4 >= 4) then
     begin
      imod4 := 0;
      Inc(PType);
     end;
   end;
end; // FindPageInPFSPage



//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRTablePageMapsManager.Create(
                       aTableData:              TACRTableData;
                       aRandomSearchRetryCount: Integer;
                       IsSynchronizationRequired: TACRIsPageSynchronizationRequired
                      );
begin
  LTableData := TACRDiskTableData(aTableData);
  if (LTableData = nil) then
    raise EACRException.Create(10500,ErrorLNilPointer);
  if (LTableData.DatabaseData = nil) then
    raise EACRException.Create(10501,ErrorLNilPointer);
  if (LTableData.DatabaseData.PageManager = nil) then
    raise EACRException.Create(10502,ErrorLNilPointer);
  LPageManager := TACRDiskPageManager(LTableData.DatabaseData.PageManager);
  FPFSItemsPerPage := ((LPageManager.PageDataSize * 8) div ACRTablePFSItemSize);
  FPageTypeOffset := FPFSItemsPerPage * 2;
  FTablePFSMapFile := TACRTablePFSMapFile.Create(LPageManager,IsSynchronizationRequired);
  FRandomSearchRetryCount := aRandomSearchRetryCount;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRTablePageMapsManager.Destroy;
begin
  FreeAndNil(FTablePFSMapFile);
  inherited;
end; // Destroy


//------------------------------------------------------------------------------
// add page
//------------------------------------------------------------------------------
procedure TACRTablePageMapsManager.OpenTablePageMapsManager(StartPageNo: TACRPageNo; OpenExclusive: Boolean);
begin
  FTablePFSMapFile.OpenFile(StartPageNo,OpenExclusive);
end; // OpenTablePageMapsManager


//------------------------------------------------------------------------------
// add page
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.AddPage(
                                          SessionID:      TACRSessionID;
                                          TablePageType:  TACRTablePageType
                                         ): TACRPage;
var NewPageCount:                     Integer;
    I, NumPagesToAdd:                 Integer;
    NewPage:                          TACRPage;
begin
  Result := LPageManager.AddPage(SessionID);
  if (Result.PageNo >= FTablePFSMapFile.PageCount) then
   begin
    // need to extend table page map
    NewPageCount := Result.PageNo+1;

    if (FTablePFSMapFile.TablePFSPageCount * FPFSItemsPerPage < NewPageCount) then
     begin
      NumPagesToAdd := NewPageCount div FPFSItemsPerPage +
                        Integer(NewPageCount mod FPFSItemsPerPage > 0) -
                        FTablePFSMapFile.TablePFSPageCount;
      for I := 1 to NumPagesToAdd do
       begin
        NewPage := LPageManager.AddPage(SessionID);
        NewPage.PageHeader.ObjectID := LTableData.TableID;
        NewPage.PageHeader.PageType := ACRPageTypeIDTablePFS;
        if (NewPage.PageNo = INVALID_PAGE_NO) then
          raise EACRException.Create(10511,ErrorLInvalidPageNo,[NewPage.PageNo]);
        FTablePFSMapFile.AddPage(SessionID,NewPage.PageNo);
        // mark all items in new pfs page as empty pages
        FillChar(NewPage.PageData^,(FPFSItemsPerPage * SizeOf(Word)),$FF);
        LPageManager.PutPage(NewPage);
       end;
     end;
    FTablePFSMapFile.PageCount := NewPageCount;
   end;
  SetPageType(SessionID,Result.PageNo,TablePageType);
end; // AddPage


//------------------------------------------------------------------------------
// remove page
//------------------------------------------------------------------------------
procedure TACRTablePageMapsManager.RemovePage(
                          SessionID:  TACRSessionID;
                          PageNo:         TACRPageNo;
                          TablePageType:  TACRTablePageType
                                             );
begin
  SetPageRecordCount(SessionID,PageNo,INVALID_PAGE_RECORD_NO,TablePageType);
end; // RemovePage


//------------------------------------------------------------------------------
// try to find a page for adding a small amount of data (< 1 page)
// if failed performs add page
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.GetPageForAddingNewData(
                      SessionID:            TACRSessionID;
                      TablePageType:        TACRTablePageType;
                      Size:                 Word;
                      out PageRecordCount:  Word
                                                          ): TACRPage;
var j,k,PFSItemNo:  Integer;
    PFSPageNo,PageNo: TACRPageNo;
    PFSPage:          TACRPage;
    ScannedPFSPages:  TACRPagesArray;
begin
  Result := nil;
  ScannedPFSPages := TACRPagesArray.Create(0,FRandomSearchRetryCount);
  try
    if (FTablePFSMapFile.TablePFSPageCount > 0) then
     begin
      for j := 1 to FRandomSearchRetryCount do
       begin
        PFSItemNo := Random(MaxInt) mod FTablePFSMapFile.TablePFSPageCount;
        PFSPageNo :=  FTablePFSMapFile.TablePFSPageMapItems[PFSItemNo].PageNo;
        if (ScannedPFSPages.IsValueExists(PFSPageNo)) then
         continue
        else
         ScannedPFSPages.Append(PFSPageNo);
        PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
        // search desired page inside the random PFS page
        if (PFSItemNo = FTablePFSMapFile.TablePFSPageCount-1) then
         k := (FTablePFSMapFile.PageCount-1) mod FPFSItemsPerPage
        else
         k := FPFSItemsPerPage-1;

        if (FindPageInPFSPage(PFSPage.PageData,k,TablePageType,Size,
                            PageNo,PageRecordCount)) then
         Result := LPageManager.GetPage(SessionID,(PFSItemNo * FPFSItemsPerPage + PageNo),LTableData.IsSynchronizationRequired);
        LPageManager.PutPage(PFSPage);
        if (Result <> nil) then break;
       end; // random search loop
     end; // page Count
    if (Result = nil) then
     begin
      Result := AddPage(SessionID,TablePageType);
      PageRecordCount := INVALID_PAGE_RECORD_NO; // empty page was found
     end;
  finally
   ScannedPFSPages.Free;
  end;
end; // GetPageForAddingNewData


//------------------------------------------------------------------------------
// return PageNo where first row begin if row count > 0, otherwise return INVALID_PAGE_NO
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.GetFirstRowBeginPageNo(SessionID: TACRSessionID): TACRPageNo;
var i,j,k:            Integer;
    PFSPage:          TACRPage;
    PFSPageNo:        TACRPageNo;
begin
  Result := INVALID_PAGE_NO;
  if (FTablePFSMapFile.TablePFSPageCount > 0) then
   begin
    for j := 0 to FTablePFSMapFile.TablePFSPageCount-1 do
     if (FTablePFSMapFile.TablePFSPageMapItems[j].PageRecordCount > 0) then
      begin
        PFSPageNo := FTablePFSMapFile.TablePFSPageMapItems[j].PageNo;
        PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
        if (j = FTablePFSMapFile.TablePFSPageCount-1) then
         k := (FTablePFSMapFile.PageCount-1) mod FPFSItemsPerPage
        else
         k := FPFSItemsPerPage-1;
        for i := 0 to k do
         if (PWord(PFSPage.PageData + i * SizeOf(Word))^ <>
             INVALID_PAGE_RECORD_NO) then
          if (InternalGetPageType(PFSPage,i) = tptRowStart) then
            begin
             Result := j * FPFSItemsPerPage + i;
             break;
            end;
        LPageManager.PutPage(PFSPage);
        if (Result <> INVALID_PAGE_NO) then
         break;
      end;
   end;
end; // GetFirstRowBeginPageNo


//------------------------------------------------------------------------------
// return page where last row begin if row count > 0, otherwise return nil
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.GetLastRowBeginPageNo(SessionID: TACRSessionID): TACRPageNo;
var i,j,k:            Integer;
    PFSPage:          TACRPage;
    PFSPageNo:        TACRPageNo;
begin
  Result := INVALID_PAGE_NO;
  if (FTablePFSMapFile.TablePFSPageCount > 0) then
   begin
    for j := FTablePFSMapFile.TablePFSPageCount-1 downto 0  do
     if (FTablePFSMapFile.TablePFSPageMapItems[j].PageRecordCount > 0) then
      begin
        PFSPageNo := FTablePFSMapFile.TablePFSPageMapItems[j].PageNo;
        PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
        if (j = FTablePFSMapFile.TablePFSPageCount-1) then
         k := (FTablePFSMapFile.PageCount-1) mod FPFSItemsPerPage
        else
         k := FPFSItemsPerPage-1;
        for i := k downto 0 do
         if (PWord(PFSPage.PageData + i * SizeOf(Word))^ <>
             INVALID_PAGE_RECORD_NO) then
           if (InternalGetPageType(PFSPage,i) = tptRowStart) then
             begin
              Result := j * FPFSItemsPerPage + i;
              break;
             end;
        LPageManager.PutPage(PFSPage);
        if (Result <> INVALID_PAGE_NO) then
         break;
      end;
   end;
end; // GetLastRowBeginPageNo


//------------------------------------------------------------------------------
// return page where next row begin if row count > 0, otherwise return nil
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.GetNextRowBeginPageNo(SessionID:  TACRSessionID;
                                                    PageNo:     TACRPageNo): TACRPageNo;
var i,j,k,n,m:        Integer;
    PFSPage:          TACRPage;
    PFSPageNo:        TACRPageNo;
begin
  Result := INVALID_PAGE_NO;
  if (FTablePFSMapFile.TablePFSPageCount > 0) then
   begin
    n := PageNo div FPFSItemsPerPage;
    for j := n to FTablePFSMapFile.TablePFSPageCount-1 do
     if (FTablePFSMapFile.TablePFSPageMapItems[j].PageRecordCount > 0) then
      begin
        PFSPageNo := FTablePFSMapFile.TablePFSPageMapItems[j].PageNo;
        PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
        if (j = FTablePFSMapFile.TablePFSPageCount-1) then
         k := (FTablePFSMapFile.PageCount-1) mod FPFSItemsPerPage
        else
         k := FPFSItemsPerPage-1;
        if (j = n) then
          m := PageNo mod FPFSItemsPerPage+1
        else
          m := 0;
        for i := m to k do
         if (PWord(PFSPage.PageData + i * SizeOf(Word))^ <>
             INVALID_PAGE_RECORD_NO) then
           if (InternalGetPageType(PFSPage,i) = tptRowStart) then
             begin
              Result := j * FPFSItemsPerPage + i;
              break;
             end;
        LPageManager.PutPage(PFSPage);
        if (Result <> INVALID_PAGE_NO) then
         break;
      end;
   end;
end; // GetNextRowBeginPageNo


//------------------------------------------------------------------------------
// return page where prior row begin if row count > 0, otherwise return nil
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.GetPriorRowBeginPageNo(SessionID: TACRSessionID;
                                                        PageNo:    TACRPageNo): TACRPageNo;
var i,j,k,n,m:        Integer;
    PFSPage:          TACRPage;
    PFSPageNo:        TACRPageNo;
begin
  Result := INVALID_PAGE_NO;
  if (FTablePFSMapFile.TablePFSPageCount > 0) then
   begin
    n := PageNo div FPFSItemsPerPage;
    for j := n downto 0  do
     if (FTablePFSMapFile.TablePFSPageMapItems[j].PageRecordCount > 0) then
      begin
        PFSPageNo := FTablePFSMapFile.TablePFSPageMapItems[j].PageNo;
        PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
        if (j = FTablePFSMapFile.TablePFSPageCount-1) then
         k := (FTablePFSMapFile.PageCount-1) mod FPFSItemsPerPage
        else
         k := FPFSItemsPerPage-1;
        if (j = n) then
          m := PageNo mod FPFSItemsPerPage-1
        else
          m := k;
        for i := m downto 0 do
         if (PWord(PFSPage.PageData + i * SizeOf(Word))^ <>
             INVALID_PAGE_RECORD_NO) then
           if (InternalGetPageType(PFSPage,i) = tptRowStart) then
             begin
              Result := j * FPFSItemsPerPage + i;
              break;
             end;
        LPageManager.PutPage(PFSPage);
        if (Result <> INVALID_PAGE_NO) then
         break;
      end;
   end;
end; // GetPriorRowBeginPageNo


//------------------------------------------------------------------------------
// return RecNo of the specified record in physical records order
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.GetRecNoByRecordID(SessionID: TACRSessionID; RecordID: TACRRecordID): TACRRecordNo;
var i,j,k:                Integer;
    PFSPage:              TACRPage;
    PFSPageNo:            TACRPageNo;
    RecCount:             TACRRecordNo;
begin
  Result := -1;
  if (FTablePFSMapFile.TablePFSPageCount > 0) then
   begin
    k := RecordID.PageNo div FPFSItemsPerPage;
    if (k < FTablePFSMapFile.TablePFSPageCount) then
     begin
      Result := 0;
      for j := 0 to k-1 do
       begin
        RecCount := FTablePFSMapFile.TablePFSPageMapItems[j].PageRecordCount;
        Inc(Result,RecCount);
       end;
      PFSPageNo := FTablePFSMapFile.TablePFSPageMapItems[k].PageNo;
      PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
      try
        k := RecordID.PageNo mod FPFSItemsPerPage;
        for i := 0 to k-1 do
         if (PWord(PFSPage.PageData + i * SizeOf(Word))^ <>
             INVALID_PAGE_RECORD_NO) then
          if (InternalGetPageType(PFSPage,i) = tptRowStart) then
           Inc(Result, Integer(PWord(PFSPage.PageData + i * 2)^));
      finally
        LPageManager.PutPage(PFSPage);
      end;
     end; // Page found
   end; // record count > 0
end; // GetRecNoByRecordID


//------------------------------------------------------------------------------
// return RecordID of the record with specified RecNo in physical order
// RecNo is 0-based
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.GetRecordIDByRecNo(SessionID: TACRSessionID; RecNo: TACRRecordNo): TACRRecordID;
var i,j,k:                Integer;
    PFSPage:              TACRPage;
    PFSPageNo:            TACRPageNo;
    CurRecNo,RecCount:    TACRRecordNo;
begin
  Result.PageNo := INVALID_PAGE_NO;
  CurRecNo := 0;
  PFSPageNo := INVALID_PAGE_NO;
  if (FTablePFSMapFile.TablePFSPageCount > 0) then
   begin
    for j := 0 to FTablePFSMapFile.TablePFSPageCount-1 do
     begin
      RecCount := FTablePFSMapFile.TablePFSPageMapItems[j].PageRecordCount;
      if (CurRecNo + RecCount > RecNo) then
       begin
        PFSPageNo := FTablePFSMapFile.TablePFSPageMapItems[j].PageNo;
        break;
       end;
      Inc(CurRecNo,RecCount); 
     end;
    if (PFSPageNo <> INVALID_PAGE_NO) then
      begin
        PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
        if (j = FTablePFSMapFile.TablePFSPageCount-1) then
         k := (FTablePFSMapFile.PageCount-1) mod FPFSItemsPerPage
        else
         k := FPFSItemsPerPage-1;
        for i := 0 to k do
         if (PWord(PFSPage.PageData + i * SizeOf(Word))^ <>
             INVALID_PAGE_RECORD_NO) then
          if (InternalGetPageType(PFSPage,i) = tptRowStart) then
           begin
            RecCount := PWord(PFSPage.PageData + i * SizeOf(Word))^;
            if (CurRecNo + RecCount > RecNo) then
             begin
              Result.PageNo := j * FPFSItemsPerPage + i;
              // relative page item no
              // real page item no is a number of PageItemNo bit = 0
              // in delete rows of this page
              Result.PageItemNo := RecNo - CurRecNo;
              break;
             end;
            Inc(CurRecNo,RecCount);
           end;
        LPageManager.PutPage(PFSPage);
      end;
   end; // record count > 0
end; // GetRecordIDByRecNo


//------------------------------------------------------------------------------
// return number of physical records in the table
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.GetRecordCount: TACRRecordNo;
begin
  Result := FTablePFSMapFile.GetTableRecordCount;
end; // GetRecordCount


//------------------------------------------------------------------------------
// LoadFile
//------------------------------------------------------------------------------
procedure TACRTablePageMapsManager.LoadFile(SessionID: TACRSessionID);
begin
 FTablePFSMapFile.LoadFile(SessionID);
end; // LoadFile


//------------------------------------------------------------------------------
// SaveFile
//------------------------------------------------------------------------------
procedure TACRTablePageMapsManager.SaveFile(SessionID: TACRSessionID);
begin
 FTablePFSMapFile.SaveFile(SessionID);
end; // SaveFile


//------------------------------------------------------------------------------
// empty all table pages with rows, varchars and blobs
//------------------------------------------------------------------------------
procedure TACRTablePageMapsManager.Empty(SessionID: TACRSessionID);
var i,j,k:                Integer;
    PFSPage:              TACRPage;
    PFSPageNo:            TACRPageNo;
begin
  for j := 0 to FTablePFSMapFile.TablePFSPageCount-1 do
    begin
      PFSPageNo := FTablePFSMapFile.TablePFSPageMapItems[j].PageNo;
      PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
      if (PFSPageNo = FTablePFSMapFile.TablePFSPageCount-1) then
       k := (FTablePFSMapFile.PageCount-1) mod FPFSItemsPerPage
      else
       k := FPFSItemsPerPage-1;
      // removing rows, varchars and blob pages of the table
      for i := 0 to k do
       if (PWord(PFSPage.PageData + i * 2)^ <> INVALID_PAGE_RECORD_NO) then
         LPageManager.RemovePage(SessionID, j * FPFSItemsPerPage + i);

      LPageManager.PutPage(PFSPage);
      // remove table pfs map page
      LPageManager.RemovePage(SessionID,PFSPageNo);
    end;
  FTablePFSMapFile.EmptyFile(SessionID);
end; // Empty


//------------------------------------------------------------------------------
// save file state
//------------------------------------------------------------------------------
function TACRTablePageMapsManager.SaveState: Pointer;
begin
  Result := FTablePFSMapFile.SaveState;
end; // SaveState


//------------------------------------------------------------------------------
// restore file state
//------------------------------------------------------------------------------
procedure TACRTablePageMapsManager.RestoreState(State: Pointer);
begin
  FTablePFSMapFile.RestoreState(State);
end; // RestoreState


//------------------------------------------------------------------------------
// free file state
//------------------------------------------------------------------------------
procedure TACRTablePageMapsManager.FreeState(var State: Pointer);
begin
  FTablePFSMapFile.FreeState(State);
end; // FreeState


//------------------------------------------------------------------------------
// get page type and record count
//------------------------------------------------------------------------------
procedure TACRTablePageMapsManager.GetPageTypeAndRecordCount(
            SessionID:           TACRSessionID;
            PageNo:              TACRPageNo;
            out TablePageType:   TACRTablePageType;
            out PageRecordCount: Word);
var PFSPage:    TACRPage;
    PFSPageNo:  TACRPageNo;
    aPageNo:    TACRPageNo;
begin
  // mark new page in accordance with page type
  PFSPageNo := PageNo div FPFSItemsPerPage;
  if (PFSPageNo >= FTablePFSMapFile.TablePFSPageCount) then
   raise EACRException.Create(10515,ErrorLInvalidPageCount,[PFSPageNo,FTablePFSMapFile.TablePFSPageCount]);
  PFSPageNo := FTablePFSMapFile.TablePFSPageMapItems[PFSPageNo].PageNo;
  PFSPage := LPageManager.GetPage(SessionID,PFSPageNo,LTableData.IsSynchronizationRequired);
  aPageNo := PageNo mod FPFSItemsPerPage;
  Move(PWord(PFSPage.PageData + aPageNo * SizeOf(Word))^,PageRecordCount,SizeOf(Word));
  TablePageType := InternalGetPageType(PFSPage,aPageNo);
  LPageManager.PutPage(PFSPage);
end; // GetPageTypeAndRecordCount


////////////////////////////////////////////////////////////////////////////////
//
// TACRDiskRecordManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// load data items map from page
//------------------------------------------------------------------------------
function TACRDiskRecordManager.LoadDataItemsMapFromPage(
        ErrorCode:            Integer;
        Page:                 TACRPage;
        var ItemCount:        Word
                                  ): TACRDataItemsMap;
var Size:     Word;
    Count:    Integer;
begin
  Move(PChar(Page.PageData + Page.PageDataSize - SizeOf(Word))^,
    ItemCount,SizeOf(Word));
  if (ItemCount = 0) then
   begin
    LPageManager.PutPage(Page);
    raise EACRException.Create(ErrorCode,ErrorLInvalidItemCount,[ItemCount]);
   end;
  Count := ItemCount;
  SetLength(Result,Count);
  Size := SizeOf(TACRDataItem) * ItemCount;
  Move(PChar(Page.PageData + Page.PageDataSize - Size - SizeOf(Word))^,
    Result[0],Size);
end; // LoadDataItemsMapFromPage


//------------------------------------------------------------------------------
// save data items map to page
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.SaveDataItemsMapToPage(
        Page:                 TACRPage;
        ItemsMap:             TACRDataItemsMap;
        ItemCount:            Word
                                  );
var Size:   Integer;
    Offset: Integer;
begin
  Size := SizeOf(TACRDataItem) * ItemCount;
  Offset := Page.PageDataSize - Size - SizeOf(Word);
  Move(ItemsMap[0],PChar(Page.PageData + Offset)^,Size);
  Move(ItemCount,PChar(Page.PageData +
        Page.PageDataSize - SizeOf(Word))^,SizeOf(Word));
end; // LoadDataItemsMapFromPage


//------------------------------------------------------------------------------
// return INVALID_PAGE_RECORD_NO if failed, otherwise return RecordNo
//------------------------------------------------------------------------------
function TACRDiskRecordManager.GetFirstRecordOnPage(Page: TACRPage): TACRPageItemID;
var i: Integer;
begin
  Result := INVALID_PAGE_RECORD_NO;
  for i := 0 to FRecordsPerPage-1 do
   if (not CheckNullFlag(i,Page.PageData)) then
    begin
      Result := i;
      break;
    end;
end; // GetFirstRecordOnPage


//------------------------------------------------------------------------------
// return INVALID_PAGE_RECORD_NO if failed, otherwise return RecordNo
//------------------------------------------------------------------------------
function TACRDiskRecordManager.GetLastRecordOnPage(Page: TACRPage): TACRPageItemID;
var i: Integer;
begin
  Result := INVALID_PAGE_RECORD_NO;
  for i := FRecordsPerPage-1 downto 0 do
   if (not CheckNullFlag(i,Page.PageData)) then
    begin
      Result := i;
      break;
    end;
end; // GetLastRecordOnPage


//------------------------------------------------------------------------------
// return INVALID_PAGE_RECORD_NO if failed, otherwise return RecordNo
//------------------------------------------------------------------------------
function TACRDiskRecordManager.GetNextRecordOnPage(Page: TACRPage; PageItemNo: TACRPageItemID): TACRPageItemID;
var i: Integer;
begin
  Result := INVALID_PAGE_RECORD_NO;
  for i := PageItemNo+1 to FRecordsPerPage-1 do
   if (not CheckNullFlag(i,Page.PageData)) then
    begin
      Result := i;
      break;
    end;
end; // GetNextRecordOnPage


//------------------------------------------------------------------------------
// return INVALID_PAGE_RECORD_NO if failed, otherwise return RecordNo
//------------------------------------------------------------------------------
function TACRDiskRecordManager.GetPriorRecordOnPage(Page: TACRPage; PageItemNo: TACRPageItemID): TACRPageItemID;
var i: Integer;
begin
  Result := INVALID_PAGE_RECORD_NO;
  for i := PageItemNo-1 downto 0 do
   if (not CheckNullFlag(i,Page.PageData)) then
    begin
      Result := i;
      break;
    end;
end; // GetPriorRecordOnPage


//------------------------------------------------------------------------------
// return result for attempt of getting record relatively to first position
// and set RecordID to new record ID
//------------------------------------------------------------------------------
function TACRDiskRecordManager.GetRecordFromFirstPosition(
            SessionID:            TACRSessionID;
            GetRecordMode:        TACRGetRecordMode;
            var RecordID:         TACRRecordID
                                   ): TACRGetRecordResult;
var Page: TACRPage;
begin
 case GetRecordMode of
  grmPrior:
   begin
    Result := grrBOF;
   end;
  grmCurrent:
   begin
    Result := grrError;
   end;
  grmNext:
   begin
    Result := grrOk;
    RecordID.PageNo := FTablePageMapsManager.GetFirstRowBeginPageNo(SessionID);
    if (RecordID.PageNo = INVALID_PAGE_NO) then
      Result := grrEOF
    else
     begin
      if (FLargeRows) then
       RecordID.PageItemNo := 0
      else
       begin
        // find first record page item no in the page
        RecordID.PageItemNo := INVALID_PAGE_RECORD_NO;
        Result := grrOk;
        try
          Page := LPageManager.GetPage(SessionID,RecordID.PageNo,LTableData.IsSynchronizationRequired);
        except
          Result := grrError;
        end;
        if (Result = grrError) then
         Exit;
        RecordID.PageItemNo := GetFirstRecordOnPage(Page);
        LPageManager.PutPage(Page);
        if (RecordID.PageItemNo = INVALID_PAGE_RECORD_NO) then
         Result := grrError;
       end;
     end; // record found
   end; // next
 end; // GetRecordMode
end; // GetRecordFromFirstPosition


//------------------------------------------------------------------------------
// return result for attempt of getting record relatively to last position
// and set RecordID to new record ID
//------------------------------------------------------------------------------
function TACRDiskRecordManager.GetRecordFromLastPosition(
            SessionID:            TACRSessionID;
            GetRecordMode:        TACRGetRecordMode;
            var RecordID:         TACRRecordID
                                  ): TACRGetRecordResult;
var Page: TACRPage;
begin
 Result := grrError;
 case GetRecordMode of
  grmPrior:
   begin
    Result := grrOk;
    RecordID.PageNo := FTablePageMapsManager.GetLastRowBeginPageNo(SessionID);
    if (RecordID.PageNo = INVALID_PAGE_NO) then
      Result := grrBOF
    else
     begin
      if (FLargeRows) then
       RecordID.PageItemNo := 0
      else
       begin
        // find first record page item no in the page
        Result := grrOk;
        try
          Page := LPageManager.GetPage(SessionID,RecordID.PageNo,LTableData.IsSynchronizationRequired);
        except
          Result := grrError;
        end;
        if (Result = grrError) then
         Exit;
        RecordID.PageItemNo := GetLastRecordOnPage(Page);
        LPageManager.PutPage(Page);
        if (RecordID.PageItemNo = INVALID_PAGE_RECORD_NO) then
         Result := grrError;
       end;
     end; // record found
   end;
  grmCurrent:
   begin
    Result := grrError;
   end;
  grmNext:
   begin
    Result := grrEOF;
   end;
 end; // GetRecordMode
end; // GetRecordFromLastPosition


//------------------------------------------------------------------------------
// return result for attempt of getting record relatively any position
// and set RecordID to new record ID
//------------------------------------------------------------------------------
function TACRDiskRecordManager.GetRecordFromAnyPosition(
            SessionID:            TACRSessionID;
            GetRecordMode:        TACRGetRecordMode;
            var RecordID:         TACRRecordID
                                  ): TACRGetRecordResult;
var Page:         TACRPage;
    NewRecordID:  TACRRecordID;
begin
  Result := grrError;
  NewRecordID := RecordID;
  case GetRecordMode of
  grmPrior:
   begin
    NewRecordID.PageItemNo := INVALID_PAGE_RECORD_NO;
    if (not FLargeRows) then
     begin
      // try to find prior record on this page
      Result := grrOk;
      try
        Page := LPageManager.GetPage(SessionID,NewRecordID.PageNo,LTableData.IsSynchronizationRequired);
      except
        Result := grrError;
      end;
      if (Result = grrError) then
       Exit;
      NewRecordID.PageItemNo := GetPriorRecordOnPage(Page,RecordID.PageItemNo);
      LPageManager.PutPage(Page);
     end;
    // if prior record was not found on this page
    if (NewRecordID.PageItemNo = INVALID_PAGE_RECORD_NO) then
     begin
      NewRecordID.PageNo := FTablePageMapsManager.GetPriorRowBeginPageNo(SessionID,RecordID.PageNo);
      if (NewRecordID.PageNo = INVALID_PAGE_NO) then
       begin
         Result := grrBOF;
         Exit;
       end
      else
       begin
        if (FLargeRows) then
         NewRecordID.PageItemNo := 0
        else
         begin
          // try to find page item no
          Result := grrOk;
          try
            Page := LPageManager.GetPage(SessionID,NewRecordID.PageNo,LTableData.IsSynchronizationRequired);
          except
            Result := grrError;
          end;
          if (Result = grrError) then
           Exit;
          NewRecordID.PageItemNo := GetLastRecordOnPage(Page);
          LPageManager.PutPage(Page);
          if (NewRecordID.PageItemNo = INVALID_PAGE_RECORD_NO) then
           begin
            Result := grrError;
            Exit;
           end;
         end;
       end; // prior record page was found
     end; // new record id found
    Result := grrOK;
   end; // go to prior record
  grmCurrent:
   begin
    Result := grrOK;
   end; // go to current record
  grmNext:
   begin
    NewRecordID.PageItemNo := INVALID_PAGE_RECORD_NO;
    if (not FLargeRows) then
     begin
      // try to find prior record on this page
      Result := grrOk;
      try
        Page := LPageManager.GetPage(SessionID,NewRecordID.PageNo,LTableData.IsSynchronizationRequired);
      except
        Result := grrError;
      end;
      if (Result = grrError) then
       Exit;
      NewRecordID.PageItemNo := GetNextRecordOnPage(Page,RecordID.PageItemNo);
      LPageManager.PutPage(Page);
     end;
    // if prior record was not found on this page
    if (NewRecordID.PageItemNo = INVALID_PAGE_RECORD_NO) then
     begin
      NewRecordID.PageNo := FTablePageMapsManager.GetNextRowBeginPageNo(SessionID,RecordID.PageNo);
      if (NewRecordID.PageNo = INVALID_PAGE_NO) then
       begin
         Result := grrEOF;
         Exit;
       end
      else
       begin
        if (FLargeRows) then
         NewRecordID.PageItemNo := 0
        else
         begin
          // try to find page item no
          Result := grrOk;
          try
            Page := LPageManager.GetPage(SessionID,NewRecordID.PageNo,LTableData.IsSynchronizationRequired);
          except
            Result := grrError;
          end;
          if (Result = grrError) then
           Exit;
          NewRecordID.PageItemNo := GetFirstRecordOnPage(Page);
          LPageManager.PutPage(Page);
          if (NewRecordID.PageItemNo = INVALID_PAGE_RECORD_NO) then
           begin
            Result := grrError;
            Exit;
           end;
         end;
       end; // next record page was found
     end; // new record id found
    Result := grrOK;
   end; // go to next record
  end;
 if (Result = grrOK) then
  RecordID := NewRecordID;
end; // GetRecordFromAnyPosition


//------------------------------------------------------------------------------
// convert data in the memory to disk record buffer
// varchar and blobs links are not set
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.GetDiskRecordBufferFromRecordBuffer(
                    SessionID:     TACRSessionID;
                    RecordBuffer:  TACRRecordBuffer
                                                );
var i: Integer;
begin
  // copy null flags
  i := LTableData.FFieldManager.FieldDefs.Count div 8 +
       Integer((LTableData.FFieldManager.FieldDefs.Count mod 8) > 0);
  Move(RecordBuffer^,FDiskRecordBuffer^,i);
  for i := 0 to LTableData.FFieldManager.FieldDefs.Count - 1 do
   // if not null
   if (not CheckNullFlag(i,RecordBuffer)) then
    if ((not IsBLOBFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType)) and
        (not IsVarcharFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType))) then
     begin
      if (LTableData.FFieldManager.FieldDefs[i].MemoryDataSize <>
          LTableData.FFieldManager.FieldDefs[i].DiskDataSize) then
        raise EACRException.Create(10517,ErrorLInvalidFieldDataSize,
          [LTableData.FFieldManager.FieldDefs[i].Name,
           Integer(LTableData.FFieldManager.FieldDefs[i].BaseFieldType),
           Integer(LTableData.FFieldManager.FieldDefs[i].AdvancedFieldType),
           LTableData.FFieldManager.FieldDefs[i].MemoryDataSize,
           LTableData.FFieldManager.FieldDefs[i].DiskDataSize]);
      Move(PChar(RecordBuffer + LTableData.FFieldManager.FieldDefs[i].MemoryOffset)^,
           PChar(FDiskRecordBuffer + LTableData.FFieldManager.FieldDefs[i].DiskOffset)^,
           LTableData.FFieldManager.FieldDefs[i].MemoryDataSize);
     end;
  // process varchar and blob pages
  if (LTableData.FFieldManager.FieldDefs.VarcharOrBLOBFieldsExists) then
   for i := 0 to LTableData.FFieldManager.FieldDefs.Count - 1 do
    if (not CheckNullFlag(i,RecordBuffer)) then
     if ((IsBLOBFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType)) or
        (IsVarcharFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType))) then
            WriteBLOBOrVarcharValue(
              SessionID,
              LTableData.FFieldManager.FieldDefs[i],RecordBuffer); // process varchar and blob pages
end; // GetDiskRecordBufferFromRecordBuffer


//------------------------------------------------------------------------------
// copy data from disk record buffer to memory record buffer
// including varchars and blobs
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.CopyDataFromDiskRecordBufferToMemoryRecordBuffer(
            SessionID:     TACRSessionID;
            RecordBuffer:     TACRRecordBuffer);
var i: Integer;
begin
  // copy null flags
  i := LTableData.FFieldManager.FieldDefs.Count div 8 +
       Integer((LTableData.FFieldManager.FieldDefs.Count mod 8) > 0);
  Move(FDiskRecordBuffer^,RecordBuffer^,i);
  for i := 0 to LTableData.FFieldManager.FieldDefs.Count - 1 do
   // if not null
   if (not CheckNullFlag(i,RecordBuffer)) then
    if (not IsVarcharFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType)) then
     begin
      if (not IsBLOBFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType)) then
       begin
        if (LTableData.FFieldManager.FieldDefs[i].MemoryDataSize <>
            LTableData.FFieldManager.FieldDefs[i].DiskDataSize) then
          raise EACRException.Create(10520,ErrorLInvalidFieldDataSize,
            [LTableData.FFieldManager.FieldDefs[i].Name,
             Integer(LTableData.FFieldManager.FieldDefs[i].BaseFieldType),
             Integer(LTableData.FFieldManager.FieldDefs[i].AdvancedFieldType),
             LTableData.FFieldManager.FieldDefs[i].MemoryDataSize,
             LTableData.FFieldManager.FieldDefs[i].DiskDataSize]);
       end
      else
       LTableData.SetBLOBModified(False,RecordBuffer,
                                  LTableData.FFieldManager.FieldDefs[i]);
      Move(PChar(FDiskRecordBuffer + LTableData.FFieldManager.FieldDefs[i].DiskOffset)^,
           PChar(RecordBuffer + LTableData.FFieldManager.FieldDefs[i].MemoryOffset)^,
           LTableData.FFieldManager.FieldDefs[i].DiskDataSize);
     end;
  // process varchar and blob pages
  if (LTableData.FFieldManager.FieldDefs.VarcharOrBLOBFieldsExists) then
   begin
     for i := 0 to LTableData.FFieldManager.FieldDefs.Count - 1 do
      if (not CheckNullFlag(i,RecordBuffer)) then
       begin
//       if ((IsBLOBFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType)) or
{
        if (IsVarcharFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType)) then
           ReadBLOBOrVarcharValue(
            SessionID,
            LTableData.FFieldManager.FieldDefs[i],
            DiskRecordBuffer,
            RecordBuffer); // process varchar and blob pages
}
         try
          if (IsVarcharFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType)) then
           ReadBLOBOrVarcharValue(
            SessionID,
            LTableData.FFieldManager.FieldDefs[i],
            RecordBuffer); // process varchar and blob pages
         except
          if (not FRepair) then
           raise
          else
           SetNullFlag(True,i,RecordBuffer);
         end;

       end;
   end;
end; // CopyDataFromDiskRecordBufferToMemoryRecordBuffer


//------------------------------------------------------------------------------
// write blob value to pages
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.InternalWriteBLOBOrVarcharValue(
              SessionID:        TACRSessionID;
              BLOBDescriptor:   TACRPartialTemporaryBLOBDescriptor;
              PData:            PChar;
              IsVarchar:        Boolean;
              out RecordID:     TACRRecordID
                                         );
var
    DataPage:               TACRPage;
    NextDataPage:           TACRPage;
    DataSize:               Int64;
    Size:                   Word;
    PageFreeSpaceSize:      Word;
    PageType:               TACRTablePageType;
    i,PageCount,Offset:     Integer;
    ItemCount:              Word;
    ItemsMap:               TACRDataItemsMap;
    ItemID:                 Word;
    bOk:                    Boolean;
begin
  if (IsVarchar) then
   PageType := tptVarchar
  else
   PageType := tptBLOB;
  PageFreeSpaceSize := INVALID_PAGE_RECORD_NO;
  DataSize := BLOBDescriptor.CompressedSize + SizeOf(BLOBDescriptor);
  if (DataSize < Int64(LPageManager.PageDataSize -
                                      SizeOf(TACRDataItem) - SizeOf(Word))) then
   begin
    Size := Word(DataSize)+SizeOf(TACRDataItem);
    DataPage := FTablePageMapsManager.GetPageForAddingNewData(
                    SessionID,PageType,Size,PageFreeSpaceSize);
   end // try to find small page
  else
   DataPage := FTablePageMapsManager.AddPage(SessionID,PageType);
  RecordID.PageNo := DataPage.PageNo;
  if ((PageFreeSpaceSize = INVALID_PAGE_RECORD_NO) and
      (DataSize >= Int64(LPageManager.PageDataSize -
                                      SizeOf(TACRDataItem) - SizeOf(Word))))then
   begin
    // write large blob or varchar item
    RecordID.PageItemNo := 0;
    Offset := 0;
    PageCount := DataSize div LPageManager.PageDataSize +
                 Integer((DataSize mod LPageManager.PageDataSize) > 0);
    for i := 1 to PageCount do
     begin
      DataPage.PageHeader.ObjectID := LTableData.TableID;
      DataPage.PageHeader.PageType := ACRPageTypeIDTableDataLarge;
      if (i = 1) then
       begin
        Move(BLOBDescriptor,DataPage.PageData^,SizeOf(BLOBDescriptor));
        if (LPageManager.PageDataSize <= DataSize) then
         Size := LPageManager.PageDataSize - SizeOf(BLOBDescriptor)
        else
         Size := BLOBDescriptor.CompressedSize;
        Move(PData^,PChar(DataPage.PageData + SizeOf(BLOBDescriptor))^,Size);
       end
      else
       begin
        if (LPageManager.PageDataSize <=
              DataSize-Offset-SizeOf(BLOBDescriptor)) then
         Size := LPageManager.PageDataSize
        else
         Size := DataSize-Offset-SizeOf(BLOBDescriptor);
        Move(PChar(PData+Offset)^,DataPage.PageData^,Size);
       end;
      FTablePageMapsManager.SetPageRecordCount(SessionID,DataPage.PageNo,0,PageType);
      if (i = PageCount) then
       begin
        NextDataPage := nil;
        DataPage.PageHeader.NextPageNo := INVALID_PAGE_NO;
       end
      else
       begin
        NextDataPage := FTablePageMapsManager.AddPage(SessionID,PageType);
        DataPage.PageHeader.NextPageNo := NextDataPage.PageNo;
       end;
      DataPage.IsDirty := True;
      LPageManager.PutPage(DataPage);
      DataPage := NextDataPage;
      Inc(Offset,Size);
     end; // write pages
   end // write large blob or varchar item
  else
   begin
    Offset := 0; // offset to new data in DataPage
    if (PageFreeSpaceSize = INVALID_PAGE_RECORD_NO) then
     begin
      DataPage.PageHeader.NextPageNo := INVALID_PAGE_NO;
      DataPage.PageHeader.PageType := ACRPageTypeIDTableDataSmall;
      DataPage.PageHeader.ObjectID := LTableData.TableID;
      ItemCount := 1;
      SetLength(ItemsMap,ItemCount);
      ItemsMap[0].ItemID := 0;
      ItemsMap[0].DataSize := Word(DataSize);
      RecordID.PageItemNo := 0;
      PageFreeSpaceSize := Word(DataPage.PageDataSize - SizeOf(Word) -
                            SizeOf(ItemsMap[0]) - Integer(DataSize));
     end // new page
    else
     begin
      ItemsMap := LoadDataItemsMapFromPage(10535,DataPage,ItemCount);
      SetLength(ItemsMap,ItemCount+1);
      ItemID := 0;
      repeat
        bOk := True;
        for i := 0 to ItemCount-1 do
         if (ItemsMap[i].ItemID = ItemID) then
          begin
           bOk := False;
           Inc(ItemID);
           break;
          end;
      until (bOk);
      for i := 0 to ItemCount-1 do
       Inc(Offset,ItemsMap[i].DataSize);
      ItemsMap[ItemCount].ItemID := ItemID;
      ItemsMap[ItemCount].DataSize := Word(DataSize);
      RecordID.PageItemNo := ItemID;
      Inc(ItemCount);
      Dec(PageFreeSpaceSize,(Word(DataSize)+SizeOf(TACRDataItem)));
     end; // existing page
    // write data to page
    Move(BLOBDescriptor,PChar(DataPage.PageData+Offset)^,SizeOf(BLOBDescriptor));
    Inc(Offset,SizeOf(BLOBDescriptor));
    Move(PData^,PChar(DataPage.PageData+Offset)^,DataSize - SizeOf(BLOBDescriptor));
    SaveDataItemsMapToPage(DataPage,ItemsMap,ItemCount);
    FTablePageMapsManager.SetPageRecordCount(SessionID,DataPage.PageNo,PageFreeSpaceSize,PageType);
    DataPage.IsDirty := True;
    LPageManager.PutPage(DataPage);
    ItemsMap := nil;
   end; // write small blob or varchar item
end; // InternalWriteBLOBOrVarcharValue


//------------------------------------------------------------------------------
// write blob or varchar value
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.WriteBLOBOrVarcharValue(
              SessionID:        TACRSessionID;
              FieldDef:         TACRFieldDef;
              RecordBuffer:     TACRRecordBuffer
                                  );
var
    RecordID:       TACRRecordID;
    BLOBDescriptor: TACRPartialTemporaryBLOBDescriptor;
    InSize,OutSize: Integer;
    PData:          PChar;
    IsVarchar:      Boolean;
begin
  if (IsVarcharFieldType(FieldDef.BaseFieldType)) then
   begin
    IsVarchar := True;
    BLOBDescriptor.NumBlocks := -1;
    InSize := GetStrLength(PChar(RecordBuffer+FieldDef.MemoryOffset),
              FieldDef.AdvancedFieldType);
    if (FieldDef.BaseFieldType = bftVarchar) then
     Inc(InSize)
    else
     Inc(InSize,2);
    BLOBDescriptor.UncompressedSize := InSize;
    if (FieldDef.BLOBCompressionAlgorithm <> acaNone) then
     begin
      ACRInternalCompressBuffer(FieldDef.BLOBCompressionAlgorithm,
                                FieldDef.BLOBCompressionMode,
                                PChar(RecordBuffer+FieldDef.MemoryOffset),
                                InSize,
                                PData,
                                OutSize
                                );
      BLOBDescriptor.CompressedSize := OutSize;
     end // compression
    else
     begin
      BLOBDescriptor.CompressedSize := InSize;
      PData := PChar(RecordBuffer+FieldDef.MemoryOffset);
     end; // no compression
    InternalWriteBLOBOrVarcharValue(SessionID,BLOBDescriptor,PData,
                                    IsVarchar,RecordID);
    Move(RecordID,PChar(FDiskRecordBuffer + FieldDef.DiskOffset)^,SizeOf(RecordID));
    if (FieldDef.BLOBCompressionAlgorithm <> acaNone) then
      FreeMem(PData);
   end // Varchar
  else
   begin
    IsVarchar := False;
    if (LTableData.IsBLOBModified(RecordBuffer,FieldDef)) then
     begin
      Cardinal(PData) := PCardinal(RecordBuffer + FieldDef.MemoryOffset)^ ;
      BLOBDescriptor.CompressedSize := MemoryManager.GetMemoryBufferSize(PData)-
                                          SizeOf(TACRPartialBLOBDescriptor);
      BLOBDescriptor.NumBlocks := PACRPartialBLOBDescriptor(PData)^.NumBlocks;
      BLOBDescriptor.UncompressedSize := PACRPartialBLOBDescriptor(PData)^.UncompressedSize;
      PData := PChar(PData + SizeOf(TACRPartialBLOBDescriptor));
      InternalWriteBLOBOrVarcharValue(SessionID,BLOBDescriptor,PData,
                                      IsVarchar,RecordID);
      // free memory with new blob value
      Cardinal(PData) := PCardinal(RecordBuffer + FieldDef.MemoryOffset)^ ;
      MemoryManager.FreeAndNilMem(PData);
      // mark this record bufer as not modified to avoid destroying it once more
      Move(RecordID,PChar(RecordBuffer + FieldDef.MemoryOffset)^,SizeOf(RecordID));
      LTableData.SetBLOBModified(False,RecordBuffer,FieldDef);

      Move(RecordID,PChar(FDiskRecordBuffer + FieldDef.DiskOffset)^,SizeOf(RecordID));
     end
    else
      Move(PChar(RecordBuffer + FieldDef.MemoryOffset)^,
           PChar(FDiskRecordBuffer + FieldDef.DiskOffset)^,FieldDef.DiskDataSize);
   end; // BLOB
end; // WriteBLOBOrVarcharValue


//------------------------------------------------------------------------------
// read BLOB or Varchar value from disk
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.InternalReadBLOBOrVarcharValue(
              SessionID:            TACRSessionID;
              RecordID:             TACRRecordID;
              IsVarchar:            Boolean;
              var BLOBDescriptor:   TACRPartialTemporaryBLOBDescriptor;
              var PData:            PChar;
              out DataSize:         Integer
              );
var DataPage:     TACRPage;
    ItemsMap:     TACRDataItemsMap;
    ItemCount:    Word;
    i,Size:       Integer;
    bOk:          Boolean;
    Offset:       Integer;
    PageNo:       TACRPageNo;
    ItemSize:     Int64;
begin
  if (FRepair) then
   if (RecordID.PageNo < 0) or (RecordID.PageNo >= LPageManager.DBHeader.TotalPageCount) then
    raise EACRException.Create(10248,ErrorLInvalidPageNo,[PageNo]);

  DataPage := LPageManager.GetPage(SessionID,RecordID.PageNo,LTableData.IsSynchronizationRequired);
  if ((DataPage.PageHeader.PageType <> ACRPageTypeIDTableDataSmall) and
      (DataPage.PageHeader.PageType <> ACRPageTypeIDTableDataLarge)) then
   begin
    LPageManager.PutPage(DataPage);
    raise EACRException.Create(10536,ErrorLInvalidPageType,
      [DataPage.PageHeader.PageType,LTableData.TableName,RecordID.PageNo]);
   end;
  if (DataPage.PageHeader.PageType = ACRPageTypeIDTableDataSmall) then
   begin
    ItemsMap := LoadDataItemsMapFromPage(10537,DataPage,ItemCount);
    bOk := False;
    Offset := 0;
    for i := 0 to ItemCount - 1 do
      if (ItemsMap[i].ItemID = RecordID.PageItemNo) then
       begin
        bOk := True;
        ItemSize := ItemsMap[i].DataSize;
        break;
       end
      else
       Inc(Offset,ItemsMap[i].DataSize);
    if (not bOk) then
     raise EACRException.Create(10549,ErrorLLoadingRecord,
      [LTableData.TableName,SessionID,RecordID.PageNo,RecordID.PageItemNo]);
    if (PData = nil) then
     PData := MemoryManager.GetMem(ItemSize - SizeOf(BLOBDescriptor));
    Move(PChar(DataPage.PageData + Offset)^,BLOBDescriptor,SizeOf(BLOBDescriptor));
    Inc(Offset,SizeOf(BLOBDescriptor));
    Move(PChar(DataPage.PageData + Offset)^,PData^,ItemSize - SizeOf(BLOBDescriptor));
    DataSize := ItemSize - SizeOf(BLOBDescriptor);
    LPageManager.PutPage(DataPage);
   end // small page (multiple items)
  else
   begin
    Move(DataPage.PageData^,BLOBDescriptor,SizeOf(BLOBDescriptor));
    if (PData = nil) then
      PData := MemoryManager.GetMem(BLOBDescriptor.CompressedSize);
    Offset := SizeOf(BLOBDescriptor);
    if (BLOBDescriptor.CompressedSize + Offset <= DataPage.PageDataSize) then
     Size := BLOBDescriptor.CompressedSize
    else
     Size := DataPage.PageDataSize - Offset;
    Move(PChar(DataPage.PageData + Offset)^,PData^,Size);
    DataSize := Size;
    LPageManager.PutPage(DataPage);
    PageNo := DataPage.PageHeader.NextPageNo;
    while ((PageNo <> INVALID_PAGE_NO) and
          (DataSize < BLOBDescriptor.CompressedSize)) do
     begin
      if (FRepair) then
       if (PageNo < 0) or (PageNo >= LPageManager.DBHeader.TotalPageCount) then
        raise EACRException.Create(10247,ErrorLInvalidPageNo,[PageNo]);

      DataPage := LPageManager.GetPage(SessionID,PageNo,LTableData.IsSynchronizationRequired);
      if (BLOBDescriptor.CompressedSize - DataSize < DataPage.PageDataSize) then
       Size := BLOBDescriptor.CompressedSize - DataSize
      else
       Size := DataPage.PageDataSize;
      Move(DataPage.PageData^,PChar(PData + DataSize)^,Size);
      Inc(DataSize,Size);
      PageNo := DataPage.PageHeader.NextPageNo;
      LPageManager.PutPage(DataPage);
     end;
    if (DataSize <> BLOBDescriptor.CompressedSize) then
     raise EACRException.Create(10538,ErrorLInvalidDataSize,
            [DataSize,BLOBDescriptor.CompressedSize,
            LTableData.TableName,RecordID.PageNo]);
   end; // large page (single items)
end; // InternalReadBLOBOrVarcharValue


//------------------------------------------------------------------------------
// read blob or varchar value
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.ReadBLOBOrVarcharValue(
              SessionID:        TACRSessionID;
              FieldDef:         TACRFieldDef;
              RecordBuffer:     TACRRecordBuffer
                                  );
var
    RecordID:       TACRRecordID;
    BLOBDescriptor: TACRPartialTemporaryBLOBDescriptor;
    InSize,OutSize: Integer;
    PData,NewPData: PChar;
    IsVarchar:      Boolean;
    x:              Cardinal;
begin
  Move(PChar(FDiskRecordBuffer + FieldDef.DiskOffset)^,RecordID,SizeOf(RecordID));
  if (IsBLOBFieldType(FieldDef.BaseFieldType)) then
   begin
    IsVarchar := False;
    PData := nil;
   end // blob
  else
   begin
    IsVarchar := True;
    if (FieldDef.BLOBCompressionAlgorithm = acaNone) then
      PData := PChar(RecordBuffer + FieldDef.MemoryOffset)
    else
      PData := nil;
   end; // varchar
  InternalReadBLOBOrVarcharValue(SessionID,RecordID,IsVarchar,
    BLOBDescriptor,PData,InSize);
  if (IsVarchar) then
   begin
    if (FieldDef.BLOBCompressionAlgorithm <> acaNone) then
     begin
      Outsize := BLOBDescriptor.UncompressedSize;
      ACRInternalDecompressBuffer(FieldDef.BLOBCompressionAlgorithm,
        PData,InSize,NewPData,OutSize);
      Move(NewPData^,PChar(RecordBuffer + FieldDef.MemoryOffset)^,OutSize);
      MemoryManager.FreeAndNilMem(PData);
      FreeMem(NewPData);
     end;
   end // varchar
  else
   begin
    NewPData := MemoryManager.GetMem(BLOBDescriptor.CompressedSize +
                SizeOf(TACRPartialBLOBDescriptor));
    Move(PData^,PChar(NewPData + SizeOf(TACRPartialBLOBDescriptor))^,InSize);
    MemoryManager.FreeAndNilMem(PData);
    PACRPartialBLOBDescriptor(NewPData)^.NumBlocks := BLOBDescriptor.NumBlocks;
    PACRPartialBLOBDescriptor(NewPData)^.UncompressedSize := BLOBDescriptor.UncompressedSize;
    x := Cardinal(NewPData);
    Move(x,PChar(RecordBuffer + FieldDef.MemoryOffset)^,
          SizeOf(Cardinal));
   end; // blob
end; // ReadBLOBOrVarcharValue


//------------------------------------------------------------------------------
// delete blob or varchar value
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.DeleteBLOBOrVarcharValue(
              SessionID:        TACRSessionID;
              FieldDef:         TACRFieldDef
                                  );
var RecordID:       TACRRecordID;
    DataPage:       TACRPage;
    PageNo,
    NextPageNo:     TACRPageNo;
    Offset,i,ItemNo,
    Size,DataSize,
    OverallSize:    Integer;
    ItemCount:      Word;
    ItemsMap:       TACRDataItemsMap;
    PageType:       TACRTablePageType;
begin
  if (IsVarcharFieldType(FieldDef.BaseFieldType)) then
   PageType := tptVarchar
  else
   PageType := tptBLOB;
  Move(PChar(FDiskRecordBuffer + FieldDef.DiskOffset)^,RecordID,SizeOf(RecordID));
  PageNo := RecordID.PageNo;
  DataPage := LPageManager.GetPage(SessionID,PageNo,LTableData.IsSynchronizationRequired);
  if ((DataPage.PageHeader.PageType <> ACRPageTypeIDTableDataSmall) and
      (DataPage.PageHeader.PageType <> ACRPageTypeIDTableDataLarge)) then
   begin
    LPageManager.PutPage(DataPage);
    raise EACRException.Create(10544,ErrorLInvalidPageType,
      [DataPage.PageHeader.PageType,LTableData.TableName,PageNo]);
   end;
  if (DataPage.PageHeader.PageType = ACRPageTypeIDTableDataSmall) then
   begin
    ItemsMap := LoadDataItemsMapFromPage(10545,DataPage,ItemCount);
    try
      if (ItemCount = 1) then
       begin
        FTablePageMapsManager.RemovePage(SessionID,PageNo,PageType);
        LPageManager.PutPage(DataPage);
        LPageManager.RemovePage(SessionID,PageNo);
       end // delete last item and page
      else
       begin
        Offset := 0;
        Size := 0;
        DataSize := 0;
        OverallSize := 0;
        ItemNo := -1;
        for i := 0 to ItemCount-1 do
         begin
          if (ItemsMap[i].ItemID <> RecordID.PageItemNo) then
            Inc(OverallSize,ItemsMap[i].DataSize);
//          if (ItemsMap[i].ItemID <= RecordID.PageItemNo) then
          if (ItemNo = -1) then
           Inc(Offset,ItemsMap[i].DataSize);
//          if (ItemsMap[i].ItemID > RecordID.PageItemNo) then
          if (ItemsMap[i].ItemID = RecordID.PageItemNo) then
           begin
            ItemNo := i;
            DataSize := ItemsMap[i].DataSize;
           end
          else
          if (ItemNo <> -1) then
           Inc(Size,ItemsMap[i].DataSize);
         end;
        if (ItemNo = -1) then
         begin
          LPageManager.PutPage(DataPage);
          raise EACRException.Create(10546,ErrorLDeleteRecordInvalidDataItem,
            [LTableData.TableName,FieldDef.Name,
             RecordID.PageNo,RecordID.PageItemNo,ItemCount]);
         end;
        if (DataSize = 0) then
         begin
          LPageManager.PutPage(DataPage);
          raise EACRException.Create(10546,ErrorLDeleteRecordInvalidDataItem,
            [LTableData.TableName,FieldDef.Name,
             RecordID.PageNo,RecordID.PageItemNo,ItemCount]);
         end;
        Move(PChar(DataPage.PageData + Offset)^,
             PChar(DataPage.PageData + Offset - DataSize)^,
             Size);
        if (ItemNo < ItemCount-1) then
         Move(ItemsMap[ItemNo+1],ItemsMap[ItemNo],
              SizeOf(TACRDataItem) * (ItemCount - 1 - ItemNo));
        Dec(ItemCount);
        SaveDataItemsMapToPage(DataPage,ItemsMap,ItemCount);
        Inc(OverallSize, SizeOf(TACRDataItem) * ItemCount + SizeOf(Word));
        FTablePageMapsManager.SetPageRecordCount(SessionID,PageNo,
          (DataPage.PageDataSize - OverallSize),PageType);
        DataPage.IsDirty := True;
        LPageManager.PutPage(DataPage);
       end; // delete this item from the page
    finally
      ItemsMap := nil;
    end;
   end // small page (multiple items on page)
  else
   begin
    repeat
     FTablePageMapsManager.RemovePage(SessionID,PageNo,PageType);
     NextPageNo := DataPage.PageHeader.NextPageNo;
     LPageManager.PutPage(DataPage);
     LPageManager.RemovePage(SessionID,PageNo);
     PageNo := NextPageNo;
     if (PageNo <> INVALID_PAGE_NO) then
      DataPage := LPageManager.GetPage(SessionID,PageNo,LTableData.IsSynchronizationRequired);
    until (PageNo = INVALID_PAGE_NO);
   end; // large page (single item on page)
end; // DeleteBLOBOrVarcharValue


//------------------------------------------------------------------------------
// load record from disk into new allocated buffer
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.LoadDiskRecord(
              SessionID:        TACRSessionID;
              RecordID:         TACRRecordID;
              DiskRecordBuffer: TACRRecordBuffer
                       );
begin
  if (FLargeRows) then
    InternalLoadDiskRecordLarge(SessionID,RecordID,DiskRecordBuffer)
  else
    InternalLoadDiskRecordSmall(SessionID,RecordID,DiskRecordBuffer);
end; // LoadDiskRecord


//------------------------------------------------------------------------------
// load record from disk into new allocated buffer
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.InternalLoadDiskRecordLarge(
              SessionID:        TACRSessionID;
              RecordID:         TACRRecordID;
              DiskRecordBuffer: TACRRecordBuffer
                       );
var Offset,Size: Integer;
    RowPage:       TACRPage;
    PageNo:        TACRPageNo;
begin
  PageNo := RecordID.PageNo;
  Offset := 0;
  repeat
   RowPage := LPageManager.GetPage(SessionID,PageNo,LTableData.IsSynchronizationRequired);
   try
     PageNo := RowPage.PageHeader.NextPageNo;
     if (PageNo = INVALID_PAGE_NO) then
      Size := FDiskRecordBufferSize - Offset
     else
      Size := RowPage.PageDataSize;
     if (Offset + Size > FDiskRecordBufferSize) then
      raise EACRException.Create(10524,ErrorLInvalidDiskRecordSize,
        [LTableData.TableName,Offset,Size,FDiskRecordBufferSize,
         RowPage.PageNo,PageNo]);
     Move(RowPage.PageData^,PChar(DiskRecordBuffer+Offset)^,Size);
     Inc(Offset,Size);
   finally
     LPageManager.PutPage(RowPage);
   end;
  until (PageNo = INVALID_PAGE_NO);
end; // LoadDiskRecordLarge


//------------------------------------------------------------------------------
// load record from disk into new allocated buffer
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.InternalLoadDiskRecordSmall(
              SessionID:        TACRSessionID;
              RecordID:         TACRRecordID;
              DiskRecordBuffer: TACRRecordBuffer
                       );
var Offset:   Integer;
    RowPage:  TACRPage;
begin
  RowPage := LPageManager.GetPage(SessionID,RecordID.PageNo,LTableData.IsSynchronizationRequired);
  try
    if (CheckNullFlag(RecordID.PageItemNo,RowPage.PageData)) then
      raise EACRException.Create(10525,ErrorLDiskRecordIsDeleted,
        [LTableData.TableName,RecordID.PageNo,RecordID.PageItemNo]);
    Offset := FRecordsPerPage div 8 + Integer((FRecordsPerPage mod 8) > 0) +
              RecordID.PageItemNo * FDiskRecordBufferSize;
    if (Offset + FDiskRecordBufferSize > RowPage.PageDataSize) then
      raise EACRException.Create(10526,ErrorLInvalidDiskRecordSize,
        [LTableData.TableName,Offset,FDiskRecordBufferSize,RowPage.PageDataSize,
         RecordID.PageNo,RecordID.PageItemNo]);
    Move(PChar(RowPage.PageData + Offset)^,
         DiskRecordBuffer^,FDiskRecordBufferSize);
  finally
    LPageManager.PutPage(RowPage);
  end;
end; // LoadDiskRecordSmall


//------------------------------------------------------------------------------
// return true if record exists
//------------------------------------------------------------------------------
function TACRDiskRecordManager.InternalIsRecordExistsLarge(
              SessionID:        TACRSessionID;
              var RecordID:     TACRRecordID
                       ): Boolean;
var
    PageRecordCount:  Word;
    TablePageType:    TACRTablePageType;
begin
  FTablePageMapsManager.GetPageTypeAndRecordCount(SessionID,RecordID.PageNo,
    TablePageType,PageRecordCount);
  if (TablePageType <> tptRowStart) then
   Result := False
  else
   Result := (PageRecordCount = 1); // in large mode only 1 record can be in the page
end; // InternalIsRecordExistsLarge


//------------------------------------------------------------------------------
// return true if record exists
//------------------------------------------------------------------------------
function TACRDiskRecordManager.InternalIsRecordExistsSmall(
              SessionID:        TACRSessionID;
              var RecordID:     TACRRecordID
                       ): Boolean;
var
    PageRecordCount:  Word;
    TablePageType:    TACRTablePageType;
    RowPage:          TACRPage;
begin
  Result := False;
  FTablePageMapsManager.GetPageTypeAndRecordCount(SessionID,RecordID.PageNo,
    TablePageType,PageRecordCount);
  if ((TablePageType = tptRowStart) and
      (PageRecordCount < INVALID_PAGE_RECORD_NO) and
      (PageRecordCount > 0)) then
   begin
    RowPage := LPageManager.GetPage(SessionID,RecordID.PageNo,LTableData.IsSynchronizationRequired);
    try
      Result := not CheckNullFlag(RecordID.PageItemNo,RowPage.PageData);
    finally
      LPageManager.PutPage(RowPage);
    end;
   end; // there are some records in this page
end; // InternalIsRecordExistsSmall


//------------------------------------------------------------------------------
// add record and return its number (RecordSize > PageDataSize div 2)
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.InternalAddRecordLarge(
                   RecordBuffer:  TACRRecordBuffer;
                   var RecordID:  TACRRecordID;
                   SessionID:     TACRSessionID
                  );
var
    i,PageCount,
     Size, Offset:      Integer;
    RowPage:            TACRPage;
    NextRowPage:        TACRPage;
begin
  GetDiskRecordBufferFromRecordBuffer(SessionID,RecordBuffer);
  // write row to pages
  PageCount := FDiskRecordBufferSize div LPageManager.PageDataSize +
                Integer((FDiskRecordBufferSize mod LPageManager.PageDataSize) > 0);
  Offset := 0;
  for i := 1 to PageCount do
   begin
    if (i = 1) then
     begin
      RowPage := FTablePageMapsManager.AddPage(SessionID,tptRowStart);
      RecordID.PageNo := RowPage.PageNo;
      RecordID.PageItemNo := 0;
     end;
    if (Offset + LPageManager.PageDataSize <= FDiskRecordBufferSize) then
     Size := LPageManager.PageDataSize
    else
     Size := FDiskRecordBufferSize - Offset;
    Move(PChar(FDiskRecordBuffer + Offset)^,RowPage.PageData^,Size);
    if (i = 1) then
      FTablePageMapsManager.SetPageRecordCount(SessionID,RowPage.PageNo,1,tptRowStart)
    else
      FTablePageMapsManager.SetPageRecordCount(SessionID,RowPage.PageNo,0,tptRowContinue);
    if (i = PageCount) then
     begin
      NextRowPage := nil;
      RowPage.PageHeader.NextPageNo := INVALID_PAGE_NO;
     end
    else
     begin
      NextRowPage := FTablePageMapsManager.AddPage(SessionID,tptRowContinue);
      RowPage.PageHeader.NextPageNo := NextRowPage.PageNo;
     end;
    RowPage.PageHeader.ObjectID := LTableData.TableID;
    RowPage.PageHeader.PageType := ACRPageTypeIDTableRows;
    RowPage.IsDirty := True;
    LPageManager.PutPage(RowPage);
    RowPage := NextRowPage;
    Inc(Offset,Size);
   end;
end; // InternalAddRecordLarge


//------------------------------------------------------------------------------
// add record and return its number (RecordSize <= PageDataSize div 2)
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.InternalAddRecordSmall(
                   RecordBuffer:  TACRRecordBuffer;
                   var RecordID:  TACRRecordID;
                   SessionID:     TACRSessionID
                  );
var
    i, Size, Offset:    Integer;
    RowPage:            TACRPage;
    PageRecordCount:    Word;
begin
  GetDiskRecordBufferFromRecordBuffer(SessionID,RecordBuffer);
  // write row to page
  RowPage := FTablePageMapsManager.GetPageForAddingNewData(
                  SessionID,tptRowStart,FRecordsPerPage,
                  PageRecordCount);
  RowPage.IsDirty := True;
  Size := FRecordsPerPage div 8 + Integer((FRecordsPerPage mod 8) > 0);
  if (PageRecordCount = INVALID_PAGE_RECORD_NO) then
   begin
    // mark all records as empty
    FillChar(RowPage.PageData^,Size,$FF);
   end;
  RowPage.PageHeader.ObjectID := LTableData.TableID;
  RowPage.PageHeader.PageType := ACRPageTypeIDTableRows;
  RecordID.PageNo := RowPage.PageNo;
  RecordID.PageItemNo := INVALID_PAGE_RECORD_NO;
  for i := 0 to FRecordsPerPage-1 do
   if (CheckNullFlag(i,RowPage.PageData)) then
    begin
     RecordID.PageItemNo := i;
     break;
    end;
  if (RecordID.PageItemNo = INVALID_PAGE_RECORD_NO) then
   begin
    LPageManager.PutPage(RowPage);
    raise EACRException.Create(10527, ErrorLCannotAddDiskRecordSmall,
      [LTableData.TableName,RecordID.PageNo]);
   end;
  if ((PageRecordCount <> INVALID_PAGE_RECORD_NO) and
      (PageRecordCount >= FRecordsPerPage)) then
   begin
    LPageManager.PutPage(RowPage);
    raise EACRException.Create(10529,ErrorLInvalidPageRecordCount,
      [LTableData.TableName,PageRecordCount,FRecordsPerPage]);
   end;
  RowPage.PageHeader.NextPageNo := INVALID_PAGE_NO;
  if (PageRecordCount = INVALID_PAGE_RECORD_NO) then
   begin
    // new empty page was allocated
    PageRecordCount := 1;
   end
  else
    Inc(PageRecordCount);
  SetNullFlag(False,RecordID.PageItemNo,RowPage.PageData);
  Offset := Size + RecordID.PageItemNo * FDiskRecordBufferSize;
  if (Offset + FDiskRecordBufferSize > RowPage.PageDataSize) then
    begin
      LPageManager.PutPage(RowPage);
      raise EACRException.Create(10528, ErrorLInvalidDiskRecordSize,
      [LTableData.TableName,Offset,RowPage.PageDataSize,FDiskRecordBufferSize,
         RowPage.PageNo,RowPage.PageHeader.NextPageNo]);
    end;
  Move(FDiskRecordBuffer^,PChar(RowPage.PageData + Offset)^,FDiskRecordBufferSize);
  FTablePageMapsManager.SetPageRecordCount(SessionID,RowPage.PageNo,PageRecordCount,tptRowStart);
  LPageManager.PutPage(RowPage);
end; // InternalAddRecordSmall


//------------------------------------------------------------------------------
// delete record, return true if record was deleted, false if record was deleted earlier
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.InternalDeleteRecordSmall(
                      var RecordID: TACRRecordID;
                      SessionID:    TACRSessionID
                     );
var RowPage:      TACRPage;
    RecordCount:  Word;
begin
  RecordCount := FTablePageMapsManager.GetPageRecordCount(SessionID,RecordID.PageNo);
  if (RecordCount = 0) then
    raise EACRException.Create(10539,ErrorLInvalidPageRecordCount,
      [LTableData.TableName,RecordCount,FRecordsPerPage]);
  if (RecordCount = INVALID_PAGE_RECORD_NO) then
    raise EACRException.Create(10540,ErrorLInvalidPageRecordCount,
      [LTableData.TableName,RecordCount,FRecordsPerPage]);
  if (RecordCount = 1) then
   begin
    FTablePageMapsManager.RemovePage(SessionID,RecordID.PageNo,tptRowStart);
    LPageManager.RemovePage(SessionID,RecordID.PageNo);
   end // last record will be deleted - empty page
  else
   begin
    FTablePageMapsManager.SetPageRecordCount(SessionID,RecordID.PageNo,(RecordCount-1),tptRowStart);
    RowPage := LPageManager.GetPage(SessionID,RecordID.PageNo,LTableData.IsSynchronizationRequired);
    try
      SetNullFlag(True,RecordID.PageItemNo,RowPage.PageData);
      RowPage.IsDirty := True;
    finally
      LPageManager.PutPage(RowPage);
    end;
   end;
end; // InternalDeleteRecordSmall


//------------------------------------------------------------------------------
// delete record, return true if record was deleted, false if record was deleted earlier
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.InternalDeleteRecordLarge(
                      var RecordID: TACRRecordID;
                      SessionID:    TACRSessionID
                     );
var PageNo:       TACRPageNo;
    RecordCount:  Word;
    RowPage:      TACRPage;
    NextPageNo:   TACRPageNo;
    DataSize:     Integer;
begin
  RecordCount := FTablePageMapsManager.GetPageRecordCount(SessionID,RecordID.PageNo);
  if (RecordCount <> 1) then
    raise EACRException.Create(10541,ErrorLInvalidPageRecordCount,
      [LTableData.TableName,RecordCount,FRecordsPerPage]);
  PageNo := RecordID.PageNo;
  DataSize := 0;
  while (PageNo <> INVALID_PAGE_NO) do
   begin
    if (DataSize > FDiskRecordBufferSize) then
     raise EACRException.Create(10542,ErrorLInvalidDataSize,
        [LTableData.TableName,RecordID.PageNo,PageNo]);
    try
      RowPage := LPageManager.GetPage(SessionID,PageNo,LTableData.IsSynchronizationRequired);
    except
      raise EACRException.Create(10543,ErrorLDeleteRecordInvalidDataSize,
        [LTableData.TableName,RecordID.PageNo,PageNo,DataSize,FDiskRecordBufferSize]);
    end;
    if (PageNo = RecordID.PageNo) then
      FTablePageMapsManager.RemovePage(SessionID,PageNo,tptRowStart)
    else
      FTablePageMapsManager.RemovePage(SessionID,PageNo,tptRowContinue);
    NextPageNo := RowPage.PageHeader.NextPageNo;
    Inc(DataSize,RowPage.PageDataSize);
    LPageManager.PutPage(RowPage);
    LPageManager.RemovePage(SessionID,PageNo);
    PageNo := NextPageNo;
   end;
end; // InternalDeleteRecordLarge


//------------------------------------------------------------------------------
// update record (RecordSize > PageDataSize div 2)
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.InternalUpdateRecordLarge(
                   RecordBuffer:  TACRRecordBuffer;
                   var RecordID:  TACRRecordID;
                   SessionID:     TACRSessionID
                  );
var RowPage:          TACRPage;
    PageNo:           TACRPageNo;
    Offset:           Integer;
    Size:             Integer;
begin
  GetDiskRecordBufferFromRecordBuffer(SessionID,RecordBuffer);
  PageNo := RecordID.PageNo;
  Offset := 0;
  while (Offset < FDiskRecordBufferSize) do
   begin
    RowPage := LPageManager.GetPage(SessionID,PageNo,LTableData.IsSynchronizationRequired);
    PageNo := RowPage.PageHeader.NextPageNo;
    if ((FDiskRecordBufferSize - Offset) < RowPage.PageDataSize) then
     Size := FDiskRecordBufferSize - Offset
    else
     Size := RowPage.PageDataSize;
    Move(PChar(FDiskRecordBuffer + Offset)^,RowPage.PageData^,Size);
    RowPage.IsDirty := True;
    LPageManager.PutPage(RowPage);
    Inc(Offset,Size);
   end;
end; // InternalUpdateRecordLarge


//------------------------------------------------------------------------------
// update record (RecordSize <= PageDataSize div 2)
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.InternalUpdateRecordSmall(
                   RecordBuffer:  TACRRecordBuffer;
                   var RecordID:  TACRRecordID;
                   SessionID:     TACRSessionID
                  );
var RowPage:          TACRPage;
    Offset:           Integer;
begin
  GetDiskRecordBufferFromRecordBuffer(SessionID,RecordBuffer);
  RowPage := LPageManager.GetPage(SessionID,RecordID.PageNo,LTableData.IsSynchronizationRequired);
  Offset := FRecordsPerPage div 8 + Integer((FRecordsPerPage mod 8) > 0) +
    RecordID.PageItemNo * FDiskRecordBufferSize;
  if (Offset + FDiskRecordBufferSize > RowPage.PageDataSize) then
   begin
    LPageManager.PutPage(RowPage);
    raise EACRException.Create(10547,ErrorLInvalidDiskRecordSize,
      [LTableData.TableName,Offset,FDiskRecordBufferSize,RecordID.PageNo,
       INVALID_PAGE_NO]);
    end;
  Move(FDiskRecordBuffer^,PChar(RowPage.PageData + Offset)^,
    FDiskRecordBufferSize);
  RowPage.IsDirty := True;
  LPageManager.PutPage(RowPage);
end; // InternalUpdateRecordSmall


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRDiskRecordManager.Create(
                        TablePFSMapFileStartPageNo: TACRPageNo;
                        RecordBufferSize:           Integer;
                        DiskRecordBufferSize:       Integer;
                        aPageManager:               TACRDiskPageManager;
                        aTableData:                 TACRDiskTableData
                      );
begin
 InitializeCriticalSection(FCSect);
 FRepair := False;
 FRecordBufferSize := RecordBufferSize;
 FDiskRecordBufferSize := DiskRecordBufferSize;
 FDiskRecordBuffer := MemoryManager.GetMem(DiskRecordBufferSize);
 LPageManager := aPageManager;
 LTableData := aTableData;
 FTablePageMapsManager := TACRTablePageMapsManager.Create(
                                LTableData,
                                LTableData.FRandomSearchRetryCount,
                                LTableData.IsSynchronizationRequired);
 FTablePageMapsManager.OpenTablePageMapsManager(TablePFSMapFileStartPageNo,LTableData.Exclusive);
 FLargeRows := (FDiskRecordBufferSize > ((LPageManager.PageDataSize-1) div 2));
 if (FLargeRows) then
  FRecordsPerPage := 1
 else
  begin
   FRecordsPerPage := LPageManager.PageDataSize div FDiskRecordBufferSize;
   while (FRecordsPerPage * FDiskRecordBufferSize +
          ((FRecordsPerPage div 8) + Integer((FRecordsPerPage mod 8) > 0)) >
          LPageManager.PageDataSize) do
    Dec(FRecordsPerPage);
   if (FRecordsPerPage <= 1) then
    raise EACRException.Create(10522,ErrorLCannotCalculateRecordsPerPage,
      [FRecordsPerPage,FDiskRecordBufferSize,LPageManager.PageDataSize]);
  end;
end; // Create;


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRDiskRecordManager.Destroy;
begin
  DeleteCriticalSection(FCSect);
  FTablePageMapsManager.Free;
  MemoryManager.FreeAndNilMem(FDiskRecordBuffer);
  inherited Destroy;
end; // Destroy


//------------------------------------------------------------------------------
// return record count
//------------------------------------------------------------------------------
function TACRDiskRecordManager.GetRecordCount: TACRRecordNo;
begin
  Lock;
  try
    Result := FTablePageMapsManager.GetRecordCount;
  finally
    Unlock;
  end;
end;


//------------------------------------------------------------------------------
// empty table (delete all records)
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.Empty(SessionID: TACRSessionID);
begin
  FTablePageMapsManager.Empty(SessionID);
end; // Empty


//------------------------------------------------------------------------------
// add record and return its number
//------------------------------------------------------------------------------
function TACRDiskRecordManager.AddRecord(
                   RecordBuffer:  TACRRecordBuffer;
                   var RecordID:  TACRRecordID;
                   SessionID:     TACRSessionID
                  ): Boolean;
begin
 Result := True;
 try
  if (FLargeRows) then
    InternalAddRecordLarge(RecordBuffer,RecordID,SessionID)
  else
    InternalAddRecordSmall(RecordBuffer,RecordID,SessionID);
 except
  raise;
 end;
end; // AddRecord


//------------------------------------------------------------------------------
// update record, return true if record was updated, false if record was deleted
//------------------------------------------------------------------------------
function TACRDiskRecordManager.UpdateRecord(
                      RecordBuffer: TACRRecordBuffer;
                      RecordID:     TACRRecordID;
                      SessionID:    TACRSessionID
                     ): Boolean;
var
    i:                    Integer;
begin
 Result := True;
 try
  if (LTableData.FieldManager.FieldDefs.VarcharOrBLOBFieldsExists) then
   begin
    LoadDiskRecord(SessionID,RecordID,FDiskRecordBuffer);
    // delete old blob values
    for i := 0 to LTableData.FieldManager.FieldDefs.Count-1 do
     if (not CheckNullFlag(i,FDiskRecordBuffer)) then
       begin
        if (IsBLOBFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType)) then
         if (LTableData.IsBLOBModified(RecordBuffer,LTableData.FFieldManager.FieldDefs[i])) then
          DeleteBLOBOrVarcharValue(SessionID,LTableData.FFieldManager.FieldDefs[i]);
        if (IsVarcharFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType)) then
          DeleteBLOBOrVarcharValue(SessionID,LTableData.FFieldManager.FieldDefs[i]);
       end;
   end; // delete old blob and varchar values
  if (FLargeRows) then
   InternalUpdateRecordLarge(RecordBuffer,RecordID,SessionID)
  else
   InternalUpdateRecordSmall(RecordBuffer,RecordID,SessionID);
 except
  raise;
 end;
end; // UpdateRecord


//------------------------------------------------------------------------------
// delete record, return true if record was deleted, false if record was deleted earlier
//------------------------------------------------------------------------------
function TACRDiskRecordManager.DeleteRecord(
                      var RecordID: TACRRecordID;
                      SessionID:    TACRSessionID
                     ): Boolean;
var 
    i:                Integer;
begin
 Result := True;
 try
  if (LTableData.FieldManager.FieldDefs.VarcharOrBLOBFieldsExists) then
   begin
    LoadDiskRecord(SessionID,RecordID,FDiskRecordBuffer);
    for i := 0 to LTableData.FieldManager.FieldDefs.Count - 1 do
     if (not CheckNullFlag(i,FDiskRecordBuffer)) then
      if ((IsBLOBFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType)) or
          (IsVarcharFieldType(LTableData.FFieldManager.FieldDefs[i].BaseFieldType))) then
        DeleteBLOBOrVarcharValue(SessionID, LTableData.FieldManager.FieldDefs[i]);
   end; // delete varchar and blobs
  if (FLargeRows) then
    InternalDeleteRecordLarge(RecordID,SessionID)
  else
    InternalDeleteRecordSmall(RecordID,SessionID);
 except
  raise;
 end;
end; // DeleteRecord


//------------------------------------------------------------------------------
// return true if record exists
//------------------------------------------------------------------------------
function TACRDiskRecordManager.IsRecordExists(var RecordID: TACRRecordID; SessionID: TACRSessionID): Boolean;
begin
  Lock;
  try
    Result := False;
    if (GetRecordCount > 0) then
     begin
      if (FLargeRows) then
       Result := InternalIsRecordExistsLarge(SessionID,RecordID)
      else
       Result := InternalIsRecordExistsSmall(SessionID,RecordID);
     end;
  finally
    Unlock;
  end;
end; // IsRecordExists


//------------------------------------------------------------------------------
// get record
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.GetRecordBuffer(var NavigationInfo: TACRNavigationInfo);
begin
  Lock;
  try
     // get record relatively to the first position
     if (NavigationInfo.FirstPosition) then
      begin
       NavigationInfo.GetRecordResult := GetRecordFromFirstPosition(
                                          NavigationInfo.SessionID,
                                          NavigationInfo.GetRecordMode,
                                          NavigationInfo.RecordID
                                                                   );
       if (NavigationInfo.GetRecordResult = grrOK) then
        NavigationInfo.FirstPosition := False;
      end
     else
     // get record relatively to the last position
     if (NavigationInfo.LastPosition) then
      begin
       NavigationInfo.GetRecordResult := GetRecordFromLastPosition(
                                          NavigationInfo.SessionID,
                                          NavigationInfo.GetRecordMode,
                                          NavigationInfo.RecordID
                                                                   );
       if (NavigationInfo.GetRecordResult = grrOK) then
        NavigationInfo.LastPosition := False;
      end
     else
      NavigationInfo.GetRecordResult := GetRecordFromAnyPosition(
                                          NavigationInfo.SessionID,
                                          NavigationInfo.GetRecordMode,
                                          NavigationInfo.RecordID
                                                                   );
     if (NavigationInfo.GetRecordResult = grrOK) then
      begin
        if (not LTableData.FieldManager.FieldDefs.VarcharOrBLOBFieldsExists) then
         begin
          // disk record buffer is equal to memory record buffer - load it directly
          try
            LoadDiskRecord(NavigationInfo.SessionID, NavigationInfo.RecordID,
                           NavigationInfo.RecordBuffer);
          except
           raise EACRException.Create(10518,ErrorLLoadingRecord,
                  [LTableData.TableName,NavigationInfo.SessionID,
                   NavigationInfo.RecordID.PageNo,
                   NavigationInfo.RecordID.PageItemNo
                   ])
          end;
         end
        else
         try
          LoadDiskRecord(NavigationInfo.SessionID, NavigationInfo.RecordID,
                         FDiskRecordBuffer);
          CopyDataFromDiskRecordBufferToMemoryRecordBuffer(
                                            NavigationInfo.SessionID,
                                            NavigationInfo.RecordBuffer
                                                           );
         except
           raise EACRException.Create(10519,ErrorLLoadingRecord,
                  [LTableData.TableName,NavigationInfo.SessionID,
                   NavigationInfo.RecordID.PageNo,
                   NavigationInfo.RecordID.PageItemNo
                   ])
         end;
      end; // record retrieved successfully
  finally
    Unlock;
  end;
end; // GetRecordBuffer


//------------------------------------------------------------------------------
// return approximate record no
//------------------------------------------------------------------------------
function TACRDiskRecordManager.GetApproximateRecNo(RecordID: TACRRecordID; SessionID: TACRSessionID): TACRRecordNo;
begin
  Result := FTablePageMapsManager.GetRecNoByRecordID(SessionID,RecordID)+1;
end; // GetApproximateRecNo


//------------------------------------------------------------------------------
// return exact RecNo
//------------------------------------------------------------------------------
function TACRDiskRecordManager.GetRecNoByRecordID(RecordID: TACRRecordID; SessionID: TACRSessionID): TACRRecordNo;
var RowPage: TACRPage;
    i,k:     Integer;
begin
  Result := FTablePageMapsManager.GetRecNoByRecordID(SessionID,RecordID);
  if (Result > -1) and (FLargeRows) then
   Inc(Result)
  else
  if (Result > -1) and (not FLargeRows) then
   begin
    RowPage := LPageManager.GetPage(SessionID,RecordID.PageNo,LTableData.IsSynchronizationRequired);
    try
      k := 0;
      for i := 0 to FRecordsPerPage-1 do
       begin
        if (not CheckNullFlag(i,RowPage.PageData)) then
          begin
           Inc(k);
           if (i >= RecordID.PageItemNo) then
            break;
          end;
       end;
      Inc(Result,k);
    finally
      LPageManager.PutPage(RowPage);
    end;
   end;
end; // GetRecNoByRecordID


//------------------------------------------------------------------------------
// return record ID
//------------------------------------------------------------------------------
function TACRDiskRecordManager.GetRecordIDByRecNo(RecNo: TACRRecordNo; SessionID: TACRSessionID): TACRRecordID;
var ItemNo:   TACRPageItemID;
    RowPage:  TACRPage;
    i,k:      Integer;
begin
  Result := FTablePageMapsManager.GetRecordIDByRecNo(SessionID,RecNo-1);
  if (Result.PageNo <> INVALID_PAGE_NO) and (not FLargeRows) then
   begin
    ItemNo := Result.PageItemNo;
    RowPage := LPageManager.GetPage(SessionID,Result.PageNo,LTableData.IsSynchronizationRequired);
    try
      Result.PageItemNo := INVALID_PAGE_RECORD_NO;
      k := 0;
      for i := 0 to FRecordsPerPage-1 do
       begin
        if (not CheckNullFlag(i,RowPage.PageData)) then
         begin
          Inc(k);
          if (k > ItemNo) then
           begin
            Result.PageItemNo := TACRPageItemID(i);
            break;
           end;
         end;
       end;
      if (Result.PageItemNo = INVALID_PAGE_RECORD_NO) then
       Result.PageNo := INVALID_PAGE_NO;
    finally
      LPageManager.PutPage(RowPage);
    end;
   end;
end; // GetRecordIDByRecNo


//------------------------------------------------------------------------------
// read PFSMapFile
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.ReadTablePFSMapFile(SessionID: TACRSessionID);
begin
  FTablePageMapsManager.LoadFile(SessionID);
end; // ReadTablePFSMapFile


//------------------------------------------------------------------------------
// write PFSMap file
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.WriteTablePFSMapFile(SessionID: TACRSessionID);
begin
  FTablePageMapsManager.SaveFile(SessionID);
end; // WriteTablePFSMapFile

//------------------------------------------------------------------------------
// save file state
//------------------------------------------------------------------------------
function TACRDiskRecordManager.SaveState: Pointer;
begin
  Result := FTablePageMapsManager.SaveState;
end; // SaveState


//------------------------------------------------------------------------------
// restore file state
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.RestoreState(State: Pointer);
begin
  FTablePageMapsManager.RestoreState(State);
end; // RestoreState


//------------------------------------------------------------------------------
// free file state
//------------------------------------------------------------------------------
procedure TACRDiskRecordManager.FreeState(var State: Pointer);
begin
  FTablePageMapsManager.FreeState(State);
end; // FreeState



////////////////////////////////////////////////////////////////////////////////
//
// TACRDiskTableData
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// get table id
//------------------------------------------------------------------------------
function TACRDiskTableData.GetTableID: TACRObjectID;
begin
 Result := FTableItem.TableID;
end; // GetTableID


//------------------------------------------------------------------------------
// GetPageManager
//------------------------------------------------------------------------------
function TACRDiskTableData.GetPageManager: TACRPageManager;
begin
  Result := LPageManager;
end;// GetPageManager


//------------------------------------------------------------------------------
// create RecordManager
//------------------------------------------------------------------------------
procedure TACRDiskTableData.CreateRecordManager(TablePFSMapFileStartPageNo: TACRPageNo);
begin
 if (FRecordManager <> nil) then
  FRecordManager.Free;
 FRecordManager := TACRDiskRecordManager.Create(
                      TablePFSMapFileStartPageNo,
                      GetRecordBufferSize,
                      GetRecordSize,
                      TACRDiskPageManager(LPageManager),
                      Self
                                                 );
end;// CreateRecordManager


//------------------------------------------------------------------------------
// lock table in X mode
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalLockTableX(const CurrentRetryNo: Integer): Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.LockTable(
              CurrentRetryNo,FTableItem,FTableName,FSessionID,ltX,nil);
end; // InternalLockTableX


//------------------------------------------------------------------------------
// unlock table in X mode
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalUnlockTableX: Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.UnlockTable(
              FTableName,FSessionID,ltX,nil,False);
end; // InternalUnlockTableX


//------------------------------------------------------------------------------
// lock table in IS mode
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalLockTableIS(const CurrentRetryNo: Integer): Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.LockTable(
              CurrentRetryNo,FTableItem,FTableName,FSessionID,ltIS,nil);
end; // InternalLockTableIS


//------------------------------------------------------------------------------
// unlock table in IS mode
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalUnlockTableIS: Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.UnlockTable(
              FTableName,FSessionID,ltIS,nil,False);
end; // InternalUnlockTableIS


//------------------------------------------------------------------------------
// lock table in S mode
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalLockTableS(const CurrentRetryNo: Integer): Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.LockTable(
              CurrentRetryNo,FTableItem,FTableName,FSessionID,ltS,nil);
end; // InternalLockTableS


//------------------------------------------------------------------------------
// unlock table in S mode
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalUnlockTableS: Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.UnlockTable(
              FTableName,FSessionID,ltS,nil,False);
end; // InternalUnlockTableS


//------------------------------------------------------------------------------
// lock table in IRW mode
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalLockTableIRW(const CurrentRetryNo: Integer): Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.LockTable(
              CurrentRetryNo,FTableItem,FTableName,FSessionID,ltIRW,nil);
end; // InternalLockTableIRW


//------------------------------------------------------------------------------
// unlock table in IRW mode
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalUnlockTableIRW: Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.UnlockTable(
              FTableName,FSessionID,ltIRW,nil,False);
end; // InternalUnlockTableIRW


//------------------------------------------------------------------------------
// lock table in RW mode
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalLockTableRW(const CurrentRetryNo: Integer): Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.LockTable(
              CurrentRetryNo,FTableItem,FTableName,FSessionID,ltRW,nil);
end; // InternalLockTableRW


//------------------------------------------------------------------------------
// unlock table in RW mode
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalUnlockTableRW: Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.UnlockTable(
              FTableName,FSessionID,ltRW,nil,False);
end; // InternalUnlockTableRW


//------------------------------------------------------------------------------
// lock record in U mode
//------------------------------------------------------------------------------
function TACRDiskTableData.LockRecordU(const CurrentRetryNo: Integer): Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.LockRecord(
              FTableName,FSessionID,FRecordID);
end; // LockRecordU


//------------------------------------------------------------------------------
// unlock record in U mode
//------------------------------------------------------------------------------
function TACRDiskTableData.UnlockRecordU: Boolean;
begin
  Result := LDiskDatabaseData.TableLocksManager.UnlockRecord(
              FTableName,FSessionID,FRecordID);
end; // UnlockRecordU


//------------------------------------------------------------------------------
// set new session ID
//------------------------------------------------------------------------------
procedure TACRDiskTableData.SetSessionID(SessionID: TACRSessionID);
begin
  FSessionID := SessionID;
end;// set new session ID


//------------------------------------------------------------------------------
// Read data from TableMostUpdatedFile
//------------------------------------------------------------------------------
procedure TACRDiskTableData.ReadMostUpdatedData(SessionID: TACRSessionID);
var ms:               TACRMemoryStream;
    i:                Integer;
    TempIndexDefs:    TACRIndexDefs;
begin
 FTableMostUpdatedFile.ReadFile(SessionID);
 ms := TACRMemoryStream.Create;
 TempIndexDefs := TACRIndexDefs.Create;
 try
  ms.WriteBuffer(FTableMostUpdatedFile.FHandle.FFileData.Buffer^,
                 FTableMostUpdatedFile.FHandle.FFileData.Size);
  ms.Position := 0;
  TempIndexDefs.Assign(FIndexManager.IndexDefs);
  LoadSequencesFromStream(ms);
  FIndexManager.LoadFromStream(ms);
  for i := 0 to TempIndexDefs.Count - 1 do
   if (TempIndexDefs[i].Temporary) then
    FIndexManager.IndexDefs.AddCreated.Assign(TempIndexDefs[i]);
 finally
  TempIndexDefs.Free;
  ms.Free;
 end;
end; // ReadMostUpdatedData


//------------------------------------------------------------------------------
// write data into TableMostUpdatedFile
//------------------------------------------------------------------------------
procedure TACRDiskTableData.WriteMostUpdatedData(SessionID: TACRSessionID);
var ms: TACRMemoryStream;
begin
 ms := TACRMemoryStream.Create;
 try
  SaveSequencesToStream(ms);
  FIndexManager.SaveToStream(ms);
  FTableMostUpdatedFile.FHandle.ResizeFileData(ms.Size);
  ms.Position := 0;
  ms.ReadBuffer(FTableMostUpdatedFile.FHandle.FFileData.Buffer^,ms.Size);
  FTableMostUpdatedFile.WriteFile(SessionID);
 finally
  ms.Free;
 end;
end; // WriteMostUpdatedData


//------------------------------------------------------------------------------
// ReadTablePFSMapFile
//------------------------------------------------------------------------------
procedure TACRDiskTableData.ReadTablePFSMapFile(SessionID: TACRSessionID);
begin
  TACRDiskRecordManager(FRecordManager).ReadTablePFSMapFile(SessionID);
end; // ReadTablePFSMapFile


//------------------------------------------------------------------------------
// WriteTablePFSMapFile
//------------------------------------------------------------------------------
procedure TACRDiskTableData.WriteTablePFSMapFile(SessionID: TACRSessionID);
begin
  TACRDiskRecordManager(FRecordManager).WriteTablePFSMapFile(SessionID);
end; // WriteTablePFSMapFile


//------------------------------------------------------------------------------
// save state
//------------------------------------------------------------------------------
procedure TACRDiskTableData.SaveTablePFSMapFileState;
begin
  if (FSavedPFSMapState <> nil) then
   raise EACRException.Create(10841,ErrorLStateAlreadySet);
  FSavedPFSMapState := TACRDiskRecordManager(FRecordManager).SaveState;
end; // SaveTablePFSMapFileState;


//------------------------------------------------------------------------------
// restore state
//------------------------------------------------------------------------------
procedure TACRDiskTableData.RestoreTablePFSMapFileState;
begin
  TACRDiskRecordManager(FRecordManager).RestoreState(FSavedPFSMapState);
end; // RestoreTablePFSMapFileState;


//------------------------------------------------------------------------------
// free state
//------------------------------------------------------------------------------
procedure TACRDiskTableData.FreeTablePFSMapFileState;
begin
  TACRDiskRecordManager(FRecordManager).FreeState(FSavedPFSMapState);
end; // FreeTablePFSMapFileState;


//------------------------------------------------------------------------------
// ReadTableMetadata
//------------------------------------------------------------------------------
procedure TACRDiskTableData.ReadTableMetadata(
          SessionID:        TACRSessionID;
          aStartPageNo:     TACRPageNo
                                );
var ms: TACRMemoryStream;
    i:  Integer;
begin
  ms := TACRMemoryStream.Create;
  // Read data fom TableMetadataFile
  try
    FTableMetadataFile.ReadFile(SessionID,
                                FTableItem.TableMetaDataFilePageNo,
                                ms);
    FFieldManager.EngineVersion := TACRDiskPageManager(FDatabaseData.PageManager).DBHeader.Version;
    FFieldManager.LoadFromStream(ms);
    FFieldManager.FieldDefs.RecalcFieldOffsets;
    FConstraintManager.LoadFromStream(ms);
    FBLOBFieldsPresent := False;
    for i := 0 to FFieldManager.FieldDefs.Count - 1 do
     if (IsBLOBFieldType(FFieldManager.FieldDefs[i].BaseFieldType)) then
      begin
       FBLOBFieldsPresent := True;
       break;
      end;
   finally
      ms.Free;
   end;
end; // ReadTableMetadata


//------------------------------------------------------------------------------
// WriteTableMetadata
//------------------------------------------------------------------------------
procedure TACRDiskTableData.WriteTableMetadata(
          SessionID:        TACRSessionID;
          aStartPageNo:     TACRPageNo
                                );
var ms: TACRMemoryStream;
begin
   ms := TACRMemoryStream.Create;
   try
    FFieldManager.EngineVersion := TACRDiskPageManager(FDatabaseData.PageManager).DBHeader.Version;
    FFieldManager.SaveToStream(ms);
    FConstraintManager.SaveToStream(ms);
    FTableMetadataFile.WriteFile(SessionID,
                                 FTableItem.TableMetaDataFilePageNo,
                                 ms);
   finally
    ms.Free;
   end;
end; // WriteTableMetadata


//------------------------------------------------------------------------------
// delete table
//------------------------------------------------------------------------------
procedure TACRDiskTableData.InternalDeleteTable(Session: TACRBaseSession; Cascade: Boolean; LockTables: Boolean = True);
var
 TableLocksFile:          TACRTableLocksFile;
begin
 FSessionID := SYSTEM_SESSION_ID;
 if (not TryToLockTableX) then
   raise EACRException.Create(10783,ErrorLCannotLockTable,['X',FTableName]);
 try
  InternalOpenTable;
  try
   inherited DeleteTable(Session,Cascade,False);
  finally
   InternalCloseTable;
  end;
  // Flush changes
  LPageManager.ApplyChanges(SYSTEM_SESSION_ID);
  LPageManager.FlushFileBuffers;
  // lock tables and remove table item from tables list
  if (LockTables) then
   if (not TryUsingTimeOut(LDiskDatabaseData.InternalLockTables, FLockParams)) then
    raise EACRException.Create(10785,ErrorLCannotLockTables,
      [LDiskDatabaseData.DatabaseName]);
  try
    LDiskDatabaseData.FTableListFile.DeleteTable(FTableName);
    LPageManager.ApplyChanges(SYSTEM_SESSION_ID);
    LPageManager.FlushFileBuffers;
  finally
  if (LockTables) then
   if (not LDiskDatabaseData.InternalUnlockTables) then
    raise EACRException.Create(10786,ErrorLCannotUnlockTables,
      [LDiskDatabaseData.DatabaseName]);
  end;
 finally
  if (not TryToUnlockTableX) then
    raise EACRException.Create(10784,ErrorLCannotUnlockTable,['X',FTableName]);
 end;
 InternalOpenTable;
 try
  try
   // Free all pages used by table rows, PFS, index, BLOBs, varchars
   if (FIndexManager.IndexDefs.Count > 0) then
     FIndexManager.DropAllIndexes(SYSTEM_SESSION_ID);
   ReadTablePFSMapFile(SYSTEM_SESSION_ID);
   FRecordManager.Empty(SYSTEM_SESSION_ID);
   TACRDiskRecordManager(FRecordManager).FTablePageMapsManager.FTablePFSMapFile.DeleteFile(
    SYSTEM_SESSION_ID,FTableItem.PFSPageMapFilePageNo);
  finally
   InternalCloseTable;
   // Delete table files
   FTableMetadataFile.DeleteFile(SYSTEM_SESSION_ID,
    FTableItem.TableMetaDataFilePageNo);
   FTableMostUpdatedFile.DeleteFile(SYSTEM_SESSION_ID,
    FTableItem.MostUpdatedTableMetadataFilePageNo);
   TableLocksFile := TACRTableLocksFile.Create(LDiskDatabaseData.PageManager,FLockParams);
   try
    TableLocksFile.OpenFile(FTableItem.LockFilePageNo);
    TableLocksFile.DeleteFile(SYSTEM_SESSION_ID,FTableItem.LockFilePageNo);
   finally
    TableLocksFile.Free;
   end;
  end;
 finally
  LPageManager.ApplyChanges(SYSTEM_SESSION_ID);
  LPageManager.FlushFileBuffers;
 end;
end; // InternalDeleteTable


//------------------------------------------------------------------------------
// write metadata changes
//------------------------------------------------------------------------------
procedure TACRDiskTableData.WriteMetadataChanges(SessionID: TACRSessionID);
begin
  WriteTableMetadata(SessionID,FTableItem.TableMetaDataFilePageNo);
  LPageManager.ApplyChanges(SessionID);
  LPageManager.FlushFileBuffers;
end; // WriteMetadataChanges


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRDiskTableData.Create(
                        aDiskDatabaseData: TACRDiskDatabaseData
                      );
begin
 inherited Create(aDiskDatabaseData);
 FSequenceManager := nil;
 FLockParams := aDiskDatabaseData.LockParams;
 FRandomSearchRetryCount := aDiskDatabaseData.Options.RandomSearchRetryCount;
 FSessionID := INVALID_SESSION_ID;
 FExclusive := False;
 LDiskDatabaseData := aDiskDatabaseData;
 LPageManager := TACRDiskPageManager(aDiskDatabaseData.FPageManager);
 FTableMetadataFile := TACRTableMetadataFile.Create(LPageManager);
 FTableMostUpdatedFile := TACRTableMostUpdatedFile.Create(LPageManager,Self.IsSynchronizationRequired);
 FIsTableOpened := False;
 FInMemory := False;
 FTemporary := False;
end; // Create;


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRDiskTableData.Destroy;
begin
 if (FSequenceManager <> nil) then
  FSequenceManager.Free;
 FSequenceManager := nil;
 FTableMetadataFile.Free;
 FTableMostUpdatedFile.Free;
 inherited Destroy;
end; // Destroy


//------------------------------------------------------------------------------
// return size of the record on disk
//------------------------------------------------------------------------------
function TACRDiskTableData.GetRecordSize: Integer;
begin
 if (FFieldManager.FieldDefs.Count <= 0) then
  raise EACRException.Create(10515,ErrorLNoFields);
 Result := FFieldManager.FieldDefs[FFieldManager.FieldDefs.Count-1].DiskOffset +
   FFieldManager.FieldDefs[FFieldManager.FieldDefs.Count-1].DiskDataSize;
end; // GetRecordSize


//------------------------------------------------------------------------------
// create table
//------------------------------------------------------------------------------
procedure TACRDiskTableData.CreateTable(
                          Cursor: TACRCursor;
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                                       );
var
    TableLocksFile:   TACRTableLocksFile;
    TablePFSMapFile:  TACRTablePFSMapFile;
    DeleteTable:      Boolean;
begin
  Lock;
  try
   if (FieldDefs.Count <= 0) then
    raise EACRException.Create(10025,ErrorLNoFields);
   CheckFieldDefinitions(FieldDefs);
   CheckIndexDefinitions(IndexDefs);
   CheckConstraintDefinitions(ConstraintDefs);
   if (FIsTableOpened) then
    raise EACRException.Create(10779,ErrorLTableIsInUse,[FTableName]);
   if (not TryUsingTimeOut(LDiskDatabaseData.InternalLockTables, FLockParams)) then
    raise EACRException.Create(11503,ErrorLCannotLockTables,
      [LDiskDatabaseData.DatabaseName]);
   try
    DeleteTable := LDiskDatabaseData.GetTableItemIfExists(FTableName,FTableItem);
   finally
     if (not LDiskDatabaseData.InternalUnlockTables) then
      raise EACRException.Create(11504,ErrorLCannotUnlockTables,
        [LDiskDatabaseData.DatabaseName]);
   end;
   if (DeleteTable) then
     InternalDeleteTable(Cursor.Session,True,False);
   if (not TryUsingTimeOut(LDiskDatabaseData.InternalLockTables, FLockParams)) then
    raise EACRException.Create(10616,ErrorLCannotLockTables,
      [LDiskDatabaseData.DatabaseName]);
   try
     FTableName := Cursor.FTableName;
     FTableItem.TableName := FTableName;
     FTableItem.TableID :=  LDiskDatabaseData.FTableListFile.GetNextTableID;
     // create table files
     // create table metadata file
     FTableItem.TableMetaDataFilePageNo :=
      FTableMetadataFile.CreateFile(SYSTEM_SESSION_ID,FTableItem.TableID);
     // create table most updated metadata file
     FTableItem.MostUpdatedTableMetadataFilePageNo :=
      FTableMostUpdatedFile.CreateFile(SYSTEM_SESSION_ID,FTableItem.TableID);
     // create table PFSPageMapFile
     TablePFSMapFile := TACRTablePFSMapFile.Create(LDiskDatabaseData.PageManager,nil);
     try
      FTableItem.PFSPageMapFilePageNo :=
       TablePFSMapFile.CreateFile(FTableItem.TableID);
     finally
      TablePFSMapFile.Free;
     end;
     // create LockFile
     TableLocksFile := TACRTableLocksFile.Create(LDiskDatabaseData.PageManager,FLockParams);
     try
      FTableItem.LockFilePageNo :=
       TableLocksFile.CreateFile(LDiskDatabaseData.Options.MaxSessionCount,
         FTableItem.TableID);
     finally
       TableLocksFile.Free;
     end;

     FTableItem.CryptoHeader.CryptoAlgorithm := ACR_Cipher_None;
     // create new record in table list
     FTableItem.TableID :=  LDiskDatabaseData.FTableListFile.CreateTable(FTableItem);
   finally
     if (not LDiskDatabaseData.InternalUnlockTables) then
      raise EACRException.Create(10617,ErrorLCannotUnlockTables,
        [LDiskDatabaseData.DatabaseName]);
   end;
     // init objects
     CreateSequenceManager;
     CreateFieldManager(FieldDefs);
     CreateIndexManager(IndexDefs);
     CreateConstraintManager(ConstraintDefs);
     BuildSequences;
     CreateRecordManager(FTableItem.PFSPageMapFilePageNo);
     try
       InitCursor(Cursor);
       try
        if (FConstraintManager.ConstraintDefs.ForeignKeysExists) then
          CheckForeignKeyDefinitionsAndCreateForeignKeyActions(Cursor);
       except
         if (not TryUsingTimeOut(LDiskDatabaseData.InternalLockTables, FLockParams)) then
          raise EACRException.Create(11501,ErrorLCannotLockTables,
            [LDiskDatabaseData.DatabaseName]);
         try
          LDiskDatabaseData.FTableListFile.DeleteTable(FTableName);
          LPageManager.ApplyChanges(SYSTEM_SESSION_ID);
         finally
           if (not LDiskDatabaseData.InternalUnlockTables) then
            raise EACRException.Create(11502,ErrorLCannotUnlockTables,
              [LDiskDatabaseData.DatabaseName]);
         end;
         raise;
       end;
       // write data into TableMetadataFile
       WriteTableMetadata(SYSTEM_SESSION_ID,FTableItem.TableMetaDataFilePageNo);
       // write data into TableMostUpdatedFile
       FTableMostUpdatedFile.OpenFile(FTableItem.MostUpdatedTableMetadataFilePageNo,True);
       WriteMostUpdatedData(SYSTEM_SESSION_ID);
       // Flush changes
       LPageManager.ApplyChanges(SYSTEM_SESSION_ID);
       LPageManager.FlushFileBuffers;
     finally
      FreeAndNil(FRecordManager);
      FreeAndNil(FFieldManager);
      FreeAndNil(FSequenceManager);
      FreeAndNil(FIndexManager);
      FreeAndNil(FConstraintManager);
     end;
  finally
   Unlock;
  end;
end;// CreateTable


//------------------------------------------------------------------------------
// DeleteTable
//------------------------------------------------------------------------------
procedure TACRDiskTableData.DeleteTable(Session: TACRBaseSession; Cascade: Boolean; DesignMode: Boolean = False);
begin
  Lock;
  try
   if (FIsTableOpened) then
    raise EACRException.Create(10782,ErrorLTableIsInUse,[FTableName]);
   if (not TryUsingTimeOut(LDiskDatabaseData.InternalLockTables, FLockParams)) then
    raise EACRException.Create(10616,ErrorLCannotLockTables,
      [LDiskDatabaseData.DatabaseName]);
   try
    if (not LDiskDatabaseData.GetTableItemIfExists(FTableName,FTableItem)) then
      raise EACRException.Create(10780, ErrorLTableDoesNotExists,[FTableName,LDiskDatabaseData.DatabaseName]);
   finally
    if (not LDiskDatabaseData.InternalUnlockTables) then
      raise EACRException.Create(10624,ErrorLCannotUnlockTables,
          [LDiskDatabaseData.DatabaseName]);
   end;
   InternalDeleteTable(Session,Cascade,True);
  finally
    Unlock;
  end;
end;// DeleteTable


//------------------------------------------------------------------------------
// delete constraint (FK,FKAction or PK) and write changes to MetaData file
//------------------------------------------------------------------------------
procedure TACRDiskTableData.DeleteConstraint(
                                        Cursor:           TACRCursor;
                                        Name:             String;
                                        Cascade:          Boolean;
                                        FKPartialDelete:  Boolean
                                        );
begin
 Lock;
 try
   inherited DeleteConstraint(Cursor,Name,Cascade,FKPartialDelete);
   WriteMetadataChanges(Cursor.Session.SessionID);
   LPageManager.ApplyChanges(Cursor.Session.SessionID);
   LPageManager.FlushFileBuffers;
 finally
   Unlock;
 end;
end; // DeleteConstraint


//------------------------------------------------------------------------------
// add foreign key
//------------------------------------------------------------------------------
procedure TACRDiskTableData.AddForeignKey(Cursor: TACRCursor; ConstraintDef: TACRConstraintDefForeignKey);
begin
 Lock;
 try
   inherited;
   WriteMetadataChanges(Cursor.Session.SessionID);
   LPageManager.ApplyChanges(Cursor.Session.SessionID);
   LPageManager.FlushFileBuffers;
 finally
   Unlock;
 end;
end; // AddForeignKey


//------------------------------------------------------------------------------
// delete constraint (FK,FKAction or PK) and write changes to MetaData file
//------------------------------------------------------------------------------
procedure TACRDiskTableData.RenameReferenceTableName(Cursor:            TACRCursor;
                                                     OldName, NewName:  String);
begin
 Lock;
 try
   inherited;
   WriteMetadataChanges(Cursor.Session.SessionID);
   LPageManager.ApplyChanges(Cursor.Session.SessionID);
   LPageManager.FlushFileBuffers;
 finally
   Unlock;
 end;
end; // DeleteConstraint


//------------------------------------------------------------------------------
// EmptyTable
//------------------------------------------------------------------------------
procedure TACRDiskTableData.EmptyTable(Cursor: TACRCursor);
begin
  Lock;
  try
   if (FIsTableOpened) then
    raise EACRException.Create(10785,ErrorLTableIsInUse,[FTableName]);
   if (Cursor.Session.InTransaction) then
    raise EACRException.Create(10836,ErrorLCannotPerformThisOperationInsideATransaction);
   OpenTable(Cursor);
   try
    FSessionID := Cursor.Session.SessionID;
    if (FConstraintManager.ConstraintDefs.ForeignKeysActionsDeleteExists) then
     raise EACRException.Create(11493,ErrorLCannotEmptyTableWithForeignKeyActionsDelete,[FTableName]);
    FRecordManager.Empty(FSessionID);
    EmptyAllIndexes(FSessionID);
    WriteMostUpdatedData(FSessionID);
    WriteTablePFSMapFile(FSessionID);
    LPageManager.ApplyChanges(FSessionID);
    LPageManager.FlushFileBuffers;
   finally
    CloseTable(Cursor);
   end;
  finally
   Unlock;
  end;
end;// EmptyTable


//------------------------------------------------------------------------------
// RenameTable
//------------------------------------------------------------------------------
procedure TACRDiskTableData.RenameTable(Cursor: TACRCursor; NewTableName: String);
var OldTableName: String;
begin
  Lock;
  try
   if (Cursor = nil) then
    raise EACRException.Create(11506,ErrorLNilPointer);
   FSessionID := Cursor.Session.SessionID;
   FExclusive := True;
   OldTableName := FTableName;
   if (FIsTableOpened) then
    raise EACRException.Create(10789,ErrorLTableIsInUse,[FTableName]);
   if (not TryUsingTimeOut(LDiskDatabaseData.InternalLockTables, FLockParams)) then
    raise EACRException.Create(10790,ErrorLCannotLockTables,
      [LDiskDatabaseData.DatabaseName]);
   try
     if (LDiskDatabaseData.GetTableItemIfExists(NewTableName,FTableItem)) then
      raise EACRException.Create(11490,ErrorLCannotRenameTableAlreadyExists,[FTableName, NewTableName]);
     if (not LDiskDatabaseData.GetTableItemIfExists(FTableName,FTableItem)) then
      raise EACRException.Create(10791, ErrorLTableDoesNotExists,[FTableName,LDiskDatabaseData.DatabaseName]);
   finally
     if (not LDiskDatabaseData.InternalUnlockTables) then
      raise EACRException.Create(10794,ErrorLCannotUnlockTables,
        [LDiskDatabaseData.DatabaseName]);
   end;
   if (not TryToLockTableX) then
    raise EACRException.Create(10792,ErrorLCannotLockTable,['X',FTableName]);
   try
     InternalOpenTable;
     try
       inherited RenameTable(Cursor,NewTableName);
       WriteMetadataChanges(FSessionID);
     finally
       InternalCloseTable;
     end;
     if (not TryUsingTimeOut(LDiskDatabaseData.InternalLockTables, FLockParams)) then
      raise EACRException.Create(11507,ErrorLCannotLockTables,
        [LDiskDatabaseData.DatabaseName]);
     try
       FTableName := OldTableName;
       LDiskDatabaseData.FTableListFile.RenameTable(FTableName, NewTableName);
       LPageManager.ApplyChanges(FSessionID);
       LPageManager.ApplyChanges(SYSTEM_SESSION_ID);
       LPageManager.FlushFileBuffers;
     finally
       if (not LDiskDatabaseData.InternalUnlockTables) then
        raise EACRException.Create(11508,ErrorLCannotUnlockTables,
          [LDiskDatabaseData.DatabaseName]);
     end;
   finally
     FTableName := OldTableName;
     try
      if (not TryToUnlockTableX) then
       raise EACRException.Create(10793,ErrorLCannotUnlockTable,['X',FTableName]);
     finally
      FTableName := NewTableName;
      FTableItem.TableName := NewTableName;
     end;
   end;
  finally
   Unlock;
  end;
end;// RenameTable


//------------------------------------------------------------------------------
// open table
//------------------------------------------------------------------------------
procedure TACRDiskTableData.InternalOpenTable;
begin
  FIsTableOpened := True;
  FNumLocksForRead := 0;
  // read metadata
  if (FConstraintManager = nil) then
    FConstraintManager := TACRBaseConstraintManager.Create(Self);
  if (FSequenceManager = nil) then
    FSequenceManager := TACRBaseSequenceManager.Create(LDiskDatabaseData);
  if (FFieldManager = nil) then
    FFieldManager := TACRBaseFieldManager.Create(Self,FSequenceManager);
  if (FIndexManager <> nil) then
    FreeAndNil(FIndexManager);
  FIndexManager := TACRBaseIndexManager.Create(Self);
  ReadTableMetadata(SessionID,FTableItem.TableMetaDataFilePageNo);

  // Read data from TableMostUpdatedFile
  FTableMostUpdatedFile.OpenFile(FTableItem.MostUpdatedTableMetadataFilePageNo,
                                 FExclusive);
  ReadMostUpdatedData(SessionID);
  // create record manager
  CreateRecordManager(FTableItem.PFSPageMapFilePageNo);
  if (FExclusive or FInTransaction) then
   begin
    // read this file in RAM for speed optimization
    ReadTablePFSMapFile(SessionID);
   end;
end; // InternalOpenTable


//------------------------------------------------------------------------------
// close table
//------------------------------------------------------------------------------
procedure TACRDiskTableData.InternalCloseTable;
begin
  if (FInTransaction) then
   begin
    WriteMostUpdatedData(SessionID);
    WriteTablePFSMapFile(SessionID);
   end;
  FreeAndNil(FRecordManager);
  FreeAndNil(FSequenceManager);
  FreeAndNil(FIndexManager);
  FreeAndNil(FConstraintManager);
  FreeAndNil(FFieldManager);
  FIsTableOpened := False;
end; // InternalCloseTable


//------------------------------------------------------------------------------
// OpenTable
//------------------------------------------------------------------------------
procedure  TACRDiskTableData.OpenTable(Cursor: TACRCursor);
begin
 Lock;
 try
   FSessionID := Cursor.Session.SessionID;
   if (FSessionID < 0) then
    raise EACRException.Create(10669,ErrorLInvalidSessionID,[FSessionID]);
   if (not FIsTableOpened) then
    begin
     if (not TryUsingTimeOut(LDiskDatabaseData.InternalLockTables, FLockParams)) then
      raise EACRException.Create(10796,ErrorLCannotLockTables,
        [LDiskDatabaseData.DatabaseName]);
     try
      FTableName := Cursor.TableName;
      if (not LDiskDatabaseData.GetTableItemIfExists(FTableName,FTableItem)) then
        raise EACRException.Create(10797, ErrorLTableDoesNotExists,[FTableName,LDiskDatabaseData.DatabaseName]);
      FExclusive := Cursor.Exclusive;
     finally
      if (not LDiskDatabaseData.InternalUnlockTables) then
        raise EACRException.Create(10798,ErrorLCannotUnlockTables,
            [LDiskDatabaseData.DatabaseName]);
     end;
    end // table is not opened
   else
    if (FExclusive) then
     raise EACRException.Create(10810,ErrorLTableIsOpenedInExclusiveMode,[FTableName]);
   // lock table in IS or X
   if (Cursor.Exclusive) then
    begin
     if (not TryToLockTableX) then
      raise EACRException.Create(10799,ErrorLCannotLockTable,['X',FTableName]);
     if (Cursor.Session.InTransaction) then
      TACRLocalSession(Cursor.Session).Transaction.OpenTableForRead(FTableName,True);
    end // lock in X
   else
    begin
     if (not TryToLockTableIS) then
      raise EACRException.Create(10800,ErrorLCannotLockTable,['IS',FTableName]);
     if (Cursor.Session.InTransaction) then
      TACRLocalSession(Cursor.Session).Transaction.OpenTableForRead(FTableName,False);
    end; // lock in IS
   if (not FIsTableOpened) then
    begin
     FInTransaction := False;
     InternalOpenTable;
     if (Cursor.Session.InTransaction) then
       StartTransaction(FSessionID);
    end;
   inherited;
 finally
   Unlock;
 end;
end;// OpenTable


//------------------------------------------------------------------------------
// CloseTable
//------------------------------------------------------------------------------
procedure TACRDiskTableData.CloseTable(Cursor: TACRCursor);
var SkipUnlock: Boolean;
begin
  if (not FIsTableOpened) then
   Exit;
 Lock;
 try
  if (FCursorList.IndexOf(Cursor) < 0) then
   Exit;
  SkipUnlock := False;
  if (Cursor is TACRLocalCursor) then
   if (TACRLocalCursor(Cursor).DoNotUnlockTable) then
    SkipUnlock := True;
  FSessionID := Cursor.Session.SessionID;
  inherited CloseTable(Cursor);
  if (FCursorList.Count = 0) then
   begin
    InternalCloseTable;
   end;
  if (FSessionID < 0) then
   raise EACRException.Create(10801,ErrorLInvalidSessionID,[FSessionID]);
  if (Cursor.FExclusive) then
   begin
    if (not SkipUnlock) then
     if (not TryToUnlockTableX) then
      raise EACRException.Create(10802,ErrorLCannotUnlockTable,['X',FTableName]);
   end
  else
   begin
    if (not SkipUnlock) then
     if (not TryToUnlockTableIS) then
      raise EACRException.Create(10803,ErrorLCannotUnlockTable,['IS',FTableName]);
   end;
  if (FCursorList.Count = 0) then
   if (FNumLocksForRead > 0) then
    if (not SkipUnlock) then
     raise EACRException.Create(11246,ErrorLUnlockTableWasNotCalled,[FNumLocksForRead]);
 finally
  Unlock;
 end;
end;// CloseTable


//------------------------------------------------------------------------------
// Rename Field by Field Index in FieldDefs
//------------------------------------------------------------------------------
procedure TACRDiskTableData.RenameField(Cursor: TACRCursor; FieldName, NewFieldName: String);
begin
  Lock;
  try
   if (not FIsTableOpened) or (not FExclusive) then
    raise EACRException.Create(10804,ErrorLTableIsNotOpenedExclusively,[FTableName]);
   if (Cursor.Session.InTransaction) then
    raise EACRException.Create(10837,ErrorLCannotPerformThisOperationInsideATransaction);
   inherited RenameField(Cursor,FieldName,NewFieldName);
   WriteTableMetadata(FSessionID,FTableItem.TableMetaDataFilePageNo);
   LPageManager.ApplyChanges(FSessionID);
   LPageManager.FlushFileBuffers;
  finally
   Unlock;
  end;
end; // RenameField


//------------------------------------------------------------------------------
// add index
//------------------------------------------------------------------------------
procedure TACRDiskTableData.AddIndex(IndexDef: TACRIndexDef; Cursor: TACRCursor);
begin
 Lock;
 try
   FSessionID := Cursor.Session.SessionID;
   if (not Cursor.Exclusive) then
    raise EACRException.Create(10609,ErrorLTableIsNotOpenedExclusively,[FTableName]);
   if (FInTransaction) then
    raise EACRException.Create(10834,ErrorLCannotPerformThisOperationInsideATransaction); 
   inherited AddIndex(IndexDef,Cursor);
   WriteMostUpdatedData(FSessionID);
   WriteTableMetadata(FSessionID,FTableItem.TableMetaDataFilePageNo);
   LPageManager.ApplyChanges(FSessionID);
   LPageManager.FlushFileBuffers;
 finally
  Unlock;
 end;
end; // AddIndex


//------------------------------------------------------------------------------
// delete index
//------------------------------------------------------------------------------
procedure TACRDiskTableData.DeleteIndex(IndexID: TACRObjectID; Cursor: TACRCursor);
begin
 Lock;
 try
   FSessionID := Cursor.Session.SessionID;
   if (not Cursor.Exclusive) then
    raise EACRException.Create(10610,ErrorLTableIsNotOpenedExclusively,[FTableName]);
   if (FInTransaction) then
    raise EACRException.Create(10835,ErrorLCannotPerformThisOperationInsideATransaction); 
   inherited DeleteIndex(IndexID,Cursor);
   WriteMostUpdatedData(FSessionID);
   WriteTableMetadata(FSessionID,FTableItem.TableMetaDataFilePageNo);
   LPageManager.ApplyChanges(FSessionID);
   LPageManager.FlushFileBuffers;
 finally
  Unlock;
 end;
end; // DeleteIndex


//------------------------------------------------------------------------------
// delete all indexes
//------------------------------------------------------------------------------
procedure TACRDiskTableData.DeleteAllIndexes(Cursor: TACRCursor);
begin
 Lock;
 try
   FSessionID := Cursor.Session.SessionID;
   if (not Cursor.Exclusive) then
    raise EACRException.Create(10611,ErrorLTableIsNotOpenedExclusively,[FTableName]);
   inherited DeleteAllIndexes(Cursor);
   WriteMostUpdatedData(FSessionID);
   LPageManager.ApplyChanges(FSessionID);
   LPageManager.FlushFileBuffers;
 finally
  Unlock;
 end;
end; // DeleteAllIndexes


//------------------------------------------------------------------------------
// return true if record was found and is visible by cursor
// prepares params for FindRecordByScan and calls it
//------------------------------------------------------------------------------
function TACRDiskTableData.FindRecord(
                    Cursor:           TACRCursor;
                    SearchExpression: TACRExpression; // locate
                    KeyCondition:     TACRScanSearchCondition; // find key
                    // find first or last record in order specified by index settings
                    Restart:          Boolean;
                    // go from first record to last or vice versa
                    GoForward:        Boolean;
                    // if record was found then return record id
                    ResultRecordID:   PACRRecordID;
                    // if specified - fill record bitmap
                    RecordBitmap:     TACRRecordBitmap;
                    InternalCall:     Boolean;
                    StopAtFirstFoundRecord: Boolean
                   ): Boolean;
begin
  Lock;
  try
    FSessionID := Cursor.Session.SessionID;
    if (InternalCall) then
     begin
       LockTableForRead(FSessionID);
       try
         Result := inherited FindRecord(Cursor,SearchExpression,KeyCondition,
                           Restart,GoForward,ResultRecordID,RecordBitmap,InternalCall,StopAtFirstFoundRecord);
       finally
         UnlockTableForRead(FSessionID);
       end;
       Exit;
     end;
    if (not FInTransaction) and (not FExclusive) and (FNumLocksForRead = 0) then
     LockTableForRead(FSessionID);
{
     if (not TryToLockTableS) then
      raise EACRException.Create(10587,ErrorLCannotLockTable,
       ['S',FTableName]);
}
    try
      if (not FInTransaction) and (not FExclusive) then
        ReadTablePFSMapFile(FSessionID);
      if (FInTransaction and (not Cursor.Session.InTransaction)) then
       begin
        SaveTablePFSMapFileState;
        ReadTablePFSMapFile(FSessionID);
       end;
      try
        Result := inherited FindRecord(Cursor,SearchExpression,KeyCondition,
                           Restart,GoForward,ResultRecordID,RecordBitmap,True);
      finally
        if (FInTransaction and (not Cursor.Session.InTransaction)) then
         begin
          RestoreTablePFSMapFileState;
          FreeTablePFSMapFileState;
         end;
      end;
    finally
     if (not FInTransaction) and (not FExclusive) then
      UnlockTableForRead(FSessionID);
{
      if (not TryToUnlockTableS) then
       raise EACRException.Create(10588,ErrorLCannotUnlockTable,
        ['S',FTableName]);
}
    end;
  finally
    Unlock;
  end;
end; // FindRecord


//------------------------------------------------------------------------------
// return true if record exists
//------------------------------------------------------------------------------
function TACRDiskTableData.IsRecordExists(Cursor: TACRCursor): Boolean;
begin
 Lock;
 try
    FSessionID := Cursor.Session.SessionID;
    if (not FInTransaction) and (not FExclusive) and (FNumLocksForRead = 0) then
     if (not TryToLockTableS) then
      raise EACRException.Create(11281,ErrorLCannotLockTable,['S',FTableName]);
    try
      if (not FInTransaction) and (not FExclusive) and (FNumLocksForRead = 0) then
        ReadTablePFSMapFile(FSessionID);
      if (FInTransaction and (not Cursor.Session.InTransaction)) then
       begin
        SaveTablePFSMapFileState;
        ReadTablePFSMapFile(FSessionID);
       end;
      try
        Result := inherited IsRecordExists(Cursor);
      finally
        if (FInTransaction and (not Cursor.Session.InTransaction)) then
         begin
          RestoreTablePFSMapFileState;
          FreeTablePFSMapFileState;
         end;
      end;
    finally
     if (not FInTransaction) and (not FExclusive) and (FNumLocksForRead = 0) then
      if (not TryToUnlockTableS) then
       raise EACRException.Create(11282,ErrorLCannotUnlockTable,['S',FTableName]);
    end;
 finally
   Unlock;
 end;
end; // IsRecordExists


//------------------------------------------------------------------------------
// return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
//------------------------------------------------------------------------------
function TACRDiskTableData.CompareRecordID(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer;
begin
 if ((RecordID1.PageNo = RecordID2.PageNo) and
     (RecordID1.PageItemNo = RecordID2.PageItemNo)) then
  Result := 0
 else
 if (RecordID1.PageNo > RecordID2.PageNo) then
  Result := 1
 else
 if (RecordID1.PageNo < RecordID2.PageNo) then
  Result := -1
 else
 if (RecordID1.PageItemNo > RecordID2.PageItemNo) then
  Result := 1
 else
  Result := -1
end; // CompareRecordID


//------------------------------------------------------------------------------
// read record with Distinct, SQLFilter, SQLTopRowCount, Filter, Range, OnFilterRecord
//------------------------------------------------------------------------------
function TACRDiskTableData.GetRecordBuffer(
                          Cursor:         TACRCursor;
                          GetRecordMode:  TACRGetRecordMode
                        ): TACRGetRecordResult;
begin
 Lock;
 try
    FSessionID := Cursor.Session.SessionID;
    if (not FInTransaction) and (not FExclusive) and (FNumLocksForRead = 0) then
     if (not TryToLockTableS) then
      raise EACRException.Create(10812,ErrorLCannotLockTable,['S',FTableName]);
    try
      if (not FInTransaction) and (not FExclusive) and (FNumLocksForRead = 0) then
        ReadTablePFSMapFile(FSessionID);
      if (FInTransaction and (not Cursor.Session.InTransaction)) then
       begin
        SaveTablePFSMapFileState;
        ReadTablePFSMapFile(FSessionID);
       end;
      try
        Result := inherited GetRecordBuffer(Cursor,GetRecordMode);
      finally
        if (FInTransaction and (not Cursor.Session.InTransaction)) then
         begin
          RestoreTablePFSMapFileState;
          FreeTablePFSMapFileState;
         end;
      end;
    finally
     if (not FInTransaction) and (not FExclusive) and (FNumLocksForRead = 0) then
      if (not TryToUnlockTableS) then
       raise EACRException.Create(10813,ErrorLCannotUnlockTable,['S',FTableName]);
    end;
 finally
   Unlock;
 end;
end; // GetRecordBuffer


//------------------------------------------------------------------------------
// return record count
//------------------------------------------------------------------------------
function TACRDiskTableData.GetRecordCount(Cursor: TACRCursor; InternalCall: Boolean): TACRRecordNo;
begin
  Lock;
  try
    FSessionID := Cursor.Session.SessionID;
    if (InternalCall) then
      Result := inherited GetRecordCount(Cursor)
    else
     begin
      if (not FExclusive) and (not FInTransaction) then
       if (not TryToLockTableS) then
        raise EACRException.Create(10589,ErrorLCannotLockTable,
        ['S',FTableName]);
      try
        if (not FExclusive) and (not FInTransaction) then
         ReadTablePFSMapFile(FSessionID);
        if (FInTransaction and (not Cursor.Session.InTransaction)) then
         begin
          SaveTablePFSMapFileState;
          ReadTablePFSMapFile(FSessionID);
         end;
        try
          Result := inherited GetRecordCount(Cursor);
        finally
          if (FInTransaction and (not Cursor.Session.InTransaction)) then
           begin
            RestoreTablePFSMapFileState;
            FreeTablePFSMapFileState;
           end;
        end;
      finally
       if (not FExclusive) and (not FInTransaction) then
        if (not TryToUnlockTableS) then
         raise EACRException.Create(10590,ErrorLCannotUnlockTable,
          ['S',FTableName]);
      end;
     end; // external call
  finally
    Unlock;
  end;
end; // GetRecordCount


//------------------------------------------------------------------------------
// set RecNo
//------------------------------------------------------------------------------
procedure TACRDiskTableData.SetRecNo(Cursor: TACRCursor; RecNo: TACRRecordNo);
begin
  Lock;
  try
    FSessionID := Cursor.Session.SessionID;
    if (not Cursor.IsViewRestricted) then
     if (not FExclusive) and (not FInTransaction) then
      if (not TryToLockTableS) then
       raise EACRException.Create(10591,ErrorLCannotLockTable,
        ['S',FTableName]);
    try
      if (not Cursor.IsViewRestricted) then
       begin
        if (not FExclusive) and (not FInTransaction) then
         ReadTablePFSMapFile(FSessionID);
        if (FInTransaction and (not Cursor.Session.InTransaction)) then
         begin
          SaveTablePFSMapFileState;
          ReadTablePFSMapFile(FSessionID);
         end;
       end;
      try
        inherited SetRecNo(Cursor,RecNo);
      finally
        if (not Cursor.IsViewRestricted) then
         begin
          if (FInTransaction and (not Cursor.Session.InTransaction)) then
           begin
            RestoreTablePFSMapFileState;
            FreeTablePFSMapFileState;
           end;
         end;
      end;
    finally
     if (not Cursor.IsViewRestricted) then
      if (not FExclusive) and (not FInTransaction) then
       if (not TryToUnlockTableS) then
        raise EACRException.Create(10592,ErrorLCannotUnlockTable,
         ['S',FTableName]);
    end;
  finally
    Unlock;
  end;
end; // SetRecNo


//------------------------------------------------------------------------------
// get RecNo
//------------------------------------------------------------------------------
function TACRDiskTableData.GetRecNo(Cursor: TACRCursor): TACRRecordNo;
begin
  Lock;
  try
    FSessionID := Cursor.Session.SessionID;
    if (not Cursor.IsViewRestricted) then
     if (not FExclusive) and (not FInTransaction) then
      if (not TryToLockTableS) then
       raise EACRException.Create(10593,ErrorLCannotLockTable,
       ['S',FTableName]);
    try
      if (not Cursor.IsViewRestricted) then
       begin
        if (not FExclusive) and (not FInTransaction) then
         ReadTablePFSMapFile(FSessionID);
        if (FInTransaction and (not Cursor.Session.InTransaction)) then
         begin
          SaveTablePFSMapFileState;
          ReadTablePFSMapFile(FSessionID);
         end;
       end;
      try
        Result := inherited GetRecNo(Cursor);
      finally
       if (not Cursor.IsViewRestricted) then
        if (FInTransaction and (not Cursor.Session.InTransaction)) then
         begin
          RestoreTablePFSMapFileState;
          FreeTablePFSMapFileState;
         end;
      end;
    finally
     if (not Cursor.IsViewRestricted) then
      if (not FExclusive) and (not FInTransaction) then
       if (not TryToUnlockTableS) then
        raise EACRException.Create(10594,ErrorLCannotUnlockTable,
         ['S',FTableName]);
    end;
  finally
    Unlock;
  end;
end; // GetRecNo


//------------------------------------------------------------------------------
// check constraint conditions
//------------------------------------------------------------------------------
procedure TACRDiskTableData.CheckConstraints(
                            Cursor:          TACRCursor;
                            SessionID:       TACRSessionID;
                            NewRecordBuffer: TACRRecordBuffer;
                            OldRecordBuffer: TACRRecordBuffer;
                            ToInsert:        Boolean;
                            CurrentRecordID: TACRRecordID;
                            SkipFKCheck:     Boolean = False
                          );
begin
 if (FConstraintManager.ConstraintDefs.Count > 0) then
  begin
    Lock;
    try
      FSessionID := SessionID;
      if (not TryToLockTableS) then
       raise EACRException.Create(10832,ErrorLCannotLockTable,['S',FTableName]);
      try
        if (not FExclusive) and (not FInTransaction) then
         ReadTablePFSMapFile(FSessionID);
        inherited;
      finally
        if (not TryToUnlockTableS) then
         raise EACRException.Create(10833,ErrorLCannotUnlockTable,['S',FTableName]);
      end;
    finally
      Unlock;
    end;
  end;
end; // CheckConstraints


//------------------------------------------------------------------------------
// apply default values
//------------------------------------------------------------------------------
procedure TACRDiskTableData.ApplyDefaultValuesToRecordBuffer(
                  Cursor:       TACRCursor;
                  RecordBuffer: TACRRecordBuffer;
                  InsertMode:   Boolean
                                          );
begin
  Lock;
  try
    FSessionID := Cursor.Session.SessionID;
    if (InsertMode) then
     begin
      if (FSequenceManager.SequenceDefs.Count > 0) then
       begin
        if (not TryToLockTableRW) then
         raise EACRException.Create(10550,ErrorLCannotLockTable,
          ['RW',FTableName]);
        try
          if (not FExclusive) and (not FInTransaction) then 
            ReadMostUpdatedData(FSessionID); 
          inherited;
          if (not FInTransaction) then
            WriteMostUpdatedData(FSessionID); 
          if (not Cursor.Session.InTransaction) and (not FExclusive) then 
            LPageManager.ApplyChanges(FSessionID);
        finally
          if (not TryToUnlockTableRW) then 
            raise EACRException.Create(10551,ErrorLCannotUnlockTable,
              ['RW',FTableName]); 
        end;
       end 
      else
       inherited; 
     end // Insert
    else 
     begin 
     end; // Edit
  finally 
    Unlock; 
  end; 
end; // ApplyDefaultValuesToRecordBuffer 
 
	
//------------------------------------------------------------------------------
// insert record
//------------------------------------------------------------------------------
function TACRDiskTableData.InsertRecord(var Cursor: TACRCursor): Boolean;
var 
    IndexDefsWillBeUpdated:   Boolean;
    Pos:                      Pointer; 
begin
 Lock; 
 try 
   if (FRecordManager = nil) then
    raise EACRException.Create(10531,ErrorLNilPointer); 
   if (Cursor.CurrentRecordBuffer = nil) then
    raise EACRException.Create(10532,ErrorLNilPointer); 
   FSessionID := Cursor.Session.SessionID; 
   if (Cursor.Session.InTransaction) then
    begin
     TACRLocalSession(Cursor.Session).Transaction.StartTableModification(FTableName); 
    end // in transaction 
   else 
    begin 
     Result := TryToLockTableRW; 
     if (not Result) then 
      begin 
  {$IFDEF DEBUG_TRACE_TABLE_LOCKS}
   aaWriteToLog('InsertRecord failed, lock type = RW, SessionID = '+IntToStr(FSessionID));
  {$ENDIF} 
        raise EACRException.Create(10533,ErrorLCannotLockTable,
          ['RW',FTableName]);
      end; 
    end; // outside transaction
    try
      if (not FExclusive) and (not FInTransaction) then 
       ReadTablePFSMapFile(FSessionID); 
      IndexDefsWillBeUpdated := (InternalGetRecordCount = 0); 
      Pos := Cursor.SavePosition;
      try
        // add record to first empty space 
        Result := FRecordManager.AddRecord(
          Cursor.CurrentRecordBuffer,Cursor.CurrentRecordID,Cursor.Session.SessionID);
 
        if (Result) then 
         begin
           if (not FInTransaction) then
            WriteTablePFSMapFile(FSessionID);
           if (FIndexManager.IndexDefs.Count > 0)  then 
            begin 
              if (IndexDefsWillBeUpdated) then 
               if (not FExclusive) and (not FInTransaction) then 
                ReadMostUpdatedData(FSessionID);
              FIndexManager.InsertRecord(Cursor); 
              if (IndexDefsWillBeUpdated) then 
               if (not FInTransaction) then 
                WriteMostUpdatedData(FSessionID);
            end;
           if (not Cursor.Session.InTransaction) then
            begin 
              if (FIndexManager.IndexDefs.Count > 0) then
               FIndexManager.GetTemporaryPageManager.ApplyChanges(FSessionID);
              LPageManager.ApplyChanges(FSessionID);
            end; 
           Cursor.FirstPosition := False; 
           Cursor.LastPosition := False;
           if (TACRRecordBitmap(Cursor.RecordBitmap).Active) then
            UpdateRecordBitmapAfterInsertRecord(Cursor,Pos);
         end;
      finally
        Cursor.FreePosition(Pos);
      end; 
    finally
     if (not Cursor.Session.InTransaction) then
      if (not TryToUnlockTableRW) then 
        raise EACRException.Create(10534,ErrorLCannotUnlockTable,
          ['RW',FTableName]);
    end; 
 finally 
   Unlock; 
 end; 
end; // InsertRecord
 
 
//------------------------------------------------------------------------------
// delete record
//------------------------------------------------------------------------------
function TACRDiskTableData.DeleteRecord(Cursor: TACRCursor): Boolean;
var 
  RecordID:                 TACRRecordID; 
  NewRecordID:              TACRRecordID;
  Buffer:                   TACRRecordBuffer;
  TempBuffer:               TACRRecordBuffer; 
  OldPos:                   Pointer; 
  IndexDefsWillBeUpdated:   Boolean; 
begin
 Result := False;
 Lock;
 try
  if (FRecordManager = nil) then
   raise EACRException.Create(10560,ErrorLNilPointer); 
  FSessionID := Cursor.Session.SessionID;
  if (Cursor.Session.InTransaction) then 
    begin
     TACRLocalSession(Cursor.Session).Transaction.StartTableModification(FTableName);
    end // in transaction
  else
    begin
      Result := TryToLockTableRW;
      if (not Result) then
       raise EACRException.Create(10561,ErrorLCannotLockTable,
         ['RW',FTableName]);
    end; // outside transaction
  try
    FRecordID := Cursor.CurrentRecordID;
    Result := TryUsingTimeOutForTableLocks(LockRecordU,LockParams);
    if (not Result) then
      raise EACRException.Create(10562,ErrorLCannotLockRecord,
       ['U',FTableName,FRecordID.PageNo,FRecordID.PageItemNo]);
    try
      RecordID := Cursor.CurrentRecordID;
      if (FConstraintManager.ConstraintDefs.ForeignKeysActionsDeleteExists) then
       ExecuteForeignKeyActions(Cursor,False);
      Buffer := Cursor.CurrentRecordBuffer;
      // clear blob fields
      ClearBLOBFieldsInRecordBuffer(Buffer);
      TempBuffer := Cursor.AllocateRecordBuffer;
      OldPos := Cursor.SavePosition;
      try
       Cursor.CurrentRecordBuffer := TempBuffer;
       if (GetRecordBuffer(Cursor, grmNext) = grrOK) then
         NewRecordID := Cursor.CurrentRecordID
       else
        begin
          Cursor.RestorePosition(OldPos);
          if (GetRecordBuffer(Cursor, grmPrior) = grrOK) then
            NewRecordID := Cursor.CurrentRecordID;
        end;
       Cursor.RestorePosition(OldPos);
       Cursor.CurrentRecordBuffer := Buffer;
       Cursor.CurrentRecordID := RecordID;
      finally
       ClearBLOBFieldsInRecordBuffer(TempBuffer);
       Cursor.FreePosition(OldPos);
       Cursor.FreeRecordBuffer(TempBuffer);
      end;
      if (not FExclusive) and (not FInTransaction) then
       ReadTablePFSMapFile(FSessionID);
      IndexDefsWillBeUpdated := (InternalGetRecordCount = 1);
      if (FIndexManager.IndexDefs.Count > 0) then
       begin
        if (IndexDefsWillBeUpdated) then
         if (not FExclusive) and (not FInTransaction) then
          ReadMostUpdatedData(FSessionID);
        FIndexManager.DeleteRecord(Cursor);
        if (IndexDefsWillBeUpdated) then
         if (not FInTransaction) then
          WriteMostUpdatedData(FSessionID);
       end;
      Result := FRecordManager.DeleteRecord(RecordID,FSessionID);
      if (Result) then
       begin
        if (not FInTransaction) then
         WriteTablePFSMapFile(FSessionID);
        if (TACRRecordBitmap(Cursor.RecordBitmap).Active) then
          TACRRecordBitmap(Cursor.RecordBitmap).HideRecord(Cursor.CurrentRecordID);
        if (Cursor.GetRecordCount = 0) then
         begin
          Cursor.FirstPosition := True;
          Cursor.LastPosition := False;
         end
        else
         begin
          Cursor.FirstPosition := False;
          Cursor.LastPosition := False;
          Cursor.CurrentRecordID := NewRecordID;
         end;
         if (not Cursor.Session.InTransaction) then
          begin
            if (FIndexManager.IndexDefs.Count > 0) then
             FIndexManager.GetTemporaryPageManager.ApplyChanges(FSessionID);
            LPageManager.ApplyChanges(FSessionID);
          end;
       end;
    finally
     if (not UnlockRecordU) then
      raise EACRException.Create(10563,ErrorLCannotUnlockRecord,
         ['U',FTableName,FRecordID.PageNo,FRecordID.PageItemNo]);
    end;
  finally
   if (not Cursor.Session.InTransaction) then
    if (not TryToUnlockTableRW) then
      raise EACRException.Create(10564,ErrorLCannotUnlockTable,
        ['RW',FTableName]);
  end;
 finally
  Unlock;
 end;
end; // DeleteRecord


//------------------------------------------------------------------------------
// update record
//------------------------------------------------------------------------------
function TACRDiskTableData.UpdateRecord(Cursor: TACRCursor): Boolean;
begin
 Lock;
 try
  if (FRecordManager = nil) then
   raise EACRException.Create(10556,ErrorLNilPointer);
  if (Cursor.CurrentRecordBuffer = nil) then
   raise EACRException.Create(10557,ErrorLNilPointer);
  FSessionID := Cursor.Session.SessionID;
  if (Cursor.Session.InTransaction) then
    begin
     TACRLocalSession(Cursor.Session).Transaction.StartTableModification(FTableName);
    end // in transaction
  else
    begin
      Result := TryToLockTableRW;
      if (not Result) then
       raise EACRException.Create(10558,ErrorLCannotLockTable,
         ['RW',FTableName]);
    end; // outside transaction
  try
    if (FConstraintManager.ConstraintDefs.ForeignKeysActionsUpdateExists) then
     ExecuteForeignKeyActions(Cursor,True);
    if (FIndexManager.IndexDefs.Count > 0) then
      FIndexManager.UpdateRecord(Cursor);
    if (not FExclusive) and (not FInTransaction) then
      ReadTablePFSMapFile(FSessionID);
    Result := FRecordManager.UpdateRecord(Cursor.CurrentRecordBuffer,
                                Cursor.CurrentRecordID,FSessionID);
    if (Result) then
     begin
       if (not FInTransaction) then
         WriteTablePFSMapFile(FSessionID);
       if (not Cursor.Session.InTransaction) then
        begin
          if (FIndexManager.IndexDefs.Count > 0) then
           FIndexManager.GetTemporaryPageManager.ApplyChanges(FSessionID);
          LPageManager.ApplyChanges(FSessionID);
        end;
       if (not UnlockRecordU) then
        raise EACRException.Create(10582,ErrorLCannotUnlockRecord,
          ['U',FTableName,FRecordID.PageNo,FRecordID.PageItemNo]);
       if (TACRRecordBitmap(Cursor.RecordBitmap).Active) then
        UpdateRecordBitmapAfterUpdateRecord(Cursor);
     end;
  finally
   if (not Cursor.Session.InTransaction) then
    if (not TryToUnlockTableRW) then
      raise EACRException.Create(10559,ErrorLCannotUnlockTable,
        ['RW',FTableName]);
  end;
 finally
  Unlock;
 end;
end; // UpdateRecord


//------------------------------------------------------------------------------
// start record editing
//------------------------------------------------------------------------------
procedure TACRDiskTableData.EditRecord(Cursor: TACRCursor);
begin
  Lock;
  try
    FRecordID := Cursor.CurrentRecordID;
    FSessionID := Cursor.Session.SessionID;
    if (not Cursor.Session.InTransaction) then
     begin
       // check for active transactions or other locks
       if (not TryToLockTableIRW) then
        raise EACRException.Create(10552,ErrorLCannotLockTable,
         ['IRW',FTableName]);
       if (not TryToUnlockTableIRW) then
            raise EACRException.Create(10554,ErrorLCannotUnlockTable,
              ['IRW',FTableName]);
     end;
    // lock record
    if (not TryUsingTimeOutForTableLocks(LockRecordU,LockParams)) then

     raise EACRException.Create(10583,ErrorLCannotLockRecord,
      ['U',FTableName,FRecordID.PageNo,FrecordID.PageItemNo]);
  finally
    Unlock;
  end;
end; // EditRecord


//------------------------------------------------------------------------------
// cancel record inserting or editing
//------------------------------------------------------------------------------
procedure TACRDiskTableData.CancelRecord(Cursor: TACRCursor; ToInsert: Boolean);
begin
  Lock;
  try
    FSessionID := Cursor.Session.SessionID;
    FRecordID := Cursor.CurrentRecordID;
    inherited CancelRecord(Cursor,ToInsert);
    if (not ToInsert) then
     begin
       if (not UnlockRecordU) then
        raise EACRException.Create(10581,ErrorLCannotUnlockRecord,
          ['U',FTableName,FRecordID.PageNo,FRecordID.PageItemNo]);
     end; // Edit
  finally
    Unlock;
  end;
end; // CancelRecord


//------------------------------------------------------------------------------
// delete all visible records
//------------------------------------------------------------------------------
procedure TACRDiskTableData.DeleteVisibleRecords(Cursor: TACRCursor);
var VisibleRecordCount: TACRRecordNo;
    Filtered:           Boolean;
begin
  Lock;
  try
    if (FRecordManager = nil) then
     raise EACRException.Create(11307,ErrorLNilPointer);
    FSessionID := Cursor.Session.SessionID;
    if (Cursor.Session.InTransaction) then
      begin
       TACRLocalSession(Cursor.Session).Transaction.StartTableModification(FTableName);
      end // in transaction
    else
      begin
        if (not TryToLockTableRW) then
         raise EACRException.Create(11308,ErrorLCannotLockTable,
           ['RW',FTableName]);
      end; // outside transaction
    try
       LockTableForRead(FSessionID);
       try
        Filtered := TACRRecordBitmap(Cursor.RecordBitmap).Active;
        if (Filtered) then
         VisibleRecordCount := TACRRecordBitmap(Cursor.RecordBitmap).GetRecordCount
        else
         VisibleRecordCount := FCachedRecordCount;
         // delete
        if (VisibleRecordCount > 0) then
         begin
           if (FCachedRecordCount < VisibleRecordCount) then
            raise EACRException.Create(11311,ErrorLInvalidRecordCount,[VisibleRecordCount,FCachedRecordCount])
           else
           if ((FCachedRecordCount = VisibleRecordCount) and
             (not FConstraintManager.ConstraintDefs.ForeignKeysActionsDeleteExists)) then
            begin
              FRecordManager.Empty(FSessionID);
              EmptyAllIndexes(FSessionID);
              if (not FInTransaction) then
                WriteMostUpdatedData(FSessionID);
            end // empty table
           else
            begin
             inherited DeleteVisibleRecords(Cursor);
            end;
         end;
         Cursor.FirstPosition := True;
         Cursor.LastPosition := False;
         if (not FInTransaction) then
           WriteTablePFSMapFile(FSessionID);
         if (not Cursor.Session.InTransaction) then
          begin
            if (FIndexManager.IndexDefs.Count > 0) then
             FIndexManager.GetTemporaryPageManager.ApplyChanges(FSessionID);
            LPageManager.ApplyChanges(FSessionID);
          end;
       finally
         UnlockTableForRead(FSessionID);
       end;
    finally
     if (not Cursor.Session.InTransaction) then
      if (not TryToUnlockTableRW) then
        raise EACRException.Create(11309,ErrorLCannotUnlockTable,
          ['RW',FTableName]);
    end;
  finally
    Unlock;
  end;
end; // DeleteVisibleRecords


//------------------------------------------------------------------------------
// Update visible records
//------------------------------------------------------------------------------
procedure TACRDiskTableData.UpdateVisibleRecords(
                                Cursor:       TACRCursor;
                                FieldNames:   TStringList;
                                values:       array of TACRVariant;
                                SkipFKCheck:  Boolean = False
                                );
begin
  Lock;
  try
    if (FRecordManager = nil) then
     raise EACRException.Create(11451,ErrorLNilPointer);
    FSessionID := Cursor.Session.SessionID;
    if (Cursor.Session.InTransaction) then
      begin
       TACRLocalSession(Cursor.Session).Transaction.StartTableModification(FTableName);
      end // in transaction
    else
      begin
        if (not TryToLockTableRW) then
         raise EACRException.Create(11452,ErrorLCannotLockTable,
           ['RW',FTableName]);
      end; // outside transaction
    try
       LockTableForRead(FSessionID);
       try
         inherited;
         if (not FInTransaction) then
           WriteTablePFSMapFile(FSessionID);
         if (not Cursor.Session.InTransaction) then
          begin
            if (FIndexManager.IndexDefs.Count > 0) then
             FIndexManager.GetTemporaryPageManager.ApplyChanges(FSessionID);
            LPageManager.ApplyChanges(FSessionID);
          end;
       finally
         UnlockTableForRead(FSessionID);
       end;
    finally
     if (not Cursor.Session.InTransaction) then
      if (not TryToUnlockTableRW) then
        raise EACRException.Create(11453,ErrorLCannotUnlockTable,
          ['RW',FTableName]);
    end;
  finally
    Unlock;
  end;
end; // UpdateVisibleRecords


//------------------------------------------------------------------------------
// create blob stream
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalCreateBlobStream(
          Cursor:   TACRCursor;
          ToInsert: Boolean;
          FieldNo:  Integer;
          OpenMode: TACRBLOBOpenMode
          ): TACRStream;
var TempStream:             TACRTemporaryStream;
    CompressedStream:       TACRCompressedBLOBStream;
    Buffer:                 PChar;
    Offset,BufferSize:      Integer;
    BLOBDescriptor:         TACRBLOBDescriptor;
    TempDescriptor:         TACRPartialTemporaryBLOBDescriptor;
    RecordID:               TACRRecordID;
begin
  Lock;
  try
    if ((ToInsert and (OpenMode = bomWrite)) or (OpenMode = bomWrite) or
       (Cursor.CurrentRecordBuffer = nil) or
       (CheckNullFlag(FieldNo,Cursor.CurrentRecordBuffer))) then
      begin
       Result := inherited InternalCreateBlobStream(Cursor,ToInsert,FieldNo,OpenMode);
      end
    else
    if (IsBLOBModified(Cursor.CurrentRecordBuffer,FieldManager.FieldDefs[FieldNo])) then
      begin
       Result := inherited InternalCreateBlobStream(Cursor,ToInsert,FieldNo,OpenMode);
      end
    else
     begin
       // read blob value from disk
       Result := nil;
       BLOBDescriptor.CompressionAlgorithm :=
           Byte(FieldManager.FieldDefs[FieldNo].BLOBCompressionAlgorithm);
       BLOBDescriptor.CompressionMode := FieldManager.FieldDefs[FieldNo].BLOBCompressionMode;
       BLOBDescriptor.BlockSize := FieldManager.FieldDefs[FieldNo].BLOBBlockSize;
       if (BLOBDescriptor.BlockSize = 0) then
        raise EACRException.Create(10628,ErrorLZeroBlockSizeIsNotAllowed);
       BLOBDescriptor.StartPosition := 0;
       TempStream := TACRTemporaryStream.Create;
       // create new compressed stream
          // copy value from TableData
          Offset := FieldManager.FieldDefs[FieldNo].MemoryOffset;
          FSessionID := Cursor.Session.SessionID;
          if (not TryToLockTableS) then
            raise EACRException.Create(10629,ErrorLCannotLockTable,
            ['S',FTableName]);
          try
           Move(PChar(Cursor.CurrentRecordBuffer + Offset)^,RecordID,SizeOf(RecordID));
           Buffer := nil;

           // read blob value from disk
             TACRDiskRecordManager(FRecordManager).InternalReadBLOBOrVarcharValue(
              FSessionID,RecordID,False,TempDescriptor,Buffer,BufferSize);
{
           try
             TACRDiskRecordManager(FRecordManager).InternalReadBLOBOrVarcharValue(
              FSessionID,RecordID,False,TempDescriptor,Buffer,BufferSize);
           except
             if (not FRepair) then
              raise
             else
              begin
               // blob value cannot be read
               SetNullFlag(True,FieldNo,Cursor.CurrentRecordBuffer);
               CompressedStream := TACRCompressedBLOBStream.Create(TempStream,
                BLOBDescriptor,False);
               Result := TACRLocalBLOBStream.Create(CompressedStream,Cursor,OpenMode,FieldNo);
               Exit;
              end;
           end;
}              
           if (Buffer = nil) then
            raise EACRException.Create(10631,ErrorLNilPointer);
           try
             // copy partial blob descriptor
             BLOBDescriptor.NumBlocks := TempDescriptor.NumBlocks;
             BLOBDescriptor.UncompressedSize := TempDescriptor.UncompressedSize;
             TempStream.Write(Buffer^,BufferSize);
             TempStream.Position := 0;
             CompressedStream := TACRCompressedBLOBStream.Create(TempStream,
              BLOBDescriptor,False);
             Result := TACRLocalBLOBStream.Create(CompressedStream,Cursor,OpenMode,FieldNo);
           finally
             MemoryManager.FreeAndNilMem(Buffer);
           end;
          finally
          if (not TryToUnlockTableS) then
           raise EACRException.Create(10630,ErrorLCannotUnlockTable,
            ['S',FTableName]);
          end;
     end;
  finally
    Unlock;
  end;
end; // InternalCreateBlobStream


//------------------------------------------------------------------------------
// clear blob field
//------------------------------------------------------------------------------
procedure TACRDiskTableData.ClearBLOBFieldInRecordBuffer(
              var RecordBuffer: TACRRecordBuffer;
              FieldNo:      Integer
              );
var Buffer:                   PChar;
begin
 Lock;
 try
   if (FBLOBFieldsPresent) then
    begin
     // check null
     if (not ((PByte(RecordBuffer + (FieldNo div 8))^) and
            (1 shl (FieldNo mod 8)) <> 0)) then
      begin
       if (IsBLOBModified(RecordBuffer,
                                FieldManager.FieldDefs[FieldNo])) then
       begin
         Move(PChar(RecordBuffer +
           FieldManager.FieldDefs[FieldNo].MemoryOffset)^,Buffer,sizeof(Buffer));
         if (Buffer = nil) then
          raise EACRException.Create(10118,ErrorLNilPointer);
         MemoryManager.FreeAndNilMem(Buffer);
       end;

       // set null
       PByte(RecordBuffer + (FieldNo div 8))^ := PByte(RecordBuffer + (FieldNo div 8))^ or
        (1 shl (FieldNo mod 8));
      end;
    end;
 finally
  Unlock;
 end;
end; // ClearBLOBFieldInRecordBuffer


//------------------------------------------------------------------------------
// return record count
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalGetRecordCount: TACRRecordNo;
begin
 if (FRecordManager = nil) then
   raise EACRException.Create(10560,ErrorLNilPointer);
 if (FNumLocksForRead = 0) then
  Result := FRecordManager.GetRecordCount
 else
  Result := FCachedRecordCount; 
end; // GetRecordCount


//------------------------------------------------------------------------------
// move cursor to specified position and set current record id in cursor
//------------------------------------------------------------------------------
procedure TACRDiskTableData.InternalSetRecNo(Cursor: TACRCursor; RecNo: TACRRecordNo);
var RecordID: TACRRecordID;
begin
  if (FRecordManager = nil) then
   raise EACRException.Create(10566,ErrorLNilPointer);
  FSessionID := Cursor.Session.SessionID;
  RecordID := TACRDiskRecordManager(FRecordManager).GetRecordIDByRecNo(
                  RecNo,FSessionID);
  if (RecordID.PageNo = INVALID_PAGE_NO) then
     raise EACRException.Create(10580,ErrorLCannotSetRecNo, [RecNo,FTableName]);
  Cursor.CurrentRecordID := RecordID;
end; // SetRecNo


//------------------------------------------------------------------------------
// get current record position from cursor
//------------------------------------------------------------------------------
function TACRDiskTableData.InternalGetRecNo(Cursor: TACRCursor): TACRRecordNo;
begin
  if (FRecordManager = nil) then
   raise EACRException.Create(10567,ErrorLNilPointer);
  FSessionID := Cursor.Session.SessionID;
  Result := TACRDiskRecordManager(FRecordManager).GetRecNoByRecordID(
                Cursor.CurrentRecordID,FSessionID);
end; // GetRecNo


//------------------------------------------------------------------------------
// return size of the bitmap
//------------------------------------------------------------------------------
function TACRDiskTableData.GetBitmapSize(Session: TACRSessionID): TACRRecordNo;
var BitSize1,BitSize,i,k: Integer;
begin
  k := TACRDiskRecordManager(FRecordManager).RecordsPerPage;
  BitSize := 0;
  for i := 0 to (SizeOf(TACRPageItemID) * 8 - 1) do
   if (((1 shl i) and k) <> 0) then
    BitSize := i+1;
  BitSize1 := 0;
  LPageManager.LoadDBHeader;
  k := LPageManager.PageCount;
  for i := 0 to 31 do
   if (((1 shl i) and k) <> 0) then
    BitSize1 := i+1;
  Result := 1 shl (BitSize + BitSize1 + 1);
end; // GetBitmapSize


//------------------------------------------------------------------------------
// return filter bitmap rec no by record id
//------------------------------------------------------------------------------
function TACRDiskTableData.GetBitmapRecNoByRecordID(RecordID: TACRRecordID): TACRRecordNo;
var BitSize,i,k,mask: Integer;
begin
  k := TACRDiskRecordManager(FRecordManager).RecordsPerPage;
  BitSize := 0;
  for i := 0 to (SizeOf(TACRPageItemID) * 8 - 1) do
   if (((1 shl i) and k) <> 0) then
    BitSize := i+1;
  mask := 0;
  for i := 0 to BitSize - 1 do
   mask := mask or (1 shl i);
  Result := (RecordID.PageNo shl BitSize) or (mask and RecordID.PageItemNo);
end; // GetBitmapRecNoByRecordID


//------------------------------------------------------------------------------
// return filter bitmap rec no by record id
//------------------------------------------------------------------------------
function TACRDiskTableData.GetRecordIDByBitmapRecNo(RecordNo: TACRRecordNo): TACRRecordID;
var BitSize,i,k,mask: Integer;
begin
  k := TACRDiskRecordManager(FRecordManager).RecordsPerPage;
  BitSize := 0;
  for i := 0 to (SizeOf(TACRPageItemID) * 8 - 1) do
   if (((1 shl i) and k) <> 0) then
    BitSize := i+1;
  mask := 0;
  for i := 0 to BitSize - 1 do
   mask := mask or (1 shl i);
  Result.PageNo := TACRPageNo(RecordNo shr BitSize);
  Result.PageItemNo := TACRPageItemID(mask and RecordNo);
end; // GetRecordIDByBitmapRecNo


//------------------------------------------------------------------------------
// LastAutoincValue
//------------------------------------------------------------------------------
function TACRDiskTableData.LastAutoincValue(FieldNo: Integer; Session: TACRBaseSession): Int64;
var
  id:           TACRObjectID;
begin
 Lock;
 try
  if (FieldNo >= FFieldManager.FieldDefs.Count) then
    raise EACRException.Create(10568, ErrorGFieldWithObjectIdNotFound, [FieldNo]);
  id := FFieldManager.FieldDefs[FieldNo].SequenceDefObjectId;
  FSessionID := Session.SessionID;
  if (not TryToLockTableS) then
    raise EACRException.Create(10569,ErrorLCannotLockTable,
      ['S',FTableName]);
  try
   ReadMostUpdatedData(FSessionID);
   Result := FSequenceManager.GetLastVal(Session, id);
  finally
    if (not TryToUnlockTableS) then
      raise EACRException.Create(10570,ErrorLCannotUnlockTable,
        ['S',FTableName])
  end;
 finally
   Unlock;
 end;
end;//LastAutoincValue


//------------------------------------------------------------------------------
// SetLastAutoincValue
//------------------------------------------------------------------------------
procedure TACRDiskTableData.SetLastAutoincValue(Value: Int64; FieldNo: Integer; Cursor: TACRCursor);
var
  id:           TACRObjectID;
begin
 Lock;
 try
  if (FieldNo >= FFieldManager.FieldDefs.Count) then
    raise EACRException.Create(10584, ErrorGFieldWithObjectIdNotFound, [FieldNo]);
  id := FFieldManager.FieldDefs[FieldNo].SequenceDefObjectId;
  FSessionID := Cursor.Session.SessionID;
  if (not TryToLockTableRW) then
    raise EACRException.Create(10586,ErrorLCannotLockTable,
      ['RW',FTableName]);
  try
   ReadMostUpdatedData(FSessionID);
   TACRSequenceDef(FSequenceManager.SequenceDefs.GetDefByObjectId(id)).LastValue :=
    Value;
   WriteMostUpdatedData(FSessionID);
   if (not Cursor.Session.InTransaction) then
     LPageManager.ApplyChanges(FSessionID);
  finally
    if (not TryToUnlockTableRW) then
      raise EACRException.Create(10585,ErrorLCannotUnlockTable,
        ['RW',FTableName])
  end;
 finally
   Unlock;
 end;
end; // SetLastAutoincValue


//------------------------------------------------------------------------------
// lock table for read and load data for fast navigation
//------------------------------------------------------------------------------
procedure TACRDiskTableData.LockTableForRead(aSessionID: TACRSessionID);
begin
  Lock;
  try
    FSessionID := aSessionID;
    TryToLockTableS;
    if (FNumLocksForRead = 0) then
     begin
      if (not FInTransaction) then
       ReadTablePFSMapFile(aSessionID);
      FCachedRecordCount := InternalGetRecordCount;
      FLastDBStateOnLockTable := LPageManager.GetState;
     end;
    inherited;
  finally
    Unlock;
  end;
end; // LockTableForRead


//------------------------------------------------------------------------------
// unlock table for read and free data for fast navigation
//------------------------------------------------------------------------------
procedure TACRDiskTableData.UnlockTableForRead(aSessionID: TACRSessionID);
begin
  Lock;
  try
    FSessionID := aSessionID;
    inherited;
    TryToUnlockTableS;
  finally
    Unlock;
  end;
end; // UnlockTableForRead


//------------------------------------------------------------------------------
// lock table in S mode
//------------------------------------------------------------------------------
function TACRDiskTableData.TryToLockTableS: Boolean;
begin
 Result := TryUsingTimeoutForTableLocks(InternalLockTableS,FLockParams);
end; // LockTableS


//------------------------------------------------------------------------------
// unlock table in S mode
//------------------------------------------------------------------------------
function TACRDiskTableData.TryToUnlockTableS: Boolean;
begin
 Result := InternalUnlockTableS;
end; // UnlockTableS


//------------------------------------------------------------------------------
// lock table in IRW mode
//------------------------------------------------------------------------------
function TACRDiskTableData.TryToLockTableIRW: Boolean;
begin
 Result := TryUsingTimeoutForTableLocks(InternalLockTableIRW,FLockParams);
end; // LockTableIRW


//------------------------------------------------------------------------------
// unlock table in IRW mode
//------------------------------------------------------------------------------
function TACRDiskTableData.TryToUnlockTableIRW: Boolean;
begin
 Result := InternalUnlockTableIRW;
end; // UnlockTableIRW


//------------------------------------------------------------------------------
// lock table in RW mode
//------------------------------------------------------------------------------
function TACRDiskTableData.TryToLockTableRW: Boolean;
begin
 Result := TryUsingTimeoutForTableLocks(InternalLockTableRW,FLockParams);
end; // LockTableRW


//------------------------------------------------------------------------------
// unlock table in RW mode
//------------------------------------------------------------------------------
function TACRDiskTableData.TryToUnlockTableRW: Boolean;
begin
 Result := InternalUnlockTableRW;
end; // UnockTableRW


//------------------------------------------------------------------------------
// lock table in X
//------------------------------------------------------------------------------
function TACRDiskTableData.TryToLockTableX: Boolean;
begin
 Result := TryUsingTimeoutForTableLocks(InternalLockTableX,FLockParams);
end;


//------------------------------------------------------------------------------
// unlock table in X
//------------------------------------------------------------------------------
function TACRDiskTableData.TryToUnlockTableX: Boolean;
begin
 Result := InternalUnlockTableX;
end; // TryToUnlockTableX


//------------------------------------------------------------------------------
// lock table in IS
//------------------------------------------------------------------------------
function TACRDiskTableData.TryToLockTableIS: Boolean;
begin
 Result := TryUsingTimeoutForTableLocks(InternalLockTableIS,FLockParams);
end;


//------------------------------------------------------------------------------
// unlock table in IS
//------------------------------------------------------------------------------
function TACRDiskTableData.TryToUnlockTableIS: Boolean;
begin
 Result := InternalUnlockTableIS;
end; // TryToUnlockTableIS


//------------------------------------------------------------------------------
// start transaction
//------------------------------------------------------------------------------
procedure TACRDiskTableData.StartTransaction(aSessionID: TACRSessionID);
begin
  if (not FInTransaction) then
   begin
    FInTransaction := True;
    if (not FExclusive) then
     begin
      FLastDBState := LPageManager.GetState;
      ReadMostUpdatedData(aSessionID);
      ReadTablePFSMapFile(aSessionID);
     end;
   end;
end; // StartTransaction


//------------------------------------------------------------------------------
// commit
//------------------------------------------------------------------------------
procedure TACRDiskTableData.Commit(aSessionID: TACRSessionID);
begin
  if (FInTransaction) then
   begin
    // restore cached data modified by the transaction
    WriteMostUpdatedData(aSessionID);
    WriteTablePFSMapFile(aSessionID);
    FInTransaction := False;
    if (FIndexManager.IndexDefs.Count > 0) then
     FIndexManager.GetTemporaryPageManager.ApplyChanges(aSessionID);
   end;
end; // Commit


//------------------------------------------------------------------------------
// rollback
//------------------------------------------------------------------------------
procedure TACRDiskTableData.Rollback(aSessionID: TACRSessionID);
begin
  if (FInTransaction) then
   begin
    FInTransaction := False;
    if (FExclusive) then
     begin
      // restore cached data modified by the transaction
      ReadMostUpdatedData(aSessionID);
      ReadTablePFSMapFile(aSessionID);
     end;
    if (FIndexManager.IndexDefs.Count > 0) then
     FIndexManager.GetTemporaryPageManager.CancelChanges(aSessionID);
   end;
end; // Rollback


//------------------------------------------------------------------------------
// return true if read page synchronization is required
//------------------------------------------------------------------------------
function TACRDiskTableData.IsSynchronizationRequired(PageState: Cardinal): Boolean;
begin
  if (FExclusive) or (FInTransaction and (PageState >= FLastDBState)) then
   Result := False
  else
   Result := True;
  if (Result and (FNumLocksForRead > 0)) then
   begin
    if (PageState >= FLastDBStateOnLockTable) then
     Result := False;
   end;
end; // IsSynchronizationRequired


////////////////////////////////////////////////////////////////////////////////
//
// TACRDatabaseFile
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Convert ACRAccessMode to Windows or Linux file open access modes
//------------------------------------------------------------------------------
function TACRDatabaseFile.ConvertAccessMode(am: TACRAccessMode): DWORD;
begin
  case am of
    amReadOnly:
{$IFDEF MSWINDOWS}
      Result := GENERIC_READ;
{$ENDIF}
{$IFDEF LINUX}
      Result := fmOpenRead;
{$ENDIF}
    amReadWrite:
{$IFDEF MSWINDOWS}
      Result := GENERIC_READ or GENERIC_WRITE;
{$ENDIF}
{$IFDEF LINUX}
      Result := fmOpenReadWrite;
{$ENDIF}
    else
      raise EACRException.Create(30366, ErrorGUnknownAccessMode, [Integer(am)]);
  end;
end;//ConvertAccessMode


//------------------------------------------------------------------------------
// Convert ACRShareMode to Windows or Linux file open share modes
//------------------------------------------------------------------------------
function TACRDatabaseFile.ConvertShareMode(sm: TACRShareMode): DWORD;
begin
  case sm of
    smExclusive:
{$IFDEF MSWINDOWS}
      Result := 0;
{$ENDIF}
{$IFDEF LINUX}
      Result := fmShareExclusive;
{$ENDIF}
    smShared:
{$IFDEF MSWINDOWS}
      Result := FILE_SHARE_READ or FILE_SHARE_WRITE;
{$ENDIF}
{$IFDEF LINUX}
      Result := fmShareDenyNone;
{$ENDIF}
    else
      raise EACRException.Create(30367, ErrorGUnknownShareMode, [Integer(sm)]);
  end;
end;//ConvertShareMode


//------------------------------------------------------------------------------
// if file opened then raise
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.CheckOpened(OperationName: String);
begin
  if not IsOpened then
    raise EACRException.Create(30376, ErrorGFileIsClosed, [OperationName]);
end;//CheckOpened


//------------------------------------------------------------------------------
// File Seek
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.CheckClosed(OperationName: String);
begin
  if IsOpened then
    raise EACRException.Create(30377, ErrorGFileIsOpened, [OperationName]);
end;//CheckClosed


//------------------------------------------------------------------------------
// File Seek
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.Seek(Offset: Int64; ErrorCode: Integer = 0);
var
  pos:      Int64;
  ECode:    Integer;
begin
  CheckOpened('Seek');
{$IFDEF MSWINDOWS}
{$R-}
  pos := Offset;
  Int64Rec(pos).Lo := SetFilePointer(THandle(FHandle), Int64Rec(pos).Lo,
    @Int64Rec(pos).Hi, 0);
//  HOffset := Int64Rec(Offset).Hi;
//  Int64Rec(pos).Lo := SetFilePointer(FHandle, Int64Rec(Offset).Lo, @HOffset, 0);
//  Int64Rec(pos).Hi := HOffset;
{$R+}
{$ENDIF}
{$IFDEF LINUX}
  pos := lseek64(FHandle, Offset, SEEK_SET);
{$ENDIF}
  if (pos <> Offset) then
    begin
      if ErrorCode = 0 then
        ECode := 30372
      else
        ECode := ErrorCode;
      raise EACRException.Create(ECode, ErrorGCannotSetFilePosition,
                                 [Offset, FFileName, Position, Size]);
    end;
{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog('Seek for '+IntToStr(Offset));
{$ENDIF}
end;//Seek


//------------------------------------------------------------------------------
// File Set Position
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.SetPosition(Offset: Int64);
begin
  CheckOpened('SetPosition');
{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog('SetPosition to '+IntToStr(Offset)+':');
{$ENDIF}
  Seek(Offset);
{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog('');
{$ENDIF}
end;//SetPosition


//------------------------------------------------------------------------------
// File Get Position
//------------------------------------------------------------------------------
function TACRDatabaseFile.GetPosition: Int64;
var
  SysErrorCode: DWORD;
{$IFDEF MSWINDOWS}
  HOffset:  Integer;
{$ENDIF}
begin
  CheckOpened('GetPosition');
{$IFDEF MSWINDOWS}
  HOffset := 0;
  Int64Rec(Result).Lo := SetFilePointer(FHandle, 0, @HOffset, FILE_CURRENT);
  Int64Rec(Result).Hi := HOffset;
{$ENDIF}
{$IFDEF LINUX}
  Result := lseek64(FHandle, 0, SEEK_CUR);
{$ENDIF}
  if (Result = INVALID_ID4) then
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(30373, ErrorGCannotGetFilePosition,
                              [FFileName, SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;
{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog('GetPosition: '+IntToStr(Result));
aaWriteToLog('');
{$ENDIF}
end;//GetPosition


//------------------------------------------------------------------------------
// Set File Size
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.SetSize(NewSize: Int64);
var
  SysErrorCode: DWORD;
begin
  CheckOpened('SetSize');
  // SetPosition
  Seek(NewSize);
  // Truncate File
{$IFDEF MSWINDOWS}
  if (not SetEndOfFile(FHandle)) then
    begin
{$ENDIF}
{$IFDEF LINUX}
  FSize := NewSize;
  if ftruncate(FHandle, Position) = -1 then
    begin
{$ENDIF}
      SysErrorCode := GetLastError;
      raise EACRException.Create(30370, ErrorGCannotTruncateFile,
                                 [FFileName, NewSize, SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;
{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog('SetSize: '+IntToStr(NewSize));
aaWriteToLog('');
{$ENDIF}
end;//SetSize


//------------------------------------------------------------------------------
// Get File Size
//------------------------------------------------------------------------------
function TACRDatabaseFile.GetSize: Int64;
var
{$IFDEF MSWINDOWS}
  HSize: Integer;
{$ENDIF}
  SysErrorCode: DWORD;
begin
  CheckOpened('GetSize');
{$IFDEF MSWINDOWS}
  Int64Rec(Result).Lo := GetFileSize(FHandle, @HSize);
  if (Int64Rec(Result).Lo = INVALID_FILE_SIZE) then
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(30371, ErrorGCannotGetFileSize,
                                 [FFileName, SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;
  Int64Rec(Result).Hi := HSize;
{$ENDIF}
{$IFDEF LINUX}
  Result := FSize;
{$ENDIF}

{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog('GetSize: '+IntToStr(Result));
aaWriteToLog('');
{$ENDIF}
end;//GetSize


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRDatabaseFile.Create;
begin
  FFileName := '';

  FAccessMode := amReadWrite;
  FShareMode  := smShared;

{$IFDEF MSWINDOWS}
  FAttrFlags  := FILE_ATTRIBUTE_NORMAL or FILE_FLAG_SEQUENTIAL_SCAN;
{$ENDIF}
{$IFDEF LINUX}
  FAttrFlags  := FileAccessRights;
  FSize := INVALID_FILE_SIZE;
{$ENDIF}

  FHandle     := INVALID_HANDLE_VALUE;
  FIsOpened   := False;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRDatabaseFile.Destroy;
begin
  if IsOpened then
    CloseFile;
end;//Destroy


//------------------------------------------------------------------------------
// Create and Open File
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.CreateAndOpenFile(FileName: String);
var
  SysErrorCode: DWORD;
begin
  CheckClosed('CreateFile');

  if FileName = '' then
    raise EACRException.Create(30374, ErrorGBlankFileName);

  FFileName := FileName;
  FAccessMode := amReadWrite;
  FShareMode := smExclusive;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabaseFile.CreateAndOpenFile starting... '+
   ', FFileName = '+FileName+
   ', AccessMode = '+IntToStr(Integer(FAccessMode))+
   ', ShareMode = '+IntToStr(Integer(FShareMode))
   );
 {$ENDIF}
{$IFDEF MSWINDOWS}
  FHandle := Windows.CreateFile(
                                  PChar(FFileName),
                                  ConvertAccessMode(FAccessMode),
                                  ConvertShareMode(FShareMode),
                                  nil,
                                  CREATE_ALWAYS,
                                  FAttrFlags,
                                  0
                               );
{$ENDIF}
{$IFDEF LINUX}
  FHandle := Integer(Libc.open(
                                  PChar(FFileName),
                                  ConvertAccessMode(FAccessMode)
                                  or ConvertShareMode(FShareMode)
                                  or O_ASYNC
                                  or O_TRUNC
                                  or O_CREAT,
                                  FileAccessRights
                               ));
{$ENDIF}
  if (FHandle = INVALID_HANDLE_VALUE) then
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(30361, ErrorGCreateFileError,
                                 [FFileName, SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;

{$IFDEF LINUX}
  FSize := 0;
{$ENDIF}

  FIsOpened := True;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabaseFile.CreateAndOpenFile starting... OK'+
   ', FFileName = '+FileName+
   ', FHandle = '+IntToStr(FHandle)+
   ', AccessMode = '+IntToStr(Integer(FAccessMode))+
   ', ShareMode = '+IntToStr(Integer(FShareMode))
   );
 {$ENDIF}

{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog(FileName+' created and opened as #'+IntToStr(Integer(FHandle)));
aaWriteToLog('');
{$ENDIF}
end;//CreateFile


//------------------------------------------------------------------------------
// Delete File
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.DeleteFile;
var
  SysErrorCode: DWORD;
begin
  CheckClosed('DeleteFile');
  if not SysUtils.DeleteFile(PChar(FFileName)) then
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(30362, ErrorGDeleteFileError,
                                 [FFileName, SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;
end;//DeleteFile


//------------------------------------------------------------------------------
// Rename Closed File
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.RenameFile(OldFileName, NewFileName: String);
var
  SysErrorCode: DWORD;
begin
  //CheckClosed('RenameFile');
  if not SysUtils.RenameFile(PChar(OldFileName), PChar(NewFileName)) then
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(30378, ErrorGRenameFileError,
          [OldFileName,NewFileName, SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;
end;//RenameFile


//------------------------------------------------------------------------------
// Open File
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.OpenFile(FileName: String; var AccessMode: TACRAccessMode; var ShareMode: TACRShareMode);
var
  SysErrorCode: DWORD;
  sr:           TSearchRec;
{$IFDEF LINUX}
  FileStat : _stat;
{$ENDIF}
begin
  CheckClosed('OpenFile');

  FFileName := FileName;
  FAccessMode := AccessMode;
  FShareMode := ShareMode;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabaseFile.OpenFile checking if file exists... '+
   ', FFileName = '+FileName+
   ', AccessMode = '+IntToStr(Integer(FAccessMode))+
   ', ShareMode = '+IntToStr(Integer(FShareMode))
   );
 {$ENDIF}
  if (SysUtils.FindFirst(FFileName,faAnyFile,sr) <> 0) then
   begin
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabaseFile.OpenFile checking if file exists... FILE DOES NOT EXIST! '+
   ', FFileName = '+FileName+
   ', AccessMode = '+IntToStr(Integer(FAccessMode))+
   ', ShareMode = '+IntToStr(Integer(FShareMode))
   );
 {$ENDIF}
    SysUtils.FindClose(sr);
    raise EACRException.Create(11351,ErrorLDatabaseFileDoesNotExists,[FFileName]);
   end
  else
   begin
    if ((sr.Attr and SysUtils.faReadOnly) <> 0) then
     begin
      FAccessMode := amReadOnly;
     end;
    SysUtils.FindClose(sr);
   end;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabaseFile.OpenFile opening file... '+
   ', FFileName = '+FileName+
   ', AccessMode = '+IntToStr(Integer(FAccessMode))+
   ', ShareMode = '+IntToStr(Integer(FShareMode))
   );
 {$ENDIF}
{$IFDEF MSWINDOWS}
  FHandle := Windows.CreateFile(
                                  PChar(FFileName),
                                  ConvertAccessMode(FAccessMode),
                                  ConvertShareMode(FShareMode),
                                  nil,
                                  OPEN_EXISTING,
                                  FAttrFlags,
                                  0
                               );
{$ENDIF}
{$IFDEF LINUX}
{
  if (Libc.stat64(PChar(FFileName), FileStat) = -1) then
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(40018, ErrorRStatFileError,
                                 ['size', FFileName, SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;
  FSize := FileStat.st_size;
}
  FHandle := Integer(Libc.open(
                                  PChar(FFileName),
                                  ConvertAccessMode(FAccessMode)
                                  or ConvertShareMode(FShareMode)
                                  or O_ASYNC,
                                  FileAccessRights
                               ));
{$ENDIF}
  if (FHandle = INVALID_HANDLE_VALUE) then
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(30364, ErrorGOpenFileError,
                                 [FFileName, SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;

{$IFDEF LINUX}
  if (Libc.fstat(FHandle, FileStat) = -1) then
//  if (Libc.stat64(PChar(FFileName), FileStat) = -1) then
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(40018, ErrorRStatFileError,
                                 ['size', FFileName, SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;
  FSize := FileStat.st_size;
{$ENDIF}

  AccessMode := FAccessMode;
  ShareMode := FShareMode;
  FIsOpened := True;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabaseFile.OpenFile finished '+
   ', FFileName = '+FileName+
   ', FHandle = '+IntToStr(FHandle)+
   ', AccessMode = '+IntToStr(Integer(FAccessMode))+
   ', ShareMode = '+IntToStr(Integer(FShareMode))+#13#10

   );
 {$ENDIF}
{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog(FileName+' opened as #'+IntToStr(Integer(FHandle)));
aaWriteToLog('');
{$ENDIF}
end;//OpenFile



//------------------------------------------------------------------------------
// Close File
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.CloseFile;
var
  SysErrorCode: DWORD;
begin
  CheckOpened('CloseFile');
{$IFDEF MSWINDOWS}
  if (not Windows.CloseHandle(FHandle))
{$ENDIF}
{$IFDEF LINUX}
  if Libc.__close(FHandle) = -1 // No need to unlock since all locks are released on close.
{$ENDIF}
   then
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(30365, ErrorGCloseFileError,
                                 [FFileName, SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;
{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog('File #'+IntToStr(Integer(FHandle))+' closed');
aaWriteToLog('');
{$ENDIF}
  FHandle := INVALID_HANDLE_VALUE;
  FIsOpened := False;
end;//CloseFile


//------------------------------------------------------------------------------
// Read Buffer
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.ReadBuffer(var Buffer; const Count, Pos: Int64; ErrorCode: Integer);
var
// ALEX: Changed the following Windows code:
//  NumberOfBytesRead:  DWORD;
  SysErrorCode:       DWORD;
begin
  CheckOpened('ReadBuffer');
  // Set Position
  Seek(Pos, ErrorCode);
  // Read Bytes
// ALEX: Changed the following Windows code:
//  if not ReadFile(FHandle, Buffer, Count, NumberOfBytesRead, nil) then
  if SysUtils.FileRead(Integer(FHandle), Buffer, Count) = -1 then
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(ErrorCode, ErrorGReadFileError,
                                 [FFileName, Pos, Count, self.Position, self.Size,
                                 SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;
{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog('From file #'+IntToStr(Integer(FHandle))+' read '+IntToStr(Count)+' bytes starting @'+IntToStr(Pos)+':');
GetPosition;
{$ENDIF}
end;//ReadBuffer


//------------------------------------------------------------------------------
// Write Buffer
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.WriteBuffer(const Buffer; const Count, Pos: Int64; ErrorCode: Integer);
var
{$IFDEF DEBUG_LOG_FILES}
  str1 : String;
{$ENDIF}
// ALEX: Changed the following Windows code:
//  NumberOfBytesWritten: DWORD;
  SysErrorCode: DWORD;
begin
  CheckOpened('WriteBuffer');
  // Set Position
  Seek(Pos, ErrorCode);
  // Write Count of bytes from Buffer
// ALEX: Changed the following Windows code:
//  if not WriteFile(FHandle, Buffer, Count, NumberOfBytesWritten, nil) then
  if (SysUtils.FileWrite(Integer(FHandle), Buffer, Count) = -1) then
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(ErrorCode, ErrorGWriteFileError,
                                 [FFileName, Pos, Count, self.Position, self.Size,
                                 SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;
{$IFDEF LINUX}
  if ((Pos + Count) > FSize) then
    FSize := Pos + Count;
{$ENDIF}
{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog('To file #'+IntToStr(Integer(FHandle))+' write '+IntToStr(Count)+' bytes starting @'+IntToStr(Pos)+':');
GetPosition;
{$ENDIF}
end;//WriteBuffer


//------------------------------------------------------------------------------
// Flush Buffers
//------------------------------------------------------------------------------
procedure TACRDatabaseFile.FlushFileBuffers;

var
  SysErrorCode: DWORD;
begin
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabaseFile.FlushFileBuffers starting ...'+
   ', FFileName = '+FFileName);
 {$ENDIF}
  CheckOpened('FlushFileBuffers');
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabaseFile.FlushFileBuffers flushing...'+
   ', FFileName = '+FFileName+
   'FHandle = '+IntToStr(FHandle));
 {$ENDIF}
{$IFDEF MSWINDOWS}
  if (not Windows.FlushFileBuffers(FHandle)) then
{$ENDIF}
{$IFDEF LINUX}
// There is no need to flush buffers because the only low-level input and output
// functions that operate on file descriptors are used.
// If open with O_SYNC:
  if (Libc.fsync(FHandle) = -1) then
{$ENDIF}
    begin
      SysErrorCode := GetLastError;
      raise EACRException.Create(30379, ErrorGFlushFileBufferError,
                       [FFileName,  SysErrorCode, SysErrorMessage(SysErrorCode)]);
    end;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabaseFile.FlushFileBuffers flushing... OK '+
   ', FFileName = '+FFileName+
   ', FHandle = '+IntToStr(FHandle));
 {$ENDIF}

{$IFDEF DEBUG_LOG_FILES}
aaWriteToLog('FlushFileBuffers of file #'+IntToStr(Integer(FHandle)));
{$ENDIF}

end;//FlushFileBuffers


//------------------------------------------------------------------------------
// Lock Byte
//------------------------------------------------------------------------------
function TACRDatabaseFile.LockByte(Offset: Int64; Count: Integer = 1): Boolean;
 {$IFDEF LINUX}
 var
  LockP: TFLock;
 {$ENDIF}
begin
  CheckOpened('LockByte');
{$IFDEF FILE_SERVER_VERSION}
  if (FShareMode = smExclusive) then
   Result := True
  else
 {$IFDEF MSWINDOWS}
    Result := Windows.LockFile(FHandle, Int64Rec(Offset).Lo, Int64Rec(Offset).Hi, Count, 0);
 {$ENDIF}
 {$IFDEF LINUX}
   begin
    LockP.l_type := F_WRLCK;
    LockP.l_whence := SEEK_SET;
    LockP.l_start := Offset;
    LockP.l_len := Int64(Count);
    if (Libc.fcntl(FHandle, F_SETLK, LockP) = -1)
     then
      Result := False
     else
      Result := True;
   end;
 {$ENDIF}
{$ELSE}
  Result := True;
{$ENDIF}
end;//LockByte


//------------------------------------------------------------------------------
// Unlock Byte
//------------------------------------------------------------------------------
function TACRDatabaseFile.UnlockByte(Offset: Int64; Count: Integer = 1): Boolean;
 {$IFDEF LINUX}
 var
  LockP: TFLock;
 {$ENDIF}
begin
  CheckOpened('UnlockByte');
{$IFDEF FILE_SERVER_VERSION}
  if (FShareMode = smExclusive) then
   Result := True
  else
 {$IFDEF MSWINDOWS}
   Result := Windows.UnLockFile(FHandle, Int64Rec(Offset).Lo, Int64Rec(Offset).Hi, Count, 0);
 {$ENDIF}
 {$IFDEF LINUX}
   begin
    LockP.l_type := F_UNLCK;
    LockP.l_whence := SEEK_SET;
    LockP.l_start := Offset;
    LockP.l_len := Int64(Count);
    if (Libc.fcntl(FHandle, F_SETLK, LockP) = -1)
     then
      Result := False
     else
      Result := True;
   end;
 {$ENDIF}
{$ELSE}
  Result := True;
{$ENDIF}
end;//UnlockByte


//------------------------------------------------------------------------------
// return TRUE if byte locked
//------------------------------------------------------------------------------
function TACRDatabaseFile.IsByteLocked(Offset: Int64): Boolean;
begin
  CheckOpened('IsByteLocked');
  Result := LockByte(Offset);
  if Result then
    UnlockByte(Offset);
  Result := not Result;
end;//IsByteNotLocked


//------------------------------------------------------------------------------
// return FALSE if any byte of region is locked
//------------------------------------------------------------------------------
function TACRDatabaseFile.IsRegionLocked(Offset: Int64; Count: Integer): Boolean;
begin
  CheckOpened('IsRegionLocked');
  Result := LockByte(Offset, Count);
  if Result then
    UnlockByte(Offset, Count);
  Result := not Result;
end;//IsByteRegionNotLocked


////////////////////////////////////////////////////////////////////////////////
//
// TACRDatabaseFreeSpaceManager
//
////////////////////////////////////////////////////////////////////////////////


{$IFDEF FREE_SPACE_MANAGER_ON}


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRDatabaseFreeSpaceManager.Create(
                        aPageManager: TACRDiskPageManager;
                        aLockParams:  TACRLockParams
                                               );
begin
  LPageManager := aPageManager;
  LockParams := aLockParams;
  GAM := nil;
  SGAM := nil;
  PFS := nil;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRDatabaseFreeSpaceManager.Destroy;
begin
  if (GAM <> nil) then
   MemoryManager.FreeAndNilMem(GAM);
  if (SGAM <> nil) then
   MemoryManager.FreeAndNilMem(SGAM);
  if (PFS <> nil) then
   MemoryManager.FreeAndNilMem(PFS);
  inherited;
end; // Destroy


//------------------------------------------------------------------------------
// CreateFreeSpaceManager
//------------------------------------------------------------------------------
procedure TACRDatabaseFreeSpaceManager.CreateFreeSpaceManager;

 var i : TACRPageNo;

begin
  if LPageManager.FDBHeader.ExtentPageCount < ACRMinExtentPageCount then
   raise EACRException.Create(50028,ErrorFInvalidExtPageCount);
  GAM :=  MemoryManager.AllocMem(LPageManager.PageDataSize);
  SGAM := MemoryManager.AllocMem(LPageManager.PageDataSize);
  PFS :=  MemoryManager.AllocMem(LPageManager.PageDataSize);
   // Set initial bits
  for i := 0 to 2 do
   SetNullFlag(true, i, PFS);
  SetNullFlag(true, 0, GAM);
  LPageManager.ExtendFile(3);
   // Write maps to file
  SavePage(ACRFirstPageNoGAM,fsmGAM);
  SavePage(ACRFirstPageNoSGAM,fsmSGAM);
  SavePage(ACRFirstPageNoPFS,fsmPFS);
  FillChar(GAM^,LPageManager.PageDataSize,0);
  FillChar(SGAM^,LPageManager.PageDataSize,0);
  FillChar(PFS^,LPageManager.PageDataSize,0);
end;// CreateFreeSpaceManager


//------------------------------------------------------------------------------
// LoadPage (0=PFS 1=GAM 2=SGAM)
//------------------------------------------------------------------------------
procedure TACRDatabaseFreeSpaceManager.LoadPage(IPageNo:    Integer;
                                                IPageType:  TACRFreeSpaceManagerPageType);
var PageHeader: TACRDiskPageHeader;
    Buffer:     PChar;
begin
  if (LPageManager.FDBHeader.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None) then
   begin
    Buffer := MemoryManager.GetMem(LPageManager.PageSize);
    try
      LPageManager.ReadPageRegion(Buffer^,IPageNo,-SizeOf(PageHeader),LPageManager.PageSize);
      ACRDecryptBuffer(LPageManager.CryptoInfo,Buffer,LPageManager.PageSize);
      Move(Buffer^,PageHeader,Sizeof(PageHeader));
      case IPageType of
       fsmPFS:
        begin
         if (PageHeader.PageType <> ACRPageTypeIDPFS) then
           raise EACRException.Create(10726,ErrorFInvalidPage,[IPageNo,Byte(IPageType)]);
         if (PFS = nil) then
          PFS := MemoryManager.GetMem(LPageManager.PageDataSize);
         Move(PChar(Buffer+SizeOf(PageHeader))^,PFS^,LPageManager.PageDataSize);
        end;
       fsmGAM:
        begin
         if (PageHeader.PageType <> ACRPageTypeIDGAM) then
           raise EACRException.Create(10727,ErrorFInvalidPage,[IPageNo,Byte(IPageType)]);
         if (GAM = nil) then
          GAM := MemoryManager.GetMem(LPageManager.PageDataSize);
         Move(PChar(Buffer+SizeOf(PageHeader))^,GAM^,LPageManager.PageDataSize);
        end;
       fsmSGAM:
        begin
         if (PageHeader.PageType <> ACRPageTypeIDSGAM) then
           raise EACRException.Create(10728,ErrorFInvalidPage,[IPageNo,Byte(IPageType)]);
         if (SGAM = nil) then
          SGAM := MemoryManager.GetMem(LPageManager.PageDataSize);
         Move(PChar(Buffer+SizeOf(PageHeader))^,SGAM^,LPageManager.PageDataSize);
        end
       else
        raise EACRException.Create(10729,ErrorFUnknownPageType,[Byte(IPageType)]);
      end;
    finally
      MemoryManager.FreeAndNilMem(Buffer);
    end;
   end
  else
   begin
    LPageManager.ReadPageRegion(PageHeader,IPageNo,-sizeOf(PageHeader),sizeOf(PageHeader));
    case IPageType of
     fsmPFS:
       begin
        if (PageHeader.PageType <> ACRPageTypeIDPFS) then
         raise EACRException.Create(50015,ErrorFInvalidPage,[IPageNo,Byte(IPageType)]);
        if (PFS = nil) then
         PFS := MemoryManager.GetMem(LPageManager.PageDataSize);
        LPageManager.ReadPageRegion(PFS^, IPageNo, 0, LPageManager.PageDataSize);
       end;
     fsmGAM:
       begin
        if (PageHeader.PageType <> ACRPageTypeIDGAM) then
         raise EACRException.Create(50016,ErrorFInvalidPage,[IPageNo,Byte(IPageType)]);
        if (GAM = nil) then
         GAM := MemoryManager.GetMem(LPageManager.PageDataSize);
        LPageManager.ReadPageRegion(GAM^, IPageNo, 0, LPageManager.PageDataSize);
       end;
     fsmSGAM:
       begin
        if (PageHeader.PageType <> ACRPageTypeIDSGAM) then
         raise EACRException.Create(50017,ErrorFInvalidPage,[IPageNo,Byte(IPageType)]);
        if (SGAM = nil) then
         SGAM := MemoryManager.GetMem(LPageManager.PageDataSize);
        LPageManager.ReadPageRegion(SGAM^, IPageNo, 0, LPageManager.PageDataSize);
       end
     else
      raise EACRException.Create(50013,ErrorFUnknownPageType,[Byte(IPageType)]);
    end;
   end;
end;// LoadPage


//------------------------------------------------------------------------------
// SavePage (0=PFS 1=GAM 2=SGAM)
//------------------------------------------------------------------------------
procedure TACRDatabaseFreeSpaceManager.SavePage(IPageNo:    Integer;
                                                IPageType:  TACRFreeSpaceManagerPageType);
var PageHeader: TACRDiskPageHeader;
    Buffer:     PChar;
begin
  PageHeader.Signature := ACRDiskPageSignature;
  PageHeader.State := 0;
  PageHeader.ObjectID := ACRObjectIDFreeSpaceManager;
  PageHeader.NextPageNo := INVALID_PAGE_NO;
  if (LPageManager.FDBHeader.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None) then
   begin
    Buffer := MemoryManager.GetMem(LPageManager.PageSize);
    try
      case IPageType of
       fsmPFS:
        begin
         PageHeader.PageType := ACRPageTypeIDPFS;
         Move(PFS^,PChar(Buffer+SizeOf(PageHeader))^,LPageManager.PageDataSize);
        end;
       fsmGAM:
        begin
         PageHeader.PageType := ACRPageTypeIDGAM;
         Move(GAM^,PChar(Buffer+SizeOf(PageHeader))^,LPageManager.PageDataSize);
        end;
       fsmSGAM:
        begin
         PageHeader.PageType := ACRPageTypeIDSGAM;
         Move(SGAM^,PChar(Buffer+SizeOf(PageHeader))^,LPageManager.PageDataSize);
        end;
      end;
      Move(PageHeader,Buffer^,Sizeof(PageHeader));
      ACREncryptBuffer(LPageManager.CryptoInfo,Buffer,LPageManager.PageSize);
      LPageManager.WritePageRegion(Buffer^,IPageNo,-SizeOf(PageHeader),LPageManager.PageSize);
    finally
      MemoryManager.FreeAndNilMem(Buffer);
    end;
   end
  else
   begin
    case IPageType of
     fsmPFS:
       begin
        PageHeader.PageType := ACRPageTypeIDPFS;
        if (PFS = nil) then
         raise EACRException.Create(50010,ErrorFNilPointer);
        LPageManager.WritePageRegion(PFS^, IPageNo, 0, LPageManager.PageDataSize);
       end;
     fsmGAM:
       begin
        PageHeader.PageType := ACRPageTypeIDGAM;
        if (GAM = nil) then
         raise EACRException.Create(50011,ErrorFNilPointer);
        LPageManager.WritePageRegion(GAM^, IPageNo, 0, LPageManager.PageDataSize);
       end;
     fsmSGAM:
       begin
        PageHeader.PageType := ACRPageTypeIDSGAM;
        if (SGAM = nil) then
         raise EACRException.Create(50012,ErrorFNilPointer);
        LPageManager.WritePageRegion(SGAM^, IPageNo, 0, LPageManager.PageDataSize);
       end
     else
      raise EACRException.Create(50014,ErrorFUnknownPageType,[Byte(IPageType)]);
    end;
    LPageManager.WritePageRegion(PageHeader,IPageNo,-SizeOf(PageHeader),SizeOf(PageHeader));
   end;
end;// SavePage


//------------------------------------------------------------------------------
// Append New Page
//------------------------------------------------------------------------------
function TACRDatabaseFreeSpaceManager.AppendNewPage(CurGAM: integer;
                                                    PFSPageNo, EmptyExt : TACRPageNo;
                                                    PFSNeed : Boolean): TACRPageNo;
 var
  GAMNo, i : TACRPageNo;
  GAMNeed : Boolean;
  PageNo, PFSNo : TACRPageNo;

begin
  PagePerExt := LPageManager.PageDataSize*8*LPageManager.FDBHeader.ExtentPageCount;

  if not PFSNeed then
   PFSNeed := (PFSCount*LPageManager.PageDataSize * 8) -
                               (LPageManager.FDBHeader.TotalPageCount + 1) < 0;
  GAMNeed := (GAMCount*PagePerExt) -
                               (LPageManager.FDBHeader.TotalPageCount + 1) < 0;

  if GAMNeed and PFSNeed then
    // Creating new GAM/SGAM pages
   begin
    FillChar(GAM^,LPageManager.PageDataSize,0);
    FillChar(SGAM^,LPageManager.PageDataSize,0);
    FillChar(PFS^,LPageManager.PageDataSize,0);
    PageNo := LPageManager.FDBHeader.TotalPageCount;
    for i:=0 to 3 do
     SetNullFlag(true, i, PFS);
    SetNullFlag(true, 0, GAM);
    if LPageManager.FDBHeader.ExtentPageCount = 4
     then
      SetNullFlag(true, 0, SGAM);
    GAMCount := GAMCount + 1;
    PFSCount := PFSCount + 1;
    LPageManager.ExtendFile(4);

    SavePage(PageNo+1,fsmGAM);
    SavePage(PageNo+2,fsmSGAM);
    SavePage(PageNo,fsmPFS);

    Result := PageNo + 3;
   end
  else
   if PFSNeed then
    begin
     FillChar(PFS^,LPageManager.PageDataSize,0);
     PageNo := LPageManager.FDBHeader.TotalPageCount;
     for i := 0 to 1 do
      SetNullFlag(true, i, PFS);
     GAMNo := (PageNo - (GAMCount-1) * PagePerExt) div
                                         LPageManager.FDBHeader.ExtentPageCount;
     if (GAMNo < 0) or (GAMNo >= LPageManager.PageDataSize * 8) then
      raise EACRException.Create(50018,ErrorFInvalidBitNo,[GAMNo,LPageManager.PageDataSize]);
     SetNullFlag(true, GAMNo, GAM);

     PFSCount := PFSCount + 1;
     LPageManager.ExtendFile(2);
     i := (PageNo - CurGAM*PagePerExt + 2) div LPageManager.FDBHeader.ExtentPageCount;
     if ((i-1) < 0) or ((i-1) >= LPageManager.PageDataSize * 8) then
      raise EACRException.Create(50019,ErrorFInvalidBitNo,[(i-1),LPageManager.PageDataSize]);
     if i > 0 then
      SetNullFlag(true, i - 1, SGAM);

     SavePage(ACRFirstPageNoGAM + (GAMCount-1) * PagePerExt,fsmGAM);
     SavePage(ACRFirstPageNoSGAM + (GAMCount-1) * PagePerExt,fsmSGAM);
     SavePage(PageNo,fsmPFS);

     Result := PageNo + 1;
    end
   else
    if (not PFSNeed) and (not GAMNeed) then
     begin
      PageNo := LPageManager.FDBHeader.TotalPageCount;
      GAMNo := (PageNo - (GAMCount-1) * PagePerExt) div
                                        LPageManager.FDBHeader.ExtentPageCount;
      if (GAMNo < 0) or (GAMNo >= LPageManager.PageDataSize * 8) then
       raise EACRException.Create(50020,ErrorFInvalidBitNo,[GAMNo,LPageManager.PageDataSize]);
      SetNullFlag(true, GAMNo, GAM);
      if EmptyExt = -1 then
       begin
        PFSNo := PageNo div (LPageManager.PageDataSize*8);
        PFSPageNo := PFSNo * LPageManager.PageDataSize*8;
        LoadPage(PFSPageNo,fsmPFS);
       end;
      if ((PageNo - PFSPageNo) < 0) or ((PageNo - PFSPageNo) >= LPageManager.PageDataSize * 8) then
       raise EACRException.Create(50021,ErrorFInvalidBitNo,[(PageNo - PFSPageNo),LPageManager.PageDataSize]);
      SetNullFlag(true, PageNo - PFSPageNo, PFS);

      i := ((PageNo - CurGAM*PagePerExt + 1) div LPageManager.FDBHeader.ExtentPageCount);
      if i > 0 then
       begin
        if ((i-1) < 0) or ((i-1) >= LPageManager.PageDataSize * 8) then
         raise EACRException.Create(50022,ErrorFInvalidBitNo,[(i-1),LPageManager.PageDataSize]);
        SetNullFlag(true, i - 1, SGAM);
       end;
      LPageManager.ExtendFile(1);
      SavePage(ACRFirstPageNoGAM + (GAMCount-1) * PagePerExt,fsmGAM);
      SavePage(ACRFirstPageNoSGAM + (GAMCount-1) * PagePerExt,fsmSGAM);
      SavePage(PFSPageNo,fsmPFS);
      Result := PageNo;
     end;
end; // AppendNewPage


//------------------------------------------------------------------------------
// checks if there are some empty pages at the end of file and cut file if found
//------------------------------------------------------------------------------
procedure TACRDatabaseFreeSpaceManager.CheckShrinking;
var PFSPageNo,PFSNo,PageNo: TACRPageNo;
    LastUsedPageNo:         TACRPageNo;
    BitNo:                  Integer;
    Delta:                  Integer;
begin
  PageNo := LPageManager.FDBHeader.TotalPageCount-1;
  LastUsedPageNo := PageNo;
  Delta := LPageManager.PageDataSize*8;
  PFSNo := PageNo div (Delta);
  PFSPageNo := PFSNo * Delta;
  BitNo := PageNo - PFSNo*Delta;
  repeat
    if ((PageNo = PFSPageNo) or (PFSPageNo >= LPageManager.FDBHeader.TotalPageCount))  then
        raise EACRException.Create(50032,ErrorFFreePFSMap,[PageNo]);
    LoadPage(PFSPageNo,fsmPFS);
    if (BitNo < 0) or (BitNo >= Delta) then
       raise EACRException.Create(50033,ErrorFInvalidBitNo, [BitNo,LPageManager.PageDataSize]);
    repeat
      if (CheckNullFlag(BitNo,PFS)) then
       begin
        PFSPageNo := 0;
        break;
       end
      else
       Dec(LastUsedPageNo);
      Dec(BitNo);
    until (BitNo < 0);
    BitNo := Delta-1;
    if (PFSPageNo >= Delta) then
      PFSPageNo := PFSPageNo - Delta
    else
      PFSPageNo := 0;
  until (PFSPageNo <= 0);
  // shrinking
  if (LastUsedPageNo < PageNo) then
    LPageManager.SetPageCount(LastUsedPageNo);
end; // CheckShrinkning


//------------------------------------------------------------------------------
// Get page
//------------------------------------------------------------------------------
function TACRDatabaseFreeSpaceManager.GetPage(DesiredStartPageNo: TACRPageNo = INVALID_PAGE_NO): TACRPageNo;
 var
  CurGAM, i, k                                  : Integer;
  EmptyExt, EmptyPage, PageNo, PFSPageNo        : TACRPageNo;
  FirstPage, LastPage, LastUsedExt, PFSNo, Mic  : TACRPageNo;

begin
  if (not TryUsingTimeOut(LPageManager.LockFreeSpaceManager,
                                                        LockParams)) then
   raise EACRException.Create(50000,ErrorFLockByTimeoutFailed,
                              [LockParams.RetryCount,LockParams.Delay]);
  try
   if (GAM = nil) then
    GAM := MemoryManager.AllocMem(LPageManager.PageDataSize)
   else
    FillChar(GAM^,LPageManager.PageDataSize,0);
   if (SGAM = nil) then
    SGAM := MemoryManager.AllocMem(LPageManager.PageDataSize)
   else
    FillChar(SGAM^,LPageManager.PageDataSize,0);
   if (PFS = nil) then
    PFS := MemoryManager.AllocMem(LPageManager.PageDataSize)
   else
    FillChar(PFS^,LPageManager.PageDataSize,0);
   LPageManager.LoadDBHeader;
   PagePerExt := LPageManager.PageDataSize*8*LPageManager.FDBHeader.ExtentPageCount;
   CurGAM := 0;
   PFSCount := ((LPageManager.FDBHeader.TotalPageCount - 1) div
                                    (LPageManager.PageDataSize*8)) + 1;
   GAMCount := ((LPageManager.FDBHeader.TotalPageCount - 1) div PagePerExt) + 1;

   repeat

    LoadPage(ACRFirstPageNoSGAM + CurGAM * PagePerExt , fsmSGAM);
    LoadPage(ACRFirstPageNoGAM + CurGAM * PagePerExt, fsmGAM);
    EmptyExt := -1;

    // Last existing extent in the GAM block

    Mic := (LPageManager.FDBHeader.TotalPageCount - (GAMCount - 1)*PagePerExt);

    if Mic = PagePerExt then
     begin
      Result := AppendNewPage(CurGAM, 0, 0, true);
      Break;
     end;

    if CurGAM = GAMCount - 1 then
     LastUsedExt :=  Mic div LPageManager.FDBHeader.ExtentPageCount
    else
     LastUsedExt := (PagePerExt div LPageManager.FDBHeader.ExtentPageCount) - 1;

    // Search mixed extent
    for i:=0 to LastUsedExt do
     begin
      if (i >= LPageManager.PageDataSize * 8) then
       raise EACRException.Create(50023,ErrorFInvalidBitNo,[i,LPageManager.PageDataSize]);
      if not CheckNullFlag(i, SGAM) then
       begin
        EmptyExt := i;
        Break;
       end;
     end;

    PageNo :=  LPageManager.FDBHeader.TotalPageCount;
    if PageNo = PFSCount*LPageManager.PageDataSize*8
    then
     begin
      Result := AppendNewPage(GAMCount - 1, 0, 0, true);
      Break;
     end;

    if EmptyExt <> -1 then
    // Search empty page using value of founded extent
     begin
      PageNo :=  EmptyExt * LPageManager.FDBHeader.ExtentPageCount;
      PFSNo := PageNo div (LPageManager.PageDataSize*8);
      PFSPageNo := PFSNo * LPageManager.PageDataSize*8 + CurGAM * PagePerExt;
      LoadPage(PFSPageNo,fsmPFS);
     // The first page of founded extent
      FirstPage := PageNo - PFSNo*LPageManager.PageDataSize*8;
     // The last page of founded extent
      LastPage := FirstPage + LPageManager.FDBHeader.ExtentPageCount - 1;
      EmptyPage := -1;
      if (FirstPage < 0) or (FirstPage >= LPageManager.PageDataSize*8) then
       raise EACRException.Create(50024,ErrorFInvalidBitNo, [FirstPage,LPageManager.PageDataSize]);
      if (LastPage < 0) or (LastPage >= LPageManager.PageDataSize*8) then
       raise EACRException.Create(50025,ErrorFInvalidBitNo, [LastPage,LPageManager.PageDataSize]);
     // Search of empty page
      for i := FirstPage to LastPage do
       if not CheckNullFlag(i, PFS) then
        begin
         EmptyPage := i + PFSNo*LPageManager.PageDataSize*8;
         Break;
        end;
      k := 0 ;
      for i := FirstPage to LastPage do
       if CheckNullFlag(i, PFS) then k := k + 1;

      if ((EmptyPage = -1) or (EmptyPage + CurGAM*PagePerExt >=
                                     LPageManager.FDBHeader.TotalPageCount))
                                          and (CurGAM + 1 = GAMCount) then
       // Create new empty page in the end of file
       begin
        Result := AppendNewPage(CurGAM, PFSPageNo, 0, false);
        Break;
       end
      else
       // Mark empty page as not empty and write PFS, GAM and SGAM pages to file
       begin
        if (EmptyPage < 0) or (EmptyPage - PFSPageNo >= LPageManager.PageDataSize*8) then
         raise EACRException.Create(50026,ErrorFInvalidBitNo, [EmptyPage,LPageManager.PageDataSize]);
        if (EmptyExt < 0) or (EmptyExt >= LPageManager.PageDataSize*8) then
         raise EACRException.Create(50027,ErrorFInvalidBitNo, [EmptyExt,LPageManager.PageDataSize]);


        SetNullFlag(true, EmptyPage - PFSPageNo, PFS);
        SetNullFlag(true, EmptyExt, GAM);

        if k >= 7 then
         SetNullFlag(true, EmptyExt, SGAM);

        Result := EmptyPage + CurGAM * PagePerExt;

        SavePage(ACRFirstPageNoGAM + CurGAM * PagePerExt,fsmGAM);
        SavePage(ACRFirstPageNoSGAM + CurGAM * PagePerExt,fsmSGAM);
        SavePage(PFSPageNo,fsmPFS);
        Break;
       end;
     end
    else
     begin
      if (CurGAM + 1 = GAMCount) then
       begin
        Result := AppendNewPage(CurGAM, 0, -1, false);
        Break;
       end
      else
       CurGAM := CurGAM + 1;
     end;
   until False;

   LPageManager.SaveDBHeader;
   FillChar(GAM^,LPageManager.PageDataSize,0);
   FillChar(SGAM^,LPageManager.PageDataSize,0);
   FillChar(PFS^,LPageManager.PageDataSize,0);

  finally
   if (not LPageManager.UnlockFreeSpaceManager) then
    raise EACRException.Create(50003,ErrorFUnlockFailed);
  end;
end; // GetPage


//------------------------------------------------------------------------------
// Free page
//------------------------------------------------------------------------------
procedure TACRDatabaseFreeSpaceManager.FreePage(PageNo: TACRPageNo);

var
  PFSPageNo, ExtNo, GAMPageNo:      TACRPageNo;
  StartPage, EndPage, PFSNo:        TACRPageNo;
  GAMNo, i, k:                      Integer;
  BitNo:                            Integer;

begin
  if (not TryUsingTimeOut(LPageManager.LockFreeSpaceManager,
                                                    LockParams)) then
   raise EACRException.Create(50002,ErrorFLockByTimeoutFailed,
                                  [LockParams.RetryCount,
                                    LockParams.Delay]);
  try
   if (GAM = nil) then
    GAM := MemoryManager.AllocMem(LPageManager.PageDataSize)
   else
    FillChar(GAM^,LPageManager.PageDataSize,0);
   if (SGAM = nil) then
    SGAM := MemoryManager.AllocMem(LPageManager.PageDataSize)
   else
    FillChar(SGAM^,LPageManager.PageDataSize,0);
   if (PFS = nil) then
    PFS := MemoryManager.AllocMem(LPageManager.PageDataSize)
   else
    FillChar(PFS^,LPageManager.PageDataSize,0);
   k := 0;
   PagePerExt := LPageManager.PageDataSize*8*LPageManager.FDBHeader.ExtentPageCount;
   LPageManager.LoadDBHeader;
   PFSNo := PageNo div (LPageManager.PageDataSize*8);
   PFSPageNo := PFSNo * LPageManager.PageDataSize*8;
   if ((PageNo = PFSPageNo) or (PFSPageNo >= LPageManager.FDBHeader.TotalPageCount))  then
    raise EACRException.Create(50001,ErrorFFreePFSMap,[PageNo]);
   LoadPage(PFSPageNo, fsmPFS);
   BitNo := PageNo - PFSNo*LPageManager.PageDataSize*8;
   if (BitNo < 0) or (BitNo >= LPageManager.PageDataSize*8) then
    raise EACRException.Create(50005,ErrorFInvalidBitNo, [BitNo,LPageManager.PageDataSize]);
   SetNullFlag(false,BitNo, PFS);

   SavePage(PFSPageNo,fsmPFS);

   ExtNo := PageNo div LPageManager.FDBHeader.ExtentPageCount;
   StartPage := ExtNo*LPageManager.FDBHeader.ExtentPageCount -
                                              PFSNo*LPageManager.PageDataSize*8;
   EndPage := StartPage + LPageManager.FDBHeader.ExtentPageCount - 1;
   if (StartPage < 0) or (StartPage >= LPageManager.PageDataSize*8) then
    raise EACRException.Create(50006,ErrorFInvalidBitNo, [StartPage,LPageManager.PageDataSize]);
   if (EndPage < 0) or (EndPage >= LPageManager.PageDataSize*8) then
    raise EACRException.Create(50007,ErrorFInvalidBitNo, [EndPage,LPageManager.PageDataSize]);
   for i := StartPage to EndPage do
    if CheckNullFlag(i, PFS) then
      k := k + 1;
   // Set null to empty extent in GAM
   if k < 1 then
    begin
     GAMNo := PageNo div PagePerExt;
     GAMPageNo := ACRFirstPageNoGAM + GAMNo * PagePerExt;
     LoadPage(GAMPageNo, fsmGAM);
     BitNo := ExtNo - GAMNo*LPageManager.PageDataSize*8;
     if (BitNo < 0) or (BitNo >= LPageManager.PageDataSize*8) then
      raise EACRException.Create(50008,ErrorFInvalidBitNo, [BitNo,LPageManager.PageDataSize]);
     SetNullFlag(false, BitNo, GAM);
     SavePage(GAMPageNo,fsmGAM);
     FillChar(GAM^,LPageManager.PageDataSize,0);
    end;
   // Set null to empty extent in SGAM
   if k < LPageManager.FDBHeader.ExtentPageCount then
    begin
     GAMNo := PageNo div PagePerExt;
     GAMPageNo := ACRFirstPageNoSGAM + GAMNo * PagePerExt;
     LoadPage(GAMPageNo, fsmSGAM);
     BitNo := ExtNo - GAMNo*LPageManager.PageDataSize*8;
     if (BitNo < 0) or (BitNo >= LPageManager.PageDataSize*8) then
      raise EACRException.Create(50009,ErrorFInvalidBitNo, [BitNo,LPageManager.PageDataSize]);
     SetNullFlag(false, BitNo, SGAM);
     SavePage(GAMPageNo,fsmSGAM);
     FillChar(SGAM^,LPageManager.PageDataSize,0);
    end;
   CheckShrinking;
   LPageManager.SaveDBHeader;
   FillChar(PFS^,LPageManager.PageDataSize,0);
  finally
   if (not LPageManager.UnlockFreeSpaceManager) then
    raise EACRException.Create(50004,ErrorFUnlockFailed);
  end;
end; // FreePage

function TACRDatabaseFreeSpaceManager.GetFreePageCount: TACRPageNo;
var
  i, j, EmptyPageCount, PFSPageNo, PFSCount, LastPage: TACRPageNo;
begin
 if (not TryUsingTimeOut(LPageManager.LockFreeSpaceManager,
                                                        LockParams)) then
  raise EACRException.Create(50030,ErrorFLockByTimeoutFailed,
                              [LockParams.RetryCount,LockParams.Delay]);
 try
  if (PFS = nil) then
   PFS := MemoryManager.AllocMem(LPageManager.PageDataSize)
  else
   FillChar(PFS^,LPageManager.PageDataSize,0);
  LPageManager.LoadDBHeader;
  PFSCount := ((LPageManager.FDBHeader.TotalPageCount - 1) div
                                    (LPageManager.PageDataSize*8)) + 1;
  EmptyPageCount := 0;

  for i := 0 to PFSCount - 1 do
   begin
    PFSPageNo := i*LPageManager.PageDataSize*8;
    LoadPage(PFSPageNo,fsmPFS);
    if LPageManager.PageDataSize*8 > LPageManager.FDBHeader.TotalPageCount - PFSPageNo
     then
      LastPage := LPageManager.FDBHeader.TotalPageCount - PFSPageNo - 1
     else
      LastPage := LPageManager.PageDataSize*8 - 1;
    for j := 0 to LastPage do
     if not CheckNullFlag(j, PFS) then
      EmptyPageCount := EmptyPageCount + 1;
    FillChar(PFS^,LPageManager.PageDataSize,0);
   end;
  LPageManager.SaveDBHeader;
 //  FillChar(PFS^,LPageManager.PageDataSize,0);
  Result := EmptyPageCount;
 finally
  if (not LPageManager.UnlockFreeSpaceManager) then
   raise EACRException.Create(50031,ErrorFUnlockFailed);
 end;
end;

{$ELSE}

//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRDatabaseFreeSpaceManager.Create(
                        aPageManager: TACRDiskPageManager;
                        aLockParams:  TACRLockParams
                                               );
begin
 LPageManager := aPageManager;
 LockParams := aLockParams;
end; // Create


//------------------------------------------------------------------------------
// CreateFreeSpaceManager
//------------------------------------------------------------------------------
procedure TACRDatabaseFreeSpaceManager.CreateFreeSpaceManager;
begin
; // do nothing
end;// CreateFreeSpaceManager


//------------------------------------------------------------------------------
// get page
//------------------------------------------------------------------------------
function TACRDatabaseFreeSpaceManager.GetPage(DesiredStartPageNo: TACRPageNo = INVALID_PAGE_NO): TACRPageNo;
begin
  if (not TryUsingTimeOut(LPageManager.LockFreeSpaceManager,
                                                        LockParams)) then
   raise EACRException.Create(10472,ErrorLLockByTimeoutFailed,
                              [LockParams.RetryCount,LockParams.Delay]);
  try
   LPageManager.ExtendFile(1);
   Result := LPageManager.FDBHeader.TotalPageCount-1;
  finally
   LPageManager.UnlockFreeSpaceManager;
  end;
end; // GetPage


//------------------------------------------------------------------------------
// free page
//------------------------------------------------------------------------------
procedure TACRDatabaseFreeSpaceManager.FreePage(PageNo: TACRPageNo);
begin
;
end; // FreePage
{$ENDIF}




////////////////////////////////////////////////////////////////////////////////
//
// TACRTableLocksManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// lock
//------------------------------------------------------------------------------
procedure TACRTableLocksManager.Lock;
begin
  EnterCriticalSection(FCSect);
end;// Lock


//------------------------------------------------------------------------------
// unlock
//------------------------------------------------------------------------------
procedure TACRTableLocksManager.Unlock;
begin
  LeaveCriticalSection(FCSect);
end;// Unlock


//------------------------------------------------------------------------------
// clear all counters
//------------------------------------------------------------------------------
procedure TACRTableLocksManager.InitTableLockInfo(var TableLockInfo: TACRTableLockInfo);
begin
 FillChar(TableLockInfo,SizeOf(TableLockInfo),$00);
end; // InitTableLockInfo


//------------------------------------------------------------------------------
// return true if table has no locks from this sessio
//------------------------------------------------------------------------------
function TACRTableLocksManager.IsLockInfoEmpty(var TableLockInfo: TACRTableLockInfo): Boolean;
begin
  Result := (TableLockInfo.NumLocksIS = 0) and
            (TableLockInfo.NumLocksX = 0) and
            (TableLockInfo.NumLocksIRW = 0) and
            (TableLockInfo.NumLocksRW = 0) and
            (TableLockInfo.NumLocksS = 0);
end; // IsLockInfoEmpty


//------------------------------------------------------------------------------
// return table lock counter value
//------------------------------------------------------------------------------
function TACRTableLocksManager.GetTableLockCount(var TableLockInfo: TACRTableLockInfo; LockType: TACRLockType): Integer;
begin
 case LockType of
  ltX: Result := TableLockInfo.NumLocksX;
  ltIS: Result := TableLockInfo.NumLocksIS;
  ltIRW: Result := TableLockInfo.NumLocksIRW;
  ltRW: Result := TableLockInfo.NumLocksRW;
  ltS: Result := TableLockInfo.NumLocksS;
 else
  raise EACRException.Create(10775,ErrorLInvalidLockType,[Integer(LockType)]);
 end;
end; // GetTableLockCount


//------------------------------------------------------------------------------
// set 0 to the table lock counter
//------------------------------------------------------------------------------
procedure TACRTableLocksManager.IncTableLockCount(var TableLockInfo: TACRTableLockInfo; LockType: TACRLockType);
begin
 case LockType of
  ltX: Inc(TableLockInfo.NumLocksX);
  ltIS: Inc(TableLockInfo.NumLocksIS);
  ltIRW: Inc(TableLockInfo.NumLocksIRW);
  ltRW: Inc(TableLockInfo.NumLocksRW);
  ltS: Inc(TableLockInfo.NumLocksS);
 else
  raise EACRException.Create(10776,ErrorLInvalidLockType,[Integer(LockType)]);
 end;
end; // IncTableLockCount


//------------------------------------------------------------------------------
// set 0 to the table lock counter
//------------------------------------------------------------------------------
procedure TACRTableLocksManager.DecTableLockCount(var TableLockInfo: TACRTableLockInfo; LockType: TACRLockType; Decrement: Integer = 1);
begin
 case LockType of
  ltX: Dec(TableLockInfo.NumLocksX,Decrement);
  ltIS: Dec(TableLockInfo.NumLocksIS,Decrement);
  ltIRW: Dec(TableLockInfo.NumLocksIRW,Decrement);
  ltRW: Dec(TableLockInfo.NumLocksRW,Decrement);
  ltS: Dec(TableLockInfo.NumLocksS,Decrement);
 else
  raise EACRException.Create(10777,ErrorLInvalidLockType,[Integer(LockType)]);
 end;
end; // DecTableLockCount


//------------------------------------------------------------------------------
// set 0 to the table lock counter
//------------------------------------------------------------------------------
procedure TACRTableLocksManager.ClearTableLockCount(var TableLockInfo: TACRTableLockInfo; LockType: TACRLockType);
begin
 case LockType of
  ltX: TableLockInfo.NumLocksX := 0;
  ltIS: TableLockInfo.NumLocksIS := 0;
  ltIRW: TableLockInfo.NumLocksIRW := 0;
  ltRW: TableLockInfo.NumLocksRW := 0;
  ltS: TableLockInfo.NumLocksS := 0;
 else
  raise EACRException.Create(10778,ErrorLInvalidLockType,[Integer(LockType)]);
 end;
end; // ClearTableLockCount


//------------------------------------------------------------------------------
// add session
//------------------------------------------------------------------------------
function TACRTableLocksManager.AddSession(PTableLockItem: PACRTableLockItem; SessionID: TACRSessionID): Integer;
var SessionItem: PACRSessionLockItem;
begin
 New(SessionItem);
 SessionItem^.SessionID := SessionID;
 InitTableLockInfo(SessionItem^.TableLockInfo);
 Result := PTableLockItem.SessionLocksList.Add(SessionItem);
end; // AddSession


//------------------------------------------------------------------------------
// find session
//------------------------------------------------------------------------------
function TACRTableLocksManager.FindSession(PTableLockItem: PACRTableLockItem; SessionID: TACRSessionID): Integer;
var i: Integer;
begin
  Result := -1;
  for i := 0 to PTableLockItem.SessionLocksList.Count - 1 do
   if (PACRSessionLockItem(PTableLockItem.SessionLocksList.Items[i])^.SessionID =
       SessionID) then
    begin
     Result := i;
     break;
    end;
end;


//------------------------------------------------------------------------------
// find or create session
//------------------------------------------------------------------------------
function TACRTableLocksManager.FindOrCreateSession(PTableLockItem: PACRTableLockItem; SessionID: TACRSessionID): Integer;
begin
  Result := FindSession(PTableLockItem,SessionID);
  if (Result < 0) then
   Result := AddSession(PTableLockItem,SessionID);
end; /// FindOrCreateSession


//------------------------------------------------------------------------------
// remove session
//------------------------------------------------------------------------------
procedure TACRTableLocksManager.RemoveSession(PTableLockItem: PACRTableLockItem; SessionItem: PACRSessionLockItem);
begin
 PTableLockItem.SessionLocksList.Remove(SessionItem);
 Dispose(SessionItem);
end; // RemoveSession


//------------------------------------------------------------------------------
// create new table lock item and return its address
//------------------------------------------------------------------------------
function TACRTableLocksManager.CreateTableLockItem(TableName: String; SessionID: TACRSessionID): Integer;
var Item:       PACRTableLockItem;
begin
 New(Item);
 Item^.TableName := AnsiUpperCase(TableName);
 Item^.SessionLocksList := TList.Create;
 AddSession(Item,SessionID);
 Item^.TableLocksFile := nil;
 Result := FTableLockItemList.Add(Item);
end; // CreateTableLockItem


//------------------------------------------------------------------------------
// remove table lock item
//------------------------------------------------------------------------------
procedure TACRTableLocksManager.DestroyTableLockItem(PTableLockItem: PACRTableLockItem);
begin
  if (PTableLockItem^.TableLocksFile <> nil) then
   PTableLockItem^.TableLocksFile.Free;
  while PTableLockItem^.SessionLocksList.Count > 0 do
    RemoveSession(PTableLockItem,PTableLockItem^.SessionLocksList.Items[0]);
  PTableLockItem^.SessionLocksList.Free;
  FTableLockItemList.Remove(PTableLockItem);
  Dispose(PTableLockItem);
end; // DestroyTableLockItem


//------------------------------------------------------------------------------
// return index of found item
//------------------------------------------------------------------------------
function TACRTableLocksManager.FindTableLockItem(TableName: String): Integer;
var i:          Integer;
    FTableName: String;
begin
 Result := -1;
 FTableName := AnsiUpperCase(TableName);
 for i := 0 to FTableLockItemList.Count-1 do
  if (PACRTableLockItem(FTableLockItemList.Items[i]).TableName = FTableName) then
    begin
      Result := i;
      break;
    end;
end; // FindTableLockItem


//------------------------------------------------------------------------------
// return index of found or created item
//------------------------------------------------------------------------------
function TACRTableLocksManager.FindOrCreateTableLockItem(TableName: String; SessionID: TACRSessionID): Integer;
begin
 Result := FindTableLockItem(TableName);
 if (Result < 0) then
  Result := CreateTableLockItem(TableName,SessionID);
end; // FindOrCreateTableLockItem


//------------------------------------------------------------------------------
// perform locking of the byte in physical database file if other lock applied
// by other applications / threads allows it
//------------------------------------------------------------------------------
function TACRTableLocksManager.InternalLockTable(
                        const CurrentRetryNo: Integer;
                        var TableItem:        TACRTableListItem;
                        PTableLockItem:       PACRTableLockItem;
                        SessionID:            TACRSessionID;
                        LockType:             TACRLockType;
                        PTableLockInfo:       PACRTableLockInfo = nil): Boolean;
begin
 if (PTableLockItem^.TableLocksFile = nil) then
  begin
    PTableLockItem^.TableLocksFile := TACRTableLocksFile.Create(LDiskDatabaseData.PageManager,LDiskDatabaseData.LockParams);
    PTableLockItem^.TableLocksFile.OpenFile(TableItem.LockFilePageNo);
  end;
 Result := PTableLockItem^.TableLocksFile.LockTableIfLockAllowed(CurrentRetryNo,SessionID,LockType);
end; // InternalLockTable


//------------------------------------------------------------------------------
// return true if lock specified by LockType is compatible with session lock item
//------------------------------------------------------------------------------
function TACRTableLocksManager.IsTableLockAllowed(
                    PSessionLockItem: PACRSessionLockItem;
                    LockType:         TACRLockType): Boolean;
begin
  case LockType of
   ltIS:
    Result := (PSessionLockItem^.TableLockInfo.NumLocksX = 0);
   ltX:
    Result := False;
   ltS:
    Result := (PSessionLockItem^.TableLockInfo.NumLocksX = 0) and
              (PSessionLockItem^.TableLockInfo.NumLocksRW = 0);
   ltIRW:
    Result := (PSessionLockItem^.TableLockInfo.NumLocksX = 0) and
              (PSessionLockItem^.TableLockInfo.NumLocksIRW = 0) and
              (PSessionLockItem^.TableLockInfo.NumLocksRW = 0);
   ltRW:
    Result := (PSessionLockItem^.TableLockInfo.NumLocksX = 0) and
              (PSessionLockItem^.TableLockInfo.NumLocksIRW = 0) and
              (PSessionLockItem^.TableLockInfo.NumLocksRW = 0) and
              (PSessionLockItem^.TableLockInfo.NumLocksS = 0)
  else
   raise EACRException.Create(10746,ErrorLInvalidLockType,[Integer(LockType)]);
  end;
end; // IsTableLockAllowed


//------------------------------------------------------------------------------
// perform unlocking of the byte in physical database file
//------------------------------------------------------------------------------
function TACRTableLocksManager.InternalUnlockTable(
                    PTableLockItem: PACRTableLockItem;
                    SessionID:      TACRSessionID;
                    LockType:       TACRLockType
                    ): Boolean;
begin
  if (PTableLockItem^.TableLocksFile = nil) then
   raise EACRException.Create(10762,ErrorLNilPointer);
  Result := PTableLockItem^.TableLocksFile.UnlockTableIfPossible(SessionID,LockType);
end; // InternalUnlockTable


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRTableLocksManager.Create(aDiskDatabaseData: TACRDiskDatabaseData);
begin
 InitializeCriticalSection(FCSect);
 LDiskDatabaseData := aDiskDatabaseData;
 FTableLockItemList := TList.Create;
 FInMemory := TACRDiskPageManager(LDiskDatabaseData.PageManager).Exclusive;
 {$IFNDEF FILE_SERVER_VERSION}
 FInMemory := True;
 {$ENDIF}
end; // Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRTableLocksManager.Destroy;
var bFailed: Boolean;
begin
 bFailed := (FTableLockItemList.Count > 0);
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
WriteCurrentLocksToLog('TACRTableLocksManager.Destroy');
aaWriteToLog('TACRTableLocksManager.Destroy bFailed = '+BoolToStr(bFailed,True));
aaWriteToLog('TACRTableLocksManager.Destroy FTableLockItemList.Count = '+IntToStr(FTableLockItemList.Count));
{$ENDIF}
 while FTableLockItemList.Count > 0 do
   DestroyTableLockItem(PACRTableLockItem(FTableLockItemList.Items[0]));
 FTableLockItemList.Free;
 DeleteCriticalSection(FCSect);
 inherited;
 if (bFailed) then
    raise EACRException.Create(10811,ErrorLNotAllLocksWereRemoved,[LDiskDatabaseData.DatabaseName]);
end; // Destroy


//------------------------------------------------------------------------------
// if result is true and PTableLockInfo is not nil -
// increment corresponding field of this record
//------------------------------------------------------------------------------
function TACRTableLocksManager.LockTable(
                        const CurrentRetryNo: Integer;
                        var TableItem:        TACRTableListItem;
                        TableName:            String;
                        SessionID:            TACRSessionID;
                        LockType:             TACRLockType;
                        PTableLockInfo:       PACRTableLockInfo = nil): Boolean;
var i,j,TableItemIndex,SessionItemIndex: Integer;
    SessionItem:                         PACRSessionLockItem;
    bCreated:                            Boolean;
begin
  Lock;
  try
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
aaWriteToLog('LockTable start, TableName = '+TableName+#13#10+
             'SessionID = '+IntToStr(SessionID)+#13#10+
             'LockType = '+IntToStr(Integer(LockType))+#13#10+
             'PTableLockInfo = '+IntToHex(Integer(PTableLockInfo),4)+#13#10
            );
WriteCurrentLocksToLog('LockTable Start');
{$ENDIF}
    Result := True;
    bCreated := False;
    TableItemIndex := FindTableLockItem(TableName);
    if (TableItemIndex < 0) then
     begin
      TableItemIndex := CreateTableLockItem(TableName,SessionID);
      if (TableItemIndex < 0) or (TableItemIndex >= FTableLockItemList.Count) then
       raise EACRException.Create(10745,ErrorLInvalidItemNumber,[TableItemIndex,FTableLockItemList.Count]);
      bCreated := True;
     end // new table is locked
    else
     begin
      if (TableItemIndex >= FTableLockItemList.Count) then
       raise EACRException.Create(10747,ErrorLInvalidItemNumber,[TableItemIndex,FTableLockItemList.Count]);
      // checking other session locks applied to this table
      j := PACRTableLockItem(FTableLockItemList.Items[TableItemIndex])^.SessionLocksList.Count-1;
      for i := 0 to j do
       begin
        SessionItem := PACRSessionLockItem(PACRTableLockItem(
          FTableLockItemList.Items[TableItemIndex])^.SessionLocksList.Items[i]);
        if (SessionItem^.SessionID <> SessionID) then
         begin
          Result := IsTableLockAllowed(SessionItem,LockType);
          if (not Result) then
           break;
         end;
       end;
     end;
   if (Result) then
    begin
     SessionItemIndex := FindOrCreateSession(FTableLockItemList.Items[TableItemIndex],SessionID);
     if ((SessionItemIndex < 0) or
          (SessionItemIndex >= PACRTableLockItem(FTableLockItemList.Items[TableItemIndex])^.SessionLocksList.Count)) then
       raise EACRException.Create(10744,ErrorLCannotFindSessionForLockingTable,
        [SessionID,TableName]);
     SessionItem := PACRSessionLockItem(PACRTableLockItem(
      FTableLockItemList.Items[TableItemIndex])^.SessionLocksList.Items[SessionItemIndex]);
     // lock bytes in file only if not exclusive mode
     if (SessionItem^.TableLockInfo.NumLocksX = 0) then
       if (not FInMemory) then
        if (GetTableLockCount(SessionItem^.TableLockInfo,LockType) = 0) then
          Result := InternalLockTable(CurrentRetryNo,TableItem,FTableLockItemList.Items[TableItemIndex],
                  SessionID,LockType,PTableLockInfo);
     if (Result) then
      begin
       IncTableLockCount(SessionItem^.TableLockInfo,LockType);
       if (PTableLockInfo <> nil) then
         IncTableLockCount(PTableLockInfo^,LockType);
      end;
    end; // lock new table or lock is compatible with other locks from this app

//aaStopTime;
  finally
   try
    if (not Result) and (bCreated) then
     DestroyTableLockItem(PACRTableLockItem(FTableLockItemList.Items[TableItemIndex]));
   except
   end;
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
aaWriteToLog('LockTable finish, TableName = '+TableName+#13#10+
             'SessionID = '+IntToStr(SessionID)+#13#10+
             'LockType = '+IntToStr(Integer(LockType))+#13#10+
             'PTableLockInfo = '+IntToHex(Integer(PTableLockInfo),4)+#13#10+
             'Result = '+BoolToStr(result,True)
            );
WriteCurrentLocksToLog('LockTable Finish');
{$ENDIF}
    Unlock;
  end;
end; // LockTable


//------------------------------------------------------------------------------
// if result is true and PTableLockInfo is not nil - decrement corresponding field of this record
// if ForceUnlock is true - unlocks all locks set by this session
//------------------------------------------------------------------------------
function TACRTableLocksManager.UnlockTable(
                      TableName:        String;
                      SessionID:        TACRSessionID;
                      LockType:         TACRLockType;
                      PTableLockInfo:   PACRTableLockInfo = nil;
                      ForceUnlock:      Boolean = False
                      ): Boolean;
var TableItemIndex,SessionItemIndex:      Integer;
    SessionItem:                          PACRSessionLockItem;
    DoNotPerformUnlock:                   Boolean;
    TableLockItem:                        PACRTableLockItem;
    Decrement, LockCount:                 Integer;
begin
  Lock;
  try
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
aaWriteToLog('UnlockTable start, TableName = '+TableName+#13#10+
             'SessionID = '+IntToStr(SessionID)+#13#10+
             'LockType = '+IntToStr(Integer(LockType))+#13#10+
             'PTableLockInfo = '+IntToHex(Integer(PTableLockInfo),4)+#13#10+
             'ForceUnlock = '+BoolToStr(ForceUnlock,True)
            );
WriteCurrentLocksToLog('UnlockTable Start');
{$ENDIF}
    Result := True;
    TableItemIndex := FindTableLockItem(TableName);
    if (TableItemIndex < 0) or (TableItemIndex >= FTableLockItemList.Count) then
       raise EACRException.Create(10754,ErrorLInvalidItemNumber,[TableItemIndex,FTableLockItemList.Count]);
    TableLockItem := PACRTableLockItem(FTableLockItemList.Items[TableItemIndex]);
    SessionItemIndex := FindSession(FTableLockItemList.Items[TableItemIndex],SessionID);
    if ((SessionItemIndex < 0) or
        (SessionItemIndex >= TableLockItem^.SessionLocksList.Count)) then
       raise EACRException.Create(10755,ErrorLCannotFindSessionForLockingTable,
        [SessionID,TableName]);
    SessionItem := PACRSessionLockItem(PACRTableLockItem(
      FTableLockItemList.Items[TableItemIndex])^.SessionLocksList.Items[SessionItemIndex]);

    if (PTableLockInfo <> nil) then
     Decrement := GetTableLockCount(PTableLockInfo^,LockType)
    else
     Decrement := 1;
    LockCount := GetTableLockCount(SessionItem^.TableLockInfo,LockType);
    if (not ForceUnlock) then
      if ((LockCount - Decrement) < 0) then
          raise EACRException.Create(10757,ErrorLInvalidLockCounter,
            [LockCount,Integer(LockType),TableName]);

    if (ForceUnlock) then
     DoNotPerformUnlock := (LockCount = 0)
    else
     DoNotPerformUnlock := ((LockCount - Decrement) > 0);
    if (SessionItem^.TableLockInfo.NumLocksX = 0) or (LockType = ltX) then
     if (not FInMemory) then
      if (not DoNotPerformUnlock) then
        Result := InternalUnlockTable(TableLockItem,SessionID,LockType);
    if (Result) then
     begin
      if (ForceUnlock) then
        ClearTableLockCount(SessionItem^.TableLockInfo,LockType)
      else
        DecTableLockCount(SessionItem^.TableLockInfo,LockType,Decrement);
     end;
    // if table is not needed by this session
    if (IsLockInfoEmpty(SessionItem^.TableLockInfo)) then
     begin
       RemoveSession(TableLockItem,SessionItem);
       if (TableLockItem^.SessionLocksList.Count = 0) then
        DestroyTableLockItem(TableLockItem);
     end;
  finally
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
aaWriteToLog('UnlockTable finish, TableName = '+TableName+#13#10+
             'SessionID = '+IntToStr(SessionID)+#13#10+
             'LockType = '+IntToStr(Integer(LockType))+#13#10+
             'PTableLockInfo = '+IntToHex(Integer(PTableLockInfo),4)+#13#10+
             'ForceUnlock = '+BoolToStr(ForceUnlock,True)+#13#10+
             'Result = '+BoolToStr(result,True)
            );
WriteCurrentLocksToLog('UnlockTable Finish');
{$ENDIF}
    Unlock;
  end;
end; // UnlockTable


//------------------------------------------------------------------------------
// decrements number of locks by this session and removes all locks if their number = 0
// if PTableLockInfo = nil - all table locks byt this session will be removed
//------------------------------------------------------------------------------
function TACRTableLocksManager.UnlockTableAll(
                        TableName:      String;
                        SessionID:      TACRSessionID;
                        PTableLockInfo: PACRTableLockInfo = nil): Boolean;
var ForceUnlock: Boolean;
begin
  try
   ForceUnlock := (PTableLockInfo = nil);
   if (FindTableLockItem(TableName) >=0 ) then
     UnlockTable(TableName,SessionID,ltS,PTableLockInfo,ForceUnlock);
   if (FindTableLockItem(TableName) >=0 ) then
     UnlockTable(TableName,SessionID,ltIRW,PTableLockInfo,ForceUnlock);
   if (FindTableLockItem(TableName) >=0 ) then
     UnlockTable(TableName,SessionID,ltRW,PTableLockInfo,ForceUnlock);
   if (FindTableLockItem(TableName) >=0 ) then
     UnlockTable(TableName,SessionID,ltIS,PTableLockInfo,ForceUnlock);
   if (FindTableLockItem(TableName) >=0 ) then
     UnlockTable(TableName,SessionID,ltX,PTableLockInfo,ForceUnlock);
   Result := True;
  except
   raise;
  end;
end; // UnlockTableAll


//------------------------------------------------------------------------------
// lock record
//------------------------------------------------------------------------------
function TACRTableLocksManager.LockRecord(
                    TableName:  String;
                    SessionID:  TACRSessionID;
                    RecordID:   TACRRecordID
                   ): Boolean;
var TableItemIndex,SessionItemIndex:      Integer;
    SessionItem:                          PACRSessionLockItem;
    TableLockItem:                        PACRTableLockItem;
begin
  Lock;
  try
    TableItemIndex := FindTableLockItem(TableName);
    if (TableItemIndex < 0) or (TableItemIndex >= FTableLockItemList.Count) then
       raise EACRException.Create(10765,ErrorLInvalidItemNumber,[TableItemIndex,FTableLockItemList.Count]);
    TableLockItem := PACRTableLockItem(FTableLockItemList.Items[TableItemIndex]);
    SessionItemIndex := FindSession(FTableLockItemList.Items[TableItemIndex],SessionID);
    if ((SessionItemIndex < 0) or
        (SessionItemIndex >= TableLockItem^.SessionLocksList.Count)) then
       raise EACRException.Create(10766,ErrorLCannotFindSessionForLockingTable,
        [SessionID,TableName]);
    SessionItem := PACRSessionLockItem(PACRTableLockItem(
      FTableLockItemList.Items[TableItemIndex])^.SessionLocksList.Items[SessionItemIndex]);
    if (SessionItem^.TableLockInfo.IsRecordLocked) then
     raise EACRException.Create(10767,ErrorLCannotLockAlreadyLockedRecord,
      [TableName,RecordID.PageNo,RecordID.PageItemNo,SessionID,
       SessionItem^.TableLockInfo.LockedRecordID.PageNo,
       SessionItem^.TableLockInfo.LockedRecordID.PageItemNo]);
    Result := not IsRecordLocked(TableName,SessionID,RecordID);
    if (Result) then
     begin
      if (not FInMemory) and (SessionItem^.TableLockInfo.NumLocksX = 0) then
       Result := TableLockItem^.TableLocksFile.LockRecordIfPossible(SessionID,RecordID);
      if (Result) then
       begin
        SessionItem^.TableLockInfo.IsRecordLocked := True;
        SessionItem^.TableLockInfo.LockedRecordID := RecordID;
       end;
     end;
  finally
    Unlock;
  end;
end; // LockRecord


//------------------------------------------------------------------------------
// unlock record
//------------------------------------------------------------------------------
function TACRTableLocksManager.UnlockRecord(
                    TableName:      String;
                    SessionID:  TACRSessionID;
                    RecordID:   TACRRecordID
                     ): Boolean;
var TableItemIndex,SessionItemIndex:      Integer;
    SessionItem:                          PACRSessionLockItem;
    TableLockItem:                        PACRTableLockItem;
begin
  Lock;
  try
    TableItemIndex := FindTableLockItem(TableName);
    if (TableItemIndex < 0) or (TableItemIndex >= FTableLockItemList.Count) then
       raise EACRException.Create(10768,ErrorLInvalidItemNumber,[TableItemIndex,FTableLockItemList.Count]);
    TableLockItem := PACRTableLockItem(FTableLockItemList.Items[TableItemIndex]);
    SessionItemIndex := FindSession(FTableLockItemList.Items[TableItemIndex],SessionID);
    if ((SessionItemIndex < 0) or
        (SessionItemIndex >= TableLockItem^.SessionLocksList.Count)) then
       raise EACRException.Create(10769,ErrorLCannotFindSessionForLockingTable,
        [SessionID,TableName]);
    SessionItem := PACRSessionLockItem(PACRTableLockItem(
      FTableLockItemList.Items[TableItemIndex])^.SessionLocksList.Items[SessionItemIndex]);
    Result := True;
    if (SessionItem^.TableLockInfo.IsRecordLocked) then
     begin
      if (not FInMemory) and (SessionItem^.TableLockInfo.NumLocksX = 0) then
        Result := TableLockItem^.TableLocksFile.UnlockRecordIfPossible(SessionID,RecordID);
      if (Result) then
        SessionItem^.TableLockInfo.IsRecordLocked := False;
     end
    else
     Result := False;
  finally
    Unlock;
  end;
end; // IsRecordLocked


//------------------------------------------------------------------------------
// return true if record is locked
//------------------------------------------------------------------------------
function TACRTableLocksManager.IsRecordLocked(
                    TableName:      String;
                    SessionID:  TACRSessionID;
                    RecordID:   TACRRecordID
                     ): Boolean;
var TableItemIndex,i:                     Integer;
    SessionItem:                          PACRSessionLockItem;
    TableLockItem:                        PACRTableLockItem;
begin
  Lock;
  try
    TableItemIndex := FindTableLockItem(TableName);
    if (TableItemIndex < 0) or (TableItemIndex >= FTableLockItemList.Count) then
       raise EACRException.Create(10770,ErrorLInvalidItemNumber,[TableItemIndex,FTableLockItemList.Count]);
    TableLockItem := PACRTableLockItem(FTableLockItemList.Items[TableItemIndex]);
    Result := False;
    for i := 0 to TableLockItem.SessionLocksList.Count - 1 do
     begin
      SessionItem := PACRSessionLockItem(PACRTableLockItem(
        FTableLockItemList.Items[TableItemIndex])^.SessionLocksList.Items[i]);
      if (SessionItem^.TableLockInfo.IsRecordLocked) then
       if (SessionItem^.TableLockInfo.LockedRecordID.PageNo = RecordID.PageNo) then
        if (SessionItem^.TableLockInfo.LockedRecordID.PageItemNo = RecordID.PageItemNo) then
         begin
          Result := True;
          break;
         end;
     end;
    if (not Result) then
     if (not FInMemory) then
      if (SessionItem^.TableLockInfo.NumLocksX = 0)  and (SessionItem^.TableLockInfo.NumLocksX = 0) then
       Result := TableLockItem^.TableLocksFile.CheckIfRecordIsLocked(SessionID,RecordID);
  finally
    Unlock;
  end;
end; // IsRecordLocked


//------------------------------------------------------------------------------
// fill TableLockList with lock info for the tables opened by this session
//------------------------------------------------------------------------------
procedure TACRTableLocksManager.GetTableLocksListForSession(
                                          SessionID:      TACRSessionID;
                                          TableLockList:  TList
                                      );
var i,j:                                  Integer;
    SessionItem:                          PACRSessionLockItem;
    TableLockItem:                        PACRTableLockItem;
    TableLockItemForSession:              PACRTableLockItemForSession;
begin
 for i := 0 to FTableLockItemList.Count -1 do
  begin
   TableLockItem := FTableLockItemList.Items[i];
   j := FindSession(TableLockItem,SessionID);
   if (j >= 0) then
    begin
     SessionItem := PACRSessionLockItem(TableLockItem^.SessionLocksList.Items[j]);
     New(TableLockItemForSession);
     TableLockItemForSession^.TableLockInfo := SessionItem^.TableLockInfo;
     TableLockItemForSession^.TableName := TableLockItem^.TableName;
     TableLockList.Add(TableLockItemForSession);
    end;
  end;
end; // GetTableLocksListForSession


//------------------------------------------------------------------------------
// remove all locks set by the session - for correct disconnecting server session
//------------------------------------------------------------------------------
procedure TACRTableLocksManager.RemoveAllLocksForSession(SessionID: TACRSessionID);
var
    TableLockItem:          PACRTableLockItemForSession;
    s:                      String;
    TableLocksList:         TList;
    i:                      Integer;
begin
  Lock;
  try
    s := '';
    TableLocksList := TList.Create;
    try
      GetTableLocksListForSession(SessionID,TableLocksList);
      try
       for i := 0 to TableLocksList.Count-1 do
        begin
          TableLockItem := TableLocksList.Items[i];
          try
            if (TableLockItem^.TableLockInfo.IsRecordLocked) then
             UnlockRecord(TableLockItem^.TableName,SessionID,TableLockItem^.TableLockInfo.LockedRecordID);
          except
           on E: Exception do
            s := s + e.Message+#13#10
           else
            begin
             if (TableLockItem = nil) then
               s := s + Format(ErrorLUnlockTableItemFailedTableLockItemNil,[i])+#13#10
             else
              s := s + ErrrorLUnlockTableItemFailed+' '+TableLockItem^.TableName+#13#10;
            end;
          end;
          try
            UnlockTableAll(TableLockItem^.TableName,SessionID,nil);
          except
           on E: Exception do
            s := s + e.Message+#13#10
           else
            begin
             if (TableLockItem = nil) then
               s := s + Format(ErrorLUnlockTableItemFailedTableLockItemNil,[i])+#13#10
             else
              s := s + ErrrorLUnlockTableItemFailed+' '+TableLockItem^.TableName+#13#10;
            end;
          end;
        end;
      finally
        for i := 0 to TableLocksList.Count-1 do
         Dispose(TableLocksList.Items[i]);
      end;
    finally
      TableLocksList.Free;
      if (s <> '') then
       raise EACRException.Create(11570,ErrorLRemoveAllLocksForSessionFailed,
            [SessionID,s]);
    end;
  finally
    Unlock;
  end;
end; // RemoveAllLocksForSession


{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
//------------------------------------------------------------------------------
// write current table locks to log
//------------------------------------------------------------------------------
procedure TACRTableLocksManager.WriteCurrentLocksToLog(Caption: String);
var i,j:                                  Integer;
    SessionItem:                          PACRSessionLockItem;
    TableLockItem:                        PACRTableLockItem;
begin
 aaWriteToLog(#13#10+
'---------------------------------------------------------------------------------'+#13#10+
 Caption +
 ', Current table items count = '+IntToStr(FTableLockItemList.Count));
 for i := 0 to FTableLockItemList.Count -1 do
  begin
   TableLockItem := FTableLockItemList.Items[i];
   aaWriteToLog(#13#10+'Table #'+IntToStr(i)+': name = '''+TableLockItem^.TableName+''''+
              ', sessions count = '+IntToStr(TableLockItem^.SessionLocksList.Count));
   for j := 0 to TableLockItem^.SessionLocksList.Count - 1 do
    begin
     SessionItem := TableLockItem^.SessionLocksList.Items[j];
     aaWriteToLog('SessionID '+IntToStr(SessionItem^.SessionID)+':');
     aaWriteToLog('NumLocksX = '+IntToStr(SessionItem^.TableLockInfo.NumLocksX));
     aaWriteToLog('NumLocksIS = '+IntToStr(SessionItem^.TableLockInfo.NumLocksIS));
     aaWriteToLog('NumLocksIRW = '+IntToStr(SessionItem^.TableLockInfo.NumLocksIRW));
     aaWriteToLog('NumLocksRW = '+IntToStr(SessionItem^.TableLockInfo.NumLocksRW));
     aaWriteToLog('NumLocksS = '+IntToStr(SessionItem^.TableLockInfo.NumLocksS));
     aaWriteToLog('Record locked = '+BoolToStr(SessionItem^.TableLockInfo.IsRecordLocked,True)+', PageNo = '+IntToStr(SessionItem^.TableLockInfo.LockedRecordID.PageNo)+
       ', PageItemNo = '+IntToStr(SessionItem^.TableLockInfo.LockedRecordID.PageItemNo))
    end;
  end;
 aaWriteToLog(
'---------------------------------------------------------------------------------'+
 #13#10+#13#10);
end; // WriteCurrentLocksToLog
{$ENDIF}


////////////////////////////////////////////////////////////////////////////////
//
// TACRDiskPageManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// init DBHeader
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.InitDBHeader;
begin
  FDBHeader.Signature := ACRDiskSignature;
  FDBHeader.Version := ACRVersion;
  FDBHeader.TotalPageCount := 0;
  FDBHeader.HeaderSize := Sizeof(FDBHeader) + DefaultDBHeaderReserved;
  FDBHeader.CryptoHeader.CryptoAlgorithm := ACR_Cipher_None;
  FLockedBytes.LockedByteCount := Sizeof(FLockedBytes) + DefaultLockedBytesReserved;
end; // InitDBHeader


//------------------------------------------------------------------------------
// load DBHeader
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.LoadDBHeader;
var Position: Int64;
begin
 Lock;
 try
  if (not FDatabaseFile.IsOpened) then
    raise EACRException.Create(10459,ErrorLDatabaseFileIsNotOpened,[]);
  Position := FOffsetToDBHeader;
  FDatabaseFile.ReadBuffer(FDBHeader,Sizeof(FDBHeader),Position,10464);
  Inc(Position,(FDBHeader.HeaderSize));
  FDatabaseFile.ReadBuffer(FLockedBytes,Sizeof(FLockedBytes),Position,10465);
 finally
   Unlock;
 end;
end; // LoadDBHeader


//------------------------------------------------------------------------------
// save DBHeader
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.SaveDBHeader;
var Position: Int64;
begin
 if (FReadOnly) then
  Exit;
 Lock;
  try
    if (not FDatabaseFile.IsOpened) then
      raise EACRException.Create(10460,ErrorLDatabaseFileIsNotOpened,[]);
    Position := FOffsetToDBHeader;
    FDatabaseFile.WriteBuffer(FDBHeader,Sizeof(FDBHeader),Position,10462);
    Inc(Position,(FDBHeader.HeaderSize));
    FDatabaseFile.WriteBuffer(FLockedBytes,Sizeof(FLockedBytes),Position,10463);
    Inc(Position,FLockedBytes.LockedByteCount + Sizeof(FLockedBytes));
    if (FDatabaseFile.Size < Position) then
      FDatabaseFile.Size := Position;
  finally
    Unlock;
  end;
end; // SaveDBHeader


//------------------------------------------------------------------------------
// return true if database file is opened
//------------------------------------------------------------------------------
function TACRDiskPageManager.GetIsOpened: Boolean;
begin
  Result := FDatabaseFile.IsOpened;
end; // GetIsOpened


//------------------------------------------------------------------------------
// return offset to beginning of the page in the file
//------------------------------------------------------------------------------
function TACRDiskPageManager.GetPageOffset(PageNo: TACRPageNo): Int64;
begin
  if (PageNo = INVALID_PAGE_NO) then
   raise EACRException.Create(10471,ErrorLInvalidPageNo,[PageNo]); 
  Result := Int64(FOffsetToFirstPage) + Int64(PageNo) * Int64(FPageSize);
end; // GetPageOffset


//------------------------------------------------------------------------------
// return page count
//------------------------------------------------------------------------------
function TACRDiskPageManager.GetPageCount: TACRPageNo;
begin
  LoadDBHeader;
  Result := FDBHeader.TotalPageCount;
end; // GetPageCount


//------------------------------------------------------------------------------
// write new state
//------------------------------------------------------------------------------
function TACRDiskPageManager.UpdateState: Cardinal;
begin
 Lock;
 try
  if (not TryUsingTimeOut(LockFreeSpaceManager,FLockParams)) then
    raise EACRException.Create(10638,ErrorLCannotLockState);
  try
   LoadDBHeader;
   Inc(FDBHeader.State);
   Result := FDBHeader.State;
   SaveDBHeader;
  finally
    if (not UnlockFreeSpaceManager) then
      raise EACRException.Create(10639,ErrorLCannotUnlockState);
  end;
 finally
   Unlock;
 end;
end; // UpdateState


//------------------------------------------------------------------------------
// return state
//------------------------------------------------------------------------------
function TACRDiskPageManager.GetState: Cardinal;
begin
 Lock;
 try
  LoadDBHeader;
  Result := FDBHeader.State;
 finally
  Unlock;
 end;
end; // GetState


//------------------------------------------------------------------------------
// read page region
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.ReadPageRegion(var Buffer; PageNo: TACRPageNo; Offset, Count: Integer);
begin
  if (Offset+Count+FPageHeaderSize > FPageSize) then
    raise EACRException.Create(10476,ErrorLInvalidPageOffset,
        [Offset+Count+FPageHeaderSize,FPageSize]);
 if (CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None) then
  begin
   FDatabaseFile.ReadBuffer(PChar(Buffer),
                           Count,
                           GetPageOffset(PageNo)+Int64(Offset)+
                           FPageHeaderSize,10724);
   ACRDecryptBuffer(CryptoInfo,PChar(@Buffer),Count);
  end
 else
  FDatabaseFile.ReadBuffer(PChar(Buffer),
                           Count,
                           GetPageOffset(PageNo)+Int64(Offset)+
                           FPageHeaderSize,10474);
end; // ReadPageRegion


//------------------------------------------------------------------------------
// write page region
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.WritePageRegion(const Buffer; PageNo: TACRPageNo; Offset, Count: Integer);
var NewBuffer: PChar;
begin
  if (Offset+Count+FPageHeaderSize > FPageSize) then
    raise EACRException.Create(10477,ErrorLInvalidPageOffset,
    [Offset+Count+FPageHeaderSize,FPageSize]);
 if (CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None) then
  begin
   NewBuffer := MemoryManager.GetMem(Count);
   try
     Move(PChar(@Buffer)^,NewBuffer^,Count);
     ACREncryptBuffer(CryptoInfo,NewBuffer,Count);
     FDatabaseFile.WriteBuffer(NewBuffer^,
                             Count,
                             GetPageOffset(PageNo)+Int64(Offset)+
                             FPageHeaderSize,10725);
   finally
     MemoryManager.FreeAndNilMem(NewBuffer);
   end;
  end
 else
  FDatabaseFile.WriteBuffer(PChar(Buffer),
                           Count,
                           GetPageOffset(PageNo)+Int64(Offset)+
                           FPageHeaderSize,10475);
end; // WritePageRegion


//------------------------------------------------------------------------------
// direct add page
//------------------------------------------------------------------------------
function TACRDiskPageManager.DirectAddPage: TACRPageNo;
begin
 Result := FDatabaseFreeSpaceManager.GetPage;
 if (Result = INVALID_PAGE_NO) or (Result >= FDBHeader.TotalPageCount) then
   raise EACRException.Create(10707,ErrorLInvalidPageNo,[Result]);
end; // DirectAddPage


//------------------------------------------------------------------------------
// internal add page
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.InternalAddPage(aPage: TACRPage);
begin
{$IFDEF DEBUG_TRACE_PAGE_MANAGER}
aaWriteToLog('0. add page...');
{$ENDIF}
  aPage.PageNo := DirectAddPage;
  aPage.AllocPageBuffer;
  if (CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None) then
   ACRGenerateRandomBuffer(aPage.PageBuffer,FPageSize);
  aPage.PageHeader.Signature := ACRDiskPageSignature;
  aPage.PageHeader.PageType := ACRPageTypeIDEmpty;
  aPage.PageHeader.State := 0;
  aPage.PageHeader.NextPageNo := INVALID_PAGE_NO;
  aPage.PageHeader.ObjectID := ACRObjectIDUnknown;
{$IFDEF DEBUG_TRACE_PAGE_MANAGER}
aaWriteToLog('1. add page '+IntToStr(aPage.PageNo));
{$ENDIF}
end; // InternalAddPage


//------------------------------------------------------------------------------
// internal remove page
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.InternalRemovePage(PageNo: TACRPageNo);
begin
{$IFDEF DEBUG_TRACE_PAGE_MANAGER}
aaWriteToLog('0. remove page '+IntToStr(PageNo));
{$ENDIF}
 FDatabaseFreeSpaceManager.FreePage(PageNo);
{$IFDEF DEBUG_TRACE_PAGE_MANAGER}
aaWriteToLog('1. remove page '+IntToStr(PageNo));
{$ENDIF}
end; // InternalRemovePage

//------------------------------------------------------------------------------
// read page data
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.InternalReadPage(aPage: TACRPage);
var Buffer: PChar;
begin
 if (aPage.PageBuffer = nil) then
   aPage.AllocPageBuffer;
 if (CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None)  then
  begin
   Buffer := MemoryManager.GetMem(PageSize);
   try
     FDatabaseFile.ReadBuffer(Buffer^,PageSize,GetPageOffset(aPage.PageNo),10716);
     ACRDecryptBuffer(CryptoInfo,Buffer,PageSize);
     Move(Buffer^,aPage.PageBuffer^,PageSize);
   finally
     MemoryManager.FreeAndNilMem(Buffer);
   end;
  end
 else
  begin
   FDatabaseFile.ReadBuffer(aPage.PageBuffer^,FPageSize,GetPageOffset(aPage.PageNo),10466);
  end;
{$IFDEF DEBUG_TRACE_PAGE_MANAGER}
aaWriteToLog('read page '+IntToStr(aPage.PageNo)+' type = '+IntToStr(aPage.PageHeader.PageType));
{$ENDIF}
end; // ReadPageData


//------------------------------------------------------------------------------
// write page data
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.InternalWritePage(aPage: TACRPage);
var Buffer: PChar;
begin
 if (CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None)  then
  begin
   Buffer := MemoryManager.GetMem(PageSize);
   try
     Move(aPage.PageBuffer^,Buffer^,PageSize);
     ACREncryptBuffer(CryptoInfo,Buffer,PageSize);
     FDatabaseFile.WriteBuffer(Buffer^,PageSize,GetPageOffset(aPage.PageNo),10715);
   finally
     MemoryManager.FreeAndNilMem(Buffer);
   end;
  end
 else
  begin
   FDatabaseFile.WriteBuffer(aPage.PageBuffer^,FPageSize,GetPageOffset(aPage.PageNo),10469);
  end;
{$IFDEF DEBUG_TRACE_PAGE_MANAGER}
aaWriteToLog('write page '+IntToStr(aPage.PageNo)+' type = '+IntToStr(aPage.PageHeader.PageType));
{$ENDIF}
end; // WritePageData


//------------------------------------------------------------------------------
// lock Free Space Manager byte
//------------------------------------------------------------------------------
function TACRDiskPageManager.LockFreeSpaceManager: Boolean;
begin
  Result := FDatabaseFile.LockByte(FOffsetToLockedBytes + OffsetToFreeSpaceManagerLockByte);
end; // LockFreeSpaceManager


//------------------------------------------------------------------------------
// unlock Free Space Manager byte
//------------------------------------------------------------------------------
function TACRDiskPageManager.UnlockFreeSpaceManager: Boolean;
begin
  Result := FDatabaseFile.UnlockByte(FOffsetToLockedBytes + OffsetToFreeSpaceManagerLockByte);
end; // UnockFreeSpaceManager


//------------------------------------------------------------------------------
// lock Tables byte
//------------------------------------------------------------------------------
function TACRDiskPageManager.InternalLockTableS: Boolean;
begin
  Result := FDatabaseFile.LockByte(FOffsetToLockedBytes + OffsetToTablesLockByte);
end; // InternalLockTableS


//------------------------------------------------------------------------------
// unlock Tables byte
//------------------------------------------------------------------------------
function TACRDiskPageManager.InternalUnlockTableS: Boolean;
begin
  Result := FDatabaseFile.UnlockByte(FOffsetToLockedBytes + OffsetToTablesLockByte);
end; // UnockTables


//------------------------------------------------------------------------------
// Lock Byte (return TRUE if success)
//------------------------------------------------------------------------------
function TACRDiskPageManager.LockPageByte(PageNo: TACRPageNo; Offset: Word): Boolean;
begin
  Result := FDatabaseFile.LockByte(GetPageOffset(PageNo) + Int64(Offset));
end; // LockPageByte


//------------------------------------------------------------------------------
// Unlock Byte
//------------------------------------------------------------------------------
function TACRDiskPageManager.UnlockPageByte(PageNo: TACRPageNo; Offset: Word): Boolean;
begin
  Result := FDatabaseFile.UnlockByte(GetPageOffset(PageNo) + Int64(Offset));
end; // UnlockPageByte


//------------------------------------------------------------------------------
// return True if byte is locked
//------------------------------------------------------------------------------
function TACRDiskPageManager.IsPageByteLocked(PageNo: TACRPageNo; Offset: Word): Boolean;
begin
  Result := FDatabaseFile.IsByteLocked(GetPageOffset(PageNo) + Int64(Offset));
end; // IsPageByteLocked


//------------------------------------------------------------------------------
// return True if any byte of region is locked
//------------------------------------------------------------------------------
function TACRDiskPageManager.IsPageRegionLocked(PageNo: TACRPageNo; Offset: Word; Count: Word): Boolean;
begin
  Result := FDatabaseFile.IsRegionLocked(
              GetPageOffset(PageNo) + Int64(Offset),Count);
end; // IsPageRegionLocked


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRDiskPageManager.Create(aLockParams: TACRLockParams);
begin
 inherited Create;
 FDatabaseFreeSpaceManager := nil;
 FDatabaseFile := TACRDatabaseFile.Create;
 LockParams := aLockParams;
end; // Create


//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------
destructor TACRDiskPageManager.Destroy;
begin
  CloseDatabase;
  FreeAndNil(FDatabaseFile);
  if (Length(FCryptoInfo.Password) > 0) then
   FillChar(FCryptoInfo.Password[1],Length(FCryptoInfo.Password),$00);
  FCryptoInfo.Password := '';
  FillChar(FCryptoInfo,SizeOf(FCryptoInfo),$00);
  inherited;
end; // Destroy


//------------------------------------------------------------------------------
// CreateAndOpenDatabase
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.CreateAndOpenDatabase(
                              DatabaseFileName: String;
                              PageSize:         Word;
                              ExtentPageCount:  Word
                                            );
begin
  if (FDatabaseFile.IsOpened) then
   raise EACRException.Create(10461,ErrorLDatabaseFileIsInUse,[FDatabaseFile.FileName]);
  FDatabaseFile.CreateAndOpenFile(DatabaseFileName);
  InitDBHeader;
  FDBHeader.PageSize := PageSize;
  FDBHeader.ExtentPageCount := ExtentPageCount;
  FDBHeader.CryptoHeader := ACRCreateCryptoHeader(FCryptoInfo);
  FPageSize := FDBHeader.PageSize;
  FPageDataSize := FDBHeader.PageSize - SizeOf(TACRDiskPageHeader);
  FPageHeaderSize := Sizeof(TACRDiskPageHeader);
  SaveDBHeader;
  FOffsetToFirstPage := FDBHeader.HeaderSize + Sizeof(FLockedBytes) +
                        FLockedBytes.LockedByteCount;
  FDatabaseFreeSpaceManager := TACRDatabaseFreeSpaceManager.Create(Self,LockParams);
  FDatabaseFreeSpaceManager.CreateFreeSpaceManager;
  FReadOnly := False;
  FExclusive := True;
end; // CreateAndOpenDatabase


//------------------------------------------------------------------------------
// Delete database
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.DeleteDatabase(DatabaseFileName: String);
begin
 ;
end; // DeleteDatabase


//------------------------------------------------------------------------------
// Rename database
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.RenameDatabase(
                                             DatabaseFileName,
                                             NewDatabaseName: String
                                            );
begin
;
end; // RenameDatabase


//------------------------------------------------------------------------------
// return true if file exists
//------------------------------------------------------------------------------
function TACRDiskPageManager.GetExists(DatabaseFileName: String): Boolean;
begin
 Result := SysUtils.FileExists(DatabaseFileName);
end; // GetExists


//------------------------------------------------------------------------------
// Open database
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.OpenDatabase(
                DatabaseFileName:     String;
                var aReadOnly:        Boolean;
                aExclusive:           Boolean = False;
                DoNotCheckPassword:   Boolean = False
                                          );
var AccessMode: TACRAccessMode;
    ShareMode:  TACRShareMode;
begin
 FExclusive := aExclusive;
 FReadOnly := aReadOnly;
 if (FDatabaseFile.IsOpened) then
  raise EACRException.Create(10462,ErrorLDatabaseFileIsInUse,[FDatabaseFile.FileName]);
 if (FReadOnly) then
  AccessMode := amReadOnly
 else
  AccessMode := amReadWrite;
 if (FExclusive) then
  ShareMode := smExclusive
 else
  ShareMode := smShared;
 FDatabaseFile.OpenFile(DatabaseFileName,AccessMode,ShareMode);
 aReadOnly := (AccessMode = amReadOnly);
 FReadOnly := aReadOnly;
 FOffsetToDBHeader := GetOffsetToDBHeader;
 if (FOffsetToDBHeader < 0) then
  raise EACRException.Create(11289,ErrorLInvalidDatabaseFile,[DatabaseFileName]);
 LoadDBHeader;
 FCryptoInfo.CryptoAlgorithm := FDBHeader.CryptoHeader.CryptoAlgorithm;
 FCryptoInfo.CryptoMode := FDBHeader.CryptoHeader.CryptoMode;
 if (FDBHeader.Signature <> ACRDiskSignature) then
  raise EACRException.Create(10470,ErrorLInvalidSignature,
        [FDBHeader.Signature,ACRDiskSignature]);
 if ((DBHeader.Version < ACRMinVersion) or (DBHeader.Version > ACRMaxVersion)) then
  raise EACRException.Create(10637,ErrorLInvalidVersion,
    [DBHeader.Version,ACRVersion]);
 if (not DoNotCheckPassword) then
  if (not ACRIsKeyValid(FDBHeader.CryptoHeader,FCryptoInfo)) then
   raise EACRException.Create(10712,ErrorLInvalidCryptoKeyInfo,
    [FDatabaseFile.FFileName]);
 FDatabaseFreeSpaceManager := TACRDatabaseFreeSpaceManager.Create(Self,LockParams);
 FOffsetToFirstPage := FOffsetToDBHeader + FDBHeader.HeaderSize + Sizeof(FLockedBytes) +
                        FLockedBytes.LockedByteCount ;
 FPageSize := FDBHeader.PageSize;
 FPageDataSize := FDBHeader.PageSize - SizeOf(TACRDiskPageHeader);
 FPageHeaderSize := Sizeof(TACRDiskPageHeader);
 FOffsetToLockedBytes := FDBHeader.HeaderSize + Sizeof(FLockedBytes);
end; // OpenDatabase


//------------------------------------------------------------------------------
// Close database
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.CloseDatabase;
begin
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskPageManager.CloseDatabase starting ...'+
   ', FDatabaseFile = '+IntToStr(Integer(FDatabaseFile)));
 if (FDatabaseFile <> nil) then
 aaWriteToLog('TACRDiskPageManager.CloseDatabase starting ...'+
   ', FDatabaseFile.FIsOpened = '+BoolToStr(FDatabaseFile.FIsOpened,True));
 {$ENDIF}
 if (FDatabaseFile.FIsOpened) then
  begin
// commented in v.4.10
//   if (not FReadOnly) then
//    SaveDBHeader;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskPageManager.CloseDatabase flushing buffes...');
 {$ENDIF}
   FlushFileBuffers;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskPageManager.CloseDatabase closing database file ...');
 {$ENDIF}
   FDatabaseFile.CloseFile;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskPageManager.CloseDatabase closing database file ... OK');
 {$ENDIF}
  end;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskPageManager.CloseDatabase destroy free space manager ...'+
   ', FDatabaseFreeSpaceManager = '+IntToStr(Integer(FDatabaseFreeSpaceManager)));
 {$ENDIF}
 if (FDatabaseFreeSpaceManager <> nil) then
  FreeAndNil(FDatabaseFreeSpaceManager);
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDiskPageManager.CloseDatabase destroy free space manager ... OK. Finished'+
   ', FDatabaseFreeSpaceManager = '+IntToStr(Integer(FDatabaseFreeSpaceManager)));
 {$ENDIF}
end; // CloseDatabase


//------------------------------------------------------------------------------
// flush file buffers
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.FlushFileBuffers;
begin
 if (not FReadOnly) then
  FDatabaseFile.FlushFileBuffers;
end; // FlushFileBuffers


//------------------------------------------------------------------------------
// extend file by number of pages specified by PageCount
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.ExtendFile(PageCount: TACRPageNo);
var
    NewSize: Int64;
begin
  if (PageCount > 0) then
   begin
    Inc(FDBHeader.TotalPageCount,PageCount);
    NewSize := FOffsetToFirstPage + Int64(FDBHeader.TotalPageCount) * Int64(FPageSize);
    if (FDatabaseFile.Size < NewSize) then
      FDatabaseFile.Size := NewSize;
//    for i := FDBHeader.TotalPageCount - PageCount to FDBHeader.TotalPageCount - 1 do
//     WriteEmptyPage(i);
    SaveDBHeader;
   end;
end; // ExtendFile


//------------------------------------------------------------------------------
// sets database file size to correspond remaining number of pages
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.SetPageCount(LastPageNo: TACRPageNo);
var
    NewSize: Int64;
begin
  FDBHeader.TotalPageCount := LastPageNo+1;
  NewSize := FOffsetToFirstPage + Int64(FDBHeader.TotalPageCount) * Int64(FPageSize);
  if (FDatabaseFile.Size <> NewSize) then
    FDatabaseFile.Size := NewSize;
  SaveDBHeader;
end; // SetPageCount


//------------------------------------------------------------------------------
// return true of db header is valid
//------------------------------------------------------------------------------
function TACRDiskPageManager.IsDBHeaderValid(Offset: Int64): Boolean;
var TempHeader: TACRDBHeader;
    oldPos:     Int64;
begin
  Result := False;
  oldPos := FDatabaseFile.Position;
  try
    if ((Offset + Int64(SizeOf(TempHeader))) < FDatabaseFile.Size) then
     begin
      FDatabaseFile.ReadBuffer(TempHeader,SizeOf(TempHeader),Offset,11295);
      if (TempHeader.Signature = ACRDiskSignature) then
       if ((TempHeader.Version >= ACRMinVersion) and (TempHeader.Version < ACRMaxVersion)) then
        Result := True;
     end;
  finally
    FDatabaseFile.Position := oldPos;
  end;
end; // IsDBHeaderValid


//------------------------------------------------------------------------------
// return offset from beginning of the file to DBHeader's signature, or -1 if no signature inthe file
//------------------------------------------------------------------------------
function TACRDiskPageManager.GetOffsetToDBHeader: Int64;
const BufSize = $FFFF;
var
    size,offset,pos,
    oldPos,i,j,k:	  Int64;
    buf:		        PChar;
    sgn:      		  TACRSignature;

begin
 Result := -1;
 sgn := ACRDiskSignature;
 if (IsDBHeaderValid(0)) then
  begin
   Result := 0;
   Exit;
  end;
 buf := MemoryManager.AllocMem(BufSize);
 oldPos := FDatabaseFile.Position;
 try
   offset := 0;
   FDatabaseFile.Position := offset;
   // find local file header for first file in archive
   while FDatabaseFile.Position < FDatabaseFile.Size do
    begin
     pos := FDatabaseFile.Position;
     if (FDatabaseFile.Size - FDatabaseFile.Position > BufSize) then
      size := BufSize
     else
      size := FDatabaseFile.Size - FDatabaseFile.Position;
     FDatabaseFile.ReadBuffer(Buf^,size,FDatabaseFile.Position,11293);
     // find local file header signature
     i := 0;
     k := -1;
     while (i < size) do
      begin
       k := -1;
       if (PChar(Buf+i)^ = sgn[0]) then
        begin
         j := 1;
         while j <= 3 do
          begin
           if ((i + j) >= size) then break;
           if (pChar(Buf + i + j)^ <> sgn[j]) then break;
           inc(j);
           // signature found
           if (j > 3) then
             k := i;
          end; // chaeck signature
         if (k >= 0) then
          if (IsDBHeaderValid(pos+k)) then
          // local file header for first file in archive found!
           break
          else
           k := -1;
        end; // signature found
       inc(i);
      end; // end of searching local header
     // stub size = difference between supposed offset for header
     // and real one
     if (k >= 0) then
      begin
       Result := k+pos-offset;
       break;
      end;
    end;
 finally
   MemoryManager.FreeAndNilMem(buf);
   FDatabaseFile.Position := oldPos;
 end;
end; // GetOffsetToDBHeader


//------------------------------------------------------------------------------
// return true if this file is an Accuracer Database File
//------------------------------------------------------------------------------
function TACRDiskPageManager.IsAccuracerDatabaseFile(DatabaseFileName: String): Boolean;
var
    bOpened:      Boolean;
    AccessMode:   TACRAccessMode;
    ShareMode:    TACRShareMode;
begin
 bOpened := FDatabaseFile.IsOpened;
 if (bOpened and (FDatabaseFile.FFileName <> DatabaseFileName)) then
  raise EACRException.Create(11296,ErrorLInvalidDatabaseFileName,[DatabaseFileName,FDatabaseFile.FFileName]);
 AccessMode := amReadOnly;
 ShareMode := smShared;
 if (not bOpened) then
  FDatabaseFile.OpenFile(DatabaseFileName,AccessMode,ShareMode);
 Result := (GetOffsetToDBHeader >= 0);
 if (not bOpened) then
  FDatabaseFile.CloseFile;
end; // IsAccuracerDatabaseFile


//------------------------------------------------------------------------------
// remove database from exe file
//------------------------------------------------------------------------------
procedure TACRDiskPageManager.RemoveDatabaseFromExe(DatabaseFileName: String);
var
    bOpened:    Boolean;
    AccessMode: TACRAccessMode;
    ShareMode:  TACRShareMode;

begin
 bOpened := FDatabaseFile.IsOpened;
 if (not bOpened) then
  begin
   if (not GetExists(DatabaseFileName)) then
    Exit;
   AccessMode := amReadWrite;
   ShareMode := smExclusive;
   FDatabaseFile.OpenFile(DatabaseFileName,AccessMode,ShareMode);
   FOffsetToDBHeader := GetOffsetToDBHeader;
  end;
 if (FOffsetToDBHeader >= 0) then
  FDatabaseFile.Size := FOffsetToDBHeader;
 if (not bOpened) then
  FDatabaseFile.CloseFile;
end; // IsAccuracerDatabaseFile


////////////////////////////////////////////////////////////////////////////////
//
// TACRInternalLockedDBFile
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// raise if file is Closed
//------------------------------------------------------------------------------
procedure TACRInternalLockedDBFile.CheckFileOpened(OperationName: String);
begin
  if (FStartPageNo = INVALID_PAGE_NO) then
    raise EACRException.Create(10642, ErrorGIFClosed, [OperationName]);
end;//CheckFileOpened


//------------------------------------------------------------------------------
// return page offset
//------------------------------------------------------------------------------
function TACRInternalLockedDBFile.GetPageOffset(Position: Int64): Word;
begin
  Result := (Position + SizeOf(TACRInternalFileHeader)) mod LPageManager.PageDataSize;
end; // GetPageOffset


//------------------------------------------------------------------------------
// return page no
//------------------------------------------------------------------------------
function TACRInternalLockedDBFile.GetPageNo(Position: Int64): Integer;
begin
  Result := (Position + SizeOf(TACRInternalFileHeader)) div LPageManager.PageDataSize;
end; // GetPageNo


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRInternalLockedDBFile.Create(PageManager: TACRDiskPageManager; PageTypeID: TACRPageTypeID);
begin
  LPageManager := PageManager;
  FPageTypeID := PageTypeID;
  FStartPageNo := INVALID_PAGE_NO;
  FPages := TACRIntegerArray.Create;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRInternalLockedDBFile.Destroy;
begin
  FPages.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// Create file
//------------------------------------------------------------------------------
procedure TACRInternalLockedDBFile.CreateFile(FileSize: Integer; ObjectID: TACRObjectID);
var
  i,PageCount:     Integer;
  NextPageNo:      TACRPageNo;
  PageNo:          TACRPageNo;
  Buffer:          PChar;
begin
 PageCount := (FileSize + SizeOf(TACRInternalFileHeader)) div LPageManager.PageDataSize +
              Integer(((FileSize + SizeOf(TACRInternalFileHeader)) mod LPageManager.PageDataSize) > 0);
 FPages.SetSize(PageCount);
 PageNo := LPageManager.DirectAddPage;
 Buffer := MemoryManager.GetMem(LPageManager.PageSize);
 try
  for i := 0 to PageCount-1 do
   begin
    if (i > 0) then
     PageNo := NextPageNo;
    if (i < (PageCount-1)) then
     NextPageNo := LPageManager.DirectAddPage
    else
     NextPageNo := INVALID_PAGE_NO;
    if (LPageManager.FDBHeader.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None) then
     ACRGenerateRandomBuffer(Buffer,LPageManager.PageSize);
    FillChar(Buffer^,LPageManager.PageHeaderSize,$00);
    PACRDiskPageHeader(Buffer)^.Signature := ACRDiskPageSignature;
    PACRDiskPageHeader(Buffer)^.PageType := FPageTypeID;
    PACRDiskPageHeader(Buffer)^.ObjectID := ObjectID;
    PACRDiskPageHeader(Buffer)^.NextPageNo := NextPageNo;
    LPageManager.WritePageRegion(Buffer^,PageNo,-LPageManager.PageHeaderSize,LPageManager.PageHeaderSize);
    if (i = 0) then
     begin
      FStartPageNo := PageNo;
      PACRInternalFileHeader(Buffer+LPageManager.PageHeaderSize).Signature := ACRInternalDBFileSignature;
      PACRInternalFileHeader(Buffer+LPageManager.PageHeaderSize).FileSize := FileSize;
      PACRInternalFileHeader(Buffer+LPageManager.PageHeaderSize).CompressionMode := 0;
      LPageManager.WritePageRegion(PChar(Buffer+LPageManager.PageHeaderSize)^,PageNo,0,LPageManager.PageDataSize);
      LPageManager.WritePageRegion(PChar(Buffer+LPageManager.PageHeaderSize)^,PageNo,0,SizeOf(TACRInternalFileHeader));
     end
    else
     begin
      LPageManager.WritePageRegion(PChar(Buffer+LPageManager.PageHeaderSize)^,PageNo,0,LPageManager.PageDataSize);
     end;
  end;
 finally
  MemoryManager.FreeAndNilMem(Buffer);
 end;
 FSize := FileSize;
end; // CreateFile


//------------------------------------------------------------------------------
// Delete file and free all file pages
//------------------------------------------------------------------------------
procedure TACRInternalLockedDBFile.DeleteFile(StartPageNo: TACRPageNo);
var i: Integer;
begin
  CheckFileOpened('DeleteFile');
  for i := 0 to FPages.ItemCount - 1 do
   LPageManager.RemovePage(SYSTEM_SESSION_ID,FPages.Items[i]);
end; // DeleteFile


//------------------------------------------------------------------------------
// Open file (reading file header)
//------------------------------------------------------------------------------
procedure TACRInternalLockedDBFile.OpenFile(StartPageNo: TACRPageNo);
var
  i,PageCount:  Integer;
  PageNo:       TACRPageNo;
  Buf:          PChar;
begin
 FStartPageNo := StartPageNo;
 if (FStartPageNo <= ACRFirstPageNoSystemDirectory) then
  raise EACRException.Create(10697,ErrorLCannotOpenFileInvalidStartPage,
    [FPageTypeID,FStartPageNo]);
 FSize := 0;
 i := 0;
 PageNo := FStartPageNo;
 Buf := MemoryManager.GetMem(LPageManager.PageSize);
 try
   while (PageNo <> INVALID_PAGE_NO) do
    begin
     // avoid reading header with locked byte #0
     while (LPageManager.IsPageByteLocked(PageNo,0)) do
      Sleep(LPageManager.LockParams.Delay);
     LPageManager.ReadPageRegion(PChar(Buf)^,PageNo,-LPageManager.PageHeaderSize,LPageManager.PageHeaderSize);
     if (PageNo = FStartPageNo) then
      LPageManager.ReadPageRegion(PChar(Buf+LPageManager.PageHeaderSize)^,PageNo,
        0,SizeOf(TACRInternalFileHeader));

//     LPageManager.FDatabaseFile.ReadBuffer(Buf^,Count,StartPos,10672);
     if (PACRDiskPageHeader(Buf)^.PageType <> FPageTypeID) then
      raise EACRException.Create(10643,ErrorLInvalidPageType1,
       [PACRDiskPageHeader(Buf)^.PageType,FPageTypeID,PageNo]);
     if (PageNo = FStartPageNo) then
      begin
        if (PACRInternalFileHeader(Buf + LPageManager.PageHeaderSize).Signature <>
            ACRInternalDBFileSignature) then
         raise EACRException.Create(10644,ErrorLInvalidSignature,
         [PACRInternalFileHeader(Buf + LPageManager.PageHeaderSize).Signature,
          ACRInternalDBFileSignature]);
        FSize := PACRInternalFileHeader(Buf + LPageManager.PageHeaderSize)^.FileSize;
        if (FSize <= 0) then
         raise EACRException.Create(10645,ErrorLInvalidFileSize,[FSize]);
        PageCount := (FSize + SizeOf(TACRInternalFileHeader)) div LPageManager.PageDataSize +
                     Integer(((FSize + SizeOf(TACRInternalFileHeader)) mod LPageManager.PageDataSize) > 0);
        FPages.SetSize(PageCount);
      end;
     FPages.Items[i] := PageNo;
     PageNo := PACRDiskPageHeader(Buf)^.NextPageNo;
     Inc(i);
    end; // reading page headers
 finally
   MemoryManager.FreeAndNilMem(Buf);
 end;
end; // OpenFile



//------------------------------------------------------------------------------
// direct (without sessions, cache and transactions) Read buffer from DB file
//------------------------------------------------------------------------------
procedure TACRInternalLockedDBFile.DirectReadBuffer(
            var Buffer;
            const Count:    Integer;
            const Position: Integer
                                                   );
var
  i:            Integer;
  FirstPageNo:  Integer;
  LastPageNo:   Integer;
  PageNo:       TACRPageNo;
  DBFileSize,
  Offset:       Word;
  DBFilePos:    Int64;
  BytesRead:    Integer;
begin
  CheckFileOpened('DirectReadBuffer');
  FirstPageNo := GetPageNo(Position);
  if (FirstPageNo >= FPages.ItemCount) then
   raise EACRException.Create(10647,ErrorLCannotReadFromFile,
    [Position,Count,FPages.ItemCount,FSize]);
  LastPageNo := GetPageNo(Count + Position);
  if (LastPageNo >= FPages.ItemCount) then
   raise EACRException.Create(10648,ErrorLCannotReadFromFile,
    [Position,Count,FPages.ItemCount,FSize]);
  if (Position + Count > FSize) then
   raise EACRException.Create(10649,ErrorLCannotReadFromFile,
    [Position,Count,FPages.ItemCount,FSize]);
  BytesRead := 0;

  for i := FirstPageNo to LastPageNo do
   begin
    PageNo := FPages.Items[i];
    DBFilePos := LPageManager.GetPageOffset(PageNo) + LPageManager.PageHeaderSize;
    DBFileSize := LPageManager.PageDataSize;
    if ((i = FirstPageNo) or (i = LastPageNo)) then
     begin
      if (i = FirstPageNo) then
       begin
        Offset := GetPageOffset(Position);
        Inc(DBFilePos,Offset);
        Dec(DBFileSize,Offset);
       end;
      if (i = LastPageNo) then
       begin
        Offset := GetPageOffset(Position+Count);
        Dec(DBFileSize,(LPageManager.PageDataSize - Offset));
       end;
     end;
    LPageManager.DatabaseFile.ReadBuffer(
      PChar(PChar(@Buffer) + BytesRead)^,DBFileSize,DBFilePos,10650);
    Inc(BytesRead,DBFileSize);
   end;
 if (LPageManager.DBHeader.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None) then
  ACRDecryptBuffer(LPageManager.CryptoInfo,PChar(@Buffer),Count);
end;//DirectReadBuffer


//------------------------------------------------------------------------------
// direct (without sessions, cache and transactions) Write buffer to DB file
//------------------------------------------------------------------------------
procedure TACRInternalLockedDBFile.DirectWriteBuffer(
            const Buffer;
            const Count:    Integer;
            const Position: Integer
                                                   );
var
  i:            Integer;
  FirstPageNo:  Integer;
  LastPageNo:   Integer;
  PageNo:       TACRPageNo;
  DBFileSize,
  Offset:       Word;
  DBFilePos:    Int64;
  BytesRead:    Integer;
  NewBuffer:    PChar;
begin
  CheckFileOpened('DirectWriteBuffer');
  if (LPageManager.DBHeader.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None) then
   begin
    NewBuffer := MemoryManager.GetMem(Count);
    Move(Buffer,NewBuffer^,Count);
    ACREncryptBuffer(LPageManager.CryptoInfo,NewBuffer,Count);
   end
  else
   NewBuffer := @Buffer;

  FirstPageNo := GetPageNo(Position);
  if (FirstPageNo >= FPages.ItemCount) then
   raise EACRException.Create(10651,ErrorLCannotReadFromFile,
    [Position,Count,FPages.ItemCount,FSize]);
  LastPageNo := GetPageNo(Count + Position);
  if (LastPageNo >= FPages.ItemCount) then
   raise EACRException.Create(10652,ErrorLCannotReadFromFile,
    [Position,Count,FPages.ItemCount,FSize]);
  if (Position + Count > FSize) then
   raise EACRException.Create(10653,ErrorLCannotReadFromFile,
    [Position,Count,FPages.ItemCount,FSize]);
  BytesRead := 0;

  for i := FirstPageNo to LastPageNo do
   begin
    PageNo := FPages.Items[i];
    DBFilePos := LPageManager.GetPageOffset(PageNo) + LPageManager.PageHeaderSize;
    DBFileSize := LPageManager.PageDataSize;
    if ((i = FirstPageNo) or (i = LastPageNo)) then
     begin
      if (i = FirstPageNo) then
       begin
        Offset := GetPageOffset(Position);
        Inc(DBFilePos,Offset);
        Dec(DBFileSize,Offset);
       end;
      if (i = LastPageNo) then
       begin
        Offset := GetPageOffset(Position+Count);
        Dec(DBFileSize,(LPageManager.PageDataSize - Offset));
       end;
     end;
    LPageManager.DatabaseFile.WriteBuffer(
      PChar(PChar(NewBuffer) + BytesRead)^,DBFileSize,DBFilePos,10654);
    Inc(BytesRead,DBFileSize);
   end;
  if (LPageManager.DBHeader.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None) then
   MemoryManager.FreeAndNilMem(NewBuffer);
end;//DirectWriteBuffer


//------------------------------------------------------------------------------
// LockFile (Lock byte in 1-st page header
//------------------------------------------------------------------------------
function TACRInternalLockedDBFile.LockFile: Boolean;
begin
  CheckFileOpened('LockFile');
  Result := LPageManager.LockPageByte(FStartPageNo, 0);
end;//LockFile


//------------------------------------------------------------------------------
// UnockFile
//------------------------------------------------------------------------------
function TACRInternalLockedDBFile.UnlockFile: Boolean;
begin
  CheckFileOpened('UnlockFile');
  Result := LPageManager.UnlockPageByte(FStartPageNo, 0);
end;//UnlockFile


//------------------------------------------------------------------------------
// IsFileLocked
//------------------------------------------------------------------------------
function TACRInternalLockedDBFile.IsFileLocked: Boolean;
begin
  CheckFileOpened('IsFileLocked');
  Result := LPageManager.IsPageByteLocked(FStartPageNo, 0);
end;//IsFileLocked


//------------------------------------------------------------------------------
// LockByte
//------------------------------------------------------------------------------
function TACRInternalLockedDBFile.LockByte(ByteNo: Integer): Boolean;
var
  PageOffset: Word;
  PageNo:     Integer;
begin
  CheckFileOpened('LockByte');
  PageNo := GetPageNo(ByteNo);
  if (PageNo >= FPages.ItemCount) or (ByteNo >= FSize) then
   raise EACRException.Create(10655,ErrorLCannotAccessByteInFile,
    [ByteNo,1,FPages.ItemCount,FSize]);
  PageOffset := GetPageOffset(ByteNo) + LPageManager.PageHeaderSize;
  Result := LPageManager.LockPageByte(FPages.Items[PageNo], PageOffset);
end;//LockByte


//------------------------------------------------------------------------------
// UnlockByte
//------------------------------------------------------------------------------
function TACRInternalLockedDBFile.UnlockByte(ByteNo: Integer): Boolean;
var
  PageOffset: Word;
  PageNo:     Integer;
begin
  CheckFileOpened('UnlockByte');
  PageNo := GetPageNo(ByteNo);
  if (PageNo >= FPages.ItemCount) then
   raise EACRException.Create(10656,ErrorLCannotAccessByteInFile,
    [ByteNo,1,FPages.ItemCount,FSize]);
  PageOffset := GetPageOffset(ByteNo) + LPageManager.PageHeaderSize;
  Result := LPageManager.UnlockPageByte(FPages.Items[PageNo], PageOffset);
end;//UnlockByte


//------------------------------------------------------------------------------
// IsByteLocked
//------------------------------------------------------------------------------
function TACRInternalLockedDBFile.IsByteLocked(ByteNo: Integer): Boolean;
var
  PageOffset: Word;
  PageNo:     Integer;
begin
  CheckFileOpened('IsByteLocked');
  PageNo := GetPageNo(ByteNo);
  if (PageNo >= FPages.ItemCount) then
   raise EACRException.Create(10657,ErrorLCannotAccessByteInFile,
    [ByteNo,1,FPages.ItemCount,FSize]);
  PageOffset := GetPageOffset(ByteNo) + LPageManager.PageHeaderSize;
  Result := LPageManager.IsPageByteLocked(FPages.Items[PageNo], PageOffset);
end;//IsByteLocked


//------------------------------------------------------------------------------
// IsRegionLocked
//------------------------------------------------------------------------------
function TACRInternalLockedDBFile.IsRegionLocked(const Position: Integer; const Count: Integer): Boolean;
var
  i:            Integer;
  FirstPageNo:  Integer;
  LastPageNo:   Integer;
  PageNo:       TACRPageNo;
  DBFileSize,
  Offset:       Word;
  DBFilePos:    Int64;
begin
  CheckFileOpened('IsRegionLocked');
  if (Count <= 0) then
   raise EACRException.Create(10729,ErrorLInvalidItemCount,[Count]);
  Result := False;
  FirstPageNo := GetPageNo(Position);
  if (FirstPageNo >= FPages.ItemCount) then
   raise EACRException.Create(10658,ErrorLCannotReadFromFile,
    [Position,Count,FPages.ItemCount,FSize]);
  LastPageNo := GetPageNo(Count + Position);
  if (LastPageNo >= FPages.ItemCount) then
   raise EACRException.Create(10659,ErrorLCannotReadFromFile,
    [Position,Count,FPages.ItemCount,FSize]);
  if (Position + Count > FSize) then
   raise EACRException.Create(10660,ErrorLCannotReadFromFile,
    [Position,Count,FPages.ItemCount,FSize]);

  for i := FirstPageNo to LastPageNo do
   begin
    PageNo := FPages.Items[i];
    DBFilePos := LPageManager.GetPageOffset(PageNo) + LPageManager.PageHeaderSize;
    DBFileSize := LPageManager.PageDataSize;
    if ((i = FirstPageNo) or (i = LastPageNo)) then
     begin
      if (i = FirstPageNo) then
       begin
        Offset := GetPageOffset(Position);
        Inc(DBFilePos,Offset);
        Dec(DBFileSize,Offset);
       end;
      if (i = LastPageNo) then
       begin
        Offset := GetPageOffset(Position+Count);
        Dec(DBFileSize,(LPageManager.PageDataSize - Offset));
       end;
     end;
    Result := LPageManager.DatabaseFile.IsRegionLocked(DBFilePos,DBFileSize);
    if (Result) then
     break;
   end;
end;//IsRegionLocked


////////////////////////////////////////////////////////////////////////////////
//
// TACRInternalDBFile
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// InitFile
//------------------------------------------------------------------------------
procedure TACRInternalDBFile.InitFile;
begin
  FFileData.Pages.SetSize(1);
  FFileData.Pages.Items[0] := FStartPageNo;
  if (FFileData.Buffer <> nil) then
   MemoryManager.FreeAndNilMem(FFileData.Buffer);
  FFileData.Size := 0;
end; // InitFile


//------------------------------------------------------------------------------
// raise if file is Closed
//------------------------------------------------------------------------------
procedure TACRInternalDBFile.CheckFileOpened(OperationName: String);
begin
  if (FStartPageNo = INVALID_PAGE_NO) then
    raise EACRException.Create(30389, ErrorGIFClosed, [OperationName]);
end;//CheckFileOpened


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRInternalDBFile.Create(
                       PageManager:               TACRDiskPageManager;
                       PageTypeID:                TACRPageTypeID;
                       IsSynchronizationRequired: TACRIsPageSynchronizationRequired = nil);
begin
  LPageManager := PageManager;
  FPageTypeID := PageTypeID;
  FStartPageNo := INVALID_PAGE_NO;
  FIsLoaded := False;
  FExclusive := False;
  FFileData.Buffer := nil;
  FFileData.Pages := TACRIntegerArray.Create(0,1,10);
  FIsSynchronizationRequired := IsSynchronizationRequired;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRInternalDBFile.Destroy;
begin
  FreeAndNil(FFileData.Pages);
  if (FFileData.Buffer <> nil) then
   MemoryManager.FreeAndNilMem(FFileData.Buffer);
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// Resize File Data Buffer
//------------------------------------------------------------------------------
procedure TACRInternalDBFile.ResizeFileData(NewSize: Integer);
begin
 if (FFileData.Buffer <> nil) then
  MemoryManager.FreeAndNilMem(FFileData.Buffer);
 FFileData.Size := NewSize;
 if (NewSize > 0) then
  FFileData.Buffer := MemoryManager.GetMem(NewSize);
end; // ResizeFileData


//------------------------------------------------------------------------------
// Create file
//------------------------------------------------------------------------------
procedure TACRInternalDBFile.CreateFile(SessionID: TACRSessionID; ObjectID: TACRObjectID);
var
  Page: TACRPage;
begin
  // Check File Opened
  Page := LPageManager.AddPage(SessionID);
  try
    FStartPageNo := Page.PageNo;
    // Init Page Header
    Page.PageHeader.PageType := FPageTypeID;
    Page.PageHeader.ObjectID := ObjectID;
    // Init File Header
    PACRInternalFileHeader(Page.PageData).Signature := ACRInternalDBFileSignature;
    PACRInternalFileHeader(Page.PageData).FileSize := 0;
    PACRInternalFileHeader(Page.PageData).CompressionMode := 0;
  finally
    LPageManager.PutPage(Page);
  end;
  InitFile;
end;//CreateFile


//------------------------------------------------------------------------------
// Delete file. And free all file pages
//------------------------------------------------------------------------------
procedure TACRInternalDBFile.DeleteFile(SessionID: TACRSessionID; StartPageNo: TACRPageNo);
var
  Page: TACRPage;
  PageNo, NextPageNo: TACRPageNo;
begin
  if (FFileData.Buffer <> nil) then
   MemoryManager.FreeAndNilMem(FFileData.Buffer);
  FFileData.Pages.SetSize(0);
  FFileData.Size := 0;
  FStartPageNo := StartPageNo;
  PageNo := FStartPageNo;
  repeat
    Page := LPageManager.GetPage(SessionID, PageNo, FIsSynchronizationRequired);
    NextPageNo := Page.PageHeader.NextPageNo;
    LPageManager.PutPage(Page);
    LPageManager.RemovePage(SessionID, PageNo);
    PageNo := NextPageNo;
  until (PageNo = INVALID_PAGE_NO);
  FStartPageNo := INVALID_PAGE_NO;
end;//DeleteFile


//------------------------------------------------------------------------------
// Open file (reading file header)
//------------------------------------------------------------------------------
procedure TACRInternalDBFile.OpenFile(StartPageNo: TACRPageNo; OpenExclusive: Boolean);
begin
  FStartPageNo := StartPageNo;
  if (FStartPageNo < ACRFirstPageNoSystemDirectory) then
   raise EACRException.Create(10698,ErrorLCannotOpenFileInvalidStartPage,
    [FPageTypeID,FStartPageNo]);
  InitFile;
  FExclusive := OpenExclusive;
  FIsLoaded := False;
end;//OpenFile


//------------------------------------------------------------------------------
// Read File Data
//------------------------------------------------------------------------------
procedure TACRInternalDBFile.ReadFile(SessionID: TACRSessionID);
var Page:       TACRPage;
    PageNo:     TACRPageNo;
    BytesRead:  Integer;
    DataSize:   Integer;
    Offset:     Integer;
begin
  CheckFileOpened('ReadFile');
  if (not (FExclusive and FIsLoaded)) then
   begin
    PageNo := FStartPageNo;
    BytesRead := 0;
    FFileData.Pages.SetSize(0);
    repeat
     Page := LPageManager.GetPage(SessionID,PageNo, FIsSynchronizationRequired);
     try
       FFileData.Pages.Append(PageNo);
       if (PageNo = FStartPageNo) then
        begin
         FFileData.Size := PACRInternalFileHeader(Page.PageData)^.FileSize;
         if (FFileData.Size <= 0) then
          break;
         if (FFileData.Buffer <> nil) then
          MemoryManager.FreeAndNilMem(FFileData.Buffer);
         FFileData.Buffer := MemoryManager.GetMem(FFileData.Size);
         Offset := SizeOf(TACRInternalFileHeader);
        end
       else
        Offset := 0;
       if ((FFileData.Size - BytesRead) + Offset <= Page.PageDataSize) then
        DataSize := FFileData.Size - BytesRead
       else
        DataSize := Page.PageDataSize - Offset;
       Move(PChar(Page.PageData + Offset)^,
            PChar(FFileData.Buffer + BytesRead)^,
            DataSize);
       Inc(BytesRead,DataSize);
       PageNo := Page.PageHeader.NextPageNo;
       if (BytesRead >= FFileData.Size) then
        break;
     finally
       LPageManager.PutPage(Page);
     end;
    until (PageNo = INVALID_PAGE_NO);
    if (FFileData.Size <> BytesRead) then
     begin
      if (FFileData.Buffer <> nil) then
       MemoryManager.FreeAndNilMem(FFileData.Buffer);
      FFileData.Size := 0;
      raise EACRException.Create(10663,ErrorLCannotReadFile,[FFileData.Size,BytesRead]);
     end;
    FIsLoaded := True;
   end;
end; // ReadFile


//------------------------------------------------------------------------------
// Write File
//------------------------------------------------------------------------------
procedure TACRInternalDBFile.WriteFile(SessionID: TACRSessionID);
var Page,NextPage:   TACRPage;
    BytesWrite:      Integer;
    DataSize:        Integer;
    Offset,i:        Integer;

procedure WritePage;
begin
 Offset := 0;
 if (BytesWrite = 0) then
  begin
    PACRInternalFileHeader(Page.PageData)^.Signature := ACRInternalDBFileSignature;
    PACRInternalFileHeader(Page.PageData)^.FileSize := FFileData.Size;
    PACRInternalFileHeader(Page.PageData)^.CompressionMode := 0;
    Offset := SizeOf(TACRInternalFileHeader);
  end;
 if ((FFileData.Size - BytesWrite) + Offset <= Page.PageDataSize) then
    DataSize := FFileData.Size - BytesWrite
 else
    DataSize := Page.PageDataSize - Offset;
 Move(PChar(FFileData.Buffer + BytesWrite)^,
        PChar(Page.PageData + Offset)^,
        DataSize);
 Inc(BytesWrite,DataSize);
end; // WritePage

begin
  CheckFileOpened('WriteFile');
  if (FFileData.Size = 0) then
   EmptyFile(SessionID)
  else
   begin
    BytesWrite := 0;
    NextPage := nil;
    for i := 0 to FFileData.Pages.ItemCount - 1 do
     begin
      if (BytesWrite >= FFileData.Size) then
       begin
        // Shrinking file...
        if (i > 0) then
         LPageManager.RemovePage(SessionID,FFileData.Pages.Items[i])
       end // Shrinking file ... 
      else
       begin
        // Rewriting file
        Page := LPageManager.GetPage(SessionID,FFileData.Pages.Items[i], FIsSynchronizationRequired);
        try
         Page.IsDirty := True;
         WritePage;
         // Is this page last? If yes then set NextPageNo := INVALID_PAGE_NO
         if (BytesWrite >= FFileData.Size) then
          Page.PageHeader.NextPageNo := INVALID_PAGE_NO
         else
          if (i = FFileData.Pages.ItemCount - 1) then
           begin
            // extend the file
            NextPage := LPageManager.AddPage(SessionID);
            NextPage.PageHeader.PageType := FPageTypeID;
            Page.PageHeader.NextPageNo := NextPage.PageNo;
           end;
        finally
         LPageManager.PutPage(Page);
        end; // put page
       end; // Rewriting file
     end; // scan all loaded pages
     if (BytesWrite > FFileData.Size) then
      raise EACRException.Create(10663,ErrorLCannotWriteFile,
        [FFileData.Size,BytesWrite]);
     // add new pages
     while (BytesWrite < FFileData.Size) do
      begin
       if (NextPage = nil) then
        raise EACRException.Create(10666,ErrorLNilPointer)
       else
        Page := NextPage;
       try
        WritePage;
        if (BytesWrite >= FFileData.Size) then
         begin
          Page.PageHeader.NextPageNo := INVALID_PAGE_NO;
          NextPage := nil;
         end
        else
         begin
          NextPage := LPageManager.AddPage(SessionID);
          Page.PageHeader.PageType := FPageTypeID;
          Page.PageHeader.NextPageNo := NextPage.PageNo;
         end;
       finally
        LPageManager.PutPage(Page);
       end;
      end; // write bytes
     if (BytesWrite > FFileData.Size) then
      raise EACRException.Create(10664,ErrorLCannotWriteFile,
        [FFileData.Size,BytesWrite]);
   end; // not empty file
end; // WriteFile


//------------------------------------------------------------------------------
// EmptyFile
//------------------------------------------------------------------------------
procedure TACRInternalDBFile.EmptyFile(SessionID: TACRSessionID);
var Page:   TACRPage;
    PageNo:     TACRPageNo;
    NextPageNo: TACRPageNo;
begin
  CheckFileOpened('EmptyFile');
  InitFile;
  PageNo := FStartPageNo;
  while (PageNo <> INVALID_PAGE_NO) do
   begin
    Page := LPageManager.GetPage(SessionID,PageNo, FIsSynchronizationRequired);
    try
      NextPageNo := Page.PageHeader.NextPageNo;
      if (PageNo = FStartPageNo) then
       begin
        Page.IsDirty := True;
        Page.PageHeader.NextPageNo := INVALID_PAGE_NO;
        PACRInternalFileHeader(Page.PageData)^.Signature := ACRInternalDBFileSignature;
        PACRInternalFileHeader(Page.PageData)^.FileSize := 0;
        PACRInternalFileHeader(Page.PageData)^.CompressionMode := 0;
        Page.PageHeader.PageType := FPageTypeID;
       end;
    finally
      LPageManager.PutPage(Page);
    end;
    if (PageNo <> FStartPageNo) then
      LPageManager.RemovePage(SessionID,PageNo);
    PageNo := NextPageNo;
   end;
end; // EmptyFile


////////////////////////////////////////////////////////////////////////////////
//
// TACRActiveSessionsFile
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRActiveSessionsFile.Create(
                                    PageManager: TACRPageManager;
                                    aLockParams: TACRLockParams
                                           );
begin
  LPageManager := TACRDiskPageManager(PageManager);
  FHandle := nil;
  LockParams := aLockParams;
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRActiveSessionsFile.Destroy;
begin
  CloseFile;
  inherited;
end;// Destroy


//------------------------------------------------------------------------------
// CreateFile
//------------------------------------------------------------------------------
function TACRActiveSessionsFile.CreateFile(aMaxSessionCount: Integer): TACRPageNo;
begin
  CloseFile;
  FHandle := TACRInternalLockedDBFile.Create(LPageManager, ACRPageTypeIDActiveSessionList);
  FHandle.CreateFile(aMaxSessionCount,ACRObjectIDActiveSessionsFile);
  Result := FHandle.StartPageNo;
  CloseFile;
end;// CreateFile


//------------------------------------------------------------------------------
// OpenFile
//------------------------------------------------------------------------------
procedure TACRActiveSessionsFile.OpenFile(aStartPageNo: TACRPageNo);
begin
  CloseFile;
  FHandle := TACRInternalLockedDBFile.Create(LPageManager, ACRPageTypeIDActiveSessionList);
  FHandle.OpenFile(aStartPageNo);
end;// OpenFile


//------------------------------------------------------------------------------
// CloseFile
//------------------------------------------------------------------------------
procedure TACRActiveSessionsFile.CloseFile;
begin
  if (FHandle <> nil) then
    FreeAndNil(FHandle);
end;// CloseFile


//------------------------------------------------------------------------------
// Connect
//------------------------------------------------------------------------------
function TACRActiveSessionsFile.Connect: TACRSessionID;
var
  i: Integer;
begin
  if (not TryUsingTimeOut(FHandle.LockFile, LockParams)) then
   raise EACRException.Create(10503,ErrorLLockByTimeoutFailed,
                                [LockParams.RetryCount,LockParams.Delay]);
  try
    Result := INVALID_SESSION_ID;
    for i := 0 to FHandle.Size-1 do
      if (FHandle.LockByte(i)) then
        begin
          Result := i;
          break;
        end;
    if (Result = INVALID_SESSION_ID) then
     raise EACRException.Create(10842, ErrorLMaximumSessionCountExceeded,
             [FHandle.Size]);
  finally
    if (not FHandle.UnlockFile) then
      raise EACRException.Create(20093, ErrorACannotUnlockSessionsFile);
  end;
end;// Connect


//------------------------------------------------------------------------------
// Disconnect
//------------------------------------------------------------------------------
procedure TACRActiveSessionsFile.Disconnect(SessionID: TACRSessionID);
begin
  if (not TryUsingTimeOut(FHandle.LockFile, LockParams)) then
   raise EACRException.Create(10504,ErrorLLockByTimeoutFailed,
                                      [LockParams.RetryCount,LockParams.Delay]);
  try
     if (not FHandle.UnlockByte(SessionID)) then
      raise EACRException.Create(20095, ErrorACannotDisconnectSession);
  finally
    if (not FHandle.UnlockFile) then
      raise EACRException.Create(20096, ErrorACannotUnlockSessionsFile);
  end;
end;// Disconnect




////////////////////////////////////////////////////////////////////////////////
//
// TACRTableListFile
//
////////////////////////////////////////////////////////////////////////////////
{
function TACRTableListFile.GetStartPageNo: TACRPageNo;
begin
  Result := FHandle.StartPageNo;
end;// GetStartPageNo
}


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRTableListFile.Create(PageManager: TACRPageManager);
begin
  LPageManager := TACRDiskPageManager(PageManager);
  FHandle := TACRInternalDBFile.Create(LPageManager, ACRPageTypeIDTableList);
  FTableList := nil;
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRTableListFile.Destroy;
begin
  if (FHandle <> nil) then
    FreeAndNil(FHandle);
  inherited;
end;// Destroy


//------------------------------------------------------------------------------
// CreateFile
//------------------------------------------------------------------------------
function TACRTableListFile.CreateFile: TACRPageNo;
begin
  FHandle.CreateFile(SYSTEM_SESSION_ID,ACRObjectIDTableListFile);
  Result := FHandle.StartPageNo;
end;// CreateFile


//------------------------------------------------------------------------------
// OpenFile
//------------------------------------------------------------------------------
procedure TACRTableListFile.OpenFile(aStartPageNo: TACRPageNo);
begin
  FHandle.OpenFile(aStartPageNo,False);
end;// OpenFile


//------------------------------------------------------------------------------
// Load Table List
//------------------------------------------------------------------------------
procedure TACRTableListFile.Load;
begin
 FHandle.ReadFile(SYSTEM_SESSION_ID);
 if FHandle.FileData.Size <> 0 then
  begin
    if (FHandle.FileData.Size mod SizeOf(TACRTableListItem)) <> 0 then
      EACRException.Create(40002, ErrorRTableListFileInvalidStructure,
                            [FHandle.FileData.Size, SizeOf(TACRTableListItem)]);
    SetLength(FTableList,FHandle.FileData.Size div SizeOf(TACRTableListItem));
    Move(FHandle.FFileData.Buffer^,FTableList[0],FHandle.FileData.Size);
  end
 else
  SetLength(FTableList, 0);
end;//Load


//------------------------------------------------------------------------------
// Save Table List
//------------------------------------------------------------------------------
procedure TACRTableListFile.Save;
var
  Len: Integer;
begin
  Len := Length(FTableList);
  if Len <> 0 then
   begin
    FHandle.ResizeFileData(Len*SizeOf(TACRTableListItem));
    Move(FTableList[0],FHandle.FFileData.Buffer^,FHandle.FileData.Size);
    FHandle.WriteFile(SYSTEM_SESSION_ID);
   end
  else
   FHandle.EmptyFile(SYSTEM_SESSION_ID);
end;//Save


//------------------------------------------------------------------------------
// IsTableExist
//------------------------------------------------------------------------------
function TACRTableListFile.IsTableExist(TableName: String): Boolean;
var i: Integer;
begin
  Load;
  Result := False;
  for i := 0 to Length(FTableList)-1 do
    if AnsiUpperCase(FTableList[i].TableName) = AnsiUpperCase(TableName) then
      begin
        Result := True;
        Break;
      end;
end;//IsTableExist


//------------------------------------------------------------------------------
// CreateTable
//------------------------------------------------------------------------------
function TACRTableListFile.CreateTable(TableItem: TACRTableListItem):TACRTableID;
var
  Len: Integer;
begin
  if IsTableExist(TableItem.TableName) then
    raise EACRException.Create(30394, ErrorGTableAlreadyExists, [TableItem.TableName]);
  Load;
  Len := Length(FTableList);
  SetLength(FTableList, Len+1);
  TableItem.TableID := Len+1;
  FTableList[Len] := TableItem;
  Save;
  Result:= Len+1;
end;//CreateTable


//------------------------------------------------------------------------------
// DeleteTable
//------------------------------------------------------------------------------
procedure TACRTableListFile.DeleteTable(TableName: String);
var
  Len, i: Integer;
  TableNotFound : Boolean;
begin
  Load;
  Len := Length(FTableList);
  TableNotFound := True;
  for i:=0 to Len-1 do
    if AnsiUpperCase(FTableList[i].TableName) = AnsiUpperCase(TableName) then
      begin
        if (i < Len-1) then
          Move(FTableList[i+1],FTableList[i],(Len-i-1)*SizeOf(TACRTableListItem));
        SetLength(FTableList, Len-1);
        Save;
        TableNotFound := False;
        break;
      end;
  if TableNotFound then
    raise EACRException.Create(40003, ErrorRTableNotFound, [TableName]);
end;//DeleteTable


//------------------------------------------------------------------------------
// RenameTable
//------------------------------------------------------------------------------
procedure TACRTableListFile.RenameTable(OldTableName, NewTableName: String);
var
  i: Integer;
  TableNotFound : Boolean;
begin
  Load;
  TableNotFound := True;
  for i:=0 to Length(FTableList)-1 do
    if AnsiUpperCase(FTableList[i].TableName) = AnsiUpperCase(OldTableName) then
      begin
        FTableList[i].TableName := NewTableName;
        Save;
        TableNotFound := False;
      end;
  if TableNotFound then
    raise EACRException.Create(40003, ErrorRTableNotFound, [OldTableName]);
end;//RenameTable


//------------------------------------------------------------------------------
// GetTablesList
//------------------------------------------------------------------------------
procedure TACRTableListFile.GetTablesList(List: TStrings);
var i: Integer;
begin
  Load;
  for i := 0 to Length(FTableList)-1 do
    List.Add(FTableList[i].TableName);
end; // GetTablesList


//------------------------------------------------------------------------------
// Return TableID for next call CreateTable
//------------------------------------------------------------------------------
function TACRTableListFile.GetNextTableID: TACRObjectID;
begin
 Load;
 Result := Length(FTableList) + 1;
end; // GetNextTableID;


//------------------------------------------------------------------------------
// GetTableItem
//------------------------------------------------------------------------------
function TACRTableListFile.GetTableItem(TableName: String): PACRTableListItem;
var
  i: Integer;
  TableNotFound : Boolean;
begin
  Load;
  TableNotFound := True;
  for i := 0 to Length(FTableList)-1 do
    if AnsiUpperCase(FTableList[i].TableName) = AnsiUpperCase(TableName) then
      begin
        Result := @FTableList[i];
        TableNotFound := False;
        break;
      end;
  if TableNotFound then
    raise EACRException.Create(40003, ErrorRTableNotFound, [TableName]);
end;


//------------------------------------------------------------------------------
// GetTableItemIfExists
//------------------------------------------------------------------------------
function TACRTableListFile.GetTableItemIfExists(TableName: String; var TableItem: TACRTableListItem): Boolean;
var
  i: Integer;
begin
  Load;
  Result := False;
  for i := 0 to Length(FTableList)-1 do
    if AnsiUpperCase(FTableList[i].TableName) = AnsiUpperCase(TableName) then
      begin
        TableItem := FTableList[i];
        Result := True;
        break;
      end;
end; // GetTableItemIfExists


////////////////////////////////////////////////////////////////////////////////
//
// TACRTableMetadataFile
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRTableMetadataFile.Create(
                       PageManager: TACRPageManager
                                      );

begin
  LPageManager := TACRDiskPageManager(PageManager);
  FHandle := TACRInternalDBFile.Create(LPageManager, ACRPageTypeIDTableMetadata);
end; // Create;


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRTableMetadataFile.Destroy;
begin
  if (FHandle <> nil) then
    FreeAndNil(FHandle);
//  CloseFile;
  inherited;
end; // Destroy;


//------------------------------------------------------------------------------
// CreateFile
//------------------------------------------------------------------------------
function TACRTableMetadataFile.CreateFile(
              SessionID:        TACRSessionID;
              ObjectID:         TACRObjectID
                                          ): TACRPageNo;
begin
  FHandle.FStartPageNo := -1;
  FHandle.CreateFile(SessionID,ObjectID);
  Result := FHandle.StartPageNo;
end;// CreateFile


//------------------------------------------------------------------------------
// DeleteFile
//------------------------------------------------------------------------------
procedure TACRTableMetadataFile.DeleteFile(
              SessionID:        TACRSessionID;
              aStartPageNo: TACRPageNo
                                              );
begin
  FHandle.DeleteFile(SessionID, aStartPageNo);
end;// DeleteFile


//------------------------------------------------------------------------------
// EmptyFile
//------------------------------------------------------------------------------
procedure TACRTableMetadataFile.EmptyFile(SessionID: TACRSessionID);
begin
  FHandle.EmptyFile(SessionID);
end;// EmptyFile


//------------------------------------------------------------------------------
// ReadFile
//------------------------------------------------------------------------------
procedure TACRTableMetadataFile.ReadFile(
              SessionID:        TACRSessionID;
              aStartPageNo:     TACRPageNo;
              ms:               TACRStream
                                         );
begin
  FHandle.OpenFile(aStartPageNo,False);
  FHandle.ReadFile(SessionID);
  if (FHandle.FileData.Size <= 0) then
   raise EACRException.Create(10667,ErrorLInvalidFileSize,[FHandle.FileData.Size]);
  ms.Size := FHandle.FileData.Size;
  ms.WriteBuffer(FHandle.FFileData.Buffer^,FHandle.FFileData.Size);
  ms.Position := 0;
end; // ReadFile


//------------------------------------------------------------------------------
// WriteFile
//------------------------------------------------------------------------------
procedure TACRTableMetadataFile.WriteFile(
              SessionID:        TACRSessionID;
              aStartPageNo:     TACRPageNo;
              ms:               TACRStream
                                           );
begin
  FHandle.OpenFile(aStartPageNo,False);
  if (ms.Size <= 0) then
   raise EACRException.Create(10668,ErrorLInvalidFileSize,[FHandle.FileData.Size]);
  FHandle.ReadFile(SessionID);
  FHandle.ResizeFileData(ms.Size);
  ms.Position := 0;
  ms.ReadBuffer(FHandle.FFileData.Buffer^,FHandle.FFileData.Size);
  ms.Position := 0;
  FHandle.WriteFile(SessionID);
end; // WriteFile


////////////////////////////////////////////////////////////////////////////////
//
// TACRTableMostUpdatedFile
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRTableMostUpdatedFile.Create(
                       PageManager:               TACRPageManager;
                       IsSynchronizationRequired: TACRIsPageSynchronizationRequired
                                      );

begin
  LPageManager := TACRDiskPageManager(PageManager);
  FHandle := TACRInternalDBFile.Create(LPageManager, ACRPageTypeIDTableMostUpdated, IsSynchronizationRequired);
end; // Create;


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRTableMostUpdatedFile.Destroy;
begin
  FreeAndNil(FHandle);
  inherited;
end; // Destroy;


//------------------------------------------------------------------------------
// CreateFile
//------------------------------------------------------------------------------
function TACRTableMostUpdatedFile.CreateFile(
                              SessionID:        TACRSessionID;
                              ObjectID:         TACRObjectID
                                              ): TACRPageNo;
begin
  FHandle.CreateFile(SessionID,ObjectID);
  Result := FHandle.StartPageNo;
end;// CreateFile


//------------------------------------------------------------------------------
// OpenFile
//------------------------------------------------------------------------------
procedure TACRTableMostUpdatedFile.OpenFile(
          aStartPageNo:   TACRPageNo;
          OpenExclusive:  Boolean
          );
begin
  FHandle.OpenFile(aStartPageNo,OpenExclusive);
end; // OpenFile


//------------------------------------------------------------------------------
// DeleteFile
//------------------------------------------------------------------------------
procedure TACRTableMostUpdatedFile.DeleteFile(
                      SessionID:    TACRSessionID;
                      aStartPageNo: TACRPageNo
                                              );
begin
  FHandle.DeleteFile(SessionID, aStartPageNo);
end;// DeleteFile


//------------------------------------------------------------------------------
// ReadFile
//------------------------------------------------------------------------------
procedure TACRTableMostUpdatedFile.ReadFile(SessionID: TACRSessionID);
begin
 FHandle.ReadFile(SessionID);
end; // ReadFile


//------------------------------------------------------------------------------
// WriteFile
//------------------------------------------------------------------------------
procedure TACRTableMostUpdatedFile.WriteFile(SessionID: TACRSessionID);
begin
 FHandle.WriteFile(SessionID);
end; // WriteFile


////////////////////////////////////////////////////////////////////////////////
//
// TACRTableLocksFile
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// return number of start page
//------------------------------------------------------------------------------
function TACRTableLocksFile.GetStartPageNo: TACRPageNo;
begin
  if (FHandle = nil) then
   raise EACRException.Create(10483,ErrorLNilPointer);
  Result := FHandle.StartPageNo;
end;// GetStartPageNo


//------------------------------------------------------------------------------
// return Max Sessions Count
//------------------------------------------------------------------------------
function TACRTableLocksFile.GetMaxSessionCount: Integer;
begin
 Result := FHandle.Size div (ACRMaxLockTypes + SizeOf(TACRRecordID));
end; // GetMaxSessionCount


//------------------------------------------------------------------------------
// return number of byte for lock / unlock /is locked operations
//------------------------------------------------------------------------------
function TACRTableLocksFile.GetByteNo(
                        SessionID:    TACRSessionID;
                        LockType:     TACRLockType;
                        UseRecordID:  Boolean
                                     ): Integer;
begin
 if (not UseRecordID) then
   Result := SessionID + Integer(LockType) * MaxSessionCount
 else
   Result := SessionID * SizeOf(TACRRecordID) +
              (Integer(LockType) + 1) * MaxSessionCount;
end; // GetByteNo


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRTableLocksFile.Create(
                        PageManager:  TACRPageManager;
                        aLockParams:  TACRLockParams
                                     );
begin
  LPageManager := TACRDiskPageManager(PageManager);
  FHandle := TACRInternalLockedDBFile.Create(LPageManager, ACRPageTypeIDTableLocksFile);
  LockParams := aLockParams;
  FLockDelayOn := False;
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRTableLocksFile.Destroy;
begin
  if (FHandle <> nil) then
    FreeAndNil(FHandle);
  inherited;
end;// Destroy


//------------------------------------------------------------------------------
// CreateFile
//------------------------------------------------------------------------------
function TACRTableLocksFile.CreateFile(aMaxSessionCount: Integer; ObjectID: TACRObjectID): TACRPageNo;
begin
  FMaxSessionCount := aMaxSessionCount;
  FHandle.FStartPageNo := -1;
  FHandle.CreateFile(FMaxSessionCount * (ACRMaxLockTypes + SizeOf(TACRRecordID)),
                     ObjectID);
  Result := FHandle.StartPageNo;
end;// CreateFile


//------------------------------------------------------------------------------
// DeleteFile
//------------------------------------------------------------------------------
procedure TACRTableLocksFile.DeleteFile(
              SessionID:        TACRSessionID;
              aStartPageNo: TACRPageNo
                                              );
begin
  FHandle.DeleteFile(aStartPageNo);
end;// DeleteFile


//------------------------------------------------------------------------------
// OpenFile
//------------------------------------------------------------------------------
procedure TACRTableLocksFile.OpenFile(aStartPageNo: TACRPageNo);
begin
  FHandle.OpenFile(aStartPageNo);
  FMaxSessionCount := GetMaxSessionCount;
end;// OpenFile


//------------------------------------------------------------------------------
// lock table and return true if table locked successfully, otherwise return false
//------------------------------------------------------------------------------
function TACRTableLocksFile.LockTableIfLockAllowed(
                          const CurrentRetryNo: Integer;
                          SessionID: TACRSessionID;
                          LockType: TACRLockType): Boolean;
var LowCount,HighCount:                     Integer;
    StartTime:                              DWORD;
    FIsOtherSessionsConnectedToThisTable:   Boolean;

 function IsTableLockedInISByOtherSessions: Boolean;
 begin
  Result := False;
  if (SessionID > 0) then
   Result := FHandle.IsRegionLocked(GetByteNo(0,ltIS),LowCount);
  if (not Result) then
    if (SessionID < MaxSessionCount-1) then
     Result := FHandle.IsRegionLocked(GetByteNo(SessionID+1,ltIS),
                                            HighCount);
 end; // IsTableLockedInISByOtherSessions

 function IsOtherSessionsConnectedToThisTable: Boolean;
 begin
   Result := FHandle.IsFileLocked;
   if (not Result) then
    Result := IsTableLockedInISByOtherSessions;
 end; // IsOtherSessionsConnectedToThisTable

 function IsSleepNeeded: Boolean;
 begin
  FLockDelayOn := FIsOtherSessionsConnectedToThisTable;
  Result := FLockDelayOn and
              (Integer(GetTickCount - FLockDelayTime) > 1000);
 end; // IsSleepNeeded

 function IsLastRetry: Boolean;
 begin
  Result := (CurrentRetryNo >= FLockParams.RetryCount);
 end;

 function IsLastRetryFinished: Boolean;
 begin
   Result := ((GetTickCount - StartTime) > ACRMAXLastLockRetryTime);
  {$IFDEF DEBUG_TRACE_TABLE_LOCKS}
     if (Result) then
      aaWriteToLog(aaGetCurrentTimeAsString+'  TableLocksFile last retry finished, lock type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID));
  {$ENDIF}
 end;

begin
 LowCount := SessionID;
 HighCount := MaxSessionCount - LowCount - 1;
 FIsOtherSessionsConnectedToThisTable := True;
// FIsOtherSessionsConnectedToThisTable := IsOtherSessionsConnectedToThisTable;
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
 aaWriteToLog(aaGetCurrentTimeAsString+' TableLocksFile starting locking file... , type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID)+', CurrentRetryNo = '+IntToStr(CurrentRetryNo));
{$ENDIF}
 // try to lock file
 StartTime := GetTickCount;
 repeat
  Result := FHandle.LockFile;
  if (not Result) then
    if (GetTickCount - StartTime > ACRMAXLockFileTime) then
        break;
 until (Result);
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
if (not Result) then
 aaWriteToLog(aaGetCurrentTimeAsString+' TableLocksFile lock set failed - cannot lock file, type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID))
else
 aaWriteToLog(aaGetCurrentTimeAsString+' TableLocksFile lock set started - file locked ok, type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID));
{$ENDIF}
 if (Result) then
  try
   if (Result) then
    begin
      // check for compatibility with other locks
      if (LockType = ltX) then
       begin
        Result := not IsTableLockedInISByOtherSessions;
       end
      else
       begin
        if (SessionID < 0) or (SessionID >= MaxSessionCount) then
         raise EACRException.Create(10809,ErrorLInvalidSessionID,[SessionID]);
        case LockType of
         ltIS:
           Result := True;
         ltIRW, ltS, ltRW:
          begin
           if (not FIsOtherSessionsConnectedToThisTable) then
            Result := True
           else
            begin
             if (IsLastRetry) then
              begin
               StartTime := GetTickCount;
              end;
             // check for other sessions locks
             repeat
  {$IFDEF DEBUG_TRACE_TABLE_LOCKS}
     if (IsLastRetry) then
      aaWriteToLog(aaGetCurrentTimeAsString+'  TableLocksFile last retry started, lock type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID));
  {$ENDIF}
               Result := True;
               if (LockType <> ltS) then
                begin
                 if (SessionID > 0) then
                  begin
                    Result := not FHandle.IsRegionLocked(GetByteNo(0,ltIRW),LowCount);
  {$IFDEF DEBUG_TRACE_TABLE_LOCKS}
    if (not Result) then
      aaWriteToLog(aaGetCurrentTimeAsString+'  TableLocksFile lock set failed - IRW locked before, lock type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID));
  {$ENDIF}
                  end;
                 if (Result) then
                  if (SessionID < MaxSessionCount-1) then
                   begin
                     Result := not FHandle.IsRegionLocked(GetByteNo(SessionID+1,ltIRW),
                                                          HighCount);
  {$IFDEF DEBUG_TRACE_TABLE_LOCKS}
    if (not Result) then         
      aaWriteToLog(aaGetCurrentTimeAsString+'  TableLocksFile lock set failed - IRW locked after, lock type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID));
  {$ENDIF}
                   end;
                end; // IRW
               // check RW
               if (Result) then
                if (SessionID > 0) then
                 begin
                   Result := not FHandle.IsRegionLocked(GetByteNo(0,ltRW),LowCount);
  {$IFDEF DEBUG_TRACE_TABLE_LOCKS}
    if (not Result) then
      aaWriteToLog(aaGetCurrentTimeAsString+'  TableLocksFile lock set failed - RW locked before, lock type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID));
  {$ENDIF}
                 end;
               if (Result) then
                if (SessionID < MaxSessionCount-1) then
                 begin
                   Result := not FHandle.IsRegionLocked(GetByteNo(SessionID+1,ltRW),
                                                        HighCount);
  {$IFDEF DEBUG_TRACE_TABLE_LOCKS}
    if (not Result) then
      aaWriteToLog(aaGetCurrentTimeAsString+'  TableLocksFile lock set failed - RW locked after, lock type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID));
  {$ENDIF}
                 end;
               // check S
               if (Result) and (LockType = ltRW) then
                begin
                 if (Result) then
                  if (SessionID > 0) then
                   begin
                     Result := not FHandle.IsRegionLocked(GetByteNo(0,ltS),LowCount);
  {$IFDEF DEBUG_TRACE_TABLE_LOCKS}
    if (not Result) then
      aaWriteToLog(aaGetCurrentTimeAsString+'  TableLocksFile lock set failed - S locked before, lock type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID));
  {$ENDIF}
                   end;
                 if (Result) then
                  if (SessionID < MaxSessionCount-1) then
                   begin
                     Result := not FHandle.IsRegionLocked(GetByteNo(SessionID+1,ltS),
                                                          HighCount);
  {$IFDEF DEBUG_TRACE_TABLE_LOCKS}
    if (not Result) then
      aaWriteToLog(aaGetCurrentTimeAsString+'  TableLocksFile lock set failed - S locked after, lock type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID));
  {$ENDIF}
                   end;
                end; // check S
              if (not Result) then
               if (IsLastRetry) then
                begin
  {$IFDEF DEBUG_TRACE_TABLE_LOCKS}
      aaWriteToLog(aaGetCurrentTimeAsString+'  TableLocksFile last retry failed, lock type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID));
  {$ENDIF}
                  if (IsLastRetryFinished) then break;
                end;
             until (Result) or (not IsLastRetry);
            end; // check for other sessions locks
          end // ltS, ltIRW, ltRW
         else
          raise EACRException.Create(10753,ErrorLInvalidLockType,[Integer(LockType)]);
        end; // case - check for locks compatibility
           // check IRW
        if (Result) then
          Result := FHandle.LockByte(GetByteNo(SessionID,LockType));
       end; // not X lock
    end; // if Result
  finally
     if ((LockType <> ltX) or ((LockType = ltX) and (not Result))) then
      if (not FHandle.UnlockFile) then
       begin
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
 aaWriteToLog(aaGetCurrentTimeAsString+'  TableLocksFile Unlock failed, lock type = '+IntToStr(Integer(LockType))+', SessionID = '+IntToStr(SessionID));
{$ENDIF}
         raise EACRException.Create(10752, ErrorLCannotUnlockTableLocksFile);
       end;
  end; // finally unlock file
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
if (Result) then
 aaWriteToLog(aaGetCurrentTimeAsString+' TableLocksFile lock set OK, type = '+IntToStr(Integer(LockType))+', ID = '+IntToStr(SessionID))
else
 aaWriteToLog(aaGetCurrentTimeAsString+' TableLocksFile lock set failed, file unlocked, type = '+IntToStr(Integer(LockType))+', ID = '+IntToStr(SessionID));
{$ENDIF}

   if (Result) then
    begin
      if (not FLockDelayOn) then
       begin
        FLockDelayTime := GetTickCount;
        FLockDelayOn := True;
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
 aaWriteToLog(IntToStr(FLockDelayTime)+'     FLockDelay is on '+', SessionID = '+IntToStr(SessionID)+#13#10);
{$ENDIF}
       end;
    end
   else
    begin
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
 aaWriteToLog('');
{$ENDIF}
      FLockDelayOn := False;
    end;
end; // LockTableIfLockAllowed


//------------------------------------------------------------------------------
// unlock table if possible
//------------------------------------------------------------------------------
function TACRTableLocksFile.UnlockTableIfPossible(SessionID: TACRSessionID; LockType: TACRLockType): Boolean;
begin
  if (LockType = ltX) then
   Result := FHandle.UnlockFile
  else
   Result := FHandle.UnlockByte(GetByteNo(SessionID,LockType));
{$IFDEF DEBUG_TRACE_TABLE_LOCKS}
if (Result) then
 aaWriteToLog(aaGetCurrentTimeAsString+'   TableLocksFile unlock OK, type = '+IntToStr(Integer(LockType))+', ID = '+IntToStr(SessionID)+#13#10)
else
 aaWriteToLog(aaGetCurrentTimeAsString+'   TableLocksFile unlock failed, type = '+IntToStr(Integer(LockType))+', ID = '+IntToStr(SessionID)+#13#10);
{$ENDIF}
end; // UnlockTableIfPossible


//------------------------------------------------------------------------------
// lock record if possible
//------------------------------------------------------------------------------
function TACRTableLocksFile.LockRecordIfPossible(SessionID: TACRSessionID; RecordID: TACRRecordID): Boolean;
begin
  Result := TryUsingTimeOut(FHandle.LockFile, LockParams);
  try
    if (Result) then
     begin
       Result := FHandle.LockByte(GetByteNo(SessionID,ltU));
       if (Result) then
           FHandle.DirectWriteBuffer(RecordID,SizeOf(RecordID),
             GetByteNo(SessionID,ltU,True));
     end;
  finally
    if (not FHandle.UnlockFile) then
     raise EACRException.Create(10771, ErrorLCannotUnlockTableLocksFile);
  end;
end; // LockRecordIfPossible


//------------------------------------------------------------------------------
// unlock record if possible
//------------------------------------------------------------------------------
function TACRTableLocksFile.UnlockRecordIfPossible(SessionID: TACRSessionID; RecordID: TACRRecordID): Boolean;
begin
 Result := TryUsingTimeOut(FHandle.LockFile, LockParams);
 try
   if (Result) then
     Result := FHandle.UnLockByte(GetByteNo(SessionID,ltU));
 finally
    if (not FHandle.UnlockFile) then
     raise EACRException.Create(10772, ErrorLCannotUnlockTableLocksFile);
 end;
end; // UnlockRecordIfPossible


//------------------------------------------------------------------------------
// return true if record is locked
//------------------------------------------------------------------------------
function TACRTableLocksFile.CheckIfRecordIsLocked(SessionID: TACRSessionID; RecordID: TACRRecordID): Boolean;
var i:                  Integer;
    LockedRecordID:     TACRRecordID;
    LowCount,HighCount: Integer;
begin
 if (not TryUsingTimeOut(FHandle.LockFile, LockParams)) then
  raise EACRException.Create(10774,ErrorLCannotLockTableLocksFile);
 try
   LowCount := SessionID;
   HighCount := MaxSessionCount - LowCount - 1;
   Result := False;
   if (SessionID > 0) then
     Result := FHandle.IsRegionLocked(GetByteNo(0,ltU),LowCount);
   if (not Result) then
    if (SessionID < MaxSessionCount-1) then
      Result := FHandle.IsRegionLocked(GetByteNo(SessionID+1,ltU),
                                            HighCount);
//     Result := FHandle.IsRegionLocked(GetByteNo(0,ltU),MaxSessionCount-1);
     if (Result) then
      begin
       Result := False;
       for i := 0 to MaxSessionCount - 1 do
        if (i <> SessionID) then
         if (FHandle.IsByteLocked(GetByteNo(i,ltU))) then
          begin
           FHandle.DirectReadBuffer(LockedRecordID,SizeOf(RecordID),
                                    GetByteNo(i,ltU,True));
           if  (RecordID.PageNo = LockedRecordID.PageNo) and
              (RecordID.PageItemNo = LockedRecordID.PageItemNo) then
            begin
             Result := True;
             break;
            end;
          end;
      end;
 finally
    if (not FHandle.UnlockFile) then
     raise EACRException.Create(10773, ErrorLCannotUnlockTableLocksFile);
 end;
end; // CheckIfRecordIsLocked


////////////////////////////////////////////////////////////////////////////////
//
// TACRTablePFSMapFile
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// load file
//------------------------------------------------------------------------------
procedure TACRTablePFSMapFile.LoadFile(SessionID: TACRSessionID);
begin
 FHandle.ReadFile(SessionID);
 if (FHandle.FileData.Size = 0) then
  begin
   TablePFSPageMapItems := nil;
   TablePFSPageCount := 0;
   PageCount := 0;
  end // empty file
 else
  begin
   Move(FHandle.FFileData.Buffer^, PageCount, Sizeof(PageCount));
   TablePFSPageCount := (FHandle.FileData.Size - Sizeof(PageCount))
                        div sizeof(TACRTablePFSPageMapItem);
   SetLength(TablePFSPageMapItems,TablePFSPageCount);
   Move(PChar(FHandle.FFileData.Buffer + Sizeof(PageCount))^,
    TablePFSPageMapItems[0],
    TablePFSPageCount * sizeof(TACRTablePFSPageMapItem));
  end; // not empty file
end; // LoadFile


//------------------------------------------------------------------------------
// Save file
//------------------------------------------------------------------------------
procedure TACRTablePFSMapFile.SaveFile(SessionID: TACRSessionID);
begin
 if (TablePFSPageCount = 0) then
  begin
   FHandle.EmptyFile(SessionID);
  end // empty file
 else
  begin
   FHandle.ResizeFileData(Sizeof(PageCount) +
                          TablePFSPageCount * sizeof(TACRTablePFSPageMapItem));
   Move(PageCount,FHandle.FFileData.Buffer^,Sizeof(PageCount));
   Move(TablePFSPageMapItems[0],
        PChar(FHandle.FFileData.Buffer + Sizeof(PageCount))^,
        FHandle.FFileData.Size - SizeOf(PageCount));
   FHandle.WriteFile(SessionID);
  end; // not empty file
end; // SaveFile


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRTablePFSMapFile.Create(
                       aPageManager:              TACRPageManager;
                       IsSynchronizationRequired: TACRIsPageSynchronizationRequired
                                      );

begin
  LPageManager := TACRDiskPageManager(aPageManager);
  FHandle := TACRInternalDBFile.Create(LPageManager, ACRPageTypeIDTablePFSMap, IsSynchronizationRequired);
end; // Create;


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRTablePFSMapFile.Destroy;
begin
  TablePFSPageMapItems := nil;
  if (FHandle <> nil) then
   begin
    FHandle.Free;
    FHandle := nil;
   end;
  inherited;
end; // Destroy;


//------------------------------------------------------------------------------
// CreateFile
//------------------------------------------------------------------------------
function TACRTablePFSMapFile.CreateFile(ObjectID: TACRObjectID): TACRPageNo;
begin
  FHandle.FStartPageNo := -1;
  FHandle.CreateFile(SYSTEM_SESSION_ID,ObjectID);
  Result := FHandle.StartPageNo;
end;// CreateFile


//------------------------------------------------------------------------------
// OpenFile
//------------------------------------------------------------------------------
procedure TACRTablePFSMapFile.OpenFile(aStartPageNo: TACRPageNo; OpenExclusive: Boolean);
begin
  FHandle.OpenFile(aStartPageNo,OpenExclusive);
end;// OpenFile


//------------------------------------------------------------------------------
// EmptyFile
//------------------------------------------------------------------------------
procedure TACRTablePFSMapFile.EmptyFile(SessionID: TACRSessionID);
begin
 FHandle.EmptyFile(SessionID);
 TablePFSPageMapItems := nil;
 TablePFSPageCount := 0;
 PageCount := 0;
end;// EmptyFile


//------------------------------------------------------------------------------
// DeleteFile
//------------------------------------------------------------------------------
procedure TACRTablePFSMapFile.DeleteFile(
                      SessionID:    TACRSessionID;
                      aStartPageNo: TACRPageNo
                                              );
begin
  FHandle.DeleteFile(SessionID, aStartPageNo);
end;// DeleteFile


//------------------------------------------------------------------------------
// add PFS page
//------------------------------------------------------------------------------
procedure TACRTablePFSMapFile.AddPage(SessionID: TACRSessionID; PageNo: TACRPageNo);
var x: TACRPageNo;
    i: Integer;
begin
 x := INVALID_PAGE_NO;
 if (TablePFSPageCount < 0) then
  raise EACRException.Create(10703,ErrorLInvalidPageNo,[TablePFSPageCount]);
 for i := 0 to TablePFSPageCount-1 do
  if (TablePFSPageMapItems[i].PageNo = PageNo) then
   raise EACRException.Create(10480,ErrorLCannotAddTablePFSPage,
          [TablePFSPageCount,i,TablePFSPageMapItems[i].PageNo,PageNo])
  else
   if (TablePFSPageMapItems[i].PageNo > PageNo) then
    begin
     x := i;
     break;
    end;
 Inc(TablePFSPageCount);
 if (x = INVALID_PAGE_NO) then
  begin
   x := TablePFSPageCount - 1;
   SetLength(TablePFSPageMapItems,TablePFSPageCount);
  end
 else
  begin
   SetLength(TablePFSPageMapItems,TablePFSPageCount);
   Move(TablePFSPageMapItems[x],TablePFSPageMapItems[x+1],
    (TablePFSPageCount - x) * SizeOf(TACRTablePFSPageMapItem));
  end;
 TablePFSPageMapItems[x].PageNo := PageNo;
 TablePFSPageMapItems[x].PageRecordCount := 0;
end; // AddPage


//------------------------------------------------------------------------------
// return number of records stored on all pages
//------------------------------------------------------------------------------
function TACRTablePFSMapFile.GetTableRecordCount: TACRRecordNo;
var
    i: Integer;
begin
 Result := 0;
 for i := 0 to TablePFSPageCount-1 do
  Inc(Result,TablePFSPageMapItems[i].PageRecordCount);
end; // GetTableRecordCount


//------------------------------------------------------------------------------
// save state
//------------------------------------------------------------------------------
function TACRTablePFSMapFile.SaveState: Pointer;
var State:  PACRTablePFSMapState;
    l:      Integer;
begin
  New(State);
  State^.PageCount := PageCount;
  State^.TablePFSPageCount := TablePFSPageCount;
  l := Length(TablePFSPageMapItems);
  SetLength(State^.TablePFSPageMapItems,l);
  if (l > 0) then
   Move(TablePFSPageMapItems[0],State^.TablePFSPageMapItems[0],
        l * SizeOf(TACRTablePFSPageMapItem));
  Result := State;      
end; // SaveState


//------------------------------------------------------------------------------
// restore state
//------------------------------------------------------------------------------
procedure TACRTablePFSMapFile.RestoreState(State: Pointer);
var
    l:      Integer;
begin
  if (State = nil) then
   raise EACRException.Create(10840,ErrorLNilPointer);
  PageCount := PACRTablePFSMapState(State)^.PageCount;
  TablePFSPageCount := PACRTablePFSMapState(State)^.TablePFSPageCount;
  l := Length(PACRTablePFSMapState(State)^.TablePFSPageMapItems);
  SetLength(TablePFSPageMapItems,l);
  if (l > 0) then
   Move(PACRTablePFSMapState(State)^.TablePFSPageMapItems[0],
        TablePFSPageMapItems[0],
        l * SizeOf(TACRTablePFSPageMapItem));
end; // RestoreState


//------------------------------------------------------------------------------
// free state
//------------------------------------------------------------------------------
procedure TACRTablePFSMapFile.FreeState(var State: Pointer);
begin
  if (State <> nil) then
   begin
    SetLength(PACRTablePFSMapState(State)^.TablePFSPageMapItems,0);
    PACRTablePFSMapState(State)^.TablePFSPageMapItems := nil;
    Dispose(PACRTablePFSMapState(State));
    State := nil;
   end;
end; // FreeState


////////////////////////////////////////////////////////////////////////////////
//
// TACRSystemDirectory
//
////////////////////////////////////////////////////////////////////////////////



//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSystemDirectory.Create(PageManager: TACRDiskPageManager);
begin
  LPageManager := PageManager;
  FHandle := TACRInternalDBFile.Create(LPageManager, ACRPageTypeIDFileSystemDirectory);
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRSystemDirectory.Destroy;
begin
  FreeAndNil(FHandle);
end;//Destroy


//------------------------------------------------------------------------------
// Create SystemDirectory
//------------------------------------------------------------------------------
procedure TACRSystemDirectory.CreateDirectory;
begin
  FHandle.CreateFile(SYSTEM_SESSION_ID,ACRObjectIDSystemDirectory);
  if (FHandle.StartPageNo <> ACRFirstPageNoSystemDirectory) then
    raise EACRException.Create(30392, ErrorGSystemDirInvalidFirstPageNo,
                               [FHandle.StartPageNo, ACRFirstPageNoSystemDirectory]);
end;//CreateDirectory


//------------------------------------------------------------------------------
// Load SystemDirectory
//------------------------------------------------------------------------------
procedure TACRSystemDirectory.LoadDirectory;
var
    Count:  Integer;
begin
  FHandle.OpenFile(ACRFirstPageNoSystemDirectory,True);
  FHandle.ReadFile(SYSTEM_SESSION_ID);
  Count := FHandle.FileData.Size div SizeOf(TACRSystemDirectoryListItem);
  if (Count < ACRMinSystemFilesCount) then
   raise EACRException.Create(10633,ErrorLInvalidSystemDirectoryFilesCount,
    [Count,ACRFirstPageNoSystemDirectory]);
  SetLength(FFileList, count);
  Move(FHandle.FileData.Buffer^, FFileList[0], FHandle.FileData.Size);
end;//LoadDirectory


//------------------------------------------------------------------------------
// Save SystemDirectory
//------------------------------------------------------------------------------
procedure TACRSystemDirectory.SaveDirectory;
begin
 FHandle.ResizeFileData(Length(FFileList)*SizeOf(TACRSystemDirectoryListItem));
 Move(FFileList[0],FHandle.FFileData.Buffer^,FHandle.FileData.Size);
 FHandle.WriteFile(SYSTEM_SESSION_ID);
end;//SaveDirectory


//------------------------------------------------------------------------------
// CreateFile
//------------------------------------------------------------------------------
procedure TACRSystemDirectory.CreateFile(FileType: TACRDBFileType; FirstPageNo: TACRPageNo);
var
  Count: Integer;
begin
  Count := Length(FFileList);
  SetLength(FFileList, count+1);
  FFileList[count].FileID := FileType;
  FFileList[count].FirstPageNo := FirstPageNo;
end;//CreateFile


//------------------------------------------------------------------------------
// GetFileFirstPageNo
//------------------------------------------------------------------------------
function TACRSystemDirectory.GetFileFirstPageNo(FileType: TACRDBFileType): TACRPageNo;
var
  i: Integer;
begin
  Result := INVALID_PAGE_NO;
  for i:= 0 to Length(FFileList)-1 do
   if (FFileList[i].FileID = FileType) then
    begin
     Result := FFileList[i].FirstPageNo;
     break;
    end;
end;//GetFileFirstPageNo



//------------------------------------------------------------------------------
// convert PageTypeID to String (for exceptions messages)
//------------------------------------------------------------------------------
function PageTypeToStr(PageType: TACRPageTypeID): String;
begin
  if PageType > High(ACRPageTypeNames) then
      raise EACRException.Create(30382, ErrorGUnknownPageType, [PageType]);
  Result := ACRPageTypeNames[PageType];
end;//PageTypeToStr


//------------------------------------------------------------------------------
// retrun delay in ms
//------------------------------------------------------------------------------
function GetDelay(const Delay: Integer; const CurrentRetryNo, RetryCount: Integer): Integer;
begin
  Result := Delay;
end; // GetDelay


//------------------------------------------------------------------------------
// Try call Func before Success or TimeOut off
//------------------------------------------------------------------------------
function TryUsingTimeOut(
                           Func:        TBooleanFunctionForTimeOutCall;
                           LockParams:  TACRLockParams
                         ): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 1 to LockParams.RetryCount do
    if Func then
      begin
       Result := True;
       break;
      end
    else
      Sleep(GetDelay(LockParams.Delay,i,LockParams.RetryCount));
end;//TryUsingTimeOut


//------------------------------------------------------------------------------
// Try call Func before Success or TimeOut off
//------------------------------------------------------------------------------
function TryUsingTimeOutForTableLocks(
                           Func:        TBooleanFunctionForTimeOutCallWithRetryNo;
                           LockParams:  TACRLockParams
                         ): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 1 to LockParams.RetryCount do
    if Func(i) then
      begin
       Result := True;
       break;
      end
    else
      Sleep(GetDelay(LockParams.Delay,i,LockParams.RetryCount));
end;//TryUsingTimeOutForTableLocks

initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRDiskEngine> initialized');
{$ENDIF}

end.
