//------------------------------------------------------------------------------
//
// Local and File-Server Engine
// Server uses Local Engine
//
//------------------------------------------------------------------------------

unit ACRBaseEngine;

{$I ACRVer.inc}

interface


uses SysUtils,
     Classes,
{$IFDEF MSWINDOWS}
     Controls,
     Windows,
{$ENDIF}
{$IFDEF LINUX}
     Libc,
{$ENDIF}
     Math,

// Accuracer units

     ACRCriticalSection,
 {$IFDEF DEBUG_LOG}
     ACRDebug,
 {$ENDIF}
     ACRPage,
     ACRExcept,
     ACRBase,
     ACRMemory,
     ACRCompression,
{$IFNDEF SQLMEMTABLE}
     ACRCrypto,
{$ENDIF}
     ACRTypes,
     ACRConverts,
     ACRVariant,
     ACRConst,
     ACRExpressions
     ;

{$IFDEF LINUX}
  // Windows.pas
const
  MAXWORD = 65535;
{$ENDIF}

type

  TACRDatabaseData = class;
  TACRTableData = class;
  TACRIndex = class;
  TACRBaseSequenceManager = class;
  TACRRecordBitmap = class;
  TACRTransaction = class;

////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseRecordManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRBaseRecordManager = class (TObject)
   protected
    FCSect:                 TRTLCriticalSection;
    FRecordBufferSize:      Integer;
    FRecordCount:           TACRRecordNo;
    FRepair:                Boolean;

   public
    // lock
    procedure Lock;
    // unlock
    procedure Unlock;
    function GetRecordCount: TACRRecordNo; virtual;
    procedure Empty(SessionID: TACRSessionID = INVALID_SESSION_ID); virtual; abstract;
    // add record and return its number
    function AddRecord(
                       RecordBuffer:  TACRRecordBuffer;
                       var RecordID:  TACRRecordID;
                       SessionID:     TACRSessionID = INVALID_SESSION_ID
                      ): Boolean; virtual; abstract;
    // update record, return true if record was updated, false if record was deleted
    function UpdateRecord(
                          RecordBuffer: TACRRecordBuffer;
                          RecordID:     TACRRecordID;
                          SessionID:    TACRSessionID = INVALID_SESSION_ID
                         ): Boolean; virtual;
    // delete record, return true if record was deleted, false if record was deleted earlier
    function DeleteRecord(
                          var RecordID: TACRRecordID;
                          SessionID:    TACRSessionID = INVALID_SESSION_ID
                         ): Boolean; virtual;
    // return true if record exists
    function IsRecordExists(var RecordID: TACRRecordID; SessionID: TACRSessionID = INVALID_SESSION_ID): Boolean; virtual; abstract;
    // get record using physical order
    procedure GetRecordBuffer(var NavigationInfo: TACRNavigationInfo); virtual; abstract;
    // return record no
    function GetApproximateRecNo(RecordID: TACRRecordID; SessionID: TACRSessionID): TACRRecordNo; virtual; abstract;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    // add loaded record
    procedure AddLoadedRecord(RecordBuffer: TACRRecordBuffer; var RecordPos: Integer); virtual;
   public
    property RecordBufferSize: Integer read FRecordBufferSize write FRecordBufferSize;
    property Repair: Boolean read FRepair write FRepair;
  end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseFieldManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRBaseFieldManager = class (TObject)
   private
    FFieldDefs:       TACRFieldDefs;
    FTableData:       TACRTableData;
    FSequenceManager: TACRBaseSequenceManager;
    FEngineVersion:   Double;
   public
    constructor Create(TableData: TACRTableData; SequenceManager: TACRBaseSequenceManager);
    destructor Destroy; override;

    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToStream(Stream: TStream); virtual;

    // Set default values to fields
    procedure ApplyDefaultValuesToRecordBuffer(
                                Session:      TACRBaseSession;
                                RecordBuffer: TACRRecordBuffer;
                                InsertMode:   Boolean
                                );
   public
    property FieldDefs: TACRFieldDefs read FFieldDefs;
    property EngineVersion: Double read FEngineVersion write FEngineVersion;
  end; // TACRBaseFieldManager



////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseIndexManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRBaseIndexManager = class (TObject)
   protected
    FTableData:             TACRTableData;
    FIndexDefs:             TACRIndexDefs;
    FOpenIndexList:         TList; // list of TACRIndex objects
    FPageManager:           TACRPageManager;
    FTemporaryPageManager:  TACRPageManager;

    procedure InternalCreateIndex(
            Cursor: TACRCursor;
            IndexDef: TACRIndexDef
                                  );
    function InternalOpenIndex(
            IndexID: TACRObjectID
                               ): TACRIndex;

   public
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToStream(Stream: TStream);  virtual;
    constructor Create(aTableData: TACRTableData);
    destructor Destroy; override;

    function GetTemporaryPageManager: TACRPageManager; virtual;
    function GetPageManager(Index: TACRIndex): TACRPageManager;
    procedure DropTemporaryIndexes(SessionID: TACRSessionID);
    function IsIndexExists(FieldNames, AscDescList, CaseSensitivityList: TStringList): Boolean;
    function FindIndex(FieldNames, AscDescList, CaseSensitivityList: TStringList): TACRObjectID;
    function CreateTemporaryIndex(
                    Cursor:         TACRCursor;
                    FieldNamesList, AscDescList, CaseSensitivityList: TStringList
                                 ): TACRObjectID;
    function FindOrCreateIndex(Cursor: TACRCursor; FieldNamesList, AscDescList, CaseSensitivityList: TStringList): TACRObjectID;

    // create index definitions list
    procedure CreateIndexDefs(aIndexDefs: TACRIndexDefs); virtual;

    function FindOpenIndex(IndexID: TACRObjectID): TACRIndex;
    function CreateIndex(
                         Cursor:    TACRCursor;
                         IndexDef:  TACRIndexDef
                        ): TACRObjectID;
    function OpenIndex(IndexID: TACRObjectID): TACRIndex;
    procedure CloseIndex(IndexID: TACRObjectID);
    procedure DropIndex(SessionID: TACRSessionID; IndexID: TACRObjectID);
    procedure DropAllIndexes(SessionID: TACRSessionID);
    procedure EmptyIndex(SessionID: TACRSessionID; IndexID: TACRObjectID);
    procedure EmptyAllIndexes(SessionID: TACRSessionID);

    procedure GetRecordBuffer(
                               SessionID: TACRSessionID;
                               var NavigationInfo: TACRNavigationInfo
                             );
    procedure InsertRecord(Cursor: TACRCursor);
    procedure UpdateRecord(Cursor: TACRCursor);
    procedure DeleteRecord(Cursor: TACRCursor);
    // return 0 if record buffers are equal in this index
    // return 1 if Buffer1 is higher than Buffer 2 (Pos1 > Pos2)
    // return -1 if Buffer1 is lower than Buffer 2 (Pos1 < Pos2)
    // IndexFieldCount = 0 means to compare by all fields in index
    function CompareRecordBuffersByIndex(
                        IndexID: TACRObjectID;
                        Buffer1: TACRRecordBuffer;
                        Buffer2: TACRRecordBuffer;
                        IndexFieldCount: Integer = 0
                                        ): Integer;
   public
    property IndexDefs: TACRIndexDefs read FIndexDefs;
    property TableData: TACRTableData read FTableData;
    property PageManager: TACRPageManager read FPageManager;
  end; // TACRBaseIndexManager


////////////////////////////////////////////////////////////////////////////////
//
// TACRIndex
//
////////////////////////////////////////////////////////////////////////////////

  TACRIndex = class (TObject)
   private
    FIndexDef:      TACRIndexDef;
    FIndexManager:  TACRBaseIndexManager;
    FTemporary:     Boolean;
   public
    constructor Create(aIndexManager: TACRBaseIndexManager);
    destructor Destroy; override;
    procedure CreateIndex(Cursor: TACRCursor; aIndexDef: TACRIndexDef); virtual;
    procedure DropIndex(SessionID: TACRSessionID); virtual; abstract;
    procedure OpenIndex(aIndexDef: TACRIndexDef); virtual;

    procedure GetRecordBuffer(SessionID: TACRSessionID; var NavigationInfo: TACRNavigationInfo); virtual; abstract;
    function CreateIndexPosition: TACRIndexPosition; virtual; abstract;
    procedure FreeIndexPosition(var IndexPosition: TACRIndexPosition); virtual; abstract;
    function GetIndexPosition(
                               SessionID:      TACRSessionID;
                               RecordID:       TACRRecordID;
                               RecordBuffer:   TACRRecordBuffer;
                               IndexPosition:  TACRIndexPosition
                             ): Boolean; virtual; abstract;
    // return 0, 1, -1 if (Pos1 = Pos2), (Pos1 > Pos2), (Pos1 < Pos2)
    function CompareRecordPositionsInIndex(
                        RecordPosition1: TACRIndexPosition;
                        RecordPosition2: TACRIndexPosition
                                          ): Integer; virtual; abstract;
    function GetRecNoByRecordID(
                                SessionID:      TACRSessionID;
                                RecordID:       TACRRecordID;
                                RecordBuffer:   TACRRecordBuffer;
                                Bitmap:         TACRRecordBitmap
                               ): TACRRecordNo; virtual; abstract;
    function GetRecordIDByRecNo(
                                SessionID:      TACRSessionID;
                                RecNo:          TACRRecordNo;
                                Bitmap:         TACRRecordBitmap
                               ): TACRRecordID; virtual; abstract;
    procedure ApplyDistinctToRecordBitmap(
                                SessionID:      TACRSessionID;
                                Bitmap:         TACRRecordBitmap
                                         ); virtual; abstract;
    function CreateSearchInfo: TACRSearchInfo; virtual; abstract;
    procedure FreeSearchInfo(SearchInfo: TACRSearchInfo); virtual; abstract;
    function FindRecord(
                       SessionID:           TACRSessionID;
                       Restart:             Boolean;
                       GoForward:           Boolean;
                       StartScanCondition:  TACRScanSearchCondition;
                       EndScanCondition:    TACRScanSearchCondition;
                       RecordBuffer:        TACRRecordBuffer;
                       var RecordID:        TACRRecordID;
                       SearchInfo:          TACRSearchInfo
                      ): Boolean; virtual; abstract;
    // return 0 if record buffers are equal in this index
    // return 1 if Buffer1 is higher than Buffer 2 (Pos1 > Pos2)
    // return -1 if Buffer1 is lower than Buffer 2 (Pos1 < Pos2)
    function CompareRecordBuffersByIndex(
                        Buffer1: TACRRecordBuffer;
                        Buffer2: TACRRecordBuffer;
                        IndexFieldCount: Integer
                                        ): Integer; virtual; abstract;

    // return 0 if conditions are equal in this index
    // return 1 if Condition1 is higher than Condition2
    // return -1 if Condition1 is lower than Condition2
    function CompareConditions(
                    Condition1:   TACRScanSearchCondition;
                    Condition2:   TACRScanSearchCondition
                              ): Integer; virtual; abstract;
    // approximate record count berween range conditions
    function GetApproxRangeRecordCount(
                    SessionID:         TACRSessionID;
                    TableRecordCount:  TACRRecordNo;
                    RangeCondition1:   TACRScanSearchCondition;
                    RangeCondition2:   TACRScanSearchCondition
                                      ): TACRRecordNo; virtual; abstract;
    function CanInsertRecord(
                    SessionID:      TACRSessionID;
                    RecordBuffer:   TACRRecordBuffer
                            ): Boolean; virtual; abstract;
    function CanUpdateRecord(
                    SessionID:      TACRSessionID;
                    OldRecordBuffer, NewRecordBuffer: TACRRecordBuffer
                            ): Boolean; virtual; abstract;
    procedure InsertRecord(Cursor: TACRCursor); virtual; abstract;
    procedure DeleteRecord(Cursor: TACRCursor); virtual; abstract;
    procedure UpdateRecord(Cursor: TACRCursor); virtual; abstract;

    property IndexDef: TACRIndexDef read FIndexDef;
    property IndexManager: TACRBaseIndexManager read FIndexManager;
    property Temporary: Boolean read FTemporary;
  end; // TACRIndex




////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseSequenceManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRBaseSequenceManager = class (TObject)
   private
    FSequenceDefs:    TACRSequenceDefs;
    FDatabaseData:    TACRDatabaseData;
   public
    // constructor
    constructor Create(DatabaseData: TACRDatabaseData);
    // destructor
    destructor Destroy; override;

    // find or create SequenceDef by SequenceName
    function GetOrCreateSequenceDefByName(SequenceName: String): TACRSequenceDef;

    // returns sequence definition
    function GetSequenceDef(SequenceID: TACRObjectID): TACRSequenceDef; virtual;
    // get next sequence value
    function GetNextVal(Session:TACRBaseSession; SequenceID: TACRObjectID): TACRSequenceValue; virtual;
    // get last sequence value
    function GetLastVal(Session:TACRBaseSession; SequenceID: TACRObjectID): TACRSequenceValue; virtual;
    // Generate unic Sequence Name
    function GenerateSequenceName(isAutoinc: boolean; ColumnName: String; TableName: String = ''): String;
   public
    property SequenceDefs: TACRSequenceDefs read FSequenceDefs write FSequenceDefs;
  end; // TACRBaseSequenceManager



////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseConstraintManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRBaseConstraintManager = class (TObject)
   private
    FConstraintDefs:    TACRConstraintDefs;
    FTableData:         TACRTableData;
   public
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToStream(Stream: TStream);
    constructor Create(aTableData: TACRTableData); //virtual;
    destructor Destroy; override;

    // check constraint conditions
    procedure CheckConstraints(
                                Cursor:          TACRCursor;
                                SessionID:       TACRSessionID;
                                NewRecordBuffer: TACRRecordBuffer;
                                OldRecordBuffer: TACRRecordBuffer;
                                ToInsert:        Boolean;
                                CurrentRecordID: TACRRecordID;
                                SkipFKCheck:     Boolean = False
                              );

    // link object id
    procedure LinkObjectId(ConstraintDef: TACRConstraintDef);
    // link object ids between constraints and fields
    procedure LinkObjectIds;
    // generate Constraint Names for empty constraints
    procedure FillConstraintAutoNames;

    // add constraint for primary or Unique index
    procedure AddConstraintFromIndex(IndexDef: TACRIndexDef);
    // delete Constraint Linked with deleted index
    procedure DeleteConstraintForIndexID(IndexObjectID: TACRObjectID);
    // add action to master table
    function AddForeignKeyAction(
                                  ConstraintDef:           TACRConstraintDefForeignKey;
                                  ReferencedTableName:     String;
                                  ReferencedTableObjectID: TACRObjectID
                                ): TACRConstraintDefForeignKeyAction;

   public
    property ConstraintDefs: TACRConstraintDefs read FConstraintDefs write FConstraintDefs;

  end; // TACRBaseConstraintManager


////////////////////////////////////////////////////////////////////////////////
//
// TACRRecordBitmap
//
////////////////////////////////////////////////////////////////////////////////


  TACRRecordBitmap = class (TObject)
  private
   FVisibleRecords: TACRRecordIDArray;
   FTableData:      TACRTableData;
   FActive:         Boolean;
   FIndexed:        Boolean; // is index applied
   FDistinct:       Boolean; // for distinct without filter on disk tables
  protected
   procedure SetActive(Value: Boolean);
   procedure SetIndexed(Value: Boolean);
  public
   constructor Create(aTableData: TACRTableData);
   destructor Destroy; override;
   function GetRecordCount: TACRRecordNo;
   function GetRecNoByRecordID(RecordID: TACRRecordID): TACRRecordNo;
   function GetRecordIDByRecNo(RecNo: TACRRecordNo): TACRRecordID;
   function IsRecordVisible(RecordID: TACRRecordID): Boolean;
   procedure InsertVisibleRecord(
                                 var   RecordID:      TACRRecordID;
                                 const PriorRecordID: TACRRecordID;
                                 bFirst:              Boolean;
                                 bLast:                 Boolean
                                );
   procedure ShowRecord(RecordID: TACRRecordID);
   procedure HideRecord(RecordID: TACRRecordID);
   procedure ClearVisibleRecords;
  protected
   procedure GetRecordFromFirstPosition(var NavigationInfo: TACRNavigationInfo);
   procedure GetRecordFromLastPosition(var NavigationInfo: TACRNavigationInfo);
   procedure GetRecordFromAnyPosition(var NavigationInfo: TACRNavigationInfo);
  public
   procedure GetRecord(var NavigationInfo: TACRNavigationInfo);
   procedure PrepareBitmapForActivation;
  public
   property TableData: TACRTableData read FTableData;
   property Active: Boolean read FActive write SetActive;
   property Distinct: Boolean read FDistinct write FDistinct;
   property Indexed: Boolean read FIndexed write SetIndexed;
   property VisibleRecords: TACRRecordIDArray read FVisibleRecords;
  end; // TACRRecordBitmap


////////////////////////////////////////////////////////////////////////////////
//
// TACRTableData
//
////////////////////////////////////////////////////////////////////////////////


  TACRTableData = class (TObject)
  private
    FCSect:                        TRTLCriticalSection;
  protected
    FRepair:                       Boolean;
    FSessionID:                    TACRSessionID;
    FDatabaseData:                 TACRDatabaseData;
    FTableName:                    String;
    FBLOBCompression:              TACRCompression;
    FActive:                       Boolean;
    FFieldManager:                 TACRBaseFieldManager;
    FConstraintManager:            TACRBaseConstraintManager;
    FIndexManager:                 TACRBaseIndexManager;
    FRecordManager:                TACRBaseRecordManager;
    FSequenceManager:              TACRBaseSequenceManager;
    FCursorList:                   TList;
    FBLOBFieldsPresent:            Boolean;
    FInMemory:                     Boolean;
    FTemporary:                    Boolean;
    FLastDBState:                  Cardinal;
    FLastDBStateOnLockTable:       Cardinal;
    FNumLocksForRead:              Integer;
    FCachedRecordCount:            TACRRecordNo; // cached number of records for LockTableForRead
    FTempRecordBuffer:             TACRRecordBuffer;

    function GetPageManager: TACRPageManager; virtual; abstract;
    function GetTableID: TACRObjectID; virtual;
    // Fill New ObjectID for all defs
    procedure FillDefsByObjectId(Defs: TACRMetaObjectDefs);
    function GetRecordBufferSize: Integer;
    procedure CreateRecordManager(TablePFSMapFileStartPageNo: TACRPageNo = INVALID_PAGE_NO); virtual; abstract;
    procedure CreateFieldManager(FieldDefs: TACRFieldDefs);
    procedure CreateIndexManager(IndexDefs: TACRIndexDefs); virtual;
    procedure CreateConstraintManager(ConstraintDefs: TACRConstraintDefs);
    procedure CreateSequenceManager;
    procedure BuildSequences;
    procedure LoadSequencesFromStream(Stream: TStream);
    procedure SaveSequencesToStream(Stream: TStream);
    procedure InitCursor(Cursor: TACRCursor); virtual;
  public
    constructor Create(aDatabaseData: TACRDatabaseData);
    destructor Destroy; override;
    // lock
    procedure Lock;
    // unlock
    procedure Unlock;

    // table operations
    procedure CreateTable(
                          Cursor: TACRCursor;
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                         ); virtual; abstract;
    procedure DeleteTable(Session: TACRBaseSession; Cascade: Boolean; DesignMode: Boolean = False); virtual;
    procedure EmptyTable(Cursor: TACRCursor); virtual;
    procedure RenameTable(Cursor: TACRCursor; NewTableName: String); virtual;
    procedure AddForeignKey(Cursor: TACRCursor; ConstraintDef: TACRConstraintDefForeignKey); virtual;
    procedure DeleteConstraint(Cursor: TACRCursor; Name: String; Cascade: Boolean; FKPartialDelete: Boolean); virtual;
    procedure RenameReferenceTableName(Cursor: TACRCursor; OldName, NewName: String); virtual;

    procedure LoadTableFromStream(
                        Stream:               TStream
                       ); virtual;
    procedure SaveTableToStream(
                        Stream:               TStream;
                        CompressionAlgorithm: TACRCompressionAlgorithm = acaNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      ); virtual;
    procedure OpenTable(Cursor: TACRCursor); virtual;
    procedure CloseTable(Cursor: TACRCursor); virtual;

    // Rename Field by Field Index in FieldDefs
    procedure RenameField(Cursor: TACRCursor; FieldName, NewFieldName: String); virtual;

    procedure AddIndex(IndexDef: TACRIndexDef; Cursor: TACRCursor); virtual;
    procedure DeleteIndex(IndexID: TACRObjectID; Cursor: TACRCursor); virtual;
    procedure EmptyIndex(IndexID: TACRObjectID; SessionID: TACRSessionID); virtual;
    procedure DeleteAllIndexes(Cursor: TACRCursor); virtual;
    procedure EmptyAllIndexes(SessionID: TACRSessionID); virtual;
    // return index name of the index or '' if not found
    function FindIndex(
                                Cursor: TACRCursor;
                                FieldNamesList,
                                AscDescList,
                                CaseSensitivityList: TStringList
                              ): String;
    // return index name
    function FindOrCreateIndex(
                                Cursor: TACRCursor;
                                FieldNamesList,
                                AscDescList,
                                CaseSensitivityList: TStringList
                              ): String;

    // return true if Unique Constraint Failed
    function IsUniqueConstraintFailed(
                                        SessionID:        TACRSessionID;
                                        IndexID:          TACRObjectID;
                                        NewRecordBuffer:  TACRRecordBuffer;
                                        OldRecordBuffer:  TACRRecordBuffer;
                                        ToInsert:         Boolean;
                                        CurrentRecordID:  TACRRecordID
                                     ): Boolean;
   protected
    procedure RenameConstraintFKTemporaryNames(ConstraintDef: TACRConstraintDefForeignKeyAction);
    procedure CheckRecordsCompatibleWithForeignKey(Cursor: TACRCursor; ReferencedCursor: TACRCursor; ConstraintDef: TACRConstraintDefForeignKey);
    // return true if current record in Cursor does not violate foreign key constraint
    // (have null field values corresponding match type)
    function IsForeignKeyNullValuesOK(
                                        Cursor:         TACRCursor;
                                        ConstraintDef:  TACRConstraintDefForeignKey;
                                        var Failed:     Boolean
                                      ): Boolean;
    procedure CheckForeignKeyOpenReferencedTable(
                                        ReferencedCursor: TACRCursor;
                                        Cursor:           TACRCursor;
                                        ConstraintDef:    TACRConstraintDefForeignKeyAction;
                                        Exclusive:        Boolean = False
                                          ); overload;
    procedure CheckForeignKeyOpenReferencedTable(
                                        ReferencedCursor: TACRCursor;
                                        Session:          TACRBaseSession;
                                        InMemory:         Boolean;
                                        ConstraintDef:    TACRConstraintDefForeignKeyAction;
                                        Exclusive:        Boolean = False
                                          ); overload;
    procedure CheckForeignKeyCloseReferencedTable(Cursor: TACRCursor);
    function CheckForeignKeyBuildSearchExpression(
                                        ReferencedCursor: TACRCursor;
                                        Cursor:           TACRCursor;
                                        ConstraintDef:    TACRConstraintDefForeignKey
                                          ): TACRExpression;
    function CheckForeignKeyActionBuildSearchExpression(
                                        ReferencedCursor: TACRCursor;
                                        Cursor:           TACRCursor;
                                        ConstraintDef:    TACRConstraintDefForeignKeyAction;
                                        ToUpdate:         Boolean
                                          ): Boolean;
    procedure CheckFieldDefinitions(FieldDefs: TACRFieldDefs);
    procedure CheckIndexDefinitions(IndexDefs: TACRIndexDefs);
    procedure CheckConstraintDefinitions(ConstraintDefs: TACRConstraintDefs);
    function CreateForeignKeyAction(
                                  ConstraintDef:           TACRConstraintDefForeignKey;
                                  ReferencedTableName:     String;
                                  ReferencedTableObjectID: TACRObjectID
                                   ): TACRConstraintDefForeignKeyAction;
    procedure WriteMetadataChanges(SessionID: TACRSessionID); virtual;
    procedure CheckForeignKeyDefinition(ConstraintDef: TACRConstraintDefForeignKey; ReferencedCursor: TACRCursor);
    procedure CheckForeignKeyDefinitionsAndCreateForeignKeyActions(Cursor: TACRCursor);
    procedure RenameTableInForeignKeys(
                                       Cursor:       TACRCursor;
                                       OldTableName: String;
                                       NewTableName: String
                                      );
    procedure ExecuteForeignKeyActionsUpdateDetailRecords(
                                        ReferencedCursor: TACRCursor;
                                        Cursor:           TACRCursor;
                                        ConstraintDef:    TACRConstraintDefForeignKeyAction;
                                        Action:           TACRConstraintForeignKeyAction
                                                         );
   public
    // return true if current record in Cursor violates foreign key constraint
    function IsForeignKeyConstraintFailed(
                                        Cursor:         TACRCursor;
                                        ConstraintDef:  TACRConstraintDefForeignKey
                                          ): Boolean;
    procedure ExecuteForeignKeyActions(
                                      Cursor:   TACRCursor;
                                      ToUpdate: Boolean // update or delete
                              );
    //---------------- Search and navigation methods ---------------------------
   protected
    // build record bitmap for cursor
    procedure BuildCursorRecordBitmap(Cursor: TACRCursor);
    function GetBitmapSize(SessionID: TACRSessionID): TACRRecordNo; virtual;
    // return filter bitmap rec no by record id
    function GetBitmapRecNoByRecordID(RecordID: TACRRecordID): TACRRecordNo; virtual; abstract;
    // return filter bitmap rec no by record id
    function GetRecordIDByBitmapRecNo(RecordNo: TACRRecordNo): TACRRecordID; virtual; abstract;
    // return true if record is in specified range
    function IsRecordInRange(Cursor: TACRCursor): Boolean;
    // return true if record is not equal to the previous record
    function IsRecordInDistinct(Cursor: TACRCursor): Boolean;
    // return true if record is visible (with applied filters, ranges, OnFilterRecord)
    function IsRecordVisible(Cursor: TACRCursor): Boolean;

    procedure MergeAndCheckSearchConditionsCompatibility(
                                var Condition1:            TACRScanSearchCondition;
                                var Condition2:            TACRScanSearchCondition;
                                out Incompatible:          Boolean;
                                out HaveBeenMerged:        Boolean
                                           );
    // returns true if conditions are equal
    function IsEqualConditions(
                                var Condition1:            TACRScanSearchCondition;
                                var Condition2:            TACRScanSearchCondition
                                           ): Boolean;
    // removes duplicate conditions
    procedure RemoveDuplicateConditions(Conditions: TACRScanSearchConditionArray);
    // sorts Conditions array and removes unnecessary conditions
    procedure OptimizeSearchConditions(
                                var Conditions:               TACRScanSearchConditionArray;
                                out NonCompatibleConditions:  Boolean
                                     );
    // prepare conditions array
    procedure PrepareConditions(
                           Cursor:            TACRCursor;
                           Conditions:        TACRScanSearchConditionArray;
                           KeyCondition:      TACRScanSearchCondition;
                           SearchExpression:  TACRExpression;
                           GoForward:         Boolean
                               );
    // try to find the best scan condition with min range record count
    function ChooseScanConditionsWithMinRangeRecordCount(
                           SessionID:              TACRSessionID;
                           Conditions:             TACRScanSearchConditionArray;
                           CurrentIndexID:         TACRObjectID;
                           CurrentRecordID:        TACRRecordID;
                           var ScanConditionNo:    Integer;
                           var ScanEndConditionNo: Integer
                                  ): Boolean;
    // if index is defined as unique or contain unique field
    function IsIndexUnique(IndexID: TACRObjectID): Boolean;
    // try to find the best scan condition using heuristics
    function ChooseScanConditionsByHeuristsics(
                           Conditions:             TACRScanSearchConditionArray;
                           CurrentIndexID:         TACRObjectID;
                           var ScanConditionNo:    Integer;
                           var ScanEndConditionNo: Integer
                                  ): Boolean;
    // if any index exists in conditions without expression will use it
    function ChooseScanConditionsByAnyIndex(
                           Conditions:             TACRScanSearchConditionArray;
                           CurrentIndexID:         TACRObjectID;
                           var ScanConditionNo:    Integer;
                           var ScanEndConditionNo: Integer
                                            ): Boolean;
    // try to find the best scan condition
    procedure ChooseScanConditions(
                           SessionID:              TACRSessionID;
                           Conditions:             TACRScanSearchConditionArray;
                           CurrentIndexID:         TACRObjectID;
                           CurrentRecordID:        TACRRecordID;
                           GoForward:              Boolean;
                           var ScanConditionNo:    Integer;
                           var ScanEndConditionNo: Integer
                                  );
    // return true if record was found and is visible by cursor
    // prepares params for FindRecordByScan and calls it
    function FindRecord(
                        Cursor:                 TACRCursor;
                        SearchExpression:       TACRExpression; // locate
                        KeyCondition:           TACRScanSearchCondition; // find key
                        // find first or last record in order specified by index settings
                        Restart:                Boolean;
                        // go from first record to last or vice versa
                        GoForward:              Boolean;
                        // if record was found then return record id
                        ResultRecordID:         PACRRecordID;
                        // if specified - fill record bitmap
                        RecordBitmap:           TACRRecordBitmap = nil;
                        InternalCall:           Boolean = False;
                        StopAtFirstFoundRecord: Boolean = False
                       ): Boolean; virtual;
    // return true if record match specified condition
    function IsRecordMatchCondition(
                                var Condition:        TACRScanSearchCondition;
                                RecordBuffer:         TACRRecordBuffer
                                   ): Boolean;
    // return true if record match specified conditions
    function IsRecordMatchConditions(
                              Conditions:           TACRScanSearchConditionArray;
                              ExcludedConditionNo:  Integer; // INVALID_ID4 if not specified
                              ExcludedConditionNo2: Integer; // INVALID_ID4 if not specified
                              FilterRecordPtr:      Pointer;
                              Dataset:              Pointer;
                              RecordBuffer:         TACRRecordBuffer
                                    ): Boolean;

    function IsRecordMatchConditionsExists(
                              Conditions:           TACRScanSearchConditionArray;
                              ExcludedConditionNo:  Integer; // INVALID_ID4 if not specified
                              ExcludedConditionNo2: Integer; // INVALID_ID4 if not specified
                              FilterRecordPtr:      Pointer
                                    ): Boolean;


    // scan all records and check conditions
    function FindRecordByScanWithoutCondition(
                              SessionID:          TACRSessionID;
                              Conditions:         TACRScanSearchConditionArray;
                              ScanConditionNo:    Integer; // INVALID_ID4 if not specified
                              ScanEndConditionNo: Integer; // INVALID_ID4 if not specified
                              CurrentIndexID:     TACRObjectID; // INVALID_OBJECT_ID if not specified
                              FilterRecordPtr:    Pointer;
                              Dataset:            Pointer;
                              CurrentRecordID:    TACRRecordID;
                              // find first or last record in order specified by index settings
                              Restart:            Boolean;
                              // go from first record to last or vice versa
                              GoForward:          Boolean;
                              // if record was found then return record id
                              ResultRecordID:     PACRRecordID;
                              // if specified - fill record bitmap
                              RecordBitmap:           TACRRecordBitmap = nil;
                              RandomOrder:            Boolean = False;
                              StopAtFirstFoundRecord: Boolean = False
                             ): Boolean;

    // find records by index and condition specified by ScanConditionNo
    // condition index is the same as CurrentIndex
    function FindRecordByScanWithConditionAndConditionIndex(
                              SessionID:          TACRSessionID;
                              Conditions:         TACRScanSearchConditionArray;
                              ScanConditionNo:    Integer; // INVALID_ID4 if not specified
                              ScanEndConditionNo: Integer; // INVALID_ID4 if not specified
                              CurrentIndexID:     TACRObjectID; // INVALID_OBJECT_ID if not specified
                              FilterRecordPtr:    Pointer;
                              Dataset:            Pointer;
                              CurrentRecordID:    TACRRecordID;
                              // find first or last record in order specified by index settings
                              Restart:            Boolean;
                              // go from first record to last or vice versa
                              GoForward:          Boolean;
                              // if record was found then return record id
                              ResultRecordID:     PACRRecordID;
                              // if specified - fill record bitmap
                              RecordBitmap:           TACRRecordBitmap = nil;
                              RandomOrder:            Boolean = False;
                              StopAtFirstFoundRecord: Boolean = False
                             ): Boolean;

    // find records by index and condition specified by ScanConditionNo
    // there is no active index
    function FindRecordByScanWithConditionAndWihtoutCurrentIndex(
                              SessionID:          TACRSessionID;
                              Conditions:         TACRScanSearchConditionArray;
                              ScanConditionNo:    Integer; // INVALID_ID4 if not specified
                              ScanEndConditionNo: Integer; // INVALID_ID4 if not specified
                              CurrentIndexID:     TACRObjectID; // INVALID_OBJECT_ID if not specified
                              FilterRecordPtr:    Pointer;
                              Dataset:            Pointer;
                              CurrentRecordID:    TACRRecordID;
                              // find first or last record in order specified by index settings
                              Restart:            Boolean;
                              // go from first record to last or vice versa
                              GoForward:          Boolean;
                              // if record was found then return record id
                              ResultRecordID:     PACRRecordID;
                              // if specified - fill record bitmap
                              RecordBitmap:           TACRRecordBitmap = nil;
                              RandomOrder:            Boolean = False;
                              StopAtFirstFoundRecord: Boolean = False
                             ): Boolean;

    // sort record ID array by positions in the index
    procedure SortRecordIDByIndex(
                              Records:              TACRRecordIDArray;
                              IndexPositions:       TList;
                              CurrentIndexID:       TACRObjectID // INVALID_OBJECT_ID if not specified
                                 );

    // find records by index and condition specified by ScanConditionNo
    // condition index is NOT the same as CurrentIndex
    function FindRecordByScanWithConditionAndNonConditionIndex(
                              SessionID:          TACRSessionID;
                              Conditions:           TACRScanSearchConditionArray;
                              ScanConditionNo:      Integer; // INVALID_ID4 if not specified
                              ScanEndConditionNo:   Integer; // INVALID_ID4 if not specified
                              CurrentIndexID:       TACRObjectID; // INVALID_OBJECT_ID if not specified
                              FilterRecordPtr:      Pointer;
                              Dataset:              Pointer;
                              CurrentRecordID:      TACRRecordID;
                              // find first or last record in order specified by index settings
                              Restart:              Boolean;
                              // go from first record to last or vice versa
                              GoForward:            Boolean;
                              // if record was found then return record id
                              ResultRecordID:       PACRRecordID;
                              // if specified - fill record bitmap
                              RecordBitmap:           TACRRecordBitmap = nil;
                              RandomOrder:            Boolean = False;
                              StopAtFirstFoundRecord: Boolean = False
                             ): Boolean;

    // return true if record was found
    function FindRecordByScan(
                              SessionID:            TACRSessionID;
                              Conditions:           TACRScanSearchConditionArray;
                              ScanConditionNo:      Integer; // INVALID_ID4 if not specified
                              ScanEndConditionNo:   Integer; // INVALID_ID4 if not specified
                              CurrentIndexID:       TACRObjectID; // INVALID_OBJECT_ID if not specified
                              FilterRecordPtr:      Pointer;
                              Dataset:              Pointer;
                              CurrentRecordID:      TACRRecordID;
                              // find first or last record in order specified by index settings
                              Restart:              Boolean;
                              // go from first record to last or vice versa
                              GoForward:            Boolean;
                              // if record was found then return record id
                              ResultRecordID:       PACRRecordID;
                              // if specified - fill record bitmap
                              RecordBitmap:           TACRRecordBitmap = nil;
                              RandomOrder:            Boolean = False;
                              StopAtFirstFoundRecord: Boolean = False
                             ): Boolean;
    // used by GetRecordBuffer - find or get record
    function InternalFindOrGetRecordBuffer(
                              Cursor:         TACRCursor;
                              GetRecordMode:  TACRGetRecordMode;
                              // if specified - fill record bitmap
                              RecordBitmap:     TACRRecordBitmap = nil
                            ): TACRGetRecordResult;
    // get record using current index or physical order
    procedure InternalGetRecordBuffer(
                                   SessionID:          TACRSessionID;
                                   var NavigationInfo: TACRNavigationInfo
                                      );
    procedure UpdateRecordBitmapAfterInsertRecord(Cursor: TACRCursor; Pos: Pointer);
    procedure UpdateRecordBitmapAfterUpdateRecord(Cursor: TACRCursor);
   public
    function IsRecordExists(Cursor: TACRCursor): Boolean; virtual;
    // return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
    function CompareRecordID(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer; virtual; abstract;
    // read record with Distinct, SQLFilter, SQLTopRowCount, Filter, Range, OnFilterRecord
    function GetRecordBuffer(
                              Cursor:         TACRCursor;
                              GetRecordMode:  TACRGetRecordMode
                            ): TACRGetRecordResult; virtual;
    // locate
    function Locate(Cursor: TACRCursor; SearchExpression: TACRExpression): Boolean;
    // find key
    function FindKey(Cursor: TACRCursor; SearchCondition: TACRSearchCondition): Boolean;

    //---------------------------------------------------------------------------
    // BLOB methods
    //---------------------------------------------------------------------------
    function IsBLOBModified(
                            RecordBuffer: TACRRecordBuffer;
                            FieldDef:     TACRFieldDef
                           ): Boolean;

    procedure SetBLOBModified(
                            Modified:     Boolean;
                            RecordBuffer: TACRRecordBuffer;
                            FieldDef:     TACRFieldDef
                           );

    function InternalCreateBlobStream(
              Cursor:   TACRCursor;
              ToInsert: Boolean;
              FieldNo:  Integer;
              OpenMode: TACRBLOBOpenMode
              ): TACRStream; virtual;

    procedure WriteBLOBFieldToRecordBuffer(
              Cursor:     TACRCursor;
              FieldNo:    Integer;
              BLOBStream: TACRStream
              ); virtual;

    procedure ClearBLOBFieldInRecordBuffer(
              var RecordBuffer: TACRRecordBuffer;
              FieldNo:    Integer
              ); virtual;

    procedure ClearBLOBFieldsInRecordBuffer(
              var RecordBuffer: TACRRecordBuffer
              );


    // check constraint conditions
    procedure CheckConstraints(
                                Cursor:          TACRCursor;
                                SessionID:       TACRSessionID;
                                NewRecordBuffer: TACRRecordBuffer;
                                OldRecordBuffer: TACRRecordBuffer;
                                ToInsert:        Boolean;
                                CurrentRecordID: TACRRecordID;
                                SkipFKCheck:     Boolean = False
                              ); virtual;
    procedure ApplyDefaultValuesToRecordBuffer(
                      Cursor:       TACRCursor;
                      RecordBuffer: TACRRecordBuffer;
                      InsertMode:   Boolean
                                              ); virtual;

    procedure ShowRecord(Cursor: TACRCursor);
    function InsertRecord(var Cursor: TACRCursor): Boolean; virtual; abstract;
    function DeleteRecord(Cursor: TACRCursor): Boolean; virtual; abstract;
    function UpdateRecord(Cursor: TACRCursor): Boolean; virtual; abstract;
    procedure EditRecord(Cursor: TACRCursor); virtual;
    procedure ClearTemporaryBLOBValues(Cursor: TACRCursor);
    procedure SetBLOBValuesModified(Modified: Boolean; Cursor: TACRCursor);
    procedure CancelRecord(Cursor: TACRCursor; ToInsert: Boolean); virtual;
    procedure DeleteVisibleRecords(Cursor: TACRCursor); virtual;
    procedure UpdateVisibleRecords(
                                    Cursor:       TACRCursor;
                                    FieldNames:   TStringList;
                                    values:       array of TACRVariant;
                                    SkipFKCheck:  Boolean = False
                                    ); virtual;
    // record operations
    function GetRecordCount(Cursor: TACRCursor; InternalCall: Boolean = False): TACRRecordNo; virtual;
    procedure SetRecNo(Cursor: TACRCursor; RecNo: TACRRecordNo); virtual;
    function GetRecNo(Cursor: TACRCursor): TACRRecordNo; virtual;

    procedure InternalSetRecNo(Cursor: TACRCursor; RecNo: TACRRecordNo); virtual; abstract;
    function InternalGetRecNo(Cursor: TACRCursor): TACRRecordNo; virtual; abstract;
    function InternalGetRecordCount: TACRRecordNo; virtual; abstract;

    function LastAutoincValue(FieldNo: Integer; Session: TACRBaseSession): Int64; virtual;
    procedure SetLastAutoincValue(Value: Int64; FieldNo: Integer; Cursor: TACRCursor); virtual;

    procedure LockTableForRead(aSessionID: TACRSessionID); virtual;
    procedure UnlockTableForRead(aSessionID: TACRSessionID); virtual;

    function TryToLockTableS: Boolean; virtual;
    function TryToUnlockTableS: Boolean; virtual;
    function TryToLockTableIRW: Boolean; virtual;
    function TryToUnlockTableIRW: Boolean; virtual;
    function TryToLockTableRW: Boolean; virtual;
    function TryToUnlockTableRW: Boolean; virtual;
    function TryToLockTableX: Boolean; virtual;
    function TryToUnlockTableX: Boolean; virtual;
    function TryToLockTableIS: Boolean; virtual;
    function TryToUnlockTableIS: Boolean; virtual;

    procedure StartTransaction(aSessionID: TACRSessionID); virtual;
    procedure Commit(aSessionID: TACRSessionID); virtual;
    procedure Rollback(aSessionID: TACRSessionID); virtual;

    function IsSynchronizationRequired(PageState: Cardinal): Boolean; virtual;

   public
    property BLOBFieldsPresent: Boolean read FBLOBFieldsPresent;
    property TableName: String read FTableName write FTableName;
    property DatabaseData: TACRDatabaseData read FDatabaseData;
    property FieldManager: TACRBaseFieldManager read FFieldManager;
    property RecordManager: TACRBaseRecordManager read FRecordManager;
    property ConstraintManager: TACRBaseConstraintManager read FConstraintManager;
    property IndexManager: TACRBaseIndexManager read FIndexManager;
    property PageManager: TACRPageManager read GetPageManager;
    property InMemory: Boolean read FInMemory;
    property Temporary: Boolean read FTemporary;
    property TableID: TACRObjectID read GetTableID;
    property CursorList: TList read FCursorList;
  end; // TACRTableData


////////////////////////////////////////////////////////////////////////////////
//
// TACRTransaction
//
////////////////////////////////////////////////////////////////////////////////


  TACRTransaction = class (TObject)
   private
    FSession:         TACRBaseSession;
    FSessionID:       TACRSessionID;
    FDatabaseData:    TACRDatabaseData;
    FIsFinished:      Boolean;
    FTableLocksList:  TList;
   protected
    procedure SaveCursorPositions;
    procedure RestoreCursorPositions;
    procedure CallStartTransactionInTableData;
    procedure CallRollbackInTableData;
    procedure CallCommitInTableData;
    procedure UnlockTables;
    function TryToChangeIRWLockToRWLock: Boolean;
    procedure OpenTableForReadAndCheckForOtherLocks(TableLockItemForSession: PACRTableLockItemForSession);
   public
    constructor Create(aSession: TACRBaseSession; aDatabaseData: TACRDatabaseData);
    destructor Destroy; override;
    procedure Commit(FlushFileBuffers: Boolean = True);
    procedure Rollback;
    procedure OpenTableForRead(TableName: String; OpenExclusive: Boolean);
    procedure StartTableModification(TableName: String);
  end; // TACRTransaction


////////////////////////////////////////////////////////////////////////////////
//
// TACRDatabaseData
//
////////////////////////////////////////////////////////////////////////////////


  TACRDatabaseData = class (TObject)
   protected
    FCSect:               TRTLCriticalSection;
    FSessionList:         TList;
    FTableDataList:       TList;
    FDatabaseName:        String;
    FObjectIdSequence:    TACRSequenceDef;   // sequence for new ObjectID
    FPageManager:         TACRPageManager;
   public
    constructor Create;
    destructor Destroy; override;
    procedure Lock;
    procedure Unlock;
    procedure ConnectSession(Session: TACRBaseSession); virtual;
    procedure DisconnectSession(Session: TACRBaseSession); virtual;
    procedure FreeIfNoSessionsConnected; virtual;
    procedure GetTablesList(List: TStrings); virtual;
    function TableExists(TableName: String): Boolean; virtual;

    procedure CreateDatabase(Session: TACRBaseSession); virtual;
    // create table data
    function CreateTableData(Cursor: TACRCursor): TACRTableData; virtual; abstract;
    // find table data
    function FindTableData(Cursor: TACRCursor): TACRTableData; virtual;
    // find or create table data
    function FindOrCreateTableData(Cursor: TACRCursor): TACRTableData; virtual;
    // get new ObjectId
    function GetNewObjectId: TACRObjectID;
    // flush file buffers
    procedure FlushFileBuffers; virtual;
    // return database format version
    function GetFormatVersion(Session: TACRBaseSession): Double; virtual;
        // return total number of pages
    function GetTotalPageCount(Session: TACRBaseSession) : Integer; virtual;
    // return number of free pages
    function GetFreePageCount(Session: TACRBaseSession) : Integer; virtual;
    // return true if database is encrypted
    function IsDatabaseEncrypted(Session: TACRBaseSession): Boolean; virtual;
    // return true if database is encrypted by password or by key
    function IsDatabaseEncryptedByPassword(Session: TACRBaseSession): Boolean; virtual;
    // return true if CryptoParams are valid
    function IsCryptoParamsValid(Session: TACRBaseSession): Boolean; virtual;
    // makes Exe database from edb file
    procedure MakeExeDatabase(Session: TACRBaseSession; ExeFileName, ExeDatabaseFileName: string); virtual;
    // removes database file from executable database file
    procedure RemoveDatabaseFromExe(Session: TACRBaseSession); virtual;
    // returns true if this file is an Accuracer database
    function IsAccuracerDatabaseFile(Session: TACRBaseSession): Boolean; virtual;
    // fill TableLockList with lock info for the tables opened by this session
    procedure GetTableLocksListForSession(
                                              SessionID:      TACRSessionID;
                                              TableLockList:  TList
                                          ); virtual;
    // if result is true and PTableLockInfo is not nil - increment corresponding field of this record
    function TryToLockTable(
                        TableName:            String;
                        SessionID:            TACRSessionID;
                        LockType:             TACRLockType;
                        PTableLockInfo:       PACRTableLockInfo = nil): Boolean; virtual;
    // decrements number of locks by this session and removes all locks if their number = 0
    // if PTableLockInfo = nil - all table locks byt this session will be removed
    function UnlockTableAll(
                            TableName:      String;
                            SessionID:      TACRSessionID;
                            PTableLockInfo: PACRTableLockInfo = nil): Boolean; virtual;
    procedure RemoveAllLocks(SessionID: TACRSessionID); virtual;
   public
    property DatabaseName: String read FDatabaseName write FDatabaseName;
    property ObjectIdSequence: TACRSequenceDef read FObjectIdSequence;
    property PageManager: TACRPageManager read FPageManager write FPageManager;
  end; // TACRDatabaseData




////////////////////////////////////////////////////////////////////////////////
//
// General functions and procedures
//
////////////////////////////////////////////////////////////////////////////////


  // return false if field value is null
  function GetFieldData(
            // field number in FieldDefs
            FieldNo: Integer;
            // field definitions
            FieldDefs: TACRFieldDefs;
            // pointer to allocated buffer for storing field value
            Buffer: Pointer;
            // record buffer
            RecordBuffer: TACRRecordBuffer
                       ): Boolean;

  // returns false if field value is null
  procedure SetFieldData(
            // physical field No
            FieldNo: Integer;
            // field definitions
            FieldDefs: TACRFieldDefs;
            // pointer to allocated buffer with new field value
            Buffer: Pointer;
            // record buffer
            RecordBuffer: TACRRecordBuffer
                        );

 // return true if record buffers are binary identical
 function CompareRecordBuffers(RecordBuffer1,RecordBuffer2: TACRRecordBuffer;
                              RecordBufferSize: Integer): Boolean;

{$IFNDEF SQLMEMTABLE}
 // compresses and encrypts buffer
procedure CompressAndEncryptBuffer(
                        const CryptoInfo:     TACRCryptoInfo;
                        CompressionAlgorithm: Byte;
                        CompressionMode:      Byte;
                        InBuffer:             PChar;
                        InSize:               Integer;
                        out OutBuffer:        PChar;
                        out OutSize:          Integer
                                                              );
 // decompresses and decrypts buffer; return true if successful
 function DecompressAndDecryptBuffer(
                        const CryptoInfo:     TACRCryptoInfo;
                        CompressionAlgorithm: Byte;
                        CompressionMode:      Byte;
                        var Buffer:           PChar;
                        var BufferSize:       Integer
                                                              ): Boolean;

 procedure CompressAndEncryptStream(
                        const CryptoInfo:     TACRCryptoInfo;
                        CompressionAlgorithm: Byte;
                        CompressionMode:      Byte;
                        BlockSize:            Integer;
                        SourceStream:         TACRStream;
                        DestStream:           TACRStream;
                        OnProgress:           TACRProgressEvent = nil
                                    );

 function DecompressAndDecryptStream(
                        const CryptoInfo:     TACRCryptoInfo;
                        CompressionAlgorithm: Byte;
                        CompressionMode:      Byte;
                        BlockSize:            Integer;
                        SourceStream:         TACRStream;
                        DestStream:           TACRStream;
                        OnProgress:           TACRProgressEvent = nil
                                                              ): Boolean;
 {$ENDIF}

implementation


uses

// Accuracer units
     ACRMain,
     ACRBTree,
     ACRLocalEngine;



////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseRecordManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// lock
//------------------------------------------------------------------------------
procedure TACRBaseRecordManager.Lock;
begin
  EnterCriticalSection(FCSect);
end;// Lock


//------------------------------------------------------------------------------
// unlock
//------------------------------------------------------------------------------
procedure TACRBaseRecordManager.Unlock;
begin
  LeaveCriticalSection(FCSect);
end;// Unlock


//------------------------------------------------------------------------------
// return record count
//------------------------------------------------------------------------------
function TACRBaseRecordManager.GetRecordCount: TACRRecordNo;
begin
  Result := FRecordCount;
end; // GetRecordCount


//------------------------------------------------------------------------------
// update record, return true if record was updated, false if record was deleted
//------------------------------------------------------------------------------
function TACRBaseRecordManager.UpdateRecord(
                          RecordBuffer: TACRRecordBuffer;
                          RecordID:     TACRRecordID;
                          SessionID:    TACRSessionID
                         ): Boolean;
begin
 Result := False;
end; // UpdateRecord


//------------------------------------------------------------------------------
// delete record, return true if record was deleted, false if record was deleted earlier
//------------------------------------------------------------------------------
function TACRBaseRecordManager.DeleteRecord(
                          var RecordID: TACRRecordID;
                          SessionID:    TACRSessionID
                         ): Boolean; 
begin
 Result := False;
end; // DeleteRecord


//------------------------------------------------------------------------------
// Load from stream
//------------------------------------------------------------------------------
procedure TACRBaseRecordManager.LoadFromStream(Stream: TStream);
begin
;
end; // LoadFromStream


//------------------------------------------------------------------------------
// Save from stream
//------------------------------------------------------------------------------
procedure TACRBaseRecordManager.SaveToStream(Stream: TStream);
begin
;
end; // SaveToStream


//------------------------------------------------------------------------------
// add loaded record
//------------------------------------------------------------------------------
procedure TACRBaseRecordManager.AddLoadedRecord(RecordBuffer: TACRRecordBuffer; var RecordPos: Integer);
begin
;
end; // SaveToStream


////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseFieldManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRBaseFieldManager.Create(TableData: TACRTableData; SequenceManager: TACRBaseSequenceManager);
begin
  FTableData := TableData;
  FFieldDefs := TACRFieldDefs.Create;
  FSequenceManager := SequenceManager;
  FEngineVersion := ACRVersion;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRBaseFieldManager.Destroy;
begin
  FFieldDefs.Free;
end;//Destroy


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRBaseFieldManager.LoadFromStream(Stream: TStream);
begin
 FieldDefs.EngineVersion := FEngineVersion;
 FieldDefs.LoadFromStream(Stream);
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRBaseFieldManager.SaveToStream(Stream: TStream);
begin
 FieldDefs.EngineVersion := FEngineVersion;
 FieldDefs.SaveToStream(Stream);
end; // SaveToStream


//------------------------------------------------------------------------------
// Set default values to fields
//------------------------------------------------------------------------------
procedure TACRBaseFieldManager.ApplyDefaultValuesToRecordBuffer(
                                Session:      TACRBaseSession;
                                RecordBuffer: TACRRecordBuffer;
                                InsertMode:   Boolean
                                );
var
  i: Integer;
  SequenceValue: TACRSequenceValue;
begin
  if (FFieldDefs.AutoIncFieldsExists or FFieldDefs.DefaultValuesExists) then
   for i:=0 to FFieldDefs.Count-1 do
    begin
      // Apply DefaultValues
      if (not FFieldDefs[i].DefaultValue.IsNull) then
        begin
          FFieldDefs[i].DefaultValue.CopyDataToAddress(RecordBuffer + FFieldDefs[i].MemoryOffset);
          SetNullFlag(false, i, RecordBuffer);
        end;

      // Apply AutoIncs
      if (InsertMode) then
       if (IsAutoincFieldType(FFieldDefs[i].AdvancedFieldType)) then
        begin
          SequenceValue := FSequenceManager.GetNextVal(
                                     Session, FFieldDefs[i].SequenceDefObjectId);
          Move(SequenceValue, (RecordBuffer + FFieldDefs[i].MemoryOffset)^,
               FFieldDefs[i].MemoryDataSize);
          SetNullFlag(false, i, RecordBuffer);
        end;
    end;
end;//ApplyDefaultValuesToRecordBuffer


////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseIndexManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.LoadFromStream(Stream: TStream);
begin
  FIndexDefs.LoadFromStream(Stream);
  FPageManager.LoadFromStream(Stream);
end; // SaveToStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.SaveToStream(Stream: TStream);
begin
  FIndexDefs.SaveToStream(Stream);
  FPageManager.SaveToStream(Stream);
end; // SaveToStream


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRBaseIndexManager.Create(aTableData: TACRTableData);
begin
 FOpenIndexList := TList.Create; // list of TACRIndex objects
 FIndexDefs := TACRIndexDefs.Create;
 FTableData := aTableData;
 FPageManager := aTableData.PageManager;
 FTemporaryPageManager := nil;
end;// Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRBaseIndexManager.Destroy;
begin
 DropTemporaryIndexes(SYSTEM_SESSION_ID);
 while (FOpenIndexList.Count > 0) do
   CloseIndex(TACRIndex(FOpenIndexList.Items[0]).IndexDef.ObjectID);
 FOpenIndexList.Free;
 FIndexDefs.Free;
 if (FTemporaryPageManager <> nil) then
   FTemporaryPageManager.Free;
end;// Destroy


//------------------------------------------------------------------------------
// return temporary page manager
//------------------------------------------------------------------------------
function TACRBaseIndexManager.GetTemporaryPageManager: TACRPageManager;
begin
  if (FTemporaryPageManager = nil) then
   FTemporaryPageManager := TACRTemporaryPageManager.Create;
  Result := FTemporaryPageManager; 
end; // GetTempoararyPageManager


//------------------------------------------------------------------------------
// return PageManager for index
//------------------------------------------------------------------------------
function TACRBaseIndexManager.GetPageManager(Index: TACRIndex): TACRPageManager;
begin
  if (Index.Temporary) then
   Result := GetTemporaryPageManager
  else
   Result := FPageManager;
end; // GetPageManager


//------------------------------------------------------------------------------
// DropTemporaryIndexes
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.DropTemporaryIndexes(SessionID: TACRSessionID);
var i: Integer;
begin
  i := 0;
  while (i < IndexDefs.Count) do
   begin
    if (IndexDefs[i].Temporary) then
     DropIndex(SessionID, IndexDefs[i].ObjectID)
    else
     Inc(i);
   end;
end;// DropTemporaryIndexes


//------------------------------------------------------------------------------
// return true if coresponding index was found
//------------------------------------------------------------------------------
function TACRBaseIndexManager.IsIndexExists(FieldNames, AscDescList, CaseSensitivityList: TStringList): Boolean;
begin
  Result := FIndexDefs.IsIndexExists(FieldNames,AscDescList,CaseSensitivityList);
end; // IsIndexExists


//------------------------------------------------------------------------------
// return IndexID if coresponding index was found
//------------------------------------------------------------------------------
function TACRBaseIndexManager.FindIndex(FieldNames, AscDescList, CaseSensitivityList: TStringList): TACRObjectID;
begin
  Result := FIndexDefs.FindIndex(FieldNames,AscDescList,CaseSensitivityList);
end; // FindIndex


//------------------------------------------------------------------------------
// create temporary index
//------------------------------------------------------------------------------
function TACRBaseIndexManager.CreateTemporaryIndex(
          Cursor: TACRCursor;
          FieldNamesList, AscDescList, CaseSensitivityList: TStringList
                                                  ): TACRObjectID;
var IndexDef: TACRIndexDef;
begin
  raise EACRException.Create(11389,ErrorLOperationIsNotSupported);
  IndexDef := TACRIndexDef.Create;
  try
{  
   repeat
    IndexDef.Name := GetTemporaryName(ACRTemporaryIndexName);
   until (IndexDefs.GetDefNumberByName(IndexDef.Name) = -1);
   IndexDef.Temporary := True;
   IndexDef.IndexType := itBTree;
   IndexDef.Unique := False;
   IndexDef.Primary := False;
   IndexDef.AssignByNames(FieldNamesList,AscDescList,CaseSensitivityList);
   Result := CreateIndex(Cursor, IndexDef);
   GetTemporaryPageManager.ApplyChanges(Cursor.Session.SessionID);
}
  finally
   IndexDef.Free;
  end;
end; // CreateTemporaryIndex


//------------------------------------------------------------------------------
// find or create temporary index
//------------------------------------------------------------------------------
function TACRBaseIndexManager.FindOrCreateIndex(Cursor: TACRCursor; FieldNamesList, AscDescList, CaseSensitivityList: TStringList): TACRObjectID;
begin
  raise EACRException.Create(11390,ErrorLOperationIsNotSupported);
  Result := FindIndex(FieldNamesList,AscDescList,CaseSensitivityList);
  if (Result = INVALID_OBJECT_ID) then
    Result := CreateTemporaryIndex(Cursor, FieldNamesList,AscDescList,CaseSensitivityList);
end; // FindOrCreateIndex


//------------------------------------------------------------------------------
// create index definitions list
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.CreateIndexDefs(aIndexDefs: TACRIndexDefs);
begin
  FIndexDefs.Assign(aIndexDefs);
end;// CreateIndexDefs


//------------------------------------------------------------------------------
// FindOpenIndex
//------------------------------------------------------------------------------
function TACRBaseIndexManager.FindOpenIndex(IndexID: TACRObjectID): TACRIndex;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to FOpenIndexList.Count-1 do
   if (TACRIndex(FOpenIndexList.Items[i]).IndexDef.ObjectID = IndexID) then
    begin
     Result := TACRIndex(FOpenIndexList.Items[i]);
     break;
    end;
end;// FindOpenIndex


//------------------------------------------------------------------------------
// create index
//------------------------------------------------------------------------------
function TACRBaseIndexManager.CreateIndex(
                         Cursor: TACRCursor;
                         IndexDef: TACRIndexDef
                        ): TACRObjectID;
begin
  IndexDef.ObjectID := FTableData.DatabaseData.GetNewObjectID;
  FIndexDefs.AddCreated.Assign(IndexDef);
  InternalCreateIndex(Cursor, IndexDef);
  Result := IndexDef.ObjectID;
end;// CreateIndex


//------------------------------------------------------------------------------
// return index ID if Index exists, otherwise return NULL_ID
//------------------------------------------------------------------------------
function TACRBaseIndexManager.OpenIndex(IndexID: TACRObjectID): TACRIndex;
begin
  Result := FindOpenIndex(IndexID);
  if (Result = nil) then
    Result := InternalOpenIndex(IndexID);
end;// OpenIndex


//------------------------------------------------------------------------------
// close index
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.CloseIndex(IndexID: TACRObjectID);
var
  Index: TACRBTreeIndex;
begin
  Index := TACRBTreeIndex(FindOpenIndex(IndexID));
  if (Index <> nil) then
   begin
    FOpenIndexList.Remove(Index);
    Index.Free;
   end;
end;// CloseIndex


//------------------------------------------------------------------------------
// drop index
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.DropIndex(SessionID: TACRSessionID; IndexID: TACRObjectID);
begin
  EmptyIndex(SessionID, IndexID);
  FIndexDefs.Delete(FIndexDefs.GetDefNumberByObjectId(IndexID));
end;// DropIndex


//------------------------------------------------------------------------------
// DropAllIndexes
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.DropAllIndexes(SessionID: TACRSessionID);
begin
 while (IndexDefs.Count > 0) do
   DropIndex(SessionID, IndexDefs.Items[0].ObjectID);
end;// DropAllIndexes


//------------------------------------------------------------------------------
// EmptyIndex
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.EmptyIndex(SessionID: TACRSessionID; IndexID: TACRObjectID);
var
  Index: TACRIndex;
begin
  Index := OpenIndex(IndexID);
  Index.DropIndex(SessionID);
  FIndexDefs[FIndexDefs.GetDefNumberByObjectId(IndexID)].RootPageNo := INVALID_PAGE_NO;
  CloseIndex(IndexID);
end;// EmptyIndex


//------------------------------------------------------------------------------
// EmptyAllIndexes
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.EmptyAllIndexes(SessionID: TACRSessionID);
var
  i: Integer;
begin
  for i:=0 to IndexDefs.Count-1 do
   EmptyIndex(SessionID, IndexDefs.Items[i].ObjectID);
end;// EmptyAllIndexes


//------------------------------------------------------------------------------
// GetRecordBuffer
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.GetRecordBuffer(
                               SessionID: TACRSessionID;
                               var NavigationInfo: TACRNavigationInfo
                             );
var
  NavigationInfo1: TACRNavigationInfo;
begin
  NavigationInfo1.SessionID := SessionID;
  NavigationInfo1 := NavigationInfo;
  NavigationInfo1.GetRecordResult := grrOK;
  if ((not NavigationInfo.FirstPosition) and (not NavigationInfo.LastPosition)) then
   begin
    NavigationInfo1.GetRecordMode := grmCurrent;
    NavigationInfo1.FirstPosition := False;
    NavigationInfo1.LastPosition := False;
    TableData.RecordManager.GetRecordBuffer(NavigationInfo1);
   end;
  if (NavigationInfo1.GetRecordResult <> grrError) then
   OpenIndex(NavigationInfo.IndexID).GetRecordBuffer(SessionID, NavigationInfo);
end;// GetRecordBuffer


//------------------------------------------------------------------------------
// InsertRecord
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.InsertRecord(Cursor: TACRCursor);
var
  i: Integer;
  Index: TACRIndex;
  bTemporary: Boolean;
begin
  bTemporary := False;
  for i := 0 to FIndexDefs.Count-1 do
   begin
    Index := OpenIndex(FIndexDefs.Items[i].ObjectID);
    Index.InsertRecord(Cursor);
    if (FIndexDefs.Items[i].RootPageNo <> Index.IndexDef.RootPageNo) then
     FIndexDefs.Items[i].RootPageNo := Index.IndexDef.RootPageNo;
    if (Index.Temporary) then
      bTemporary := True;
   end;
 if (bTemporary) then
  if (not Cursor.Session.InTransaction) then
   GetTemporaryPageManager.ApplyChanges(Cursor.Session.SessionID);
end;// InsertRecord


//------------------------------------------------------------------------------
// UpdateRecord
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.UpdateRecord(Cursor: TACRCursor);
var
  i:           Integer;
  Index:       TACRIndex;
  bTemporary:     Boolean;
begin
  bTemporary := False;
  for i := 0 to FIndexDefs.Count-1 do
    begin
     Index := OpenIndex(FIndexDefs.Items[i].ObjectID);
     Index.UpdateRecord(Cursor);
     if (Index.Temporary) then
      bTemporary := True;
    end;
 if (bTemporary) then
  if (not Cursor.Session.InTransaction) then
   GetTemporaryPageManager.ApplyChanges(Cursor.Session.SessionID);
end;// UpdateRecord


//------------------------------------------------------------------------------
// DeleteRecord
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.DeleteRecord(Cursor: TACRCursor);
var
  i:              Integer;
  Index:          TACRIndex;
  NavigationInfo: TACRNavigationInfo;
  bTemporary:     Boolean;
begin
  bTemporary := False;
  NavigationInfo.SessionID := Cursor.Session.SessionID;
  NavigationInfo.RecordBuffer := Cursor.CurrentRecordBuffer;
  NavigationInfo.RecordID := Cursor.CurrentRecordID;
  NavigationInfo.FirstPosition := False;
  NavigationInfo.LastPosition := False;
  NavigationInfo.GetRecordMode := grmCurrent;
  TableData.RecordManager.GetRecordBuffer(NavigationInfo);

  if (NavigationInfo.GetRecordResult <> grrError) then
   begin
     for i := 0 to FIndexDefs.Count-1 do
      begin
       Index := OpenIndex(FIndexDefs.Items[i].ObjectID);
       Index.DeleteRecord(Cursor);
       if (Index.Temporary) then
         bTemporary := True;
      end;
     if (bTemporary) then
      if (not Cursor.Session.InTransaction) then
       GetTemporaryPageManager.ApplyChanges(Cursor.Session.SessionID);
   end;
end;// DeleteRecord


//------------------------------------------------------------------------------
// return 0 if record buffers are equal in this index
// return 1 if Buffer1 is higher than Buffer 2 (Pos1 > Pos2)
// return -1 if Buffer1 is lower than Buffer 2 (Pos1 < Pos2)
// IndexFieldCount = 0 means to compare by all fields in index
//------------------------------------------------------------------------------
function TACRBaseIndexManager.CompareRecordBuffersByIndex(
                    IndexID: TACRObjectID;
                    Buffer1: TACRRecordBuffer;
                    Buffer2: TACRRecordBuffer;
                    IndexFieldCount: Integer
                                    ): Integer;
var
  Index:          TACRIndex;
  FieldCount,Num: Integer;
begin
  if (IndexFieldCount = 0) then
   begin
    Num := FIndexDefs.GetDefNumberByObjectId(IndexID);
    if (Num < 0) then
     raise EACRException.Create(11497,ErrorLCannotFindIndexByID,[IndexID]);
    FieldCount := FIndexDefs.Items[Num].ColumnCount;
   end
  else
   FieldCount := IndexFieldCount;
  Index := OpenIndex(IndexID);
  try
    Result := Index.CompareRecordBuffersByIndex(Buffer1,Buffer2,FieldCount);
  finally
    CloseIndex(IndexID);
  end;
end; // CompareRecordBuffersByIndex


//------------------------------------------------------------------------------
// create index
//------------------------------------------------------------------------------
procedure TACRBaseIndexManager.InternalCreateIndex(
                Cursor: TACRCursor;
                IndexDef: TACRIndexDef
                                                   );
var
  Index: TACRBTreeIndex;
begin
  Index := TACRBTreeIndex.Create(Self);
  try
   try
    Index.CreateIndex(Cursor, IndexDef);
    FIndexDefs.Items[FIndexDefs.GetDefNumberByObjectId(IndexDef.ObjectID)].Assign(IndexDef);
   except
    FIndexDefs.Delete(FIndexDefs.GetDefNumberByObjectId(IndexDef.ObjectID));
    raise;
   end;
  finally
   Index.Free;
  end;
end;// InternalCreateIndex


//------------------------------------------------------------------------------
// Internal open index
//------------------------------------------------------------------------------
function TACRBaseIndexManager.InternalOpenIndex(
                IndexID: TACRObjectID
                                       ): TACRIndex;
var
  Index: TACRBTreeIndex;
  IndexNo: Integer;
begin
  IndexNo := FIndexDefs.GetDefNumberByObjectId(IndexID);
  if (IndexNo = -1) then
   raise EACRException.Create(20018, ErrorAIndexNotFound);
  Index := TACRBTreeIndex.Create(Self);
  Index.OpenIndex(FIndexDefs[IndexNo]);
  FOpenIndexList.Add(Index);
  Result := Index;
end;// InternalOpenIndex



////////////////////////////////////////////////////////////////////////////////
//
// TACRIndex
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRIndex.Create(aIndexManager: TACRBaseIndexManager);
begin
  FIndexManager := aIndexManager;
  FIndexDef := nil;
  FTemporary := False;
end;// Create



//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRIndex.Destroy;
begin
  if (FIndexDef <> nil) then
   FIndexDef.Free;
end;// Destroy


//------------------------------------------------------------------------------
// create index
//------------------------------------------------------------------------------
procedure TACRIndex.CreateIndex(Cursor: TACRCursor; aIndexDef: TACRIndexDef);
begin
 if (FIndexDef <> nil) then
  FIndexDef.Free;
 FIndexDef := TACRIndexDef.Create;
 FIndexDef.Assign(aIndexDef);
 FTemporary := FIndexDef.Temporary;
end;// CreateIndex


//------------------------------------------------------------------------------
// OpenIndex
//------------------------------------------------------------------------------
procedure TACRIndex.OpenIndex(aIndexDef: TACRIndexDef);
begin
 if (FIndexDef <> nil) then
  FIndexDef.Free;
 FIndexDef := TACRIndexDef.Create;
 FIndexDef.Assign(aIndexDef);
 FTemporary := FIndexDef.Temporary;
end;// OpenIndex



////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseSequenceManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRBaseSequenceManager.Create(DatabaseData: TACRDatabaseData);
begin
  FDatabaseData := DatabaseData;
  FSequenceDefs := TACRSequenceDefs.Create;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRBaseSequenceManager.Destroy;
begin
  FSequenceDefs.Free;
end;//Destroy


//------------------------------------------------------------------------------
// find or create SequenceDef by SequenceName
//------------------------------------------------------------------------------
function TACRBaseSequenceManager.GetOrCreateSequenceDefByName(SequenceName: String): TACRSequenceDef;
var i: TACRObjectID;
begin
  i := FSequenceDefs.GetDefNumberByName(SequenceName);
  if i = INVALID_ID4 then
   begin
     Result := FSequenceDefs.AddCreated;
     Result.Name := SequenceName;
     Result.ObjectID := FDatabaseData.GetNewObjectId;
   end
  else
    Result := FSequenceDefs[i];
end;//GetOrCreateSequenceDefByName


//------------------------------------------------------------------------------
// get next sequence value
//------------------------------------------------------------------------------
function TACRBaseSequenceManager.GetNextVal(Session: TACRBaseSession;
  SequenceID: TACRObjectID): TACRSequenceValue;
var
  i: TACRObjectID;
  //OldSequenceValue: TACRSequenceValue;
begin
  i := FSequenceDefs.GetDefNumberByObjectId(SequenceID);
  if i = INVALID_ID4 then
   raise EACRException.Create(30010, ErrorGSequenceIdNotFound,[SequenceID]);
  // Get Next Value
  Result := FSequenceDefs[i].GetNextVal;
  Session.SetSequenceValue(FSequenceDefs[i].Name, Result);
end;//GetNextVal


//------------------------------------------------------------------------------
// get last sequence value
//------------------------------------------------------------------------------
function TACRBaseSequenceManager.GetLastVal(Session: TACRBaseSession;
  SequenceID: TACRObjectID): TACRSequenceValue;
var
  Value: TACRSessionNamedObjectSequenceValue;
  i: TACRObjectID;
begin
  i := FSequenceDefs.GetDefNumberByObjectId(SequenceID);
  if i = INVALID_ID4 then
    raise EACRException.Create(30011, ErrorGSequenceIdNotFound,[SequenceID]);
  Value := TACRSessionNamedObjectSequenceValue(
                       Session.GetNamedObject(FSequenceDefs[i].Name));
  if value = nil then
    raise EACRException.Create(30012, ErrorGSequenceLastValueFailed, [FSequenceDefs[i].Name]);
  Result := value.Value;
end;//GetLastVal


//------------------------------------------------------------------------------
// Get GequenceDef
//------------------------------------------------------------------------------
function TACRBaseSequenceManager.GetSequenceDef(SequenceID: TACRObjectID): TACRSequenceDef;
var i: TACRObjectID;
begin
  i := FSequenceDefs.GetDefNumberByObjectId(SequenceID);
  if i = INVALID_ID4 then
    raise EACRException.Create(30013, ErrorGSequenceIdNotFound,[SequenceID]);
  Result := FSequenceDefs[i];
end;//GetSequenceDef



//------------------------------------------------------------------------------
// Generate unic Sequence Name
//------------------------------------------------------------------------------
function TACRBaseSequenceManager.GenerateSequenceName(isAutoinc: boolean;
                                        ColumnName, TableName: String): String;
var
  i: Integer;
  s: String;
begin
  Result := AutoNameSymbol;
  if isAutoinc then
    Result := Result + AutoNameSequenceAutoIncPrefix
  else
    Result := Result + AutoNameSequenceLinkedWithColumnPrefix;
  Result := Result + AutoNameSymbol + TableName + AutoNameSymbol + ColumnName;
  // Check Name for Uniqueness
  s := Result;
  i := FSequenceDefs.GetDefNumberByName(Result);
  while i <> -1 do
   begin
    Result := s + IntToStr(Random(MAXWORD));
     i := FSequenceDefs.GetDefNumberByName(Result);
   end;
end;//GenerateSequenceName


////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseConstraintManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRBaseConstraintManager.LoadFromStream(Stream: TStream);
begin
 ConstraintDefs.LoadFromStream(Stream);
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRBaseConstraintManager.SaveToStream(Stream: TStream);
begin
 ConstraintDefs.SaveToStream(Stream);
end; // SaveToStream


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRBaseConstraintManager.Create(aTableData: TACRTableData);
begin
  FTableData := aTableData;
  FConstraintDefs := TACRConstraintDefs.Create;
end;//Create


//------------------------------------------------------------------------------
// Denstructor
//------------------------------------------------------------------------------
destructor TACRBaseConstraintManager.Destroy;
begin
  FConstraintDefs.Free;
end;//Destroy


//------------------------------------------------------------------------------
// check constraint conditions
//------------------------------------------------------------------------------
procedure TACRBaseConstraintManager.CheckConstraints(
                                Cursor:          TACRCursor;
                                SessionID:       TACRSessionID;
                                NewRecordBuffer: TACRRecordBuffer;
                                OldRecordBuffer: TACRRecordBuffer;
                                ToInsert:        Boolean;
                                CurrentRecordID: TACRRecordID;
                                SkipFKCheck:     Boolean = False
                                                    );
var i,j:    Integer;
    ObjId:  TACRObjectID;
    n:      Integer;
begin
  for i:=0 to ConstraintDefs.Count-1 do
    case ConstraintDefs[i].ConstraintType of
      ctNotNull:
        begin
          ObjId := TACRConstraintDefNotNull(ConstraintDefs[i]).ColumnObjectID;
          n := FTableData.FieldManager.FieldDefs.GetDefNumberByObjectId(ObjId);
          if n = -1 then
            raise EACRException.Create(30024, ErrorGFieldWithObjectIdNotFound, [ObjId]);
          if CheckNullFlag(n, NewRecordBuffer) then
           raise EACRException.Create(30026, ErrorGConstraintNotNull,
                                   [ConstraintDefs[i].Name,
                                    FTableData.FieldManager.FieldDefs[n].Name]);
        end;
      ctCheck:
        begin
          ObjId := TACRConstraintDefCheck(ConstraintDefs[i]).ColumnObjectID;
          n := FTableData.FieldManager.FieldDefs.GetDefNumberByObjectId(ObjId);
          if n = -1 then
            raise EACRException.Create(30028, ErrorGFieldWithObjectIdNotFound, [ObjId]);
          if (not CheckNullFlag(n, NewRecordBuffer)) then
           begin
             // Check Min Constraint ( Min > Value ==> Raise )
             if (not TACRConstraintDefCheck(ConstraintDefs[i]).MinValue.IsNull) then
              if CompareValueBuffers(
                              TACRConstraintDefCheck(ConstraintDefs[i]).MinValue.pData,
                              NewRecordBuffer + FTableData.FieldManager.FieldDefs[n].MemoryOffset,
                              TACRConstraintDefCheck(ConstraintDefs[i]).MinValue.DataType,
                              FTableData.FieldManager.FieldDefs[n].BaseFieldType
                                    ) = cmprGreater then
                raise EACRException.Create(30029, ErrorGConstraintCheckMinViolated,
                     [ConstraintDefs[i].Name,
                      FTableData.FieldManager.FieldDefs[n].Name,
                      TACRConstraintDefCheck(ConstraintDefs[i]).MinValue.AsString]);

             // Check Max Constraint ( Max < Value ==> Raise )
             if (not TACRConstraintDefCheck(ConstraintDefs[i]).MaxValue.IsNull) then
              if CompareValueBuffers(
                              TACRConstraintDefCheck(ConstraintDefs[i]).MaxValue.pData,
                              NewRecordBuffer + FTableData.FieldManager.FieldDefs[n].MemoryOffset,
                              TACRConstraintDefCheck(ConstraintDefs[i]).MaxValue.DataType,
                              FTableData.FieldManager.FieldDefs[n].BaseFieldType
                                    ) = cmprLower then
                raise EACRException.Create(30030, ErrorGConstraintCheckMaxViolated,
                     [ConstraintDefs[i].Name,
                      FTableData.FieldManager.FieldDefs[n].Name,
                      TACRConstraintDefCheck(ConstraintDefs[i]).MaxValue.AsString]);

           end;
        end;
      ctPK:
        begin
          // Check NULL
          for j:=0 to Length(TACRConstraintDefPrimary(ConstraintDefs[i]).Columns)-1 do
           begin
            ObjId := TACRConstraintDefPrimary(ConstraintDefs[i]).Columns[j].ColumnObjectID;
            n := FTableData.FieldManager.FieldDefs.GetDefNumberByObjectId(ObjId);
            if n = -1 then
              raise EACRException.Create(30329, ErrorGFieldWithObjectIdNotFound, [ObjId]);

            if CheckNullFlag(n, NewRecordBuffer) then
              raise EACRException.Create(30330, ErrorGConstraintNotNull,
                                   [ConstraintDefs[i].Name,
                                    FTableData.FieldManager.FieldDefs[n].Name]);

           end;

          // Check UNIQUE
          if (FTableData.IsUniqueConstraintFailed(
              SessionID, TACRConstraintDefPrimary(ConstraintDefs[i]).IndexObjectID,
               NewRecordBuffer, OldRecordBuffer, ToInsert, CurrentRecordID)) then
           raise EACRException.Create(30319, ErrorGConstraintPrimaryKeyViolated,
                                   [ConstraintDefs[i].Name]);
        end;
      ctUnique:
        begin
          if (FTableData.IsUniqueConstraintFailed(
              SessionID, TACRConstraintDefUnique(ConstraintDefs[i]).IndexObjectID,
               NewRecordBuffer, OldRecordBuffer, ToInsert, CurrentRecordID)) then
           raise EACRException.Create(30320, ErrorGConstraintUniqueViolated,
                                   [ConstraintDefs[i].Name]);
        end;
      ctFKAction: ; // do nothing
      ctFK:
       begin
        if (not SkipFKCheck) then
         if (FTableData.IsForeignKeyConstraintFailed(Cursor,
            TACRConstraintDefForeignKey(ConstraintDefs[i]))) then
          raise EACRException.Create(11420,ErrorLConstraintForeignKeyViolated,[ConstraintDefs[i].Name,FTableData.TableName]);
       end;
      else
        raise EACRException.Create(30036, ErrorGNotImplementedYet);
    end;
end;//CheckConstraints


//------------------------------------------------------------------------------
//  Link object id
//------------------------------------------------------------------------------
procedure TACRBaseConstraintManager.LinkObjectId(ConstraintDef: TACRConstraintDef);
var j: Integer;
    ColumnName, IndexName: TACRObjectName;
    n: Integer;
begin
    case ConstraintDef.ConstraintType of
      ctNotNull,
      ctCheck:
        begin
          ColumnName := TACRConstraintDefNotNull(ConstraintDef).ColumnName;
          n := FTableData.FieldManager.FieldDefs.GetDefNumberByName(ColumnName);
          if n = -1 then
            raise EACRException.Create(30027, ErrorGFieldWithNameNotFound,[ColumnName]);
          TACRConstraintDefNotNull(ConstraintDef).ColumnObjectID :=
                                  FTableData.FieldManager.FieldDefs[n].ObjectID;
        end;
      ctPK:
        begin
          // Set Index Object ID
          IndexName := TACRConstraintDefPrimary(ConstraintDef).IndexName;
          n := FTableData.IndexManager.FIndexDefs.GetDefNumberByName(IndexName);
          if n = -1 then
            raise EACRException.Create(30325, ErrorGIndexNameNotFound,[IndexName]);
          TACRConstraintDefPrimary(ConstraintDef).IndexObjectID :=
                                 FTableData.IndexManager.FIndexDefs[n].ObjectID;

          // Set Columns Object ID
          for j:=0 to Length(TACRConstraintDefPrimary(ConstraintDef).Columns)-1 do
            begin
              ColumnName := TACRConstraintDefPrimary(ConstraintDef).Columns[j].ColumnName;
              n := FTableData.FieldManager.FieldDefs.GetDefNumberByName(ColumnName);
              if n = -1 then
                raise EACRException.Create(30331, ErrorGFieldWithNameNotFound,[ColumnName]);
              TACRConstraintDefPrimary(ConstraintDef).Columns[j].ColumnObjectID :=
                              FTableData.FieldManager.FieldDefs[n].ObjectID;

            end;
        end;
      ctUnique:
        begin
          // Set Index Object ID
          IndexName := TACRConstraintDefUnique(ConstraintDef).IndexName;
          n := FTableData.IndexManager.FIndexDefs.GetDefNumberByName(IndexName);
          if n = -1 then
            raise EACRException.Create(30326, ErrorGIndexNameNotFound,[IndexName]);
          TACRConstraintDefUnique(ConstraintDef).IndexObjectID :=
                                 FTableData.IndexManager.FIndexDefs[n].ObjectID;

          // Set Columns Object ID
          for j:=0 to Length(TACRConstraintDefUnique(ConstraintDef).Columns)-1 do
            begin
              ColumnName := TACRConstraintDefUnique(ConstraintDef).Columns[j].ColumnName;
              n := FTableData.FieldManager.FieldDefs.GetDefNumberByName(ColumnName);
              if n = -1 then
                raise EACRException.Create(30332, ErrorGFieldWithNameNotFound,[ColumnName]);
              TACRConstraintDefUnique(ConstraintDef).Columns[j].ColumnObjectID :=
                              FTableData.FieldManager.FieldDefs[n].ObjectID;

            end;
        end;
       ctFKAction: ;
       ctFK:
        begin
          // Set Columns Object ID
          for j:=0 to Length(TACRConstraintDefForeignKey(ConstraintDef).Columns)-1 do
            begin
              ColumnName := TACRConstraintDefForeignKey(ConstraintDef).Columns[j].ColumnName;
              n := FTableData.FieldManager.FieldDefs.GetDefNumberByName(ColumnName);
              if n = -1 then
                raise EACRException.Create(11419, ErrorGFieldWithNameNotFound,[ColumnName]);
              TACRConstraintDefForeignKey(ConstraintDef).Columns[j].ColumnObjectID :=
                              FTableData.FieldManager.FieldDefs[n].ObjectID;
            end;
        end;
      else
        raise EACRException.Create(30037, ErrorGNotImplementedYet);
    end;
end;//LinkObjectId


//------------------------------------------------------------------------------
//  Link object ids between constraints and fields
//------------------------------------------------------------------------------
procedure TACRBaseConstraintManager.LinkObjectIds;
var
  i: Integer;
begin
  for i:=0 to ConstraintDefs.Count-1 do
    LinkObjectId(ConstraintDefs[i]);
end;//FillObjectIds


//------------------------------------------------------------------------------
// Generate Constraint Names for empty constraints
//------------------------------------------------------------------------------
procedure TACRBaseConstraintManager.FillConstraintAutoNames;
var
  i, n: Integer;
  s, CName: String;
begin
  for i:=0 to FConstraintDefs.Count-1 do
   if (FConstraintDefs[i].Name = '') then
    begin
     s := AutoNameSymbol;
     case FConstraintDefs[i].ConstraintType of
      ctFK,ctFKAction:
        begin
         s := GetTemporaryName(ACRConstraintFKName);
        end;
      ctNotNull:
        begin
         s := s + AutoNameConstraintNotNullPrefix + AutoNameSymbol +
              TACRConstraintDefNotNull(FConstraintDefs[i]).TableName + AutoNameSymbol +
              TACRConstraintDefNotNull(FConstraintDefs[i]).ColumnName;
        end;
      ctCheck:
        begin
         s := s + AutoNameConstraintCheckPrefix + AutoNameSymbol +
              TACRConstraintDefNotNull(FConstraintDefs[i]).TableName + AutoNameSymbol +
              TACRConstraintDefNotNull(FConstraintDefs[i]).ColumnName;
        end;
      else
        raise EACRException.Create(30038, ErrorGNotImplementedYet);
     end;

     CName := s;

     // Check Unique Names
     n := FConstraintDefs.GetDefNumberByName(CName);
     while n <> -1 do
      begin
       CName := s + IntToStr(Random(MAXWORD));
       n := FConstraintDefs.GetDefNumberByName(CName);
      end;

      FConstraintDefs[i].Name := CName;
    end;

end;//FillConstraintAutoNames


//------------------------------------------------------------------------------
// add constraint for primary or Unique index
//------------------------------------------------------------------------------
procedure TACRBaseConstraintManager.AddConstraintFromIndex(IndexDef: TACRIndexDef);
var
  ConstraintDef: TACRConstraintDef;
begin
  ConstraintDef := AddConstraintForIndex(IndexDef, FConstraintDefs);
  if (ConstraintDef <> nil) then
    begin
      ConstraintDef.ObjectID := FTableData.DatabaseData.GetNewObjectID;
      LinkObjectId(ConstraintDef);
    end;
end;//AddConstraintFromIndex


//------------------------------------------------------------------------------
// Delete Constraint Linked with deleted index
//------------------------------------------------------------------------------
procedure TACRBaseConstraintManager.DeleteConstraintForIndexID(IndexObjectID: TACRObjectID);
var
  i: Integer;
begin
  for i:=0 to FConstraintDefs.Count-1 do
    case FConstraintDefs[i].ConstraintType of
      ctPK:
        if (TACRConstraintDefPrimary(FConstraintDefs[i]).IndexObjectID = IndexObjectID ) then
          begin
            FConstraintDefs.Delete(i);
            break;
          end;
      ctUnique:
        if (TACRConstraintDefUnique(FConstraintDefs[i]).IndexObjectID = IndexObjectID ) then
          begin
            FConstraintDefs.Delete(i);
            break;
          end;
    end;
end;//DeleteConstraintForIndexID


//------------------------------------------------------------------------------
// add action to master table
//------------------------------------------------------------------------------
function TACRBaseConstraintManager.AddForeignKeyAction(
                              ConstraintDef:           TACRConstraintDefForeignKey;
                              ReferencedTableName:     String;
                              ReferencedTableObjectID: TACRObjectID
                            ): TACRConstraintDefForeignKeyAction;
var i,h:            Integer;
    name:           String;
begin
  name := ConstraintDef.Name+'_Action';
  while (FConstraintDefs.GetDefNumberByName(name) >= 0) do
   name := GetTemporaryName(ConstraintDef.Name+'_Action_');
  Result := FConstraintDefs.AddFKAction;
  if (Result <> nil) then
    begin
      Result.Name := name;
      Result.ObjectID := FTableData.DatabaseData.GetNewObjectID;
      Result.ReferencedTableName := ReferencedTableName;
      Result.ReferencedTableObjectID := ReferencedTableObjectID;
      Result.ReferencedFKName := ConstraintDef.Name;
      Result.ReferencedFKObjectID := ConstraintDef.ObjectID;
      h := High(ConstraintDef.Columns);
      SetLength(Result.Columns,h+1);
      for i := 0 to h do
       Result.Columns[i] := ConstraintDef.Columns[i];
      Result.DeleteAction := ConstraintDef.DeleteAction;
      Result.UpdateAction := ConstraintDef.UpdateAction;
      Result.MatchType := ConstraintDef.MatchType;
    end;
end; // AddForeignKeyAction


////////////////////////////////////////////////////////////////////////////////
//
// TACRRecordBitmap
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// set active
//------------------------------------------------------------------------------
procedure TACRRecordBitmap.SetActive(Value: Boolean);
begin
  FActive := Value;
  FDistinct := False;
  if (not Value) then
   begin
     FVisibleRecords.SetSize(0);
     FVisibleRecords.Sorted := True;
   end;
end; // SetActive


//------------------------------------------------------------------------------
// set indexed
//------------------------------------------------------------------------------
procedure TACRRecordBitmap.SetIndexed(Value: Boolean);
begin
  FIndexed := Value;
end; // SetActive


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRRecordBitmap.Create(aTableData: TACRTableData);
begin
 if (aTableData = nil) then
  raise EACRException.Create(10425,ErrorLNilPointer);
 FTableData := aTableData;
 FVisibleRecords := TACRRecordIDArray.Create;
 FVisibleRecords.Temporary := FTableData.Temporary;
 FVisibleRecords.InMemory := FTableData.InMemory;
 FActive := False;
 FDistinct := False;
 FIndexed := False;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRRecordBitmap.Destroy;
begin
 FVisibleRecords.Free;
 inherited;
end; // Destroy


//------------------------------------------------------------------------------
// return record count
//------------------------------------------------------------------------------
function TACRRecordBitmap.GetRecordCount: TACRRecordNo;
begin
 Result := FVisibleRecords.ItemCount;
end; // GetRecordCount


//------------------------------------------------------------------------------
// return recno by RecordID
//------------------------------------------------------------------------------
function TACRRecordBitmap.GetRecNoByRecordID(RecordID: TACRRecordID): TACRRecordNo;
begin
 Result := FVisibleRecords.FindRecordByID(RecordID);
 if (Result >= 0) then
  Inc(Result);
end; // GetRecNoByRecordID


//------------------------------------------------------------------------------
// return RecordID by recno
//------------------------------------------------------------------------------
function TACRRecordBitmap.GetRecordIDByRecNo(RecNo: TACRRecordNo): TACRRecordID;
begin
 Dec(RecNo);
 if ((RecNo < 0) or (RecNo >= FVisibleRecords.ItemCount)) then
  raise EACRException.Create(11326,ErrorLInvalidItemNumber,[RecNo,FVisibleRecords.ItemCount]);
 Result := FVisibleRecords.Items[RecNo];
end; // GetRecordIDByRecNo


//------------------------------------------------------------------------------
// return true if record is visible
//------------------------------------------------------------------------------
function TACRRecordBitmap.IsRecordVisible(RecordID: TACRRecordID): Boolean;
begin
 if (FDistinct) then
  begin
    ShowRecord(RecordID);
    Result := True;
  end
 else
  Result := (FVisibleRecords.FindRecordByID(RecordID) >= 0);
end; // IsRecordVisible


//------------------------------------------------------------------------------
// insert new visible record to record map
//------------------------------------------------------------------------------
procedure TACRRecordBitmap.InsertVisibleRecord(
                                 var   RecordID:      TACRRecordID;
                                 const PriorRecordID: TACRRecordID;
                                 bFirst:              Boolean;
                                 bLast:                 Boolean
                                );
var x: Integer;
begin
  FVisibleRecords.Sorted := False;
  if (bLast) then
   FVisibleRecords.Append(RecordID)
  else
  if (bFirst) then
   FVisibleRecords.Insert(0,RecordID)
  else
   begin
    x := FVisibleRecords.FindRecordByID(PriorRecordID);
    if (x < 0) then
     raise EACRException.Create(11332,ErrorLRecordDoesNotExist,[PriorRecordID.PageNo,PriorRecordID.PageItemNo]);
    FVisibleRecords.Insert(x+1,RecordID); 
   end;
end; // InsertVisibleRecord


//------------------------------------------------------------------------------
// show record
//------------------------------------------------------------------------------
procedure TACRRecordBitmap.ShowRecord(RecordID: TACRRecordID);
var res: Integer;
begin
  if (FVisibleRecords.ItemCount >= 1) then
   begin
    if (FVisibleRecords.Sorted) then
     begin
      res := FVisibleRecords.CompareRecordID(RecordID,FVisibleRecords.MaxRecordID);
      if (res = 0) then
       raise EACRException.Create(11328,ErrorLDuplicateRecordID,
        [RecordID.PageNo,RecordID.PageItemNo]);
     end;
    FVisibleRecords.Append(RecordID);
    if (FVisibleRecords.Sorted) then
     begin
      if (res < 0) then
        FVisibleRecords.Sorted := False
      else
       FVisibleRecords.MaxRecordID := RecordID;
     end;
   end
  else
   begin
    FVisibleRecords.Append(RecordID);
    FVisibleRecords.MaxRecordID := RecordID;
   end;
end; // ShowRecord


//------------------------------------------------------------------------------
// hide record
//------------------------------------------------------------------------------
procedure TACRRecordBitmap.HideRecord(RecordID: TACRRecordID);
var RecNo: Integer;
begin
 RecNo := FVisibleRecords.FindRecordByID(RecordID);
 if (RecNo < 0) then
  raise EACRException.Create(11327,ErrorLRecordDoesNotExist,[RecordID.PageNo,RecordID.PageItemNo]);
 FVisibleRecords.Delete(RecNo);
end; // HideRecord


//------------------------------------------------------------------------------
// clear visible records map
//------------------------------------------------------------------------------
procedure TACRRecordBitmap.ClearVisibleRecords;
begin
  FVisibleRecords.SetSize(0);
  FVisibleRecords.Sorted := True;
end; // ClearVisibleRecords


//------------------------------------------------------------------------------
// GetRecordFromFirstPosition
//------------------------------------------------------------------------------
procedure TACRRecordBitmap.GetRecordFromFirstPosition(var NavigationInfo: TACRNavigationInfo);
begin
 case NavigationInfo.GetRecordMode of
  grmPrior:
   begin
    NavigationInfo.GetRecordResult := grrBOF;
   end;
  grmCurrent:
   begin
    NavigationInfo.GetRecordResult := grrError;
   end;
  grmNext:
   begin
    NavigationInfo.GetRecordResult := grrOk;
    NavigationInfo.RecordID := FVisibleRecords.Items[0];
   end;
 end; // GetRecordMode
end; // GetRecordFromFirstPosition


//------------------------------------------------------------------------------
// GetRecordFromLastPosition
//------------------------------------------------------------------------------
procedure TACRRecordBitmap.GetRecordFromLastPosition(var NavigationInfo: TACRNavigationInfo);
begin
 case NavigationInfo.GetRecordMode of
  grmPrior:
   begin
    NavigationInfo.GetRecordResult := grrOk;
    NavigationInfo.RecordID := FVisibleRecords.Items[FVisibleRecords.ItemCount-1];
   end;
  grmCurrent:
   begin
    NavigationInfo.GetRecordResult := grrError;
   end;
  grmNext:
   begin
    NavigationInfo.GetRecordResult := grrEOF;
   end;
 end; // GetRecordMode
end; // GetRecordFromLastPosition


//------------------------------------------------------------------------------
// GetRecordFromAnyPosition
//------------------------------------------------------------------------------
procedure TACRRecordBitmap.GetRecordFromAnyPosition(var NavigationInfo: TACRNavigationInfo);
var RecNo: Integer;
begin
 RecNo := FVisibleRecords.FindRecordByID(NavigationInfo.RecordID);
 if (RecNo < 0) then
  begin
   NavigationInfo.GetRecordResult := grrError;
   Exit;
  end;
 case NavigationInfo.GetRecordMode of
  grmPrior:
   begin
    if (RecNo = 0) then
     NavigationInfo.GetRecordResult := grrBOF
    else
     begin
       NavigationInfo.GetRecordResult := grrOK;
       NavigationInfo.RecordID := FVisibleRecords.Items[RecNo-1];
     end;
   end;
  grmCurrent:
   begin
    NavigationInfo.GetRecordResult := grrOk;
   end;
  grmNext:
   begin
    if (RecNo >= (FVisibleRecords.ItemCount-1)) then
     NavigationInfo.GetRecordResult := grrEOF
    else
     begin
       NavigationInfo.GetRecordResult := grrOK;
       NavigationInfo.RecordID := FVisibleRecords.Items[RecNo+1];
     end;
   end;
 end; // GetRecordMode
end; // GetRecordFromAnyPosition


//------------------------------------------------------------------------------
// get record
//------------------------------------------------------------------------------
procedure TACRRecordBitmap.GetRecord(var NavigationInfo: TACRNavigationInfo);
begin
  if (not FActive) then
   begin
    NavigationInfo.GetRecordResult := grrError;
    Exit;
   end;
  if (GetRecordCount = 0) then
   begin
    NavigationInfo.GetRecordResult := grrEOF;
    Exit;
   end;
  if (NavigationInfo.FirstPosition) then
   GetRecordFromFirstPosition(NavigationInfo)
  else
  if (NavigationInfo.LastPosition) then
   GetRecordFromLastPosition(NavigationInfo)
  else
   GetRecordFromAnyPosition(NavigationInfo);
end; // GetRecord


//------------------------------------------------------------------------------
// prepare bitmap for activation
//------------------------------------------------------------------------------
procedure TACRRecordBitmap.PrepareBitmapForActivation;
var da, mda: Integer;
begin
 mda := FTableData.InternalGetRecordCount div 10;
 if (mda <= 0) then
  mda := 1000;
 da := mda div 100;
 if (da <= 10) then
  da := 10;
 FVisibleRecords.AllocBy := da;
 FVisibleRecords.MaxAllocBy := mda;
end; // PrepareBitmapForActivation


////////////////////////////////////////////////////////////////////////////////
//
// TACRTableData
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// get table id
//------------------------------------------------------------------------------
function TACRTableData.GetTableID: TACRObjectID;
begin
 Result := INVALID_OBJECT_ID;
end; // GetTableID


//------------------------------------------------------------------------------
// Fill New ObjectID for all defs
//------------------------------------------------------------------------------
procedure TACRTableData.FillDefsByObjectId(Defs: TACRMetaObjectDefs);
var i: Integer;
begin
  for i:=0 to Defs.Count-1 do
      Defs[i].ObjectId := FDatabaseData.GetNewObjectId;
end;//FillDefsByObjectId


//------------------------------------------------------------------------------
// calc record buffer size
//------------------------------------------------------------------------------
function TACRTableData.GetRecordBufferSize: Integer;
begin
 if (FFieldManager.FieldDefs.Count <= 0) then
  raise EACRException.Create(10131,ErrorLNoFields);
 Result := FFieldManager.FieldDefs[FFieldManager.FieldDefs.Count-1].MemoryOffset +
   FFieldManager.FieldDefs[FFieldManager.FieldDefs.Count-1].MemoryDataSize;
end;// GetRecordBufferSize


//------------------------------------------------------------------------------
// Create ConstraintManager
//------------------------------------------------------------------------------
procedure TACRTableData.CreateConstraintManager(ConstraintDefs: TACRConstraintDefs);
var i: Integer;
begin
  if (FConstraintManager <> nil) then
    FConstraintManager.Free;
  // Fill New ObjectIds
  FillDefsByObjectId(ConstraintDefs);
  // Set TableName to Constraints
  for i:=0 to ConstraintDefs.Count-1 do
   case ConstraintDefs[i].ConstraintType of
    ctFK,ctFKAction: ;
    ctNotNull:
      TACRConstraintDefNotNull(ConstraintDefs[i]).TableName := FTableName;
    ctCheck:
      TACRConstraintDefCheck(ConstraintDefs[i]).TableName := FTableName;
    ctPK:
      TACRConstraintDefPrimary(ConstraintDefs[i]).TableName := FTableName;
    ctUnique:
      TACRConstraintDefUnique(ConstraintDefs[i]).TableName := FTableName;
    else
      raise EACRException.Create(30040, ErrorGNotImplementedYet);
   end;
  // Create ConstraintManager
  FConstraintManager := TACRBaseConstraintManager.Create(Self);
  FConstraintManager.ConstraintDefs.Assign(ConstraintDefs);
  // Links ObjectIds
  FConstraintManager.LinkObjectIds;
  // Set empty Names
  FConstraintManager.FillConstraintAutoNames;
end;//CreateConstraintManager


//------------------------------------------------------------------------------
// create and init FieldManager
//------------------------------------------------------------------------------
procedure TACRTableData.CreateFieldManager(FieldDefs: TACRFieldDefs);
var i: Integer;
begin
 if (FFieldManager <> nil) then
  FFieldManager.Free;
 FFieldManager := TACRBaseFieldManager.Create(Self,FSequenceManager);
 FillDefsByObjectId(FieldDefs);
 FFieldManager.FieldDefs.Assign(FieldDefs);
 FFieldManager.FieldDefs.RecalcFieldOffsets;
 FBLOBFieldsPresent := False;
 for i := 0 to FFieldManager.FieldDefs.Count - 1 do
  if (IsBLOBFieldType(FFieldManager.FieldDefs[i].BaseFieldType)) then
   begin
    FBLOBFieldsPresent := True;
    if (FFieldManager.FieldDefs[i].BLOBBlockSize = 0) then
     raise EACRException.Create(10421,ErrorLZeroBlockSizeIsNotAllowedForField,[FFieldManager.FieldDefs[i].Name]);
   end;
end;// CreateFieldManager


//------------------------------------------------------------------------------
// create and init IndexManager
//------------------------------------------------------------------------------
procedure TACRTableData.CreateIndexManager(IndexDefs: TACRIndexDefs);
begin
 if (FIndexManager <> nil) then
  FIndexManager.Free;
 FillDefsByObjectId(IndexDefs);
 FIndexManager := TACRBaseIndexManager.Create(Self);
 FIndexManager.CreateIndexDefs(IndexDefs);
end;// CreateIndexManager


//------------------------------------------------------------------------------
// CreateSequenceManager
//------------------------------------------------------------------------------
procedure TACRTableData.CreateSequenceManager;
begin
  if (FSequenceManager <> nil) then
   FSequenceManager.Free;
  FSequenceManager := TACRBaseSequenceManager.Create(FDatabaseData);
end; // CreateSequenceManager


//------------------------------------------------------------------------------
// BuildSequences
//------------------------------------------------------------------------------
procedure TACRTableData.BuildSequences;
var i: Integer;
    SName: String;
    SequenceDef: TACRSequenceDef;
begin
  for i:=0 to FFieldManager.FieldDefs.Count-1 do
   if (IsAutoincFieldType(FFieldManager.FieldDefs[i].AdvancedFieldType)) then
begin
//      SName := FFieldManager.FieldDefs[i].SequenceName;
//      // If SequenceName not found then Generate It
//      if (SName = '') then
       begin
        SName := FSequenceManager.GenerateSequenceName(
                      True,//FFieldManager.FieldDefs[i].AdvancedFieldType = aftAutoInc,
                      FFieldManager.FieldDefs[i].Name,
                      FTableName
                                                      );
        // Create Sequence
        SequenceDef := FSequenceManager.SequenceDefs.AddCreated;
        SequenceDef.Name := SName;
        SequenceDef.ObjectID := FDatabaseData.GetNewObjectId;

        SequenceDef.DataType := FFieldManager.FieldDefs[i].BaseFieldType;
        SequenceDef.MinValue := FFieldManager.FieldDefs[i].AutoincMinValue;
        SequenceDef.MaxValue := FFieldManager.FieldDefs[i].AutoincMaxValue;
        SequenceDef.LastValue := FFieldManager.FieldDefs[i].AutoincInitialValue;
        SequenceDef.Increment := FFieldManager.FieldDefs[i].AutoincIncrement;
        SequenceDef.Cycled := FFieldManager.FieldDefs[i].AutoincCycled;


//        // Set SequenceName in Field
//        FFieldManager.FieldDefs[i].SequenceName := SName;
//       end
//      else
//       begin
//        // Find Sequence by Name
//        SequenceDef := FSequenceManager.GetOrCreateSequenceDefByName(SName);
       end;
      // Make link from Field To Sequence
      FFieldManager.FieldDefs[i].SequenceDefObjectId := SequenceDef.ObjectID;
    end;
end;//BuildSequences


//------------------------------------------------------------------------------
// load sequences
//------------------------------------------------------------------------------
procedure TACRTableData.LoadSequencesFromStream(Stream: TStream);
var SequenceCount:  Integer;
    SequenceDef:    TACRSequenceDef;
    i:              Integer;
begin
  FSequenceManager.SequenceDefs.Clear;
  LoadDataFromStream(SequenceCount,Sizeof(SequenceCount),Stream,10218);
  for i := 0 to SequenceCount - 1 do
   begin
    SequenceDef := FSequenceManager.SequenceDefs.AddCreated;
    SequenceDef.LoadFromStream(Stream);
   end;
end; // LoadSequencesFromStream


//------------------------------------------------------------------------------
// save sequences
//------------------------------------------------------------------------------
procedure TACRTableData.SaveSequencesToStream(Stream: TStream);
var SequenceCount:  Integer;
    i,j,n:          Integer;
    SequenceID:     Integer;
    SequenceIDs:    array of TACRObjectID;
    SequenceExists: Boolean;
begin
  SequenceCount := 0;
  for i := 0 to FFieldManager.FieldDefs.Count - 1 do
   if (IsAutoincFieldType(FFieldManager.FieldDefs[i].AdvancedFieldType)) then
    Inc(SequenceCount);
  SaveDataToStream(SequenceCount,Sizeof(SequenceCount),Stream,10205);
  SetLength(SequenceIDs,SequenceCount);
  n := 0;
  for i := 0 to FFieldManager.FieldDefs.Count - 1 do
   if (IsAutoincFieldType(FFieldManager.FieldDefs[i].AdvancedFieldType)) then
    begin
     SequenceID := FFieldManager.FieldDefs[i].SequenceDefObjectId;
     SequenceExists := False;
     for j := 0 to n-1 do
      if (SequenceIDs[j] = SequenceID) then
       begin
        SequenceExists := True;
        break;
       end;
     if (not SequenceExists) then
      begin
       FSequenceManager.GetSequenceDef(SequenceID).SaveToStream(Stream);
       SequenceIDs[n] := SequenceID;
       Inc(n);
      end;
    end; // field linked to sequence
end; // SaveSequencesToStream


//------------------------------------------------------------------------------
// init cursor
//------------------------------------------------------------------------------
procedure TACRTableData.InitCursor(Cursor: TACRCursor);
begin
 Cursor.FieldValuesOffset := FFieldManager.FieldDefs[0].MemoryOffset;
 Cursor.RecordSize := GetRecordBufferSize;
end;// InitCursor


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRTableData.Create(aDatabaseData: TACRDatabaseData);
begin
 InitializeCriticalSection(FCSect);
 FCursorList := TList.Create;
 FFieldManager := nil;
 FRecordManager := nil;
 FIndexManager := nil;
 FDatabaseData := aDatabaseData;
 FConstraintManager := nil;
 FBLOBFieldsPresent := False;
 FRepair := False;
 FNumLocksForRead := 0;
 FCachedRecordCount := 0;
 FTempRecordBuffer := nil;
end;// Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRTableData.Destroy;
begin
 if (FIndexManager <> nil) then
  FIndexManager.Free;
 FIndexManager := nil;
 if (FFieldManager <> nil) then
  FFieldManager.Free;
 FFieldManager := nil;
 if (FConstraintManager <> nil) then
  FConstraintManager.Free;
 FConstraintManager := nil;
 if (FRecordManager <> nil) then
  FRecordManager.Free;
 FRecordManager := nil;
 DeleteCriticalSection(FCSect);
 FCursorList.Free;
 FCursorList := nil;
 FDatabaseData.FTableDataList.Remove(Self);
 if (FTempRecordBuffer <> nil) then
  MemoryManager.FreeAndNilMem(FTempRecordBuffer);
 inherited;
end;// Destroy


//------------------------------------------------------------------------------
// lock
//------------------------------------------------------------------------------
procedure TACRTableData.Lock;
begin
  EnterCriticalSection(FCSect);
end;// Lock


//------------------------------------------------------------------------------
// unlock
//------------------------------------------------------------------------------
procedure TACRTableData.Unlock;
begin
  LeaveCriticalSection(FCSect);
end;// Unlock


//------------------------------------------------------------------------------
// delete table
//------------------------------------------------------------------------------
procedure TACRTableData.DeleteTable(Session: TACRBaseSession; Cascade: Boolean; DesignMode: Boolean);
var i:                Integer;
    ConstraintDef:    TACRConstraintDef;
    ReferencedCursor: TACRLocalCursor;
begin
 if (not DesignMode) then
  if (FCursorList.Count > 0) then
   raise EACRException.Create(20017, ErrorACannotDeleteBusyTable, [FTableName]);
 if (not Cascade) then
  if (FConstraintManager.ConstraintDefs.ForeignKeysActionsExists or
      FConstraintManager.ConstraintDefs.ForeignKeysExists) then
   raise EACRException.Create(11481,ErrorLCannotDeleteTableForeignKeys,[FTableName]);
 if (Session <> nil) then
  if ((FConstraintManager.ConstraintDefs.ForeignKeysActionsExists or
      FConstraintManager.ConstraintDefs.ForeignKeysExists)) then
   begin
     for i := 0 to FConstraintManager.ConstraintDefs.Count-1 do
      begin
       ConstraintDef := FConstraintManager.ConstraintDefs.Items[i];
       if (ConstraintDef.ConstraintType = ctFK) or
          (ConstraintDef.ConstraintType = ctFKAction) then
        if (AnsiUpperCase(TACRConstraintDefForeignKeyAction(ConstraintDef).ReferencedTableName)
            <> AnsiUpperCase(FTableName)) then
         begin
          ReferencedCursor := TACRLocalCursor.Create;
          try
           CheckForeignKeyOpenReferencedTable(ReferencedCursor,Session,FInMemory,
            TACRConstraintDefForeignKeyAction(ConstraintDef),True);
           try
             ReferencedCursor.DeleteConstraint(
              TACRConstraintDefForeignKeyAction(ConstraintDef).ReferencedFKName,
              False,
              True);
           finally
             CheckForeignKeyCloseReferencedTable(ReferencedCursor);
           end;
          finally
            ReferencedCursor.Free;
          end;
         end;
      end;
   end; // remove foreign keys and actions
 FConstraintManager.ConstraintDefs.ForeignKeysActionsExists := False;
 FConstraintManager.ConstraintDefs.ForeignKeysExists := False;
end; // DeleteTable


//------------------------------------------------------------------------------
// empty table
//------------------------------------------------------------------------------
procedure TACRTableData.EmptyTable(Cursor: TACRCursor);
begin
 if (FCursorList.Count > 0) then
  raise EACRException.Create(10072, ErrorLCannotDeleteBusyTable, [FTableName]);
 if (Cursor <> nil) then
  EmptyAllIndexes(Cursor.Session.SessionID)
 else
  EmptyAllIndexes(SYSTEM_SESSION_ID);
end; // EmptyTable


//------------------------------------------------------------------------------
// rename table
//------------------------------------------------------------------------------
procedure TACRTableData.RenameTable(Cursor: TACRCursor; NewTableName: String);
begin
 if (FCursorList.Count > 0) then
  raise EACRException.Create(10150, ErrorLCannotRenameBusyTable, [FTableName]);
 if (FConstraintManager.ConstraintDefs.ForeignKeysExists or
     FConstraintManager.ConstraintDefs.ForeignKeysActionsExists) then
  RenameTableInForeignKeys(Cursor,FTableName,NewTableName);
 FTableName := NewTableName;
end; // RenameTable


//------------------------------------------------------------------------------
// Add Foreign key
//------------------------------------------------------------------------------
procedure TACRTableData.AddForeignKey(Cursor: TACRCursor; ConstraintDef: TACRConstraintDefForeignKey);
var i,n:              Integer;
    ReferencedCursor: TACRLocalCursor;
    FKAction:         TACRConstraintDefForeignKeyAction;
begin
   if (AnsiUpperCase(FTableName) = AnsiUpperCase(ConstraintDef.ReferencedTableName)) then
    begin
     // self reference
     ReferencedCursor := TACRLocalCursor(Cursor);
    end
   else
    begin
     ReferencedCursor := TACRLocalCursor.Create;
     try
      CheckForeignKeyOpenReferencedTable(ReferencedCursor,Cursor,ConstraintDef,True);
     except
       ReferencedCursor.Free;
       raise;
     end;
    end;
   try
     n := FConstraintManager.ConstraintDefs.GetDefNumberByName(ConstraintDef.Name);
     if (n >= 0) then
      raise EACRException.Create(11588,ErrorLCannotCreateForeignKeyAlreadyExists,
       [ConstraintDef.Name,FTableName]);
     for i := 0 to High(ConstraintDef.Columns) do
      begin
       //
       n := FFieldManager.FieldDefs.GetDefNumberByName(ConstraintDef.Columns[i].ColumnName);
       if (n < 0) then
        raise EACRException.Create(11587,
          ErrorLCannotCreateForeignKeyColumnDoesNotExist,
          [ConstraintDef.Name,ConstraintDef.Columns[i].ColumnName,FTableName]);
       ConstraintDef.Columns[i].ColumnObjectID := FFieldManager.FieldDefs.Items[n].ObjectID;
      end;
     CheckForeignKeyDefinition(ConstraintDef,ReferencedCursor);
     CheckRecordsCompatibleWithForeignKey(Cursor,ReferencedCursor,ConstraintDef);
     FKAction := ReferencedCursor.TableData.CreateForeignKeyAction(ConstraintDef,FTableName,INVALID_OBJECT_ID);
     ConstraintDef.ReferencedFKName := FKAction.Name;
     ConstraintDef.ReferencedFKObjectID := FKAction.ObjectID;
     // write metadata changes - added foreign key actions
     if (ReferencedCursor <> Cursor) then
       ReferencedCursor.TableData.WriteMetadataChanges(ReferencedCursor.Session.SessionID);
     FConstraintManager.ConstraintDefs.AddFK.Assign(ConstraintDef);
   finally
     if (ReferencedCursor <> Cursor) then
      begin
       CheckForeignKeyCloseReferencedTable(ReferencedCursor);
       ReferencedCursor.Free;
      end;
   end;
end; // AddForeignKey


//------------------------------------------------------------------------------
// delete constraint (FK,FKAction or PK) and write changes to MetaData file
//------------------------------------------------------------------------------
procedure TACRTableData.DeleteConstraint(Cursor:          TACRCursor;
                                         Name:            String;
                                         Cascade:         Boolean;
                                         FKPartialDelete: Boolean
                                        );
var Num,i:            Integer;
    ConstraintDef:    TACRConstraintDef;
    ReferencedCursor: TACRLocalCursor;
begin
  Num := FConstraintManager.ConstraintDefs.GetDefNumberByName(Name);
  if (Num < 0) then
   raise EACRException.Create(11482,ErrorLConstraintNotFound,[Name,FTableName]);
  ConstraintDef := FConstraintManager.ConstraintDefs.Items[Num];
  if (ConstraintDef.ConstraintType = ctPK) then
   begin
     if (not Cascade) and (FConstraintManager.ConstraintDefs.ForeignKeysActionsExists) then
      raise EACRException.Create(11483,ErrorLCannotDeletePKFKActionsExists,[ConstraintDef.Name,FTableName]);
     if (FConstraintManager.ConstraintDefs.ForeignKeysActionsExists) then
      for i := 0 to FConstraintManager.ConstraintDefs.Count -1 do
       if (FConstraintManager.ConstraintDefs.Items[i].ConstraintType = ctFKAction) then
        DeleteConstraint(Cursor,FConstraintManager.ConstraintDefs.Items[i].Name,True,False);
     DeleteIndex(TACRConstraintDefPrimary(ConstraintDef).IndexObjectID,Cursor);
     Exit;
   end;

  if ((ConstraintDef.ConstraintType = ctFKAction) or (ConstraintDef.ConstraintType = ctFK))
      and (not FKPartialDelete) then
   begin
    ReferencedCursor := TACRLocalCursor.Create;
    try
     CheckForeignKeyOpenReferencedTable(ReferencedCursor,Cursor,TACRConstraintDefForeignKeyAction(ConstraintDef),True);
     try
       ReferencedCursor.DeleteConstraint(TACRConstraintDefForeignKeyAction(ConstraintDef).ReferencedFKName,False,True);
     finally
       CheckForeignKeyCloseReferencedTable(ReferencedCursor);
     end;
    finally
     ReferencedCursor.Free;
    end;
   end; // Delete FK
  FConstraintManager.ConstraintDefs.Delete(Num);
end; // DeleteConstraint


//------------------------------------------------------------------------------
// rename ReferenceTableName in foreign keys
//------------------------------------------------------------------------------
procedure TACRTableData.RenameReferenceTableName(Cursor: TACRCursor; OldName, NewName: String);
var
    ConstraintDef:    TACRConstraintDefForeignKeyAction;
    name:             String;
    i:                Integer;
begin
  name := AnsiUpperCase(OldName);
  for i := 0 to FConstraintManager.ConstraintDefs.Count - 1 do
   if (FConstraintManager.ConstraintDefs.Items[i].ConstraintType = ctFK) or
      (FConstraintManager.ConstraintDefs.Items[i].ConstraintType = ctFKAction) then
    begin
     ConstraintDef := TACRConstraintDefForeignKeyAction(FConstraintManager.ConstraintDefs.Items[i]);
     if (AnsiUpperCase(ConstraintDef.ReferencedTableName) = name) then
      begin
       ConstraintDef.ReferencedTableName := NewName;
      end;
     RenameConstraintFKTemporaryNames(ConstraintDef);
    end;
end; // RenameReferenceTableName


//------------------------------------------------------------------------------
// load table from stream
//------------------------------------------------------------------------------
procedure TACRTableData.LoadTableFromStream(
                        Stream: TStream
                       );
begin
 if (FCursorList.Count > 0) then
  raise EACRException.Create(10152, ErrorLCannotLoadBusyTable, [FTableName]);
end; // LoadTable


//------------------------------------------------------------------------------
// save table to stream
//------------------------------------------------------------------------------
procedure TACRTableData.SaveTableToStream(
                        Stream:               TStream;
                        CompressionAlgorithm: TACRCompressionAlgorithm;
                        CompressionMode:      Byte;
                        BlockSize:            Integer
                      );
begin
 if (FCursorList.Count > 0) then
  raise EACRException.Create(10153, ErrorLCannotSaveBusyTable, [FTableName]);
end; // SaveTable


//------------------------------------------------------------------------------
// open table
//------------------------------------------------------------------------------
procedure TACRTableData.OpenTable(Cursor: TACRCursor);
begin
 FCursorList.Add(Cursor);
 if (FTempRecordBuffer = nil) then
  FTempRecordBuffer := MemoryManager.GetMem(GetRecordBufferSize);
end;// OpenTable


//------------------------------------------------------------------------------
// close table
//------------------------------------------------------------------------------
procedure TACRTableData.CloseTable(Cursor: TACRCursor);
begin
 FCursorList.Remove(Cursor);
 if (FCursorList.Count = 0) then
  if (FTempRecordBuffer <> nil) then
   MemoryManager.FreeAndNilMem(FTempRecordBuffer);
end;// CloseTable


//------------------------------------------------------------------------------
// Rename Field by Field Index in FieldDefs
//------------------------------------------------------------------------------
procedure TACRTableData.RenameField(Cursor: TACRCursor; FieldName, NewFieldName: String);
var
  fd:             TACRFieldDef;
  i,j:            Integer;
  ConstraintDef:  TACRConstraintDefForeignKey;
  IndexDef:       TACRIndexDef;
begin
  // Check Field Exists
  fd := FFieldManager.FieldDefs.GetFieldDefByName(FieldName);
  if (fd = nil) then
    raise EACRException.Create(30343, ErrorGFieldWithNameNotFound, [FieldName]);

  // Check For Duplicate FieldName
  if (FFieldManager.FieldDefs.GetFieldDefByName(NewFieldName) <> nil ) then
   raise EACRException.Create(30344, ErrorGCannotRenameField,
                                     [FieldName, NewFieldName]);
  for i := 0 to FIndexManager.IndexDefs.Count - 1 do
   begin
    IndexDef := FIndexManager.IndexDefs.Items[i];
    for j := 0 to IndexDef.ColumnCount-1 do
     if (AnsiUpperCase(IndexDef.Columns[j].FieldName) = AnsiUpperCase(fd.Name)) then
      raise EACRException.Create(11489,ErrorLCannotRenameFieldInIndex,[fd.Name,FTableName,IndexDef.Name]);
   end;

  if (FConstraintManager.ConstraintDefs.ForeignKeysExists) then
   begin
    for i := 0 to FConstraintManager.ConstraintDefs.Count - 1 do
     if (FConstraintManager.ConstraintDefs.Items[i].ConstraintType = ctFK) or
        (FConstraintManager.ConstraintDefs.Items[i].ConstraintType = ctFKAction) then
      begin
       ConstraintDef := TACRConstraintDefForeignKey(
                          FConstraintManager.ConstraintDefs.Items[i]);
       for j := 0 to High(ConstraintDef.Columns) do
        if (ConstraintDef.Columns[j].ColumnObjectID = fd.ObjectID) then
         raise EACRException.Create(11488,ErrorLCannotRenameFieldInForeignKey,[fd.Name,FTableName,ConstraintDef.Name]);
      end;
   end;
  fd.Name := NewFieldName;
end;//RenameField


//------------------------------------------------------------------------------
// add index
//------------------------------------------------------------------------------
procedure TACRTableData.AddIndex(IndexDef: TACRIndexDef; Cursor: TACRCursor);
var
  OldPos: Pointer;
begin
 Lock;
 try
  OldPos := Cursor.SavePosition;
  FIndexManager.CreateIndex(Cursor, IndexDef);
  if ( (not IndexDef.Temporary) and (not Cursor.Temporary) ) then
    FConstraintManager.AddConstraintFromIndex(IndexDef);
 finally
  Cursor.RestorePosition(OldPos);
  Cursor.FreePosition(OldPos);
  Unlock;
 end;
end;// AddIndex


//------------------------------------------------------------------------------
// DeleteIndex
//------------------------------------------------------------------------------
procedure TACRTableData.DeleteIndex(IndexID: TACRObjectID; Cursor: TACRCursor);
var
  IndexDef: TACRIndexDef;
begin
 Lock;
 try
  // Drop Constraint
  if (Cursor <> nil) then
   begin
    IndexDef := TACRIndexDef(FIndexManager.IndexDefs.GetDefByObjectID(IndexID));
    if ( (not IndexDef.Temporary) and (not Cursor.Temporary) ) then
       FConstraintManager.DeleteConstraintForIndexID(IndexID);
    FIndexManager.DropIndex(Cursor.Session.SessionID, IndexID);
    PageManager.ApplyChanges(Cursor.Session.SessionID);
   end
  else
    begin
      FIndexManager.DropIndex(SYSTEM_SESSION_ID, IndexID);
      PageManager.ApplyChanges(SYSTEM_SESSION_ID);
    end;
 finally
  Unlock;
 end;
end;// DeleteIndex


//------------------------------------------------------------------------------
// EmptyIndex
//------------------------------------------------------------------------------
procedure TACRTableData.EmptyIndex(IndexID: TACRObjectID; SessionID: TACRSessionID);
begin
 Lock;
 try
  FIndexManager.EmptyIndex(SessionID, IndexID)
 finally
  Unlock;
 end;
end;// EmptyIndex


//------------------------------------------------------------------------------
// DeleteAllIndexes
//------------------------------------------------------------------------------
procedure TACRTableData.DeleteAllIndexes(Cursor: TACRCursor);
begin
  if (FIndexManager <> nil) then
   while (FIndexManager.IndexDefs.Count > 0) do
    DeleteIndex(FIndexManager.IndexDefs.Items[0].ObjectID, Cursor);
end;// DeleteAllIndexes


//------------------------------------------------------------------------------
// EmptyAllIndexes
//------------------------------------------------------------------------------
procedure TACRTableData.EmptyAllIndexes(SessionID: TACRSessionID);
begin
 if (FIndexManager <> nil) then
  FIndexManager.EmptyAllIndexes(SessionID);
end;// EmptyAllIndexes


//------------------------------------------------------------------------------
// return index name of the index or '' if not found
//------------------------------------------------------------------------------
function TACRTableData.FindIndex(
                                Cursor:         TACRCursor;
                                FieldNamesList, AscDescList, CaseSensitivityList: TStringList
                                        ): String;
var ID:       TACRObjectID;
    IndexDef: TACRMetaObjectDef;
begin
 Lock;
 try
  ID := FIndexManager.FindIndex(FieldNamesList, AscDescList, CaseSensitivityList);
  if (ID = INVALID_OBJECT_ID) then
   Result := ''
  else
   begin
    IndexDef := FIndexManager.IndexDefs.GetDefByObjectId(ID);
    if (IndexDef = nil) then
     raise EACRException.Create(11384,ErrorLCannotFindIndexByID,[ID]);
    Result := IndexDef.Name;
   end;
 finally
  Unlock;
 end;
end; // FindIndex


//------------------------------------------------------------------------------
// return index name
//------------------------------------------------------------------------------
function TACRTableData.FindOrCreateIndex(
                                Cursor:         TACRCursor;
                                FieldNamesList, AscDescList, CaseSensitivityList: TStringList
                                        ): String;
var ID:       TACRObjectID;
    IndexDef: TACRMetaObjectDef;
    OldPos: Pointer;
begin
 raise EACRException.Create(11392,ErrorLOperationIsNotSupported);
 Lock;
 try
  OldPos := Cursor.SavePosition;
  ID := FIndexManager.FindOrCreateIndex(Cursor, FieldNamesList, AscDescList, CaseSensitivityList);
  IndexDef := FIndexManager.IndexDefs.GetDefByObjectId(ID);
  if (IndexDef = nil) then
   raise EACRException.Create(10324,ErrorLCannotFindIndexByID,[ID]);
  Result := IndexDef.Name;
 finally
  Cursor.RestorePosition(OldPos);
  Cursor.FreePosition(OldPos);
  Unlock;
 end;
end; // FindOrCreateIndex


//------------------------------------------------------------------------------
// return true if Unique Constraint Failed
//------------------------------------------------------------------------------
function TACRTableData.IsUniqueConstraintFailed(
                                        SessionID:        TACRSessionID;
                                        IndexID:          TACRObjectID;
                                        NewRecordBuffer:  TACRRecordBuffer;
                                        OldRecordBuffer:  TACRRecordBuffer;
                                        ToInsert:         Boolean;
                                        CurrentRecordID:  TACRRecordID
                                                ): Boolean;
begin
  if (InternalGetRecordCount > 0) then
   if (ToInsert) then
     Result := not FIndexManager.OpenIndex(IndexID).CanInsertRecord(SessionID, NewRecordBuffer)
   else
     Result := not FIndexManager.OpenIndex(IndexID).CanUpdateRecord(SessionID, OldRecordBuffer, NewRecordBuffer)
  else
   Result := False;
end;//IsUniqueConstraintFailed


//------------------------------------------------------------------------------
// check Name and ReferencedFKName for tempoary FK name prefix (set by RestructureTable / RepairTable in ACRMain)
// and if found remove name prefix;
//------------------------------------------------------------------------------
procedure TACRTableData.RenameConstraintFKTemporaryNames(ConstraintDef: TACRConstraintDefForeignKeyAction);

 function CheckAndRename(const Name: TACRObjectName): TACRObjectName;
 var i,l,j: Integer;
     s:     String;
 begin
  Result := Name;
  s := AnsiUpperCase(Name);
  i := Pos(ACRConstraintFKTemporaryNamePrefix,s);
  l := Length(ACRConstraintFKTemporaryNamePrefix);
  j := Length(Name);
  if (i = 1) then
   begin
    if (j - l <= 0) then
     raise EACRException.Create(11585,ErrorLInvalidFKTempName,[Name])
    else
     Result := Copy(Name,l+1,j-l);
   end;
 end; // CheckAndRename

begin
 ConstraintDef.Name := CheckAndRename(ConstraintDef.Name);
 ConstraintDef.ReferencedFKName := CheckAndRename(ConstraintDef.ReferencedFKName);
end; // RenameConstraintFKTemporaryNames


//------------------------------------------------------------------------------
// raise an exception if record not compatible with foreign key was found
//------------------------------------------------------------------------------
procedure TACRTableData.CheckRecordsCompatibleWithForeignKey(
                            Cursor:           TACRCursor;
                            ReferencedCursor: TACRCursor;
                            ConstraintDef:    TACRConstraintDefForeignKey);
var
    Buf, CurrentRecordBuffer:  TACRRecordBuffer;
    OldPos:                    Pointer;

  function IsConstraintFailed: Boolean;
  var
      SearchExpression:     TACRExpression;
      Failed:               Boolean;
  begin
    Result := not IsForeignKeyNullValuesOK(Cursor,ConstraintDef,Failed);
    if (Failed) then
     Result := True
    else
    if (Result) then
     begin
       SearchExpression := CheckForeignKeyBuildSearchExpression(
                              ReferencedCursor,Cursor,ConstraintDef);
       try
         // search record
         Result := not TACRLocalCursor(ReferencedCursor).TableData.FindRecord(ReferencedCursor,
                              SearchExpression,
                              nil,True,True,nil,nil,False,True);
       finally
         SearchExpression.Free;
       end;
     end; // if not null and not failed
  end; // IsConstraintFailed

begin
  Buf := Cursor.CurrentRecordBuffer;
  CurrentRecordBuffer := Cursor.AllocateRecordBuffer;
  Cursor.CurrentRecordBuffer := CurrentRecordBuffer;
  OldPos := Cursor.SavePosition;
  Cursor.FirstPosition := True;
  Cursor.LastPosition := False;
  try
    while (Cursor.GetRecordBuffer(grmNext) = grrOK) do
     begin
      if (IsConstraintFailed) then
       raise EACRException.Create(11586,ErrorLConstraintForeignKeyViolated,[ConstraintDef.Name,FTableName]);
     end;
  finally
    Cursor.FreeRecordBuffer(CurrentRecordBuffer);
    Cursor.CurrentRecordBuffer := Buf;
    Cursor.RestorePosition(OldPos);
    Cursor.FreePosition(OldPos);
  end;
end; // CheckRecordsCompatibleWithForeignKey


//------------------------------------------------------------------------------
// return true if current record in Cursor does not violate foreign key constraint
// (have null field values corresponding match type)
//------------------------------------------------------------------------------
function TACRTableData.IsForeignKeyNullValuesOK(
                                    Cursor:         TACRCursor;
                                    ConstraintDef:  TACRConstraintDefForeignKey;
                                    var Failed:     Boolean
                                  ): Boolean;
var i, FieldNo: Integer;
    value:      TACRVariant;
    NullFound:  Boolean;
begin
 value := TACRVariant.Create;
 try
  Failed := False;
  NullFound := False;
  Result := (ConstraintDef.MatchType <> cfkmtDefault);
  for i := 0 to High(ConstraintDef.Columns) do
   begin
     FieldNo := FieldManager.FieldDefs.GetDefNumberByObjectId(ConstraintDef.Columns[i].ColumnObjectID);
     if (FieldNo < 0) then
       raise EACRException.Create(11422,ErrorLForeignKeyInvalidColumnID,[ConstraintDef.Name,FTableName,i,ConstraintDef.Columns[i].ColumnObjectID]);
     Cursor.GetFieldValue(value,FieldNo,True,False);
     case ConstraintDef.MatchType of
      cfkmtDefault:
       begin
        if (value.IsNull) then
         begin
          Result := True;
          break;
         end;
       end;
      cfkmtFull:
       begin
        if (not value.IsNull) then
         begin
          Result := False;
          if (NullFound) then
           Failed := True;
         end
        else
         begin
          if (not Result) then
           Failed := True;
          NullFound := True;
         end;
       end;
      cfkmtPartial:
       begin
        if (not value.IsNull) then
         begin
          Result := False;
          break;
         end;
       end
      else
       raise EACRException.Create(11421,ErrorLForeignKeyInvalidMatchType,[ConstraintDef.Name,FTableName,Byte(ConstraintDef.MatchType)]);
     end; // case
   end; // for
 finally
   value.Free;
 end;
end; // IsForeignKeyNullValuesOK


//------------------------------------------------------------------------------
// referenced cursor open
//------------------------------------------------------------------------------
procedure TACRTableData.CheckForeignKeyOpenReferencedTable(
                                    ReferencedCursor: TACRCursor;
                                    Cursor:           TACRCursor;
                                    ConstraintDef:    TACRConstraintDefForeignKeyAction;
                                    Exclusive:        Boolean
                                      );
begin
 ReferencedCursor.Session := Cursor.Session;
 ReferencedCursor.InMemory := Cursor.InMemory;
 ReferencedCursor.FTableName := ConstraintDef.ReferencedTableName;
 ReferencedCursor.IsDesignMode := Cursor.IsDesignMode;
 ReferencedCursor.Exclusive := Exclusive;
 ReferencedCursor.ReadOnly := True;
 if (not ReferencedCursor.Session.TableExists(ReferencedCursor.FTableName)) then
  raise EACRException.Create(11435,ErrorLInvalidFKReferencedTable,[ConstraintDef.Name,FTableName,ReferencedCursor.FTableName]);
 ReferencedCursor.OpenTableByFieldDefs(nil,nil,nil);
 ReferencedCursor.InternalInitFieldDefs;
 ReferencedCursor.InternalFirst;
 ReferencedCursor.CurrentRecordBuffer := ReferencedCursor.AllocateRecordBuffer;
end; // CheckForeignKeyOpenReferencedTable


//------------------------------------------------------------------------------
// referenced cursor open
//------------------------------------------------------------------------------
procedure TACRTableData.CheckForeignKeyOpenReferencedTable(
                                    ReferencedCursor: TACRCursor;
                                    Session:          TACRBaseSession;
                                    InMemory:         Boolean;
                                    ConstraintDef:    TACRConstraintDefForeignKeyAction;
                                    Exclusive:        Boolean
                                      );
begin
 ReferencedCursor.Session := Session;
 ReferencedCursor.InMemory := InMemory;
 ReferencedCursor.FTableName := ConstraintDef.ReferencedTableName;
 ReferencedCursor.IsDesignMode := False;
 ReferencedCursor.Exclusive := Exclusive;
 ReferencedCursor.ReadOnly := True;
 if (not ReferencedCursor.Session.TableExists(ReferencedCursor.FTableName)) then
  raise EACRException.Create(11487,ErrorLInvalidFKReferencedTable,[ConstraintDef.Name,FTableName,ReferencedCursor.FTableName]);
 ReferencedCursor.OpenTableByFieldDefs(nil,nil,nil);
 ReferencedCursor.InternalInitFieldDefs;
 ReferencedCursor.InternalFirst;
 ReferencedCursor.CurrentRecordBuffer := ReferencedCursor.AllocateRecordBuffer;
end; // CheckForeignKeyOpenReferencedTable


//------------------------------------------------------------------------------
// close referenced cursor
//------------------------------------------------------------------------------
procedure TACRTableData.CheckForeignKeyCloseReferencedTable(Cursor: TACRCursor);
var buffer: PChar;
begin
 buffer := Cursor.CurrentRecordBuffer;
 Cursor.FreeRecordBuffer(buffer);
 Cursor.CurrentRecordBuffer := nil;
 Cursor.CloseTable;
end; // CheckForeignKeyCloseReferencedTable


//------------------------------------------------------------------------------
// build search expression for the foreign key check
//------------------------------------------------------------------------------
function TACRTableData.CheckForeignKeyBuildSearchExpression(
                                    ReferencedCursor: TACRCursor;
                                    Cursor:           TACRCursor;
                                    ConstraintDef:    TACRConstraintDefForeignKey
                                      ): TACRExpression;
var
  PrimaryIndex:   TACRIndexDef;
begin
  PrimaryIndex := ReferencedCursor.IndexDefs.FindPrimaryIndex;
  if (PrimaryIndex = nil) then
   raise EACRException.Create(11424,ErrorLForeignKeyNoPrimaryIndex,[ConstraintDef.Name,FTableName,ReferencedCursor.TableName]);
  if (PrimaryIndex.ColumnCount < Length(ConstraintDef.Columns)) then
   raise EACRException.Create(11425,ErrorLForeignKeyPrimaryIndexInvalidColumnCount,[ConstraintDef.Name,FTableName,PrimaryIndex.Name,ReferencedCursor.TableName]);
  Result := TACRExpression.Create;
  Result.PrepareForeignKeyCheck(ReferencedCursor,Cursor,ConstraintDef,PrimaryIndex);
end; // CheckForeignKeyBuildSearchExpression


//------------------------------------------------------------------------------
// build foreign key action filter and return true if there is no detail records
//------------------------------------------------------------------------------
function TACRTableData.CheckForeignKeyActionBuildSearchExpression(
                                    ReferencedCursor: TACRCursor;
                                    Cursor:           TACRCursor;
                                    ConstraintDef:    TACRConstraintDefForeignKeyAction;
                                    ToUpdate:         Boolean
                                      ): Boolean;
var
  PrimaryIndex:   TACRIndexDef;
  Buffer:         PChar;
begin
  PrimaryIndex := FIndexManager.IndexDefs.FindPrimaryIndex;
  if (PrimaryIndex = nil) then
   raise EACRException.Create(11457,ErrorLForeignKeyNoPrimaryIndex,[ConstraintDef.ReferencedFKName,ReferencedCursor.TableName,FTableName]);
  if (PrimaryIndex.ColumnCount < Length(ConstraintDef.Columns)) then
   raise EACRException.Create(11458,ErrorLForeignKeyPrimaryIndexInvalidColumnCount,[ConstraintDef.ReferencedFKName,ReferencedCursor.TableName,PrimaryIndex.Name,FTableName]);
  if (ToUpdate) then
   begin
    Buffer := Cursor.CurrentRecordBuffer;
    Cursor.CurrentRecordBuffer := Cursor.EditRecordBuffer;
   end;
  try
    ReferencedCursor.SQLFilterExpression := TACRExpression.Create;
    TACRExpression(ReferencedCursor.SQLFilterExpression).PrepareForeignKeyActionFilter(
      ReferencedCursor,Cursor,ConstraintDef,PrimaryIndex);
    if (not ReferencedCursor.InMemory) then
     TACRLocalCursor(ReferencedCursor).TableData.GetRecordCount(ReferencedCursor,False);
    TACRLocalCursor(ReferencedCursor).TableData.BuildCursorRecordBitmap(ReferencedCursor);
    Result := (TACRRecordBitmap(TACRLocalCursor(ReferencedCursor).RecordBitmap).GetRecordCount <= 0);
  finally
   if (ToUpdate) then
      Cursor.CurrentRecordBuffer := Buffer;
  end;
end; // CheckForeignKeyActionBuildSearchExpression


//------------------------------------------------------------------------------
// check field definitions for CreateTable
//------------------------------------------------------------------------------
procedure TACRTableData.CheckFieldDefinitions(FieldDefs: TACRFieldDefs);
var i,k:   Integer;
    Name:  String;
begin
 if (FieldDefs.Count <= 0) then
  raise EACRException.Create(11429,ErrorLCreateTableNoFields,[FTableName]);
 for i := 0 to FieldDefs.Count - 1 do
  begin
   if (FieldDefs.Items[i].Name = '') then
    raise EACRException.Create(11432,ErrorLCreateTableNoFieldName,[FTableName]);
   Name := AnsiUpperCase(FieldDefs.Items[i].Name);
   for k := 0 to FieldDefs.Count - 1 do
    if (k <> i) then
     if (AnsiUpperCase(FieldDefs.Items[k].Name) = Name) then
      raise EACRException.Create(11430,ErrorLCreateTableDuplicateFieldName,[FTableName,Name]);
  end;
end; // CheckFieldDefinitions


//------------------------------------------------------------------------------
// check for duplicates in index definitions and for multiple primary keys for CreateTable
//------------------------------------------------------------------------------
procedure TACRTableData.CheckIndexDefinitions(IndexDefs: TACRIndexDefs);
var i,j,k: Integer;
    Name:  String;
begin
 j := 0;
 for i := 0 to IndexDefs.Count - 1 do
  begin
   if (IndexDefs.Items[i].Name = '') then
    raise EACRException.Create(11433,ErrorLCreateTableNoIndexName,[FTableName]);
   Name := AnsiUpperCase(IndexDefs.Items[i].Name);
   if (IndexDefs.Items[i].ColumnCount <= 0) then
    raise EACRException.Create(11434,ErrorLCreateTableNoFieldsInIndex,[FTableName,Name]);
   for k := 0 to IndexDefs.Count - 1 do
    if (k <> i) then
     if (AnsiUpperCase(IndexDefs.Items[k].Name) = Name) then
      raise EACRException.Create(11428,ErrorLCreateTableDuplicateIndexName,[FTableName,Name]);
   if (IndexDefs.Items[i].Primary) then
    Inc(j);
  end;
 if (j > 1) then
  raise EACRException.Create(11427,ErrorLCreateTableMultiplePrimaryIndexes,[FTableName]);
end; // CheckPrimaryIndexDefinitions


//------------------------------------------------------------------------------
// check constraint names
//------------------------------------------------------------------------------
procedure TACRTableData.CheckConstraintDefinitions(ConstraintDefs: TACRConstraintDefs);
var i,k:   Integer;
    Name:  String;
begin
 for i := 0 to ConstraintDefs.Count - 1 do
  begin
   Name := AnsiUpperCase(ConstraintDefs.Items[i].Name);
   if (Name = '') then
    continue;
   for k := 0 to ConstraintDefs.Count - 1 do
    if (k <> i) then
     if (AnsiUpperCase(ConstraintDefs.Items[k].Name) = Name) then
      raise EACRException.Create(11431,ErrorLCreateTableDuplicateConstraintName,[FTableName,Name]);
  end;
end; // CheckConstraintDefinitions


//------------------------------------------------------------------------------
// create foreign key action
//------------------------------------------------------------------------------
function TACRTableData.CreateForeignKeyAction(
                              ConstraintDef:           TACRConstraintDefForeignKey;
                              ReferencedTableName:     String;
                              ReferencedTableObjectID: TACRObjectID
                               ): TACRConstraintDefForeignKeyAction;
begin
  Result := FConstraintManager.AddForeignKeyAction(ConstraintDef,
                ReferencedTableName,ReferencedTableObjectID);
end; // CreateForeignKeyAction


//------------------------------------------------------------------------------
// write metadata changes
//------------------------------------------------------------------------------
procedure TACRTableData.WriteMetadataChanges(SessionID: TACRSessionID);
begin
; // do nothing
end; // WriteMetadataChanges


procedure TACRTableData.CheckForeignKeyDefinition(ConstraintDef: TACRConstraintDefForeignKey; ReferencedCursor: TACRCursor);
var
    FieldDefFK:       TACRFieldDef;
    FieldDefIndex:    TACRFieldDef;
    PrimaryIndex:     TACRIndexDef;
    j,h:              Integer;
begin
 PrimaryIndex := ReferencedCursor.IndexDefs.FindPrimaryIndex;
 if (PrimaryIndex = nil) then
  raise EACRException.Create(11442,ErrorLCreateTableInvalidFKDefNoPrimaryIndex,[ConstraintDef.Name,FTableName,ConstraintDef.ReferencedTableName]);
 if (PrimaryIndex.ColumnCount <> Length(ConstraintDef.Columns)) then
  raise EACRException.Create(11443,ErrorLCreateTableInvalidFKDefPrimaryIndexColumnCount,[ConstraintDef.Name,FTableName,ConstraintDef.ReferencedTableName]);
 h := High(ConstraintDef.Columns);
 for j := 0 to h do
  begin
   FieldDefFK := FFieldManager.FieldDefs.GetFieldDefByName(ConstraintDef.Columns[j].ColumnName);
   if (FieldDefFK = nil) then
    raise EACRException.Create(11444,ErrorLCreateTableInvalidFKDefInvalidColumnName,[ConstraintDef.Name,FTableName,ConstraintDef.Columns[j].ColumnName]);
   FieldDefIndex := ReferencedCursor.FFieldDefs.GetFieldDefByName(PrimaryIndex.Columns[j].FieldName);
   if (FieldDefIndex = nil) then
    raise EACRException.Create(11445,ErrorLCreateTableInvalidFKDefInvalidPrimaryIndexColumnName,[ConstraintDef.Name,FTableName,PrimaryIndex.Columns[j].FieldName,ConstraintDef.ReferencedTableName]);
   if (GetCommonDataType(FieldDefFK.BaseFieldType,FieldDefIndex.BaseFieldType) = bftUnknown) or
      (not IsConvertableFieldType(FieldDefFK.AdvancedFieldType)) or
      (not IsConvertableFieldType(FieldDefIndex.AdvancedFieldType)) then
    raise EACRException.Create(11446,ErrorLCreateTableInvalidFKDefTypesMismatch,[ConstraintDef.Name,FTableName,ConstraintDef.Columns[j].ColumnName,PrimaryIndex.Columns[j].FieldName,ConstraintDef.ReferencedTableName]);
  end;
end; // CheckForeignKeyDefinition


//------------------------------------------------------------------------------
// checks foreign key definitions and creates foreign key actions in master tables
//------------------------------------------------------------------------------
procedure TACRTableData.CheckForeignKeyDefinitionsAndCreateForeignKeyActions(Cursor: TACRCursor);
var i,J:              Integer;
    ReferencedCursor: TACRLocalCursor;
    MasterTablesList: TList;
    ConstraintDef:    TACRConstraintDefForeignKey;
    FKAction:         TACRConstraintDefForeignKeyAction;
    bOK:              Boolean;
begin
  bOK := False;
  MasterTablesList := TList.Create;
  try
    for i := 0 to FConstraintManager.ConstraintDefs.Count - 1 do
     if (FConstraintManager.ConstraintDefs[i].ConstraintType = ctFK) then
      begin
       ConstraintDef := TACRConstraintDefForeignKey(FConstraintManager.ConstraintDefs[i]);
       if (AnsiUpperCase(FTableName) = AnsiUpperCase(ConstraintDef.ReferencedTableName)) then
        begin
         // self reference
         ReferencedCursor := TACRLocalCursor(Cursor);
        end
       else
        begin
         ReferencedCursor := nil;
         for j := 0 to MasterTablesList.Count-1 do
          if (AnsiUpperCase(TACRLocalCursor(MasterTablesList.Items[j]).TableName) =
              AnsiUpperCase(ConstraintDef.ReferencedTableName)) then
            begin
             ReferencedCursor := TACRLocalCursor(MasterTablesList.Items[j]);
             break;
            end;
         if (ReferencedCursor = nil) then
          begin
           ReferencedCursor := TACRLocalCursor.Create;
           try
            CheckForeignKeyOpenReferencedTable(ReferencedCursor,Cursor,ConstraintDef,True);
           except
             ReferencedCursor.Free;
             raise;
           end;
           MasterTablesList.Add(ReferencedCursor);
          end;
        end;
       CheckForeignKeyDefinition(ConstraintDef,ReferencedCursor);
       FKAction := ReferencedCursor.TableData.CreateForeignKeyAction(ConstraintDef,FTableName,INVALID_OBJECT_ID);
       ConstraintDef.ReferencedFKName := FKAction.Name;
       ConstraintDef.ReferencedFKObjectID := FKAction.ObjectID;
      end;
    bOK := True;
  finally
    for i := 0 to MasterTablesList.Count - 1 do
     begin
      try
       ReferencedCursor := TACRLocalCursor(MasterTablesList.Items[i]);
       // write metadata changes - added foreign key actions
       if (bOK) then
        ReferencedCursor.TableData.WriteMetadataChanges(ReferencedCursor.Session.SessionID);
       CheckForeignKeyCloseReferencedTable(ReferencedCursor);
       ReferencedCursor.Free;
      except
       ;
      end;
     end;
    MasterTablesList.Free;
  end;
end; // CheckForeignKeyDefinitionsAndCreateForeignKeyActions


//------------------------------------------------------------------------------
// rename table in foreign keys
//------------------------------------------------------------------------------
procedure TACRTableData.RenameTableInForeignKeys(
                                   Cursor:       TACRCursor;
                                   OldTableName: String;
                                   NewTableName: String
                                  );
var i,j:              Integer;
    ReferencedCursor: TACRLocalCursor;
    MasterTablesList: TList;
    ConstraintDef:    TACRConstraintDefForeignKeyAction;
begin
  MasterTablesList := TList.Create;
  try
    for i := 0 to FConstraintManager.ConstraintDefs.Count - 1 do
     if (FConstraintManager.ConstraintDefs[i].ConstraintType = ctFK) or
        (FConstraintManager.ConstraintDefs[i].ConstraintType = ctFKAction) then
      begin
       ConstraintDef := TACRConstraintDefForeignKeyAction(FConstraintManager.ConstraintDefs[i]);
       if (AnsiUpperCase(OldTableName) <>
           AnsiUpperCase(ConstraintDef.ReferencedTableName)) then
        begin
         ReferencedCursor := nil;
         for j := 0 to MasterTablesList.Count-1 do
          if (AnsiUpperCase(TACRLocalCursor(MasterTablesList.Items[j]).TableName) =
              AnsiUpperCase(ConstraintDef.ReferencedTableName)) then
            begin
             ReferencedCursor := TACRLocalCursor(MasterTablesList.Items[j]);
             break;
            end;
         if (ReferencedCursor = nil) then
          begin
            ReferencedCursor := TACRLocalCursor.Create;
             try
              CheckForeignKeyOpenReferencedTable(ReferencedCursor,Cursor,ConstraintDef,True);
             except
               ReferencedCursor.Free;
               raise;
             end;
             MasterTablesList.Add(ReferencedCursor);
          end;
        end;
      end;
    for i := 0 to MasterTablesList.Count - 1 do
     begin
      ReferencedCursor := TACRLocalCursor(MasterTablesList.Items[i]);
      ReferencedCursor.TableData.RenameReferenceTableName(ReferencedCursor,OldTableName,NewTableName);
     end;
    RenameReferenceTableName(Cursor,OldTableName,NewTableName);
  finally
    for i := 0 to MasterTablesList.Count - 1 do
     begin
      try
       ReferencedCursor := TACRLocalCursor(MasterTablesList.Items[i]);
       CheckForeignKeyCloseReferencedTable(ReferencedCursor);
       ReferencedCursor.Free;
      except
       ;
      end;
     end;
    MasterTablesList.Free;
  end;
end; // RenameTableInForeignKeys


//------------------------------------------------------------------------------
// update detail records for foreign key actions
//------------------------------------------------------------------------------
procedure TACRTableData.ExecuteForeignKeyActionsUpdateDetailRecords(
                                    ReferencedCursor: TACRCursor;
                                    Cursor:           TACRCursor;
                                    ConstraintDef:    TACRConstraintDefForeignKeyAction;
                                    Action:           TACRConstraintForeignKeyAction
                                                     );
var
  PrimaryIndex:   TACRIndexDef;
  FieldNames:             TStringList;
  values:         array of TACRVariant;
  i,h,FieldNo:    Integer;
  FieldDef:       TACRFieldDef;
begin
  if (Action = cfkaCascade) then
   begin
    PrimaryIndex := FIndexManager.IndexDefs.FindPrimaryIndex;
    if (PrimaryIndex = nil) then
     raise EACRException.Create(11460,ErrorLForeignKeyNoPrimaryIndex,[ConstraintDef.ReferencedFKName,ReferencedCursor.TableName,FTableName]);
   end;
  h := High(ConstraintDef.Columns);
  FieldNames := TStringList.Create;
  SetLength(values,h+1);
  for i := 0 to h do
   values[i] := TACRVariant.Create;
  try
    for i := 0 to h do
     begin
      FieldDef := ReferencedCursor.FFieldDefs.GetFieldDefByName(
                    ConstraintDef.Columns[i].ColumnName);
      if (FieldDef = nil) then
       raise EACRException.Create(11458,ErrorLCreateTableInvalidFKDefInvalidColumnName,
                [ConstraintDef.Name,FTableName,ConstraintDef.Columns[i].ColumnName]);
      FieldNames.Add(FieldDef.Name);
      values[i].SetNull();
      if (Action = cfkaSetNull) then
       values[i].SetNull(FieldDef.BaseFieldType)
      else
      if (Action = cfkaSetDefault) then
       values[i].Assign(FieldDef.DefaultValue)
      else
      if (Action = cfkaCascade) then
       begin
        FieldNo := FFieldManager.FieldDefs.GetDefNumberByName(
                    PrimaryIndex.Columns[i].FieldName);
        if (FieldNo < 0) then
         raise EACRException.Create(11461,ErrorLCreateTableInvalidFKDefInvalidColumnName,
                [ConstraintDef.Name,FTableName,PrimaryIndex.Columns[i].FieldName]);
        Cursor.GetFieldValue(values[i],FieldNo,True,False);
       end
      else
       raise EACRException.Create(11462,ErrorLForeignKeyInvalidAction,
                      [Byte(Action),FTableName,ConstraintDef.ReferencedFKName,ConstraintDef.ReferencedTableName])
     end;
   ReferencedCursor.UpdateVisibleRecords(FieldNames,values,(Action = cfkaCascade));
  finally
    for i := 0 to h do
     values[i].Free;
    SetLength(values,0);
    FieldNames.Free;
  end;
end; // ExecuteForeignKeyActionsUpdateDetailRecords


//------------------------------------------------------------------------------
// return true if current record in Cursor violates foreign key constraint
//------------------------------------------------------------------------------
function TACRTableData.IsForeignKeyConstraintFailed(
                                    Cursor:         TACRCursor;
                                    ConstraintDef:  TACRConstraintDefForeignKey
                                      ): Boolean;
var ReferencedCursor: TACRCursor;
    SearchExpression: TACRExpression;
    Failed:           Boolean;
begin
  Result := not IsForeignKeyNullValuesOK(Cursor,ConstraintDef,Failed);
  if (Failed) then
   Result := True
  else
    if (Result) then
     begin
      Result := False;
      ReferencedCursor := TACRLocalCursor.Create;
      try
       try
        CheckForeignKeyOpenReferencedTable(ReferencedCursor,Cursor,ConstraintDef,False);
       except
        on E: Exception do
         raise EACRException.Create(11423,ErrorLForeignKeyCannotOpenReferencedTable,
                [ConstraintDef.Name,FTableName,e.Message]);
       end;
       try
         SearchExpression := CheckForeignKeyBuildSearchExpression(
                                ReferencedCursor,Cursor,ConstraintDef);
         try
           // search record
           Result := not TACRLocalCursor(ReferencedCursor).TableData.FindRecord(ReferencedCursor,
                                SearchExpression,
                                nil,True,True,nil,nil,False,True);
         finally
           SearchExpression.Free;
         end;
       finally
         CheckForeignKeyCloseReferencedTable(ReferencedCursor);
       end;
      finally
        ReferencedCursor.Free;
      end;
     end; // if not null and not failed
end; // IsForeignKeyConstraintFailed


//------------------------------------------------------------------------------
// execute foreign key actions
//------------------------------------------------------------------------------
procedure TACRTableData.ExecuteForeignKeyActions(
                                  Cursor:   TACRCursor;
                                  ToUpdate: Boolean // update or delete
                          );
var i:                  Integer;
    ConstraintDef:      TACRConstraintDefForeignKeyAction;
    Action:             TACRConstraintForeignKeyAction;
    ReferencedCursor:   TACRLocalCursor;
    NoRecords:          Boolean;
begin
  for i := 0 to FConstraintManager.ConstraintDefs.Count - 1 do
   if (FConstraintManager.ConstraintDefs.Items[i].ConstraintType = ctFKAction) then
    begin
     ConstraintDef := TACRConstraintDefForeignKeyAction(FConstraintManager.ConstraintDefs.Items[i]);
     if (ToUpdate) then
      Action := ConstraintDef.UpdateAction
     else
      Action := ConstraintDef.DeleteAction;
     if (Action = cfkaNoAction) then
      continue;
     ReferencedCursor := TACRLocalCursor.Create;
     try
       CheckForeignKeyOpenReferencedTable(ReferencedCursor,Cursor,ConstraintDef,False);
       try
         NoRecords := CheckForeignKeyActionBuildSearchExpression(ReferencedCursor,Cursor,ConstraintDef,ToUpdate);
         if (not NoRecords)then
           case Action of
            cfkaNoAction: ;
            cfkaDefault:
              begin
                if (ToUpdate) then
                 raise EACRException.Create(11454,ErrorLForeignKeyUpdateInMasterTableProhibited,
                        [FTableName,ConstraintDef.ReferencedFKName,ConstraintDef.ReferencedTableName])
                else
                 raise EACRException.Create(11455,ErrorLForeignKeyDeleteInMasterTableProhibited,
                        [FTableName,ConstraintDef.ReferencedFKName,ConstraintDef.ReferencedTableName])
              end;
            cfkaCascade, cfkaSetNull, cfkaSetDefault:
             begin
              if (not ToUpdate) and (Action = cfkaCascade) then
               ReferencedCursor.DeleteVisibleRecords
              else
               begin
                ExecuteForeignKeyActionsUpdateDetailRecords(ReferencedCursor,
                  Cursor,ConstraintDef,Action);
               end; // not cascade delete
             end
            else
             raise EACRException.Create(11456,ErrorLForeignKeyInvalidAction,
                      [Byte(Action),FTableName,ConstraintDef.ReferencedFKName,ConstraintDef.ReferencedTableName])
           end; // case Action
       finally
         CheckForeignKeyCloseReferencedTable(ReferencedCursor);
       end;
     finally
       ReferencedCursor.Free;
     end;
    end; // FKAction
end; // ExecuteForeigKeyActions


//------------------------------------------------------------------------------
// build record bitmap for cursor
//------------------------------------------------------------------------------
procedure TACRTableData.BuildCursorRecordBitmap(Cursor: TACRCursor);
var Buffer:         TACRRecordBuffer;
    OldPos:         Pointer;
    OldBuffer:      TACRRecordBuffer;
begin
 Lock;
 try
   TACRRecordBitmap(Cursor.RecordBitmap).Active := False;
   TACRRecordBitmap(Cursor.RecordBitmap).Indexed := Cursor.IsIndexApplied;
   TACRRecordBitmap(Cursor.RecordBitmap).PrepareBitmapForActivation;
   if (InternalGetRecordCount > 0) then
    begin
     OldBuffer := Cursor.CurrentRecordBuffer;
     OldPos := Cursor.SavePosition;
     Cursor.CurrentRecordBuffer := MemoryManager.AllocMem(GetRecordBufferSize);
     Cursor.FirstPosition := True;
     Cursor.LastPosition := False;
     try
      // build cursor bitmap
      FindRecord(Cursor,nil,nil,True,True,nil,Cursor.RecordBitmap,True);
{
// CURSOR DISTINCT IS NOT SUPPORTED 
      // apply distinct 
      if (Cursor.DistinctFieldCount > 0) then
        FIndexManager.OpenIndex(Cursor.IndexID).
            ApplyDistinctToRecordBitmap(Cursor.Session.SessionID, 
                                        TACRRecordBitmap(Cursor.RecordBitmap)); 
}
      TACRRecordBitmap(Cursor.RecordBitmap).Active := True;
     finally 
      Buffer := Cursor.CurrentRecordBuffer;
      MemoryManager.FreeAndNilMem(Buffer); 
      Cursor.CurrentRecordBuffer := OldBuffer;
      Cursor.RestorePosition(OldPos); 
      Cursor.FreePosition(OldPos);
     end; 
    end;
 finally 
  Unlock; 
 end;
end; // BuildCursorRecordBitmap 
 
 
//------------------------------------------------------------------------------ 
// return bitmap size 
//------------------------------------------------------------------------------
function TACRTableData.GetBitmapSize(SessionID: TACRSessionID): TACRRecordNo;
begin
  Result := InternalGetRecordCount;
end; // GetBitmapSize
 
	
//------------------------------------------------------------------------------ 
// return true if record is in specified range
//------------------------------------------------------------------------------ 
function TACRTableData.IsRecordInRange(Cursor: TACRCursor): Boolean; 
var Res:  Integer;
begin 
  Result := True;
  if (Cursor.RangeStartBuffer <> nil) then 
    begin 
     Res := FIndexManager.OpenIndex(Cursor.IndexID).
              CompareRecordBuffersByIndex(
                Cursor.CurrentRecordBuffer, 
                Cursor.RangeStartBuffer, 
                Cursor.RangeStartKeyFieldCount); 
     if (Cursor.RangeStartExclusive) then 
      Result := (Res > 0) 
     else 
      Result := (Res >= 0); 
    end;
  if (Result) then
   if (Cursor.RangeEndBuffer <> nil) then 
    begin
     Res := FIndexManager.OpenIndex(Cursor.IndexID).
              CompareRecordBuffersByIndex( 
                Cursor.CurrentRecordBuffer,
                Cursor.RangeEndBuffer,
                Cursor.RangeEndKeyFieldCount); 
     if (Cursor.RangeEndExclusive) then 
      Result := (Res < 0) 
     else
      Result := (Res <= 0);
    end; 
end; // IsRecordInRange
	
 
//------------------------------------------------------------------------------ 
// return true if record is not equal to the previous record
//------------------------------------------------------------------------------
function TACRTableData.IsRecordInDistinct(Cursor: TACRCursor): Boolean;
var Res:  Integer; 
begin 
  Res := FIndexManager.OpenIndex(Cursor.IndexID). 
          CompareRecordBuffersByIndex( 
            Cursor.CurrentRecordBuffer,
            Cursor.DistinctRecordBuffer, 
            Cursor.DistinctFieldCount); 
  Result := (Res <> 0); 
end; // IsRecordInDistinct
	
	
//------------------------------------------------------------------------------ 
// return true if record is visible (with applied filters, ranges, OnFilterRecord)
//------------------------------------------------------------------------------
function TACRTableData.IsRecordVisible(Cursor: TACRCursor): Boolean;
var 
  aDataset:         Pointer; 
  aRecordBuffer:    TACRRecordBuffer;
  FunctionAddress:  Pointer;
begin
 Result := True;
 if (Result) then
  Result := IsRecordInRange(Cursor);
 if (Result) then 
   if (Cursor.FilterExpression <> nil) then
    begin
     TACRExpression(Cursor.FilterExpression).AssignCursorBuffer(Cursor.CurrentRecordBuffer); 
     Result := TACRExpression(Cursor.FilterExpression).GetResult;
    end;
 if (Result) then 
   if (Cursor.SQLFilterExpression <> nil) then 
    begin 
     TACRExpression(Cursor.SQLFilterExpression).AssignCursorBuffer(Cursor.CurrentRecordBuffer); 
     Result := TACRExpression(Cursor.SQLFilterExpression).GetResult;
    end; 
 if (Result) then 
  if (Pointer(Cursor.FilterRecord) <> nil) then
   begin
    aDataset := Cursor.Dataset;
    aRecordBuffer := Cursor.CurrentRecordBuffer;
    // Borland parameters transmitting bug fix 
    // Get Function address from Virtual Method Table 
    FunctionAddress := Pointer(PCardinal(Pointer(Cardinal(@Cursor.FilterRecord)))^);
    asm
      MOV    ECX, aDataset                // Dataset 
      MOV    EDX, aRecordBuffer           // Buffer 
      MOV    EAX, FunctionAddress         // Address of Function 
      CALL   EAX                          // Call Function
      MOV    Result, AL                   // Save Result
    end;
   end;
end; // IsRecordVisible
 
	
//------------------------------------------------------------------------------ 
// return true if conditions are incompatible
//------------------------------------------------------------------------------
procedure TACRTableData.MergeAndCheckSearchConditionsCompatibility(
                                var Condition1:            TACRScanSearchCondition;
                                var Condition2:            TACRScanSearchCondition;
                                out Incompatible:          Boolean;
                                out HaveBeenMerged:        Boolean
                                           );
var ResMin:         Integer;
    IndexID:        TACRObjectID;
    MinFieldCount:  Integer;
    MaxFieldCount:  Integer;
    NewCondition:   TACRScanSearchCondition;

 procedure SetFirstCondition;
 begin
  HaveBeenMerged := True;
  NewCondition.KeyFieldCount := Condition1.KeyFieldCount;
  NewCondition.KeyRecordBuffer := Condition1.KeyRecordBuffer;
  NewCondition.Condition := Condition1.Condition;
 end; // SetFirstCondition

 procedure SetSecondCondition;
 begin
  HaveBeenMerged := True;
  NewCondition.KeyFieldCount := Condition2.KeyFieldCount;
  NewCondition.KeyRecordBuffer := Condition2.KeyRecordBuffer;
  NewCondition.Condition := Condition2.Condition;
 end; // SetMaxKFCCondition

 procedure SetMinKFCCondition;
 begin
  HaveBeenMerged := True;
  NewCondition.KeyFieldCount := MinFieldCount;
  if (Condition1.KeyFieldCount = MinFieldCount) then
   begin
    NewCondition.KeyRecordBuffer := Condition1.KeyRecordBuffer;
    NewCondition.Condition := Condition1.Condition;
   end
  else
   begin
    NewCondition.KeyRecordBuffer := Condition2.KeyRecordBuffer;
    NewCondition.Condition := Condition2.Condition;
   end
 end; // SetMinKFCCondition

 procedure SetMaxKFCCondition;
 begin
  HaveBeenMerged := True;
  NewCondition.KeyFieldCount := MaxFieldCount;
  if (Condition1.KeyFieldCount = MaxFieldCount) then
   begin
    NewCondition.KeyRecordBuffer := Condition1.KeyRecordBuffer;
    NewCondition.Condition := Condition1.Condition;
   end
  else
   begin
    NewCondition.KeyRecordBuffer := Condition2.KeyRecordBuffer;
    NewCondition.Condition := Condition2.Condition;
   end
 end; // SetMaxKFCCondition

begin
 IndexID := Condition1.IndexID;
 MinFieldCount := Min(Condition1.KeyFieldCount,Condition2.KeyFieldCount);
 MaxFieldCount := Max(Condition1.KeyFieldCount,Condition2.KeyFieldCount);
 Incompatible := False;
 HaveBeenMerged := False;
 // check for distinct
 if (Condition1.KeyRecordBuffer = nil) or
    (Condition2.KeyRecordBuffer = nil) then
    Exit;
 ResMin := FIndexManager.OpenIndex(IndexID).CompareRecordBuffersByIndex(
          Condition1.KeyRecordBuffer,Condition2.KeyRecordBuffer,MinFieldCount);
 NewCondition := TACRScanSearchCondition.Create;
 try
 NewCondition.IndexID := IndexID;
 NewCondition.Expression := nil;
 if (Condition1.Condition <> scNone) and (Condition2.Condition <> scNone) then
   case (Condition1.Condition) of
     scEqual:
       case (Condition2.Condition) of
         scEqual:
          begin
            if ((ResMin > 0) or (ResMin < 0)) then
             Incompatible := True
            else
             SetMaxKFCCondition;
          end;
         scGreater:
          begin
            if ((ResMin = 0) or (ResMin < 0)) then
             Incompatible := True
            else
             SetFirstCondition;
          end;
         scLower:
          begin
            if ((ResMin = 0) or (ResMin > 0)) then
             Incompatible := True
            else
             SetFirstCondition;
          end;
         scGreaterEqual:
          begin
            if (ResMin < 0) then
             Incompatible := True
            else
            if ((ResMin > 0) or
                ((ResMin = 0) and (Condition1.KeyFieldCount = MaxFieldCount))) then
             SetFirstCondition; // otherwise not merge
          end;
         scLowerEqual:
          begin
            if (ResMin > 0) then
             Incompatible := True
            else
            if ((ResMin < 0) or
                ((ResMin = 0) and (Condition1.KeyFieldCount = MaxFieldCount))) then
             SetFirstCondition; // otherwise not merge
          end;
       end; // Condition2
     scGreater:
       case (Condition2.Condition) of
         scEqual:
          begin
            if ((ResMin = 0) or (ResMin > 0)) then
             Incompatible := True
            else
             SetSecondCondition;
          end;
         scGreater:
          begin
            if (ResMin > 0) then
             SetFirstCondition
            else
            if (ResMin < 0) then
             SetSecondCondition
            else
             SetMinKFCCondition;
          end;
         scLower:
          begin
            if ((ResMin > 0) or (ResMin = 0)) then
             Incompatible := True; // otherwise not merge
          end;
         scGreaterEqual:
          begin
            if ((ResMin = 0) or (ResMin > 0)) then
             SetFirstCondition
            else
            if (ResMin < 0) then
             SetSecondCondition;
          end;
         scLowerEqual:
          begin
            if ((ResMin > 0) or
                ((ResMin = 0) and (Condition1.KeyFieldCount = MinFieldCount))) then
             Incompatible := True; // otherwise not merge
          end;
       end; // Condition2
     scLower:
       case (Condition2.Condition) of
         scEqual:
          begin
            if ((ResMin = 0) or (ResMin < 0)) then
             Incompatible := True
            else
             SetSecondCondition;
          end;
         scGreater:
          begin
            if ((ResMin < 0) or (ResMin = 0)) then
             Incompatible := True; // otherwise not merge
          end;
         scLower:
          begin
            if (ResMin < 0) then
             SetFirstCondition
            else
            if (ResMin > 0) then
             SetSecondCondition
            else
             SetMinKFCCondition;
          end;
         scGreaterEqual:
          begin
            if ((ResMin < 0) or
                ((ResMin = 0) and (Condition1.KeyFieldCount = MinFieldCount))) then
             Incompatible := True; // otherwise not merge
          end;
         scLowerEqual:
          begin
            if ((ResMin < 0) or (ResMin = 0)) then
             SetFirstCondition
            else
            if (ResMin > 0) then
             SetSecondCondition;
          end;
       end; // Condition2
     scGreaterEqual:
       case (Condition2.Condition) of
         scEqual:
          begin
            if (ResMin > 0) then
             Incompatible := True
            else
            if ((ResMin < 0) or
                ((ResMin = 0) and
                 (Condition2.KeyFieldCount = MaxFieldCount))) then
             SetSecondCondition;
          end;
         scGreater:
          begin
            if (ResMin > 0) then
             SetFirstCondition
            else
             SetSecondCondition;
          end;
         scLower:
          begin
            if ((ResMin > 0) or
                ((ResMin = 0) and (Condition2.KeyFieldCount = MinFieldCount))) then
             Incompatible := True; // otherwise not merge
          end;
         scGreaterEqual:
          begin
            if (ResMin > 0) then
             SetFirstCondition
            else
            if (ResMin < 0) then
             SetSecondCondition
            else
             SetMaxKFCCondition;
          end;
         scLowerEqual:
          begin
            if (ResMin > 0) then
             Incompatible := True // otherwise not merge
            else
             if ((ResMin = 0) and (MinFieldCount = MaxFieldCount)) then
              begin
               SetMaxKFCCondition;
               NewCondition.Condition := scEqual;
              end;
          end;
       end; // Condition2
     scLowerEqual:
       case (Condition2.Condition) of
         scEqual:
          begin
            if (ResMin < 0) then
             Incompatible := True
            else
            if ((ResMin > 0) or
                ((ResMin = 0) and (Condition2.KeyFieldCount = MaxFieldCount)))
               then
             SetSecondCondition; // otherwise not merge
          end;
         scGreater:
          begin
            if ((ResMin < 0) or
                ((ResMin = 0) and (Condition2.KeyFieldCount = MinFieldCount))) then
             Incompatible := True; // otherwise not merge
          end;
         scLower:
          begin
            if (ResMin < 0) then
             SetFirstCondition
            else
            if (ResMin > 0) then
             SetSecondCondition;
          end;
         scGreaterEqual:
          begin
            if (ResMin < 0) then
             Incompatible := True // otherwise not merge
            else
             if ((ResMin = 0) and (MinFieldCount = MaxFieldCount)) then
              begin
               SetMaxKFCCondition;
               NewCondition.Condition := scEqual;
              end;
          end;
         scLowerEqual:
          begin
            if (ResMin < 0) then
             SetFirstCondition
            else
            if (ResMin > 0) then
             SetSecondCondition
            else
             SetMaxKFCCondition;
          end;
       end; // Condition2
   end; // Condition1
  if (HaveBeenMerged) then
   begin
      Condition1.Assign(NewCondition);
     end;
 finally
  NewCondition.Free;
   end;
end; // MergeAndCheckSearchConditionsCompatibility


//------------------------------------------------------------------------------
// returns true if conditions are equal
//------------------------------------------------------------------------------
function TACRTableData.IsEqualConditions(
                            var Condition1:            TACRScanSearchCondition;
                            var Condition2:            TACRScanSearchCondition
                                       ): Boolean;
begin
 Result := False;
 if (Condition1.IndexId = Condition2.IndexId) then
  if ((Condition1.Condition = Condition2.Condition) and
      (Condition1.Condition <> scNone)) then
   if (Condition1.KeyFieldCount = Condition2.KeyFieldCount) then
    begin
     Result := (FIndexManager.OpenIndex(Condition1.IndexId).
                  CompareRecordBuffersByIndex(
                            Condition1.KeyRecordBuffer,
                            Condition2.KeyRecordBuffer,
                            Condition1.KeyFieldCount) = 0);
    end;
end; // IsEqualConditions


//------------------------------------------------------------------------------
// removes duplicate conditions
//------------------------------------------------------------------------------
procedure TACRTableData.RemoveDuplicateConditions(Conditions: TACRScanSearchConditionArray);
var i,j:        Integer;
    ToRemove:   Boolean;
begin
 i := 0;
 while (i < Conditions.Count) and (Conditions.Count > 1) do
  begin
   ToRemove := False;
   for j := 0 to Conditions.Count - 1 do
    if (j <> i) then
     if (IsEqualConditions(Conditions.Items[i],Conditions.Items[j])) then
      ToRemove := True;
   if (ToRemove) then
    Conditions.Delete(i)
   else
    Inc(i);
  end;
end; // RemoveDuplicateConditions


//------------------------------------------------------------------------------
// sorts Conditions array and removes unnecessary conditions
//------------------------------------------------------------------------------
procedure TACRTableData.OptimizeSearchConditions(
                            var Conditions:               TACRScanSearchConditionArray;
                            out NonCompatibleConditions:  Boolean
                                 );
var IndexIDs:       TACRIntegerArray;
    i,j,k:          Integer;
    IndexID:        TACRObjectID;
    NewConditions:  TACRScanSearchConditionArray;
    HaveBeenMerged: Boolean;
    DoAppend:       Boolean;
begin
 IndexIDs := TACRIntegerArray.Create(0,Conditions.Count,Conditions.Count);
 NewConditions := TACRScanSearchConditionArray.Create;
 NonCompatibleConditions := False;
 try
  for i := 0 to Conditions.Count - 1 do
   if (Conditions.Items[i].Expression = nil) then
    begin
     IndexID := Conditions.Items[i].IndexID;
     if (not IndexIDs.IsValueExists(IndexID)) then
       IndexIDs.Append(IndexID);
    end
   else
    NonCompatibleConditions := TACRExpression(Conditions.Items[i].Expression).IsIncompatible;
  for i := 0 to IndexIDs.ItemCount-1 do
   if (NonCompatibleConditions) then
    break
   else
     begin
      IndexID := IndexIDs.Items[i];
      for k := 0 to Conditions.Count-1 do
       if (NonCompatibleConditions) then
        break
       else
         if (Conditions.Items[k].Expression = nil) then
           if (Conditions.Items[k].IndexID = IndexID) then
             begin
              DoAppend := True;
              for j := 0 to NewConditions.Count-1 do
               if (NewConditions.Items[j].IndexID = IndexID) then
                begin
                 MergeAndCheckSearchConditionsCompatibility(
                   NewConditions.Items[j],
                   Conditions.Items[k],
                   NonCompatibleConditions,
                   HaveBeenMerged
                                                           );
                 if (NonCompatibleConditions) then
                  break;
                 if (HaveBeenMerged) then
                  DoAppend := False;
                end; // j
               if (DoAppend) then
                NewConditions.AddCondition(Conditions.Items[k]);
             end; // k condition with current index
     end; // i
  if (not NonCompatibleConditions) then
    RemoveDuplicateConditions(NewConditions);

  if (not NonCompatibleConditions) then
    for k := 0 to Conditions.Count-1 do
     if (Conditions.Items[k].Expression <> nil) then
      NewConditions.AddCondition(Conditions.Items[k]);
 finally
  IndexIDs.Free;
  if (NonCompatibleConditions) then
   NewConditions.Free
  else
   begin
    Conditions.Free;
    Conditions := NewConditions;
   end;
 end;
end; // OptimizeSearchConditions


//------------------------------------------------------------------------------
// prepare conditions array
//------------------------------------------------------------------------------
procedure TACRTableData.PrepareConditions(
                           Cursor:            TACRCursor;
                       Conditions:        TACRScanSearchConditionArray;
                       KeyCondition:      TACRScanSearchCondition;
                       SearchExpression:  TACRExpression;
                           GoForward:         Boolean
                           );
var
  RangeCondition: TACRScanSearchCondition;
  DistinctCondition:  TACRScanSearchCondition;
begin
  RangeCondition := TACRScanSearchCondition.Create;
  DistinctCondition := TACRScanSearchCondition.Create;
  try
    if (KeyCondition <> nil) then
      Conditions.AddCondition(KeyCondition);
    if (SearchExpression <> nil) then
       Conditions.AddExpression(SearchExpression);
    if (Cursor.FilterExpression <> nil) then
      Conditions.AddExpression(Cursor.FilterExpression);
    if (Cursor.SQLFilterExpression <> nil) then
      Conditions.AddExpression(Cursor.SQLFilterExpression);
    if (Cursor.IsRangeApplied) then
     begin
       if (Cursor.RangeStartBuffer <> nil) then
        begin
          RangeCondition.KeyRecordBuffer := Cursor.RangeStartBuffer;
          RangeCondition.KeyFieldCount := Cursor.RangeStartKeyFieldCount;
          if (Cursor.RangeStartExclusive) then
            RangeCondition.Condition := scGreater
          else
            RangeCondition.Condition := scGreaterEqual;
          RangeCondition.IndexID := Cursor.IndexID;
          RangeCondition.Expression := nil;
          Conditions.AddCondition(RangeCondition);
        end;
       if (Cursor.RangeEndBuffer <> nil) then
        begin
          RangeCondition.KeyRecordBuffer := Cursor.RangeEndBuffer;
          RangeCondition.KeyFieldCount := Cursor.RangeEndKeyFieldCount;
          if (Cursor.RangeEndExclusive) then
            RangeCondition.Condition := scLower
          else
            RangeCondition.Condition := scLowerEqual;
          RangeCondition.IndexID := Cursor.IndexID;
          RangeCondition.Expression := nil;
          Conditions.AddCondition(RangeCondition);
        end;
     end;
    if ((Cursor.IsDistinctApplied) and (Cursor.DistinctRecordBuffer <> nil) and
        ((not Cursor.FirstPosition) and (not Cursor.LastPosition))) then
     begin
      DistinctCondition.KeyRecordBuffer := Cursor.DistinctRecordBuffer;
      DistinctCondition.KeyFieldCount := Cursor.DistinctFieldCount;
      if (GoForward) then
        DistinctCondition.Condition := scGreater
      else
        DistinctCondition.Condition := scLower;
      DistinctCondition.IndexID := Cursor.IndexID;
      DistinctCondition.Expression := nil;
      Conditions.AddCondition(DistinctCondition);
     end;
  finally
    RangeCondition.Free;
    DistinctCondition.Free;
  end;
end;// PrepareConditions


//------------------------------------------------------------------------------
// try to find the best scan condition with min range record count
//------------------------------------------------------------------------------
function TACRTableData.ChooseScanConditionsWithMinRangeRecordCount(
                       SessionID:              TACRSessionID;
                       Conditions:             TACRScanSearchConditionArray;
                       CurrentIndexID:         TACRObjectID;
                       CurrentRecordID:        TACRRecordID;
                       var ScanConditionNo:    Integer;
                       var ScanEndConditionNo: Integer
                              ): Boolean;
var
  i:                        Integer;
  StartConditionNo:         Integer;
  EndConditionNo:           Integer;
  ScanConditionRecordCount: TACRRecordNo;
  Index:                    TACRIndex;
  ScanIndexRecCount:        TACRRecordNo;
  ScanNoCondRecCount:       TACRRecordNo;
begin
  // scan without condition
  ScanConditionNo := INVALID_ID4;
  ScanConditionRecordCount := InternalGetRecordCount;
  // index scan
  i := 0;
  while (i < Conditions.Count) do
   begin
     if (Conditions.Items[i].Expression = nil) then
      begin
       StartConditionNo := i;
       if (i < Conditions.Count-1) then
        if (Conditions.Items[i].IndexID = Conditions.Items[i+1].IndexID) then
         Inc(i);
       EndConditionNo := i;
       Index := FIndexManager.OpenIndex(Conditions.Items[i].IndexID);
       if (Index = nil) then
        raise EACRException.Create(20058, ErrorANilPointer);
       ScanIndexRecCount := Index.GetApproxRangeRecordCount(
                      SessionID,
                      InternalGetRecordCount,
                      Conditions.Items[StartConditionNo],
                      Conditions.Items[EndConditionNo]
                                                           );
       if (ScanIndexRecCount < ScanConditionRecordCount) then
        begin
          ScanConditionNo := StartConditionNo;
          if (StartConditionNo <> EndConditionNo) then
            ScanEndConditionNo := EndConditionNo
          else
            ScanEndConditionNo := INVALID_ID4;
          ScanConditionRecordCount := ScanIndexRecCount;
        end;
      end;
     Inc(i);
   end;
  // compare best conditional scan with with non-conditional scan
  if (ScanConditionNo <> INVALID_OBJECT_ID) then
    begin
      if (CurrentIndexID = INVALID_OBJECT_ID) then
         ScanNoCondRecCount := (InternalGetRecordCount-
                     FRecordManager.GetApproximateRecNo(CurrentRecordID,SessionID)) div 5
      else
         ScanNoCondRecCount := InternalGetRecordCount;
      if (ScanNoCondRecCount <
          ScanConditionRecordCount * ScanConditionRecordCount) then
        begin
          ScanConditionNo := INVALID_OBJECT_ID;
          ScanEndConditionNo := INVALID_OBJECT_ID;
        end;
    end;
  Result := (ScanConditionNo <> INVALID_OBJECT_ID);
end;// ChooseScanConditionsWithMinRangeRecordCount


//------------------------------------------------------------------------------
// if index is defined as unique or contain unique field
//------------------------------------------------------------------------------
function TACRTableData.IsIndexUnique(IndexID: TACRObjectID): Boolean;
var
  IndexDef: TACRIndexDef;
  FieldDef: TACRFieldDef;
begin
  IndexDef := TACRIndexDef(FIndexManager.IndexDefs.GetDefByObjectId(IndexID));
  Result := IndexDef.Unique;
  if (not Result) then
    begin
      FieldDef := FFieldManager.FieldDefs.GetFieldDefByName(IndexDef.Columns[0].FieldName);
      if (IsAutoincFieldType(FieldDef.AdvancedFieldType)) then
        Result := True;
    end;
end;// IsIndexUnique


//------------------------------------------------------------------------------
// try to find the best scan condition using heuristics
//------------------------------------------------------------------------------
function TACRTableData.ChooseScanConditionsByHeuristsics(
                       Conditions:             TACRScanSearchConditionArray;
                       CurrentIndexID:         TACRObjectID;
                       var ScanConditionNo:    Integer;
                       var ScanEndConditionNo: Integer
                              ): Boolean;
var
  i:       Integer;
  Found:   Boolean;
  FoundNo: Integer;
begin
  FoundNo := INVALID_ID4;
  ScanConditionNo := INVALID_ID4;
  ScanEndConditionNo := INVALID_ID4;
  Result := False;
  if (Conditions.Count > 0) then
   begin
     // any condition '=' for unique index?
     for i := 0 to Conditions.Count-1 do
       if (Conditions.Items[i].Expression = nil) and
          (Conditions.Items[i].Condition = scEqual) and
          (IsIndexUnique(Conditions.Items[i].IndexID)) then
        begin
          ScanConditionNo := i;
          Result := True;
          break;
        end;
     // check for single condition with current index
     if (not Result) then
      begin
        Found := False;
        for i := 0 to Conditions.Count-1 do
         if (Conditions.Items[i].Expression = nil) then
          begin
           if (Conditions.Items[i].IndexID = CurrentIndexID) then
            begin
              if (not Found) then
               begin
                 Found := True;
                 FoundNo := i;
               end
              else
               begin
                 Found := False;
                 break;
               end;
            end
           else
             begin
               Found := False;
               break;
             end;
          end;
        if (Found) then
         begin
          ScanConditionNo := FoundNo;
          Result := True;
         end;
      end;
   end;
end;// ChooseScanConditionsByHeuristsics


//------------------------------------------------------------------------------
// if any index exists in conditions without expression will use it
//------------------------------------------------------------------------------
function TACRTableData.ChooseScanConditionsByAnyIndex(
                       Conditions:             TACRScanSearchConditionArray;
                       CurrentIndexID:         TACRObjectID;
                       var ScanConditionNo:    Integer;
                       var ScanEndConditionNo: Integer
                                        ): Boolean;
var
  i:       Integer;
begin
  ScanConditionNo := INVALID_ID4;
  ScanEndConditionNo := INVALID_ID4;
  Result := False;
  if (Conditions.Count > 0) then
   begin
     for i := 0 to Conditions.Count-1 do
       if (Conditions.Items[i].Expression = nil) and
          (Conditions.Items[i].IndexID <> INVALID_OBJECT_ID) then
        begin
          ScanConditionNo := i;
          Result := True;
          break;
        end;
   end;
end; // ChooseScanConditionsByAnyIndex


//------------------------------------------------------------------------------
// try to find the best scan condition
//------------------------------------------------------------------------------
procedure TACRTableData.ChooseScanConditions(
                           SessionID:              TACRSessionID;
                           Conditions:             TACRScanSearchConditionArray;
                           CurrentIndexID:         TACRObjectID;
                           CurrentRecordID:        TACRRecordID;
                           GoForward:              Boolean;
                           var ScanConditionNo:    Integer;
                           var ScanEndConditionNo: Integer
                                            );
var
  StartConditionNo:         Integer;
  EndConditionNo:           Integer;
  Index:                    TACRIndex;
  bOK:                      Boolean;
begin
  bOK := ChooseScanConditionsByHeuristsics(
                                            Conditions,
                                            CurrentIndexID,
                                            ScanConditionNo,
                                            ScanEndConditionNo
                                           );
  if (not bOK) then
    bOK := ChooseScanConditionsWithMinRangeRecordCount(
                                            SessionID,
                                            Conditions,
                                            CurrentIndexID,
                                            CurrentRecordID,
                                            ScanConditionNo,
                                            ScanEndConditionNo
                                                 );
  if (not bOK) then
    bOK := ChooseScanConditionsByAnyIndex(
                                            Conditions,
                                            CurrentIndexID,
                                            ScanConditionNo,
                                            ScanEndConditionNo
                                           );
  if (bOK) then
   begin
      // need to swap ScanEnd and Scan conditions?
      if (ScanEndConditionNo <> INVALID_ID4) then
       begin
        Index := FIndexManager.OpenIndex(Conditions.Items[ScanConditionNo].IndexID);
        if (Index = nil) then
         raise EACRException.Create(20059, ErrorANilPointer);
         if (Index.CompareConditions(
                       Conditions.Items[ScanConditionNo],
                       Conditions.Items[ScanEndConditionNo]) < 0) then
          begin
            StartConditionNo := ScanConditionNo;
            EndConditionNo := ScanEndConditionNo;
          end
         else
          begin
            StartConditionNo := ScanEndConditionNo;
            EndConditionNo := ScanConditionNo;
          end;
         if (GoForward) then
          begin
            ScanConditionNo := StartConditionNo;
            ScanEndConditionNo := EndConditionNo;
          end
         else
          begin
            ScanConditionNo := EndConditionNo;
            ScanEndConditionNo := StartConditionNo;
          end;
       end;
   end; // MinRange
end;// ChooseScanConditions


//------------------------------------------------------------------------------
// return true if record was found and is visible by cursor
// prepares params for FindRecordByScan and calls it
//------------------------------------------------------------------------------
function TACRTableData.FindRecord(
                    Cursor:                 TACRCursor;
                    SearchExpression:       TACRExpression; // locate
                    KeyCondition:           TACRScanSearchCondition; // find key
                    Restart:                Boolean;
                    GoForward:              Boolean;
                    // if record was found then return record id
                    ResultRecordID:         PACRRecordID;
                    // if specified - fill record bitmap
                    RecordBitmap:           TACRRecordBitmap = nil;
                    InternalCall:           Boolean = False;
                    StopAtFirstFoundRecord: Boolean = False
                   ): Boolean;
var
  Conditions:                TACRScanSearchConditionArray;
  ScanConditionNo:           Integer;
  ScanEndConditionNo:        Integer;
  bNonCompatibleConditions:  Boolean;
  ExtractedConditionsInfo:   TList;
begin
  Lock;
  try
    if (InternalGetRecordCount = 0) then
     Result := False
    else
     begin
      Conditions := TACRScanSearchConditionArray.Create;
      PrepareConditions(Cursor, Conditions, KeyCondition, SearchExpression, GoForward);
      ExtractedConditionsInfo := TList.Create;
      try
        Conditions.ExtractConditionsFromExpressions(FIndexManager.IndexDefs,
                                                    ExtractedConditionsInfo);
        try
          OptimizeSearchConditions(Conditions, bNonCompatibleConditions);
          if (bNonCompatibleConditions) then
           Result := False
          else
           begin
             ChooseScanConditions(
                                   Cursor.Session.SessionID,
                                   Conditions,
                                   Cursor.IndexID,
                                   Cursor.CurrentRecordID,
                                   GoForward,
                                   ScanConditionNo,
                                   ScanEndConditionNo
                                 );
             Result := FindRecordByScan(
                                       Cursor.Session.SessionID,
                                       Conditions,
                                       ScanConditionNo,
                                       ScanEndConditionNo,
                                       Cursor.IndexID,
                                       Cursor.FilterRecord,
                                       Cursor.Dataset,
                                       Cursor.CurrentRecordID,
                                       Restart,
                                       GoForward,
                                       ResultRecordID,
                                       RecordBitmap,
                                       Cursor.RandomOrder,
                                       StopAtFirstFoundRecord
                                       );
           end;
        finally
          Conditions.ReturnConditionsToExpressions(ExtractedConditionsInfo);
        end;
      finally
        ExtractedConditionsInfo.Free;
        Conditions.Free;
      end;
     end;
  finally
     Unlock;
  end;
end; // FindRecord


//------------------------------------------------------------------------------
// return true if record match specified condition
//------------------------------------------------------------------------------
function TACRTableData.IsRecordMatchCondition(
                                var Condition:        TACRScanSearchCondition;
                                RecordBuffer:         TACRRecordBuffer
                               ): Boolean;
var
    Res:            Integer;
begin
 Res := FIndexManager.OpenIndex(Condition.IndexID).
        CompareRecordBuffersByIndex(
        RecordBuffer,
        Condition.KeyRecordBuffer,
        Condition.KeyFieldCount);
 case Condition.Condition of
  scEqual: Result := (Res = 0);
  scGreater: Result := (Res > 0);
  scLower: Result := (Res < 0);
  scGreaterEqual: Result := (Res >= 0);
  scLowerEqual: Result := (Res <= 0)
 else
  Result := False
 end;
end; // IsRecordMatchCondition


//------------------------------------------------------------------------------
// return true if record match specified conditions
//------------------------------------------------------------------------------
function TACRTableData.IsRecordMatchConditions(
                          Conditions:           TACRScanSearchConditionArray;
                          ExcludedConditionNo:  Integer; // INVALID_ID4 if not specified
                          ExcludedConditionNo2: Integer; // INVALID_ID4 if not specified
                          FilterRecordPtr:      Pointer;
                          Dataset:              Pointer;
                          RecordBuffer:         TACRRecordBuffer
                                ): Boolean;
var
  FunctionAddress:  Pointer;
  i:                Integer;
begin
  Result := True;
  if (Result) then
   begin
    for i := 0 to Conditions.Count - 1 do
     if (i <> ExcludedConditionNo) and (i <> ExcludedConditionNo2) then
      begin
       if (Conditions.Items[i].Expression <> nil) then
        begin
         if (not TACRExpression(Conditions.Items[i].Expression).IsEmpty) then
          Result := TACRExpression(Conditions.Items[i].Expression).GetResult;
        end // expression
       else
        begin
         Result := IsRecordMatchCondition(Conditions.Items[i],RecordBuffer);
        end; // key condition
       if (not Result) then
        break;
      end;
   end;
 if (Result) then
  if (Pointer(FilterRecordPtr) <> nil) then
   begin
    // Borland parameters transmitting bug fix
    // Get Function address from Virtual Method Table
    FunctionAddress := Pointer(PCardinal(Pointer(Cardinal(@FilterRecordPtr)))^);
    asm
      MOV    ECX, Dataset                // Dataset
      MOV    EDX, RecordBuffer           // Buffer
      MOV    EAX, FunctionAddress        // Address of Function
      CALL   EAX                         // Call Function
      MOV    Result, AL                  // Save Result
    end;
   end;
end; // IsRecordMatchConditions


//------------------------------------------------------------------------------
// return true if there are some active conditions except excluded
//------------------------------------------------------------------------------
function TACRTableData.IsRecordMatchConditionsExists(
                          Conditions:           TACRScanSearchConditionArray;
                          ExcludedConditionNo:  Integer; // INVALID_ID4 if not specified
                          ExcludedConditionNo2: Integer; // INVALID_ID4 if not specified
                          FilterRecordPtr:      Pointer
                                ): Boolean;
var
  i:                Integer;
begin
  Result := False;
  for i := 0 to Conditions.Count - 1 do
   if (i <> ExcludedConditionNo) and (i <> ExcludedConditionNo2) then
    begin
     if (Conditions.Items[i].Expression <> nil) then
      begin
       if (not TACRExpression(Conditions.Items[i].Expression).IsEmpty) then
        begin
         Result := True;
         break;
        end;
      end // expression
     else
      begin
        Result := True;
        break;
      end; // key condition
    end;
 if (not Result) then
  if (Pointer(FilterRecordPtr) <> nil) then
   Result := True;
end; // IsRecordMatchConditionsExists


//------------------------------------------------------------------------------
// scan all records and check conditions
//------------------------------------------------------------------------------
function TACRTableData.FindRecordByScanWithoutCondition(
                          SessionID:          TACRSessionID;
                          Conditions:         TACRScanSearchConditionArray;
                          ScanConditionNo:    Integer; // INVALID_ID4 if not specified
                          ScanEndConditionNo: Integer; // INVALID_ID4 if not specified
                          CurrentIndexID:     TACRObjectID; // INVALID_OBJECT_ID if not specified
                          FilterRecordPtr:    Pointer;
                          Dataset:            Pointer;
                          CurrentRecordID:    TACRRecordID;
                          // find first or last record in order specified by index settings
                          Restart:            Boolean;
                          // go from first record to last or vice versa
                          GoForward:          Boolean;
                          // if record was found then return record id
                          ResultRecordID:     PACRRecordID;
                          // if specified - fill record bitmap
                          RecordBitmap:           TACRRecordBitmap = nil;
                          RandomOrder:            Boolean = False;
                          StopAtFirstFoundRecord: Boolean = False
                         ): Boolean;
var
  NavigationInfo:   TACRNavigationInfo;
  Acceptable:       Boolean;
  i:                Integer;
begin
  if ((RecordBitmap <> nil) and
      (Conditions.Count = 0) and (FilterRecordPtr = nil)) then
   begin
    Result := False;
    if ((not InMemory) and (not Temporary)) then
     RecordBitmap.Distinct := True
   end
  else
   begin
    NavigationInfo.RecordID := CurrentRecordID;
    NavigationInfo.FirstPosition := (Restart and GoForward);
    NavigationInfo.LastPosition := (Restart and (not GoForward));
    if (GoForward) then
     NavigationInfo.GetRecordMode := grmNext
    else
     NavigationInfo.GetRecordMode := grmPrior;
    NavigationInfo.IndexID := CurrentIndexID;
    NavigationInfo.RecordBuffer := FTempRecordBuffer;
    for i := 0 to Conditions.Count-1 do
     if (Conditions.Items[i].Expression <> nil) then
      TACRExpression(Conditions.Items[i].Expression).AssignCursorBuffer(
          NavigationInfo.RecordBuffer);
    // scan without active condition - check each record
    repeat
     InternalGetRecordBuffer(SessionID, NavigationInfo);
     Acceptable := False;
     if (NavigationInfo.GetRecordResult = grrOK) then
      begin
       Acceptable := IsRecordMatchConditions(
                        Conditions,
                        ScanConditionNo,ScanEndConditionNo,FilterRecordPtr,Dataset,
                        NavigationInfo.RecordBuffer
                                            );
       if (RecordBitmap <> nil) then
        begin
         if (Acceptable) then
          RecordBitmap.ShowRecord(NavigationInfo.RecordID);
         Acceptable := False;
        end;
      end;
    until (Acceptable or (NavigationInfo.GetRecordResult <> grrOK));

    Result := (NavigationInfo.GetRecordResult = grrOK);
    if (Result) then
      if (ResultRecordID <> nil) then
       Move(NavigationInfo.RecordID,ResultRecordID^,sizeof(NavigationInfo.RecordID));
 end;
end; // FindRecordByScanWithoutCondition


//------------------------------------------------------------------------------
// find records by index and condition specified by ScanConditionNo
// condition index is the same as CurrentIndex
//------------------------------------------------------------------------------
function TACRTableData.FindRecordByScanWithConditionAndConditionIndex(
                          SessionID:          TACRSessionID;
                          Conditions:         TACRScanSearchConditionArray;
                          ScanConditionNo:    Integer; // INVALID_ID4 if not specified
                          ScanEndConditionNo: Integer; // INVALID_ID4 if not specified
                          CurrentIndexID:     TACRObjectID; // INVALID_OBJECT_ID if not specified
                          FilterRecordPtr:    Pointer;
                          Dataset:            Pointer;
                          CurrentRecordID:    TACRRecordID;
                          // find first or last record in order specified by index settings
                          Restart:            Boolean;
                          // go from first record to last or vice versa
                          GoForward:          Boolean;
                          // if record was found then return record id
                          ResultRecordID:     PACRRecordID;
                          // if specified - fill record bitmap
                          RecordBitmap:           TACRRecordBitmap = nil;
                          RandomOrder:            Boolean = False;
                          StopAtFirstFoundRecord: Boolean = False
                         ): Boolean;
var
  NavigationInfo:       TACRNavigationInfo;
  Acceptable:           Boolean;
  i:                    Integer;
  SearchInfo:           TACRSearchInfo;
  ConditionsExists:     Boolean;
begin
  Result := False;
  ConditionsExists := IsRecordMatchConditionsExists(
                      Conditions,
                      ScanConditionNo,ScanEndConditionNo,FilterRecordPtr
                                          );
  NavigationInfo.IndexID := CurrentIndexID;
  NavigationInfo.RecordID := CurrentRecordID;
  NavigationInfo.FirstPosition := False;
  NavigationInfo.LastPosition := False;
  NavigationInfo.GetRecordMode := grmCurrent;
  NavigationInfo.RecordBuffer := FTempRecordBuffer;
  SearchInfo := FIndexManager.OpenIndex(CurrentIndexID).CreateSearchInfo;
  // find current position in index
  try
    for i := 0 to Conditions.Count-1 do
     if (Conditions.Items[i].Expression <> nil) then
      TACRExpression(Conditions.Items[i].Expression).AssignCursorBuffer(
          NavigationInfo.RecordBuffer);
    // load current record
    if (not Restart) then
      InternalGetRecordBuffer(SessionID, NavigationInfo);
    if ((Restart) or
        ((not Restart) and (NavigationInfo.GetRecordResult = grrOK))) then
     begin
      repeat
        if (ScanEndConditionNo <> INVALID_ID4) then
         // find record using the end condition
         Result := FIndexManager.OpenIndex(CurrentIndexID).FindRecord(
                     SessionID,
                     Restart,
                     GoForward,
                     Conditions.Items[ScanConditionNo],
                     Conditions.Items[ScanEndConditionNo],
                     NavigationInfo.RecordBuffer,
                     NavigationInfo.RecordID,
                     SearchInfo
                   )
        else
         // find record not using the end condition
         Result := FIndexManager.OpenIndex(CurrentIndexID).FindRecord(
                     SessionID,
                     Restart,
                     GoForward,
                     Conditions.Items[ScanConditionNo],
                     nil,
                     NavigationInfo.RecordBuffer,
                     NavigationInfo.RecordID,
                     SearchInfo
                   );
        Restart := False;
        Acceptable := False;
        if (Result) then
         begin
          if (ConditionsExists) then
           begin
            InternalGetRecordBuffer(SessionID, NavigationInfo);
            Acceptable := IsRecordMatchConditions(
                        Conditions,
                        ScanConditionNo,ScanEndConditionNo,FilterRecordPtr,Dataset,
                        NavigationInfo.RecordBuffer
                                          );
           end
          else
           Acceptable := True;
          if (RecordBitmap <> nil) then
           begin
            if (Acceptable) then
              RecordBitmap.ShowRecord(NavigationInfo.RecordID);
            Acceptable := False;
           end;
         end;
      until ((not Result) or (Result and Acceptable));
     end;

    if (Result) then
      if (ResultRecordID <> nil) then
       Move(NavigationInfo.RecordID,ResultRecordID^,sizeof(NavigationInfo.RecordID));
  finally
    FIndexManager.OpenIndex(CurrentIndexID).FreeSearchInfo(SearchInfo);
  end;
end; // FindRecordByScanWithConditionAndConditionIndex


//------------------------------------------------------------------------------
// find records by index and condition specified by ScanConditionNo
// there is no active index
//------------------------------------------------------------------------------
function TACRTableData.FindRecordByScanWithConditionAndWihtoutCurrentIndex(
                          SessionID:          TACRSessionID;
                          Conditions:         TACRScanSearchConditionArray;
                          ScanConditionNo:    Integer; // INVALID_ID4 if not specified
                          ScanEndConditionNo: Integer; // INVALID_ID4 if not specified
                          CurrentIndexID:     TACRObjectID; // INVALID_OBJECT_ID if not specified
                          FilterRecordPtr:    Pointer;
                          Dataset:            Pointer;
                          CurrentRecordID:    TACRRecordID;
                          // find first or last record in order specified by index settings
                          Restart:            Boolean;
                          // go from first record to last or vice versa
                          GoForward:          Boolean;
                          // if record was found then return record id
                          ResultRecordID:     PACRRecordID;
                          // if specified - fill record bitmap
                          RecordBitmap:           TACRRecordBitmap = nil;
                          RandomOrder:            Boolean = False;
                          StopAtFirstFoundRecord: Boolean = False
                         ): Boolean;
var
  NavigationInfo:       TACRNavigationInfo;
  Acceptable:           Boolean;
  i,Res:                Integer;
  ResRecordID:          TACRRecordID;
  RecordFound:          Boolean;
  FindRestart:          Boolean;
  ConditionIndexID:     TACRObjectID;
  SearchInfo:           TACRSearchInfo;
  ConditionsExists:     Boolean;
begin
  RecordFound := False;
  ConditionsExists := IsRecordMatchConditionsExists(
                      Conditions,
                      ScanConditionNo,ScanEndConditionNo,FilterRecordPtr
                                          );
  ConditionIndexID := Conditions.Items[ScanConditionNo].IndexID;
  NavigationInfo.RecordID := CurrentRecordID;
  NavigationInfo.IndexID := INVALID_OBJECT_ID;
  NavigationInfo.GetRecordMode := grmCurrent;
  NavigationInfo.FirstPosition := False;
  NavigationInfo.LastPosition := False;
  NavigationInfo.RecordBuffer := FTempRecordBuffer;
  SearchInfo := FIndexManager.OpenIndex(ConditionIndexID).CreateSearchInfo;
  try
    for i := 0 to Conditions.Count-1 do
     if (Conditions.Items[i].Expression <> nil) then
      TACRExpression(Conditions.Items[i].Expression).AssignCursorBuffer(
          NavigationInfo.RecordBuffer);
    // find current position in index
    FindRestart := True;
    repeat
      if (ScanEndConditionNo <> INVALID_ID4) then
        Result := FIndexManager.OpenIndex(ConditionIndexID).FindRecord(
                    SessionID,
                    FindRestart,
                    GoForward,
                    Conditions.Items[ScanConditionNo],
                    Conditions.Items[ScanEndConditionNo],
                    NavigationInfo.RecordBuffer,
                    NavigationInfo.RecordID,
                    SearchInfo
                  )
      else
        Result := FIndexManager.OpenIndex(ConditionIndexID).FindRecord(
                    SessionID,
                    FindRestart,
                    GoForward,
                    Conditions.Items[ScanConditionNo],
                    nil,
                    NavigationInfo.RecordBuffer,
                    NavigationInfo.RecordID,
                    SearchInfo
                  );
      FindRestart := False;
      Acceptable := False;
      if (Result) then
       begin
        if (ConditionsExists) then
         begin
          InternalGetRecordBuffer(SessionID, NavigationInfo);
          Acceptable := IsRecordMatchConditions(
                      Conditions,
                      ScanConditionNo,ScanEndConditionNo,FilterRecordPtr,Dataset,
                      NavigationInfo.RecordBuffer
                                        );
         end
        else
         Acceptable := True;
       end;
      if (Result and Acceptable) then
       begin
        if (StopAtFirstFoundRecord) then
         begin
          ResRecordID := NavigationInfo.RecordID;
          RecordFound := True;
          break;
         end;
        if (RecordBitmap <> nil) then
         RecordBitmap.ShowRecord(NavigationInfo.RecordID)
        else
        if (Restart) then
         begin
          // find minimum or maximum RecordID that meets all conditions
          if (not RecordFound) then
           begin
            // first record found
            ResRecordID := NavigationInfo.RecordID;
            RecordFound := True;
           end // first record found
          else
           begin
            // check current record for < or > result record
            Res := CompareRecordID(
                      NavigationInfo.RecordID,
                      ResRecordID
                   );
            if ((GoForward and (Res < 0)) or
                ((not GoForward) and (Res > 0))) then
              ResRecordID := NavigationInfo.RecordID;
           end; // not first record found
         end // Restart
        else
         begin
          // find next or prior record that mathces all conditions
          Res := CompareRecordID(
                   NavigationInfo.RecordID,
                   CurrentRecordID
                 );
          if ((GoForward) and (Res > 0)) then
           begin
            // find next record
            if (not RecordFound) then
             begin
              RecordFound := True;
              ResRecordID := NavigationInfo.RecordID;
             end // first record found
            else
             begin
              Res := CompareRecordID(
                      NavigationInfo.RecordID,
                      ResRecordID
                     );
              if (Res < 0) then
                ResRecordID := NavigationInfo.RecordID;
             end; // not first record found
           end // find next record
          else
          if ((not GoForward) and (Res < 0)) then
           begin
            // find next record
            if (not RecordFound) then
             begin
              RecordFound := True;
              ResRecordID := NavigationInfo.RecordID;
             end // first record found
            else
             begin
              Res := CompareRecordID(
                      NavigationInfo.RecordID,
                      ResRecordID
                     );
              if (Res > 0) then
                ResRecordID := NavigationInfo.RecordID;
             end; // not first record found
           end; // find prior record
         end; // Not restart
       end; // record meets conditions
    until (not Result);
    if (RecordBitmap <> nil) then
     begin
      Result := False;
      if (not RandomOrder) then
        RecordBitmap.VisibleRecords.Sort;
     end
    else
      Result := RecordFound;
    if (Result) then
      if (ResultRecordID <> nil) then
       Move(ResRecordID,ResultRecordID^,sizeof(ResRecordID));
  finally
    FIndexManager.OpenIndex(ConditionIndexID).FreeSearchInfo(SearchInfo);
  end;
end; // FindRecordByScanWithConditionAndWihtoutCurrentIndex


//------------------------------------------------------------------------------
// sort record ID array by positions in the index
//------------------------------------------------------------------------------
procedure TACRTableData.SortRecordIDByIndex(
                          Records:              TACRRecordIDArray;
                          IndexPositions:       TList;
                          CurrentIndexID:       TACRObjectID // INVALID_OBJECT_ID if not specified
                             );
var
 aLo, aHi:          Integer;
 TempRecordID:      TACRRecordID;
 TempIndexPosition: Pointer;

 function Compare(const Index1: Integer; IndexPos2: TObject): Integer;
 begin
   Result := FIndexManager.OpenIndex(CurrentIndexID).
              CompareRecordPositionsInIndex(
                IndexPositions[Index1],
                IndexPos2);
 end;

 procedure Swap(const Index1: Integer; const Index2: Integer);
 begin
  TempRecordID := Records.Items[Index1];
  TempIndexPosition := IndexPositions[Index1];
  Records.Items[Index1] := Records.Items[Index2];
  IndexPositions[Index1] := IndexPositions[Index2];
  Records.Items[Index2] := TempRecordID;
  IndexPositions[Index2] := TempIndexPosition;
 end;

 procedure QuickSort (
                    var iLo, iHi : Integer
                    );
  var
    Lo, Hi: Integer;
    Mid:    TObject;
  begin
    Lo := iLo;
    Hi := iHi;
    Mid := IndexPositions[(Lo + Hi) shr 1];
    repeat
     while (Compare(Lo,Mid) < 0) and (Lo < iHi) do
      Inc(Lo);
     while (Compare(Hi,Mid) > 0) and (Hi > 0) do
       Dec(Hi);
      if (Lo <= Hi) then
       begin
        Swap(Lo,Hi);
        Inc(Lo);
        Dec(Hi);
       end;
    until (Lo > Hi);
    if (Hi > iLo) then
     begin
      // check infinite recurse
      if (iHi = Hi) then
       raise EACRException.Create(11330,ErrorLErrorSoringRecordsByID,[Hi,Lo,iHi,iLo,Records.ItemCount]);
      QuickSort(iLo, Hi);
     end;
    if (Lo < iHi) then
      QuickSort(Lo, iHi);
  end; //QuickSort
begin
  if (Records.ItemCount > 1) then
   begin
    aLo := 0;
    aHi := Records.ItemCount-1;
    QuickSort (aLo, aHi);
   end;
end; // SortRecordIDByIndex


//------------------------------------------------------------------------------
// find records by index and condition specified by ScanConditionNo
// condition index is NOT the same as CurrentIndex
//------------------------------------------------------------------------------
function TACRTableData.FindRecordByScanWithConditionAndNonConditionIndex(
                          SessionID:          TACRSessionID;
                          Conditions:         TACRScanSearchConditionArray;
                          ScanConditionNo:    Integer; // INVALID_ID4 if not specified
                          ScanEndConditionNo: Integer; // INVALID_ID4 if not specified
                          CurrentIndexID:     TACRObjectID; // INVALID_OBJECT_ID if not specified
                          FilterRecordPtr:    Pointer;
                          Dataset:            Pointer;
                          CurrentRecordID:    TACRRecordID;
                          // find first or last record in order specified by index settings
                          Restart:            Boolean;
                          // go from first record to last or vice versa
                          GoForward:          Boolean;
                          // if record was found then return record id
                          ResultRecordID:     PACRRecordID;
                          // if specified - fill record bitmap
                          RecordBitmap:           TACRRecordBitmap = nil;
                          RandomOrder:            Boolean = False;
                          StopAtFirstFoundRecord: Boolean = False
                         ): Boolean;
var
  NavigationInfo:       TACRNavigationInfo;
  Acceptable:           Boolean;
  i,Res:                Integer;
  ResultRecPosInIndex:  TACRIndexPosition;
  ResRecordID:          TACRRecordID;
  ResRecPosInIndex:     TACRIndexPosition;
  CurRecPosInIndex:     TACRIndexPosition;
  RecordFound:          Boolean;
  FindRestart:          Boolean;
  ConditionIndexID:     TACRObjectID;
  SearchInfo:           TACRSearchInfo;
  IndexPositions:       TList;
  ConditionsExists:     Boolean;
begin
  Result := False;
  RecordFound := False;
  ConditionsExists := IsRecordMatchConditionsExists(
                      Conditions,
                      ScanConditionNo,ScanEndConditionNo,FilterRecordPtr
                                          );
  ConditionIndexID := Conditions.Items[ScanConditionNo].IndexID;
  NavigationInfo.RecordID := CurrentRecordID;
  NavigationInfo.IndexID := INVALID_OBJECT_ID;
  NavigationInfo.GetRecordMode := grmCurrent;
  NavigationInfo.FirstPosition := False;
  NavigationInfo.LastPosition := False;
  ResRecPosInIndex := nil;
  ResultRecPosInIndex := nil;
  NavigationInfo.RecordBuffer := FTempRecordBuffer;
  SearchInfo := FIndexManager.OpenIndex(ConditionIndexID).CreateSearchInfo;
  CurRecPosInIndex := FIndexManager.OpenIndex(CurrentIndexID).CreateIndexPosition;
  if (RecordBitmap <> nil) then
   IndexPositions := TList.Create;
  try
    for i := 0 to Conditions.Count-1 do
     if (Conditions.Items[i].Expression <> nil) then
      TACRExpression(Conditions.Items[i].Expression).AssignCursorBuffer(
          NavigationInfo.RecordBuffer);
    // find current position in index
    FindRestart := True;
    if (not Restart) then
     begin
       // load current record position in index
       NavigationInfo.RecordID := CurrentRecordID;
       NavigationInfo.IndexID := CurrentIndexID;
       InternalGetRecordBuffer(SessionID, NavigationInfo);
       if (NavigationInfo.GetRecordResult <> grrOK) then
        raise EACRException.Create(10387,ErrorLCannotFindRecordPositionInIndex,
                [CurrentIndexID,NavigationInfo.RecordID.PageNo,
                NavigationInfo.RecordID.PageItemNo]);
       if ((NavigationInfo.RecordID.PageNo <> CurrentRecordID.PageNo) or
            (NavigationInfo.RecordID.PageItemNo <> CurrentRecordID.PageItemNo)) then
         raise EACRException.Create(10415,ErrorLCannotFindRecordPositionInIndex,
                [CurrentIndexID,NavigationInfo.RecordID.PageNo,
                NavigationInfo.RecordID.PageItemNo]);
       // find record pos in index
       if (not FIndexManager.OpenIndex(CurrentIndexID).GetIndexPosition(
                SessionID,
                CurrentRecordID,
                NavigationInfo.RecordBuffer,
                CurRecPosInIndex
               )
          ) then
        raise EACRException.Create(10386,ErrorLCannotFindRecordPositionInIndex,
                [CurrentIndexID,NavigationInfo.RecordID.PageNo,
                NavigationInfo.RecordID.PageItemNo]);
        if ((NavigationInfo.RecordID.PageNo <> CurrentRecordID.PageNo) or
            (NavigationInfo.RecordID.PageItemNo <> CurrentRecordID.PageItemNo)) then
         raise EACRException.Create(10414,ErrorLCannotFindRecordPositionInIndex,
                [CurrentIndexID,NavigationInfo.RecordID.PageNo,
                NavigationInfo.RecordID.PageItemNo]);
     end; // not restart

    repeat
      if (ScanEndConditionNo <> INVALID_ID4) then
        Result := FIndexManager.OpenIndex(ConditionIndexID).FindRecord(
                    SessionID,
                    FindRestart,
                    GoForward,
                    Conditions.Items[ScanConditionNo],
                    Conditions.Items[ScanEndConditionNo],
                    NavigationInfo.RecordBuffer,
                    NavigationInfo.RecordID,
                    SearchInfo
                  )
      else
        Result := FIndexManager.OpenIndex(ConditionIndexID).FindRecord(
                    SessionID,
                    FindRestart,
                    GoForward,
                    Conditions.Items[ScanConditionNo],
                    nil,
                    NavigationInfo.RecordBuffer,
                    NavigationInfo.RecordID,
                    SearchInfo
                  );
      if (Result) then
       InternalGetRecordBuffer(SessionID, NavigationInfo);
      FindRestart := False;
      Acceptable := False;
      if (Result) then
       begin
        if (ConditionsExists) then
         begin
          Acceptable := IsRecordMatchConditions(
                      Conditions,
                      ScanConditionNo,ScanEndConditionNo,FilterRecordPtr,Dataset,
                      NavigationInfo.RecordBuffer
                                        );
         end
        else
         Acceptable := True;
       end;
      if (Result and Acceptable) then
       begin
        if (StopAtFirstFoundRecord) then
         begin
          ResRecordID := NavigationInfo.RecordID;
          RecordFound := True;
          break;
         end;

        if (Acceptable and (RecordBitmap <> nil)) then
         begin
          RecordBitmap.ShowRecord(NavigationInfo.RecordID);
          ResultRecPosInIndex := FIndexManager.OpenIndex(CurrentIndexID).CreateIndexPosition;
          // find record pos in current index
          if (not FIndexManager.OpenIndex(CurrentIndexID).GetIndexPosition(
                     SessionID,
                     NavigationInfo.RecordID,
                     NavigationInfo.RecordBuffer,
                     ResultRecPosInIndex
                   )
              ) then
            raise EACRException.Create(11331,ErrorLCannotFindRecordPositionInIndex,
                    [CurrentIndexID,NavigationInfo.RecordID.PageNo,
                    NavigationInfo.RecordID.PageItemNo]);
          IndexPositions.Add(ResultRecPosInIndex);
         end // build record bitmap
        else
         begin
          ResultRecPosInIndex := FIndexManager.OpenIndex(CurrentIndexID).CreateIndexPosition;
          try
            // find record pos in current index
            if (not FIndexManager.OpenIndex(CurrentIndexID).GetIndexPosition(
                       SessionID,
                       NavigationInfo.RecordID,
                       NavigationInfo.RecordBuffer,
                       ResultRecPosInIndex
                     )
                ) then
              raise EACRException.Create(10432,ErrorLCannotFindRecordPositionInIndex,
                      [CurrentIndexID,NavigationInfo.RecordID.PageNo,
                      NavigationInfo.RecordID.PageItemNo]);

            if (Restart) then
             begin
              // find minimum or maximum RecordID that meets all conditions
              if (not RecordFound) then
               begin
                // first record found
                ResRecordID := NavigationInfo.RecordID;
                FIndexManager.OpenIndex(CurrentIndexID).
                                  FreeIndexPosition(ResRecPosInIndex);
                ResRecPosInIndex := ResultRecPosInIndex;
                RecordFound := True;
               end // first record found
              else
               begin
                // check current record position for < or > result record position in index
                Res := FIndexManager.OpenIndex(CurrentIndexID).
                        CompareRecordPositionsInIndex(
                          ResultRecPosInIndex,
                          ResRecPosInIndex
                        );
                if ((GoForward and (Res < 0)) or ((not GoForward) and (Res > 0))) then
                 begin
                   ResRecordID := NavigationInfo.RecordID;
                   FIndexManager.OpenIndex(CurrentIndexID).
                                  FreeIndexPosition(ResRecPosInIndex);
                   ResRecPosInIndex := ResultRecPosInIndex;
                 end;
               end; // not first record found
             end // Restart
            else
             begin
              // find next or prior record that mathces all conditions
              Res := FIndexManager.OpenIndex(CurrentIndexID).
                        CompareRecordPositionsInIndex(
                          ResultRecPosInIndex,
                          CurRecPosInIndex
                        );
              if ((GoForward) and (Res > 0)) then
               begin
                if (not RecordFound) then
                 begin
                  // first record found
                  ResRecordID := NavigationInfo.RecordID;
                  FIndexManager.OpenIndex(CurrentIndexID).
                                  FreeIndexPosition(ResRecPosInIndex);
                  ResRecPosInIndex := ResultRecPosInIndex;
                  RecordFound := True;
                 end // first record found
                else
                 // find next record
                 begin
                  Res := FIndexManager.OpenIndex(CurrentIndexID).
                        CompareRecordPositionsInIndex(
                          ResultRecPosInIndex,
                          ResRecPosInIndex
                        );
                  if (Res < 0) then
                   begin
                     ResRecordID := NavigationInfo.RecordID;
                     FIndexManager.OpenIndex(CurrentIndexID).
                                    FreeIndexPosition(ResRecPosInIndex);
                     ResRecPosInIndex := ResultRecPosInIndex;
                   end;
                 end; // not first record found
               end // find next record
              else
              if ((not GoForward) and (Res < 0)) then
               begin
                if (not RecordFound) then
                 begin
                  // first record found
                  ResRecordID := NavigationInfo.RecordID;
                  FIndexManager.OpenIndex(CurrentIndexID).
                                  FreeIndexPosition(ResRecPosInIndex);
                  ResRecPosInIndex := ResultRecPosInIndex;
                  RecordFound := True;
                 end // first record found
                else
                 begin
                  Res := FIndexManager.OpenIndex(CurrentIndexID).
                        CompareRecordPositionsInIndex(
                          ResultRecPosInIndex,
                          ResRecPosInIndex
                        );
                  if (Res > 0) then
                   begin
                     ResRecordID := NavigationInfo.RecordID;
                     FIndexManager.OpenIndex(CurrentIndexID).
                                    FreeIndexPosition(ResRecPosInIndex);
                     ResRecPosInIndex := ResultRecPosInIndex;
                   end;
                 end; // prior record found
               end; // not first record found
             end; // Not restart
          finally
            if (ResRecPosInIndex <> ResultRecPosInIndex) then
              FIndexManager.OpenIndex(CurrentIndexID).FreeIndexPosition(ResultRecPosInIndex);
          end;
         end; // do not create record bitmap
       end; // record meets conditions
    until (not Result);
    if (RecordBitmap <> nil) and (not StopAtFirstFoundRecord) then
     begin
      Result := False;
      if (not RandomOrder) then
       begin
        SortRecordIDByIndex(RecordBitmap.VisibleRecords,IndexPositions,CurrentIndexID);
        RecordBitmap.VisibleRecords.Sorted := False;
       end;
     end
    else
      Result := RecordFound;
    if (Result) then
     begin
      if (ResultRecordID <> nil) then
       Move(ResRecordID,ResultRecordID^,sizeof(ResRecordID));
     end;
  finally
    FIndexManager.OpenIndex(ConditionIndexID).FreeSearchInfo(SearchInfo);
    FIndexManager.OpenIndex(CurrentIndexID).FreeIndexPosition(CurRecPosInIndex);
    if (RecordBitmap <> nil) then
     begin
      for i := 0 to IndexPositions.Count - 1 do
       begin
        ResultRecPosInIndex := IndexPositions[i];
        FIndexManager.OpenIndex(CurrentIndexID).FreeIndexPosition(ResultRecPosInIndex);
       end;
      IndexPositions.Free;
     end
    else
     begin
      if (ResultRecPosInIndex <> ResRecPosInIndex) then
          FIndexManager.OpenIndex(CurrentIndexID).FreeIndexPosition(ResRecPosInIndex);
      FIndexManager.OpenIndex(CurrentIndexID).FreeIndexPosition(ResultRecPosInIndex);
     end;
  end;
end; // FindRecordByScanWithConditionAndNonConditionIndex


//------------------------------------------------------------------------------
// return true if record was found
//------------------------------------------------------------------------------
function TACRTableData.FindRecordByScan(
                          SessionID:            TACRSessionID;
                          Conditions:           TACRScanSearchConditionArray;
                          ScanConditionNo:      Integer; // INVALID_ID4 if not specified
                          ScanEndConditionNo:   Integer; // INVALID_ID4 if not specified
                          CurrentIndexID:       TACRObjectID; // INVALID_OBJECT_ID if not specified
                          FilterRecordPtr:      Pointer;
                          Dataset:              Pointer;
                          CurrentRecordID:      TACRRecordID;
                          // find first or last record in order specified by index settings
                          Restart:              Boolean;
                          // go from first record to last or vice versa
                          GoForward:            Boolean;
                          // if record was found then return record id
                          ResultRecordID:       PACRRecordID;
                          // if specified - fill record bitmap
                          RecordBitmap:           TACRRecordBitmap = nil;
                          RandomOrder:            Boolean = False;
                          StopAtFirstFoundRecord: Boolean = False
                         ): Boolean;
begin
 if (ScanConditionNo = INVALID_ID4) then
  Result := FindRecordByScanWithoutCondition(
                          SessionID,
                          Conditions,
                          ScanConditionNo,
                          ScanEndConditionNo,
                          CurrentIndexID,
                          FilterRecordPtr,
                          Dataset,
                          CurrentRecordID,
                          Restart,
                          GoForward,
                          ResultRecordID,
                          RecordBitmap,
                          RandomOrder,
                          StopAtFirstFoundRecord
            ) // no condition
 else
  begin
   if (ScanConditionNo >= Integer(Conditions.Count)) then
    raise EACRException.Create(10382,ErrorLInvalidScanConditionNo,
      [ScanConditionNo,Conditions.Count]);
   if (CurrentIndexID = INVALID_OBJECT_ID) then
    Result := FindRecordByScanWithConditionAndWihtoutCurrentIndex(
                            SessionID,
                            Conditions,
                            ScanConditionNo,
                            ScanEndConditionNo,
                            CurrentIndexID,
                            FilterRecordPtr,
                            Dataset,
                            CurrentRecordID,
                            Restart,
                            GoForward,
                            ResultRecordID,
                            RecordBitmap,
                            RandomOrder,
                            StopAtFirstFoundRecord
              ) // condition and physical order
   else
    begin
     if (CurrentIndexID = Conditions.Items[ScanConditionNo].IndexID) then
      Result := FindRecordByScanWithConditionAndConditionIndex(
                              SessionID,
                              Conditions,
                              ScanConditionNo,
                              ScanEndConditionNo,
                              CurrentIndexID,
                              FilterRecordPtr,
                              Dataset,
                              CurrentRecordID,
                              Restart,
                              GoForward,
                              ResultRecordID,
                              RecordBitmap,
                              RandomOrder,
                              StopAtFirstFoundRecord
                ) // condition and condition.indexID = CurrentIndexID
     else
      Result := FindRecordByScanWithConditionAndNonConditionIndex(
                              SessionID,
                              Conditions,
                              ScanConditionNo,
                              ScanEndConditionNo,
                              CurrentIndexID,
                              FilterRecordPtr,
                              Dataset,
                              CurrentRecordID,
                              Restart,
                              GoForward,
                              ResultRecordID,
                              RecordBitmap,
                              RandomOrder,
                              StopAtFirstFoundRecord
                ) // condition and condition.indexID <> CurrentIndexID
    end;
  end; // condition
end; // FindRecordByScan


//------------------------------------------------------------------------------
// used by GetRecordBuffer - find or get record
//------------------------------------------------------------------------------
function TACRTableData.InternalFindOrGetRecordBuffer(
                              Cursor:         TACRCursor;
                              GetRecordMode:  TACRGetRecordMode;
                              // if specified - fill record bitmap
                              RecordBitmap:     TACRRecordBitmap = nil
                            ): TACRGetRecordResult;
var Restart, GoForward,
    RecordFound:        Boolean;
    RecordID:           TACRRecordID;
    NavigationInfo:     TACRNavigationInfo;
begin
   NavigationInfo.SessionID := Cursor.Session.SessionID;
   if (GetRecordMode <> grmCurrent) and (Cursor.IsViewRestricted) then
    begin
     // find
     Restart := (Cursor.FirstPosition or Cursor.LastPosition);
     GoForward := (GetRecordMode = grmNext);
     RecordFound := FindRecord(Cursor,nil,nil,Restart,GoForward,
                               @RecordID,RecordBitmap,True);
     if (RecordFound) then
      begin
       NavigationInfo.RecordBuffer := Cursor.CurrentRecordBuffer;
       NavigationInfo.RecordID := RecordID;
       NavigationInfo.FirstPosition := False;
       NavigationInfo.LastPosition := False;
       NavigationInfo.GetRecordMode := grmCurrent;
       FRecordManager.GetRecordBuffer(NavigationInfo);
       Result := NavigationInfo.GetRecordResult;
       Cursor.CurrentRecordID := NavigationInfo.RecordID;
       Cursor.FirstPosition := NavigationInfo.FirstPosition;
       Cursor.LastPosition := NavigationInfo.LastPosition;
      end
     else
      begin
       if (GetRecordMode = grmNext) then
        Result := grrEOF
       else
        Result := grrBOF;
      end;
    end
   else
    begin
       NavigationInfo.IndexID := Cursor.IndexID;
       NavigationInfo.RecordBuffer := Cursor.CurrentRecordBuffer;
       NavigationInfo.RecordID := Cursor.CurrentRecordID;
       NavigationInfo.FirstPosition := Cursor.FirstPosition;
       NavigationInfo.LastPosition := Cursor.LastPosition;
       NavigationInfo.GetRecordMode := GetRecordMode;
       InternalGetRecordBuffer(Cursor.Session.SessionID, NavigationInfo);
       Result := NavigationInfo.GetRecordResult;
       Cursor.CurrentRecordID := NavigationInfo.RecordID;
       Cursor.FirstPosition := NavigationInfo.FirstPosition;
       Cursor.LastPosition := NavigationInfo.LastPosition;
       if (Result = grrOK) and (Cursor.IsViewRestricted) then
        if (not IsRecordVisible(Cursor)) then
          Result := grrError;
    end;
end; // InternalFindOrGetRecordBuffer


//------------------------------------------------------------------------------
// get record using current index or physical order
//------------------------------------------------------------------------------
procedure TACRTableData.InternalGetRecordBuffer(
                                   SessionID:          TACRSessionID;
                                   var NavigationInfo: TACRNavigationInfo
                                      );
begin
  if (FRecordManager = nil) then
   raise EACRException.Create(10015,ErrorLNilPointer);
  if (FIndexManager = nil) then
   raise EACRException.Create(10380,ErrorLNilPointer);
  if (NavigationInfo.RecordBuffer = nil) then
   raise EACRException.Create(10016,ErrorLNilPointer);
  NavigationInfo.SessionID := SessionID;
  if ((NavigationInfo.IndexID <> INVALID_OBJECT_ID) and
    (NavigationInfo.GetRecordMode <> grmCurrent)) then
   begin
    FIndexManager.GetRecordBuffer(SessionID, NavigationInfo);
   end
  else
   begin
    FRecordManager.Repair := FRepair;
    FRecordManager.GetRecordBuffer(NavigationInfo);
   end;
end; // InternalGetRecordBuffer


//------------------------------------------------------------------------------
// update bitmap after insert record
//------------------------------------------------------------------------------
procedure TACRTableData.UpdateRecordBitmapAfterInsertRecord(Cursor: TACRCursor; Pos: Pointer);
var RecordVisible: Boolean;
begin
  RecordVisible := IsRecordVisible(Cursor);
  if (not RecordVisible) then
   begin
    Cursor.RestorePosition(Pos);
    if ((not Cursor.FirstPosition) and (not Cursor.LastPosition)) then
      GetRecordBuffer(Cursor,grmCurrent);
   end
  else
   begin
    if (not Cursor.IsIndexApplied) then
     TACRRecordBitmap(Cursor.RecordBitmap).InsertVisibleRecord(
      Cursor.CurrentRecordID,Cursor.CurrentRecordID,False,True)
    else
     ShowRecord(Cursor); // index applied
   end;
end; // UpdateRecordBitmapAfterInsertRecord


//------------------------------------------------------------------------------
// update bitmap after update record
//------------------------------------------------------------------------------
procedure TACRTableData.UpdateRecordBitmapAfterUpdateRecord(Cursor: TACRCursor);
var RecordVisible: Boolean;
    RecordID:      TACRRecordID;
begin
  RecordID := Cursor.CurrentRecordID;
  RecordVisible := IsRecordVisible(Cursor);
  if (not Cursor.IsIndexApplied) then
   begin
     if (not RecordVisible) then
      begin
       if (GetRecordBuffer(Cursor,grmNext) <> grrOK) then
        GetRecordBuffer(Cursor,grmPrior);
       TACRRecordBitmap(Cursor.RecordBitmap).HideRecord(RecordID);
      end;
   end // no index applied
  else
    begin
     if (not RecordVisible) then
      begin
        // move to next visible record
        if (GetRecordBuffer(Cursor,grmNext) <> grrOK) then
         GetRecordBuffer(Cursor,grmPrior);
        TACRRecordBitmap(Cursor.RecordBitmap).HideRecord(RecordID);
      end
     else
      begin
        if (FIndexManager.CompareRecordBuffersByIndex(Cursor.IndexID,
              Cursor.CurrentRecordBuffer,
              Cursor.EditRecordBuffer,
              0) <> 0) then
         begin
          TACRRecordBitmap(Cursor.RecordBitmap).HideRecord(RecordID);
          ShowRecord(Cursor);
         end; // visible record changed its poisition in current index
      end; // visible record
    end; // index applied
end; // UpdateRecordBitmapAfterUpdateRecord


//------------------------------------------------------------------------------
// return true if record exists
//------------------------------------------------------------------------------
function TACRTableData.IsRecordExists(Cursor: TACRCursor): Boolean;
begin
  Result := FRecordManager.IsRecordExists(Cursor.CurrentRecordID,Cursor.Session.SessionID);
end; // IsRecordExists


//------------------------------------------------------------------------------
// get record buffer
//------------------------------------------------------------------------------
function TACRTableData.GetRecordBuffer(
              Cursor:         TACRCursor;
              GetRecordMode:  TACRGetRecordMode
              ): TACRGetRecordResult;
var IsDistinctApplied:  Boolean;
    Acceptable:         Boolean;
    Buffer:             TACRRecordBuffer;
    NavigationInfo:     TACRNavigationInfo;
begin
 Lock;
 FRepair := Cursor.Repair;
 try
  FSessionID := Cursor.Session.SessionID;
  if (InternalGetRecordCount = 0) then
   begin
    Result := grrEOF;
    Exit;
   end;
  NavigationInfo.SessionID := Cursor.Session.SessionID;
  Result := grrError;
  // distinct
  IsDistinctApplied := ((Cursor.IsDistinctApplied) and
       ((GetRecordMode = grmPrior) or (GetRecordMode = grmNext)) and
       (not Cursor.FirstPosition) and (not Cursor.LastPosition));
  if (IsDistinctApplied) then
    Cursor.DistinctRecordBuffer := Cursor.AllocateRecordBuffer
  else
    Cursor.DistinctRecordBuffer := nil;
  try
  repeat
   Acceptable := True;
   if (Cursor.IsViewRestricted) then
    begin
     if (not TACRRecordBitmap(Cursor.RecordBitmap).Active) then
      BuildCursorRecordBitmap(Cursor);
     NavigationInfo.RecordBuffer := Cursor.CurrentRecordBuffer;
     NavigationInfo.RecordID := Cursor.CurrentRecordID;
     NavigationInfo.FirstPosition := Cursor.FirstPosition;
     NavigationInfo.LastPosition := Cursor.LastPosition;
     NavigationInfo.GetRecordMode := GetRecordMode;
     TACRRecordBitmap(Cursor.RecordBitmap).GetRecord(NavigationInfo);
     if (NavigationInfo.GetRecordResult = grrOK) then
      begin
       NavigationInfo.GetRecordMode := grmCurrent;
       NavigationInfo.FirstPosition := False;
       NavigationInfo.LastPosition := False;
       FRecordManager.GetRecordBuffer(NavigationInfo);
      end;
     Result := NavigationInfo.GetRecordResult;
     Cursor.CurrentRecordID := NavigationInfo.RecordID;
     Cursor.FirstPosition := NavigationInfo.FirstPosition;
     Cursor.LastPosition := NavigationInfo.LastPosition;
    end
   else
    begin
     // distinct
     if (IsDistinctApplied) then
      begin
       NavigationInfo.RecordBuffer := Cursor.DistinctRecordBuffer;
       NavigationInfo.RecordID := Cursor.CurrentRecordID;
       NavigationInfo.FirstPosition := Cursor.FirstPosition;
       NavigationInfo.LastPosition := Cursor.LastPosition;
       NavigationInfo.GetRecordMode := grmCurrent;
       FRecordManager.GetRecordBuffer(NavigationInfo);
       if (NavigationInfo.GetRecordResult <> grrOK) then
        break;
      end;
      Result := InternalFindOrGetRecordBuffer(Cursor,GetRecordMode);
    end;
   if ((GetRecordMode = grmCurrent) and (not Acceptable)) then
     Result := grrError;
  until ((Acceptable) or (Result <> grrOK));
 finally
//  if (IsDistinctApplied) then
   if (Cursor.DistinctRecordBuffer <> nil) then
    begin
      Buffer := Cursor.DistinctRecordBuffer;
      Cursor.FreeRecordBuffer(Buffer);
      Cursor.DistinctRecordBuffer := nil;
    end;
  end;
 finally
  FRepair := False;
  Unlock;
 end;
end; // GetRecordBuffer


//------------------------------------------------------------------------------
// locate
//------------------------------------------------------------------------------
function TACRTableData.Locate(Cursor: TACRCursor; SearchExpression: TACRExpression): Boolean;
var OldPos:      Pointer;
    RecordID:    TACRRecordID;
begin
 if (SearchExpression = nil) then
  raise EACRException.Create(11375,ErrorLNilPointer);
 Lock;
 Result := False;
 OldPos := Cursor.SavePosition;
 try
  Result := FindRecord(Cursor,SearchExpression,nil,True,True,@RecordID);
  if (Result) then
   begin
    Cursor.FirstPosition := False;
    Cursor.LastPosition := False;
    Cursor.CurrentRecordID := RecordID;
    GetRecordBuffer(Cursor,grmCurrent);
   end;
 finally
  if (not Result) then
   Cursor.RestorePosition(OldPos);
  Cursor.FreePosition(OldPos);
  Unlock;
 end;
end;// Locate


//------------------------------------------------------------------------------
// find key
//------------------------------------------------------------------------------
function TACRTableData.FindKey(Cursor: TACRCursor; SearchCondition: TACRSearchCondition): Boolean;
var OldPos:             Pointer;
    KeyCondition:       TACRScanSearchCondition;
    RecordID:           TACRRecordID;
begin
 Result := False;
 Lock;
 OldPos := Cursor.SavePosition;
 KeyCondition := TACRScanSearchCondition.Create;
 try
  KeyCondition.Condition := SearchCondition;
  KeyCondition.KeyFieldCount := Cursor.KeyFieldCount;
  KeyCondition.KeyRecordBuffer := Cursor.KeyBuffer;
  KeyCondition.IndexID := Cursor.IndexID;
  KeyCondition.Expression := nil;
  Result := FindRecord(Cursor,nil,KeyCondition,True,True,@RecordID);
  if (Result) then
   begin
    Cursor.FirstPosition := False;
    Cursor.LastPosition := False;
    Cursor.CurrentRecordID := RecordID;
   end;
 finally
  if (not Result) then
    Cursor.RestorePosition(OldPos);
  Cursor.FreePosition(OldPos);
  KeyCondition.Free;
  Unlock;
 end;
end; // FindKey


//------------------------------------------------------------------------------
// return true if BLOB value was modified before Post or Cancel
//------------------------------------------------------------------------------
function TACRTableData.IsBLOBModified(
                        RecordBuffer: TACRRecordBuffer;
                        FieldDef:     TACRFieldDef
                       ): Boolean;
begin
 Result := (PWord(RecordBuffer +
                  FieldDef.MemoryOffset + FieldDef.DiskDataSize)^ =
                    ACR_BLOB_MODIFIED);
end; // IsBLOBModified


//------------------------------------------------------------------------------
// set modified or not modified flag for the blob value in specified field
//------------------------------------------------------------------------------
procedure TACRTableData.SetBLOBModified(
                        Modified:     Boolean;
                        RecordBuffer: TACRRecordBuffer;
                        FieldDef:     TACRFieldDef
                       );
begin
  if (Modified) then
   PWord(RecordBuffer + FieldDef.MemoryOffset + FieldDef.DiskDataSize)^ :=
    ACR_BLOB_MODIFIED
  else
   PWord(RecordBuffer + FieldDef.MemoryOffset + FieldDef.DiskDataSize)^ :=
    ACR_BLOB_NOT_MODIFIED;
end; // SetBLOBModified


//------------------------------------------------------------------------------
// create blob stream
//------------------------------------------------------------------------------
function TACRTableData.InternalCreateBlobStream(
              Cursor:   TACRCursor;
              ToInsert: Boolean;
              FieldNo:  Integer;
              OpenMode: TACRBLOBOpenMode
              ): TACRStream;
var TempStream:             TACRTemporaryStream;
    CompressedStream:       TACRCompressedBLOBStream;
    Buffer:                 PChar;
    Offset,BufferSize:      Integer;
    BLOBDescriptor:         TACRBLOBDescriptor;
begin
 Lock;
 try
   Result := nil;
   BLOBDescriptor.CompressionAlgorithm :=
       Byte(FieldManager.FieldDefs[FieldNo].BLOBCompressionAlgorithm);
   BLOBDescriptor.CompressionMode := FieldManager.FieldDefs[FieldNo].BLOBCompressionMode;
   BLOBDescriptor.BlockSize := FieldManager.FieldDefs[FieldNo].BLOBBlockSize;
   if (BLOBDescriptor.BlockSize = 0) then
    raise EACRException.Create(10420,ErrorLZeroBlockSizeIsNotAllowed);
   BLOBDescriptor.StartPosition := 0;
   TempStream := TACRTemporaryStream.Create;
   // create new compressed stream
   if ((ToInsert and (OpenMode = bomWrite)) or (OpenMode = bomWrite) or
       (Cursor.CurrentRecordBuffer = nil) or
       (CheckNullFlag(FieldNo,Cursor.CurrentRecordBuffer))) then
    begin
     // empty stream
     BLOBDescriptor.NumBlocks := 0;
     BLOBDescriptor.UncompressedSize := 0;
     CompressedStream := TACRCompressedBLOBStream.Create(TempStream,
      BLOBDescriptor,True);
     Result := TACRLocalBLOBStream.Create(CompressedStream,Cursor,OpenMode,FieldNo);
    end // empty stream
   else
    begin
     // copy value from TableData
     Offset := FieldManager.FieldDefs[FieldNo].MemoryOffset;
     // get pointer to memory buffer with blob stream content

     Buffer := PChar(PCardinal(Cursor.CurrentRecordBuffer + Offset)^);
     if (Buffer = nil) then
      raise EACRException.Create(10112,ErrorLNilPointer);
     // creating source memory stream
     Offset := sizeof(TACRPartialBLOBDescriptor);
     BufferSize := MemoryManager.GetMemoryBufferSize(Buffer) - Offset;
     // copy partial blob descriptor
     BLOBDescriptor.NumBlocks := PACRPartialBLOBDescriptor(Buffer)^.NumBlocks;
     BLOBDescriptor.UncompressedSize := PACRPartialBLOBDescriptor(Buffer)^.UncompressedSize;
     TempStream.Write(PChar(Buffer + Offset)^,BufferSize);
     TempStream.Position := 0;
     CompressedStream := TACRCompressedBLOBStream.Create(TempStream,
      BLOBDescriptor,False);
     Result := TACRLocalBLOBStream.Create(CompressedStream,Cursor,OpenMode,FieldNo);
    end; // copy value from TableData
 finally
  Unlock;
 end;
end; // InternalCreateBlobStream


//------------------------------------------------------------------------------
// Write BLOB Field To Record Buffer
//------------------------------------------------------------------------------
procedure TACRTableData.WriteBLOBFieldToRecordBuffer(
              Cursor:     TACRCursor;
              FieldNo:    Integer;
              BLOBStream: TACRStream
              );
var Buffer:                   PChar;
    BufferSize,Offset:        Integer;
    ReadBytes:                Integer;
begin
 Lock;
 try
   Buffer := Cursor.CurrentRecordBuffer;
   if (BLOBStream.Modified) then
    begin
     if (BLOBStream.Size = 0) then
      begin
       // empty stream
       if (IsBLOBModified(Buffer,FieldManager.FieldDefs[FieldNo])) then
         ClearBLOBFieldInRecordBuffer(Buffer,FieldNo);
       Buffer := nil;
       Move(Buffer,PChar(Cursor.CurrentRecordBuffer +
        FieldManager.FieldDefs[FieldNo].MemoryOffset)^,sizeof(Buffer));
       SetNullFlag(True,FieldNo,Cursor.CurrentRecordBuffer);
      end
     else
      begin
       if (IsBLOBModified(Buffer,FieldManager.FieldDefs[FieldNo])) then
         ClearBLOBFieldInRecordBuffer(Buffer,FieldNo);
       Offset := sizeof(TACRPartialBLOBDescriptor);
       BufferSize := Integer(TACRCompressedBLOBStream(
          TACRLocalBLOBStream(BLOBStream).TemporaryStream).CompressedStream.Size) +
          Offset;
       Buffer := MemoryManager.GetMem(BufferSize);
       PACRPartialBLOBDescriptor(Buffer)^.NumBlocks := TACRCompressedBLOBStream(
          TACRLocalBLOBStream(BLOBStream).TemporaryStream).BLOBDescriptor.NumBlocks;
       PACRPartialBLOBDescriptor(Buffer)^.UncompressedSize := TACRCompressedBLOBStream(
          TACRLocalBLOBStream(BLOBStream).TemporaryStream).BLOBDescriptor.UncompressedSize;
       TACRCompressedBLOBStream(
          TACRLocalBLOBStream(BLOBStream).TemporaryStream).CompressedStream.Position := 0;

       ReadBytes := TACRCompressedBLOBStream(
          TACRLocalBLOBStream(BLOBStream).TemporaryStream).CompressedStream.
          Read(PChar(Buffer + Offset)^,(BufferSize - Offset));

       if (ReadBytes <> (BufferSize - Offset)) then
        begin
          MemoryManager.FreeAndNilMem(Buffer);
          raise EACRException.Create(10117,ErrorLCannotReadFromStream,
          [0,(BufferSize - Offset),(BufferSize - Offset),ReadBytes]);
        end;

       //Move(Buffer,PChar(Cursor.CurrentRecordBuffer +
       //  FieldManager.FieldDefs[FieldNo].MemoryOffset)^, sizeof(Buffer));
       PCardinal(Cursor.CurrentRecordBuffer +
          FieldManager.FieldDefs[FieldNo].MemoryOffset)^ := Cardinal(Buffer);
       SetBLOBModified(True,Cursor.CurrentRecordBuffer,
          FieldManager.FieldDefs[FieldNo]);
       SetNullFlag(False,FieldNo,Cursor.CurrentRecordBuffer);
      end; // not empty stream
    end; // Modified
 finally
  Unlock;
 end;
end; // WriteBLOBFieldToRecordBuffer


//------------------------------------------------------------------------------
// clear blob field
//------------------------------------------------------------------------------
procedure TACRTableData.ClearBLOBFieldInRecordBuffer(
              var RecordBuffer: TACRRecordBuffer;
              FieldNo:      Integer
              );
var Buffer:                   PChar;
begin
 Lock;
 try
   if (FBLOBFieldsPresent) then
    begin
     // check null
     if (not ((PByte(RecordBuffer + (FieldNo div 8))^) and
            (1 shl (FieldNo mod 8)) <> 0)) then
      begin
       Move(PChar(RecordBuffer +
         FieldManager.FieldDefs[FieldNo].MemoryOffset)^,Buffer,sizeof(Buffer));
       if (Buffer = nil) then
        raise EACRException.Create(10118,ErrorLNilPointer);
       MemoryManager.FreeAndNilMem(Buffer);

       // set null
       PByte(RecordBuffer + (FieldNo div 8))^ := PByte(RecordBuffer + (FieldNo div 8))^ or
        (1 shl (FieldNo mod 8));
      end;
    end;
 finally
  Unlock;
 end;
end; // ClearBLOBFieldInRecordBuffer


//------------------------------------------------------------------------------
// clear all blob fields in record buffer - for cancel, delete operation
//------------------------------------------------------------------------------
procedure TACRTableData.ClearBLOBFieldsInRecordBuffer(
              var RecordBuffer: TACRRecordBuffer
              );
var i: Integer;
begin
 if (FBLOBFieldsPresent) then
   for i := 0 to FieldManager.FieldDefs.Count - 1 do
    if (IsBLOBFieldType(FieldManager.FieldDefs[i].BaseFieldType)) then
     begin
      ClearBLOBFieldInRecordBuffer(RecordBuffer,i);
     end;
end; // ClearBLOBFieldsInRecordBuffer


//------------------------------------------------------------------------------
// check constraint conditions
//------------------------------------------------------------------------------
procedure TACRTableData.CheckConstraints(
                            Cursor:          TACRCursor;
                            SessionID:       TACRSessionID;
                            NewRecordBuffer: TACRRecordBuffer;
                            OldRecordBuffer: TACRRecordBuffer;
                            ToInsert:        Boolean;
                            CurrentRecordID: TACRRecordID;
                            SkipFKCheck:     Boolean = False
                          );
begin
  Lock;
  try
   FConstraintManager.CheckConstraints(
          Cursor,
          SessionID, NewRecordBuffer, OldRecordBuffer,
          ToInsert, CurrentRecordID, SkipFKCheck);
  finally
    Unlock;
  end;
end; // CheckConstraints


//------------------------------------------------------------------------------
// apply default values
//------------------------------------------------------------------------------
procedure TACRTableData.ApplyDefaultValuesToRecordBuffer(
                  Cursor:       TACRCursor;
                  RecordBuffer: TACRRecordBuffer;
                  InsertMode:   Boolean
                                          );
begin
  FieldManager.ApplyDefaultValuesToRecordBuffer(Cursor.Session, RecordBuffer,
    InsertMode);
end; // ApplyDefaultValuesToRecordBuffer


//------------------------------------------------------------------------------
// show current record
//------------------------------------------------------------------------------
procedure TACRTableData.ShowRecord(Cursor: TACRCursor);
var NavigationInfo: TACRNavigationInfo;
    RecordVisible:  Boolean;
    Size:           Integer;
begin
  Size := GetRecordBufferSize;
  Move(Cursor.CurrentRecordBuffer^,FTempRecordBuffer^,Size);
  try
    NavigationInfo.IndexID := Cursor.IndexID;
    NavigationInfo.RecordBuffer := Cursor.CurrentRecordBuffer;
    NavigationInfo.FirstPosition := False;
    NavigationInfo.LastPosition := False;
    FSessionID := Cursor.Session.SessionID;
    NavigationInfo.RecordID := Cursor.CurrentRecordID;
    NavigationInfo.SessionID := FSessionID;
    NavigationInfo.GetRecordMode := grmPrior;
    repeat
     InternalGetRecordBuffer(FSessionID,NavigationInfo);
     if (NavigationInfo.GetRecordResult <> grrOK) then
      break;
     RecordVisible := IsRecordVisible(Cursor);
    until (RecordVisible);
    if (NavigationInfo.GetRecordResult <> grrOK) then
     TACRRecordBitmap(Cursor.RecordBitmap).InsertVisibleRecord(
      Cursor.CurrentRecordID,Cursor.CurrentRecordID,True,False)
     else
     TACRRecordBitmap(Cursor.RecordBitmap).InsertVisibleRecord(
      Cursor.CurrentRecordID,NavigationInfo.RecordID,False,False);
  finally
    Move(FTempRecordBuffer^,Cursor.CurrentRecordBuffer^,Size);
  end;
end; // ShowRecord


//------------------------------------------------------------------------------
// start record editing
//------------------------------------------------------------------------------
procedure TACRTableData.EditRecord(Cursor: TACRCursor);
begin
;
end; // EditRecord


//------------------------------------------------------------------------------
// lear temporary blob values created between Insert / Edit and Post / Cancel
//------------------------------------------------------------------------------
procedure TACRTableData.ClearTemporaryBLOBValues(Cursor: TACRCursor);
var i:            Integer;
    RecordBuffer: TACRRecordBuffer;
begin
  RecordBuffer := Cursor.CurrentRecordBuffer;
  if (FBLOBFieldsPresent) then
   for i := 0 to FFieldManager.FieldDefs.Count - 1 do
    if (IsBLOBFieldType(FFieldManager.FieldDefs[i].BaseFieldType)) then
     if (IsBLOBModified(Cursor.CurrentRecordBuffer,FFieldManager.FieldDefs[i])) then
      ClearBLOBFieldInRecordBuffer(RecordBuffer,i);
end; // CancelRecord


//------------------------------------------------------------------------------
// lear temporary blob values created between Insert / Edit and Post / Cancel
//------------------------------------------------------------------------------
procedure TACRTableData.SetBLOBValuesModified(Modified: Boolean; Cursor: TACRCursor);
var i:            Integer;
    RecordBuffer: TACRRecordBuffer;
begin
  RecordBuffer := Cursor.CurrentRecordBuffer;
  if (FBLOBFieldsPresent) then
   for i := 0 to FFieldManager.FieldDefs.Count - 1 do
    if (IsBLOBFieldType(FFieldManager.FieldDefs[i].BaseFieldType)) then
     if (not CheckNullFlag(i,RecordBuffer)) then
      SetBLOBModified(Modified,Cursor.CurrentRecordBuffer,FFieldManager.FieldDefs[i]);
end; // SetBLOBValuesModified


//------------------------------------------------------------------------------
// cancel record inserting or editing
//------------------------------------------------------------------------------
procedure TACRTableData.CancelRecord(Cursor: TACRCursor; ToInsert: Boolean);
begin
  ClearTemporaryBLOBValues(Cursor);
end; // CancelRecord


//------------------------------------------------------------------------------
// delete all visible records
//------------------------------------------------------------------------------
procedure TACRTableData.DeleteVisibleRecords(Cursor: TACRCursor);
var
    i,RecCount:     TACRRecordNo;
    NavigationInfo: TACRNavigationInfo;
    Filtered:       Boolean;
begin
  NavigationInfo.RecordBuffer := Cursor.CurrentRecordBuffer;
  NavigationInfo.IndexID := INVALID_OBJECT_ID;
  Filtered := TACRRecordBitmap(Cursor.RecordBitmap).Active;
  Cursor.FirstPosition := False;
  Cursor.LastPosition := False;
  if (Filtered) then
   begin
    RecCount := TACRRecordBitmap(Cursor.RecordBitmap).GetRecordCount;
    NavigationInfo.GetRecordMode := grmCurrent;
    NavigationInfo.FirstPosition := False;
    NavigationInfo.LastPosition := False;
   end // filtered
  else
   begin
    RecCount := InternalGetRecordCount;
    NavigationInfo.GetRecordMode := grmNext;
    NavigationInfo.FirstPosition := True;
    NavigationInfo.LastPosition := False;
   end; // not filtered - all records are visible
  i := 1;
  while (i <= RecCount) do
   begin
    if (Filtered) then
     NavigationInfo.RecordID := TACRRecordBitmap(Cursor.RecordBitmap).GetRecordIDByRecNo(i);
    InternalGetRecordBuffer(FSessionID,NavigationInfo);
    Cursor.CurrentRecordID := NavigationInfo.RecordID;
    if (NavigationInfo.GetRecordResult <> grrOK) then
     raise EACRException.Create(11466,ErrorLCannotLoadRecord,[NavigationInfo.RecordID.PageNo,NavigationInfo.RecordID.PageItemNo,FTableName]);
    if (FConstraintManager.ConstraintDefs.ForeignKeysActionsUpdateExists) then
     ExecuteForeignKeyActions(Cursor,False);
    if (FIndexManager.IndexDefs.Count > 0) then
     FIndexManager.DeleteRecord(Cursor);
    FRecordManager.DeleteRecord(Cursor.CurrentRecordID,FSessionID);
    Inc(i);
   end;
end; // DeleteVisibleRecords


//------------------------------------------------------------------------------
// update all visible records
//------------------------------------------------------------------------------
procedure TACRTableData.UpdateVisibleRecords(
                                Cursor:       TACRCursor;
                                FieldNames:   TStringList;
                                values:       array of TACRVariant;
                                SkipFKCheck:  Boolean = False
                                            );
var j,FieldNo:      Integer;
    i,RecCount:     TACRRecordNo;
    NavigationInfo: TACRNavigationInfo;
    Buffer:         PChar;
    Filtered:       Boolean;
begin
  j := Length(values);
  if ((j <= 0) or (j <> FieldNames.Count)) then
   raise EACRException.Create(11448,ErrorLInvalidItemNumber,[j,FieldNames.Count]);
  NavigationInfo.RecordBuffer := Cursor.CurrentRecordBuffer;
  NavigationInfo.IndexID := INVALID_OBJECT_ID;
  Filtered := TACRRecordBitmap(Cursor.RecordBitmap).Active;
  Cursor.FirstPosition := False;
  Cursor.LastPosition := False;
  if (Filtered) then
   begin
    RecCount := TACRRecordBitmap(Cursor.RecordBitmap).GetRecordCount;
    NavigationInfo.GetRecordMode := grmCurrent;
    NavigationInfo.FirstPosition := False;
    NavigationInfo.LastPosition := False;
   end // filtered
  else
   begin
    RecCount := InternalGetRecordCount;
    NavigationInfo.GetRecordMode := grmNext;
    NavigationInfo.FirstPosition := True;
    NavigationInfo.LastPosition := False;
   end; // not filtered - all records are visible
  Cursor.EditRecordBuffer := Cursor.AllocateRecordBuffer;
  try
    i := 1;
    while (i <= RecCount) do
     begin
      if (Filtered) then
       NavigationInfo.RecordID := TACRRecordBitmap(Cursor.RecordBitmap).GetRecordIDByRecNo(i);
      InternalGetRecordBuffer(FSessionID,NavigationInfo);
      Cursor.CurrentRecordID := NavigationInfo.RecordID;
      if (NavigationInfo.GetRecordResult <> grrOK) then
       raise EACRException.Create(11449,ErrorLCannotLoadRecord,[NavigationInfo.RecordID.PageNo,NavigationInfo.RecordID.PageItemNo,FTableName]);
      Move(Cursor.CurrentRecordBuffer^,Cursor.EditRecordBuffer^,Cursor.RecordBufferSize);
      for j := 0 to FieldNames.Count -1 do
       begin
        FieldNo := Cursor.VisibleFieldDefs.GetDefNumberByName(FieldNames.Strings[j]);
        if (FieldNo < 0) then
         raise EACRException.Create(11447,ErrorLCannotFindFieldInTable,[FieldNames.Strings[j],FTableName]);
        Cursor.SetFieldValue(values[j],FieldNo,False);
       end;
      CheckConstraints(Cursor,FSessionID,Cursor.CurrentRecordBuffer,
                       Cursor.FEditRecordBuffer,False,Cursor.CurrentRecordID,SkipFKCheck);
      if (FConstraintManager.ConstraintDefs.ForeignKeysActionsUpdateExists) then
       ExecuteForeignKeyActions(Cursor,True);
      if (FIndexManager.IndexDefs.Count > 0) then
       FIndexManager.UpdateRecord(Cursor);
      FRecordManager.UpdateRecord(Cursor.CurrentRecordBuffer,Cursor.CurrentRecordID,FSessionID);
      Inc(i);
     end;
  finally
    Buffer := Cursor.EditRecordBuffer;
    Cursor.FreeRecordBuffer(Buffer);
    Cursor.EditRecordBuffer := nil;
  end;
end; // UpdateVisibleRecords


//------------------------------------------------------------------------------
// return record count
//------------------------------------------------------------------------------
function TACRTableData.GetRecordCount(Cursor: TACRCursor; InternalCall: Boolean): TACRRecordNo;
begin
 if (Cursor.IsViewRestricted) then
  begin
   if (not TACRRecordBitmap(Cursor.RecordBitmap).Active) then
    BuildCursorRecordBitmap(Cursor);
   Result := TACRRecordBitmap(Cursor.RecordBitmap).GetRecordCount;
  end
 else
  Result := InternalGetRecordCount;
end; // GetRecordCount


//------------------------------------------------------------------------------
// set rec no
//------------------------------------------------------------------------------
procedure TACRTableData.SetRecNo(Cursor: TACRCursor; RecNo: TACRRecordNo);
var NavigationInfo: TACRNavigationInfo;
    RecCount:       TACRRecordNo;
begin
 Lock;
 try
   RecCount := GetRecordCount(Cursor,True);
   if (RecCount > 0) then
    begin
     if (RecNo >= RecCount) then
      RecNo := RecCount;
     Cursor.FirstPosition := False;
     Cursor.LastPosition := False;
     if (Cursor.IsViewRestricted) then
      begin
       Cursor.CurrentRecordID := TACRRecordBitmap(Cursor.RecordBitmap).
          GetRecordIDByRecNo(RecNo);
      end
     else
     if (Cursor.IsIndexApplied) then
      begin
       NavigationInfo.RecordBuffer := MemoryManager.AllocMem(GetRecordBufferSize);
       try
         Cursor.CurrentRecordID :=
            FIndexManager.OpenIndex(Cursor.IndexID).GetRecordIDByRecNo(
                                                  Cursor.Session.SessionID,
                                                  RecNo,
                                                  Cursor.RecordBitmap
                                                                   );
       finally
         MemoryManager.FreeAndNilMem(NavigationInfo.RecordBuffer);
       end;
      end // index applied
     else
      InternalSetRecNo(Cursor,RecNo);
    end; // get recno
 finally
  Unlock;
 end;
end; // SetRecNo


//------------------------------------------------------------------------------
// return rec no
//------------------------------------------------------------------------------
function TACRTableData.GetRecNo(Cursor: TACRCursor): TACRRecordNo;
var NavigationInfo: TACRNavigationInfo;
begin
 Lock;
 try
   if (Cursor.FirstPosition) or (Cursor.LastPosition) or (GetRecordCount(Cursor,True) = 0) then
    Result := -1
   else
    begin
     if (Cursor.IsViewRestricted) then
      Result := TACRRecordBitmap(Cursor.RecordBitmap).GetRecNoByRecordID(
                                                      Cursor.CurrentRecordID
                                                                        )
     else
     if (Cursor.IsIndexApplied) then
      begin
         NavigationInfo.RecordBuffer := MemoryManager.AllocMem(GetRecordBufferSize);
         try
           NavigationInfo.FirstPosition := False;
           NavigationInfo.LastPosition := False;
           NavigationInfo.RecordID := Cursor.CurrentRecordID;
           NavigationInfo.GetRecordMode := grmCurrent;
           InternalGetRecordBuffer(Cursor.Session.SessionID, NavigationInfo);
           if (NavigationInfo.GetRecordResult <> grrOK) then
            Result := -1
           else
            Result := FIndexManager.OpenIndex(Cursor.IndexID).GetRecNoByRecordID(
                                                    Cursor.Session.SessionID,
                                                    Cursor.CurrentRecordID,
                                                    NavigationInfo.RecordBuffer,
                                                    Cursor.RecordBitmap
                                                                     );
         finally
           MemoryManager.FreeAndNilMem(NavigationInfo.RecordBuffer);
         end;
      end // index applied
     else
      Result := InternalGetRecNo(Cursor);
    end; // get recno
 finally
  Unlock;
 end;
end; // GetRecNo


//------------------------------------------------------------------------------
// return last auto inc value
//------------------------------------------------------------------------------
function TACRTableData.LastAutoincValue(FieldNo: Integer; Session: TACRBaseSession): Int64;
begin
 Result := 0;
end; // LastAutoincValue


//------------------------------------------------------------------------------
// set autoinc value
//------------------------------------------------------------------------------
procedure TACRTableData.SetLastAutoincValue(Value: Int64; FieldNo: Integer; Cursor: TACRCursor);
begin
;
end; // SetLastAutoincValue


//------------------------------------------------------------------------------
// lock table for read and free data for fast navigation
//------------------------------------------------------------------------------
procedure TACRTableData.LockTableForRead(aSessionID: TACRSessionID);
begin
  Inc(FNumLocksForRead);
end; // UnlockTableForRead


//------------------------------------------------------------------------------
// unlock table for read and free data for fast navigation
//------------------------------------------------------------------------------
procedure TACRTableData.UnlockTableForRead(aSessionID: TACRSessionID);
begin
  Dec(FNumLocksForRead);
end; // UnlockTableForRead


//------------------------------------------------------------------------------
// lock table in S mode
//------------------------------------------------------------------------------
function TACRTableData.TryToLockTableS: Boolean;
begin
 Result := True;
end; // LockTableS


//------------------------------------------------------------------------------
// unlock table in S mode
//------------------------------------------------------------------------------
function TACRTableData.TryToUnlockTableS: Boolean;
begin
 Result := True;
end; // UnlockTableS


//------------------------------------------------------------------------------
// lock table in IRW mode
//------------------------------------------------------------------------------
function TACRTableData.TryToLockTableIRW: Boolean;
begin
 Result := True;
end; // LockTableIRW


//------------------------------------------------------------------------------
// unlock table in IRW mode
//------------------------------------------------------------------------------
function TACRTableData.TryToUnlockTableIRW: Boolean;
begin
 Result := True;
end; // UnlockTableIRW


//------------------------------------------------------------------------------
// lock table in RW mode
//------------------------------------------------------------------------------
function TACRTableData.TryToLockTableRW: Boolean;
begin
 Result := True;
end; // LockTableRW


//------------------------------------------------------------------------------
// unlock table in RW mode
//------------------------------------------------------------------------------
function TACRTableData.TryToUnlockTableRW: Boolean;
begin
 Result := True;
end; // UnockTableRW


//------------------------------------------------------------------------------
// lock table in RW mode
//------------------------------------------------------------------------------
function TACRTableData.TryToLockTableX: Boolean;
begin
 Result := True;
end; // LockTableX


//------------------------------------------------------------------------------
// unlock table in RW mode
//------------------------------------------------------------------------------
function TACRTableData.TryToUnlockTableX: Boolean;
begin
 Result := True;
end; // UnockTableX


//------------------------------------------------------------------------------
// lock table in RW mode
//------------------------------------------------------------------------------
function TACRTableData.TryToLockTableIS: Boolean;
begin
 Result := True;
end; // LockTableIS


//------------------------------------------------------------------------------
// unlock table in RW mode
//------------------------------------------------------------------------------
function TACRTableData.TryToUnlockTableIS: Boolean;
begin
 Result := True;
end; // UnockTableIS


//------------------------------------------------------------------------------
// start transaction
//------------------------------------------------------------------------------
procedure TACRTableData.StartTransaction(aSessionID: TACRSessionID);
begin
;
end; // StartTransaction


//------------------------------------------------------------------------------
// Commit
//------------------------------------------------------------------------------
procedure TACRTableData.Commit(aSessionID: TACRSessionID);
begin
;
end; // Commit


//------------------------------------------------------------------------------
// Rollback
//------------------------------------------------------------------------------
procedure TACRTableData.Rollback(aSessionID: TACRSessionID);
begin
;
end; // Rollback


//------------------------------------------------------------------------------
// return true if read page synchronization is required
//------------------------------------------------------------------------------
function TACRTableData.IsSynchronizationRequired(PageState: Cardinal): Boolean;
begin
  Result := False;
end; // IsSynchronizationRequired


////////////////////////////////////////////////////////////////////////////////
//
// TACRTransaction
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// save cursor positions
//------------------------------------------------------------------------------
procedure TACRTransaction.SaveCursorPositions;
var i,j:        Integer;
    TableData:  TACRTableData;
    Cursor:     TACRLocalCursor;
begin
  for i := 0 to FDatabaseData.FTableDataList.Count - 1 do
   begin
    TableData := TACRTableData(FDatabaseData.FTableDataList.Items[i]);
    for j := 0 to TableData.CursorList.Count - 1 do
     begin
      Cursor := TACRLocalCursor(TableData.CursorList.Items[j]);
      if (Cursor.Session.SessionID = FSessionID) then
        Cursor.SaveCurrentPosition;
     end;
   end;
end; // SaveCursorPositions


//------------------------------------------------------------------------------
// restore cursor positions
//------------------------------------------------------------------------------
procedure TACRTransaction.RestoreCursorPositions;
var i,j:        Integer;
    TableData:  TACRTableData;
    Cursor:     TACRLocalCursor;
begin
  for i := 0 to FDatabaseData.FTableDataList.Count - 1 do
   begin
    TableData := TACRTableData(FDatabaseData.FTableDataList.Items[i]);
    for j := 0 to TableData.CursorList.Count - 1 do
     begin
      Cursor := TACRLocalCursor(TableData.CursorList.Items[j]);
      if (Cursor.Session.SessionID = FSessionID) then
        Cursor.RestoreSavedPosition;
     end;
   end;
end; // SaveCursorPositions


//------------------------------------------------------------------------------
// call start transaction
//------------------------------------------------------------------------------
procedure TACRTransaction.CallStartTransactionInTableData;
var i,j:        Integer;
    TableData:  TACRTableData;
    Cursor:     TACRLocalCursor;
    bOk:        Boolean;
begin
  for i := 0 to FDatabaseData.FTableDataList.Count - 1 do
   begin
    TableData := TACRTableData(FDatabaseData.FTableDataList.Items[i]);
    bOk := False;
    for j := 0 to TableData.CursorList.Count - 1 do
     begin
      Cursor := TACRLocalCursor(TableData.CursorList.Items[j]);
      if (Cursor.Session.SessionID = FSessionID) then
       begin
        bOk := True;
        break;
       end;
     end;
    if (bOk) then
     TableData.StartTransaction(FSessionID);
   end;
end; // CallStartTransactionInTableData


//------------------------------------------------------------------------------
// call rollback
//------------------------------------------------------------------------------
procedure TACRTransaction.CallRollbackInTableData;
var i,j:        Integer;
    TableData:  TACRTableData;
    Cursor:     TACRLocalCursor;
    bOk:        Boolean;
begin
  for i := 0 to FDatabaseData.FTableDataList.Count - 1 do
   begin
    TableData := TACRTableData(FDatabaseData.FTableDataList.Items[i]);
    bOk := False;
    for j := 0 to TableData.CursorList.Count - 1 do
     begin
      Cursor := TACRLocalCursor(TableData.CursorList.Items[j]);
      if (Cursor.Session.SessionID = FSessionID) then
       begin
        bOk := True;
        break;
       end;
     end;
    if (bOk) then
     TableData.Rollback(FSessionID);
   end;
end; // CallRollbackInTableData


//------------------------------------------------------------------------------
// call commit
//------------------------------------------------------------------------------
procedure TACRTransaction.CallCommitInTableData;
var i,j:        Integer;
    TableData:  TACRTableData;
    Cursor:     TACRLocalCursor;
    bOk:        Boolean;
begin
  for i := 0 to FDatabaseData.FTableDataList.Count - 1 do
   begin
    TableData := TACRTableData(FDatabaseData.FTableDataList.Items[i]);
    bOk := False;
    for j := 0 to TableData.CursorList.Count - 1 do
     begin
      Cursor := TACRLocalCursor(TableData.CursorList.Items[j]);
      if (Cursor.Session.SessionID = FSessionID) then
       begin
        bOk := True;
        break;
       end;
     end;
    if (bOk) then
     TableData.Commit(FSessionID);
   end;
end; // CallCommitInTableData


//------------------------------------------------------------------------------
// unlock tables
//------------------------------------------------------------------------------
procedure TACRTransaction.UnlockTables;
var i:                        Integer;
    TableLockItemForSession:  PACRTableLockItemForSession;
begin
  try
   for i := 0 to FTableLocksList.Count - 1 do
    begin
     TableLockItemForSession := PACRTableLockItemForSession(FTableLocksList.Items[i]);
     if (not FDatabaseData.UnlockTableAll(TableLockItemForSession^.TableName,
              FSessionID,@TableLockItemForSession^.TableLockInfo)) then
      raise EACRException.Create(10825,ErrorLCannotUnlockTableInTransaction,[TableLockItemForSession^.TableName]);
    end;
  finally
    while (FTableLocksList.Count > 0) do
     begin
      Dispose(FTableLocksList.Items[0]);
      FTableLocksList.Remove(FTableLocksList.Items[0]);
     end;
  end;
end; // UnlockTables


//------------------------------------------------------------------------------
// return true if all tables locke in IRW was successfully locked in RW
//------------------------------------------------------------------------------
function TACRTransaction.TryToChangeIRWLockToRWLock: Boolean;
var i:                        Integer;
    TableLockItemForSession:  PACRTableLockItemForSession;
begin
   Result := True;
   for i := 0 to FTableLocksList.Count - 1 do
    begin
     TableLockItemForSession := PACRTableLockItemForSession(FTableLocksList.Items[i]);
     if (TableLockItemForSession^.TableLockInfo.NumLocksIRW > 0) then
      begin
       Result := FDatabaseData.TryToLockTable(TableLockItemForSession^.TableName,
                  FSessionID,ltRW,@TableLockItemForSession^.TableLockInfo);
       if (not Result) then
        break;
      end;
    end;
end; // TryToChangeIRWLockToRWLock


//------------------------------------------------------------------------------
// check for other locks (not IS and X) and raise exception if found
// otherwise opentable in transaction
//------------------------------------------------------------------------------
procedure TACRTransaction.OpenTableForReadAndCheckForOtherLocks(
              TableLockItemForSession: PACRTableLockItemForSession);
begin
 if ((TableLockItemForSession^.TableLockInfo.NumLocksS > 0) or
     (TableLockItemForSession^.TableLockInfo.NumLocksIRW > 0) or
     (TableLockItemForSession^.TableLockInfo.NumLocksRW > 0) or
     (TableLockItemForSession^.TableLockInfo.IsRecordLocked)) then
  raise EACRException.Create(10824,ErrorLCannotStrartTransaction,[TableLockItemForSession^.TableName]);
 OpenTableForRead(TableLockItemForSession^.TableName,(TableLockItemForSession^.TableLockInfo.NumLocksX > 0));
end; // OpenTableForReadAndCheckForOtherLocks


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRTransaction.Create(aSession: TACRBaseSession; aDatabaseData: TACRDatabaseData);
var TempList: TList;
    i:        Integer;
begin
 FIsFinished := False;
 FSession := aSession;
 FSessionID := aSession.SessionID;
 FDatabaseData := aDatabaseData;
 FTableLocksList := TList.Create;
 TempList := TList.Create;
 try
   FDatabaseData.GetTableLocksListForSession(FSessionID,TempList);
   for i := 0 to TempList.Count - 1 do
    OpenTableForReadAndCheckForOtherLocks(PACRTableLockItemForSession(TempList.Items[i]));
   CallStartTransactionInTableData;
   SaveCursorPositions;
 finally
   for i := 0 to TempList.Count - 1 do
    Dispose(TempList.Items[i]);
   TempList.Free;
 end;
end; // Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRTransaction.Destroy;
begin
 FTableLocksList.Free;
 inherited;
 if (not FIsFinished) then
  raise EACRException.Create(10819,ErrorLTransactionIsNotFinished);
end; // Destroy


//------------------------------------------------------------------------------
// commit
//------------------------------------------------------------------------------
procedure TACRTransaction.Commit(FlushFileBuffers: Boolean);
begin
 if (not TryToChangeIRWLockToRWLock) then
  begin
   Rollback;
   raise EACRException.Create(10826,ErrorLTransactionCannotBeCommittedDueToOtherTableLocks);
  end;
 try
   CallCommitInTableData;
   FDatabaseData.PageManager.ApplyChanges(FSession.SessionID);
   if (FlushFileBuffers) then
     FDatabaseData.PageManager.FlushFileBuffers;
 finally
   FIsFinished := True;
   UnlockTables;
 end;
end; // Commit


//------------------------------------------------------------------------------
// rollback
//------------------------------------------------------------------------------
procedure TACRTransaction.Rollback;
begin
 try
   FDatabaseData.PageManager.CancelChanges(FSession.SessionID);
   CallRollbackInTableData;
 finally
   RestoreCursorPositions;
   FIsFinished := True;
   UnlockTables;
 end;
end; // Rollback


//------------------------------------------------------------------------------
// open table for read in IS + S or X modes
//------------------------------------------------------------------------------
procedure TACRTransaction.OpenTableForRead(TableName: String; OpenExclusive: Boolean);
var i:                        Integer;
    bOK:                      Boolean;
    TableLockItemForSession:  PACRTableLockItemForSession;
begin
 bOK := True;
 for i := 0 to FTableLocksList.Count - 1 do
  begin
   TableLockItemForSession := FTableLocksList.Items[i];
   if (AnsiUpperCase(TableName) = TableLockItemForSession^.TableName) then
    begin
     bOK := false;
     break;
    end;
  end;
 if (bOK) then
  begin
   // add new item
   New(TableLockItemForSession);
   FTableLocksList.Add(TableLockItemForSession);
   FillChar(TableLockItemForSession^.TableLockInfo,
            SizeOf(TableLockItemForSession^.TableLockInfo),$0);
   TableLockItemForSession^.TableName := AnsiUpperCase(TableName);
  end;
 if (OpenExclusive) then
  begin
   bOK := FDatabaseData.TryToLockTable(TableName,FSessionID,ltX,@TableLockItemForSession^.TableLockInfo);
   if (not bOK) then
    raise EACRException.Create(10827,ErrorLCannotLockTable,['X',TableName]);
  end
 else
  begin
   bOK := FDatabaseData.TryToLockTable(TableName,FSessionID,ltIS,@TableLockItemForSession^.TableLockInfo);
   if (not bOK) then
    raise EACRException.Create(10828,ErrorLCannotLockTable,['IS',TableName]);
   if (bOK) then
    begin
     bOK := FDatabaseData.TryToLockTable(TableName,FSessionID,ltS,@TableLockItemForSession^.TableLockInfo);
     if (not bOK) then
      raise EACRException.Create(10829,ErrorLCannotLockTable,['S',TableName]);
    end;
  end;
end; // OpenTableForRead


//------------------------------------------------------------------------------
// open table in IRW mode
//------------------------------------------------------------------------------
procedure TACRTransaction.StartTableModification(TableName: String);
var i:                        Integer;
    bOK:                      Boolean;
    TableLockItemForSession:  PACRTableLockItemForSession;
begin
 bOK := False;
 for i := 0 to FTableLocksList.Count - 1 do
  begin
   TableLockItemForSession := FTableLocksList.Items[i];
   if (AnsiUpperCase(TableName) = TableLockItemForSession^.TableName) then
    begin
     bOK := True;
     break;
    end;
  end;
 if (not bOK) then
  raise EACRException.Create(10830,ErrorLTableIsNotOpened,[TableName]);
 if (TableLockItemForSession^.TableLockInfo.NumLocksIRW = 0) then
  begin
   bOK := FDatabaseData.TryToLockTable(TableName,FSessionID,ltIRW,@TableLockItemForSession^.TableLockInfo);
   if (not bOK) then
    raise EACRException.Create(10831,ErrorLCannotLockTable,['IRW',TableName]);
  end;
end; // StartTableModification


////////////////////////////////////////////////////////////////////////////////
//
// TACRDatabaseData
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRDatabaseData.Create;
begin
 InitializeCriticalSection(FCSect);
 FSessionList := TList.Create;
 FTableDataList := TList.Create;
 FObjectIdSequence := TACRSequenceDef.Create;
 FPageManager := nil;
end;// Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRDatabaseData.Destroy;
var
  DBDatas: TList;
begin
 FObjectIdSequence.Free;
 while (FTableDataList.Count > 0) do
   TACRTableData(FTableDataList.Items[0]).Free;
 FTableDataList.Free;
 FSessionList.Free;

  DBDatas := DBDataList.LockList;
  DBDatas.Remove(Self);
  DBDataList.UnlockList;

 DeleteCriticalSection(FCSect);
end;// Destroy


//------------------------------------------------------------------------------
// lock
//------------------------------------------------------------------------------
procedure TACRDatabaseData.Lock;
begin
  EnterCriticalSection(FCSect);
end;// Lock


//------------------------------------------------------------------------------
// unlock
//------------------------------------------------------------------------------
procedure TACRDatabaseData.Unlock;
begin
  LeaveCriticalSection(FCSect);
end;// Unlock


//------------------------------------------------------------------------------
// Connect local session
//------------------------------------------------------------------------------
procedure TACRDatabaseData.ConnectSession(Session: TACRBaseSession);
begin
  Lock;
  try
   FSessionList.Add(Session);
  finally
   Unlock;
  end;
end;// ConnectSession


//------------------------------------------------------------------------------
// disconnect local session
//------------------------------------------------------------------------------
procedure TACRDatabaseData.DisconnectSession(Session: TACRBaseSession);
begin
  Lock;
  try
   FSessionList.Remove(Session);
  finally
   Unlock;
  end;
end;// DisconnectSession


//------------------------------------------------------------------------------
// FreeIfNoSessionsConnected
//------------------------------------------------------------------------------
procedure TACRDatabaseData.FreeIfNoSessionsConnected;
begin

end;// FreeIfNoSessionsConnected


//------------------------------------------------------------------------------
// get list of tables
//------------------------------------------------------------------------------
procedure TACRDatabaseData.GetTablesList(List: TStrings);
var
  i: Integer;
begin
  Lock;
  try
   for i := 0 to FTableDataList.Count-1 do
    List.Add(TACRTableData(FTableDataList.Items[i]).TableName);
  finally
   Unlock;
  end;
end;// GetTablesList


//------------------------------------------------------------------------------
// TableExists
//------------------------------------------------------------------------------
function TACRDatabaseData.TableExists(TableName: String): Boolean;
var
  List: TStringList;
begin
  List := TStringList.Create;
  try
    GetTablesList(List);
    Result := List.IndexOf(TableName) > -1;
  finally
    List.Free;
  end;
end;// TableExists


//------------------------------------------------------------------------------
// CreateDatabase
//------------------------------------------------------------------------------
procedure TACRDatabaseData.CreateDatabase(Session: TACRBaseSession);
begin
end;// CreateDatabase


//------------------------------------------------------------------------------
// find table data
//------------------------------------------------------------------------------
function TACRDatabaseData.FindTableData(Cursor: TACRCursor): TACRTableData;
var
  i: Integer;
begin
  Lock;
  try
   Result := nil;
   for i := 0 to FTableDataList.Count-1 do
    if (AnsiUpperCase(TACRTableData(FTableDataList.Items[i]).TableName) =
        AnsiUpperCase(Cursor.TableName)) then
     begin
      Result := FTableDataList.Items[i];
      break;
     end;
  finally
   Unlock;
  end;
end;// FindTableData


//------------------------------------------------------------------------------
// find or create table data
//------------------------------------------------------------------------------
function TACRDatabaseData.FindOrCreateTableData(Cursor: TACRCursor): TACRTableData;
begin
  Lock;
  try
   Result := FindTableData(Cursor);
   if (Result = nil) then
     Result := CreateTableData(Cursor);
  finally
   Unlock;
  end;
end;// FindOrCreateTableData


//------------------------------------------------------------------------------
// GetNewObjectId
//------------------------------------------------------------------------------
function TACRDatabaseData.GetNewObjectId: TACRObjectID;
begin
  Result := FObjectIdSequence.GetNextVal;
end;//GetNewObjectId


//------------------------------------------------------------------------------
// flush file buffers
//------------------------------------------------------------------------------
procedure TACRDatabaseData.FlushFileBuffers;
begin
end; // FlushFileBuffers


//------------------------------------------------------------------------------
// GetFormatVersion
//------------------------------------------------------------------------------
function TACRDatabaseData.GetFormatVersion(Session: TACRBaseSession): Double;
begin
 Result := 0.0;
end;// GetFormatVersion


//------------------------------------------------------------------------------
// get count of free pages
//------------------------------------------------------------------------------
function TACRDatabaseData.GetFreePageCount(Session: TACRBaseSession): Integer;
begin
 Result := 0;
end;// GetFreePageCount


//------------------------------------------------------------------------------
// get total count of pages
//------------------------------------------------------------------------------
function TACRDatabaseData.GetTotalPageCount(Session: TACRBaseSession): Integer;
begin
 Result := 0;
end;// GetTotalPageCount


//------------------------------------------------------------------------------
// return true if database is encrypted
//------------------------------------------------------------------------------
function TACRDatabaseData.IsDatabaseEncrypted(Session: TACRBaseSession): Boolean;
begin
  Result := False;
end; // IsDatabaseEncrypted


//------------------------------------------------------------------------------
// return true if database is encrypted by password or by key
//------------------------------------------------------------------------------
function TACRDatabaseData.IsDatabaseEncryptedByPassword(Session: TACRBaseSession): Boolean;
begin
  Result := True;
end; // IsDatabaseEncryptedByPassword


//------------------------------------------------------------------------------
// return true if CryptoParams are valid
//------------------------------------------------------------------------------
function TACRDatabaseData.IsCryptoParamsValid(Session: TACRBaseSession): Boolean;
begin
  Result := True;
end; // IsCryptoParamsValid


//------------------------------------------------------------------------------
// makes Exe database from edb file
//------------------------------------------------------------------------------
procedure TACRDatabaseData.MakeExeDatabase(Session: TACRBaseSession; ExeFileName, ExeDatabaseFileName: string);
begin
  ;
end; // MakeExeDatabase


//------------------------------------------------------------------------------
// removes database file from executable database file
//------------------------------------------------------------------------------
procedure TACRDatabaseData.RemoveDatabaseFromExe(Session: TACRBaseSession);
begin
;
end; // RemoveDatabaseFromExe


//------------------------------------------------------------------------------
// returns true if this file is an Accuracer database
//------------------------------------------------------------------------------
function TACRDatabaseData.IsAccuracerDatabaseFile(Session: TACRBaseSession): Boolean;
begin
  Result := False;
end; // IsAccuracerDatabaseFile


//------------------------------------------------------------------------------
// fill TableLockList with lock info for the tables opened by this session
//------------------------------------------------------------------------------
procedure TACRDatabaseData.GetTableLocksListForSession(
                                          SessionID:      TACRSessionID;
                                          TableLockList:  TList
                                      );
begin
;
end; // GetTableLocksListForSession


//------------------------------------------------------------------------------
// if result is true and PTableLockInfo is not nil - increment corresponding field of this record
//------------------------------------------------------------------------------
function TACRDatabaseData.TryToLockTable(
                    TableName:            String;
                    SessionID:            TACRSessionID;
                    LockType:             TACRLockType;
                    PTableLockInfo:       PACRTableLockInfo = nil): Boolean;
begin
 Result := True;
end; // LockTable


//------------------------------------------------------------------------------
// decrements number of locks by this session and removes all locks if their number = 0
// if PTableLockInfo = nil - all table locks byt this session will be removed
//------------------------------------------------------------------------------
function TACRDatabaseData.UnlockTableAll(
                        TableName:      String;
                        SessionID:      TACRSessionID;
                        PTableLockInfo: PACRTableLockInfo = nil): Boolean;
begin
 Result := True;
end; // UnlockTableAll


//------------------------------------------------------------------------------
// called by disconnect of server session
//------------------------------------------------------------------------------
procedure TACRDatabaseData.RemoveAllLocks(SessionID: TACRSessionID);
begin
end; // RemoveAllLocks



////////////////////////////////////////////////////////////////////////////////
//
// General functions and procedures
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// return false if field value is null
//------------------------------------------------------------------------------
function GetFieldData(
            // field number in FieldDefs
            FieldNo: Integer;
            // field definitions
            FieldDefs: TACRFieldDefs;
            // pointer to allocated buffer for storing field value
            Buffer: Pointer;
            // record buffer
            RecordBuffer: TACRRecordBuffer
            ): Boolean;
const Zero:         Word = $0000;
var Size,ZeroSize:  Integer;
begin
 Result := False;
 // check if field data is NULL
 if (CheckNullFlag(FieldNo,RecordBuffer)) then
  Exit; // field data is NULL

 Result := True;

 // function was used only for checking a null flag - no result buffer
 if (Buffer = nil) then
  Exit;

 if (FieldNo >= Integer(FieldDefs.Count)) then
   raise EACRException.Create(10000,ErrorLInvalidFieldNumber,[FieldNo,FieldDefs.Count]);

 // if blob field was specified then exit
 if (IsBLOBFieldType(FieldDefs[FieldNo].AdvancedFieldType)) then
  Exit;

 if (Buffer = nil) then
   raise EACRException.Create(10001,ErrorLNilPointer);

 if (RecordBuffer = nil) then
   raise EACRException.Create(10003,ErrorLNilPointer);

 if (IsStringFieldType(FieldDefs[FieldNo].AdvancedFieldType)) then
  begin
   if (IsWideStringFieldType(FieldDefs[FieldNo].AdvancedFieldType)) then
    begin
      PWideChar(Buffer^) := PWideChar(RecordBuffer + FieldDefs.Items[FieldNo].MemoryOffset);
//      WideString(Dest^) := WideString(PWideChar(RecordBuffer + FieldDefs.Items[FieldNo].MemoryOffset));
//      ZeroSize := 2;
    end
   else
    begin
      ZeroSize := 1;
      Size := GetStrLength(
                  PChar(RecordBuffer + FieldDefs.Items[FieldNo].MemoryOffset),
                  FieldDefs[FieldNo].AdvancedFieldType);
       try
         Move(PChar(RecordBuffer + FieldDefs.Items[FieldNo].MemoryOffset)^,Buffer^,
                Size+ZeroSize);
       except
        raise EACRException.Create(10733,
              ErrorLErrorGettingFieldData,[FieldNo,
              FieldDefs.Items[FieldNo].MemoryOffset,
              Size]);
       end;
    end;
  end // string or wide string field
 else
  try
   Move(PChar(RecordBuffer + FieldDefs.Items[FieldNo].MemoryOffset)^,Buffer^,
    FieldDefs.Items[FieldNo].MemoryDataSize);
  except
   raise EACRException.Create(10002,
          ErrorLErrorGettingFieldData,[FieldNo,
          FieldDefs.Items[FieldNo].MemoryOffset,
          FieldDefs.Items[FieldNo].MemoryDataSize]);
  end;
end; //GetFieldData


//------------------------------------------------------------------------------
// return false if field value is null
//------------------------------------------------------------------------------
procedure SetFieldData(
            // physical field No
            FieldNo: Integer;
            // field definitions
            FieldDefs: TACRFieldDefs;
            // pointer to allocated buffer with new field value
            Buffer: Pointer;
            // record buffer
            RecordBuffer: TACRRecordBuffer
            );
const Zero:         Word = $0000;
var Size,ZeroSize:  Integer;
    Source:         PChar;
begin
 if (FieldNo > Integer(FieldDefs.Count)) then
   raise EACRException.Create(10004,ErrorLInvalidFieldNumber,[FieldNo,FieldDefs.Count]);

 if (Buffer = nil) then
  begin
   // set null flag only
   SetNullFlag(True,FieldNo,RecordBuffer);
  end // set null flag only
 else
  begin
     // not empty field
     SetNullFlag(False,FieldNo,RecordBuffer);
     // copy data
     if (IsStringFieldType(FieldDefs[FieldNo].AdvancedFieldType)) then
      begin
       if (IsWideStringFieldType(FieldDefs[FieldNo].AdvancedFieldType)) then
        begin
         Source := PChar(Buffer^);
         ZeroSize := 2;
        end
       else
        begin
         ZeroSize := 1;
         Source := Buffer;
        end;
       Size := GetStrLength(
                  PChar(Source),
                  FieldDefs[FieldNo].AdvancedFieldType);
       // empty string
       if (Size <= 0) then
        SetNullFlag(True,FieldNo,RecordBuffer)
       else
        begin
         try
          if (Size + ZeroSize > FieldDefs.Items[FieldNo].MemoryDataSize) then
           begin
            Size := FieldDefs.Items[FieldNo].MemoryDataSize - ZeroSize;
            Move(Source^,PChar(RecordBuffer + FieldDefs.Items[FieldNo].MemoryOffset)^,
                  Size);
           end
          else
           begin
            Move(Source^,PChar(RecordBuffer + FieldDefs.Items[FieldNo].MemoryOffset)^,
                 Size);
           end;
            // move zero to the end of string if it is possible
          Move(Zero,
              PChar(RecordBuffer + FieldDefs.Items[FieldNo].MemoryOffset +  Size)^,
              ZeroSize);
         except
          raise EACRException.Create(10734,
                ErrorLErrorGettingFieldData,[FieldNo,
                FieldDefs.Items[FieldNo].MemoryOffset,
                Size]);
         end;
        end;
      end // string or wide string field
     else
     try
      Move(Buffer^, PChar(RecordBuffer + FieldDefs.Items[FieldNo].MemoryOffset)^,
       FieldDefs.Items[FieldNo].MemoryDataSize);
     except
      raise EACRException.Create(10732,
          ErrorLErrorGettingFieldData,[FieldNo,
          FieldDefs.Items[FieldNo].MemoryOffset,
          FieldDefs.Items[FieldNo].MemoryDataSize]);
     end;
  end; // set field data
end; //SetFieldData


//------------------------------------------------------------------------------
// return true if record buffers are binary identical
//------------------------------------------------------------------------------
function CompareRecordBuffers(RecordBuffer1,RecordBuffer2: TACRRecordBuffer;
                              RecordBufferSize: Integer): Boolean;
var i: Integer;
begin
 if (RecordBuffer1 = nil) then
  raise EACRException.Create(10122,ErrorLNilPointer);
 if (RecordBuffer2 = nil) then
  raise EACRException.Create(10123,ErrorLNilPointer);
 Result := True;
 for i := 0 to RecordBufferSize-1 do
  if (PChar(RecordBuffer1 + i)^ <> PChar(RecordBuffer2 + i)^) then
   begin
    Result := False;
    break;
   end;
end; // CompareRecordBuffers


{$IFNDEF SQLMEMTABLE}
//------------------------------------------------------------------------------
// compresses and encrypts buffer
//------------------------------------------------------------------------------
procedure CompressAndEncryptBuffer(
                        const CryptoInfo:     TACRCryptoInfo;
                        CompressionAlgorithm: Byte;
                        CompressionMode:      Byte;
                        InBuffer:             PChar;
                        InSize:               Integer;
                        out OutBuffer:        PChar;
                        out OutSize:          Integer
                                                              );
var
  ca:         TACRCompressionAlgorithm;
  Buffer:     PChar;
  BufferSize: Integer;
  CRC32:      Cardinal;
begin
  ca := TACRCompressionAlgorithm(CompressionAlgorithm);
  if ((ca = acaNone) and (CryptoInfo.CryptoAlgorithm = ACR_Cipher_None))
  or (InSize <= 0)
  or (InBuffer = nil)
  then
   begin
    OutBuffer := InBuffer;
    OutSize := InSize;
    Exit;
   end;
     if (ca <> acaNone) then
      begin
       ACRInternalCompressBuffer(ca,CompressionMode,InBuffer,InSize,Buffer,BufferSize);
{$IFDEF DEBUG_COMPRESSANDENCRYPTBUFFER}
if (BufferSize > 0) and (Buffer <> nil) then
CRC32 := ACRCountCRC(0,Buffer,BufferSize)
else
CRC32 := 0;
aaWriteToLog('COMPRESS> InSize = '+IntToStr(InSize)+#13#10+
             'COMPRESS> BufferSize = '+IntToStr(BufferSize)+#13#10+
             'COMPRESS> Buffer = '+IntToHex(Integer(Buffer),4)+#13#10+
             'COMPRESS> CRC32  of compressed buffer = '+IntToHex(CRC32,4)+#13#10);
{$ENDIF}
       try
         if (Buffer <> nil) and (BufferSize > 0) then
          begin
           if (CryptoInfo.CryptoAlgorithm = ACR_Cipher_None) then
            begin
              OutSize := BufferSize+SizeOf(BufferSize);
              OutBuffer := MemoryManager.GetMem(OutSize);
              Move(InSize,OutBuffer^,SizeOf(InSize));
              Move(Buffer^,PChar(OutBuffer+SizeOf(BufferSize))^,BufferSize);
            end // no ecnryption
           else
            begin
              OutSize := BufferSize+SizeOf(CRC32)+SizeOf(BufferSize);
              OutBuffer := MemoryManager.GetMem(OutSize);
              Move(InSize,PChar(OutBuffer+SizeOf(CRC32))^,SizeOf(InSize));
              Move(Buffer^,PChar(OutBuffer+SizeOf(CRC32)+SizeOf(BufferSize))^,BufferSize);
              CRC32 := ACRCountCRC(0,PChar(OutBuffer+SizeOf(CRC32)),OutSize-SizeOf(CRC32));
              Move(CRC32,OutBuffer^,SizeOf(CRC32));
              ACREncryptBuffer(CryptoInfo,PChar(OutBuffer+SizeOf(CRC32)),OutSize-SizeOf(CRC32));
            end; // encryption
          end;
{$IFDEF DEBUG_COMPRESSANDENCRYPTBUFFER}
if (OutSize > 0) and (OutBuffer <> nil) then
CRC32 := ACRCountCRC(0,OutBuffer,OutSize)
else
CRC32 := 0;
aaWriteToLog('COMPRESS> BufferSize = '+IntToStr(BufferSize)+#13#10+
             'COMPRESS> OutSize = '+IntToStr(OutSize)+#13#10+
             'COMPRESS> MemMgrOutBufferSize = '+IntToStr(MemoryManager.GetMemoryBufferSize(OutBuffer))+#13#10+
             'COMPRESS> OutBuffer = '+IntToHex(Integer(OutBuffer),4)+#13#10+
             'COMPRESS> CRC32 = '+IntToHex(CRC32,4)+#13#10#13#10);
{$ENDIF}
       finally
         if (Buffer <> nil) then
          FreeMem(Buffer);
       end;
      end // compression with or without encryption
     else
      begin
       OutSize := InSize+SizeOf(CRC32);
       OutBuffer := MemoryManager.GetMem(OutSize);
       CRC32 := ACRCountCRC(0,InBuffer,InSize);
       Move(CRC32,OutBuffer^,SizeOf(CRC32));
       Move(InBuffer^,PChar(OutBuffer+SizeOf(CRC32))^,InSize);
       ACREncryptBuffer(CryptoInfo,PChar(OutBuffer+SizeOf(CRC32)),InSize);
{$IFDEF DEBUG_COMPRESSANDENCRYPTBUFFER}
if (OutSize > 0) and (OutBuffer <> nil) then
CRC32 := ACRCountCRC(0,OutBuffer,OutSize)
else
CRC32 := 0;
aaWriteToLog('COMPRESS> InSize = '+IntToStr(InSize)+#13#10+
             'COMPRESS> OutSize = '+IntToStr(OutSize)+#13#10+
             'COMPRESS> MemMgrOutBufferSize = '+IntToStr(MemoryManager.GetMemoryBufferSize(OutBuffer))+#13#10+
             'COMPRESS> OutBuffer = '+IntToHex(Integer(OutBuffer),4)+#13#10+
             'COMPRESS> CRC32 = '+IntToHex(CRC32,4)+#13#10#13#10);
{$ENDIF}
      end; // encryption without compression
end; // CompressAndEncryptBuffer


//------------------------------------------------------------------------------
// decompresses and decrypts buffer; return true if successful
//------------------------------------------------------------------------------
function DecompressAndDecryptBuffer(
                        const CryptoInfo:     TACRCryptoInfo;
                        CompressionAlgorithm: Byte;
                        CompressionMode:      Byte;
                        var Buffer:           PChar;
                        var BufferSize:       Integer
                                                              ): Boolean;
var
  ca:         TACRCompressionAlgorithm;
  OutBuffer:  PChar;
  OutSize:    Integer;
  TempBuffer: PChar;
  TempSize:   Integer;
  CRC32:      Cardinal;
begin
  Result := True;
  if (BufferSize <= 0) or (Buffer = nil) then
   begin
    Buffer := nil;
    BufferSize := 0;
   end
  else
   begin
     ca := TACRCompressionAlgorithm(CompressionAlgorithm);
     if ((ca <> acaNone) or (CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None)) then
      begin
       if (ca <> acaNone) then
        begin
         if (CryptoInfo.CryptoAlgorithm = ACR_Cipher_None) then
          begin
           Move(Buffer^,OutSize,SizeOf(OutSize));
{$IFDEF DEBUG_COMPRESSANDENCRYPTBUFFER}
if (BufferSize > 0) and (Buffer <> nil) then
CRC32 := ACRCountCRC(0,Buffer,BufferSize)
else
CRC32 := 0;
aaWriteToLog('DECOMPRESS> BufferSize = '+IntToStr(BufferSize)+#13#10+
             'DECOMPRESS> OutSize = '+IntToStr(OutSize)+#13#10+
             'DECOMPRESS> MemMgrBufferSize = '+IntToStr(MemoryManager.GetMemoryBufferSize(Buffer))+#13#10+
             'DECOMPRESS> Buffer = '+IntToHex(Integer(Buffer),4)+#13#10+
             'DECOMPRESS> CRC32 = '+IntToHex(CRC32,4)+#13#10);
CRC32 := ACRCountCRC(0,PChar(Buffer+SizeOf(OutSize)),BufferSize-SizeOf(BufferSize));
aaWriteToLog('DECOMPRESS> CRC32 of compressed buffer = '+IntToHex(CRC32,4)+#13#10);
{$ENDIF}
           try
{
sz := BufferSize-SizeOf(BufferSize);
buf := AllocMem(sz);
try
  Move(PChar(Buffer+SizeOf(OutSize))^,buf^,sz);
             ACRInternalDecompressBuffer(ca,
                buf,sz,
                OutBuffer,OutSize);

finally
  FreeMem(buf);
end;
}

             ACRInternalDecompressBuffer(ca,
                PChar(Buffer+SizeOf(OutSize)),BufferSize-SizeOf(BufferSize),
                OutBuffer,OutSize);

{$IFDEF DEBUG_COMPRESSANDENCRYPTBUFFER}
aaWriteToLog('DECOMPRESS> OutSize = '+IntToStr(OutSize)+#13#10);
{$ENDIF}
           except
{$IFDEF DEBUG_COMPRESSANDENCRYPTBUFFER}
aaWriteToLog('DECOMPRESS> FAILED!!! OutSize = '+IntToStr(OutSize)+#13#10);
{$ENDIF}
             Result := False;
             OutBuffer := nil;
           end;
           try
{$IFDEF DEBUG_COMPRESSANDENCRYPTBUFFER}
aaWriteToLog('DECOMPRESS> Trying to free buffer = '+IntToHex(Integer(Buffer),4)+#13#10);
{$ENDIF}
             MemoryManager.FreeAndNilMem(Buffer);
{$IFDEF DEBUG_COMPRESSANDENCRYPTBUFFER}
aaWriteToLog('DECOMPRESS> Buffer freed'+#13#10);
{$ENDIF}
             BufferSize := 0;
             if (Result and (OutBuffer <> nil) and (OutSize > 0)) then
              begin
               BufferSize := OutSize;
               Buffer := MemoryManager.GetMem(OutSize);
               Move(OutBuffer^,Buffer^,OutSize);
              end;
           finally
{$IFDEF DEBUG_COMPRESSANDENCRYPTBUFFER}
aaWriteToLog('DECOMPRESS> Trying to free out buffer = '+IntToHex(Integer(OutBuffer),4)+#13#10);
{$ENDIF}
             if (OutBuffer <> nil) then
               FreeMem(OutBuffer);
{$IFDEF DEBUG_COMPRESSANDENCRYPTBUFFER}
aaWriteToLog('DECOMPRESS> OutBuffer freed'+#13#10);
{$ENDIF}
           end;
          end // no encryption
         else
          begin
           if (BufferSize > SizeOf(CRC32) + SizeOf(TempSize)) then
            begin
             Move(Buffer^,CRC32,SizeOf(CRC32));
             TempSize := BufferSize-SizeOf(CRC32);
             TempBuffer := MemoryManager.GetMem(TempSize);
             try
               Move(PChar(Buffer+SizeOf(CRC32))^,TempBuffer^,TempSize);
               MemoryManager.FreeAndNilMem(Buffer);
               BufferSize := 0;
               ACRDecryptBuffer(CryptoInfo,TempBuffer,TempSize);
               if (ACRCountCRC(0,TempBuffer,TempSize) <> CRC32) then
                Result := False
               else
                begin
                 Move(TempBuffer^,OutSize,SizeOf(OutSize));
                 try
                   ACRInternalDecompressBuffer(ca,
                     PChar(TempBuffer+SizeOf(OutSize)),TempSize-SizeOf(OutSize),
                     OutBuffer,OutSize);
                   try
                     if (OutBuffer <> nil) and (OutSize > 0) then
                      begin
                       BufferSize := OutSize;
                       Buffer := MemoryManager.GetMem(BufferSize);
                       Move(OutBuffer^,Buffer^,OutSize);
                      end
                     else
                      Result := False;
                   finally
                     if (OutBuffer <> nil) then
                      FreeMem(OutBuffer);
                   end;
                 except
                   Result := False;
                 end;
                end;
             finally
               MemoryManager.FreeAndNilMem(TempBuffer);
             end;
            end // buffersize is correct
           else
             Result := False;
          end; // encryption
        end // compression with or without encryption
       else
        begin
         if (BufferSize <= SizeOf(CRC32)) then
          Result := False
         else
          begin
           Move(Buffer^,CRC32,SizeOf(CRC32));
           TempSize := BufferSize - SizeOf(CRC32);
           TempBuffer := MemoryManager.GetMem(TempSize);
           try
             Move(PChar(Buffer+SizeOf(CRC32))^,TempBuffer^,TempSize);
             ACRDecryptBuffer(CryptoInfo,TempBuffer,TempSize);
             if (ACRCountCRC(0,TempBuffer,TempSize) <> CRC32) then
              begin
               Result := False;
               MemoryManager.FreeAndNilMem(TempBuffer);
              end
           except
             Result := False;
             MemoryManager.FreeAndNilMem(TempBuffer);
           end;
           if (Result) then
            begin
             MemoryManager.FreeAndNilMem(Buffer);
             Buffer := TempBuffer;
             BufferSize := TempSize;
            end;
          end; // buffersize is correct
        end; // encryption without compression
      end; // compression or encryption
   end; // source buffer is not nil
end;


//------------------------------------------------------------------------------
// compress and encrypt stream
//------------------------------------------------------------------------------
procedure CompressAndEncryptStream(
                        const CryptoInfo:     TACRCryptoInfo;
                        CompressionAlgorithm: Byte;
                        CompressionMode:      Byte;
                        BlockSize:            Integer;
                        SourceStream:         TACRStream;
                        DestStream:           TACRStream;
                        OnProgress:           TACRProgressEvent = nil
                                    );
var
  ca:         TACRCompressionAlgorithm;
  BufSize:    Integer;
  Buf:        PChar;
  OutSize:    Integer;
  OutBuf:     PChar;
  Abort:      Boolean;
  d:          Double;
  Header:     TACRBlockHeader;
begin
  Abort := False;
  ca := TACRCompressionAlgorithm(CompressionAlgorithm);
  SourceStream.BlockSize := BlockSize;
  SourceStream.OnProgress := OnProgress;
  if (ca = acaNone) and (CryptoInfo.CryptoAlgorithm = ACR_Cipher_None) then
   SourceStream.SaveToStream(DestStream)
  else
   begin
    d := 0;
    if (Assigned(OnProgress)) then
      OnProgress(SourceStream,d,Abort);
    Buf := MemoryManager.GetMem(BlockSize);
    try
      while ((not Abort) and (SourceStream.Position < SourceStream.Size)) do
       begin
        if ((SourceStream.Size - SourceStream.Position) > BlockSize) then
         BufSize := BlockSize
        else
         BufSize := SourceStream.Size - SourceStream.Position;
        LoadDataFromStream(Buf^,BufSize,SourceStream,11360);
        if (ca = acaNone) then
         begin
          ACREncryptBuffer(CryptoInfo,Buf,BufSize);
          SaveDataToStream(Buf^,BufSize,DestStream,11361);
         end
        else
         begin
          ACRInternalCompressBuffer(ca,CompressionMode,Buf,BufSize,OutBuf,OutSize);
          try
           if (CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None) then
            ACREncryptBuffer(CryptoInfo,OutBuf,OutSize);
           Header.CompressedSize := OutSize;
           Header.UncompressedSize := BufSize;
           SaveDataToStream(Header,SizeOf(Header),DestStream,11362);
           SaveDataToStream(OutBuf^,OutSize,DestStream,11363);
          finally
           FreeMem(OutBuf);
          end;
         end; // compression
        d := SourceStream.Position / SourceStream.Size * 100.0;
        if (Assigned(OnProgress)) then
          OnProgress(SourceStream,d,Abort);
       end; // coping data
    finally
      MemoryManager.FreeAndNilMem(Buf);
    end;
    d := 100;
    if (not Abort) then
     if (Assigned(OnProgress)) then
      OnProgress(SourceStream,d,Abort);
   end; // compression or encryption
end; // CompressAndEncryptStream


//------------------------------------------------------------------------------
// decompress and decrypt stream
//------------------------------------------------------------------------------
function DecompressAndDecryptStream(
                        const CryptoInfo:     TACRCryptoInfo;
                        CompressionAlgorithm: Byte;
                        CompressionMode:      Byte;
                        BlockSize:            Integer;
                        SourceStream:         TACRStream;
                        DestStream:           TACRStream;
                        OnProgress:           TACRProgressEvent = nil
                                                              ): Boolean;
var
  ca:         TACRCompressionAlgorithm;
  BufSize:    Integer;
  Buf:        PChar;
  OutSize:    Integer;
  OutBuf:     PChar;
  Abort:      Boolean;
  d:          Double;
  Header:     TACRBlockHeader;
begin
  ca := TACRCompressionAlgorithm(CompressionAlgorithm);
  DestStream.BlockSize := BlockSize;
  DestStream.OnProgress := OnProgress;
  if (ca = acaNone) and (CryptoInfo.CryptoAlgorithm = ACR_Cipher_None) then
   DestStream.LoadFromStreamWithPosition(SourceStream,SourceStream.Position,SourceStream.Size - SourceStream.Position)
  else
   begin
    d := 0;
    if (Assigned(OnProgress)) then
      OnProgress(SourceStream,d,Abort);
    if (ca = acaNone) then
     begin
      Buf := MemoryManager.GetMem(BlockSize);
      try
        while ((not Abort) and (SourceStream.Position < SourceStream.Size)) do
         begin
          if ((SourceStream.Size - SourceStream.Position) > BlockSize) then
           BufSize := BlockSize
          else
           BufSize := SourceStream.Size - SourceStream.Position;
          LoadDataFromStream(Buf^,BufSize,SourceStream,11364);
          ACRDecryptBuffer(CryptoInfo,Buf,BufSize);
          SaveDataToStream(Buf^,BufSize,DestStream,11365);
          d := SourceStream.Position / SourceStream.Size * 100.0;
          if (Assigned(OnProgress)) then
            OnProgress(SourceStream,d,Abort);
         end; // coping data
      finally
        MemoryManager.FreeAndNilMem(Buf);
      end;
     end // no compression
    else
     begin
      while ((not Abort) and (SourceStream.Position < SourceStream.Size)) do
       begin
        LoadDataFromStream(Header,SizeOf(Header),SourceStream,11366);
        BufSize := Header.CompressedSize;
        Buf := MemoryManager.GetMem(BufSize);
        try
          LoadDataFromStream(Buf^,BufSize,SourceStream,11367);
          if (CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None) then
           ACRDecryptBuffer(CryptoInfo,Buf,BufSize);
          OutSize := Header.UncompressedSize;
          ACRInternalDecompressBuffer(ca,Buf,BufSize,OutBuf,OutSize);
          try
           SaveDataToStream(OutBuf^,OutSize,DestStream,11368);
          finally
           FreeMem(OutBuf);
          end;
        finally
          MemoryManager.FreeAndNilMem(Buf);
        end;
        d := SourceStream.Position / SourceStream.Size * 100.0;
        if (Assigned(OnProgress)) then
          OnProgress(SourceStream,d,Abort);
       end; // coping data
     end; // compression
    d := 100;
    if (not Abort) then
     if (Assigned(OnProgress)) then
      OnProgress(SourceStream,d,Abort);
   end; // decompression or decryption
end; // DecompressAndDecryptStream
{$ENDIF}



initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRBaseEngine> initialized');
{$ENDIF}

end.
